/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.commons.service.client;

import java.io.IOException;

import org.apache.http.HttpHeaders;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpEntityEnclosingRequestBase;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.impl.client.CloseableHttpClient;

/**
 * Allows to formulate requests against Teamscale with an exchangeable http
 * client.
 */
public class HttpRequest<T> {

	/**
	 * Helper method for starting a service call using a given
	 * {@link CloseableHttpClient}.
	 */
	public static <T> T withHttpClient(CloseableHttpClient client, ServiceClientCallable<T> callable)
			throws ServiceCallException {
		try (CloseableHttpClient c = client) {
			return callable.call(c);
		} catch (IOException e) {
			throw new ServiceCallException("Problem closing HTTP client: " + e.getMessage(), e);
		}
	}

	/**
	 * Helper method for starting a service call with the given
	 * {@link ServerDetails}.
	 */
	public static <T> T with(ServerDetails serverDetails, ServiceClientCallable<T> callable)
			throws ServiceCallException {
		return withHttpClient(ServiceClientUtils.getHttpClient(serverDetails), callable);
	}

	/**
	 * @return a {@link ServiceClientCallable} for a HTTP GET on the given URI.
	 */
	public static <R> ServiceClientCallable<R> get(IDeserializationFormat<R> deserializationFormat, String uri) {
		return execute(new HttpGet(uri), deserializationFormat);
	}

	/**
	 * @return a {@link ServiceClientCallable} for a HTTP POST on the given URI.
	 */
	public static <S, T extends S, R> ServiceClientCallable<R> post(ISerializationFormat<S> serializationFormat, T data,
			IDeserializationFormat<R> deserializationFormat, String uri) {
		return execute(new HttpPost(uri), serializationFormat, data, deserializationFormat);
	}

	/**
	 * @return a {@link ServiceClientCallable} for a HTTP POST on the given URI.
	 */
	public static <R> ServiceClientCallable<R> post(IDeserializationFormat<R> deserializationFormat, String uri) {
		return execute(new HttpPost(uri), deserializationFormat);
	}

	/**
	 * @return a {@link ServiceClientCallable} for a HTTP PUT on the given URI.
	 */
	public static <S, T extends S, R> ServiceClientCallable<R> put(ISerializationFormat<S> serializationFormat, T data,
			IDeserializationFormat<R> deserializationFormat, String uri) {
		return execute(new HttpPut(uri), serializationFormat, data, deserializationFormat);
	}

	/**
	 * @return a {@link ServiceClientCallable} for a HTTP PUT on the given URI
	 *         without any payload.
	 * @deprecated PUT requests should always have a body
	 */
	@Deprecated
	public static <R> ServiceClientCallable<R> put(IDeserializationFormat<R> deserializationFormat, String uri) {
		return execute(new HttpPut(uri), deserializationFormat);
	}

	/**
	 * @return a {@link ServiceClientCallable} for a HTTP DELETE on the given URI.
	 */
	public static <R> ServiceClientCallable<R> delete(IDeserializationFormat<R> deserializationFormat, String uri) {
		return execute(new HttpDelete(uri), deserializationFormat);
	}

	/**
	 * Executes the request and deserializes the response using the given
	 * deserialization format.
	 */
	private static <R> ServiceClientCallable<R> execute(HttpRequestBase httpRequest,
			IDeserializationFormat<R> deserializationFormat) {
		return client -> ServiceClientUtils.executeRequest(client, httpRequest,
				deserializationFormat.getMimeType().orElse(null), deserializationFormat::deserialize);
	}

	/**
	 * Executes the request and deserializes the response using the given
	 * deserialization format.
	 */
	private static <S, T extends S, R> ServiceClientCallable<R> execute(HttpEntityEnclosingRequestBase httpRequest,
			ISerializationFormat<S> serializationFormat, T data, IDeserializationFormat<R> deserializationFormat) {
		serializationFormat.getMimeType()
				.ifPresent(mimeType -> httpRequest.setHeader(HttpHeaders.CONTENT_TYPE, mimeType.getType()));
		httpRequest.setEntity(serializationFormat.toHttpEntity(data));
		return execute(httpRequest, deserializationFormat);
	}
}
