/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.commons.service.client;

import java.io.IOException;
import java.io.Reader;
import java.util.List;
import java.util.Optional;

import org.conqat.engine.commons.util.JsonUtils;
import org.conqat.engine.core.core.ConQATException;

import com.fasterxml.jackson.databind.type.MapType;
import com.google.common.io.CharStreams;
import com.teamscale.commons.service.EMimeType;

/**
 * Provides a deserialization strategy for service calls.
 */
public interface IDeserializationFormat<T> {

	/**
	 * Returns the MIME type that should be sent to the server as "Accept" header.
	 * May return empty to indicate that the Accept header should not be set.
	 */
	Optional<EMimeType> getMimeType();

	/** Deserializes the given data to an object. */
	T deserialize(Reader serializedData) throws IOException;

	/** Constructs a deserializer for the given data type from JSON. */
	static <T> IDeserializationFormat<T> fromJson(Class<T> elementType) {
		return new IDeserializationFormat<T>() {
			@Override
			public Optional<EMimeType> getMimeType() {
				return Optional.of(EMimeType.JSON);
			}

			@Override
			public T deserialize(Reader serializedData) throws IOException {
				try {
					return JsonUtils.deserializeFromJson(serializedData, elementType);
				} catch (ConQATException e) {
					throw new IOException(e.getMessage(), e);
				}
			}
		};
	}

	/** Constructs a deserializer for the given map type from JSON. */
	static <T> IDeserializationFormat<T> fromJson(MapType mapType) {
		return new IDeserializationFormat<T>() {
			@Override
			public Optional<EMimeType> getMimeType() {
				return Optional.of(EMimeType.JSON);
			}

			@Override
			public T deserialize(Reader serializedData) throws IOException {
				try {
					return JsonUtils.deserializeFromJson(serializedData, mapType);
				} catch (ConQATException e) {
					throw new IOException(e.getMessage(), e);
				}
			}
		};
	}

	/** Constructs a deserializer for a list of the given data type from JSON. */
	static <T> IDeserializationFormat<List<T>> fromJsonList(Class<T> elementType) {
		return new IDeserializationFormat<List<T>>() {
			@Override
			public Optional<EMimeType> getMimeType() {
				return Optional.of(EMimeType.JSON);
			}

			@Override
			public List<T> deserialize(Reader serializedData) throws IOException {
				try {
					return JsonUtils.deserializeFromJson(serializedData, JsonUtils.getJavaListType(elementType));
				} catch (ConQATException e) {
					throw new IOException(e.getMessage(), e);
				}
			}
		};
	}

	/** Constructs a deserializer for plain text responses. */
	static IDeserializationFormat<String> asPlainText() {
		return new IDeserializationFormat<String>() {
			@Override
			public Optional<EMimeType> getMimeType() {
				return Optional.of(EMimeType.PLAIN);
			}

			@Override
			public String deserialize(Reader serializedData) throws IOException {
				return CharStreams.toString(serializedData);
			}
		};
	}

	/**
	 * Constructs a deserializer for a request without setting the Accept header.
	 * The response is returned as a plain string.
	 */
	static IDeserializationFormat<String> asUnspecifiedPlainText() {
		return new IDeserializationFormat<String>() {
			@Override
			public Optional<EMimeType> getMimeType() {
				return Optional.empty();
			}

			@Override
			public String deserialize(Reader serializedData) throws IOException {
				return CharStreams.toString(serializedData);
			}
		};
	}
}
