/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.client;

import java.util.Objects;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;

import com.teamscale.ide.commons.client.services.ITeamscaleConnectorService;
import com.teamscale.ide.commons.servers.ETeamscaleServerState;
import com.teamscale.ide.eclipse.plugin.EclipseServiceRegistry;
import com.teamscale.ide.eclipse.util.LegacyConversionUtils;

/**
 * A job to check the connection to a Teamscale Server.
 */
public class ConnectionValidationJob extends ServerDetailsJobBase {

	/** The callback function object to call after the job has finished. */
	private final ValidationFinishedCallback callback;

	private final ITeamscaleConnectorService connectorService;

	/**
	 * Creates a new ConnectionValidationJob.
	 * 
	 * @param title
	 *            the job's title
	 * @param serverDetails
	 *            the server's connection details
	 * @param callback
	 *            the callback function object
	 */
	public ConnectionValidationJob(String title, NamedServerDetails serverDetails,
			ValidationFinishedCallback callback) {
		super(title, serverDetails);

		this.callback = Objects.requireNonNull(callback);

		connectorService = EclipseServiceRegistry.getInstance().getService(ITeamscaleConnectorService.class);
	}

	/** {@inheritDoc} */
	@Override
	protected IStatus run(IProgressMonitor monitor) {
		boolean isReachable = true;
		String notReachableMessage = null;

		ETeamscaleServerState serverState = connectorService
				.getAndBroadcastServerState(LegacyConversionUtils.toCommonsNamedServerDetails(serverDetails));
		if (serverState != ETeamscaleServerState.REACHABLE && serverState != ETeamscaleServerState.LICENSE_INVALID) {
			isReachable = false;
			notReachableMessage = serverState.getDescription();
		}

		if (monitor.isCanceled()) {
			return Status.CANCEL_STATUS;
		}

		callback.execute(serverDetails, isReachable, notReachableMessage);
		return Status.OK_STATUS;
	}

	/** Interface for the job's callback function object. */
	public interface ValidationFinishedCallback {

		/**
		 * The callback function.
		 * 
		 * @param isReachable
		 *            the server's status, either reachable or unreachable
		 * @param errorMessage
		 *            an optional error message from the server's response
		 */
		void execute(NamedServerDetails serverDetails, boolean isReachable, String errorMessage);
	}
}
