/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.client;

import static java.util.Comparator.comparing;

import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.conqat.engine.commons.util.JsonUtils;
import org.conqat.engine.resource.util.UniformPathUtils;
import org.conqat.ide.commons.ui.logging.LoggingUtils;
import org.conqat.lib.commons.uniformpath.RelativeUniformPath;
import org.conqat.lib.commons.uniformpath.UniformPath;
import org.conqat.lib.commons.uniformpath.UniformPathCompatibilityUtil;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ProjectScope;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.QualifiedName;
import org.eclipse.core.runtime.preferences.IEclipsePreferences;
import org.osgi.service.prefs.BackingStoreException;

import com.fasterxml.jackson.databind.JavaType;
import com.google.common.base.Optional;
import com.google.common.base.Preconditions;
import com.google.common.collect.Iterables;
import com.google.common.collect.Ordering;
import com.teamscale.commons.service.XmlSerializationUtils;
import com.teamscale.ide.TeamscaleIDEActivator;
import com.teamscale.ide.commons.mappings.ILegacyResourceMapping;
import com.teamscale.ide.commons.mappings.MappingProblem;
import com.teamscale.ide.commons.mappings.ResourceMappingResult;
import com.teamscale.ide.commons.model.PrefixMapping;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.model.ResourceReferenceUtils;
import com.teamscale.ide.eclipse.EclipseIds;

/**
 * Manages a project-specific list of Teamscale prefix mappings. The prefix
 * mappings are stored in the project's property store. The prefixes are
 * serialized into XML using XStream, because the property store cannot handle
 * complex types directly.
 *
 * This is a singleton per Eclipse project. A concrete registry instance can be
 * retrieved with {@link #getInstance(IProject)}.
 */
public class PrefixMappingRegistry implements ILegacyResourceMapping {

	/** Property key for storing the enabled flag. */
	private static final String TEAMSCALE_ENABLED_KEY = "teamscaleEnabled";

	/** Property key for storing the symbolic name of the Teamscale server. */
	private static final String TEAMSCALE_SERVER_KEY = "teamscaleServer";

	/** Property key for storing the name of the Teamscale project. */
	private static final String TEAMSCALE_PROJECT_KEY = "teamscaleProject";

	/** The property key for storing the serialized prefix mappings. */
	private static final String PREFIX_MAPPINGS_KEY = "prefixMappings";

	/** The property key for storing the override branch name. */
	private static final String OVERRIDE_BRANCH_KEY = "overrideBranch";

	/** The property key for storing the Teamscale SAP sync timeout. */
	private static final String ABAP_TEAMSCALE_SYNC_TIMEOUT_KEY = "sapSystemUpdateTimeout";

	/** The Teamscale SAP sync timeout in seconds */
	private int teamscaleSapSyncTimeout = ABAP_TEAMSCALE_SYNC_TIMEOUT_DEFAULT;

	/**
	 * The default value for the
	 * {@link PrefixMappingRegistry#teamscaleSapSyncTimeout} in seconds.
	 */
	private static final int ABAP_TEAMSCALE_SYNC_TIMEOUT_DEFAULT = 60;

	/** If Teamscale should be used for the associated project at all. */
	private boolean teamscaleEnabled;

	/** The symbolic name of the Teamscale server. */
	private String teamscaleServer;

	/** The name of the Teamscale project. */
	private String teamscaleProject;

	/** The name of the override branch. */
	private String overrideBranch;

	/** The ordered list of {@link PrefixMapping}s for a project. */
	private final List<PrefixMapping> mappings = new ArrayList<>();

	/** The underlying project. */
	private final IProject project;

	/** The preferences used for storing the settings. */
	private final IEclipsePreferences preferences;

	/**
	 * Returns the PrefixMappingRegistry responsible for the given project. The
	 * given project is expected to be accessible. May return <code>null</code> if
	 * reading the registry from the session property fails.
	 */
	public static synchronized PrefixMappingRegistry getInstance(IProject project) {
		Preconditions.checkArgument(project.isAccessible(), "Project %s is not accessible", project.getName());

		QualifiedName qualifiedName = new QualifiedName(EclipseIds.PLUGIN, PREFIX_MAPPINGS_KEY);

		PrefixMappingRegistry instance = null;

		try {
			instance = (PrefixMappingRegistry) project.getSessionProperty(qualifiedName);

			if (instance == null) {
				instance = new PrefixMappingRegistry(project);
				project.setSessionProperty(qualifiedName, instance);
			}
		} catch (CoreException e) {
			LoggingUtils.error(TeamscaleIDEActivator.getDefault(), e);
		}

		return instance;
	}

	/** Constructor */
	private PrefixMappingRegistry(IProject project) {
		this.project = project;
		preferences = new ProjectScope(project).getNode(EclipseIds.PLUGIN);
		restore();
	}

	/** Restores mappings from the persisted settings. */
	public void restore() {
		teamscaleEnabled = preferences.getBoolean(TEAMSCALE_ENABLED_KEY, false);

		teamscaleServer = preferences.get(TEAMSCALE_SERVER_KEY, null);

		teamscaleProject = preferences.get(TEAMSCALE_PROJECT_KEY, null);

		overrideBranch = preferences.get(OVERRIDE_BRANCH_KEY, null);

		mappings.clear();
		mappings.addAll(loadMappings(preferences));

		teamscaleSapSyncTimeout = preferences.getInt(ABAP_TEAMSCALE_SYNC_TIMEOUT_KEY,
				ABAP_TEAMSCALE_SYNC_TIMEOUT_DEFAULT);
	}

	/** Load prefix mappings from the project-scoped preferences. */
	private static Collection<PrefixMapping> loadMappings(IEclipsePreferences preferences) {
		String mappingsXml = preferences.get(PREFIX_MAPPINGS_KEY, null);

		if (mappingsXml != null) {
			try {
				// ad-hoc migration for moved PrefixMapping class
				mappingsXml = mappingsXml.replace("com.teamscale.ide.client.PrefixMapping",
						"com.teamscale.ide.commons.model.PrefixMapping");
				JavaType detailedType = JsonUtils.getJavaListType(PrefixMapping.class);
				return XmlSerializationUtils.deserializeFromXML(mappingsXml, detailedType);
			} catch (IOException e) {
				LoggingUtils.error(TeamscaleIDEActivator.getDefault(), e);
			}
		}

		return Collections.emptyList();
	}

	@Override
	public boolean isTeamscaleEnabled() {
		return teamscaleEnabled;
	}

	/** Turn Teamscale on or off for the project. */
	public void setTeamscaleEnabled(boolean enabled) {
		teamscaleEnabled = enabled;
	}

	/**
	 * Returns symbolic Teamscale server name or <code>null</code> if none has been
	 * set.
	 */
	@Override
	public String getTeamscaleServer() {
		return teamscaleServer;
	}

	/** Sets teamscaleServer. */
	public void setTeamscaleServer(String teamscaleServer) {
		this.teamscaleServer = teamscaleServer;
	}

	/** Returns teamscaleProject . */
	@Override
	public String getTeamscaleProject() {
		return teamscaleProject;
	}

	/** Sets teamscaleProject. */
	public void setTeamscaleProject(String teamscaleProject) {
		this.teamscaleProject = teamscaleProject;
	}

	/**
	 * Returns the name of the override branch, or {@code null} if not specified.
	 */
	public String getOverrideBranch() {
		if (StringUtils.isBlank(overrideBranch)) {
			return null;
		}
		return overrideBranch;
	}

	/** Sets the override branch */
	public void setOverrideBranch(String overrideBranch) {
		this.overrideBranch = overrideBranch;
	}

	/** @see #teamscaleSapSyncTimeout */
	public int getTeamscaleSapSyncTimeout() {
		return teamscaleSapSyncTimeout;
	}

	/** @see #teamscaleSapSyncTimeout */
	public void setTeamscaleSapSyncTimeout(int teamscaleSapSyncTimeout) {
		this.teamscaleSapSyncTimeout = teamscaleSapSyncTimeout;
	}

	/**
	 * Registers a new prefix mapping for the project.
	 *
	 * @throws AssertionError
	 *             If a mapping with the same local or Teamscale prefix already
	 *             exists.
	 */
	public void addMapping(PrefixMapping mapping) {
		Preconditions.checkArgument(!localPrefixExists(mapping.getLocalPrefix()), "Duplicate local mapping: %s",
				mapping);

		mappings.add(mapping);
	}

	/** Removes all mappings from the project. */
	public void removeAll() {
		mappings.clear();
	}

	/** Removes the given prefix mapping from the project. */
	public void removeMapping(PrefixMapping mapping) {
		mappings.remove(mapping);
	}

	/**
	 * Returns all registered prefix mappings for the project, sorted by
	 * localPrefix.
	 */
	public List<PrefixMapping> getMappings() {
		return Ordering.from(comparing(PrefixMapping::getLocalPrefix)).sortedCopy(mappings);
	}

	/** Persists the project's settings into the project-specific prefs. */
	public void persist() {
		preferences.putBoolean(TEAMSCALE_ENABLED_KEY, teamscaleEnabled);
		preferences.put(TEAMSCALE_SERVER_KEY, teamscaleServer);
		preferences.put(TEAMSCALE_PROJECT_KEY, teamscaleProject);
		preferences.put(PREFIX_MAPPINGS_KEY, XmlSerializationUtils.serializeToXML(getMappings()));
		preferences.put(OVERRIDE_BRANCH_KEY, overrideBranch);
		preferences.putInt(ABAP_TEAMSCALE_SYNC_TIMEOUT_KEY, teamscaleSapSyncTimeout);

		try {
			preferences.flush();
		} catch (BackingStoreException e) {
			LoggingUtils.error(TeamscaleIDEActivator.getDefault(), e);
		}
	}

	/** Checks if a mapping for the given local prefix exists. */
	public boolean localPrefixExists(String prefix) {
		return Iterables.tryFind(mappings, mapping -> prefix.equals(mapping.getLocalPrefix())).isPresent();
	}

	/**
	 * Maps a uniform path of a Teamscale resource to a file in the workspace.
	 * Returns <code>null</code> if the path cannot be mapped.
	 *
	 * @see #localToTeamscaleResource(IResource)
	 */
	/* package */ IFile uniformPathToFile(String uniformPath) {
		IResource resource = uniformPathToResource(uniformPath);
		if (resource instanceof IFile) {
			return (IFile) resource;
		}
		return null;
	}

	/**
	 * Finds a file in this registry's project that corresponds to the uniform path.
	 * Returns <code>null</code> if the path could not be mapped.
	 */
	private IResource uniformPathToResource(String uniformPath) {
		PrefixMapping mapping = findPrefixMappingForTeamscalePath(uniformPath);
		if (mapping == null) {
			return null;
		}

		return project.findMember(makeResourcePath(uniformPath, mapping));
	}

	/**
	 * Returns the most specific mapping for the given Teamscale uniform path (i.e.,
	 * path on the Teamscale server) or <code>null</code> if no such mapping exists.
	 */
	public PrefixMapping findPrefixMappingForTeamscalePath(String uniformPath) {
		if (mappings.isEmpty()) {
			return getEmptyDefaultMapping();
		}

		String[] segments = UniformPathUtils.splitPath(uniformPath);
		for (int i = 0; i <= segments.length; i++) {
			String path = UniformPathUtils.removeLastSegments(uniformPath, i);
			Optional<PrefixMapping> match = Iterables.tryFind(mappings,
					mapping -> path.equals(mapping.getTeamscalePrefix()));
			if (match.isPresent()) {
				return match.get();
			}
		}

		return null;
	}

	/**
	 * Returns the most specific mapping for the given local path (the given path
	 * starts with the localPathPrefix of the returned PrefixMapping). Returns
	 * <code>null</code> if no match could be found.
	 *
	 * This always chooses the most specific mapping, e.g. local prefix "src/module"
	 * is chosen over local prefixes "src" or just "".
	 */
	private PrefixMapping findPrefixMappingForLocalResource(RelativeUniformPath projectRelativePath) {
		if (mappings.isEmpty()) {
			return null;
		}
		List<String> segments = projectRelativePath.getSegments();
		for (int i = 0; i <= segments.size(); i++) {
			String path = StringUtils.join(segments.subList(0, segments.size() - i), "/");
			Optional<PrefixMapping> match = Iterables.tryFind(mappings,
					mapping -> path.equals(mapping.getLocalPrefix()));
			if (match.isPresent()) {
				return match.get();
			}
		}
		return null;
	}

	/**
	 * Maps a project relative path (local, in IDE) below the project root, if
	 * possible.
	 */
	@Override
	public ResourceMappingResult map(URI serverUrl, ResourceReference projectRoot,
			RelativeUniformPath projectRelativePath) {
		Preconditions.checkArgument(teamscaleProject != null);

		PrefixMapping applicableMapping = findPrefixMappingForLocalResource(projectRelativePath);
		if (applicableMapping == null) {
			return ResourceMappingResult.failure(MappingProblem.noMappingFound(projectRoot));
		}

		ResourceReference mappingRoot = ResourceReferenceUtils.resolveFolder(projectRoot,
				UniformPathCompatibilityUtil.convertRelative(applicableMapping.getLocalPrefix()));
		UniformPath mappingRootPath = UniformPathCompatibilityUtil.convert(applicableMapping.getTeamscalePrefix());

		return ResourceMappingResult.success(serverUrl, teamscaleProject, getOverrideBranch(), mappingRoot,
				mappingRootPath);
	}

	/**
	 * Transforms a Teamscale uniform path to a local workspace path using a
	 * {@link PrefixMapping}.
	 */
	private static IPath makeResourcePath(String uniformPath, PrefixMapping mapping) {
		String relativeCommonPath = uniformPath;
		if (uniformPath.startsWith(mapping.getTeamscalePrefix())) {
			relativeCommonPath = uniformPath.substring(mapping.getTeamscalePrefix().length());
		}
		return new Path(mapping.getLocalPrefix()).append(relativeCommonPath);
	}

	/**
	 * Returns a mapping <code>"" &lt;-&gt; ""</code> to be used if Teamscale is
	 * enabled but no mappings are setup at all.
	 */
	private static PrefixMapping getEmptyDefaultMapping() {
		return PrefixMapping.EMPTY_PREFIX_MAPPING;
	}
}
