/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.client;

import static java.util.Objects.requireNonNull;

import org.conqat.ide.commons.ui.logging.LoggingUtils;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.ISchedulingRule;
import org.eclipse.core.runtime.jobs.Job;

import com.google.common.base.Preconditions;
import com.teamscale.ide.TeamscaleIDEActivator;
import com.teamscale.ide.eclipse.EclipseIds;

/**
 * A base job to get information from the Teamscale server regarding an
 * {@link IResource}.
 */
public abstract class ResourceJobBase extends Job {

	/**
	 * The workspace resource for which this job loads information from the
	 * Teamscale server.
	 */
	protected final IResource resource;

	/** Rule used to stop two jobs from running at the same time */
	private static final ISchedulingRule RULE = new ISchedulingRule() {

		@Override
		public boolean isConflicting(ISchedulingRule rule) {
			return rule == ResourceJobBase.RULE;
		}

		@Override
		public boolean contains(ISchedulingRule rule) {
			return rule == ResourceJobBase.RULE;
		}
	};

	/**
	 * Create a job for the given resource with an optional jump afterwards.
	 */
	protected ResourceJobBase(String message, IResource resource) {
		super(message);
		this.resource = requireNonNull(resource);
		setPriority(DECORATE);
		setRule(RULE);
	}

	/**
	 * Cancels previously started jobs and schedules a new one. Does nothing if a
	 * job is already scheduled (running, waiting, or sleeping) for the given
	 * resource.
	 */
	protected static void schedule(ResourceJobBase job) {
		if (isJobScheduledForResource(job)) {
			return;
		}
		job.schedule();
	}

	/**
	 * Checks whether a job is already scheduled for the given resource.
	 */
	private static boolean isJobScheduledForResource(ResourceJobBase jobToSchedule) {
		Job[] scheduledJobs = Job.getJobManager().find(jobToSchedule.getClass());
		for (Job job : scheduledJobs) {
			Preconditions.checkState(job instanceof ResourceJobBase);
			if (((ResourceJobBase) job).resource.equals(jobToSchedule.resource)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Returns <code>true</code> if Teamscale is disabled for the project of this
	 * resource.
	 */
	private boolean isTeamscaleDisabled() {
		return !PrefixMappingRegistry.getInstance(resource.getProject()).isTeamscaleEnabled();
	}

	/** {@inheritDoc} */
	@Override
	protected IStatus run(IProgressMonitor monitor) {
		if (!resource.isAccessible() || isTeamscaleDisabled()) {
			return Status.OK_STATUS;
		}

		try {
			doJob();
		} catch (CoreException e) {
			LoggingUtils.error(TeamscaleIDEActivator.getDefault(), e);
			return new Status(IStatus.ERROR, EclipseIds.PLUGIN, e.getMessage());
		}
		return Status.OK_STATUS;
	}

	/** {@inheritDoc} */
	@Override
	public boolean belongsTo(Object family) {
		return this.getClass().equals(family);
	}

	/** Runs the resource job. */
	protected abstract void doJob() throws CoreException;

}
