/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.client;

import java.util.Map;
import java.util.stream.Collectors;

import org.checkerframework.checker.nullness.qual.NonNull;

import com.fasterxml.jackson.annotation.JsonProperty;

/**
 * Represents the name of a server. Used as keys in maps containing servers. We
 * use this as a String wrapper to avoid serializing maps with string keys.
 * Using string keys causes issues with Jackson XML, as they are serialized as
 * XML tags (instead of values), which breaks for certain characters in the
 * string (most often whitespace). <br/>
 * This is intended to be used only for serialization and the inner Strings
 * should be retrieved ASAP after deserialization. Use
 * {@link #withServerNameKeys(Map)} and {@link #withStringKeys(Map)} to convert
 * maps.
 *
 * @see <a href="https://cqse.atlassian.net/browse/TS-28686">TS-28686</a>
 */
public class ServerNameKey {
	/** Wrapped server name. */
	private final String serverName;

	/** Creates a new instance. Also used for deserialization. */
	public ServerNameKey(@JsonProperty("serverName") String serverName) {
		this.serverName = serverName;
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		return serverName;
	}

	/** {@inheritDoc} */
	@Override
	public boolean equals(Object obj) {
		// Allows direct comparison with Strings.
		if (obj instanceof String) {
			return serverName.equals(obj);
		}
		if (!(obj instanceof ServerNameKey)) {
			return false;
		}
		return serverName.equals(((ServerNameKey) obj).serverName);
	}

	/** {@inheritDoc} */
	@Override
	public int hashCode() {
		return serverName.hashCode();
	}

	/**
	 * Utility for converting keys of a deserialized map back to string keys.
	 */
	public static Map<String, NamedServerDetails> withStringKeys(Map<ServerNameKey, NamedServerDetails> map) {
		return map.entrySet().stream()
				.collect(Collectors.toMap(e -> e.getKey().toString(), java.util.Map.Entry::getValue));
	}

	/**
	 * Utility for converting string keys of a map to this type to prepare for
	 * serialization.
	 */
	public static Map<@NonNull ServerNameKey, NamedServerDetails> withServerNameKeys(
			Map<String, NamedServerDetails> map) {
		return map.entrySet().stream()
				.collect(Collectors.toMap(e -> new ServerNameKey(e.getKey()), java.util.Map.Entry::getValue));
	}
}
