/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.client;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.conqat.ide.commons.ui.logging.LoggingUtils;
import org.eclipse.equinox.security.storage.ISecurePreferences;
import org.eclipse.equinox.security.storage.SecurePreferencesFactory;
import org.eclipse.equinox.security.storage.StorageException;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.net.UrlEscapers;
import com.teamscale.commons.service.client.ServerDetails;
import com.teamscale.ide.TeamscaleIDEActivator;
import com.teamscale.ide.eclipse.EclipseIds;

/**
 * Manages a list of available Teamscale servers. The server list is stored
 * securely in the encrypted Eclipse SecureStorage container. The server list is
 * serialized into XML using XStream because the preferences cannot store
 * complex types directly.
 * 
 * This is a singleton.
 */
public class TeamscaleConnectionRegistry {

	private static final Logger LOGGER = LogManager.getLogger();

	private static final String URL = "url";
	private static final String USERNAME = "username";
	@VisibleForTesting
	static final String PASSWORD = "password";
	private static final String ENABLED = "enabled";
	private static final String TRUST_ALL_CERTIFICATES = "trustAllCertificates";

	/**
	 * Preferences key under {@link EclipseIds#PLUGIN} that stores all server
	 * details in the preferred hierarchical format.
	 */
	@VisibleForTesting
	static final String SERVERS_PREFERENCES_KEY = "teamscaleServers";

	/** The secure preferences used to save the connection configurations. */
	private final ISecurePreferences preferences;

	/** Singleton instance. */
	private static TeamscaleConnectionRegistry instance;

	/**
	 * Holds details for a server by name. This is used to cache access to the
	 * secure storage.
	 */
	private Map<String, NamedServerDetails> teamscaleConnections;

	/**
	 * Servers which were deleted or modified temporarily on the UI and which should
	 * be removed from the preferences when the changes are be permanently
	 * persisted.
	 */
	private List<String> serversMarkedForDeletion;

	/** Returns the singleton instance. */
	public static synchronized TeamscaleConnectionRegistry getInstance() {
		if (instance == null) {
			instance = new TeamscaleConnectionRegistry();
		}
		return instance;
	}

	/** Constructor. */
	private TeamscaleConnectionRegistry() {
		preferences = SecurePreferencesFactory.getDefault().node(EclipseIds.PLUGIN);
		teamscaleConnections = restoreServers(preferences);
		serversMarkedForDeletion = new ArrayList<>();
	}

	/**
	 * Tries to read Map of {@link ServerDetails} from the preference store.
	 * 
	 * @return Map of named {@link ServerDetails}, which may be empty on error.
	 */
	private static Map<String, NamedServerDetails> restoreServers(ISecurePreferences preferences) {
		Map<String, NamedServerDetails> servers = doRetrieveServers(preferences);
		TeamscaleServerStatus.getInstance().initializeStatus(servers.values());
		return servers;
	}

	@VisibleForTesting
	static Map<String, NamedServerDetails> doRetrieveServers(ISecurePreferences preferencesRoot) {
		Map<String, NamedServerDetails> servers = new HashMap<>();
		ISecurePreferences serversPreferences = preferencesRoot.node(SERVERS_PREFERENCES_KEY);
		for (String escapedServerName : serversPreferences.childrenNames()) {
			ISecurePreferences childServerNode = serversPreferences.node(escapedServerName);

			String unescapedServerName = unescapePreferenceNode(escapedServerName);
			try {
				String url = childServerNode.get(URL, "");
				String username = childServerNode.get(USERNAME, "");
				String password = childServerNode.get(PASSWORD, "");
				boolean enabled = childServerNode.getBoolean(ENABLED, true);
				boolean trustAllCertificates = childServerNode.getBoolean(TRUST_ALL_CERTIFICATES, false);
				servers.put(unescapedServerName, new NamedServerDetails(unescapedServerName, url, username, password,
						enabled, trustAllCertificates));
			} catch (StorageException e) {
				LoggingUtils.error(TeamscaleIDEActivator.getDefault(), "Failed to read server: " + escapedServerName,
						e);
				LOGGER.catching(e);
			}
		}

		return servers;
	}

	/**
	 * Escapes special characters in a server name so it can be saved as an Eclipse
	 * Preference node
	 */
	@VisibleForTesting
	static String escapePreferenceNode(String serverName) {
		return UrlEscapers.urlFormParameterEscaper().escape(serverName);
	}

	/**
	 * Un-escapes a previously escaped server name to its user-readable form.
	 */
	@VisibleForTesting
	static String unescapePreferenceNode(String escapedServerName) {
		try {
			return URLDecoder.decode(escapedServerName, StandardCharsets.UTF_8.name());
		} catch (UnsupportedEncodingException e) {
			throw new AssertionError(e);
		}
	}

	/**
	 * Saves server connections defined in the map into a hierarchical model rooted
	 * at {@link EclipseIds#PLUGIN}.
	 */
	@VisibleForTesting
	static void doPersistServers(ISecurePreferences preferencesRoot, Map<String, NamedServerDetails> servers,
			List<String> serversMarkedForDeletion) throws StorageException {
		ISecurePreferences serversPreferencesRoot = preferencesRoot.node(SERVERS_PREFERENCES_KEY);
		for (String serverName : servers.keySet()) {
			ISecurePreferences childServerNode = serversPreferencesRoot.node(escapePreferenceNode(serverName));
			NamedServerDetails serverDetail = servers.get(serverName);

			childServerNode.put(URL, serverDetail.getUrl(), false);
			childServerNode.putBoolean(ENABLED, serverDetail.isEnabled(), false);
			childServerNode.putBoolean(TRUST_ALL_CERTIFICATES, serverDetail.isTrustAllCertificates(), false);
			childServerNode.put(USERNAME, serverDetail.getUsername(), false);
			childServerNode.put(PASSWORD, serverDetail.getPassword(), true);
		}

		serversMarkedForDeletion
				.forEach(serverName -> serversPreferencesRoot.node(escapePreferenceNode(serverName)).removeNode());
	}

	/** Returns all registered servers. */
	public ImmutableList<NamedServerDetails> getServers() {
		return ImmutableList.copyOf(teamscaleConnections.values());
	}

	/** Get all servers, mapped with their symbolic names. */
	public ImmutableMap<String, NamedServerDetails> getAllServers() {
		return ImmutableMap.copyOf(teamscaleConnections);
	}

	/**
	 * Get the symbolic server names of enabled servers in no specific order.
	 */
	public Collection<String> getEnabledServerNames() {
		Collection<String> enabledServers = new ArrayList<>();
		for (String key : teamscaleConnections.keySet()) {
			NamedServerDetails connection = teamscaleConnections.get(key);
			if (connection.isEnabled()) {
				enabledServers.add(key);
			}
		}
		return enabledServers;
	}

	/**
	 * Returns whether the given symbolic name belongs to a registered server.
	 */
	public boolean containsServer(String symbolicName) {
		return teamscaleConnections.containsKey(symbolicName);
	}

	/** Returns whether there are any registered servers. */
	public boolean hasServers() {
		return !teamscaleConnections.isEmpty();
	}

	/**
	 * Returns the details of a server or <code>null</code> if no such server
	 * exists.
	 */
	public NamedServerDetails getServer(String symbolicName) {
		return teamscaleConnections.get(symbolicName);
	}

	/**
	 * Adds a server to the connections and persists it. If a server with the same
	 * name already exists, it will be overwritten.
	 */
	public void putServer(String serverName, NamedServerDetails serverDetails) {
		putServerTemporarily(serverName, serverDetails);
		persistServers();
	}

	/** Puts a server to the connections. */
	public void putServerTemporarily(String serverName, NamedServerDetails serverDetails) {
		teamscaleConnections.put(serverName, serverDetails);
		TeamscaleServerStatus.getInstance().initializeStatus(serverDetails);
	}

	/**
	 * Removes the server with the given symbolic name and persists the changes.
	 */
	public void removeServer(String symbolicName) {
		removeServerTemporarily(symbolicName);
		persistServers();
	}

	/** Remove a server from the connections. */
	public void removeServerTemporarily(String symbolicName) {
		if (teamscaleConnections.remove(symbolicName) != null) {
			TeamscaleServerStatus.getInstance().removeStatus(symbolicName);
			serversMarkedForDeletion.add(symbolicName);
		}
	}

	/** Removes all servers temporarily. */
	public void removeAllServersTemporarily() {
		for (String symbolicName : teamscaleConnections.keySet()) {
			TeamscaleServerStatus.getInstance().removeStatus(symbolicName);
			serversMarkedForDeletion.add(symbolicName);
		}
		teamscaleConnections.clear();
	}

	/** Restore the connections from the storage. */
	public void restoreServers() {
		teamscaleConnections = restoreServers(preferences);
	}

	/** Writes the server list to the preferences store. */
	public void persistServers() {
		try {
			doPersistServers(preferences, teamscaleConnections, serversMarkedForDeletion);
			serversMarkedForDeletion.clear();
			preferences.flush();
		} catch (StorageException | IOException e) {
			LoggingUtils.error(TeamscaleIDEActivator.getDefault(), e);
		}
	}
}
