/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.client;

import static java.util.Objects.requireNonNull;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;

import com.teamscale.ide.findings.view.FindingsMarkerView;

/**
 * A singleton registry for storing server status.
 */
public class TeamscaleServerStatus {
	/**
	 * The maximum number of allowed server connection failures. If reached, the
	 * server will be disabled.
	 */
	private static final int MAXIMUM_FAILURE_COUNT = 6;

	/**
	 * The duration after a temporarily disabled server is enabled again in
	 * milliseconds.
	 */
	private static final long TEMPORARILY_DISABLED_MILLIS = 60000;

	/** The singleton instance. */
	private static TeamscaleServerStatus instance;

	/** Mapping from server names to their status. */
	private final Map<String, StatusEntry> serverStatusMap = new ConcurrentHashMap<>();

	/**
	 * List of registered status change listeners. CopyOnWriteArrayList is
	 * thread-safe, write is expensive.
	 */
	private final List<ITeamscaleServerStatusChangeListener> statusChangeListeners = new CopyOnWriteArrayList<>();

	/**
	 * The callback to be called after finishing a status initialization job.
	 */
	private final InitialConnectionCheckCallback connectionValidationCallback = new InitialConnectionCheckCallback();

	/** Returns the singleton instance. */
	public static synchronized TeamscaleServerStatus getInstance() {
		if (instance == null) {
			instance = new TeamscaleServerStatus();
		}
		return instance;
	}

	/** Private constructor to avoid initialization. */
	private TeamscaleServerStatus() {
		// empty
	}

	/** Initializes the status for the given servers. */
	public void initializeStatus(Collection<NamedServerDetails> serverDetails) {
		for (NamedServerDetails server : serverDetails) {
			initializeStatus(server);
		}
	}

	/** Initializes the status for the given server. */
	public void initializeStatus(NamedServerDetails serverDetails) {
		String symbolicName = serverDetails.getSymbolicName();
		if (serverDetails.isEnabled()) {
			setStatus(symbolicName, ETeamscaleServerStatus.UNREACHABLE);
			ConnectionValidationJob job = new ConnectionValidationJob(
					"Checking connection to Teamscale Server: " + symbolicName, serverDetails,
					connectionValidationCallback);
			job.schedule();
		} else {
			setStatus(symbolicName, ETeamscaleServerStatus.DISABLED);
		}
	}

	/** Sets the status for the given server. */
	public void setStatus(String symbolicName, ETeamscaleServerStatus status) {
		requireNonNull(status);

		StatusEntry oldEntry = serverStatusMap.get(symbolicName);
		if (oldEntry == null) {
			createNewEntry(symbolicName, status);
		} else {
			updateExistingEntry(symbolicName, status, oldEntry);
		}
	}

	/** Create a new status entry. */
	private void createNewEntry(String symbolicName, ETeamscaleServerStatus status) {
		StatusEntry newEntry = new StatusEntry(status);
		serverStatusMap.put(symbolicName, newEntry);
		fireStatusChanged();
		if (status == ETeamscaleServerStatus.UNREACHABLE) {
			updateFailureCount(symbolicName, newEntry);
		}
	}

	/** Updates the given status entry. */
	private void updateExistingEntry(String symbolicName, ETeamscaleServerStatus status, StatusEntry entry) {
		boolean changed = false;
		if (entry.getStatus() != status) {
			entry.setStatus(status);
			changed = true;
		}
		if (changed) {
			fireStatusChanged();
		}
		if (status == ETeamscaleServerStatus.UNREACHABLE) {
			updateFailureCount(symbolicName, entry);
		}
	}

	/** Updates the failure count of a server. */
	private void updateFailureCount(final String symbolicName, StatusEntry entry) {
		if (entry.incrementFailureCount()) {
			entry.setStatus(ETeamscaleServerStatus.TEMPORARILY_DISABLED);
			FindingsMarkerView.displayTemporarilyDisabledMessage(symbolicName);
			fireStatusChanged();
		}
	}

	/** Removes the status entry for the given server. */
	public void removeStatus(String symbolicName) {
		if (serverStatusMap.remove(symbolicName) != null) {
			fireStatusChanged();
		}
	}

	/** Returns the status for the given server or null if no status is set. */
	public Optional<ETeamscaleServerStatus> getStatus(String symbolicName) {
		StatusEntry entry = serverStatusMap.get(symbolicName);
		if (entry == null) {
			return Optional.empty();
		}
		return Optional.of(entry.getStatus());
	}

	/** Returns a globally aggregated status. */
	public ETeamscaleGlobalServerStatus getGlobalStatus() {
		if (serverStatusMap.isEmpty() || allHaveStatus(ETeamscaleServerStatus.DISABLED)) {
			return ETeamscaleGlobalServerStatus.DISABLED;
		}

		if (allEnabledHaveStatus(ETeamscaleServerStatus.REACHABLE)) {
			return ETeamscaleGlobalServerStatus.REACHABLE;
		} else if (allEnabledHaveStatus(ETeamscaleServerStatus.UNREACHABLE)) {
			return ETeamscaleGlobalServerStatus.UNREACHABLE;
		}

		return ETeamscaleGlobalServerStatus.MIXED;
	}

	/** Returns whether all servers have the given status. */
	private boolean allHaveStatus(ETeamscaleServerStatus status) {
		for (StatusEntry currentEntry : serverStatusMap.values()) {
			if (currentEntry.getStatus() != status) {
				return false;
			}
		}
		return true;
	}

	/** Returns whether all enabled servers have the given status. */
	private boolean allEnabledHaveStatus(ETeamscaleServerStatus status) {
		for (StatusEntry currentEntry : serverStatusMap.values()) {
			if (currentEntry.getStatus() != ETeamscaleServerStatus.DISABLED && currentEntry.getStatus() != status) {
				return false;
			}
		}
		return true;
	}

	/** Registers the given status change listener. */
	public void addStatusChangeListener(ITeamscaleServerStatusChangeListener listener) {
		requireNonNull(listener);
		statusChangeListeners.add(listener);
	}

	/** Removes the given status change listener. */
	public void removeStatusChangeListener(ITeamscaleServerStatusChangeListener listener) {
		statusChangeListeners.remove(listener);
	}

	/** Notifies all registered listeners that a server's status has changed. */
	private void fireStatusChanged() {
		for (ITeamscaleServerStatusChangeListener listener : statusChangeListeners) {
			listener.statusChanged();
		}
	}

	/**
	 * The connection validation job that is scheduled for initializing a new
	 * server's status.
	 */
	private class InitialConnectionCheckCallback implements ConnectionValidationJob.ValidationFinishedCallback {
		/** {@inheritDoc} */
		@Override
		public void execute(NamedServerDetails serverDetails, boolean isReachable, String errorMessage) {
			String symbolicName = serverDetails.getSymbolicName();
			if (isReachable) {
				setStatus(symbolicName, ETeamscaleServerStatus.REACHABLE);
			} else {
				setStatus(symbolicName, ETeamscaleServerStatus.UNREACHABLE);
			}
		}
	}

	/** An entry inside the server status table. */
	private static class StatusEntry {

		/** The server's status. */
		private ETeamscaleServerStatus status;

		/** How many times connecting to the server failed. */
		private int failureCount;

		/**
		 * The point of time, when the server has been temporarily disabled in
		 * milliseconds since unix epoch. -1 if the server is not temporarily disabled.
		 */
		private long disabledTime;

		/** Constructor. */
		public StatusEntry(ETeamscaleServerStatus status) {
			this.status = status;
			this.failureCount = 0;
			this.disabledTime = -1;
		}

		/** Sets status. */
		public void setStatus(ETeamscaleServerStatus status) {
			this.status = status;
			if (status == ETeamscaleServerStatus.TEMPORARILY_DISABLED) {
				this.disabledTime = System.currentTimeMillis();
			} else {
				this.disabledTime = -1;
			}
		}

		/** Returns status. */
		public ETeamscaleServerStatus getStatus() {
			return status;
		}

		/**
		 * Returns whether the server is temporarily disabled. If it was disabled longer
		 * than {@link TeamscaleServerStatus#TEMPORARILY_DISABLED_MILLIS}, false is
		 * returned and the status is set to unreachable, as we don't know yet, whether
		 * the server is reachable again. The next request to the server will then
		 * determine, whether the server is reachable and set and appropriate status.
		 */
		public boolean isTemporarilyDisabled() {
			if (disabledTime == -1) {
				return false;
			}

			long duration = System.currentTimeMillis() - this.disabledTime;
			if (duration >= TEMPORARILY_DISABLED_MILLIS) {
				status = ETeamscaleServerStatus.UNREACHABLE;
				disabledTime = -1;
				return false;
			}
			return true;
		}

		/**
		 * Increments the failure count. If
		 * {@link TeamscaleServerStatus#MAXIMUM_FAILURE_COUNT} is reached, it will be
		 * reset to zero.
		 * 
		 * @return true, if the failure count has
		 *         {@link TeamscaleServerStatus#MAXIMUM_FAILURE_COUNT}
		 */
		public boolean incrementFailureCount() {
			failureCount++;
			if (failureCount == MAXIMUM_FAILURE_COUNT) {
				failureCount = 0;
				return true;
			}
			return false;
		}

	}
}
