package com.teamscale.ide.client.adt;

import java.net.URI;
import java.util.HashMap;
import java.util.Optional;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.checkerframework.checker.nullness.qual.Nullable;
import org.eclipse.core.resources.IProject;

import com.google.common.annotations.VisibleForTesting;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.model.ProjectMapping;
import com.teamscale.ide.commons.plugin.services.IResourceResolutionService;
import com.teamscale.ide.eclipse.plugin.EclipseServiceRegistry;

/**
 * Maps a URI for fugr modules to a Teamscale URI.
 */
class AdtFugrModuleUriMapper implements IAdtUriMapper {

	private static final Logger LOGGER = LogManager.getLogger();

	private static final String NAME_GROUP = "name";

	private static final String FUGR_NAME_GROUP = "fugrname";

	private static final Pattern PATTERN = Pattern
			.compile("/sap/bc/adt/functions/groups/(?<fugrname>.+)/fmodules/(?<name>.+)/source/main");

	private static final String TEAMSCALE_URI_TEMPLATE = "FUGR/%s/%s.abap";

	/** Pattern to match an include statement with corresponding comment. */
	private final static Pattern INCLUDE_WITH_COMMENT_PATTERN = Pattern.compile("(?m)^\\s*INCLUDE\\s+" + // INCLUDE
			"([/a-zA-Z0-9_#$@!?äüöÄÜÖ*%]([/a-zA-Z0-9_#$@!?äüöÄÜÖ*%|-]*[/a-zA-Z0-9_#$@!?äüöÄÜÖ*%|])?)" + // Identifier
			"\\s*.\\s+\"(.*)"); // Comment at the end of line

	/**
	 * Gets the mapping for a function module within a function group. To get this
	 * mapping it is necessary to get the Teamscale name of the function module.
	 * This name is in the object "L<FunctionGroupName>UXX" as comment. Therefore,
	 * the following steps are necessary:
	 * 
	 * 1. Get the content of the "L<FunctionGroupName>UXX" object.
	 * 
	 * 2. Parse the comment to get the name.
	 * 
	 * 3. Get the mapping with the name of the function module.
	 */
	@Override
	public Optional<String> map(String uri, IProject project) {
		LOGGER.traceEntry(null, uri, project);
		Matcher matcher = PATTERN.matcher(uri);
		if (!matcher.matches()) {
			return LOGGER.traceExit(Optional.empty());
		}
		String functionGroupName = matcher.group(FUGR_NAME_GROUP);
		String fugrUxxName = AdtUriMappingUtils.getFunctionGroupNameWithPrefixInAdt(functionGroupName, "l") + "uxx";

		ResourceReference fugrUxxResource = ResourceReference
				.forFile(URI.create(String.format("semanticfs:/%s/.adt/functions/groups/%s/includes/%s/%s.asfinc",
						project.getName(), functionGroupName.replace("/", "%252f"), fugrUxxName, fugrUxxName)));

		LOGGER.debug("Mapping Fugr Module {} with fugrUxx {} at {}", functionGroupName, fugrUxxName, fugrUxxResource);

		String fileContent = EclipseServiceRegistry.getInstance().getService(IResourceResolutionService.class)
				.readContents(fugrUxxResource).orElse(null);

		if (fileContent == null) {
			LOGGER.debug("Found no file content");
			return LOGGER.traceExit(Optional.empty());
		}
		LOGGER.debug("Found file content of length {}", fileContent.length());

		String teamscaleIncludeName = getTeamscaleIncludeName(fileContent, matcher.group(NAME_GROUP).toUpperCase());

		if (teamscaleIncludeName == null) {
			LOGGER.debug("Was not able to resolve Teamscale Include Name");
			return Optional.empty();
		}
		LOGGER.debug("The resolved Teamscale Include Name is {}", teamscaleIncludeName);

		return LOGGER.traceExit(Optional.of(String.format(TEAMSCALE_URI_TEMPLATE,
				AdtUriMappingUtils.convertFileNameToAbapName(functionGroupName), teamscaleIncludeName)));
	}

	/**
	 * Parses the file content of a UXX file and searches for the commented include
	 * name, which is used by Teamscale.
	 */
	@VisibleForTesting
	@Nullable
	String getTeamscaleIncludeName(String content, String includeName) {
		Matcher matcher = INCLUDE_WITH_COMMENT_PATTERN.matcher(content);
		HashMap<String, String> mapAbapNameToTeamscaleName = new HashMap<>();

		while (matcher.find()) {
			String abapIncludeName = matcher.group(3).replace("\"", "").trim();
			String teamscaleIncludeName = matcher.group(1).toUpperCase();
			mapAbapNameToTeamscaleName.put(abapIncludeName, teamscaleIncludeName);
		}

		if (!mapAbapNameToTeamscaleName.containsKey(includeName)) {
			return null;
		}

		return AdtUriMappingUtils.convertFileNameToAbapName(mapAbapNameToTeamscaleName.get(includeName));
	}
}
