/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.client.adt;

import java.util.Optional;

import org.eclipse.core.resources.IProject;

/**
 * Functions to map the Eclipse-ADT-resource to a Teamscale resource.
 */
public class AdtUriMapping {

	/**
	 * We are currently missing {@link IAdtUriMapper} for the following types:
	 * <dl>
	 * 	<dt><dd> FMOD Not used in the path of the file. The path contains FUGR as type.
	 * 	<dt><dd> SUSO Only opened in the SAP GUI. No need to support it.
	 * 	<dt><dd> TRAN Only opened in the SAP GUI. No need to support it.
	 * 	<dt><dd> VIEW Only opened in the SAP GUI. No need to support it.
	 * </dl>
	 */
	private static final IAdtUriMapper[] ADT_URI_MAPPER = new IAdtUriMapper[] { //
			new SimpleAdtUriMapper(
					"/sap/bc/adt/oo/classes/(?<name>[^/]+|/[^/]+/[^/]+|.+?=(/source/main))(/source/main)?$",
					"CLAS/%s.abap"),
			new SimpleAdtUriMapper("/sap/bc/adt/programs/programs/(?<name>.+)/source/main", "PROG/%s.abap"),
			new SimpleAdtUriMapper("/sap/bc/adt/programs/includes/(?<name>.+)/source/main", "INTF/%s.abap"),
			new SimpleAdtUriMapper("/sap/bc/adt/oo/interfaces/(?<name>.+)/source/main", "INTF/%s.abap"),
			new SimpleAdtUriMapper("/sap/bc/adt/bo/behaviordefinitions/(?<name>.+)/source/main", "BDEF/%s.abap_cds"),
			new SimpleAdtUriMapper("/sap/bc/adt/acm/dcl/sources/(?<name>.+)/source/main", "DCLS/%s.abap_cds"),
			new SimpleAdtUriMapper("/sap/bc/adt/ddic/srvd/sources/(?<name>.+)/source/main", "SRVD/%s.abap_cds"),
			new SimpleAdtUriMapper("/sap/bc/adt/ddic/ddl/sources/(?<name>.+)/source/main", "DDLS/%s.abap_cds"),
			new SimpleAdtUriMapper("/sap/bc/adt/ddic/ddlx/sources/(?<name>.+)/source/main", "DDLX/%s.abap_cds"),
			new SimpleAdtUriMapper("/sap/bc/adt/ddic/tables/(?<name>.+)/source/main", "TABL/%s.abap_ddic"),
			new SimpleAdtUriMapper("/sap/bc/adt/ddic/structures/(?<name>.+)/source/main", "STRU/%s.abap_ddic"),
			new SimpleAdtUriMapper("/sap/bc/adt/ddic/tabletypes/(?<name>.+)", "TTYP/%s.abap_ddic"),
			new SimpleAdtUriMapper("/sap/bc/adt/ddic/domains/(?<name>.+)", "DOMA/%s.abap_ddic"),
			new SimpleAdtUriMapper("/sap/bc/adt/ddic/dataelements/(?<name>.+)", "DTEL/%s.abap_ddic"),
			new SimpleAdtUriMapper("/sap/bc/adt/enhancements/enhoxhh/(?<name>.+)/source/main", "ENHO/%s.abap"),
			// Class Additions
			new AdtClassAdditionsUriMapper("/sap/bc/adt/oo/classes/(?<name>.+)/includes/testclasses",
					"CLAS/%s/%sCCAU.abap"),
			new AdtClassAdditionsUriMapper("/sap/bc/adt/oo/classes/(?<name>.+)/includes/definitions",
					"CLAS/%s/%sCCDEF.abap"),
			new AdtClassAdditionsUriMapper("/sap/bc/adt/oo/classes/(?<name>.+)/includes/implementations",
					"CLAS/%s/%sCCIMP.abap"),
			new AdtClassAdditionsUriMapper("/sap/bc/adt/oo/classes/(?<name>.+)/includes/macros",
					"CLAS/%s/%sCCMAC.abap"),
			new AdtClassAdditionsUriMapper("/sap/bc/adt/oo/classes/(?<name>.+)/includes/localtypes",
					"CLAS/%s/%sCL.abap"),
			// Function Groups
			new AdtFugrIncludeUriMapper(), //
			new AdtFugrModuleUriMapper(), //
			new AdtFugrPathUriMapper(), // Must be at the end, otherwise this matcher matches the includes and modules
	};

	/** Tries to get the Teamscale name for an ADT-URI. */
	/* package */ static Optional<String> getTeamscaleName(String uri, IProject project) {

		for (IAdtUriMapper mapper : ADT_URI_MAPPER) {
			Optional<String> teamscaleName = mapper.map(uri, project);
			if (teamscaleName.isPresent()) {
				return teamscaleName;
			}
		}

		return Optional.empty();
	}

}
