package com.teamscale.ide.client.adt;

import java.net.URI;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

import org.conqat.lib.commons.uniformpath.UniformPath;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.ResourcesPlugin;

import com.sap.adt.tools.core.urimapping.AdtUriMappingServiceFactory;
import com.sap.adt.tools.core.urimapping.IAdtUriMappingService;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.services.ITeamscaleConnectorService;
import com.teamscale.ide.commons.mappings.MappingProblem;
import com.teamscale.ide.commons.mappings.ResourceMappingResult;
import com.teamscale.ide.commons.model.ProjectMapping;
import com.teamscale.ide.commons.model.RemoteResourceReference;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.model.VcsResourceReference;
import com.teamscale.ide.commons.servers.NamedServerDetails;
import com.teamscale.ide.commons.servers.services.IServerCredentialsManager;
import com.teamscale.ide.eclipse.plugin.EclipseServiceRegistry;

/**
 * Stores result of mapping a ADT resource to a remote Teamscale resource.
 */
public class SuccessfullAdtResourceMappingResult extends ResourceMappingResult {

	private final URI serverUrl;

	private final String project;

	public SuccessfullAdtResourceMappingResult(URI serverUrl, String project) {
		this.serverUrl = serverUrl;
		this.project = project;
	}

	@Override
	public boolean isSuccessful() {
		return true;
	}

	@Override
	public URI getServerUrl() {
		return serverUrl;
	}

	@Override
	public String getProjectId() {
		return project;
	}

	@Override
	public Optional<String> getBranchOverride() {
		return Optional.empty();
	}

	@Override
	public ResourceReference getMappingRoot() {
		throw new IllegalStateException("ADT has no mapping roots.");
	}

	@Override
	public RemoteResourceReference getMappedResource(VcsResourceReference localResource) {
		ResourceReference resourceReference = localResource.getResource();

		if (!resourceReference.isFile()) {
			// For ADT findings are only available for files
			return createUnmappedRemote(resourceReference);
		}

		IFile[] files = ResourcesPlugin.getWorkspace().getRoot().findFilesForLocationURI(resourceReference.toUri());
		if (files.length != 1) {
			return createUnmappedRemote(resourceReference);
		}
		IFile file = files[0];

		IAdtUriMappingService mappingService = AdtUriMappingServiceFactory.createUriMappingService();

		try {
			URI adtUri = mappingService.getAdtUri(file);

			String teamscaleName = AdtUriMapping.getTeamscaleName(adtUri.getPath(), file.getProject()).orElse(null);
			
			if (teamscaleName == null) {
				return createUnmappedRemote(resourceReference);
			}

			NamedServerDetails server = EclipseServiceRegistry.getInstance().getService(IServerCredentialsManager.class)
					.getServer(serverUrl).orElse(null);
			List<ProjectMapping> mappings = EclipseServiceRegistry.getInstance()
					.getService(ITeamscaleConnectorService.class)
					.getProjectMappings(server, project, "", Collections.singletonList(teamscaleName));

			String fullFilePath;
			if (mappings == null || mappings.size() != 1) {
				// The file is not known to Teamscale or is not unique in Teamscale
				// This is a workaround. We dont know how to get the correct package name from
				// ADT for a file.
				fullFilePath = "$PRECOMMIT/" + teamscaleName;
			} else {
				fullFilePath = mappings.get(0).fullFilePath;
			}

			String branch = localResource.getBranch().orElse(null);
			return RemoteResourceReference.forFile(serverUrl, project, branch, fullFilePath);

		} catch (NetworkExceptionBase e) {
			return createUnmappedRemote(resourceReference);
		}
	}

	@Override
	public Optional<ResourceReference> getLocalFile(UniformPath remotePath) {
		return Optional.empty();
	}

	@Override
	public MappingProblem getMappingFailure() {
		throw new IllegalStateException("Resource mapping was successful.");
	}

	private RemoteResourceReference createUnmappedRemote(ResourceReference resource) {
		return RemoteResourceReference.forFile(serverUrl, project, null, resource.toUri().toString());
	}
}
