/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.teamscale.ide.commons.client;

import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.conqat.engine.index.shared.PreCommitUploadData;

import com.teamscale.ide.commons.client.exceptions.ApiNotSupportedException;
import com.teamscale.ide.commons.client.exceptions.BranchNotFoundException;
import com.teamscale.ide.commons.client.exceptions.CheckDescriptionNotFoundException;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.exceptions.ProjectNotFoundException;
import com.teamscale.ide.commons.client.exceptions.SAPSystemNotFoundException;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.IIdeIdentifiableFinding;
import com.teamscale.ide.commons.findings.IdeCheckDescription;
import com.teamscale.ide.commons.findings.PreCommitResult;
import com.teamscale.ide.commons.model.IdeCommit;
import com.teamscale.ide.commons.model.ProjectMapping;
import com.teamscale.ide.commons.util.TeamscaleVersion;
import com.teamscale.index.findings.calculation.FindingsWithCount;

/**
 * This interface is used by IDEs to communicate with Teamscale servers.
 */
public interface IIdeServiceClient {

	/** Returns the Teamscale version of the server. */
	TeamscaleVersion getServerVersion();

	/**
	 * Retrieves all project names with their corresponding publicIds.
	 */
	Set<String> getProjects() throws NetworkExceptionBase;

	/**
	 * Retrieves the list of active branch of the given project.
	 *
	 * @throws NetworkExceptionBase
	 *             if the service call fails
	 * @throws ProjectNotFoundException
	 *             if the project with ID {@code publicProjectId} was not found
	 */
	Set<String> getActiveBranches(String publicProjectId) throws NetworkExceptionBase, ProjectNotFoundException;

	/**
	 * Retrieves the default branch of the given project.
	 *
	 * @throws NetworkExceptionBase
	 *             if the service call fails
	 * @throws ProjectNotFoundException
	 *             if the project with ID {@code publicProjectId} was not found
	 */
	String getDefaultBranch(String publicProjectId) throws NetworkExceptionBase, ProjectNotFoundException;

	/**
	 * Resolves the given teamscale project id to the associated SAP system id.
	 *
	 * @throws NetworkExceptionBase
	 *             if the service call fails
	 * @throws ProjectNotFoundException
	 *             if the project with ID {@code publicProjectId} was not found
	 */
	String resolveTeamscaleIdToSapSystemID(String publicProjectId)
			throws NetworkExceptionBase, ProjectNotFoundException;

	/**
	 * Triggers a ABAP Teamscale Sync (i.e. synchronizes all projects associated
	 * with the given SAP system).
	 *
	 * @throws NetworkExceptionBase
	 *             if the service call fails
	 * @throws SAPSystemNotFoundException
	 *             if the project with ID {@code sapSystemId} was not found
	 */
	String synchronizeAbapProjects(String sapSystemId) throws NetworkExceptionBase, SAPSystemNotFoundException;

	/**
	 * Checks if the synchronization with the given id on the given SAP system is
	 * still running.
	 *
	 * @throws NetworkExceptionBase
	 *             if the service call fails
	 */
	boolean isSynchronizingAbapProjects(String sapSystemId, String updateId) throws NetworkExceptionBase;

	/**
	 * Retrieves the findings for the given uniform path or {@code null} if the
	 * uniform path does not exist.
	 *
	 * @throws NetworkExceptionBase
	 *             if the service call fails
	 * @throws ProjectNotFoundException
	 *             if the project with ID {@code publicProjectId} was not found
	 * @throws BranchNotFoundException
	 *             if the branch is not found on the Teamscale server
	 */
	FindingsWithCount getFindings(String publicProjectId, String uniformPath, boolean caseInsensitivePath,
			IdeCommit commit, boolean includeChangedCodeFindings)
			throws NetworkExceptionBase, ProjectNotFoundException, BranchNotFoundException;

	/**
	 * Retrieves the findings for the given uniform path and all its sub paths or
	 * {@code null} if the uniform path does not exist.
	 *
	 * @throws NetworkExceptionBase
	 *             if the service call fails
	 * @throws ProjectNotFoundException
	 *             if the project with ID {@code publicProjectId} was not found
	 * @throws BranchNotFoundException
	 *             if the branch is not found on the Teamscale server
	 */
	FindingsWithCount getFindingsRecursive(String publicProjectId, String uniformPath, boolean caseInsensitivePath,
			IdeCommit commit, boolean includeChangedCodeFindings)
			throws NetworkExceptionBase, ProjectNotFoundException, BranchNotFoundException;

	/**
	 * Retrieves the check description for the given {@link Finding}.
	 *
	 * @throws NetworkExceptionBase
	 *             if the service call fails
	 * @throws ApiNotSupportedException
	 *             if the servers version is not supported anymore
	 * @throws ProjectNotFoundException
	 *             if the project with ID {@code publicProjectId} was not found
	 * @throws CheckDescriptionNotFoundException
	 *             if the check of the {@code finding} was not found
	 */
	IdeCheckDescription getCheckDescription(String publicProjectId, Finding finding) throws NetworkExceptionBase,
			ApiNotSupportedException, ProjectNotFoundException, CheckDescriptionNotFoundException;

	/**
	 * Flags the given finding as false-positive.
	 *
	 * @return whether the process was successful.
	 * @throws NetworkExceptionBase
	 *             if the service call fails
	 */
	boolean flagFindingAsFalsePositive(String publicProjectId, String user, IIdeIdentifiableFinding finding,
			long timestamp, String branchName, String rationale) throws NetworkExceptionBase;

	/**
	 * Flags the given finding as tolerated.
	 *
	 * @return whether the process was successful.
	 * @throws NetworkExceptionBase
	 *             if the service call fails
	 */
	boolean flagFindingAsTolerated(String publicProjectId, String user, IIdeIdentifiableFinding finding, long timestamp,
			String branchName, String rationale) throws NetworkExceptionBase;

	/**
	 * Calculates project mappings for the given project.
	 *
	 * @param paths
	 *            List of relative file paths
	 * @param publicProjectId
	 *            Some project on server
	 * @param branchName
	 *            Name of current VCS branch
	 * @throws NetworkExceptionBase
	 *             if the service call fails
	 * @throws ProjectNotFoundException
	 *             if the project with ID {@code publicProjectId} was not found
	 */
	List<ProjectMapping> calculateProjectMappings(String publicProjectId, List<String> paths, String branchName)
			throws NetworkExceptionBase, ProjectNotFoundException;

	/**
	 * Requests a new instant pre-commit analysis with the given upload data.
	 *
	 * @throws NetworkExceptionBase
	 *             if the service call fails
	 * @throws ApiNotSupportedException
	 *             if the servers version is not supported anymore
	 * @throws ProjectNotFoundException
	 *             if the project with ID @code publicProjectId was not found
	 */
	PreCommitResult requestPreCommit3Result(String publicProjectId, String branch, PreCommitUploadData uploadData)
			throws NetworkExceptionBase, ApiNotSupportedException, ProjectNotFoundException;

	/**
	 * Polls a running instant pre-commit analysis with the given query token.
	 *
	 * @throws NetworkExceptionBase
	 *             if the service call fails
	 * @throws ApiNotSupportedException
	 *             if the servers version is not supported anymore
	 * @throws ProjectNotFoundException
	 *             if the project with ID @code publicProjectId was not found
	 */
	PreCommitResult pollPreCommit3Result(String publicProjectId, String token, Collection<String> allUniformPaths)
			throws NetworkExceptionBase, ApiNotSupportedException, ProjectNotFoundException;

	/**
	 * Removes data associated with the pre-commit branches created for requests
	 * with the given parameters. If the branch and the commit are {@code null}, all
	 * pre-commit branches for the current user are deleted.
	 */
	boolean clearPreCommitBranch(String publicProjectId, String commit, String branch)
			throws NetworkExceptionBase, ApiNotSupportedException, ProjectNotFoundException;

	/** Checks if the server has a valid license */
	boolean hasValidLicense() throws NetworkExceptionBase;
}
