package com.teamscale.ide.commons.client;

import java.io.IOException;
import java.net.URI;
import java.security.KeyManagementException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.cert.X509Certificate;
import java.time.Duration;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;

import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;

import org.apache.http.ssl.SSLContexts;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.checkerframework.checker.nullness.qual.Nullable;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.annotation.PropertyAccessor;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.json.JsonReadFeature;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.json.JsonMapper;
import com.fasterxml.jackson.datatype.guava.GuavaModule;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;
import com.google.common.base.Preconditions;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.exceptions.TeamscaleUnreachableException;
import com.teamscale.ide.commons.servers.AuthenticationInfo;

import okhttp3.Credentials;
import okhttp3.OkHttpClient;
import okhttp3.Request;
import okhttp3.ResponseBody;
import retrofit2.Call;
import retrofit2.Response;
import retrofit2.Retrofit;
import retrofit2.converter.jackson.JacksonConverterFactory;
import retrofit2.converter.scalars.ScalarsConverterFactory;

/**
 * Utility methods for performing service calls.
 */
public class RestClientUtils {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(RestClientUtils.class);

	/** The default timeout for requests in seconds. */
	private static final Duration DEFAULT_TIMEOUT = Duration.ofSeconds(10);

	/**
	 * The {@link ObjectMapper} to use during serialization and deserialization.
	 */
	public static final ObjectMapper OBJECT_MAPPER = defaultObjectMapperBuilder().build();

	/**
	 * Performs a REST request using the given API.
	 * 
	 * @param client
	 *            the {@link Retrofit} client which should be used to perform the
	 *            service call
	 * @param apiClass
	 *            the class of the interface which {@link Retrofit} should use to
	 *            generate the request URL and make the respective request (see
	 *            <a href="https://square.github.io/retrofit/">Retrofit docs</a>).
	 * @param serviceCall
	 *            specifies which function from the {@link Retrofit} interface
	 *            should be executed.
	 * @return the result of the request
	 * @throws NetworkExceptionBase
	 *             when a problem occurred talking to server
	 */
	public static <A, R> Response<R> request(Retrofit client, Class<A> apiClass, Function<A, Call<R>> serviceCall)
			throws NetworkExceptionBase {
		A api = client.create(apiClass);

		Call<R> call = serviceCall.apply(api);
		Response<R> response;
		try {
			response = call.execute();
		} catch (IOException e) {
			throw new TeamscaleUnreachableException(call.request().url().uri(), e);
		}

		return response;
	}

	/**
	 * Returns the {@link URI} of the request which led to the given
	 * {@link Response}.
	 */
	public static <T> URI getRequestUri(Response<T> response) {
		// response.raw() Closeable can be ignored:
		// https://github.com/square/retrofit/issues/3350#issuecomment-609756373
		return response.raw().request().url().uri();
	}

	/**
	 * Returns the {@link Response#errorBody() error body} of the given response as
	 * a {@link String}.
	 */
	public static @Nullable <T> String getErrorBodyAsString(Response<T> response) {
		Preconditions.checkArgument(!response.isSuccessful(),
				"Error body can only be accessed if the response was not successful.");
		try (ResponseBody errorBody = response.errorBody()) {
			if (errorBody == null) {
				return null;
			}
			return errorBody.string();
		} catch (IOException e) {
			LOGGER.error("Could not read body from error response", e);
			return null;
		}
	}

	/**
	 * Repeatedly performs a given service call operations until a result becomes
	 * available or the service call fails. Waits half the given polling interval
	 * before making the initial service call. Thereafter, waits the full polling
	 * interval between service calls. Only waiting half the polling interval before
	 * the first call is meant to return a result quickly in case the service call
	 * operation finishes fast, even if the polling interval is chosen
	 * conservatively, i.e., it is rather large.
	 * <p>
	 * This method <strong>blocks</strong> until the service call operation returns
	 * {@link Optional#isPresent() a result}.
	 *
	 * @throws InterruptedException
	 *             if the calling thread is interrupted while polling
	 */
	public static <T> T pollRepeatedly(IServiceCallOperation<Optional<T>> operation, long pollingInterval,
			TimeUnit pollingIntervalUnit) throws NetworkExceptionBase, InterruptedException {
		long intervalWaitMs = pollingIntervalUnit.toMillis(pollingInterval);

		Optional<T> result;
		do {
			Thread.sleep(intervalWaitMs);
			result = operation.perform();
		} while (!result.isPresent());

		return result.get();
	}

	/**
	 * Creates a {@link Retrofit} client that can be used to perform requests to
	 * Teamscale.
	 */
	public static Retrofit createRetrofitClient(String serverUrl, AuthenticationInfo authenticationInfo,
			List<String> userAgents) {
		OkHttpClient.Builder httpBuilder = new OkHttpClient.Builder();
		try {
			configureSsl(httpBuilder, authenticationInfo);
			configureBasicAuth(httpBuilder, authenticationInfo);
			configureUserAgent(httpBuilder, userAgents);
			configureTimeouts(httpBuilder);
		} catch (NoSuchAlgorithmException | KeyStoreException | KeyManagementException e) {
			throw new IllegalStateException("Error creating HTTP client: " + e.getMessage(), e);
		}

		Retrofit.Builder retrofitBuilder = new Retrofit.Builder();
		retrofitBuilder.baseUrl(serverUrl) //
				.addConverterFactory(ScalarsConverterFactory.create()) // Converts Java types
				.addConverterFactory(JacksonConverterFactory.create(OBJECT_MAPPER)) // Converts JSON
				.client(httpBuilder.build());
		return retrofitBuilder.build();
	}

	/** Configures SSL in the given {@link OkHttpClient.Builder}. */
	private static void configureSsl(OkHttpClient.Builder builder, AuthenticationInfo authenticationInfo)
			throws NoSuchAlgorithmException, KeyStoreException, KeyManagementException {
		Optional<X509TrustManager> trustManager = getCertificateTrustManager(authenticationInfo);
		if (!trustManager.isPresent()) {
			return;
		}

		SSLContext sslContext = SSLContexts.createDefault();
		sslContext.init(null, new TrustManager[] { trustManager.get() }, new SecureRandom());
		builder.sslSocketFactory(sslContext.getSocketFactory(), trustManager.get());
	}

	/** Returns the {@link X509TrustManager trust manager} that should be used. */
	private static Optional<X509TrustManager> getCertificateTrustManager(AuthenticationInfo authenticationInfo) {
		if (authenticationInfo.isTrustAllCertificates()) {
			return Optional.of(new TrustAllCertificatesManager());
		}

		return Optional.empty();
	}

	/**
	 * Configures basic authentication in the given {@link OkHttpClient.Builder}.
	 */
	private static void configureBasicAuth(OkHttpClient.Builder builder, AuthenticationInfo authenticationInfo) {
		builder.authenticator((route, response) -> {
			String credential = Credentials.basic(authenticationInfo.getUsername(), authenticationInfo.getPassword());
			return response.request().newBuilder().header("Authorization", credential).build();
		});
	}

	/**
	 * In the given {@link OkHttpClient.Builder}, configures that request will have
	 * the user agent header set.
	 */
	private static void configureUserAgent(OkHttpClient.Builder builder, List<String> userAgents) {
		builder.addInterceptor(chain -> {
			Request originalRequest = chain.request();
			Request newRequest = originalRequest.newBuilder().header("User-Agent", String.join(" ", userAgents))
					.build();
			return chain.proceed(newRequest);
		});
	}

	/**
	 * Configures the request timeouts in the given {@link OkHttpClient.Builder}.
	 *
	 * @see #DEFAULT_TIMEOUT
	 */
	private static void configureTimeouts(OkHttpClient.Builder builder) {
		builder.connectTimeout(DEFAULT_TIMEOUT) //
				.callTimeout(DEFAULT_TIMEOUT) //
				.readTimeout(DEFAULT_TIMEOUT) //
				.writeTimeout(DEFAULT_TIMEOUT);
	}

	/**
	 * Creates the {@link ObjectMapper} which should be used during serialization
	 * and deserialization.
	 */
	private static JsonMapper.Builder defaultObjectMapperBuilder() {
		JsonFactory jsonFactory = JsonFactory.builder() //
				.enable(JsonReadFeature.ALLOW_TRAILING_COMMA) //
				.enable(JsonReadFeature.ALLOW_UNQUOTED_FIELD_NAMES) //
				.enable(JsonReadFeature.ALLOW_SINGLE_QUOTES) //
				.enable(JsonReadFeature.ALLOW_JAVA_COMMENTS) //
				.build();
		JsonMapper.Builder builder = JsonMapper.builder(jsonFactory);
		builder.addModule(new GuavaModule()) //
				.addModule(new JavaTimeModule()) //
				.visibility(PropertyAccessor.ALL, JsonAutoDetect.Visibility.NONE) //
				.visibility(PropertyAccessor.FIELD, JsonAutoDetect.Visibility.ANY) //
				.serializationInclusion(JsonInclude.Include.NON_NULL) //
				.disable(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES);

		return builder;
	}

	/**
	 * A simple implementation of {@link X509TrustManager} that simple trusts every
	 * certificate.
	 */
	private static class TrustAllCertificatesManager implements X509TrustManager {

		@Override
		public X509Certificate[] getAcceptedIssuers() {
			return new X509Certificate[0];
		}

		@Override
		public void checkServerTrusted(X509Certificate[] certs, String authType) {
			// Nothing to do
		}

		@Override
		public void checkClientTrusted(X509Certificate[] certs, String authType) {
			// Nothing to do
		}
	}
}
