/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.client.impl;

import java.net.URI;
import java.util.function.Function;

import com.google.common.collect.ImmutableList;
import com.teamscale.ide.commons.client.IIdeServiceClient;
import com.teamscale.ide.commons.client.RestClientUtils;
import com.teamscale.ide.commons.client.exceptions.ApiExceptionBase;
import com.teamscale.ide.commons.client.exceptions.AuthenticationFailedException;
import com.teamscale.ide.commons.client.exceptions.HttpException;
import com.teamscale.ide.commons.client.exceptions.InvalidTeamscaleResponseException;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.exceptions.ProjectNotConfiguredCorrectlyException;
import com.teamscale.ide.commons.client.exceptions.TeamscaleApiVersionMismatchException;
import com.teamscale.ide.commons.servers.AuthenticationInfo;
import com.teamscale.ide.commons.util.TeamscaleVersion;
import com.teamscale.ide.commons.util.TeamscaleVersionRange;

import retrofit2.Call;
import retrofit2.Response;
import retrofit2.Retrofit;

/**
 * Base class for TeamscaleXXServiceClients
 */
public abstract class TeamscaleIdeServiceClientBase implements IIdeServiceClient {

	/** The URL of the server */
	protected final String serverUrl;

	/** The information needed to authenticate at the sever */

	private final AuthenticationInfo authenticationInfo;

	/** A list of user agents which should be sent in the User-Agent header. */
	private final ImmutableList<String> userAgents;

	/** The Teamscale version of the server. */
	private final TeamscaleVersion serverVersion;

	protected TeamscaleIdeServiceClientBase(String serverUrl, AuthenticationInfo authenticationInfo,
			ImmutableList<String> userAgents, TeamscaleVersionRange serverSupportedApiVersions)
			throws NetworkExceptionBase {

		TeamscaleVersion serverVersion = serverSupportedApiVersions.maxInclusive;
		if (!serverSupportedApiVersions.isInRange(getUsedApiVersion())) {
			throw new TeamscaleApiVersionMismatchException(URI.create(serverUrl), serverVersion, getUsedApiVersion());
		}

		this.serverUrl = serverUrl;
		this.authenticationInfo = authenticationInfo;
		this.userAgents = userAgents;
		this.serverVersion = serverVersion;
	}

	/**
	 * Returns the API version that will be used when making service calls.
	 * Subclasses should override this method with their targeted version.
	 */
	protected abstract TeamscaleVersion getUsedApiVersion();

	@Override
	public TeamscaleVersion getServerVersion() {
		return serverVersion;
	}

	/**
	 * Performs a REST request using the given API.
	 *
	 * @param apiClass
	 *            the class of the interface which {@link Retrofit} should use to
	 *            generate the request URL and make the respective request (see
	 *            <a href="https://square.github.io/retrofit/">Retrofit docs</a>).
	 * @param serviceCall
	 *            specifies which function from the {@link Retrofit} interface
	 *            should be executed.
	 * @return the result of the request
	 * @throws NetworkExceptionBase
	 *             when a problem occurred talking to server
	 */
	protected final <A, R> Response<R> request(Class<A> apiClass, Function<A, Call<R>> serviceCall)
			throws NetworkExceptionBase {
		return request(serverUrl, authenticationInfo, userAgents, apiClass, serviceCall);
	}

	/**
	 * Performs a REST request using the given API.
	 *
	 * @see #request(Class, Function)
	 */
	protected static <A, R> Response<R> request(String serverUrl, AuthenticationInfo authenticationInfo,
			ImmutableList<String> userAgents, Class<A> apiClass, Function<A, Call<R>> serviceCall)
			throws NetworkExceptionBase {
		Retrofit client = RestClientUtils.createRetrofitClient(serverUrl, authenticationInfo, userAgents);
		return RestClientUtils.request(client, apiClass, serviceCall);
	}

	/**
	 * Returns the given object if it is not null, otherwise throws a
	 * {@link NetworkExceptionBase}.
	 */
	protected static <T> T ensureNotNull(T object) throws NetworkExceptionBase {
		if (object == null) {
			throw new InvalidTeamscaleResponseException("Body of Teamscale server's response was null");
		}
		return object;
	}

	/**
	 * @see #ensureResponseIsSuccessful(Response)
	 */
	protected static <T> T ensureResponseIsSuccessful(Response<T> response, String serverUrl)
			throws NetworkExceptionBase {
		if (!response.isSuccessful()) {
			throwExceptionForErrorCode(response, serverUrl);
		}
		return response.body();
	}

	/**
	 * Ensures that the given response {@link Response#isSuccessful() is
	 * successful}. Throws some exception based on {@link NetworkExceptionBase} if
	 * it is not.
	 */
	protected <T> T ensureResponseIsSuccessful(Response<T> response) throws NetworkExceptionBase {
		if (!response.isSuccessful()) {
			throwExceptionForErrorCode(response, serverUrl);
		}
		return response.body();
	}

	/**
	 * Ensures that the given response {@link Response#isSuccessful() is
	 * successful}. Throws some exception based on {@link NetworkExceptionBase} or a
	 * specific {@link ApiExceptionBase} (in case of 404s) if it is not.
	 */
	protected <T, E extends ApiExceptionBase> T ensureResponseIsSuccessfulWithGivenExceptionFor404(Response<T> response,
			E correspondingException) throws NetworkExceptionBase, E {
		if (!response.isSuccessful()) {
			if (response.code() == 404) {
				throw correspondingException;
			}
			throwExceptionForErrorCode(response, serverUrl);
		}
		return response.body();
	}

	/**
	 * Throws the appropriate exception for error codes which could be thrown by any
	 * service call.
	 */
	protected static <T> void throwExceptionForErrorCode(Response<T> response, String serverUrl)
			throws NetworkExceptionBase {
		URI serverUri = URI.create(serverUrl);
		switch (response.code()) {
		case 401: // Unauthorized
		case 403: // Forbidden
			throw new AuthenticationFailedException(serverUri, response.code());
		case 500: // Internal server error
			throw new ProjectNotConfiguredCorrectlyException(serverUri);
		default:
			throw new HttpException(RestClientUtils.getRequestUri(response), response.message(), response.code(),
					RestClientUtils.getErrorBodyAsString(response));
		}
	}

}
