/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.client.impl.v67;

import java.util.Collections;
import java.util.List;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.teamscale.ide.commons.findings.IIdeIdentifiableFinding;
import com.teamscale.ide.commons.lang.ToStringHelpers;

/**
 * DTO for {@link com.teamscale.index.blacklisting.FindingBlacklistInfo} as
 * returned by version 6.7 of the Teamscale API.
 */
/* package */ class FindingBlacklistInfo67Dto {

	private static final String BLACKLIST_INFO_PROPERTY_NAME = "blacklistInfo";
	private static final String FINDING_ID_LIST_PROPERTY_NAME = "findingIds";

	@JsonProperty(BLACKLIST_INFO_PROPERTY_NAME)
	private final Info blacklistInfo;

	@JsonProperty(FINDING_ID_LIST_PROPERTY_NAME)
	private final List<String> findingIds;

	public FindingBlacklistInfo67Dto(IIdeIdentifiableFinding finding, long timestamp, String user, String rationale,
			EFindingBlacklistType type) {
		blacklistInfo = new Info(finding, timestamp, user, rationale, type);
		findingIds = Collections.singletonList(finding.getId());
	}

	private static class Info {

		/** The name of the JSON property name for {@link #findingId}. */
		private static final String FINDING_ID_PROPERTY = "findingId";

		/** The name of the JSON property name for {@link #timestamp}. */
		private static final String TIMESTAMP_PROPERTY = "timestamp";

		/** The name of the JSON property name for {@link #user}. */
		private static final String USER_PROPERTY = "user";

		/** The name of the JSON property name for {@link #rationale}. */
		private static final String RATIONALE_PROPERTY = "rationale";

		/** The name of the JSON property name for {@link #type}. */
		private static final String TYPE_PROPERTY = "type";

		/** The ID of the blacklisted finding */
		@JsonProperty(FINDING_ID_PROPERTY)
		public final String findingId;

		/** The time when the finding was blacklisted */
		@JsonProperty(TIMESTAMP_PROPERTY)
		public final long timestamp;

		/** The name of the user who blacklisted the finding */
		@JsonProperty(USER_PROPERTY)
		public final String user;

		/** A rationale why the finding was blacklisted */
		@JsonProperty(RATIONALE_PROPERTY)
		public final String rationale;

		@JsonProperty(TYPE_PROPERTY)
		private final EFindingBlacklistType type;

		private Info(IIdeIdentifiableFinding finding, long timestamp, String user, String rationale,
				EFindingBlacklistType type) {
			findingId = finding.getId();
			this.timestamp = timestamp;
			this.user = user;
			this.rationale = rationale;
			this.type = type;
		}

		@Override
		public String toString() {
			return ToStringHelpers.toReflectiveStringHelper(this).toString();
		}
	}

	/**
	 * Defines different types of blacklisting depending on the user's intent.
	 */
	public enum EFindingBlacklistType {

		/** A false-positive is a finding that is mistakenly reported by the analysis */
		FALSE_POSITIVE,

		/**
		 * A toleration is a finding that the user does not intend to fix (i.e. is
		 * accepted as is)
		 */
		TOLERATION
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
