/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.client.impl.v67;

import java.net.URI;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.commons.lang3.StringUtils;
import org.conqat.engine.index.shared.PreCommitUploadData;

import com.google.common.collect.ImmutableList;
import com.teamscale.ide.commons.client.IIdeServiceClient;
import com.teamscale.ide.commons.client.Uris;
import com.teamscale.ide.commons.client.exceptions.BranchNotFoundException;
import com.teamscale.ide.commons.client.exceptions.CheckDescriptionNotFoundException;
import com.teamscale.ide.commons.client.exceptions.LicenseExpiredException;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.exceptions.ProjectNotFoundException;
import com.teamscale.ide.commons.client.exceptions.SAPSystemNotFoundException;
import com.teamscale.ide.commons.client.impl.TeamscaleIdeServiceClientBase;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.IIdeIdentifiableFinding;
import com.teamscale.ide.commons.findings.IdeCheckDescription;
import com.teamscale.ide.commons.findings.PreCommitResult;
import com.teamscale.ide.commons.model.IdeCommit;
import com.teamscale.ide.commons.model.ProjectMapping;
import com.teamscale.ide.commons.servers.AuthenticationInfo;
import com.teamscale.ide.commons.util.ICheckedSupplier;
import com.teamscale.ide.commons.util.TeamscaleVersion;
import com.teamscale.ide.commons.util.TeamscaleVersionRange;
import com.teamscale.index.findings.calculation.FindingsWithCount;

import retrofit2.Call;
import retrofit2.Response;
import retrofit2.http.Body;
import retrofit2.http.DELETE;
import retrofit2.http.GET;
import retrofit2.http.POST;
import retrofit2.http.PUT;
import retrofit2.http.Path;
import retrofit2.http.Query;

/**
 * The service client implementation used for communication with Teamscale
 * servers of versions 6.7 or newer.
 */
public class Teamscale67IdeServiceClient extends TeamscaleIdeServiceClientBase {

	private static final TeamscaleVersion VERSION_6_7 = new TeamscaleVersion(6, 7, 0);

	/** The license check needs to be disabled during testing */
	private final boolean disableLicenseCheckForTesting;

	/**
	 * Creates a service client that can be used to communicate with Teamscale 6.7
	 * servers. Needs to wrap constructor as the (overloaded) cannot be used as
	 * method reference.
	 */
	public static IIdeServiceClient create(String serverUrl, AuthenticationInfo authenticationInfo,
			ImmutableList<String> userAgents) throws NetworkExceptionBase {
		return new Teamscale67IdeServiceClient(serverUrl, authenticationInfo, userAgents,
				queryServerSupportedApiVersions(serverUrl, authenticationInfo, userAgents), false);
	}

	protected Teamscale67IdeServiceClient(String serverUrl, AuthenticationInfo authenticationInfo,
			ImmutableList<String> userAgents, TeamscaleVersionRange serverSupportedApiVersions,
			boolean disableLicenseCheckForTesting) throws NetworkExceptionBase {
		super(serverUrl, authenticationInfo, userAgents, serverSupportedApiVersions);
		this.disableLicenseCheckForTesting = disableLicenseCheckForTesting;

	}

	/**
	 * Determines the range of API versions supported by the given Teamscale server.
	 */
	protected static TeamscaleVersionRange queryServerSupportedApiVersions(String serverUrl,
			AuthenticationInfo authenticationInfo, ImmutableList<String> userAgents) throws NetworkExceptionBase {
		Response<ServiceApiInfo67Dto> response = request(serverUrl, authenticationInfo, userAgents,
				IRetrofit67TeamscaleApi.class, IRetrofit67TeamscaleApi::getServiceApiInfo);

		ServiceApiInfo67Dto serviceApiInfo = ensureResponseIsSuccessful(response, serverUrl);

		return new TeamscaleVersionRange(convertVersion(serviceApiInfo.minApiVersion),
				convertVersion(serviceApiInfo.maxApiVersion));
	}

	private static TeamscaleVersion convertVersion(TeamscaleVersion67Dto versionDto) {
		return new TeamscaleVersion(versionDto.major, versionDto.minor, versionDto.patch);
	}

	@Override
	protected TeamscaleVersion getUsedApiVersion() {
		return VERSION_6_7;
	}

	protected <R> Response<R> executeWithLicenseCheck(ICheckedSupplier<Response<R>, NetworkExceptionBase> serviceCall,
			String message) throws NetworkExceptionBase {
		if (!disableLicenseCheckForTesting && !hasValidLicense()) {
			throw new LicenseExpiredException(URI.create(serverUrl), message);
		}
		return serviceCall.get();
	}

	@Override
	public Set<String> getProjects() throws NetworkExceptionBase {
		Response<List<ProjectInfo67Dto>> response = request(IRetrofit67TeamscaleApi.class,
				IRetrofit67TeamscaleApi::getProjects);

		if (!response.isSuccessful()) {
			if (response.code() == 404) {
				return Collections.emptySet();
			}
			throwExceptionForErrorCode(response, serverUrl);
		}

		List<ProjectInfo67Dto> projectInfos = response.body();
		if (projectInfos == null) {
			return Collections.emptySet();
		}

		return projectInfos.stream().map(projectInfo -> projectInfo.id).collect(Collectors.toSet());
	}

	@Override
	public Set<String> getActiveBranches(String publicProjectId) throws NetworkExceptionBase, ProjectNotFoundException {
		Response<BranchesInfo67Dto> response = request(IRetrofit67TeamscaleApi.class,
				api -> api.getActiveBranches(publicProjectId));

		BranchesInfo67Dto branchesInfo = ensureNotNull(ensureResponseIsSuccessfulWithGivenExceptionFor404(response,
				new ProjectNotFoundException(publicProjectId)));
		return new HashSet<>(branchesInfo.getLiveBranches());
	}

	@Override
	public String getDefaultBranch(String publicProjectId) throws NetworkExceptionBase, ProjectNotFoundException {
		Response<String> response = request(IRetrofit67TeamscaleApi.class,
				api -> api.getDefaultBranch(publicProjectId));

		return ensureNotNull(ensureResponseIsSuccessfulWithGivenExceptionFor404(response,
				new ProjectNotFoundException(publicProjectId)));
	}

	@Override
	public String resolveTeamscaleIdToSapSystemID(String publicProjectId)
			throws NetworkExceptionBase, ProjectNotFoundException {
		Response<String> response = request(IRetrofit67TeamscaleApi.class,
				api -> api.resolveTeamscaleIdToSapSystemID(publicProjectId));

		return ensureNotNull(ensureResponseIsSuccessfulWithGivenExceptionFor404(response,
				new ProjectNotFoundException(publicProjectId)));
	}

	@Override
	public String synchronizeAbapProjects(String sapSystemId) throws NetworkExceptionBase, SAPSystemNotFoundException {
		Response<String> response = request(IRetrofit67TeamscaleApi.class,
				api -> api.synchronizeAbapProjects(sapSystemId));

		return ensureNotNull(ensureResponseIsSuccessfulWithGivenExceptionFor404(response,
				new SAPSystemNotFoundException(sapSystemId)));
	}

	@Override
	public boolean isSynchronizingAbapProjects(String sapSystemId, String updateId) throws NetworkExceptionBase {
		Response<Boolean> response = request(IRetrofit67TeamscaleApi.class,
				api -> api.isSynchronizingAbapProjects(sapSystemId, updateId));

		return ensureNotNull(ensureResponseIsSuccessful(response));
	}

	@Override
	public FindingsWithCount getFindings(String publicProjectId, String uniformPath, boolean caseInsensitivePath,
			IdeCommit commit, boolean includeChangedCodeFindings)
			throws NetworkExceptionBase, ProjectNotFoundException, BranchNotFoundException {
		return getFindings(publicProjectId, uniformPath, caseInsensitivePath, commit, includeChangedCodeFindings, false,
				true);
	}

	@Override
	public FindingsWithCount getFindingsRecursive(String publicProjectId, String uniformPath,
			boolean caseInsensitivePath, IdeCommit commit, boolean includeChangedCodeFindings)
			throws NetworkExceptionBase, ProjectNotFoundException, BranchNotFoundException {
		return getFindings(publicProjectId, uniformPath, caseInsensitivePath, commit, includeChangedCodeFindings, true,
				false);
	}

	private FindingsWithCount getFindings(String publicProjectId, String uniformPath, boolean caseInsensitivePath,
			IdeCommit commit, boolean includeChangedCodeFindings, boolean recursive, boolean fetchAll)
			throws NetworkExceptionBase, ProjectNotFoundException, BranchNotFoundException {
		Response<FindingsWithCount> response = request(IRetrofit67TeamscaleApi.class,
				api -> api.getFindings(publicProjectId, Uris.asTimestampQueryParameter(commit), uniformPath,
						caseInsensitivePath, includeChangedCodeFindings, recursive, fetchAll));

		FindingsWithCount findings = ensureNotNull(ensureResponseIsSuccessfulWithGivenExceptionFor404(response,
				new ProjectNotFoundException(publicProjectId)));

		// For the "findings/list/with-count" endpoint, an empty list can happen if the
		// branch is not known to the server. Check whether the branch is unknown and
		// throw an appropriate exception.
		Optional<String> branch = commit.getBranchName();
		if (branch.isPresent() && findings.getFindings().isEmpty()
				&& !getActiveBranches(publicProjectId).contains(branch.get())) {
			throw new BranchNotFoundException(branch.get());
		}

		return findings;
	}

	@Override
	public IdeCheckDescription getCheckDescription(String publicProjectId, Finding finding)
			throws NetworkExceptionBase, ProjectNotFoundException, CheckDescriptionNotFoundException {
		Response<CheckDescription67Dto[]> response = request(IRetrofit67TeamscaleApi.class,
				api -> api.getCheckDescriptions(publicProjectId, Collections.singletonList(finding.getTypeId())));

		CheckDescription67Dto[] descriptions = ensureNotNull(ensureResponseIsSuccessfulWithGivenExceptionFor404(
				response, new ProjectNotFoundException(publicProjectId)));
		if (descriptions.length == 0) {
			throw new CheckDescriptionNotFoundException(finding.getTypeId());
		}

		return new IdeCheckDescription(descriptions[0].checkName, descriptions[0].description, URI.create(serverUrl));
	}

	@Override
	public boolean flagFindingAsFalsePositive(String publicProjectId, String user, IIdeIdentifiableFinding finding,
			long timestamp, String branchName, String rationale) throws NetworkExceptionBase {
		FindingBlacklistInfo67Dto blacklistInfo = new FindingBlacklistInfo67Dto(finding, timestamp, user, rationale,
				FindingBlacklistInfo67Dto.EFindingBlacklistType.FALSE_POSITIVE);
		String commit = Uris.asTimestampQueryParameter(IdeCommit.head(branchName));

		Response<Void> response = request(IRetrofit67TeamscaleApi.class,
				api -> api.flagFindingAsFalsePositive(publicProjectId, commit, "add", blacklistInfo));

		if (!response.isSuccessful()) {
			if (response.code() == 404) {
				return false;
			}
			throwExceptionForErrorCode(response, serverUrl);
		}

		return true;
	}

	@Override
	public boolean flagFindingAsTolerated(String publicProjectId, String user, IIdeIdentifiableFinding finding,
			long timestamp, String branchName, String rationale) throws NetworkExceptionBase {
		FindingBlacklistInfo67Dto blacklistInfo = new FindingBlacklistInfo67Dto(finding, timestamp, user, rationale,
				FindingBlacklistInfo67Dto.EFindingBlacklistType.TOLERATION);
		String commit = Uris.asTimestampQueryParameter(IdeCommit.head(branchName));

		Response<Void> response = request(IRetrofit67TeamscaleApi.class,
				api -> api.flagFindingAsTolerated(publicProjectId, commit, "add", blacklistInfo));

		if (!response.isSuccessful()) {
			if (response.code() == 404) {
				return false;
			}
			throwExceptionForErrorCode(response, serverUrl);
		}

		return true;
	}

	@Override
	public List<ProjectMapping> calculateProjectMappings(String publicProjectId, List<String> paths, String branchName)
			throws NetworkExceptionBase, ProjectNotFoundException {
		Response<ProjectMapping67Dto[]> response = request(IRetrofit67TeamscaleApi.class,
				api -> api.calculateProjectMappings(publicProjectId, branchName, paths));

		ProjectMapping67Dto[] projectMappings = ensureNotNull(ensureResponseIsSuccessfulWithGivenExceptionFor404(
				response, new ProjectNotFoundException(publicProjectId)));
		return Arrays.stream(projectMappings)
				.map(dto -> new ProjectMapping(dto.fullFilePath, dto.projectId, dto.from, dto.to))
				.collect(Collectors.toList());
	}

	@Override
	public PreCommitResult requestPreCommit3Result(String publicProjectId, String branch,
			PreCommitUploadData uploadData) throws NetworkExceptionBase, ProjectNotFoundException {
		ICheckedSupplier<Response<PreCommit3Result67Dto>, NetworkExceptionBase> requestPreCommit3Result = () -> request(
				IRetrofit67TeamscaleApi.class,
				api -> api.requestPreCommit3Result(publicProjectId, branch, uploadData.getUniformPathToContentMap()));
		Response<PreCommit3Result67Dto> response = executeWithLicenseCheck(requestPreCommit3Result,
				"Pre-commit findings could not be retrieved due to an invalid license");

		PreCommit3Result67Dto result = ensureNotNull(ensureResponseIsSuccessfulWithGivenExceptionFor404(response,
				new ProjectNotFoundException(publicProjectId)));
		Set<String> allUniformPaths = new HashSet<>();
		allUniformPaths.addAll(uploadData.getDeletedUniformPaths());
		allUniformPaths.addAll(uploadData.getUniformPathToContentMap().keySet());
		return convertDto(result, allUniformPaths);
	}

	@Override
	public PreCommitResult pollPreCommit3Result(String publicProjectId, String token,
			Collection<String> allUniformPaths) throws NetworkExceptionBase, ProjectNotFoundException {
		ICheckedSupplier<Response<PreCommit3Result67Dto>, NetworkExceptionBase> pollPreCommit3Result = () -> request(
				IRetrofit67TeamscaleApi.class, api -> api.pollPreCommit3Result(publicProjectId, token));
		Response<PreCommit3Result67Dto> response = executeWithLicenseCheck(pollPreCommit3Result,
				"Pre-commit findings could not be retrieved due to an invalid license");

		return convertDto(ensureNotNull(ensureResponseIsSuccessfulWithGivenExceptionFor404(response,
				new ProjectNotFoundException(publicProjectId))), allUniformPaths);

	}

	private static PreCommitResult convertDto(PreCommit3Result67Dto result, Collection<String> allUniformPaths) {
		if (StringUtils.isEmpty(result.error)) {
			return PreCommitResult.pending(result.token).addFindings(result.findings).addGeneralPaths(allUniformPaths)
					.build();
		}

		return PreCommitResult.finished()
				.addGlobalError(PreCommitResult.PreCommitErrorDetail.LEGACY_ERROR_TYPE, result.error, true)
				.addGeneralPaths(allUniformPaths).build();
	}

	@Override
	public boolean clearPreCommitBranch(String publicProjectId, String commit, String branch)
			throws NetworkExceptionBase, ProjectNotFoundException {
		ICheckedSupplier<Response<Void>, NetworkExceptionBase> clearPreCommit3Branch = () -> request(
				IRetrofit67TeamscaleApi.class, api -> api.clearPreCommitBranch(publicProjectId, commit, branch));
		Response<Void> response = executeWithLicenseCheck(clearPreCommit3Branch,
				"Pre-commit findings could not be retrieved due to an invalid license");

		if (response.isSuccessful()) {
			return true;
		}
		if (response.code() == 404) {
			throw new ProjectNotFoundException(publicProjectId);
		}
		return false;
	}

	@Override
	public boolean hasValidLicense() throws NetworkExceptionBase {
		Response<String> response = request(IRetrofit67TeamscaleApi.class, IRetrofit67TeamscaleApi::getHealthStatus);

		String healthStatus = ensureNotNull(ensureResponseIsSuccessful(response));
		return StringUtils.containsAnyIgnoreCase(healthStatus, "LICENSE OK",
				"License Warning: License will expire in less than");
	}

	/**
	 * Represents the version 6.7 of Teamscale servers API used in requests made
	 * with {@link retrofit2.Retrofit}.
	 */
	private interface IRetrofit67TeamscaleApi {

		/** Retrieves information about the version of the Teamscale server's API. */
		@GET("api/v6.7/version")
		Call<ServiceApiInfo67Dto> getServiceApiInfo();

		/**
		 * Retrieves all project names with their corresponding publicIds.
		 */
		@GET("api/v6.7/projects")
		Call<List<ProjectInfo67Dto>> getProjects();

		/**
		 * Retrieves the list of active branch of the given project.
		 */
		@GET("api/v6.7/projects/{projectId}/branches")
		Call<BranchesInfo67Dto> getActiveBranches(@Path("projectId") String projectId);

		/**
		 * Retrieves the default branch of the given project.
		 */
		@GET("api/v6.7/projects/{projectId}/default-branch")
		Call<String> getDefaultBranch(@Path("projectId") String projectId);

		/**
		 * Resolves the given teamscale project id to the associated SAP system id.
		 */
		@GET("api/v6.7/sap-system-id-by-project/{publicProjectId}")
		Call<String> resolveTeamscaleIdToSapSystemID(@Path("publicProjectId") String publicProjectId);

		/**
		 * Triggers a ABAP Teamscale Sync.
		 */
		@POST("api/v6.7/sap-sysid-projects/{sapSystemId}/update")
		Call<String> synchronizeAbapProjects(@Path("sapSystemId") String sapSystemId);

		/**
		 * Retrieves whether the synchronization with the given id on the given SAP
		 * system is still running.
		 */
		@GET("api/v6.7/sap-sysid-projects/{sapSystemId}/is-updating/{updateId}")
		Call<Boolean> isSynchronizingAbapProjects(@Path("sapSystemId") String sapSystemId,
				@Path("updateId") String updateId);

		/**
		 * Retrieves the findings for the given uniform path (including sub-paths, if
		 * {@code recursive} parameter is set to {@code true}) starting from baseline or
		 * {@code null} if the uniform path does not exist.
		 */
		@GET("api/v6.7/projects/{publicProjectId}/findings/list/with-count")
		Call<FindingsWithCount> getFindings(@Path("publicProjectId") String publicProjectId, @Query("t") String commit,
				@Query("uniform-path") String uniformPath, @Query("case-insensitive-path") boolean caseInsensitivePath,
				@Query("include-changed-code-findings") boolean includeChangedCodeFindings,
				@Query("recursive") boolean recursive, @Query("all") boolean fetchAll);

		/**
		 * Retrieves the {@link com.teamscale.ide.commons.findings.IdeCheckDescription}
		 * for the given {@link Finding#getTypeId() finding type id}s.
		 */
		@POST("api/v6.7/projects/{projectId}/finding-type-descriptors")
		Call<CheckDescription67Dto[]> getCheckDescriptions(@Path("projectId") String projectId,
				@Body List<String> findingTypeIds);

		/** Flags the given finding as false-positive. */
		@PUT("api/v6.7/projects/{publicProjectId}/findings/false-positive")
		Call<Void> flagFindingAsFalsePositive(@Path("publicProjectId") String publicProjectId,
				@Query("t") String commit, @Query("operation") String operation,
				@Body FindingBlacklistInfo67Dto blacklistInfo);

		/** Flags the given finding as tolerated. */
		@PUT("api/v6.7/projects/{publicProjectId}/findings/tolerated")
		Call<Void> flagFindingAsTolerated(@Path("publicProjectId") String publicProjectId, @Query("t") String commit,
				@Query("operation") String operation, @Body FindingBlacklistInfo67Dto blacklistInfo);

		/** Calculates project mappings for the given project. */
		@POST("api/v6.7/projects/{publicProjectId}/auto-prefix-mappings")
		Call<ProjectMapping67Dto[]> calculateProjectMappings(@Path("publicProjectId") String publicProjectId,
				@Query("branch") String branchName, @Body List<String> paths);

		/**
		 * Requests a new instant pre-commit analysis with the given upload data.
		 */
		@PUT("api/v6.7/projects/{publicProjectId}/pre-commit3")
		Call<PreCommit3Result67Dto> requestPreCommit3Result(@Path("publicProjectId") String publicProjectId,
				@Query("branch") String branch, @Body Map<String, String> uniformPathToContent);

		/**
		 * Polls a running instant pre-commit analysis with the given query token.
		 */
		@GET("api/v6.7/projects/{publicProjectId}/pre-commit3/{token}")
		Call<PreCommit3Result67Dto> pollPreCommit3Result(@Path("publicProjectId") String publicProjectId,
				@Path("token") String token);

		/**
		 * Clears a pre-commit branch.
		 */
		@DELETE("api/v6.7/projects/{publicProjectId}/pre-commit3")
		Call<Void> clearPreCommitBranch(@Path("publicProjectId") String publicProjectId, @Query("commit") String commit,
				@Query("branch") String branchName);

		/**
		 * Retrieves the health status of the Teamscale server (including information
		 * about the license validity).
		 */
		@GET("api/v6.7/health-check")
		Call<String> getHealthStatus();
	}
}
