/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.client.impl.v71;

import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import com.google.common.collect.ImmutableList;
import com.teamscale.ide.commons.client.IIdeServiceClient;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.impl.v67.Teamscale67IdeServiceClient;
import com.teamscale.ide.commons.servers.AuthenticationInfo;
import com.teamscale.ide.commons.util.TeamscaleVersion;
import com.teamscale.ide.commons.util.TeamscaleVersionRange;

import retrofit2.Call;
import retrofit2.Response;
import retrofit2.http.GET;

/**
 * The service client implementation used for communication with Teamscale
 * servers of versions 7.1 or newer.
 */
public class Teamscale71IdeServiceClient extends Teamscale67IdeServiceClient {

	private static final TeamscaleVersion VERSION_7_1 = new TeamscaleVersion(7, 1, 0);

	/**
	 * Creates a service client that can be used to communicate with Teamscale 7.1
	 * servers. Needs to wrap constructor as the (overloaded) cannot be used as
	 * method reference.
	 */
	public static IIdeServiceClient create(String serverUrl, AuthenticationInfo authenticationInfo,
			ImmutableList<String> userAgents) throws NetworkExceptionBase {
		return new Teamscale71IdeServiceClient(serverUrl, authenticationInfo, userAgents,
				queryServerSupportedApiVersions(serverUrl, authenticationInfo, userAgents), false);
	}

	protected Teamscale71IdeServiceClient(String serverUrl, AuthenticationInfo authenticationInfo,
			ImmutableList<String> userAgents, TeamscaleVersionRange serverSupportedApiVersions,
			boolean disableLicenseCheckForTesting) throws NetworkExceptionBase {
		super(serverUrl, authenticationInfo, userAgents, serverSupportedApiVersions, disableLicenseCheckForTesting);
	}

	@Override
	protected TeamscaleVersion getUsedApiVersion() {
		return VERSION_7_1;
	}

	@Override
	public Set<String> getProjects() throws NetworkExceptionBase {
		Response<List<ProjectInfo71Dto>> response = request(IRetrofit71TeamscaleApi.class,
				IRetrofit71TeamscaleApi::getProjects);

		if (!response.isSuccessful()) {
			if (response.code() == 404) {
				return Collections.emptySet();
			}
			throwExceptionForErrorCode(response, serverUrl);
		}

		List<ProjectInfo71Dto> projectInfos = ensureNotNull(response.body());
		return projectInfos.stream().map(projectInfo -> projectInfo.publicIds).flatMap(List::stream)
				.collect(Collectors.toSet());
	}

	/**
	 * Represents the version 7.1 of Teamscale servers API used in requests made
	 * with {@link retrofit2.Retrofit}.
	 */
	private interface IRetrofit71TeamscaleApi {

		/**
		 * Retrieves all project names with their corresponding publicIds.
		 */
		@GET("api/v7.1/projects")
		Call<List<ProjectInfo71Dto>> getProjects();
	}
}
