/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.client.impl.v87;

import static org.conqat.engine.index.shared.PreCommit3Result.EPrecommit3ErrorType.PRECOMMIT_BRANCH_NOT_AVAILABLE_DURING_POLL;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import org.conqat.engine.index.shared.PreCommit3Result;
import org.conqat.engine.index.shared.PreCommitUploadData;

import com.google.common.collect.ImmutableList;
import com.teamscale.ide.commons.client.IIdeServiceClient;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.exceptions.ProjectNotFoundException;
import com.teamscale.ide.commons.client.impl.v71.Teamscale71IdeServiceClient;
import com.teamscale.ide.commons.findings.PreCommitResult;
import com.teamscale.ide.commons.servers.AuthenticationInfo;
import com.teamscale.ide.commons.util.ICheckedSupplier;
import com.teamscale.ide.commons.util.TeamscaleVersion;
import com.teamscale.ide.commons.util.TeamscaleVersionRange;

import retrofit2.Call;
import retrofit2.Response;
import retrofit2.http.Body;
import retrofit2.http.GET;
import retrofit2.http.PUT;
import retrofit2.http.Path;
import retrofit2.http.Query;

/**
 * The service client implementation used for communication with Teamscale
 * servers of versions 8.7 or newer.
 */
public class Teamscale87IdeServiceClient extends Teamscale71IdeServiceClient {

	private static final TeamscaleVersion VERSION_8_7 = new TeamscaleVersion(8, 7, 0);

	/**
	 * Creates a service client that can be used to communicate with Teamscale 8.7
	 * servers. Needs to wrap constructor as the (overloaded) cannot be used as
	 * method reference.
	 */
	public static IIdeServiceClient create(String serverUrl, AuthenticationInfo authenticationInfo,
			ImmutableList<String> userAgents) throws NetworkExceptionBase {
		return new Teamscale87IdeServiceClient(serverUrl, authenticationInfo, userAgents,
				queryServerSupportedApiVersions(serverUrl, authenticationInfo, userAgents), false);
	}

	protected Teamscale87IdeServiceClient(String serverUrl, AuthenticationInfo authenticationInfo,
			ImmutableList<String> userAgents, TeamscaleVersionRange serverSupportedApiVersions,
			boolean disableLicenseCheckForTesting) throws NetworkExceptionBase {
		super(serverUrl, authenticationInfo, userAgents, serverSupportedApiVersions, disableLicenseCheckForTesting);
	}

	@Override
	protected TeamscaleVersion getUsedApiVersion() {
		return VERSION_8_7;
	}

	@Override
	public PreCommitResult requestPreCommit3Result(String publicProjectId, String branch,
			PreCommitUploadData uploadData) throws NetworkExceptionBase, ProjectNotFoundException {
		ICheckedSupplier<Response<PreCommit3Result>, NetworkExceptionBase> requestPreCommit3Result = () -> request(
				IRetrofit87TeamscaleApi.class,
				api -> api.requestPreCommit3Result(publicProjectId, branch, uploadData.getUniformPathToContentMap()));
		Response<PreCommit3Result> response = executeWithLicenseCheck(requestPreCommit3Result,
				"Pre-commit findings could not be retrieved due to an invalid license");

		PreCommit3Result result = ensureNotNull(ensureResponseIsSuccessfulWithGivenExceptionFor404(response,
				new ProjectNotFoundException(publicProjectId)));
		Set<String> allUniformPaths = new HashSet<>();
		allUniformPaths.addAll(uploadData.getDeletedUniformPaths());
		allUniformPaths.addAll(uploadData.getUniformPathToContentMap().keySet());
		return convertDto(result, allUniformPaths);
	}

	@Override
	public PreCommitResult pollPreCommit3Result(String publicProjectId, String token,
			Collection<String> allUniformPaths) throws NetworkExceptionBase, ProjectNotFoundException {
		ICheckedSupplier<Response<PreCommit3Result>, NetworkExceptionBase> pollPreCommit3Result = () -> request(
				IRetrofit87TeamscaleApi.class, api -> api.pollPreCommit3Result(publicProjectId, token));
		Response<PreCommit3Result> response = executeWithLicenseCheck(pollPreCommit3Result,
				"Pre-commit findings could not be retrieved due to an invalid license");

		return convertDto(ensureNotNull(ensureResponseIsSuccessfulWithGivenExceptionFor404(response,
				new ProjectNotFoundException(publicProjectId))), allUniformPaths);
	}

	private static PreCommitResult convertDto(PreCommit3Result result, Collection<String> allUniformPaths) {
		List<PreCommitResult.PreCommitErrorDetail> errors = new ArrayList<>();

		if (result.detailedErrors != null) {
			for (PreCommit3Result.PreCommit3ErrorDetail error : result.detailedErrors) {
				String type = error.type;
				if (Objects.equals(type, "PRECOMMIT_BRANCH_NOT_AVILABLE_DURING_POLL")) {
					// The enum had a TYPO between 8.7 and 8.9
					// Convert it to the correct ENUM
					type = PRECOMMIT_BRANCH_NOT_AVAILABLE_DURING_POLL.toString();
				}
				errors.add(new PreCommitResult.PreCommitErrorDetail(type, error.message, error.affectedUniformPath,
						error.resultsInvalid));
			}
		}

		return PreCommitResult.pending(result.getToken()).addErrors(errors).addFindings(result.findings)
				.addGeneralPaths(allUniformPaths).build();
	}

	/**
	 * Represents the version 8.7 of Teamscale servers API used in requests made
	 * with {@link retrofit2.Retrofit}.
	 */
	private interface IRetrofit87TeamscaleApi {
		/**
		 * Requests a new instant pre-commit analysis with the given upload data.
		 */
		@PUT("api/v8.7/projects/{publicProjectId}/pre-commit3")
		Call<PreCommit3Result> requestPreCommit3Result(@Path("publicProjectId") String publicProjectId,
				@Query("branch") String branch, @Body Map<String, String> uniformPathToContent);

		/**
		 * Polls a running instant pre-commit analysis with the given query token.
		 */
		@GET("api/v8.7/projects/{publicProjectId}/pre-commit3/{token}")
		Call<PreCommit3Result> pollPreCommit3Result(@Path("publicProjectId") String publicProjectId,
				@Path("token") String token);
	}
}
