/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.client.services;

import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.conqat.engine.index.shared.PreCommitUploadData;
import org.conqat.lib.commons.uniformpath.UniformPath;

import com.teamscale.ide.commons.client.exceptions.ApiNotSupportedException;
import com.teamscale.ide.commons.client.exceptions.BranchNotFoundException;
import com.teamscale.ide.commons.client.exceptions.CheckDescriptionNotFoundException;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.exceptions.ProjectNotFoundException;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.IIdeIdentifiableFinding;
import com.teamscale.ide.commons.findings.IdeCheckDescription;
import com.teamscale.ide.commons.findings.PreCommitResult;
import com.teamscale.ide.commons.model.IdeCommit;
import com.teamscale.ide.commons.model.ProjectMapping;
import com.teamscale.ide.commons.plugin.IPluginService;
import com.teamscale.ide.commons.servers.ETeamscaleServerState;
import com.teamscale.ide.commons.servers.NamedServerDetails;
import com.teamscale.ide.commons.servers.services.IServerStateBroadcastService;
import com.teamscale.index.findings.calculation.FindingsWithCount;

/**
 * Manages requests to Teamscale. Also manages whether Teamscale is enabled in
 * general.
 */
public interface ITeamscaleConnectorService extends IPluginService {

	/**
	 * Checks and broadcasts the state of a server
	 */
	ETeamscaleServerState getAndBroadcastServerState(NamedServerDetails server);

	/**
	 * Fetches the findings from the given Teamscale server. Updates the state of
	 * the server using the {@link IServerStateBroadcastService}.
	 *
	 * @param server
	 *            Details of configuration of server to use.
	 * @param publicProjectId
	 *            The ID of the public project for which the findings should be
	 *            fetched
	 * @param uniformPath
	 *            The path to the file for which the findings should be fetched
	 * @param caseInsensitivePath
	 *            Whether uniform path should be treated case-insensitive
	 * @param commit
	 *            The commit for which the findings should be fetched
	 * @param includeSubpaths
	 *            Whether findings should be fetched recursive for this resource
	 * @return A list of findings for the given resource
	 * @throws NetworkExceptionBase
	 *             if the service call fails or the connection to the Teamscale
	 *             server cannot be established
	 * @throws ProjectNotFoundException,
	 *             if the project is not found on the Teamscale server
	 */
	FindingsWithCount getFindings(NamedServerDetails server, String publicProjectId, UniformPath uniformPath,
			boolean caseInsensitivePath, IdeCommit commit, boolean includeSubpaths)
			throws NetworkExceptionBase, ProjectNotFoundException, BranchNotFoundException;

	/**
	 * Fetches the check description for the given {@link Finding}.
	 *
	 * @param server
	 *            Details of configuration of server to use.
	 * @param publicProjectId
	 *            The ID of the public project from which the description should be
	 *            fetched
	 * @param finding
	 *            for which the description should be fetched
	 * @return The description of the check that created the {@link Finding}
	 * @throws NetworkExceptionBase
	 *             if the service call fails or the connection to the Teamscale
	 *             server cannot be established
	 * @throws ProjectNotFoundException,
	 *             if the project is not found on the Teamscale server
	 * @throws ApiNotSupportedException
	 *             if the server does not support the API (e.g. server version too
	 *             old) needed for this service to work
	 * @throws CheckDescriptionNotFoundException
	 *             if no description could be found for the given
	 *             {@code findingTypeId}
	 */
	IdeCheckDescription getCheckDescription(NamedServerDetails server, String publicProjectId, Finding finding)
			throws NetworkExceptionBase, ApiNotSupportedException, ProjectNotFoundException,
			CheckDescriptionNotFoundException;

	/**
	 * Gets mappings for a Teamscale project given the relative paths of files found
	 * on the server. The idea behind using more than one file is to increase the
	 * chances of getting correct mappings for the IDE project.
	 *
	 * @param server
	 *            Details of configuration of server to use.
	 * @param publicProjectId
	 *            Some Teamscale project.
	 * @param branchName
	 *            Name of VCS branch from which the files were taken
	 * @param relativePaths
	 *            List of relative file paths
	 * @return Project Mappings for Teamscale project.
	 * @throws NetworkExceptionBase
	 *             if the service call fails or the connection to the Teamscale
	 *             server cannot be established
	 */
	List<ProjectMapping> getProjectMappings(NamedServerDetails server, String publicProjectId, String branchName,
			List<String> relativePaths) throws NetworkExceptionBase;

	/**
	 * Flags the given {@link Finding} as false-positive on the given
	 * {@link NamedServerDetails Teamscale server}. Updates the state of the server
	 * using the {@link IServerStateBroadcastService}.
	 *
	 * @param server
	 *            The Teamscale server on which the finding should be flagged
	 * @param publicProjectId
	 *            The according project on the Teamscale server
	 * @param finding
	 *            The finding to flag as false-positive
	 * @param timestamp
	 *            The time when the finding was flagged
	 * @param branchName
	 *            The name of the branch on which the finding should be flagged
	 * @param rationale
	 *            The rationale why the finding was flagged as false-positive
	 * @return whether the process was successful.
	 * @throws NetworkExceptionBase
	 *             if the service call fails or the connection to the Teamscale
	 *             server cannot be established
	 */
	boolean flagFindingAsFalsePositive(NamedServerDetails server, String publicProjectId,
			IIdeIdentifiableFinding finding, long timestamp, String branchName, String rationale)
			throws NetworkExceptionBase;

	/**
	 * Flags the given {@link Finding} as tolerated on the given
	 * {@link NamedServerDetails Teamscale server}. Updates the state of the server
	 * using the {@link IServerStateBroadcastService}.
	 *
	 * @param server
	 *            The Teamscale server on which the finding should be flagged
	 * @param publicProjectId
	 *            The according project on the Teamscale server
	 * @param finding
	 *            The finding to flag as false-positive
	 * @param timestamp
	 *            The time when the finding was flagged
	 * @param branchName
	 *            The name of the branch on which the finding should be flagged
	 * @param rationale
	 *            The rationale why the finding was flagged as false-positive
	 * @return whether the process was successful.
	 * @throws NetworkExceptionBase
	 *             if the service call fails or the connection to the Teamscale
	 *             server cannot be established
	 */
	boolean flagFindingAsTolerated(NamedServerDetails server, String publicProjectId, IIdeIdentifiableFinding finding,
			long timestamp, String branchName, String rationale) throws NetworkExceptionBase;

	/**
	 * Fetches a list of projects from the given Teamscale server. Updates the state
	 * of the server using the {@link IServerStateBroadcastService}.
	 *
	 * @param server
	 *            The Teamscale server from which the projects should be fetched
	 * @return the projects on the given Teamscale server
	 * @throws NetworkExceptionBase
	 *             if the service call fails or the connection to the Teamscale
	 *             server cannot be established
	 */
	Set<String> getProjects(NamedServerDetails server) throws NetworkExceptionBase;

	/**
	 * Retrieves the branches for the given Teamscale project from the given
	 * Teamscale server. Updates the state of the server using the
	 * {@link IServerStateBroadcastService}.
	 *
	 * @throws NetworkExceptionBase
	 *             if the service call fails or the connection to the Teamscale
	 *             server cannot be established
	 */
	Set<String> getBranches(NamedServerDetails server, String publicProjectId) throws NetworkExceptionBase;

	/**
	 * Retrieves the default branch for the given Teamscale project from the given
	 * Teamscale server. Updates the state of the server using the
	 * {@link IServerStateBroadcastService}.
	 *
	 * @throws NetworkExceptionBase
	 *             if the service call fails or the connection to the Teamscale
	 *             server cannot be established
	 */
	String getDefaultBranch(NamedServerDetails server, String publicProjectId) throws NetworkExceptionBase;

	/**
	 * Requests a new instant pre-commit analysis with the given upload data.
	 */
	PreCommitResult requestPreCommit3Result(NamedServerDetails server, String publicProjectId, String branch,
			PreCommitUploadData uploadData)
			throws NetworkExceptionBase, ApiNotSupportedException, ProjectNotFoundException;

	/**
	 * Polls a running instant pre-commit analysis with the given query token.
	 */
	PreCommitResult pollPreCommit3Result(NamedServerDetails server, String publicProjectId, String token,
			Collection<String> allUniformPaths) throws NetworkExceptionBase, ApiNotSupportedException;

	/**
	 * Removes data associated with the pre-commit branches created for requests
	 * with the given parameters. If the branch and the commit are {@code null}, all
	 * pre-commit branches for the current user are deleted.
	 */
	boolean clearPreCommitBranch(NamedServerDetails server, String publicProjectId, String commit, String branch)
			throws NetworkExceptionBase, ApiNotSupportedException, ProjectNotFoundException;

	/**
	 * Retrieves the SAP system id associated to the given teamscale project from
	 * the Teamscale server.
	 * 
	 * @param server
	 *            The teamscale server details
	 * @param publicProjectId
	 *            The teamscale public project ID
	 * @return The SAP system id, or empty if there was a problem retrieving the id.
	 * @throws NetworkExceptionBase
	 *             if the service call fails or the connection to the Teamscale
	 *             server cannot be established
	 */
	String resolveTeamscaleIdToSapSystemID(NamedServerDetails server, String publicProjectId)
			throws NetworkExceptionBase;

	/**
	 * Triggers a ABAP Teamscale Sync (i.e. synchronizes all projects associated
	 * with the given SAP system).
	 *
	 * @param server
	 *            The teamscale server details
	 * @param sapSystemId
	 *            The id of the SAP system
	 * @return The synchronization id needed to periodically poll the status of the
	 *         synchronization, empty if the synchronization could not be started
	 * @throws NetworkExceptionBase
	 *             if the service call fails or the connection to the Teamscale
	 *             server cannot be established
	 */
	String synchronizeAbapProjects(NamedServerDetails server, String sapSystemId) throws NetworkExceptionBase;

	/**
	 * Checks the status of the ABAP projects synchronization. Must be called after
	 * a synchronization has been requested via
	 * {@link ITeamscaleConnectorService#synchronizeAbapProjects(NamedServerDetails, String)}
	 *
	 * @param server
	 *            The teamscale server details
	 * @param sapSystemId
	 *            The id of the updating SAP system
	 * @param syncId
	 *            The synchronization id of the requested synchronization
	 * @return {@code false} if the synchronization is finished, {@code true} in any
	 *         other state of the synchronization cycle
	 *
	 * @throws NetworkExceptionBase
	 *             if the service call fails or the connection to the Teamscale
	 *             server cannot be established
	 */
	boolean isSynchronizingAbapProjects(NamedServerDetails server, String sapSystemId, String syncId)
			throws NetworkExceptionBase;
}
