/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.client.services.impl;

import static java.util.Objects.requireNonNull;

import java.net.URI;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.ExecutionException;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.checkerframework.checker.nullness.qual.Nullable;
import org.conqat.engine.index.shared.PreCommitUploadData;
import org.conqat.lib.commons.uniformpath.UniformPath;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.base.Preconditions;
import com.google.common.base.Strings;
import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.collect.ImmutableList;
import com.teamscale.ide.commons.client.IIdeServiceClient;
import com.teamscale.ide.commons.client.IServiceCallFunction;
import com.teamscale.ide.commons.client.IServiceClientCreator;
import com.teamscale.ide.commons.client.exceptions.ApiExceptionBase;
import com.teamscale.ide.commons.client.exceptions.ApiNotSupportedException;
import com.teamscale.ide.commons.client.exceptions.BranchNotFoundException;
import com.teamscale.ide.commons.client.exceptions.CheckDescriptionNotFoundException;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.exceptions.ProjectNotFoundException;
import com.teamscale.ide.commons.client.exceptions.TeamscaleUnreachableException;
import com.teamscale.ide.commons.client.impl.v67.Teamscale67IdeServiceClient;
import com.teamscale.ide.commons.client.impl.v71.Teamscale71IdeServiceClient;
import com.teamscale.ide.commons.client.impl.v87.Teamscale87IdeServiceClient;
import com.teamscale.ide.commons.client.services.ITeamscaleConnectorService;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.IIdeIdentifiableFinding;
import com.teamscale.ide.commons.findings.IdeCheckDescription;
import com.teamscale.ide.commons.findings.PreCommitResult;
import com.teamscale.ide.commons.lang.ToStringHelpers;
import com.teamscale.ide.commons.model.IdeCommit;
import com.teamscale.ide.commons.model.ProjectMapping;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.notification.Notifications;
import com.teamscale.ide.commons.notification.services.INotificationService;
import com.teamscale.ide.commons.servers.AuthenticationInfo;
import com.teamscale.ide.commons.servers.ETeamscaleServerState;
import com.teamscale.ide.commons.servers.NamedServerDetails;
import com.teamscale.ide.commons.servers.services.IServerStateBroadcastService;
import com.teamscale.ide.commons.util.TeamscaleVersion;
import com.teamscale.index.findings.calculation.FindingsWithCount;

/** Manages requests to Teamscale. */
public class TeamscaleConnectorService implements ITeamscaleConnectorService {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(TeamscaleConnectorService.class);

	/**
	 * The message to include in an error message, that indicates that the project
	 * ID {@link Strings#isNullOrEmpty(String) is null or empty}.
	 */
	private static final String PROJECT_ID_NULL_ERROR_MESSAGE = "Project ID cannot be empty";

	/**
	 * The version of the plugin
	 * <p>
	 * Might be null if plugin version cannot be determined
	 */
	private final TeamscaleVersion pluginVersion;

	/**
	 * The plugin's server state broadcast service instance.
	 */
	private final IServerStateBroadcastService stateBroadcast;

	/** The plugin's server to send notification to the user. */
	private final INotificationService notificationService;

	private final ImmutableList<String> userAgents;

	/**
	 * The userAgents needs to be in the format of the User-Agent. Use
	 * {@link TeamscaleConnectorService#buildUserAgent(String, String, String)} to
	 * create the right format.
	 *
	 * @see <a href=
	 *      "https://www.rfc-editor.org/rfc/rfc9110.html#name-user-agent">RFC 9110:
	 *      User-Agent header</a>
	 */
	public TeamscaleConnectorService(TeamscaleVersion pluginVersion, IServerStateBroadcastService stateBroadcast,
			INotificationService notificationService, String... userAgents) {
		this.pluginVersion = pluginVersion;
		this.stateBroadcast = requireNonNull(stateBroadcast);
		this.notificationService = requireNonNull(notificationService);
		this.userAgents = ImmutableList.copyOf(userAgents);
	}

	private final LoadingCache<CacheKey, IIdeServiceClient> serviceClientCache = CacheBuilder.newBuilder().build( //
			new CacheLoader<CacheKey, IIdeServiceClient>() {

				@Override
				public IIdeServiceClient load(CacheKey key) throws NetworkExceptionBase {
					return createServiceClient(key.server);
				}
			});

	/**
	 * Wrapper class for NamedServerDetails to load a new service client whenever
	 * URL, password, or username changes.
	 */
	@VisibleForTesting
	static final class CacheKey {

		private final NamedServerDetails server;

		private CacheKey(NamedServerDetails server) {
			this.server = requireNonNull(server);
		}

		@Override
		public int hashCode() {
			return Objects.hash(server.getUrl(), server.getPassword(), server.getUsername());
		}

		@Override
		public boolean equals(Object other) {
			if (this == other) {
				return true;
			}
			if (!(other instanceof CacheKey)) {
				return false;
			}
			CacheKey that = (CacheKey) other;
			return Objects.equals(server.getUrl(), that.server.getUrl()) && //
					Objects.equals(server.getUsername(), that.server.getUsername()) && //
					Objects.equals(server.getPassword(), that.server.getPassword());
		}

		@Override
		public String toString() {
			return ToStringHelpers.toReflectiveStringHelper(this).toString();
		}
	}

	/**
	 * Tries loading the service client to communicate with a Teamscale server.
	 */
	@VisibleForTesting
	IIdeServiceClient createServiceClient(NamedServerDetails server) throws NetworkExceptionBase {
		ImmutableList<IServiceClientCreator> creators = ImmutableList.of(Teamscale87IdeServiceClient::create,
				Teamscale71IdeServiceClient::create, Teamscale67IdeServiceClient::create);
		AuthenticationInfo authenticationInfo = new AuthenticationInfo(server.getUsername(), server.getPassword(),
				server.isTrustAllCertificates());
		NetworkExceptionBase exception = new TeamscaleUnreachableException(URI.create(server.getUrl()));
		for (IServiceClientCreator creator : creators) {
			try {
				IIdeServiceClient serviceClient = creator.create(server.getUrl(), authenticationInfo, userAgents);
				TeamscaleVersion serverVersion = serviceClient.getServerVersion();

				if (pluginVersion != null && shouldRecommendPluginUpdate(serverVersion, pluginVersion)) {
					notificationService.reportNotification(ENotificationCause.INDIRECT, Notifications
							.higherServerVersionWarning(URI.create(server.getUrl()), serverVersion, pluginVersion));
				}

				return serviceClient;
			} catch (NetworkExceptionBase e) {
				exception.addSuppressed(e);
			}
		}
		throw exception;
	}

	/**
	 * Returns whether the user should be recommended to update the plugin. This is
	 * the case if the major/minor release number of the server is higher than the
	 * one from the plugin.
	 */
	@VisibleForTesting
	static boolean shouldRecommendPluginUpdate(TeamscaleVersion serverVersion, TeamscaleVersion pluginVersion) {
		int pluginMajor = pluginVersion.major;
		int pluginMinor = pluginVersion.minor;

		if (pluginMajor == 0 && pluginMinor == 0) {
			// This means we are in Dev environment. No recommendation needed.
			return false;
		}

		int serverMajor = serverVersion.major;
		int serverMinor = serverVersion.minor;

		return serverMajor > pluginMajor || (serverMajor == pluginMajor && serverMinor > pluginMinor);
	}

	private IIdeServiceClient loadServiceClientFromCache(NamedServerDetails server) throws NetworkExceptionBase {
		try {
			return serviceClientCache.get(new CacheKey(server));
		} catch (ExecutionException e) {
			NetworkExceptionBase unreachableException = new TeamscaleUnreachableException(URI.create(server.getUrl()));
			unreachableException.addSuppressed(e);
			throw unreachableException;
		}
	}

	@Override
	public ETeamscaleServerState getAndBroadcastServerState(NamedServerDetails server) {
		LOGGER.traceEntry(null, server);
		try {
			tryCall(false, server, IIdeServiceClient::getProjects);
			boolean licenseValid = tryCall(false, server, IIdeServiceClient::hasValidLicense);
			if (!licenseValid) {
				return LOGGER.traceExit(ETeamscaleServerState.LICENSE_INVALID);
			}
			return LOGGER.traceExit(ETeamscaleServerState.REACHABLE);
		} catch (NetworkExceptionBase e) {
			LOGGER.catching(e);
			return LOGGER.traceExit(e.getServerState());
		}
	}

	@Override
	public FindingsWithCount getFindings(NamedServerDetails server, String publicProjectId, UniformPath uniformPath,
			boolean caseInsensitivePath, IdeCommit commit, boolean includeSubpaths)
			throws NetworkExceptionBase, ProjectNotFoundException, BranchNotFoundException {
		Preconditions.checkArgument(!Strings.isNullOrEmpty(publicProjectId), PROJECT_ID_NULL_ERROR_MESSAGE);
		LOGGER.traceEntry(null, server, publicProjectId, uniformPath, commit, includeSubpaths);
		FindingsWithCount result;
		if (includeSubpaths) {
			result = tryCall(
					false, server, client -> client.getFindingsRecursive(publicProjectId, uniformPath.toString(),
							caseInsensitivePath, commit, false),
					ProjectNotFoundException.class, BranchNotFoundException.class);
		} else {
			result = tryCall(
					false, server, client -> client.getFindings(publicProjectId, uniformPath.toString(),
							caseInsensitivePath, commit, false),
					ProjectNotFoundException.class, BranchNotFoundException.class);
		}
		return LOGGER.traceExit(result);
	}

	@Override
	public IdeCheckDescription getCheckDescription(NamedServerDetails server, String publicProjectId, Finding finding)
			throws NetworkExceptionBase, ApiNotSupportedException, ProjectNotFoundException,
			CheckDescriptionNotFoundException {
		Preconditions.checkArgument(!Strings.isNullOrEmpty(publicProjectId), PROJECT_ID_NULL_ERROR_MESSAGE);
		LOGGER.traceEntry(null, server, publicProjectId, finding);
		IdeCheckDescription result = tryCall(false, server,
				client -> client.getCheckDescription(publicProjectId, finding), ProjectNotFoundException.class,
				ApiNotSupportedException.class, CheckDescriptionNotFoundException.class);
		return LOGGER.traceExit(result);
	}

	@Override
	public List<ProjectMapping> getProjectMappings(NamedServerDetails server, String publicProjectId, String branchName,
			List<String> relativePaths) throws NetworkExceptionBase {
		Preconditions.checkArgument(!Strings.isNullOrEmpty(publicProjectId), PROJECT_ID_NULL_ERROR_MESSAGE);
		LOGGER.traceEntry(null, server, publicProjectId, branchName, relativePaths);
		List<ProjectMapping> result = tryCall(false, server,
				client -> client.calculateProjectMappings(publicProjectId, relativePaths, branchName));
		return LOGGER.traceExit(result);
	}

	@Override
	public boolean flagFindingAsFalsePositive(NamedServerDetails server, String publicProjectId,
			IIdeIdentifiableFinding finding, long timestamp, String branchName, String rationale)
			throws NetworkExceptionBase {
		Preconditions.checkArgument(!Strings.isNullOrEmpty(publicProjectId), PROJECT_ID_NULL_ERROR_MESSAGE);
		LOGGER.traceEntry(null, server, publicProjectId, finding, timestamp, branchName, rationale);
		boolean wasSuccessful = tryCall(false, server, client -> client.flagFindingAsFalsePositive(publicProjectId,
				server.getUsername(), finding, timestamp, branchName, rationale));
		return LOGGER.traceExit(wasSuccessful);
	}

	@Override
	public boolean flagFindingAsTolerated(NamedServerDetails server, String publicProjectId,
			IIdeIdentifiableFinding finding, long timestamp, String branchName, String rationale)
			throws NetworkExceptionBase {
		Preconditions.checkArgument(!Strings.isNullOrEmpty(publicProjectId), PROJECT_ID_NULL_ERROR_MESSAGE);
		LOGGER.traceEntry(null, server, publicProjectId, finding, timestamp, branchName, rationale);
		boolean wasSuccessful = tryCall(false, server, client -> client.flagFindingAsTolerated(publicProjectId,
				server.getUsername(), finding, timestamp, branchName, rationale));
		return LOGGER.traceExit(wasSuccessful);
	}

	@Override
	public Set<String> getProjects(NamedServerDetails server) throws NetworkExceptionBase {
		LOGGER.traceEntry(null, server);
		Set<String> result = tryCall(false, server, IIdeServiceClient::getProjects);
		return LOGGER.traceExit(result);
	}

	@Override
	public Set<String> getBranches(NamedServerDetails server, String publicProjectId) throws NetworkExceptionBase {
		Preconditions.checkArgument(!Strings.isNullOrEmpty(publicProjectId), PROJECT_ID_NULL_ERROR_MESSAGE);
		LOGGER.traceEntry(null, server, publicProjectId);
		Set<String> result = tryCall(false, server, client -> client.getActiveBranches(publicProjectId));
		return LOGGER.traceExit(result);
	}

	@Override
	public String getDefaultBranch(NamedServerDetails server, String publicProjectId) throws NetworkExceptionBase {
		Preconditions.checkArgument(!Strings.isNullOrEmpty(publicProjectId), PROJECT_ID_NULL_ERROR_MESSAGE);
		LOGGER.traceEntry(null, server, publicProjectId);
		String result = tryCall(false, server, client -> client.getDefaultBranch(publicProjectId));
		return LOGGER.traceExit(result);
	}

	@Override
	public PreCommitResult requestPreCommit3Result(NamedServerDetails server, String publicProjectId, String branch,
			PreCommitUploadData uploadData)
			throws NetworkExceptionBase, ApiNotSupportedException, ProjectNotFoundException {
		Preconditions.checkArgument(!Strings.isNullOrEmpty(publicProjectId), PROJECT_ID_NULL_ERROR_MESSAGE);
		LOGGER.traceEntry(null, server, publicProjectId);
		PreCommitResult result = tryCall(true, server,
				client -> client.requestPreCommit3Result(publicProjectId, branch, uploadData),
				ApiNotSupportedException.class, ProjectNotFoundException.class);
		return LOGGER.traceExit(result);
	}

	@Override
	public PreCommitResult pollPreCommit3Result(NamedServerDetails server, String publicProjectId, String token,
			Collection<String> allUniformPaths) throws NetworkExceptionBase, ApiNotSupportedException {
		Preconditions.checkArgument(!Strings.isNullOrEmpty(publicProjectId), PROJECT_ID_NULL_ERROR_MESSAGE);
		LOGGER.traceEntry(null, server);
		PreCommitResult result = tryCall(true, server,
				client -> client.pollPreCommit3Result(publicProjectId, token, allUniformPaths),
				ApiNotSupportedException.class);
		return LOGGER.traceExit(result);
	}

	@Override
	public boolean clearPreCommitBranch(NamedServerDetails server, String publicProjectId, String commit, String branch)
			throws NetworkExceptionBase, ApiNotSupportedException, ProjectNotFoundException {
		Preconditions.checkArgument(!Strings.isNullOrEmpty(publicProjectId), PROJECT_ID_NULL_ERROR_MESSAGE);
		LOGGER.traceEntry(null, server);
		boolean successful = tryCall(true, server,
				client -> client.clearPreCommitBranch(publicProjectId, commit, branch), ApiNotSupportedException.class,
				ProjectNotFoundException.class);
		return LOGGER.traceExit(successful);
	}

	@Override
	public String resolveTeamscaleIdToSapSystemID(NamedServerDetails server, String publicProjectId)
			throws NetworkExceptionBase {
		Preconditions.checkArgument(!Strings.isNullOrEmpty(publicProjectId), PROJECT_ID_NULL_ERROR_MESSAGE);
		LOGGER.traceEntry(null, server, publicProjectId);
		String result = tryCall(false, server, client -> client.resolveTeamscaleIdToSapSystemID(publicProjectId));
		return LOGGER.traceExit(result);
	}

	@Override
	public String synchronizeAbapProjects(NamedServerDetails server, String sapSystemId) throws NetworkExceptionBase {
		LOGGER.traceEntry(null, server, sapSystemId);
		String result = tryCall(false, server, client -> client.synchronizeAbapProjects(sapSystemId));
		return LOGGER.traceExit(result);
	}

	@Override
	public boolean isSynchronizingAbapProjects(NamedServerDetails server, String sapSystemId, String syncId)
			throws NetworkExceptionBase {
		LOGGER.traceEntry(null, server, sapSystemId);
		boolean result = tryCall(false, server, client -> client.isSynchronizingAbapProjects(sapSystemId, syncId));
		return LOGGER.traceExit(result);
	}

	private <T> T tryCall(boolean callRequiresLicense, NamedServerDetails server,
			IServiceCallFunction<IIdeServiceClient, T, ApiExceptionBase> serviceCall) throws NetworkExceptionBase {
		try {
			return callAndBroadcastServerState(callRequiresLicense, server, serviceCall);
		} catch (NetworkExceptionBase e) {
			throw LOGGER.throwing(e);
		} catch (ApiExceptionBase e) {
			throw LOGGER.throwing(new AssertionError("Unexpected subtype of ApiExceptionBase:" + e.getClass(), e));
		}
	}

	private <T, E1 extends ApiExceptionBase> T tryCall(boolean callRequiresLicense, NamedServerDetails server,
			IServiceCallFunction<IIdeServiceClient, T, ApiExceptionBase> serviceCall, Class<E1> possibleException1)
			throws NetworkExceptionBase, E1 {
		try {
			return callAndBroadcastServerState(callRequiresLicense, server, serviceCall);
		} catch (NetworkExceptionBase e) {
			throw LOGGER.throwing(e);
		} catch (ApiExceptionBase e) {
			if (possibleException1.isInstance(e)) {
				throw LOGGER.throwing(possibleException1.cast(e));
			} else {
				throw LOGGER.throwing(new AssertionError("Unexpected subtype of ApiExceptionBase:" + e.getClass(), e));
			}
		}
	}

	private <T, E1 extends ApiExceptionBase, E2 extends ApiExceptionBase> T tryCall(boolean callRequiresLicense,
			NamedServerDetails server, IServiceCallFunction<IIdeServiceClient, T, ApiExceptionBase> serviceCall,
			Class<E1> possibleException1, Class<E2> possibleException2) throws NetworkExceptionBase, E1, E2 {
		try {
			return callAndBroadcastServerState(callRequiresLicense, server, serviceCall);
		} catch (NetworkExceptionBase e) {
			throw LOGGER.throwing(e);
		} catch (ApiExceptionBase e) {
			if (possibleException1.isInstance(e)) {
				throw LOGGER.throwing(possibleException1.cast(e));
			} else if (possibleException2.isInstance(e)) {
				throw LOGGER.throwing(possibleException2.cast(e));
			} else {
				throw LOGGER.throwing(new AssertionError("Unexpected subtype of ApiExceptionBase:" + e.getClass(), e));
			}
		}
	}

	private <T, E1 extends ApiExceptionBase, E2 extends ApiExceptionBase, E3 extends ApiExceptionBase> T tryCall(
			boolean callRequiresLicense, NamedServerDetails server,
			IServiceCallFunction<IIdeServiceClient, T, ApiExceptionBase> serviceCall, Class<E1> possibleException1,
			Class<E2> possibleException2, Class<E3> possibleException3) throws NetworkExceptionBase, E1, E2, E3 {
		try {
			return callAndBroadcastServerState(callRequiresLicense, server, serviceCall);
		} catch (NetworkExceptionBase e) {
			throw LOGGER.throwing(e);
		} catch (ApiExceptionBase e) {
			if (possibleException1.isInstance(e)) {
				throw LOGGER.throwing(possibleException1.cast(e));
			} else if (possibleException2.isInstance(e)) {
				throw LOGGER.throwing(possibleException2.cast(e));
			} else if (possibleException3.isInstance(e)) {
				throw LOGGER.throwing(possibleException3.cast(e));
			} else {
				throw LOGGER.throwing(new AssertionError("Unexpected subtype of ApiExceptionBase:" + e.getClass(), e));
			}
		}
	}

	/**
	 * Tries to execute the callback server call function and updates the server
	 * connection status accordingly.
	 *
	 * @param callRequiresLicense
	 *            ture if the call requires a license
	 */
	private <T, E extends ApiExceptionBase> T callAndBroadcastServerState(boolean callRequiresLicense,
			NamedServerDetails server, IServiceCallFunction<IIdeServiceClient, T, E> serviceCall)
			throws NetworkExceptionBase, E {
		try {
			IIdeServiceClient client = loadServiceClientFromCache(server);
			T result = serviceCall.apply(client);
			ETeamscaleServerState currentState = stateBroadcast.getServerState(server)
					.orElse(ETeamscaleServerState.UNREACHABLE);
			ETeamscaleServerState newState = deduceNextServerState(currentState, callRequiresLicense);
			stateBroadcast.updateServerState(server, newState);
			return result;
		} catch (NetworkExceptionBase e) {
			stateBroadcast.updateServerState(server, e.getServerState());
			throw e;
		}
	}

	/**
	 * If a call requires no license and the current server state is
	 * {@link ETeamscaleServerState#LICENSE_INVALID} we do not know if we can update
	 * it to {@link ETeamscaleServerState#REACHABLE}, therefore it stays
	 * {@link ETeamscaleServerState#LICENSE_INVALID}.
	 */
	private static ETeamscaleServerState deduceNextServerState(ETeamscaleServerState previousState,
			boolean callRequiresLicense) {
		if (!callRequiresLicense && previousState == ETeamscaleServerState.LICENSE_INVALID) {
			return ETeamscaleServerState.LICENSE_INVALID;
		}
		return ETeamscaleServerState.REACHABLE;
	}

	/**
	 * Creates the User-Agent string from the parameters.
	 *
	 * @param product
	 *            the identifier of the agent. If empty or null will be substituted
	 *            with "unknown"
	 * @param version
	 *            the version of the product. If empty or null will be substituted
	 *            with "unknown"
	 * @param comment
	 *            more information about the product. Can be empty or null
	 * @return the correctly formatted User-Agent string
	 * @see <a href=
	 *      "https://www.rfc-editor.org/rfc/rfc9110.html#name-user-agent">RFC 9110:
	 *      User-Agent header</a>
	 */
	public static String buildUserAgent(@Nullable String product, @Nullable String version, @Nullable String comment) {
		if (Strings.isNullOrEmpty(product)) {
			product = "unknown";
		}
		if (Strings.isNullOrEmpty(version)) {
			version = "unknown";
		}
		if (Strings.isNullOrEmpty(comment)) {
			return String.format("%s/%s", product, version);
		}
		return String.format("%s/%s (%s)", product, version, comment);
	}

	/**
	 * Creates the User-Agent string for the Java VM. E.g. "Java/11.0.11 (OpenJDK
	 * 64-Bit Server VM)"
	 *
	 * @return the correctly formatted User-Agent string for the Java VM
	 * @see <a href=
	 *      "https://www.rfc-editor.org/rfc/rfc9110.html#name-user-agent">RFC 9110:
	 *      User-Agent header</a>
	 */
	public static String buildJavaUserAgent() {
		return buildUserAgent("Java", System.getProperty("java.version"), System.getProperty("java.vm.name"));
	}
}
