/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings;

import static com.google.common.collect.ImmutableSortedSet.toImmutableSortedSet;
import static java.util.Comparator.naturalOrder;
import static java.util.Objects.requireNonNull;

import java.io.IOException;
import java.util.Collection;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Predicate;
import java.util.stream.Stream;

import org.conqat.engine.commons.util.JsonUtils;
import org.conqat.engine.core.core.ConQATException;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.google.common.collect.ImmutableSortedSet;
import com.teamscale.ide.commons.lang.ToStringHelpers;
import com.teamscale.ide.commons.model.RemoteResourceReference;
import com.teamscale.ide.commons.model.ResourceReference;

/** Finding representation used by IDE plugins **/
public final class Finding implements IIdeIdentifiableFinding {

	/** The name of the JSON property name for {@link #id}. */
	private static final String ID_PROPERTY = "id";

	/** The name of the JSON property name for {@link #assessment}. */
	private static final String ASSESSMENT_PROPERTY = "assessment";

	/** The name of the JSON property name for {@link #message}. */
	private static final String MESSAGE_PROPERTY = "message";

	/** The name of the JSON property name for {@link #group}. */
	private static final String GROUP_PROPERTY = "group";

	/** The name of the JSON property name for {@link #category}. */
	private static final String CATEGORY_PROPERTY = "category";

	/** The name of the JSON property name for {@link #typeId}. */
	private static final String TYPE_ID_PROPERTY = "typeId";

	/** The name of the JSON property name for {@link #baseBranch}. */
	private static final String BASE_BRANCH_PROPERTY = "baseBranch";

	/** The name of the JSON property name for {@link #localResource}. */
	private static final String LOCAL_RESOURCE_PROPERTY = "localResource";

	/** The name of the JSON property name for {@link #remoteResource}. */
	private static final String REMOTE_RESOURCE_PROPERTY = "remoteResource";

	/** The name of the JSON property name for {@link #primaryLineRange}. */
	private static final String PRIMARY_LINE_RANGE_PROPERTY = "primaryLineRange";

	/** The name of the JSON property name for {@link #secondaryLineRanges}. */
	private static final String SECONDARY_LINE_RANGES_PROPERTY = "secondaryLineRanges";

	/** The name of the JSON property name for {@link #findingFlaggingBranch}. */
	private static final String FINDING_FLAGGING_BRANCH_PROPERTY = "findingFlaggingBranch";

	/** The name of the JSON property name for {@link #isPreCommitFinding}. */
	private static final String IS_PRE_COMMIT_FINDING_PROPERTY = "isPreCommitFinding";

	@JsonProperty(ID_PROPERTY)
	private final String id;

	@JsonProperty(ASSESSMENT_PROPERTY)
	private final EAssessment assessment;

	/** A human-readable message describing this finding. */
	@JsonProperty(MESSAGE_PROPERTY)
	private final String message;

	/** The (human-readable) group this finding belongs to. */
	@JsonProperty(GROUP_PROPERTY)
	private final String group;

	/** The (human-readable) category this finding belongs to. */
	@JsonProperty(CATEGORY_PROPERTY)
	private final String category;

	/** A unique ID for the type of finding. */
	@JsonProperty(TYPE_ID_PROPERTY)
	private final String typeId;

	/**
	 * The base branch of this finding. {@code null} if this finding was introduced
	 * on a pre-commit branch.
	 */
	@JsonProperty(BASE_BRANCH_PROPERTY)
	private final String baseBranch;

	/**
	 * The local resource in which this finding was found (but not necessarily
	 * introduced)
	 */
	@JsonProperty(LOCAL_RESOURCE_PROPERTY)
	private final ResourceReference localResource;

	/**
	 * The remote resource in which this finding was found (but not necessarily
	 * introduced).
	 */
	@JsonProperty(REMOTE_RESOURCE_PROPERTY)
	private final RemoteResourceReference remoteResource;

	/** The line range primarily affected by this finding. */
	@JsonProperty(PRIMARY_LINE_RANGE_PROPERTY)
	private final LineRange primaryLineRange;

	/**
	 * The line ranges (within the same {@linkplain #localResource resource}) also
	 * affected by this finding, if any.
	 */
	@JsonProperty(SECONDARY_LINE_RANGES_PROPERTY)
	private final ImmutableSortedSet<LineRange> secondaryLineRanges;

	/**
	 * The branch where this finding should be marked as tolerated/false positive if
	 * necessary.
	 */
	@JsonProperty(FINDING_FLAGGING_BRANCH_PROPERTY)
	private final String findingFlaggingBranch;
	/**
	 * Whether this finding was created by a pre-commit analysis or existed before
	 * (i.e., introduced by a "real" commit).
	 */
	@JsonProperty(IS_PRE_COMMIT_FINDING_PROPERTY)
	private final boolean isPreCommitFinding;

	@JsonCreator
	public Finding(@JsonProperty(ID_PROPERTY) String id, @JsonProperty(ASSESSMENT_PROPERTY) EAssessment assessment,
			@JsonProperty(MESSAGE_PROPERTY) String message, @JsonProperty(GROUP_PROPERTY) String group,
			@JsonProperty(CATEGORY_PROPERTY) String category, @JsonProperty(TYPE_ID_PROPERTY) String typeId,
			@JsonProperty(BASE_BRANCH_PROPERTY) String baseBranch,
			@JsonProperty(LOCAL_RESOURCE_PROPERTY) ResourceReference localResource,
			@JsonProperty(REMOTE_RESOURCE_PROPERTY) RemoteResourceReference remoteResource,
			@JsonProperty(PRIMARY_LINE_RANGE_PROPERTY) LineRange primaryLineRange,
			@JsonProperty(SECONDARY_LINE_RANGES_PROPERTY) Collection<LineRange> secondaryLineRanges,
			@JsonProperty(FINDING_FLAGGING_BRANCH_PROPERTY) String findingFlaggingBranch,
			@JsonProperty(IS_PRE_COMMIT_FINDING_PROPERTY) boolean isPreCommitFinding) {
		this.id = requireNonNull(id);
		this.assessment = requireNonNull(assessment);
		this.message = requireNonNull(message);
		this.group = requireNonNull(group);
		this.category = requireNonNull(category);
		this.typeId = requireNonNull(typeId);
		this.baseBranch = baseBranch;
		this.localResource = requireNonNull(localResource);
		this.remoteResource = requireNonNull(remoteResource);
		this.primaryLineRange = requireNonNull(primaryLineRange);
		this.secondaryLineRanges = ImmutableSortedSet.copyOf(secondaryLineRanges);
		this.findingFlaggingBranch = findingFlaggingBranch;
		this.isPreCommitFinding = isPreCommitFinding;
	}

	@Override
	public String getId() {
		return id;
	}

	public EAssessment getAssessment() {
		return assessment;
	}

	/** @see #message */
	public String getMessage() {
		return message;
	}

	/** @see #group */
	public String getGroup() {
		return group;
	}

	/** @see #category */
	public String getCategory() {
		return category;
	}

	/** @see #typeId */
	public String getTypeId() {
		return typeId;
	}

	/** @see #baseBranch */
	public Optional<String> getBaseBranch() {
		return Optional.ofNullable(baseBranch);
	}

	/** @see #localResource */
	public ResourceReference getLocalResource() {
		return localResource;
	}

	/** @see #remoteResource */
	@Override
	public RemoteResourceReference getRemoteResource() {
		return remoteResource;
	}

	/** @see #primaryLineRange */
	public LineRange getPrimaryLineRange() {
		return primaryLineRange;
	}

	/** @see #secondaryLineRanges */
	public ImmutableSortedSet<LineRange> getSecondaryLineRanges() {
		return secondaryLineRanges;
	}

	/** Returns the line ranges to highlight, e.g., in an editor's sidebar. */
	public ImmutableSortedSet<LineRange> getHighlightedLineRanges() {
		ImmutableSortedSet<LineRange> specifiedLineRanges = Stream
				.concat(Stream.of(primaryLineRange), secondaryLineRanges.stream()) //
				.filter(LineRange::isSpecified).collect(toImmutableSortedSet(naturalOrder()));
		if (!specifiedLineRanges.isEmpty()) {
			return specifiedLineRanges;
		}
		return ImmutableSortedSet.of(LineRange.unspecified());
	}

	/**
	 * Returns the line range to jump to when a finding is to be opened in an
	 * editor. This is always one of the {@linkplain #getHighlightedLineRanges()
	 * line ranges to highlight}.
	 */
	public LineRange getJumpLineRange() {
		if (primaryLineRange.isSpecified()) {
			return primaryLineRange;
		}
		ImmutableSortedSet<LineRange> highlightedLineRanges = getHighlightedLineRanges();
		if (!highlightedLineRanges.isEmpty()) {
			return highlightedLineRanges.first();
		}
		return LineRange.unspecified();
	}

	/**
	 * Returns a predicate matching all findings that have a given
	 * {@linkplain #getAssessment() assessment}.
	 */
	public static Predicate<Finding> hasAssessment(EAssessment assessment) {
		return finding -> finding.getAssessment() == assessment;
	}

	/**
	 * Serializes the given finding as JSON. The serialization format is
	 * <strong>not</strong> guaranteed to be compatible across versions, so be
	 * careful when persisting serialized findings across IDE re-starts.
	 */
	public static String toJson(Finding finding) {
		return JsonUtils.serializeToJSON(finding);
	}

	/**
	 * Deserializes a finding from the given {@linkplain #toJson(Finding) JSON}.
	 *
	 * @throws IOException
	 *             on errors during deserialization
	 */
	public static Finding fromJson(String json) throws IOException {
		try {
			return JsonUtils.deserializeFromJson(json, Finding.class);
		} catch (ConQATException e) {
			throw new IOException(e);
		}
	}

	/**
	 * Returns the branch where this finding should be marked as tolerated/false
	 * positive if necessary.
	 */
	public String getFindingFlaggingBranch() {
		return findingFlaggingBranch;
	}

	/**
	 * Returns whether this finding was created by a pre-commit analysis or existed
	 * before (i.e., introduced by a "real" commit).
	 */
	public boolean isPreCommitFinding() {
		return isPreCommitFinding;
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		Finding that = (Finding) other;
		return Objects.equals(this.id, that.id) //
				&& this.assessment == that.assessment //
				&& Objects.equals(this.message, that.message) //
				&& Objects.equals(this.group, that.group) //
				&& Objects.equals(this.category, that.category) //
				&& Objects.equals(this.typeId, that.typeId) //
				&& Objects.equals(this.baseBranch, that.baseBranch) //
				&& Objects.equals(this.localResource, that.localResource) //
				&& Objects.equals(this.remoteResource, that.remoteResource) //
				&& Objects.equals(this.primaryLineRange, that.primaryLineRange) //
				&& Objects.equals(this.secondaryLineRanges, that.secondaryLineRanges)//
				&& Objects.equals(this.findingFlaggingBranch, that.findingFlaggingBranch)//
				&& this.isPreCommitFinding == that.isPreCommitFinding;
	}

	@Override
	public int hashCode() {
		return Objects.hash(id, assessment, message, group, category, typeId, baseBranch, localResource, remoteResource,
				primaryLineRange, secondaryLineRanges, findingFlaggingBranch, isPreCommitFinding);
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
