/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings;

import java.util.Objects;

import org.checkerframework.checker.nullness.qual.Nullable;

import com.teamscale.ide.commons.lang.ToStringHelpers;

/**
 * Describes the state of a fetch finding process.
 */
public abstract class FindingFetchState {

	/**
	 * The given resource has not yet been fetched in the current session. Can only
	 * transition to {@link PendingState}.
	 */
	private static final class NotYetFetchedState extends FindingFetchState {

		private static final NotYetFetchedState INSTANCE = new NotYetFetchedState();

		@Override
		public boolean hasNotBeenFetched() {
			return true;
		}
	}

	/**
	 * The given resource is currently being fetched. Can only transition to
	 * {@link SuccessfulState} or {@link FailedState}.
	 */
	private static final class PendingState extends FindingFetchState {

		private static final PendingState INSTANCE = new PendingState();

		@Override
		public boolean isPending() {
			return true;
		}
	}

	/**
	 * The given resource has already been fetched successfully. Can only transition
	 * to {@link PendingState} or {@link OutdatedState}.
	 */
	private static final class SuccessfulState extends FindingFetchState {

		private static final SuccessfulState INSTANCE = new SuccessfulState();

		@Override
		public boolean isSuccessful() {
			return true;
		}
	}

	/**
	 * The last fetch from the resource failed. Can transition only to
	 * {@link PendingState} or {@link OutdatedState}.
	 */
	private static final class FailedState extends FindingFetchState {

		private final EFindingFetchError reason;
		private final String errorMessage;

		private FailedState(EFindingFetchError reason, @Nullable String errorMessage) {
			Objects.requireNonNull(reason);
			this.reason = reason;
			this.errorMessage = errorMessage;
		}

		@Override
		public boolean isFailure() {
			return true;
		}

		@Override
		@Nullable
		public String getErrorMessage() {
			return errorMessage;
		}

		@Override
		public EFindingFetchError getErrorReason() {
			return reason;
		}

		@Override
		public int hashCode() {
			return Objects.hash(reason, errorMessage);
		}

		@Override
		public boolean equals(Object other) {
			if (other == null) {
				return false;
			}
			if (other == this) {
				return true;
			}
			if (other.getClass() != getClass()) {
				return false;
			}
			FailedState that = (FailedState) other;

			return that.getErrorReason() == getErrorReason()
					&& Objects.equals(that.getErrorMessage(), getErrorMessage());
		}
	}

	/**
	 * The last fetch from the resource is outdated. Can only transition to
	 * {@link PendingState}.
	 */
	private static final class OutdatedState extends FindingFetchState {

		private static final OutdatedState INSTANCE = new OutdatedState();

		@Override
		public boolean isOutdated() {
			return true;
		}
	}

	/**
	 * Create a {@link FindingFetchState} that indicates that the findings has not
	 * been fetched, yet.
	 */
	public static FindingFetchState notYetFetched() {
		return NotYetFetchedState.INSTANCE;
	}

	/**
	 * Create a {@link FindingFetchState} that indicates that the fetch is currently
	 * pending.
	 */
	public static FindingFetchState pending() {
		return PendingState.INSTANCE;
	}

	/**
	 * Create a {@link FindingFetchState} that indicates that the fetch was
	 * successful.
	 */
	public static FindingFetchState success() {
		return SuccessfulState.INSTANCE;
	}

	/**
	 * Create a {@link FindingFetchState} that indicates that the fetch did fail.
	 */
	public static FindingFetchState failure(EFindingFetchError reason, String message) {
		return new FailedState(reason, message);
	}

	/**
	 * Create a {@link FindingFetchState} that indicates that the findings are
	 * outdated.
	 */
	public static FindingFetchState outdated() {
		return OutdatedState.INSTANCE;
	}

	/**
	 * Checks if it never got fetched.
	 */
	public boolean hasNotBeenFetched() {
		return false;
	}

	public boolean isPending() {
		return false;
	}

	public boolean isSuccessful() {
		return false;
	}

	public boolean isFailure() {
		return false;
	}

	public boolean isOutdated() {
		return false;
	}

	/**
	 * Returns the reason for the failure. Must only be called if the fetch
	 * {@linkplain #isFailure() failed}.
	 */
	public EFindingFetchError getErrorReason() {
		throw new IllegalStateException("Finding fetch has not failed.");
	}

	/**
	 * Returns the message of the failure. Must only be called if the fetch
	 * {@linkplain #isFailure() failed}.
	 */
	@Nullable
	public String getErrorMessage() {
		throw new IllegalStateException("Finding fetch has not failed.");
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}

	@Override
	public int hashCode() {
		return getClass().hashCode();
	}

	@Override
	public boolean equals(Object obj) {
		return obj != null && getClass() == obj.getClass();
	}
}
