/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.conqat.engine.commons.findings.location.ElementLocation;
import org.conqat.engine.commons.findings.location.TextRegionLocation;
import org.conqat.engine.index.shared.TrackedFinding;
import org.conqat.lib.commons.assessment.ETrafficLightColor;
import org.conqat.lib.commons.uniformpath.UniformPath;
import org.conqat.lib.commons.uniformpath.UniformPathCompatibilityUtil;

import com.google.common.base.Preconditions;
import com.teamscale.ide.commons.model.RemoteResourceReference;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.precommit.PreCommitUtils;

/** Utility methods associated with findings. */
public class FindingUtils {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(FindingUtils.class);

	/**
	 * Converts a list of {@linkplain TrackedFinding finding DTOs} into a list of
	 * {@linkplain Finding findings} used by IDE plugins, if possible.
	 */
	public static List<Finding> extractFindings(List<TrackedFinding> findingDtos, String baseBranch,
			ResourceReference localResource, RemoteResourceReference remoteResource) {
		List<Finding> findings = new ArrayList<>();
		for (TrackedFinding findingDto : findingDtos) {
			extractFinding(findingDto, baseBranch, localResource, remoteResource).ifPresent(findings::add);
		}
		return findings;
	}

	/**
	 * Converts the {@link TrackedFinding} Data Transfer Object into a
	 * {@link Finding} used by IDE plugins, if possible.
	 */
	public static Optional<Finding> extractFinding(TrackedFinding findingDto, String baseBranch,
			ResourceReference localResource, RemoteResourceReference remoteResource) {
		LOGGER.traceEntry(null, findingDto, baseBranch, localResource, remoteResource);
		try {
			String id = findingDto.getId();
			EAssessment assessment = extractAssessmentFromColor(findingDto.getAssessment());
			String message = findingDto.getMessage();
			String group = findingDto.getGroupName();
			String category = findingDto.getCategoryName();
			String typeId = findingDto.getTypeId();

			checkLocation(findingDto.getLocation(), remoteResource);
			String findingFlaggingBranch = baseBranch;
			boolean isPreCommitFinding = false;
			if (PreCommitUtils.isPreCommitBranch(findingDto.getBirthCommit().getBranchName())) {
				// Findings introduced on the pre-commit branch do not have a base branch
				baseBranch = null;
				findingFlaggingBranch = findingDto.getBirthCommit().getBranchName();
				isPreCommitFinding = true;
			}

			LineRange primaryLineRange = extractLineRangeFromLocation(findingDto.getLocation(), remoteResource);
			List<LineRange> secondaryLineRanges = extractLineRangesFromLocations(findingDto.getSecondaryLocations(),
					remoteResource);

			Finding finding = new Finding(id, assessment, message, group, category, typeId, baseBranch, localResource,
					remoteResource, primaryLineRange, secondaryLineRanges, findingFlaggingBranch, isPreCommitFinding);
			return LOGGER.traceExit(Optional.of(finding));
		} catch (IllegalArgumentException e) {
			LOGGER.catching(e);
			LOGGER.warn("Could not extract finding from DTO: {}", findingDto);
			return LOGGER.traceExit(Optional.empty());
		}
	}

	private static EAssessment extractAssessmentFromColor(ETrafficLightColor colorAssessment) {
		Preconditions.checkArgument(colorAssessment != null, "Null assessment");

		switch (colorAssessment) {
		case RED:
			return EAssessment.RED;
		case YELLOW:
			return EAssessment.YELLOW;
		default:
			throw new IllegalArgumentException("Unexpected assessment: " + colorAssessment);
		}
	}

	private static List<LineRange> extractLineRangesFromLocations(List<ElementLocation> locations,
			RemoteResourceReference remoteReference) {
		return locations.stream().map(location -> extractLineRangeFromLocation(location, remoteReference))
				.collect(Collectors.toList());
	}

	/**
	 * Builds a {@link LineRange} that represents the given {@link ElementLocation}.
	 * If the location is a {@link TextRegionLocation}, we build a line-precise
	 * representation ({@link LineRange}).
	 * <p>
	 * We can't build an offset-precise representation, however, since Teamscale's
	 * offsets in {@link TextRegionLocation} are based on files having
	 * single-character (Unix) line endings. That may or may not be case for the
	 * actual files present in the repository or locally. For example, because the
	 * file has multi-character (Windows) or even mixed line endings in the IDE.
	 */
	private static LineRange extractLineRangeFromLocation(ElementLocation location,
			RemoteResourceReference remoteReference) {
		checkLocation(location, remoteReference);

		if (!(location instanceof TextRegionLocation)) {
			return LineRange.unspecified();
		}

		TextRegionLocation textRegionLocation = (TextRegionLocation) location;
		int startLine = textRegionLocation.getRawStartLine();
		int endLine = textRegionLocation.getRawEndLine();

		if (startLine <= 0 || endLine <= 0) {
			return LineRange.unspecified();
		}

		return LineRange.specified(startLine, endLine);
	}

	/** Sanity-check the location reported by the Teamscale server. */
	private static void checkLocation(ElementLocation location, RemoteResourceReference remoteReference) {
		UniformPath actualUniformPath = UniformPathCompatibilityUtil.convert(location.getUniformPath());
		UniformPath expectedUniformPath = remoteReference.getUniformPath();

		Preconditions.checkArgument(actualUniformPath.equals(expectedUniformPath),
				"Expected uniform path %s but was %s", expectedUniformPath, actualUniformPath);
	}
}
