/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings;

import java.util.Comparator;
import java.util.Objects;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.google.common.base.Preconditions;
import com.teamscale.ide.commons.lang.ToStringHelpers;

/** A closed range of lines. */
public final class LineRange implements Comparable<LineRange> {

	/** The number indicating that a finding does not have an associated line */
	public static final int UNSPECIFIED = 0;

	/** The name of the JSON property name for {@link #firstLine}. */
	private static final String FIRST_LINE_PROPERTY = "firstLine";

	/** The name of the JSON property name for {@link #lastLine}. */
	private static final String LAST_LINE_PROPERTY = "lastLine";

	/**
	 * The first line part of this range (inclusive, one-based) or
	 * {@link #UNSPECIFIED} for {@linkplain #unspecified() unspecified ranges}.
	 */
	@JsonProperty(FIRST_LINE_PROPERTY)
	private final int firstLine;

	/**
	 * The last line part of this range (inclusive, one-based) or
	 * {@link #UNSPECIFIED} for {@linkplain #unspecified() unspecified ranges}.
	 */
	@JsonProperty(LAST_LINE_PROPERTY)
	private final int lastLine;

	/**
	 * Creates a line range from {@code firstLine} to {@code lastLine} (inclusive,
	 * one-based).
	 */
	public static LineRange specified(int firstLine, int lastLine) {
		Preconditions.checkArgument(firstLine > UNSPECIFIED, "First line must be positive: %s", firstLine);
		Preconditions.checkArgument(lastLine > UNSPECIFIED, "Last line must be positive: %s", lastLine);
		Preconditions.checkArgument(lastLine >= firstLine, "Last line must be not be less than first line: %s < %s",
				lastLine, firstLine);

		return new LineRange(firstLine, lastLine);
	}

	/**
	 * Creates an unspecified line range, implicitly spanning an entire file.
	 */
	public static LineRange unspecified() {
		return new LineRange(UNSPECIFIED, UNSPECIFIED);
	}

	@JsonCreator
	private LineRange(@JsonProperty(FIRST_LINE_PROPERTY) int firstLine,
			@JsonProperty(LAST_LINE_PROPERTY) int lastLine) {
		Preconditions.checkArgument(firstLine >= UNSPECIFIED, "First line must be non-negative: %s", firstLine);
		Preconditions.checkArgument(lastLine >= UNSPECIFIED, "Last line must be non-negative: %s", lastLine);

		this.firstLine = firstLine;
		this.lastLine = lastLine;
	}

	/** @see #firstLine */
	public int getFirstLine() {
		return firstLine;
	}

	/** @see #lastLine */
	public int getLastLine() {
		return lastLine;
	}

	/**
	 * Returns whether this range spans an {@linkplain #unspecified() unspecified}
	 * range of lines.
	 */
	public boolean isUnspecified() {
		return firstLine == UNSPECIFIED && lastLine == UNSPECIFIED;
	}

	/**
	 * Returns whether this range spans a {@linkplain #specified(int, int)
	 * specified} range of lines.
	 */
	public boolean isSpecified() {
		return !isUnspecified();
	}

	/** Returns whether this range spans a single line only. */
	public boolean isSingleLine() {
		return isSpecified() && firstLine == lastLine;
	}

	@Override
	public int compareTo(LineRange that) {
		return Comparator.comparingInt(LineRange::getFirstLine).thenComparing(LineRange::getLastLine).compare(this,
				that);
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		LineRange lineRange = (LineRange) other;
		return firstLine == lineRange.firstLine && lastLine == lineRange.lastLine;
	}

	@Override
	public int hashCode() {
		return Objects.hash(firstLine, lastLine);
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
