package com.teamscale.ide.commons.findings;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.stream.Collectors;

import org.checkerframework.checker.nullness.qual.Nullable;
import org.conqat.engine.index.shared.PreCommit3Result;
import org.conqat.engine.index.shared.TrackedFinding;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableCollection;
import com.google.common.collect.ImmutableMultimap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;
import com.teamscale.ide.commons.lang.ToStringHelpers;

/**
 * The IDE representation of the dto {@link PreCommit3Result}
 */
public class PreCommitResult {

	/**
	 * The token to poll for final results. If this is provided, you should poll for
	 * finished results later on. If this is null, the returned results are final
	 * and not further analysis will happen.
	 */
	private final @Nullable String token;

	/**
	 * Detailed descriptions/classifications of global errors that occurred.
	 * <p>
	 * The {@link PreCommitErrorDetail#affectedUniformPath path of the error} is
	 * always {@code null}.
	 */
	private final ImmutableSet<PreCommitErrorDetail> globalErrors;

	private final ImmutableSet<String> allPaths;
	private final ImmutableSet<String> errorPaths;
	private final ImmutableSet<String> findingPaths;

	private final ImmutableMultimap<String, PreCommitErrorDetail> pathToErrors;
	private final ImmutableMultimap<String, TrackedFinding> pathToFindings;

	private PreCommitResult(@Nullable String token, ImmutableSet<PreCommitErrorDetail> globalErrors,
			ImmutableSet<String> allPaths, ImmutableSet<String> errorPaths, ImmutableSet<String> findingPaths,
			ImmutableMultimap<String, PreCommitErrorDetail> pathToErrors,
			ImmutableMultimap<String, TrackedFinding> pathToFindings) {
		this.token = token;
		this.globalErrors = globalErrors;
		this.allPaths = allPaths;
		this.errorPaths = errorPaths;
		this.findingPaths = findingPaths;
		this.pathToErrors = pathToErrors;
		this.pathToFindings = pathToFindings;

	}

	/**
	 * Returns the token of this result.
	 * <p>
	 * It is only possible to get the token if the result is not yet
	 * {@link #isFinished() finished}.
	 */
	public String getToken() {
		Preconditions.checkState(!isFinished(), "A finished result can not have a token.");
		return token;
	}

	/**
	 * Returns true, if the pre-commit result represent finished analysis results.
	 */
	public boolean isFinished() {
		return token == null;
	}

	/**
	 * Returns true, if the pre-commit analysis is pending.
	 */
	public boolean isPending() {
		return token != null;
	}

	/**
	 * Returns true, if the pre-commit result represent successfully finished
	 * analysis results.
	 */
	public boolean isSuccessful() {
		// No findingsPaths indicates that all path have an invalidating local error.
		return !hasGlobalErrors() && token == null && !findingPaths.isEmpty();
	}

	public ImmutableCollection<String> getAllAnalyzedPaths() {
		return allPaths;
	}

	/**
	 * Returns true, if the pre-commit result could not be fetched due to an error.
	 */
	public boolean hasGlobalErrors() {
		return globalErrors.stream().anyMatch(detail -> detail.resultsInvalid);
	}

	/**
	 * Returns true, if the pre-commit result has any kind of error including those
	 * which not invalidate the result.
	 */
	public boolean hasErrors() {
		return !globalErrors.isEmpty() || !errorPaths.isEmpty();
	}

	public ImmutableCollection<PreCommitErrorDetail> getGlobalErrorDetails() {
		return globalErrors;
	}

	public ImmutableCollection<String> getAnalyzedPathsWithErrors() {
		return errorPaths;
	}

	/**
	 * Returns the errors of the specific path.
	 */
	public ImmutableCollection<PreCommitErrorDetail> getErrors(String analyzedPath) {
		return pathToErrors.get(analyzedPath);
	}

	public ImmutableCollection<String> getAnalyzedPathsWithFindings() {
		return findingPaths;
	}

	/**
	 * Returns the findings of the specific path.
	 */
	public ImmutableCollection<TrackedFinding> getFindings(String analyzedPath) {
		return pathToFindings.get(analyzedPath);
	}

	/**
	 * Returns all findings of this result.
	 */
	public ImmutableCollection<TrackedFinding> getAllFindings() {
		return pathToFindings.values();
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}

	/**
	 * Returns all detailed errors in this response of the precommit 3 service. This
	 * string is not meant for user consumption. More for logging.
	 */
	public String getSquashedGlobalErrors() {
		return globalErrors.stream().map(errorDetail -> errorDetail.message).collect(Collectors.joining("\n"));
	}

	/**
	 * Returns all detailed errors in this response of the precommit 3 service. This
	 * string is not meant for user consumption. More for logging.
	 */
	public String getSquashedErrors(String analyzedPath) {
		ImmutableCollection<PreCommitErrorDetail> errors = getErrors(analyzedPath);
		return errors.stream().map(errorDetail -> errorDetail.message).collect(Collectors.joining("\n"));
	}

	/**
	 * Detailed information on one error that occurred during pre-commit analysis.
	 * Depending on the information in this object clients (IDE plugins) may decide
	 * to display the error more or less visible or hide it completely.
	 */
	public static class PreCommitErrorDetail {
		/**
		 * Error type for errors that were returned in a response from precommit3
		 * service call from a pre-8.7.x Teamscale server.
		 * <p>
		 * All errors are squashed in one string. Each error is prefixed by "* ".
		 */
		public static final String LEGACY_ERROR_TYPE = "LEGACY_PRE_COMMIT_ERRORS";

		/**
		 * Error type for a {@link PreCommitResult} that was generated on the IDE side
		 * (in the service client) if the Teamscale Server could not be reached.
		 */
		public static final String SERVER_NOT_REACHABLE = "SERVER_NOT_REACHABLE";

		/**
		 * Error type for a {@link PreCommitResult} that was generated on the IDE side
		 * (in the service client) if the Teamscale Server does not support pre-commit3.
		 */
		public static final String PRE_COMMIT_API_NOT_SUPPORTED = "PRE_COMMIT_NOT_SUPPORTED";

		/**
		 * The type of this error.
		 * <p>
		 * It is {@link PreCommit3Result.EPrecommit3ErrorType#toString()} of one of
		 * {@link PreCommit3Result.EPrecommit3ErrorType}, {@link #LEGACY_ERROR_TYPE},
		 * {@link #SERVER_NOT_REACHABLE}, or a type that is not yet known in the current
		 * IDE plugin-version (but known on the Teamscale Server).
		 */
		private final String type;

		/**
		 * The message of this error. Intended for display to the user.
		 */
		private final String message;

		/**
		 * Uniform path that is affected by this error (or null if it is an error that
		 * affects the entire pre-commit analysis).
		 */
		@Nullable
		private final String affectedUniformPath;

		/**
		 * If true this error causes results in this pre-commit-result object to be
		 * invalid, for example because the analysis was aborted.
		 * <p>
		 * If {@link #affectedUniformPath} is non-null, this could abort only the
		 * analysis of that file. Other files might still be analyzed.
		 * <p>
		 * If {@link #affectedUniformPath} is null (i.e., this is a file-independent
		 * problem such as "pre-commit analysis is disabled on server"), then the entire
		 * analysis request is aborted.
		 */
		private final boolean resultsInvalid;

		public PreCommitErrorDetail(String type, String message, @Nullable String affectedUniformPath,
				boolean resultsInvalid) {
			this.type = type;
			this.message = message;
			this.affectedUniformPath = affectedUniformPath;
			this.resultsInvalid = resultsInvalid;
		}

		public String getType() {
			return type;
		}

		public String getMessage() {
			return message;
		}

		public @Nullable String getAffectedUniformPath() {
			return affectedUniformPath;
		}

		/**
		 * Returns whether this error invalidates the result of the
		 * {@link #getAffectedUniformPath() affected path}.
		 */
		public boolean invalidatesResults() {
			return resultsInvalid;
		}
	}

	/**
	 * Creates a builder for a result that could be pending.
	 * <p>
	 * If the token is {@code null} then the result is finished.
	 */
	public static PreCommitResultBuilder pending(@Nullable String token) {
		return new PreCommitResultBuilder(token);
	}

	/**
	 * Creates a builder for a result that is finished.
	 */
	public static PreCommitResultBuilder finished() {
		return new PreCommitResultBuilder(null);
	}

	/**
	 * A builder to create {@link PreCommitResult PreCommitResults}.
	 */
	public static class PreCommitResultBuilder {

		@Nullable
		private final String token;

		private final ImmutableSet.Builder<String> allPaths;
		private final ImmutableSet.Builder<PreCommitErrorDetail> globalErrors;

		private final ImmutableMultimap.Builder<String, TrackedFinding> pathToFindings;
		private final ImmutableMultimap.Builder<String, PreCommitErrorDetail> pathToErrors;

		private PreCommitResultBuilder(@Nullable String token) {
			this.token = token;
			allPaths = new ImmutableSet.Builder<>();
			globalErrors = new ImmutableSet.Builder<>();
			pathToFindings = new ImmutableMultimap.Builder<>();
			pathToErrors = new ImmutableMultimap.Builder<>();
		}

		/**
		 * Adds all errors to the result and automatically determines if a error is a
		 * global or local error.
		 *
		 * @see #addGlobalErrors(Collection)
		 * @see #addLocalErrors(Collection)
		 */
		public PreCommitResultBuilder addErrors(@Nullable Collection<PreCommitErrorDetail> errors) {
			if (errors == null) {
				return this;
			}
			for (PreCommitErrorDetail error : errors) {
				if (error.getAffectedUniformPath() == null) {
					addGlobalError(error);
				} else {
					addLocalError(error);
				}
			}
			return this;
		}

		/** @see #addGlobalErrors(Collection) */
		public PreCommitResultBuilder addGlobalError(String type, String message, boolean resultsInvalid) {
			return addGlobalError(new PreCommitErrorDetail(type, message, null, resultsInvalid));
		}

		/** @see #addGlobalErrors(Collection) */
		public PreCommitResultBuilder addGlobalError(PreCommitErrorDetail errorDetail) {
			return addGlobalErrors(Collections.singleton(errorDetail));
		}

		/**
		 * Adds all errors as global errors to the result.
		 * <p>
		 * A global error needs to have no
		 * {@link PreCommitErrorDetail#getAffectedUniformPath() path}. This means the
		 * path needs to be {@code null}.
		 */
		public PreCommitResultBuilder addGlobalErrors(@Nullable Collection<PreCommitErrorDetail> errors) {
			if (errors == null) {
				return this;
			}
			Preconditions.checkArgument(errors.stream().allMatch(error -> error.getAffectedUniformPath() == null),
					"Global errors can not have a affected uniform path.");
			globalErrors.addAll(errors);
			return this;
		}

		/** @see #addLocalErrors(Collection) */
		public PreCommitResultBuilder addLocalError(String type, String message, String affectedUniformPath,
				boolean resultsInvalid) {
			return addLocalError(new PreCommitErrorDetail(type, message, affectedUniformPath, resultsInvalid));
		}

		/** @see #addLocalErrors(Collection) */
		public PreCommitResultBuilder addLocalError(PreCommitErrorDetail errorDetail) {
			return addLocalErrors(Collections.singleton(errorDetail));
		}

		/**
		 * Adds all errors as local errors to the result.
		 * <p>
		 * A local error needs to have a
		 * {@link PreCommitErrorDetail#getAffectedUniformPath() path}.
		 */
		public PreCommitResultBuilder addLocalErrors(@Nullable Collection<PreCommitErrorDetail> errors) {
			if (errors == null) {
				return this;
			}
			for (PreCommitErrorDetail error : errors) {
				Preconditions.checkNotNull(error.getAffectedUniformPath(),
						"Local errors need to have a affected uniform path.");
				pathToErrors.put(error.getAffectedUniformPath(), error);
			}
			return this;
		}

		/** @see #addFindings(Collection) */
		public PreCommitResultBuilder addFinding(TrackedFinding finding) {
			return addFindings(Collections.singleton(finding));
		}

		/**
		 * Adds all findings to the result.
		 */
		public PreCommitResultBuilder addFindings(@Nullable Collection<TrackedFinding> findings) {
			if (findings == null) {
				return this;
			}
			for (TrackedFinding finding : findings) {
				pathToFindings.put(finding.getLocation().getUniformPath(), finding);
			}
			return this;
		}

		/**
		 * Adds the paths to the result.
		 * <p>
		 * If a path has nether {@link PreCommitResult#getFindings(String) findings} nor
		 * {@link PreCommitResult#getErrors(String) errors} then these paths are used to
		 * indicate that this result has no findings.
		 */
		public PreCommitResultBuilder addGeneralPaths(@Nullable Collection<String> paths) {
			if (paths == null) {
				return this;
			}
			allPaths.addAll(paths);
			return this;
		}

		/**
		 * Returns the resulting {@link PreCommitResult} that matches this builder.
		 */
		public PreCommitResult build() {
			ImmutableMultimap<String, TrackedFinding> pathToFindings = this.pathToFindings.build();
			ImmutableMultimap<String, PreCommitErrorDetail> pathToErrors = this.pathToErrors.build();

			ImmutableSet<String> errorPaths = pathToErrors.entries().stream()
					.filter(entry -> entry.getValue().invalidatesResults()).map(Map.Entry::getKey)
					.collect(ImmutableSet.toImmutableSet());

			// Paths without explicit findings or errors are results with zero findings
			ImmutableSet<String> findingsPaths = new ImmutableSet.Builder<String>().addAll(pathToFindings.keySet())
					.addAll(Sets.filter(allPaths.build(), path -> !errorPaths.contains(path))).build();

			allPaths.addAll(errorPaths).addAll(findingsPaths);

			return new PreCommitResult(token, globalErrors.build(), allPaths.build(), errorPaths, findingsPaths,
					pathToErrors, pathToFindings);
		}
	}
}
