/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings;

import static java.util.Objects.requireNonNull;

import java.util.Objects;

import com.teamscale.ide.commons.lang.ToStringHelpers;
import com.teamscale.ide.commons.model.RemoteResourceReference;
import com.teamscale.ide.commons.model.ResourceReference;

/**
 * A local resource (or a tombstone thereof, in case it has been removed)
 * corresponding to a remote resource and scheduled for upload during pre-commit
 * 3.0 analysis.
 */
public final class PreCommitUploadItem {

	/** The local resource that should be uploaded. */
	private final ResourceReference localResource;

	/** The respective remote resource. */
	private final RemoteResourceReference remoteResource;

	/** The content hash for the code content. */
	private final String contentHash;

	public PreCommitUploadItem(ResourceReference localResource, RemoteResourceReference remoteResource,
			String contentHash) {
		this.localResource = requireNonNull(localResource);
		this.remoteResource = requireNonNull(remoteResource);
		this.contentHash = requireNonNull(contentHash);
	}

	/** The resource to be uploaded. */
	public ResourceReference getLocalResource() {
		return localResource;
	}

	/**
	 * The remote resource corresponding to the {@linkplain #getLocalResource()
	 * uploaded resource}.
	 */
	public RemoteResourceReference getRemoteResource() {
		return remoteResource;
	}

	/** see {@link #contentHash} */
	public String getContentHash() {
		return contentHash;
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		PreCommitUploadItem that = (PreCommitUploadItem) other;
		return Objects.equals(this.localResource, that.localResource) //
				&& Objects.equals(this.remoteResource, that.remoteResource) //
				&& Objects.equals(this.contentHash, that.contentHash);
	}

	@Override
	public int hashCode() {
		return Objects.hash(localResource);
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
