/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings;

import java.net.URI;
import java.util.Objects;

import com.teamscale.ide.commons.lang.ToStringHelpers;

/**
 * A combination of Teamscale server, project and branch at which the pre-commit
 * request will be rooted. Together they uniquely identify a pre-commit upload
 * target.
 */
public final class PreCommitUploadTarget {

	private final URI serverUrl;
	private final String projectId;
	private final String branch;

	public PreCommitUploadTarget(URI serverUrl, String projectId, String branch) {
		this.serverUrl = serverUrl;
		this.projectId = projectId;
		this.branch = branch;
	}

	/**
	 * @return the URL of the Teamscale server associated with a pre-commit
	 *         resource.
	 */
	public URI getServerUrl() {
		return serverUrl;
	}

	/**
	 * @return the Teamscale project on the {@link #serverUrl Teamscale server}
	 *         associated with a pre-commit resource.
	 */
	public String getProjectId() {
		return projectId;
	}

	/**
	 * @return the branch associated with a pre-commit resource.
	 */
	public String getBranch() {
		return branch;
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		PreCommitUploadTarget that = (PreCommitUploadTarget) other;
		return Objects.equals(this.serverUrl, that.serverUrl) && Objects.equals(this.projectId, that.projectId)
				&& Objects.equals(this.branch, that.branch);
	}

	@Override
	public int hashCode() {
		return Objects.hash(serverUrl, projectId, branch);
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
