/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings.services;

import java.util.List;
import java.util.Optional;

import org.checkerframework.checker.nullness.qual.Nullable;

import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.plugin.IPluginService;

/**
 * Keeps a cache of findings for different files with a specific content hash
 * (which may be {@code null} for non-pre-commit findings) on different
 * branches.
 */
public interface IFetchFindingsTaskCache extends IPluginService {

	/**
	 * @return Returns a list of findings, if any, for the given file on the given
	 *         branch and with the given content hash, or {@link Optional#empty()}
	 *         if no findings have been cached yet.
	 */
	Optional<? extends List<Finding>> getFindings(String branch, ResourceReference resource,
			@Nullable String contentHash);

	/**
	 * Removes a finding from the cache with the given content hash.
	 */
	void removeFinding(String branch, ResourceReference resource, @Nullable String contentHash, Finding finding);

	/**
	 * Removes a finding from the cache.
	 */
	void removeFinding(String branch, ResourceReference resource, Finding finding);

	/**
	 * Update findings of a file in the cache with the given content hash.
	 */
	void updateFindings(String branch, ResourceReference resource, @Nullable String contentHash,
			List<Finding> findings);

	/**
	 * Invalidates cache for given resource, branch, and content hash.
	 */
	void invalidate(ResourceReference resource, String branch, @Nullable String contentHash);

	/**
	 * Invalidates the cache and removes all cached findings.
	 */
	void invalidateAll();
}
