/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings.services;

import java.util.Collections;
import java.util.List;

import org.checkerframework.checker.nullness.qual.Nullable;

import com.teamscale.ide.commons.findings.EFindingFetchError;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.FindingFetchState;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.plugin.IPluginService;
import com.teamscale.ide.commons.util.IEventListenable;

/**
 * Tracks the state of all findings which get fetched.
 */
public interface IFindingFetchStateService extends IPluginService, IEventListenable<IFindingsListener> {

	/**
	 * Callback for the start of a new findings fetch.
	 *
	 * @param resource
	 *            The resource for which new findings gets fetched
	 */
	void setFindingsFetchPending(ResourceReference resource);

	/**
	 * Callback for new findings. Should start a new thread for handling the update.
	 *
	 * @param resource
	 *            The resource for which new findings are available
	 * @param findings
	 *            The new findings
	 */
	void setFindingsFetchSuccessful(ResourceReference resource, List<Finding> findings);

	/**
	 * Callback for an error in the fetch of new findings.
	 *
	 * @param resource
	 *            The resource for which the fetch failed
	 * @param error
	 *            The error type of the fetch task
	 * @param errorMessage
	 *            The error message of the fetch task
	 */
	default void setFindingsFetchFailure(ResourceReference resource, EFindingFetchError error,
			@Nullable String errorMessage) {
		setFindingsFetchFailure(resource, error, errorMessage, Collections.emptyList());
	}

	/**
	 * Callback for an error in the fetch of new findings.
	 *
	 * @param resource
	 *            The resource for which the fetch failed
	 * @param error
	 *            The error type of the fetch task
	 * @param errorMessage
	 *            The error message of the fetch task
	 * @param findingsOfPreviousCommit
	 *            The already known findings of the file. If the pre-commit is
	 *            rejected because of the upload limit the server answers with the
	 *            current findings of the file.
	 */
	void setFindingsFetchFailure(ResourceReference resource, EFindingFetchError error, @Nullable String errorMessage,
			List<Finding> findingsOfPreviousCommit);

	/**
	 * Callback for an outdated fetch of an old finding
	 *
	 * @param resource
	 *            The resource for which the fetch is outdated
	 */
	void setFindingsFetchOutdated(ResourceReference resource);

	/**
	 * Set all findings to outdated. Should only happen if the cache gets
	 * invalidated.
	 */
	void setAllFindingsFetchOutdated();

	/**
	 * Get the {@link FindingFetchState} of a specific resource
	 */
	FindingFetchState getFindingState(ResourceReference resource);

	/**
	 * Adds a listener which gets trigger by the specific resources.
	 * <p>
	 * To prevent memory leaks, a listener must be
	 * {@link #removeListener(ResourceReference, IFindingsListener) removed}, when
	 * it is not used anymore.
	 */
	void addListener(ResourceReference resource, IFindingsListener listener);

	/** Removes a listener which gets trigger by the specific resources. */
	void removeListener(ResourceReference resource, IFindingsListener listener);

	/**
	 * Adds a listener which gets trigger by all resources.
	 * <p>
	 * To prevent memory leaks, a listener must be
	 * {@link #removeListener(IFindingsListener) removed}, when it is not used
	 * anymore.
	 */
	@Override
	void addListener(IFindingsListener listener);

	/** Removes a listener which gets trigger by all resources. */
	@Override
	void removeListener(IFindingsListener listener);
}
