/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings.services;

import java.util.EventListener;
import java.util.List;

import org.checkerframework.checker.nullness.qual.Nullable;

import com.teamscale.ide.commons.findings.EFindingFetchError;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.model.ResourceReference;

/**
 * Listens and receives updates to any state changes from a
 * {@link IFindingFetchStateService}.
 * <p>
 * Every update starts with
 * {@link IFindingsListener#onFindingsFetchStarted(ResourceReference)} and ends
 * with either
 * {@link IFindingsListener#onFindingsFetchSucceeded(ResourceReference, List)}
 * or
 * {@link IFindingsListener#onFindingsFetchFailed(ResourceReference, EFindingFetchError, String, List)}.
 * <p>
 * {@link IFindingsListener#onFindingsFetchOutdated(ResourceReference)} can
 * happen at any time.
 */
public interface IFindingsListener extends EventListener {

	/**
	 * Callback for the start of a new findings fetch.
	 *
	 * @param fetchResource
	 *            The resource for which new findings gets fetched
	 */
	default void onFindingsFetchStarted(ResourceReference fetchResource) {
		// Nothing to do.
	}

	/**
	 * Callback for new findings. Should start a new thread for handling the update.
	 *
	 * @param fetchResource
	 *            The resource for which new findings are available
	 * @param findings
	 *            The new findings
	 */
	default void onFindingsFetchSucceeded(ResourceReference fetchResource, List<Finding> findings) {
		// Nothing to do.
	}

	/**
	 * Callback for an error in the fetch of new findings.
	 *
	 * @param fetchResource
	 *            The resource for which the fetch failed
	 * @param error
	 *            The error type of the fetch task
	 * @param message
	 *            The failure message of the fetch task
	 * @param findingsOfPreviousCommit
	 *            The already known findings of the file. If the pre-commit is
	 *            rejected because of the upload limit the server answers with the
	 *            current findings of the file.
	 */
	default void onFindingsFetchFailed(ResourceReference fetchResource, EFindingFetchError error,
			@Nullable String message, List<Finding> findingsOfPreviousCommit) {
		// Nothing to do.
	}

	/**
	 * Callback for an outdated fetch of an old finding
	 *
	 * @param resource
	 *            The resource for which the fetch is outdated
	 */
	default void onFindingsFetchOutdated(ResourceReference resource) {
		// Nothing to do.
	}
}
