/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings.services.impl;

import static com.google.common.base.Predicates.not;
import static com.google.common.collect.ImmutableList.toImmutableList;

import java.util.List;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.checkerframework.checker.nullness.qual.Nullable;

import com.google.common.cache.Cache;
import com.google.common.cache.CacheBuilder;
import com.google.common.collect.ImmutableList;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.services.IFetchFindingsTaskCache;
import com.teamscale.ide.commons.model.ContentHashedResourceReference;
import com.teamscale.ide.commons.model.ResourceReference;

/**
 * Caches findings for code files in order to help all finding views display
 * updated findings. A single file can have different cached findings for
 * different code states and Version Control System (VCS) branches.
 */
public class FetchFindingsTaskCache implements IFetchFindingsTaskCache {

	/** Maximum number of cached findings. */
	private static final int MAX_CACHED_FINDINGS = 1000;

	/**
	 * In-memory findings cache, limited to {@value #MAX_CACHED_FINDINGS} overall.
	 */
	private final Cache<ContentHashedResourceReference, ImmutableList<Finding>> cache = CacheBuilder.newBuilder() //
			.weigher((ContentHashedResourceReference reference, List<Finding> findings) -> findings.size()) //
			.maximumWeight(MAX_CACHED_FINDINGS) //
			.build();

	@Override
	public void updateFindings(String branch, ResourceReference resource, String contentHash, List<Finding> findings) {
		synchronized (cache) {
			ContentHashedResourceReference key = new ContentHashedResourceReference(resource, branch, contentHash);
			cache.put(key, ImmutableList.copyOf(findings));
		}
	}

	@Override
	public Optional<ImmutableList<Finding>> getFindings(String branch, ResourceReference resource, String contentHash) {
		synchronized (cache) {
			ContentHashedResourceReference key = new ContentHashedResourceReference(resource, branch, contentHash);
			return Optional.ofNullable(cache.getIfPresent(key));
		}
	}

	@Override
	public void removeFinding(String branch, ResourceReference resource, String contentHash, Finding finding) {
		synchronized (cache) {
			ImmutableList<Finding> findings = getFindings(branch, resource, contentHash).orElse(null);
			if (findings == null) {
				return;
			}

			ImmutableList<Finding> updatedFindings = findings.stream() //
					.filter(not(finding::equals)) //
					.collect(toImmutableList());
			if (findings.size() != updatedFindings.size()) {
				updateFindings(branch, resource, contentHash, updatedFindings);
			}
		}
	}

	@Override
	public void removeFinding(String branch, ResourceReference resource, Finding finding) {
		synchronized (cache) {
			for (ContentHashedResourceReference contentHashedResourceReference : cache.asMap().keySet()) {
				removeFinding(branch, resource,
						contentHashedResourceReference.getContentHash().orElse(StringUtils.EMPTY), finding);
			}
		}
	}

	@Override
	public void invalidateAll() {
		synchronized (cache) {
			cache.invalidateAll();
		}
	}

	@Override
	public void invalidate(ResourceReference resource, String branch, @Nullable String contentHash) {
		synchronized (cache) {
			ContentHashedResourceReference key = new ContentHashedResourceReference(resource, branch, contentHash);
			cache.invalidate(key);
		}
	}
}
