/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings.services.impl;

import java.util.List;
import java.util.Map;
import java.util.function.Consumer;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.checkerframework.checker.nullness.qual.Nullable;

import com.google.common.collect.LinkedListMultimap;
import com.google.common.collect.Maps;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multimaps;
import com.teamscale.ide.commons.findings.EFindingFetchError;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.FindingFetchState;
import com.teamscale.ide.commons.findings.services.IFindingFetchStateService;
import com.teamscale.ide.commons.findings.services.IFindingsListener;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.util.EventListenableBase;

/**
 * Tracks the state of all findings which get fetched.
 */
public class FindingFetchStateService extends EventListenableBase<IFindingsListener>
		implements IFindingFetchStateService {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(FindingFetchStateService.class);

	private final Map<ResourceReference, FindingFetchState> stateMap = Maps.newConcurrentMap();

	private final Multimap<ResourceReference, IFindingsListener> resourceSpecificFindingsListeners = Multimaps
			.synchronizedMultimap(LinkedListMultimap.create());

	@Override
	public void setFindingsFetchPending(ResourceReference resource) {
		LOGGER.traceEntry(null, resource);

		LOGGER.debug("Findings are being fetched for {}", resource);
		stateMap.put(resource, FindingFetchState.pending());
		notifyListeners(listener -> listener.onFindingsFetchStarted(resource));
		notifyListeners(resource, listener -> listener.onFindingsFetchStarted(resource));

		LOGGER.traceExit();
	}

	@Override
	public void setFindingsFetchSuccessful(ResourceReference resource, List<Finding> findings) {
		LOGGER.traceEntry(null, resource, findings);

		LOGGER.debug("Findings have been successfully fetched for {} ({} findings)", resource, findings.size());
		stateMap.put(resource, FindingFetchState.success());
		notifyListeners(listener -> listener.onFindingsFetchSucceeded(resource, findings));
		notifyListeners(resource, listener -> listener.onFindingsFetchSucceeded(resource, findings));

		LOGGER.traceExit();
	}

	@Override
	public void setFindingsFetchFailure(ResourceReference resource, EFindingFetchError error,
			@Nullable String errorMessage, List<Finding> findingsOfPreviousCommit) {
		LOGGER.traceEntry(null, resource, error, errorMessage);

		LOGGER.debug("Findings could not be fetched for {} ({} error: {})", resource, error, errorMessage);
		stateMap.put(resource, FindingFetchState.failure(error, errorMessage));
		notifyListeners(
				listener -> listener.onFindingsFetchFailed(resource, error, errorMessage, findingsOfPreviousCommit));
		notifyListeners(resource,
				listener -> listener.onFindingsFetchFailed(resource, error, errorMessage, findingsOfPreviousCommit));

		LOGGER.traceExit();
	}

	@Override
	public void setFindingsFetchOutdated(ResourceReference resource) {
		LOGGER.traceEntry(null, resource);

		LOGGER.debug("Findings fetched for {} are considered outdated", resource);
		stateMap.put(resource, FindingFetchState.outdated());
		notifyListeners(listener -> listener.onFindingsFetchOutdated(resource));
		notifyListeners(resource, listener -> listener.onFindingsFetchOutdated(resource));

		LOGGER.traceExit();
	}

	@Override
	public void setAllFindingsFetchOutdated() {
		LOGGER.traceEntry();
		synchronized (stateMap) {
			for (ResourceReference resource : stateMap.keySet()) {
				if (getFindingState(resource).isPending()) {
					continue;
				}
				setFindingsFetchOutdated(resource);
			}
		}
		LOGGER.traceExit();
	}

	@Override
	public FindingFetchState getFindingState(ResourceReference resource) {
		return stateMap.getOrDefault(resource, FindingFetchState.notYetFetched());
	}

	@Override
	public void addListener(ResourceReference resource, IFindingsListener listener) {
		synchronized (resourceSpecificFindingsListeners) {
			resourceSpecificFindingsListeners.get(resource).add(listener);
		}
	}

	@Override
	public void removeListener(ResourceReference resource, IFindingsListener listener) {
		synchronized (resourceSpecificFindingsListeners) {
			resourceSpecificFindingsListeners.get(resource).remove(listener);
		}
	}

	/**
	 * Notifies resource specific listeners
	 */
	private void notifyListeners(ResourceReference resource, Consumer<IFindingsListener> consumer) {
		synchronized (resourceSpecificFindingsListeners) {
			resourceSpecificFindingsListeners.get(resource).forEach(consumer);
		}
	}
}
