/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings.tasks;

import java.net.URI;
import java.util.Collections;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import com.google.common.base.Preconditions;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.services.ITeamscaleConnectorService;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.model.RemoteResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.notification.Notifications;
import com.teamscale.ide.commons.notification.services.INotificationService;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.ITaskExecutor;
import com.teamscale.ide.commons.plugin.services.TaskBase;
import com.teamscale.ide.commons.servers.NamedServerDetails;
import com.teamscale.ide.commons.servers.services.IServerCredentialsManager;

/**
 * Base task for blacklisting findings.
 */
public abstract class BlacklistFindingTaskBase extends TaskBase {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(BlacklistFindingTaskBase.class);

	/** The finding which should be blacklisted. */
	protected final Finding finding;

	/** The rationale why a finding should be blacklisted. */
	protected final String rationale;

	/**
	 * The {@link ITeamscaleConnectorService} used to perform the blacklisting on
	 * the Teamscale server.
	 */
	protected final ITeamscaleConnectorService connectorService;
	private final INotificationService notificationService;

	private final IServerCredentialsManager credentialsManager;
	private final ITaskExecutor taskExecutor;
	private final IServiceRegistry serviceRegistry;

	protected BlacklistFindingTaskBase(IServiceRegistry serviceRegistry, Finding finding, String rationale,
			String label) {
		super(finding, label);

		Preconditions.checkArgument(finding.getBaseBranch().isPresent(), "Findings need a base branch to be flagged!");

		this.finding = finding;
		this.rationale = rationale;

		this.serviceRegistry = serviceRegistry;
		notificationService = serviceRegistry.getService(INotificationService.class);
		connectorService = serviceRegistry.getService(ITeamscaleConnectorService.class);
		credentialsManager = serviceRegistry.getService(IServerCredentialsManager.class);
		taskExecutor = serviceRegistry.getService(ITaskExecutor.class);
	}

	/**
	 * Performs the actual blacklisting by calling the respective service.
	 */
	protected abstract boolean performBlacklistingService(NamedServerDetails serverDetails,
			RemoteResourceReference remoteResource, String baseBranch) throws NetworkExceptionBase;

	@Override
	protected void run() {
		LOGGER.traceEntry(null, finding, rationale);

		RemoteResourceReference remoteResource = finding.getRemoteResource();
		URI serverUrl = remoteResource.getServerUrl();
		NamedServerDetails serverDetails = credentialsManager.getServer(serverUrl).orElse(null);

		if (serverDetails == null) {
			LOGGER.debug("Server is unknown: {}", serverUrl);
			notificationService.reportNotification(ENotificationCause.DIRECT,
					Notifications.serverUrlNotFound(serverUrl));
			return;
		}

		// Presence of base branch asserted in constructor
		String baseBranch = finding.getBaseBranch().orElseThrow(IllegalStateException::new);
		try {
			LOGGER.debug("Attempting to flag: {}", finding.getId());
			boolean wasSuccessfullyBlacklisted = performBlacklistingService(serverDetails, remoteResource, baseBranch);

			if (wasSuccessfullyBlacklisted) {
				// When blacklisting is not successful, an exception is thrown.
				LOGGER.debug("Flagging successful.");

				taskExecutor.execute(FetchFindingsTask.createForceFetchTask(serviceRegistry,
						ENotificationCause.INDIRECT, Collections.singleton(finding.getLocalResource()), false));
			}
		} catch (NetworkExceptionBase e) {
			LOGGER.catching(e);
			notificationService.reportNotification(ENotificationCause.DIRECT,
					Notifications.connectionError(URI.create(serverDetails.getUrl()), e.getServerState()));
		} finally {
			LOGGER.traceExit();
		}
	}
}
