package com.teamscale.ide.commons.findings.tasks;

import java.net.URI;
import java.util.Optional;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import com.teamscale.ide.commons.client.exceptions.ApiNotSupportedException;
import com.teamscale.ide.commons.client.exceptions.CheckDescriptionNotFoundException;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.exceptions.ProjectNotFoundException;
import com.teamscale.ide.commons.client.services.ITeamscaleConnectorService;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.IdeCheckDescription;
import com.teamscale.ide.commons.findings.services.ICheckDescriptionStoreService;
import com.teamscale.ide.commons.model.RemoteResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.notification.Notifications;
import com.teamscale.ide.commons.notification.services.INotificationService;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.TaskBase;
import com.teamscale.ide.commons.servers.NamedServerDetails;
import com.teamscale.ide.commons.servers.services.IServerCredentialsManager;

/** Task for fetching a check description. */
public class FetchCheckDescriptionTask extends TaskBase {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(FetchCheckDescriptionTask.class);

	/** The finding for which the description should be fetched. */
	private final Finding finding;

	private final ICheckDescriptionStoreService checkDescriptionStoreService;
	private final ITeamscaleConnectorService connectorService;
	private final INotificationService notificationService;
	private final IServerCredentialsManager credentialsManager;

	public FetchCheckDescriptionTask(IServiceRegistry serviceRegistry, Finding finding) {
		super(finding, "Loading check description");

		this.finding = finding;

		checkDescriptionStoreService = serviceRegistry.getService(ICheckDescriptionStoreService.class);
		connectorService = serviceRegistry.getService(ITeamscaleConnectorService.class);
		notificationService = serviceRegistry.getService(INotificationService.class);
		credentialsManager = serviceRegistry.getService(IServerCredentialsManager.class);
	}

	@Override
	protected void run() {
		LOGGER.traceEntry(null, finding);

		RemoteResourceReference remoteResource = finding.getRemoteResource();
		URI serverUrl = remoteResource.getServerUrl();

		Optional<IdeCheckDescription> cachedCheckDescription = checkDescriptionStoreService
				.getCheckDescription(serverUrl, finding);
		if (cachedCheckDescription.isPresent()) {
			// Trigger update so that tool window focuses on requested check description
			checkDescriptionStoreService.setCheckDescription(serverUrl, finding, cachedCheckDescription.get());
			LOGGER.traceExit();
			return;
		}

		NamedServerDetails serverDetails = credentialsManager.getServer(serverUrl).orElse(null);
		if (serverDetails == null) {
			LOGGER.debug("Server is unknown: {}", serverUrl);
			notificationService.reportNotification(ENotificationCause.DIRECT,
					Notifications.serverUrlNotFound(serverUrl));
			LOGGER.traceExit();
			return;
		}

		fetchCheckDescription(remoteResource, serverDetails);

		LOGGER.traceExit();
	}

	/** Fetches the check description. */
	private void fetchCheckDescription(RemoteResourceReference remoteResource, NamedServerDetails serverDetails) {
		URI serverUrl = remoteResource.getServerUrl();
		try {
			IdeCheckDescription description = connectorService.getCheckDescription(serverDetails,
					remoteResource.getTeamscaleProject(), finding);

			checkDescriptionStoreService.setCheckDescription(serverUrl, finding, description);
		} catch (NetworkExceptionBase e) {
			LOGGER.catching(e);
			notificationService.reportNotification(ENotificationCause.DIRECT,
					Notifications.connectionError(serverUrl, e.getServerState()));
		} catch (ProjectNotFoundException e) {
			LOGGER.debug("Project is not known to the Teamscale server", e);
			notificationService.reportNotification(ENotificationCause.DIRECT,
					Notifications.projectNotFound(serverUrl, e.getProjectId()));
		} catch (ApiNotSupportedException e) {
			LOGGER.catching(e);
			notificationService.reportNotification(ENotificationCause.DIRECT,
					Notifications.apiNotSupported(serverUrl, e.getLowestSupportedServerVersion(),
							"Check description cannot be loaded because the server does not support this feature"));
		} catch (CheckDescriptionNotFoundException e) {
			LOGGER.catching(e);
			notificationService.reportNotification(ENotificationCause.DIRECT,
					Notifications.checkDescriptionNotFound(e.getFindingTypeId()));
		}
	}
}
