/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings.tasks;

import java.util.Collection;
import java.util.HashSet;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.PreCommitUploadItem;
import com.teamscale.ide.commons.findings.PreCommitUploadTarget;
import com.teamscale.ide.commons.findings.services.IFindingsListener;
import com.teamscale.ide.commons.findings.tasks.fetcher.ECacheBehaviour;
import com.teamscale.ide.commons.findings.tasks.fetcher.FolderFindingsFetcher;
import com.teamscale.ide.commons.findings.tasks.fetcher.IFindingsFetcher;
import com.teamscale.ide.commons.findings.tasks.fetcher.LiveFileFindingsFetcher;
import com.teamscale.ide.commons.findings.tasks.fetcher.PreCommitFileFindingsFetcher;
import com.teamscale.ide.commons.findings.tasks.fetcher.PreCommitFindingsDeleteFetcher;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.ETaskInterruptionMode;
import com.teamscale.ide.commons.plugin.services.ETaskVisibility;
import com.teamscale.ide.commons.plugin.services.ITaskExecutor;
import com.teamscale.ide.commons.plugin.services.TaskBase;

/**
 * Task which fetches {@linkplain Finding findings} for the given
 * {@linkplain ResourceReference resource references}. Depending on the
 * constructor's {@code enablePreCommit} parameter, this fetching is performed
 * using the pre-commit (3.0) service (<code>true</code>) or the legacy findings
 * service (<code>false</code>) from Teamscale.
 * <p>
 * In the pre-commit mode, collections of {@linkplain PreCommitUploadItem upload
 * items} are built and uploaded to the respective
 * {@linkplain PreCommitUploadTarget upload targets}. If the results are pending
 * on the Teamscale Server, the task polls until finished.
 * <p>
 * Whenever findings are available (e.g. on cache hits, on initial requests, or
 * on final polling requests), the registered listener of the
 * {@link IFindingsListener} get informed.
 * <p>
 * Note, that the task has the {@link ETaskInterruptionMode#WAIT} interruption
 * mode, which lets the {@link ITaskExecutor} finish already running tasks with
 * the same class and token before starting new tasks.
 */
public class FetchFindingsTask extends TaskBase {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(FetchFindingsTask.class);

	/** Default polling interval in seconds for pending pre-commit requests. **/
	private static final int DEFAULT_POLLING_INTERVAL_IN_SEC = 1;

	// Dependencies
	private final IFindingsFetcher findingsFetcher;
	private final IServiceRegistry serviceRegistry;
	private final ENotificationCause causeForFetchingFindings;
	private final boolean enablePreCommit;
	private final ECacheBehaviour cacheBehaviour;
	private final int pollingIntervalInSec;

	/**
	 * Creates a task that deletes the given {@link ResourceReference resource
	 * references} in a single commit to their respective pre-commit branches.
	 * <p>
	 * {@link ResourceReference Folder references} are not allowed as input.
	 */
	public static FetchFindingsTask createDeletionTask(IServiceRegistry serviceRegistry,
			Collection<ResourceReference> deleteResources) {
		return new FetchFindingsTask(serviceRegistry, ENotificationCause.INDIRECT, true, true,
				ECacheBehaviour.ONLY_WRITE, DEFAULT_POLLING_INTERVAL_IN_SEC, deleteResources);
	}

	/**
	 * Creates a task that fetches findings for the given {@link ResourceReference
	 * resource references}.
	 * <p>
	 * No combination of folder and file {@link ResourceReference references} may be
	 * passed. The list must consist of only one type of references.
	 *
	 * @see #createFetchTask(IServiceRegistry, ENotificationCause, Collection,
	 *      boolean, int)
	 */
	public static FetchFindingsTask createFetchTask(IServiceRegistry serviceRegistry,
			ENotificationCause causeForFetchingFindings, Collection<ResourceReference> fetchResources,
			boolean enablePreCommit) {
		return new FetchFindingsTask(serviceRegistry, causeForFetchingFindings, false, enablePreCommit,
				ECacheBehaviour.READ_AND_WRITE, DEFAULT_POLLING_INTERVAL_IN_SEC, fetchResources);
	}

	/**
	 * Creates a task that fetches pre-commit findings for the given
	 * {@link ResourceReference resource references}.
	 * <p>
	 * No combination of folder and file {@link ResourceReference references} may be
	 * passed. The list must consist of only one type of references.
	 */
	public static FetchFindingsTask createPreCommitTask(IServiceRegistry serviceRegistry,
			Collection<ResourceReference> fetchResources, ENotificationCause causeForFetchingFindings,
			ECacheBehaviour cacheBehaviour) {
		return new FetchFindingsTask(serviceRegistry, causeForFetchingFindings, false, true, cacheBehaviour,
				DEFAULT_POLLING_INTERVAL_IN_SEC, fetchResources);
	}

	/**
	 * Creates a task that fetches pre-commit findings for the given
	 * {@link ResourceReference resource references}.
	 * <p>
	 * No combination of folder and file {@link ResourceReference references} may be
	 * passed. The list must consist of only one type of references.
	 */
	public static FetchFindingsTask createPreCommitTask(IServiceRegistry serviceRegistry,
			Collection<ResourceReference> fetchResources, ENotificationCause causeForFetchingFindings) {
		return new FetchFindingsTask(serviceRegistry, causeForFetchingFindings, false, true,
				ECacheBehaviour.READ_AND_WRITE, DEFAULT_POLLING_INTERVAL_IN_SEC, fetchResources);
	}

	/**
	 * Creates a task that fetches findings for the given {@link ResourceReference
	 * resource references}.
	 * <p>
	 * No combination of folder and file {@link ResourceReference references} may be
	 * passed. The list must consist of only one type of references.
	 * 
	 * @see #createFetchTask(IServiceRegistry, ENotificationCause, Collection,
	 *      boolean)
	 */
	public static FetchFindingsTask createFetchTask(IServiceRegistry serviceRegistry,
			ENotificationCause causeForFetchingFindings, Collection<ResourceReference> fetchResources,
			boolean enablePreCommit, int pollingIntervalInSec) {
		return new FetchFindingsTask(serviceRegistry, causeForFetchingFindings, false, enablePreCommit,
				ECacheBehaviour.READ_AND_WRITE, pollingIntervalInSec, fetchResources);
	}

	/**
	 * Creates a task that fetches findings direct from the server for the given
	 * {@link ResourceReference resource references}. This means the task will
	 * ignore the local cache.
	 * <p>
	 * No combination of folder and file {@link ResourceReference references} may be
	 * passed. The list must consist of only one type of references.
	 *
	 */
	public static FetchFindingsTask createForceFetchTask(IServiceRegistry serviceRegistry,
			ENotificationCause causeForFetchingFindings, Collection<ResourceReference> fetchResources,
			boolean enablePreCommit) {
		return new FetchFindingsTask(serviceRegistry, causeForFetchingFindings, false, enablePreCommit,
				ECacheBehaviour.ONLY_WRITE, DEFAULT_POLLING_INTERVAL_IN_SEC, fetchResources);
	}

	private FetchFindingsTask(IServiceRegistry serviceRegistry, ENotificationCause causeForFetchingFindings,
			boolean deleteResources, boolean enablePreCommit, ECacheBehaviour cacheBehaviour, int pollingIntervalInSec,
			Collection<ResourceReference> fetchResources) {
		super(fetchResources, "Loading findings", ETaskVisibility.UNOBTRUSIVE, ETaskInterruptionMode.WAIT);
		this.serviceRegistry = serviceRegistry;
		this.causeForFetchingFindings = causeForFetchingFindings;
		this.enablePreCommit = enablePreCommit;
		this.cacheBehaviour = cacheBehaviour;
		this.pollingIntervalInSec = pollingIntervalInSec;

		findingsFetcher = createFindingsFetcher(fetchResources, deleteResources);
	}

	@Override
	protected void run() {
		LOGGER.traceEntry();
		try {
			findingsFetcher.fetchFindings();
			LOGGER.traceExit();
		} catch (IllegalArgumentException | InterruptedException e) {
			LOGGER.catching(e);
			LOGGER.traceExit(e);
		}
	}

	/**
	 * Creates the {@link IFindingsFetcher fetcher} matching the given parameters.
	 * This includes the possibility to handle the deletion of the given
	 * {@link ResourceReference resources}.
	 */
	private IFindingsFetcher createFindingsFetcher(Collection<ResourceReference> fetchResources,
			boolean deleteResources) {
		if (deleteResources) {
			return createFindingsDeleter(fetchResources);
		} else {
			return createFindingsFetcher(fetchResources);
		}
	}

	/**
	 * Creates the {@link IFindingsFetcher fetcher} matching the given parameters.
	 */
	private IFindingsFetcher createFindingsFetcher(Collection<ResourceReference> fetchResources) {
		if (fetchResources.stream().allMatch(ResourceReference::isFolder)) {
			return new FolderFindingsFetcher(serviceRegistry, new HashSet<>(fetchResources), causeForFetchingFindings,
					this, enablePreCommit);
		} else if (fetchResources.stream().allMatch(ResourceReference::isFile)) {

			if (enablePreCommit) {
				return new PreCommitFileFindingsFetcher(serviceRegistry, fetchResources, causeForFetchingFindings,
						pollingIntervalInSec, this, cacheBehaviour);
			} else {
				return new LiveFileFindingsFetcher(serviceRegistry, fetchResources, causeForFetchingFindings, this,
						cacheBehaviour);
			}
		} else {
			throw new IllegalArgumentException("A mix of file and folder resources is not supported.");
		}
	}

	/**
	 * Creates the {@link IFindingsFetcher fetcher} matching the given parameters,
	 * which handles all given {@link ResourceReference resources} as if they were
	 * deleted.
	 */
	private IFindingsFetcher createFindingsDeleter(Collection<ResourceReference> fetchResources) {
		if (enablePreCommit) {
			if (fetchResources.stream().allMatch(ResourceReference::isFile)) {
				return new PreCommitFindingsDeleteFetcher(serviceRegistry, fetchResources, causeForFetchingFindings,
						pollingIntervalInSec, this);
			} else {
				throw new IllegalArgumentException("Pre-commit deletion is only supported for files.");
			}
		} else {
			throw new IllegalStateException("Pre-commit deletion is only usable with pre-commit.");
		}
	}
}
