package com.teamscale.ide.commons.findings.tasks;

import java.time.Duration;
import java.util.Collections;
import java.util.Optional;
import java.util.concurrent.TimeUnit;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import com.teamscale.ide.commons.client.RestClientUtils;
import com.teamscale.ide.commons.client.exceptions.ApiNotSupportedException;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.exceptions.ProjectNotFoundException;
import com.teamscale.ide.commons.client.services.ITeamscaleConnectorService;
import com.teamscale.ide.commons.mappings.ResourceMappingResult;
import com.teamscale.ide.commons.mappings.services.IResourceMappingService;
import com.teamscale.ide.commons.model.RemoteResourceReference;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.model.VcsResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.notification.Notifications;
import com.teamscale.ide.commons.notification.services.INotificationService;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.ITaskExecutor;
import com.teamscale.ide.commons.plugin.services.TaskBase;
import com.teamscale.ide.commons.servers.NamedServerDetails;
import com.teamscale.ide.commons.servers.services.IServerCredentialsManager;

/**
 * Forces the Teamscale server to fetch new findings from the SAP system.
 * <p>
 * This task has two side effects. It fetches the newest findings from Teamscale
 * afterward and deletes the pre-commit branch of the user to keep it in sync
 * with the newest findings from Code Inspector.
 */
public class TeamscaleSapSyncTask extends TaskBase {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(TeamscaleSapSyncTask.class);

	private static final String PRE_COMMIT_API_NOT_SUPPORTED_ERROR_MESSAGE = "Pre-commit branch could not be updated "
			+ "because the server does not support this feature";

	private static final int DEFAULT_POLLING_INTERVAL_IN_SEC = 1;

	private final IServiceRegistry serviceRegistry;
	private final ITaskExecutor taskExecutor;
	private final ITeamscaleConnectorService connectorService;
	private final IServerCredentialsManager serverManager;
	private final INotificationService notificationService;
	private final IResourceMappingService mappingService;

	private final ResourceReference resource;

	private final int sapSystemUpdateTimeout;

	public TeamscaleSapSyncTask(IServiceRegistry serviceRegistry, ResourceReference resource,
			int sapSystemUpdateTimeout) {
		super(resource, "Sync SAP with Teamscale");
		this.serviceRegistry = serviceRegistry;
		this.resource = resource;

		this.taskExecutor = serviceRegistry.getService(ITaskExecutor.class);
		this.connectorService = serviceRegistry.getService(ITeamscaleConnectorService.class);
		this.serverManager = serviceRegistry.getService(IServerCredentialsManager.class);
		this.notificationService = serviceRegistry.getService(INotificationService.class);
		this.mappingService = serviceRegistry.getService(IResourceMappingService.class);
		this.sapSystemUpdateTimeout = sapSystemUpdateTimeout;
	}

	@Override
	protected void run() {
		ResourceMappingResult mappingResult = mappingService.map(resource);

		if (!mappingResult.isSuccessful()) {
			notificationService.reportNotification(ENotificationCause.DIRECT,
					Notifications.mappingProblem(resource, mappingResult.getMappingFailure()));
			return;
		}

		RemoteResourceReference remoteResourceReference = mappingResult
				.getMappedResource(new VcsResourceReference(resource, null));

		NamedServerDetails serverDetails = serverManager.getServer(remoteResourceReference.getServerUrl()).orElse(null);

		if (serverDetails == null) {
			notificationService.reportNotification(ENotificationCause.DIRECT,
					Notifications.serverUrlNotFound(remoteResourceReference.getServerUrl()));
			return;
		}

		try {
			updateSapSystem(remoteResourceReference, serverDetails);
		} catch (NetworkExceptionBase e) {
			LOGGER.catching(e);
			notificationService.reportNotification(ENotificationCause.DIRECT,
					Notifications.connectionError(remoteResourceReference.getServerUrl(), e.getServerState()));
		} catch (InterruptedException e) {
			LOGGER.catching(e);
		}
	}

	private void updateSapSystem(RemoteResourceReference remoteResourceReference, NamedServerDetails serverDetails)
			throws NetworkExceptionBase, InterruptedException {
		long startTime = System.currentTimeMillis();

		String sapSystemId = connectorService.resolveTeamscaleIdToSapSystemID(serverDetails,
				remoteResourceReference.getTeamscaleProject());

		if (isSapSystemUpdateTimeout(startTime)) {
			notificationService.reportNotification(ENotificationCause.DIRECT,
					Notifications.sapTeamscaleSyncTimeout(sapSystemId));
			return;
		}

		String updateId = connectorService.synchronizeAbapProjects(serverDetails, sapSystemId);

		boolean waitedSuccessful = RestClientUtils.pollRepeatedly(() -> {
			if (isSapSystemUpdateTimeout(startTime)) {
				return Optional.of(false);
			}
			boolean isUpdating = connectorService.isSynchronizingAbapProjects(serverDetails, sapSystemId, updateId);

			if (!isUpdating) {
				return Optional.of(true);
			}

			return Optional.empty();
		}, DEFAULT_POLLING_INTERVAL_IN_SEC, TimeUnit.SECONDS);

		if (!waitedSuccessful) {
			notificationService.reportNotification(ENotificationCause.DIRECT,
					Notifications.sapTeamscaleSyncTimeout(sapSystemId));
			return;
		}

		// We need to delete the pre-commit branch of the user, so they see new findings
		// from Code Inspector
		clearPreCommitBranch(remoteResourceReference, serverDetails);

		// Fetch the new findings of Code Inspector from Teamscale
		taskExecutor.execute(FetchFindingsTask.createForceFetchTask(serviceRegistry, ENotificationCause.INDIRECT,
				Collections.singleton(resource), false));
	}

	private void clearPreCommitBranch(RemoteResourceReference remoteResourceReference, NamedServerDetails serverDetails)
			throws NetworkExceptionBase {
		try {
			// In ABAP the user is always working on the default branch
			String defaultBranch = connectorService.getDefaultBranch(serverDetails,
					remoteResourceReference.getTeamscaleProject());

			if (!connectorService.clearPreCommitBranch(serverDetails, remoteResourceReference.getTeamscaleProject(),
					null, defaultBranch)) {
				// Only a warning because it can happen that the user has no pre-commit branch
				// in the first place
				LOGGER.warn("Not able to delete the pre-commit branch of the user.");
			}
		} catch (ApiNotSupportedException e) {
			notificationService.reportNotification(ENotificationCause.INDIRECT,
					Notifications.apiNotSupported(remoteResourceReference.getServerUrl(),
							e.getLowestSupportedServerVersion(), PRE_COMMIT_API_NOT_SUPPORTED_ERROR_MESSAGE));
		} catch (ProjectNotFoundException e) {
			notificationService.reportNotification(ENotificationCause.INDIRECT, Notifications.projectNotFound(
					remoteResourceReference.getServerUrl(), remoteResourceReference.getTeamscaleProject()));
		}
	}

	private boolean isSapSystemUpdateTimeout(long startTime) {
		return (System.currentTimeMillis() - startTime) / Duration.ofSeconds(1).toMillis() > sapSystemUpdateTimeout;
	}
}
