/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings.tasks.fetcher;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.plugin.services.TaskBase;

/**
 * Encapsulates the basic attributes every {@link IFindingsFetcher fetcher}
 * needs and checks if the {@link ResourceReference input resources} are valid.
 * 
 * @see LiveFindingsFetcherBase fetching non-pre-commit findings
 * @see PreCommitFindingsFetcherBase fetching pre-commit findings
 */
/* package */ abstract class FindingsFetcherBase implements IFindingsFetcher {

	/**
	 * Indicates if fetching of findings was initiated by a direct user action e.g.
	 * clicking refresh findings button or by some indirect action e.g. opening of
	 * files in IDE when IDE first opens up. This is used to control notifications.
	 */
	protected final ENotificationCause causeForFetchingFindings;

	/**
	 * The calling task, used to detect if the task has been canceled.
	 */
	protected final TaskBase ownerTask;

	/** Input resources for which server interaction is required. */
	private final Set<ResourceReference> fetchResources;

	/* package */ FindingsFetcherBase(Collection<ResourceReference> fetchResources,
			ENotificationCause causeForFetchingFindings, TaskBase ownerTask) {
		this.fetchResources = new HashSet<>(fetchResources);
		this.causeForFetchingFindings = causeForFetchingFindings;
		this.ownerTask = ownerTask;
	}

	/**
	 * Fetches {@linkplain Finding findings} for the given
	 * {@linkplain ResourceReference references}.
	 *
	 * @throws InterruptedException
	 *             whenever the calling task got interrupted
	 */
	protected abstract void doFetchFindings(Set<ResourceReference> fetchFileResources) throws InterruptedException;

	@Override
	public final void fetchFindings() throws IllegalArgumentException, InterruptedException {
		if (fetchResources.isEmpty()) {
			throw new IllegalArgumentException("No resources have been selected.");
		}

		checkIfTaskCanceled(fetchResources);

		doFetchFindings(fetchResources);
	}

	/**
	 * Checks if the {@link #ownerTask task} got canceled and throws an
	 * {@link InterruptedException}. If the task got canceled,
	 * {@link #onTaskCanceled(ResourceReference)} gets called.
	 */
	protected final void checkIfTaskCanceled(ResourceReference resourceReference) throws InterruptedException {
		if (ownerTask.isCancelRequested()) {
			onTaskCanceled(resourceReference);
			throw new InterruptedException("Fetching findings canceled.");
		}
	}

	/**
	 * Checks if the {@link #ownerTask task} got canceled and throws an
	 * {@link InterruptedException}. If the task got canceled,
	 * {@link #onTaskCanceled(ResourceReference)} gets called for each
	 * {@link ResourceReference resource}.
	 */
	protected final void checkIfTaskCanceled(Set<ResourceReference> resourceReferences) throws InterruptedException {
		if (ownerTask.isCancelRequested()) {
			resourceReferences.forEach(this::onTaskCanceled);
			throw new InterruptedException("Fetching findings canceled.");
		}
	}

	/**
	 * Get called if the task got canceled.
	 */
	protected void onTaskCanceled(ResourceReference resourceReference) {
		// Nothing to do in the base case
	}
}
