/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings.tasks.fetcher;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.conqat.engine.index.shared.TrackedFinding;
import org.conqat.lib.commons.uniformpath.UniformPath;
import org.conqat.lib.commons.uniformpath.UniformPathCompatibilityUtil;

import com.google.common.base.Preconditions;
import com.teamscale.ide.commons.client.exceptions.BranchNotFoundException;
import com.teamscale.ide.commons.findings.EFindingFetchError;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.FindingUtils;
import com.teamscale.ide.commons.findings.services.IFindingFetchStateService;
import com.teamscale.ide.commons.findings.services.IFindingsListener;
import com.teamscale.ide.commons.findings.tasks.FetchFindingsTask;
import com.teamscale.ide.commons.mappings.MappingUtils;
import com.teamscale.ide.commons.mappings.ResourceMappingResult;
import com.teamscale.ide.commons.mappings.services.IResourceMappingService;
import com.teamscale.ide.commons.model.RemoteResourceReference;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.model.VcsResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.notification.Notifications;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.TaskBase;
import com.teamscale.ide.commons.precommit.PreCommitUtils;
import com.teamscale.ide.commons.servers.NamedServerDetails;
import com.teamscale.ide.commons.servers.services.IServerCredentialsManager;
import com.teamscale.ide.commons.vcs.services.ICheckoutFactory;

/**
 * Encapsulates logic for a {@link FetchFindingsTask} to fetch findings for the
 * given {@linkplain ResourceReference folder references}.
 * <p>
 * Whenever findings are available on the server, the {@link IFindingsListener}
 * registered listeners of the {@link FetchFindingsTask} get informed.
 */
public class FolderFindingsFetcher extends LiveFindingsFetcherBase {

	private static final Logger LOGGER = LogManager.getLogger(FolderFindingsFetcher.class);

	private final boolean usePreCommitBranch;

	private final ICheckoutFactory checkoutFactory;
	private final IServerCredentialsManager serverManager;
	private final IResourceMappingService resourceMappingService;
	private final IFindingFetchStateService findingStateService;

	public FolderFindingsFetcher(IServiceRegistry serviceRegistry, Collection<ResourceReference> fetchResources,
			ENotificationCause causeForFetchingFindings, TaskBase ownerTask, boolean usePreCommitBranch) {
		super(serviceRegistry, fetchResources, causeForFetchingFindings, ownerTask);
		Preconditions.checkArgument(fetchResources.stream().allMatch(ResourceReference::isFolder),
				"The FolderFindingsFetcher does not support files as ResourceReference.");

		this.usePreCommitBranch = usePreCommitBranch;

		checkoutFactory = serviceRegistry.getService(ICheckoutFactory.class);
		serverManager = serviceRegistry.getService(IServerCredentialsManager.class);
		resourceMappingService = serviceRegistry.getService(IResourceMappingService.class);
		findingStateService = serviceRegistry.getService(IFindingFetchStateService.class);
	}

	@Override
	protected void doFetchFindings(Set<ResourceReference> originFolderResources) throws InterruptedException {
		LOGGER.traceEntry(null, originFolderResources);

		Set<ResourceReference> roots = computeFetchRoots(originFolderResources);
		fetchFolderFindings(roots);

		LOGGER.traceExit();
	}

	/**
	 * @return All subfolders that need a separate server call including the
	 *         {@code originFolderResources}.
	 */
	private Set<ResourceReference> computeFetchRoots(Set<ResourceReference> originFolderResources) {
		LOGGER.traceEntry(null, originFolderResources);
		Set<ResourceReference> fetchRoots = new HashSet<>(originFolderResources);

		for (ResourceReference originResource : originFolderResources) {
			Set<ResourceReference> additionalRoots = resourceMappingService.findPotentialMappingRoots(originResource)
					.stream().filter(root -> root.isAtOrBelow(originResource)).collect(Collectors.toSet());
			LOGGER.trace(additionalRoots);
			fetchRoots.addAll(additionalRoots);
		}

		return LOGGER.traceExit(fetchRoots);
	}

	private void fetchFolderFindings(Collection<ResourceReference> folderResources) throws InterruptedException {
		for (ResourceReference folderResource : folderResources) {
			findingStateService.setFindingsFetchPending(folderResource);

			ResourceMappingResult mappingResult = resourceMappingService.map(folderResource);
			if (!mappingResult.isSuccessful()) {
				notificationService.reportNotification(causeForFetchingFindings,
						Notifications.mappingProblem(folderResource, mappingResult.getMappingFailure()));
				findingStateService.setFindingsFetchFailure(folderResource, EFindingFetchError.MAPPING,
						"Mapping problem.");
				continue;
			}

			URI serverUrl = mappingResult.getServerUrl();
			NamedServerDetails server = serverManager.getServer(serverUrl).orElse(null);
			if (server == null) {
				notificationService.reportNotification(causeForFetchingFindings,
						Notifications.serverUrlNotFound(serverUrl));
				findingStateService.setFindingsFetchFailure(folderResource, EFindingFetchError.NETWORK,
						"Server URL not found.");
				continue;
			}

			fetchFolderFindings(folderResource, mappingResult, server);
		}
	}

	private void fetchFolderFindings(ResourceReference folderResource, ResourceMappingResult mappingResult,
			NamedServerDetails server) throws InterruptedException {
		String baseBranch = MappingUtils.getBranch(checkoutFactory, folderResource, mappingResult).orElse(null);

		if (usePreCommitBranch) {
			try {
				RemoteResourceReference remoteResource = mappingResult.getMappedResource(new VcsResourceReference(
						folderResource, PreCommitUtils.getPreCommit3BranchName(server, baseBranch)));

				doFetchFolderFindings(folderResource, remoteResource, server, baseBranch, mappingResult);
				return;
			} catch (BranchNotFoundException e) {
				// Pre-commit branch does not exist. Use the baseBranch as fallback.
				LOGGER.debug("Pre-Commit branch {} does not exist.", e.getBranch());
				LOGGER.catching(e);
			}
		}

		try {
			RemoteResourceReference remoteResource = mappingResult
					.getMappedResource(new VcsResourceReference(folderResource, baseBranch));

			doFetchFolderFindings(folderResource, remoteResource, server, baseBranch, mappingResult);
		} catch (BranchNotFoundException e) {
			notificationService.reportNotification(causeForFetchingFindings, Notifications
					.teamscaleBranchNotFound(mappingResult.getServerUrl(), mappingResult.getProjectId(), baseBranch));
			findingStateService.setFindingsFetchFailure(folderResource, EFindingFetchError.API,
					String.format("Branch %s not found.", e.getBranch()));
		}
	}

	private void doFetchFolderFindings(ResourceReference localResource, RemoteResourceReference remoteResource,
			NamedServerDetails server, String baseBranch, ResourceMappingResult mappingResult)
			throws InterruptedException, BranchNotFoundException {

		List<TrackedFinding> findingDtos = fetchFindingsDtos(server, mappingResult, localResource, remoteResource,
				true);
		List<Finding> findings = convertFindingDtos(findingDtos, baseBranch, mappingResult);
		findingStateService.setFindingsFetchSuccessful(localResource, findings);
	}

	/**
	 * Transforms a list of {@linkplain TrackedFinding finding DTOs} into a list of
	 * {@link Finding}s.
	 */
	private static List<Finding> convertFindingDtos(List<TrackedFinding> findingDtos, String baseBranch,
			ResourceMappingResult mappingResult) {
		List<Finding> findings = new ArrayList<>();
		for (TrackedFinding findingDto : findingDtos) {
			UniformPath uniformPath = UniformPathCompatibilityUtil.convert(findingDto.getLocation().getUniformPath());

			// Findings are always located in files, never in folders.
			ResourceReference localFileResource = mappingResult.getLocalFile(uniformPath).orElse(null);
			if (localFileResource == null) {
				continue;
			}
			RemoteResourceReference remoteFileResource = mappingResult
					.getMappedResource(new VcsResourceReference(localFileResource, baseBranch));

			FindingUtils.extractFinding(findingDto, baseBranch, localFileResource, remoteFileResource)
					.ifPresent(findings::add);
		}
		return findings;
	}
}
