/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings.tasks.fetcher;

import java.net.URI;
import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.conqat.engine.index.shared.TrackedFinding;

import com.google.common.base.Preconditions;
import com.teamscale.ide.commons.client.exceptions.BranchNotFoundException;
import com.teamscale.ide.commons.findings.EFindingFetchError;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.FindingUtils;
import com.teamscale.ide.commons.findings.services.IFetchFindingsTaskCache;
import com.teamscale.ide.commons.findings.services.IFindingsListener;
import com.teamscale.ide.commons.findings.tasks.FetchFindingsTask;
import com.teamscale.ide.commons.mappings.MappingUtils;
import com.teamscale.ide.commons.mappings.ResourceMappingResult;
import com.teamscale.ide.commons.mappings.services.IResourceMappingService;
import com.teamscale.ide.commons.model.RemoteResourceReference;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.model.VcsResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.notification.Notifications;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.TaskBase;
import com.teamscale.ide.commons.servers.NamedServerDetails;
import com.teamscale.ide.commons.servers.services.IServerCredentialsManager;
import com.teamscale.ide.commons.vcs.services.ICheckoutFactory;

/**
 * Encapsulates logic for a {@link FetchFindingsTask} to fetch (non-pre-commit)
 * findings for the given {@linkplain ResourceReference file references}.
 * <p>
 * Whenever findings are available (e.g. on cache hits, or as request result),
 * the {@link IFindingsListener} registered listeners of the
 * {@link FetchFindingsTask} get informed.
 * <p>
 * 
 * @see PreCommitFileFindingsFetcher fetching pre-commit findings
 */
public class LiveFileFindingsFetcher extends LiveFindingsFetcherBase {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(LiveFileFindingsFetcher.class);

	/** The checkout factory */
	private final ICheckoutFactory checkoutFactory;
	/** The server manager */
	private final IServerCredentialsManager serverManager;
	/** The resource mapping service */
	private final IResourceMappingService resourceMappingService;
	/** The findings cache */
	private final IFetchFindingsTaskCache findingsCache;

	/** Controls if the fetch ignores the cache. */
	private final ECacheBehaviour cacheBehaviour;

	public LiveFileFindingsFetcher(IServiceRegistry serviceRegistry, Collection<ResourceReference> fetchResources,
			ENotificationCause causeForFetchingFindings, TaskBase ownerTask, ECacheBehaviour cacheBehaviour) {
		super(serviceRegistry, fetchResources, causeForFetchingFindings, ownerTask);
		Preconditions.checkArgument(fetchResources.stream().allMatch(ResourceReference::isFile),
				"The LiveFileFindingsFetcher does not support folders as ResourceReference.");

		this.cacheBehaviour = cacheBehaviour;

		checkoutFactory = serviceRegistry.getService(ICheckoutFactory.class);
		serverManager = serviceRegistry.getService(IServerCredentialsManager.class);
		resourceMappingService = serviceRegistry.getService(IResourceMappingService.class);
		findingsCache = serviceRegistry.getService(IFetchFindingsTaskCache.class);
	}

	/**
	 * Fetches {@linkplain Finding findings} for the given
	 * {@linkplain ResourceReference references} by
	 * <ul>
	 * <li>searching in the {@link IFetchFindingsTaskCache}, and if not cached</li>
	 * <li>requesting and transforming {@linkplain TrackedFinding TrackedFindings}
	 * from the Teamscale server using the (non-pre-commit) findings service.</li>
	 * </ul>
	 *
	 * @throws InterruptedException
	 *             when fetching {@linkplain Finding findings} is canceled by the
	 *             user
	 */
	@Override
	public void doFetchFindings(Set<ResourceReference> fetchFileResources) throws InterruptedException {
		LOGGER.traceEntry("Fetching findings for selected resources {}.", fetchFileResources);
		fetchFileResources.forEach(findingStateService::setFindingsFetchPending);

		for (ResourceReference localResource : fetchFileResources) {
			ResourceMappingResult mappingResult = resourceMappingService.map(localResource);
			if (!mappingResult.isSuccessful()) {
				notificationService.reportNotification(causeForFetchingFindings,
						Notifications.mappingProblem(localResource, mappingResult.getMappingFailure()));
				findingStateService.setFindingsFetchFailure(localResource, EFindingFetchError.MAPPING,
						"Mapping problem.");
				continue;
			}
			String baseBranch = MappingUtils.getBranch(checkoutFactory, localResource, mappingResult).orElse(null);
			RemoteResourceReference remoteResource = mappingResult
					.getMappedResource(new VcsResourceReference(localResource, baseBranch));

			URI serverUrl = remoteResource.getServerUrl();
			NamedServerDetails server = serverManager.getServer(serverUrl).orElse(null);
			if (server == null) {
				notificationService.reportNotification(causeForFetchingFindings,
						Notifications.serverUrlNotFound(serverUrl));
				findingStateService.setFindingsFetchFailure(localResource, EFindingFetchError.NETWORK,
						"Server URL not found.");
				continue;
			}

			fetchAndUpdateFindings(localResource, remoteResource, server, baseBranch, mappingResult);
		}
		LOGGER.traceExit();
	}

	private void fetchAndUpdateFindings(ResourceReference localResource, RemoteResourceReference remoteResource,
			NamedServerDetails server, String baseBranch, ResourceMappingResult mappingResult)
			throws InterruptedException {
		if (cacheBehaviour.shouldRead()) {
			List<Finding> cachedFindings = findingsCache.getFindings(baseBranch, localResource, null).orElse(null);
			if (cachedFindings != null) {
				LOGGER.debug("Using cached {} findings for {}.", cachedFindings.size(), localResource);
				findingStateService.setFindingsFetchSuccessful(localResource, cachedFindings);
				return;
			}
		}

		try {
			List<TrackedFinding> findingDtos = fetchFindingsDtos(server, mappingResult, localResource, remoteResource,
					false);
			List<Finding> findings = FindingUtils.extractFindings(findingDtos, baseBranch, localResource,
					remoteResource);

			findingStateService.setFindingsFetchSuccessful(localResource, findings);
			if (cacheBehaviour.shouldWrite()) {
				LOGGER.debug("Writing {} findings for {} to the cache.", findings.size(), localResource);
				findingsCache.updateFindings(baseBranch, localResource, null, findings);
			}
		} catch (BranchNotFoundException e) {
			notificationService.reportNotification(causeForFetchingFindings, Notifications
					.teamscaleBranchNotFound(mappingResult.getServerUrl(), mappingResult.getProjectId(), baseBranch));
			findingStateService.setFindingsFetchFailure(localResource, EFindingFetchError.API,
					String.format("Branch %s not found", e.getBranch()));
		}
	}
}
