/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings.tasks.fetcher;

import java.util.Collection;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.conqat.engine.index.shared.TrackedFinding;

import com.teamscale.ide.commons.client.exceptions.BranchNotFoundException;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.exceptions.ProjectNotFoundException;
import com.teamscale.ide.commons.client.services.ITeamscaleConnectorService;
import com.teamscale.ide.commons.findings.EFindingFetchError;
import com.teamscale.ide.commons.findings.services.IFindingFetchStateService;
import com.teamscale.ide.commons.findings.services.IFindingsListener;
import com.teamscale.ide.commons.findings.tasks.FetchFindingsTask;
import com.teamscale.ide.commons.mappings.ResourceMappingResult;
import com.teamscale.ide.commons.model.IdeCommit;
import com.teamscale.ide.commons.model.RemoteResourceReference;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.notification.Notifications;
import com.teamscale.ide.commons.notification.services.INotificationService;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.TaskBase;
import com.teamscale.ide.commons.servers.NamedServerDetails;
import com.teamscale.index.findings.calculation.FindingsWithCount;

/**
 * Encapsulates logic for a {@link FetchFindingsTask} to fetch non-pre-commit
 * findings for the given {@linkplain ResourceReference resources}.
 * <p>
 * Whenever the fetching fails, the {@link IFindingsListener} registered
 * listeners of the {@link IFindingFetchStateService} get informed.
 * 
 * @see PreCommitFindingsFetcherBase fetching pre-commit findings
 */
/* package */ abstract class LiveFindingsFetcherBase extends FindingsFetcherBase {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(LiveFindingsFetcherBase.class);

	/** The notification service */
	protected final INotificationService notificationService;
	/** The connector service */
	protected final ITeamscaleConnectorService connectorService;
	/** The finding state service */
	protected final IFindingFetchStateService findingStateService;

	LiveFindingsFetcherBase(IServiceRegistry serviceRegistry, Collection<ResourceReference> fetchResources,
			ENotificationCause causeForFetchingFindings, TaskBase ownerTask) {
		super(fetchResources, causeForFetchingFindings, ownerTask);
		notificationService = serviceRegistry.getService(INotificationService.class);
		connectorService = serviceRegistry.getService(ITeamscaleConnectorService.class);
		findingStateService = serviceRegistry.getService(IFindingFetchStateService.class);
	}

	/**
	 * Fetches findings from the Teamscale server and returns them. Sets the
	 * findings fetch state to failure if an {@link InterruptedException exception}
	 * is thrown.
	 */
	protected final List<TrackedFinding> fetchFindingsDtos(NamedServerDetails server,
			ResourceMappingResult mappingResult, ResourceReference localResource,
			RemoteResourceReference remoteResource, boolean includeSubPaths)
			throws InterruptedException, BranchNotFoundException {
		LOGGER.debug("Fetching findings from server for {}", remoteResource);

		checkIfTaskCanceled(localResource);

		try {
			String branch = remoteResource.getBranch()
					.orElse(connectorService.getDefaultBranch(server, remoteResource.getTeamscaleProject()));
			IdeCommit commit = IdeCommit.head(branch);

			FindingsWithCount fetchedFindings = connectorService.getFindings(server,
					remoteResource.getTeamscaleProject(), remoteResource.getUniformPath(),
					// we currently do not consider case sensitiveness (see TS-30774)
					true, commit, includeSubPaths);
			List<TrackedFinding> findingDtos = fetchedFindings.getFindings();
			LOGGER.debug("Got {} findings.", findingDtos.size());

			checkIfTaskCanceled(localResource);

			return findingDtos;
		} catch (ProjectNotFoundException e) {
			LOGGER.debug("Project is not known to the Teamscale server", e);
			findingStateService.setFindingsFetchFailure(localResource, EFindingFetchError.NETWORK,
					"Project is not known to the Teamscale server");
			notificationService.reportNotification(causeForFetchingFindings,
					Notifications.projectNotFound(remoteResource.getServerUrl(), remoteResource.getTeamscaleProject()));
			throw new InterruptedException("Project is not known to the Teamscale server");
		} catch (NetworkExceptionBase e) {
			LOGGER.catching(e);
			findingStateService.setFindingsFetchFailure(localResource, EFindingFetchError.NETWORK, e.getMessage());
			notificationService.reportNotification(causeForFetchingFindings,
					Notifications.connectionError(mappingResult.getServerUrl(), e.getServerState()));
			throw new InterruptedException("Server communication problems");
		}
	}

	@Override
	protected void onTaskCanceled(ResourceReference resourceReference) {
		findingStateService.setFindingsFetchFailure(resourceReference, EFindingFetchError.CANCELED,
				"Fetching findings canceled.");
	}
}
