/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings.tasks.fetcher;

import static com.google.common.collect.MoreCollectors.onlyElement;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.conqat.engine.index.shared.CommitDescriptor;
import org.conqat.engine.index.shared.PreCommit3Result;
import org.conqat.engine.index.shared.PreCommitUploadData;
import org.conqat.engine.index.shared.TrackedFinding;
import org.conqat.lib.commons.uniformpath.UniformPath;
import org.conqat.lib.commons.uniformpath.UniformPathCompatibilityUtil;

import com.google.common.base.Preconditions;
import com.google.common.collect.Maps;
import com.teamscale.ide.commons.findings.EFindingFetchError;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.FindingUtils;
import com.teamscale.ide.commons.findings.PreCommitResult;
import com.teamscale.ide.commons.findings.PreCommitUploadItem;
import com.teamscale.ide.commons.findings.PreCommitUploadTarget;
import com.teamscale.ide.commons.findings.services.IFetchFindingsTaskCache;
import com.teamscale.ide.commons.findings.services.IFindingsListener;
import com.teamscale.ide.commons.findings.tasks.FetchFindingsTask;
import com.teamscale.ide.commons.model.RemoteResourceReference;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.notification.Notifications;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.TaskBase;
import com.teamscale.ide.commons.precommit.PreCommitUtils;

/**
 * Encapsulates logic for a {@link FetchFindingsTask} to fetch pre-commit
 * findings for the given {@linkplain ResourceReference file references}.
 * <p>
 * Collections of {@linkplain PreCommitUploadItem upload items} are built and
 * uploaded to the respective {@linkplain PreCommitUploadTarget upload targets}.
 * If the results are pending on the Teamscale Server, the task polls until
 * finished.
 * <p>
 * Whenever findings are available (e.g. on cache hits, on initial requests, or
 * on final polling requests), the {@link IFindingsListener} registered
 * listeners of the {@link FetchFindingsTask} get informed.
 * <p>
 * 
 * @see LiveFileFindingsFetcher fetching non-pre-commit findings
 */
public class PreCommitFileFindingsFetcher extends PreCommitFindingsFetcherBase {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(PreCommitFileFindingsFetcher.class);

	/** The findings cache */
	private final IFetchFindingsTaskCache findingsCache;

	/** Controls if the fetch ignores the cache. */
	private final ECacheBehaviour cacheBehaviour;

	public PreCommitFileFindingsFetcher(IServiceRegistry serviceRegistry, Collection<ResourceReference> fetchResources,
			ENotificationCause causeForFetchingFindings, int pollingIntervalInSec, TaskBase ownerTask,
			ECacheBehaviour cacheBehaviour) {
		super(serviceRegistry, fetchResources, causeForFetchingFindings, pollingIntervalInSec, ownerTask);
		Preconditions.checkArgument(fetchResources.stream().allMatch(ResourceReference::isFile),
				"The PreCommitFileFindingsFetcher does not support folders as ResourceReference.");
		this.cacheBehaviour = cacheBehaviour;

		findingsCache = serviceRegistry.getService(IFetchFindingsTaskCache.class);
	}

	/**
	 * Fetches {@linkplain Finding findings} for the given
	 * {@link ResourceReference}s by
	 * <ul>
	 * <li>searching in the {@link IFetchFindingsTaskCache}, and if not cached</li>
	 * <li>requesting and transforming {@link PreCommit3Result}s from the Teamscale
	 * server, with polling.</li>
	 * </ul>
	 */
	@Override
	public void doFetchFindings(Set<ResourceReference> fetchFileResources) {
		LOGGER.traceEntry("Fetching pre-commit results for selected resources {}.", fetchFileResources);
		fetchFileResources.forEach(fetchFindingStateService::setFindingsFetchPending);
		Set<PreCommitUploadItem> uploadItems = buildUploadItems(fetchFileResources);

		if (!cacheBehaviour.shouldRead()) {
			LOGGER.debug("Fetch without cache.");
			fetchPreCommitResults(uploadItems);
		} else {
			Map<PreCommitUploadItem, List<Finding>> uploadItemsAndCachedFindings = getCachedFindings(uploadItems);

			Map<PreCommitUploadItem, List<Finding>> uploadItemsWithCachedFindings = Maps
					.filterValues(uploadItemsAndCachedFindings, PreCommitFileFindingsFetcher::hasFindings);
			LOGGER.debug("Found resources with cached findings: {}", uploadItemsWithCachedFindings);
			uploadItemsWithCachedFindings.forEach((uploadItem, findings) -> fetchFindingStateService
					.setFindingsFetchSuccessful(uploadItem.getLocalResource(), findings));

			Set<PreCommitUploadItem> uploadItemsWithoutCachedFindings = Maps
					.filterValues(uploadItemsAndCachedFindings, PreCommitFileFindingsFetcher::hasNoFindings).keySet();
			LOGGER.debug("Found resources without cached findings: {}", uploadItemsWithoutCachedFindings);

			if (!uploadItemsWithoutCachedFindings.isEmpty()) {
				fetchPreCommitResults(uploadItemsWithoutCachedFindings);
			}
		}

		LOGGER.traceExit();
	}

	/**
	 * @return a map of cached {@linkplain Finding findings} for each of the given
	 *         local {@link ResourceReference}s. If no cached findings exist, the
	 *         findings are {@code null}.
	 */
	private Map<PreCommitUploadItem, List<Finding>> getCachedFindings(Set<PreCommitUploadItem> uploadItems) {
		Map<PreCommitUploadItem, List<Finding>> cachedEntries = new HashMap<>();
		uploadItems.forEach(item -> cachedEntries.put(item, getCachedFindings(item).orElse(null)));
		return cachedEntries;
	}

	/**
	 * Fetches {@linkplain Finding findings} for the given
	 * {@link ResourceReference}s from the Teamscale server by requesting and
	 * polling until the results are computed.
	 * <p>
	 * The findings from the initial request and the polling are returned from this
	 * method, and updated in the {@link IFetchFindingsTaskCache} for the local
	 * {@link ResourceReference} in the current code state. This cache update also
	 * notifies its listeners, e.g. the findings tool window and code editor.
	 */
	private void fetchPreCommitResults(Collection<PreCommitUploadItem> uploadItems) {
		fetchPreCommitResults(uploadItems, preCommitResultItems -> {
			PreCommitResult preCommitResult = preCommitResultItems.getPreCommitResult();
			if (preCommitResult.hasErrors()) {
				notificationService.reportNotifications(causeForFetchingFindings,
						Notifications.preCommitError(preCommitResult));
			}
			if (preCommitResult.hasGlobalErrors()) {
				handleGlobalErrorResults(preCommitResult, preCommitResultItems.getUploadItems());
			} else if (preCommitResult.isFinished()) {
				handleResults(preCommitResult, preCommitResultItems.getUploadItems());
			}
		});
	}

	private void handleGlobalErrorResults(PreCommitResult preCommitResult,
			Collection<PreCommitUploadItem> uploadItems) {

		for (String remotePath : preCommitResult.getAllAnalyzedPaths()) {
			PreCommitUploadItem correspondingUpload = getCorrespondingUploadItem(uploadItems, remotePath);
			List<Finding> resolvedFindings = resolveFindings(preCommitResult, correspondingUpload, remotePath);
			// We don't write the results into the cache because they have a high chance to
			// be outdated
			fetchFindingStateService.setFindingsFetchFailure(correspondingUpload.getLocalResource(),
					EFindingFetchError.API, preCommitResult.getSquashedGlobalErrors(), resolvedFindings);
		}
	}

	/**
	 * Returns the cached {@linkplain Finding findings} for the
	 * {@link PreCommitUploadItem}s, i.e. for the respective local
	 * {@link ResourceReference}s with the code state of the upload item.
	 */
	private Optional<? extends List<Finding>> getCachedFindings(PreCommitUploadItem uploadItem) {
		String baseBranch = uploadItem.getRemoteResource().getBranch().orElse(null);
		return findingsCache.getFindings(baseBranch, uploadItem.getLocalResource(), uploadItem.getContentHash());
	}

	/**
	 * Updates the {@linkplain Finding findings} in {@link IFetchFindingsTaskCache}
	 * for each
	 */
	private void handleResults(PreCommitResult preCommitResult, Collection<PreCommitUploadItem> uploadItems) {
		for (String remotePath : preCommitResult.getAnalyzedPathsWithFindings()) {
			PreCommitUploadItem correspondingUpload = getCorrespondingUploadItem(uploadItems, remotePath);
			List<Finding> resolvedFindings = resolveFindings(preCommitResult, correspondingUpload, remotePath);
			if (cacheBehaviour.shouldWrite()) {
				updateCache(correspondingUpload, resolvedFindings);
			}
			fetchFindingStateService.setFindingsFetchSuccessful(correspondingUpload.getLocalResource(),
					resolvedFindings);
		}
		for (String remotePath : preCommitResult.getAnalyzedPathsWithErrors()) {
			PreCommitUploadItem correspondingUpload = getCorrespondingUploadItem(uploadItems, remotePath);
			List<Finding> resolvedFindings = resolveFindings(preCommitResult, correspondingUpload, remotePath);
			// We don't write the results into the cache because they have a high chance to
			// be outdated
			fetchFindingStateService.setFindingsFetchFailure(correspondingUpload.getLocalResource(),
					EFindingFetchError.API, preCommitResult.getSquashedErrors(remotePath), resolvedFindings);
		}
	}

	private void updateCache(PreCommitUploadItem uploadItem, List<Finding> findings) {
		String baseBranch = uploadItem.getRemoteResource().getBranch().orElse(null);
		findingsCache.updateFindings(baseBranch, uploadItem.getLocalResource(), uploadItem.getContentHash(), findings);
	}

	private static PreCommitUploadItem getCorrespondingUploadItem(Collection<PreCommitUploadItem> uploadItems,
			String remotePath) {
		return uploadItems.stream()
				.filter(item -> item.getRemoteResource().getUniformPath().toString().equals(remotePath))
				.collect(onlyElement());
	}

	private List<Finding> resolveFindings(PreCommitResult preCommitResult, PreCommitUploadItem uploadItem,
			String remotePath) {
		List<Finding> result = new ArrayList<>();
		Collection<TrackedFinding> findingDtos = preCommitResult.getFindings(remotePath);

		for (TrackedFinding findingDto : findingDtos) {
			UniformPath uniformPath = UniformPathCompatibilityUtil.convert(findingDto.getLocation().getUniformPath());
			Preconditions.checkArgument(Objects.equals(uniformPath, uploadItem.getRemoteResource().getUniformPath()),
					"UniformPath of FindingDto and RemoteResource need to match.");

			String baseBranch = null;
			CommitDescriptor birthCommit = findingDto.getBirthCommit();
			if (birthCommit != null && !birthCommit.isUnbranched()) {
				baseBranch = birthCommit.getBranchName();
			}
			String expectedPrecommitBranchName = baseBranch;
			if (!PreCommitUtils.isPreCommitBranch(baseBranch)) {
				String username = credentialsManager.getServer(uploadItem.getRemoteResource().getServerUrl())
						.orElseThrow(IllegalStateException::new).getUsername();
				expectedPrecommitBranchName = PreCommitUtils.getPreCommit3BranchName(username, baseBranch);
			}
			// we have to replace the remote resource at this point since it now points to
			// the pre-commit branch. The findings (and the file content) of this resource
			// now correspond to the file content on the precommit branch, not to the
			// findings/content on the base branch.
			RemoteResourceReference resourceOnPreCommitBranch = uploadItem.getRemoteResource()
					.withNewBranch(expectedPrecommitBranchName);
			FindingUtils
					.extractFinding(findingDto, baseBranch, uploadItem.getLocalResource(), resourceOnPreCommitBranch)
					.ifPresent(result::add);
		}

		return result;
	}

	@Override
	protected PreCommitUploadData buildPreCommitUploadData(List<PreCommitUploadItem> uploadItems) {
		Map<UniformPath, String> uniformPathToContentMap = new HashMap<>();
		for (PreCommitUploadItem uploadItem : uploadItems) {
			String content = resourceResolutionService.readContents(uploadItem.getLocalResource()).orElse(null);
			uniformPathToContentMap.put(uploadItem.getRemoteResource().getUniformPath(), content);
		}
		return new PreCommitUploadData(uniformPathToContentMap, Collections.emptyList());
	}

	@Override
	protected void onFetchError(ResourceReference localResourceReference, EFindingFetchError error, String message) {
		fetchFindingStateService.setFindingsFetchFailure(localResourceReference, error, message);
	}

	@Override
	protected void onTaskCanceled(ResourceReference resourceReference) {
		fetchFindingStateService.setFindingsFetchFailure(resourceReference, EFindingFetchError.CANCELED,
				"Fetching findings canceled.");
	}
}
