/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings.tasks.fetcher;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.conqat.engine.index.shared.PreCommitUploadData;
import org.conqat.lib.commons.uniformpath.UniformPath;

import com.google.common.base.Preconditions;
import com.teamscale.ide.commons.findings.PreCommitResult;
import com.teamscale.ide.commons.findings.PreCommitUploadItem;
import com.teamscale.ide.commons.findings.PreCommitUploadTarget;
import com.teamscale.ide.commons.findings.services.IFetchFindingsTaskCache;
import com.teamscale.ide.commons.findings.tasks.FetchFindingsTask;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.notification.Notifications;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.TaskBase;

/**
 * Encapsulates logic for a {@link FetchFindingsTask} to delete the given
 * {@linkplain ResourceReference file references} on the pre-commit branch.
 * <p>
 * Collections of {@linkplain PreCommitUploadItem upload items} are built and
 * uploaded to the respective {@linkplain PreCommitUploadTarget upload targets}.
 * If the results are pending on the Teamscale Server, the task polls until
 * finished.
 * 
 * @see PreCommitFileFindingsFetcher fetching non-deletion pre-commit findings
 */
public class PreCommitFindingsDeleteFetcher extends PreCommitFindingsFetcherBase {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(PreCommitFindingsDeleteFetcher.class);

	/** Pre-commit file content for deletion of file. */
	private static final String CONTENT_FOR_DELETION = null;

	/** The findings cache */
	private final IFetchFindingsTaskCache findingsCache;

	public PreCommitFindingsDeleteFetcher(IServiceRegistry serviceRegistry,
			Collection<ResourceReference> fetchResources, ENotificationCause causeForFetchingFindings,
			int pollingIntervalInSec, TaskBase ownerTask) {
		super(serviceRegistry, fetchResources, causeForFetchingFindings, pollingIntervalInSec, ownerTask);
		Preconditions.checkArgument(fetchResources.stream().allMatch(ResourceReference::isFile),
				"The PreCommitFindingsDeleteFetcher does not support folders as ResourceReference.");
		findingsCache = serviceRegistry.getService(IFetchFindingsTaskCache.class);
	}

	/**
	 * Sends deletion request to Teamscale's pre-commit service for the given
	 * {@linkplain ResourceReference references}.
	 */
	@Override
	public void doFetchFindings(Set<ResourceReference> deletedFileResources) {
		LOGGER.traceEntry("Send delete requests for selected resources {}.", deletedFileResources);
		Set<PreCommitUploadItem> uploadItems = buildUploadItems(deletedFileResources);
		sendDeletionRequest(uploadItems);
		LOGGER.traceExit();
	}

	/**
	 * Sends a deletion request for the given {@linkplain PreCommitUploadItem upload
	 * items}.
	 */
	private void sendDeletionRequest(Collection<PreCommitUploadItem> uploadItems) {
		fetchPreCommitResults(uploadItems, resultItems -> {
			PreCommitResult result = resultItems.getPreCommitResult();
			if (result.hasGlobalErrors()) {
				notificationService.reportNotifications(causeForFetchingFindings, Notifications.preCommitError(result));
			} else if (!result.getAnalyzedPathsWithFindings().isEmpty()) {
				// todo: workaround to avoid false positives in other files than the deleted
				// ones (see TS-31895)
				findingsCache.invalidateAll();
				fetchFindingStateService.setAllFindingsFetchOutdated();
			}
			// TODO TS-32932 Report path specific problems to the user
		});
	}

	@Override
	protected PreCommitUploadData buildPreCommitUploadData(List<PreCommitUploadItem> uploadItems) {
		Map<UniformPath, String> uniformPathToContentMap = new HashMap<>();
		for (PreCommitUploadItem uploadItem : uploadItems) {
			uniformPathToContentMap.put(uploadItem.getRemoteResource().getUniformPath(), CONTENT_FOR_DELETION);
		}
		return new PreCommitUploadData(uniformPathToContentMap, Collections.emptyList());
	}
}
