/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.findings.tasks.fetcher;

import static java.util.concurrent.TimeUnit.SECONDS;
import static java.util.stream.Collectors.groupingBy;

import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.function.Consumer;
import java.util.stream.Collectors;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.checkerframework.checker.nullness.qual.NonNull;
import org.checkerframework.checker.nullness.qual.Nullable;
import org.conqat.engine.index.shared.PreCommit3Result;
import org.conqat.engine.index.shared.PreCommitUploadData;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Maps;
import com.google.common.hash.HashCode;
import com.google.common.hash.HashFunction;
import com.google.common.hash.Hashing;
import com.teamscale.ide.commons.client.RestClientUtils;
import com.teamscale.ide.commons.client.exceptions.ApiNotSupportedException;
import com.teamscale.ide.commons.client.exceptions.HttpException;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.exceptions.ProjectNotConfiguredCorrectlyException;
import com.teamscale.ide.commons.client.exceptions.ProjectNotFoundException;
import com.teamscale.ide.commons.client.services.ITeamscaleConnectorService;
import com.teamscale.ide.commons.findings.EFindingFetchError;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.PreCommitResult;
import com.teamscale.ide.commons.findings.PreCommitUploadItem;
import com.teamscale.ide.commons.findings.PreCommitUploadTarget;
import com.teamscale.ide.commons.findings.services.IFindingFetchStateService;
import com.teamscale.ide.commons.findings.services.IFindingsListener;
import com.teamscale.ide.commons.findings.tasks.FetchFindingsTask;
import com.teamscale.ide.commons.mappings.MappingUtils;
import com.teamscale.ide.commons.mappings.ResourceMappingResult;
import com.teamscale.ide.commons.mappings.services.IResourceMappingService;
import com.teamscale.ide.commons.model.RemoteResourceReference;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.model.VcsResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.notification.INotification;
import com.teamscale.ide.commons.notification.Notifications;
import com.teamscale.ide.commons.notification.services.INotificationService;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.IResourceResolutionService;
import com.teamscale.ide.commons.plugin.services.TaskBase;
import com.teamscale.ide.commons.servers.NamedServerDetails;
import com.teamscale.ide.commons.servers.services.IServerCredentialsManager;
import com.teamscale.ide.commons.vcs.services.ICheckoutFactory;

/**
 * Encapsulates logic for a {@link FetchFindingsTask} to fetch pre-commit
 * findings for the given {@linkplain ResourceReference file references}.
 * <p>
 * Collections of {@linkplain PreCommitUploadItem upload items} are built and
 * uploaded to the respective {@linkplain PreCommitUploadTarget upload targets}.
 * If the results are pending on the Teamscale Server, the task polls until
 * finished.
 * <p>
 * Whenever findings are available (e.g. on cache hits, on initial requests, or
 * on final polling requests), the {@link IFindingsListener} registered
 * listeners of the {@link FetchFindingsTask} get informed.
 * <p>
 *
 * @see LiveFileFindingsFetcher fetching non-pre-commit findings
 */
/* package */ abstract class PreCommitFindingsFetcherBase extends FindingsFetcherBase {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	protected static final Logger LOGGER = LogManager.getLogger(PreCommitFindingsFetcherBase.class);

	/** Hash function for file content. */
	private static final HashFunction HASHER = Hashing.goodFastHash(128);

	/**
	 * The error message that should be shown if the API for pre-commit is not
	 * supported.
	 */
	private static final String PRE_COMMIT_API_NOT_SUPPORTED_ERROR_MESSAGE = "Pre-commit findings could not be retrieved "
			+ "because the server does not support this feature";

	/**
	 * The polling interval in seconds. Must be configurable to reduce test times.
	 **/
	private final int pollingIntervalInSec;

	/** The notification service */
	protected final INotificationService notificationService;
	/** The checkout factory */
	protected final ICheckoutFactory checkoutFactory;
	/** The server manager */
	protected final IServerCredentialsManager credentialsManager;
	/** The connector service */
	protected final ITeamscaleConnectorService connectorService;
	/** The resource mapping service */
	protected final IResourceMappingService resourceMappingService;

	/** The resource resolution service */
	protected final IResourceResolutionService resourceResolutionService;

	/** The fetch finding state service **/
	protected final IFindingFetchStateService fetchFindingStateService;

	/* package */ PreCommitFindingsFetcherBase(IServiceRegistry serviceRegistry,
			Collection<ResourceReference> fetchResources, ENotificationCause causeForFetchingFindings,
			int pollingIntervalInSec, TaskBase ownerTask) {
		super(fetchResources, causeForFetchingFindings, ownerTask);
		Preconditions.checkArgument(fetchResources.stream().allMatch(ResourceReference::isFile),
				"The PreCommitFindingsFetcher does not support folders as ResourceReference.");

		this.pollingIntervalInSec = pollingIntervalInSec;

		notificationService = serviceRegistry.getService(INotificationService.class);
		checkoutFactory = serviceRegistry.getService(ICheckoutFactory.class);
		credentialsManager = serviceRegistry.getService(IServerCredentialsManager.class);
		connectorService = serviceRegistry.getService(ITeamscaleConnectorService.class);
		resourceMappingService = serviceRegistry.getService(IResourceMappingService.class);
		resourceResolutionService = serviceRegistry.getService(IResourceResolutionService.class);
		fetchFindingStateService = serviceRegistry.getService(IFindingFetchStateService.class);
	}

	// static helper methods

	/** Returns if the given list has valid findings */
	protected static boolean hasFindings(List<Finding> findings) {
		return findings != null;
	}

	/** Returns if the given list has no valid findings */
	protected static boolean hasNoFindings(List<Finding> findings) {
		return !hasFindings(findings);
	}

	/**
	 * Fetches {@link PreCommit3Result}s for the given {@link PreCommitUploadItem}s
	 * from the Teamscale server by requesting and polling until the results are
	 * computed.
	 * <p>
	 * The given callback is called whenever {@link PreCommit3Result}s are
	 * available, but not for pending results where polling is necessary.
	 */
	protected final void fetchPreCommitResults(Collection<PreCommitUploadItem> uploadItems,
			Consumer<PreCommitResultItems> resultCallback) {
		LOGGER.traceEntry(null, uploadItems);
		// request pre-commit results
		Map<PreCommitUploadTarget, PreCommitResultItems> preCommitTargetsAndResults = requestPreCommitResults(
				uploadItems);
		LOGGER.debug("Pre-Commit results: {}", preCommitTargetsAndResults);

		// update (non-pending) findings in cache, which informs listeners
		preCommitTargetsAndResults.values().stream() //
				.filter(result -> !result.isPending()) //
				.forEach(resultCallback);

		// poll pending results, if available
		Map<PreCommitUploadTarget, PreCommitResultItems> pendingPreCommitTargetsAndResults = Maps
				.filterValues(preCommitTargetsAndResults, PreCommitResultItems::isPending);
		try {
			Collection<PreCommitResultItems> preCommitResults = pollPreCommitResults(pendingPreCommitTargetsAndResults);
			preCommitResults.forEach(resultCallback);
		} catch (InterruptedException e) {
			LOGGER.catching(e);
		} finally {
			LOGGER.traceExit();
		}
	}

	/**
	 * Requests {@linkplain PreCommit3Result pre-commit results} for the given
	 * {@linkplain PreCommitUploadItem upload items}.
	 *
	 * @return a {@link PreCommit3Result} for each identified
	 *         {@link PreCommitUploadTarget}
	 */
	private Map<PreCommitUploadTarget, PreCommitResultItems> requestPreCommitResults(
			Collection<PreCommitUploadItem> uploadItems) {
		ImmutableMap.Builder<PreCommitUploadTarget, PreCommitResultItems> preCommitTargetsAndResults = ImmutableMap
				.builder();
		Map<PreCommitUploadTarget, List<PreCommitUploadItem>> uploadTargetsAndItems = bundleUploadItems(uploadItems);
		uploadTargetsAndItems.forEach((uploadTarget, uploadItem) -> requestPreCommitResult(preCommitTargetsAndResults,
				uploadTarget, uploadItem));

		return preCommitTargetsAndResults.build();
	}

	/**
	 * Bundles and returns {@linkplain PreCommitUploadItem upload items} based on
	 * their respective {@linkplain PreCommitUploadTarget upload targets}, i.e. a
	 * specific combination of Teamscale server, project, branch, and commit at
	 * which the pre-commit request will be rooted for the upload items.
	 */
	private static Map<PreCommitUploadTarget, List<PreCommitUploadItem>> bundleUploadItems(
			Collection<PreCommitUploadItem> uploadItems) {
		return uploadItems.stream().collect(groupingBy(PreCommitFindingsFetcherBase::determineUploadTarget));
	}

	/**
	 * Determines the {@link PreCommitUploadTarget} by using the
	 * {@link PreCommitUploadItem}`s {@link RemoteResourceReference}.
	 */
	private static PreCommitUploadTarget determineUploadTarget(PreCommitUploadItem uploadItem) {
		RemoteResourceReference remoteResourceReference = uploadItem.getRemoteResource();
		return new PreCommitUploadTarget(remoteResourceReference.getServerUrl(),
				remoteResourceReference.getTeamscaleProject(), remoteResourceReference.getBranch().orElse(null));
	}

	private void requestPreCommitResult(
			ImmutableMap.Builder<PreCommitUploadTarget, PreCommitResultItems> preCommitTargetsAndResults,
			PreCommitUploadTarget uploadTarget, List<PreCommitUploadItem> uploadItems) {
		LOGGER.traceEntry("Requesting pre-commit results for {} @ {} ", uploadTarget.getProjectId(),
				uploadTarget.getServerUrl());
		try {
			PreCommitUploadData preCommitUploadData = buildPreCommitUploadData(uploadItems);
			PreCommitResult result = sendPreCommitRequest(uploadTarget, preCommitUploadData);
			preCommitTargetsAndResults.put(uploadTarget, new PreCommitResultItems(result, uploadItems));
		} catch (NetworkExceptionBase e) {
			LOGGER.catching(e);
			handleNetworkException(e, uploadTarget.getServerUrl(), uploadItems);
		} catch (ProjectNotFoundException e) {
			LOGGER.catching(e);
			handleException(Notifications.projectNotFound(uploadTarget.getServerUrl(), uploadTarget.getProjectId()),
					uploadItems, EFindingFetchError.NETWORK, e.getMessage());
		} catch (ApiNotSupportedException e) {
			LOGGER.catching(e);
			handleApiNotSupportedException(e, uploadTarget.getServerUrl(), uploadItems);
		}
		LOGGER.traceExit();
	}

	/**
	 * Builds the {@link PreCommitUploadData} for the given
	 * {@linkplain PreCommitUploadItem upload items} to request the
	 * {@link PreCommit3Result}.
	 */
	protected abstract PreCommitUploadData buildPreCommitUploadData(List<PreCommitUploadItem> uploadItems);

	/**
	 * Callback if the fetching process started with
	 * {@link PreCommitFindingsFetcherBase#fetchPreCommitResults(Collection, Consumer)}
	 * fails for the given {@link ResourceReference local resource}.
	 */
	protected void onFetchError(ResourceReference localResourceReference, EFindingFetchError error, String message) {
		// Default ignoring
	}

	/**
	 * Builds and returns {@link PreCommitUploadItem}s for the given local
	 * {@link ResourceReference}s.
	 * <p>
	 * In case of folder references, the method (recursively) transforms these to
	 * multiple file references contained in the folder.
	 */
	protected final Set<PreCommitUploadItem> buildUploadItems(Collection<ResourceReference> fetchFileResources) {
		LOGGER.traceEntry(null, fetchFileResources);

		ImmutableSet.Builder<PreCommitUploadItem> uploadItems = ImmutableSet.builder();
		for (ResourceReference localResource : fetchFileResources) {
			ResourceMappingResult mappingResult = resourceMappingService.map(localResource);
			if (!mappingResult.isSuccessful()) {
				notificationService.reportNotification(causeForFetchingFindings,
						Notifications.mappingProblem(localResource, mappingResult.getMappingFailure()));
				fetchFindingStateService.setFindingsFetchFailure(localResource, EFindingFetchError.MAPPING,
						"Mapping problem.");
				continue;
			}
			String baseBranch = MappingUtils.getBranch(checkoutFactory, localResource, mappingResult).orElse(null);
			RemoteResourceReference remoteResource = mappingResult
					.getMappedResource(new VcsResourceReference(localResource, baseBranch));
			HashCode hashCode = generateContentHashKey(baseBranch, localResource);
			uploadItems.add(new PreCommitUploadItem(localResource, remoteResource, hashCode.toString()));
		}

		return LOGGER.traceExit(uploadItems.build());
	}

	/**
	 * Requests a {@link PreCommit3Result} from the given
	 * {@link PreCommitUploadTarget} for the given {@linkplain PreCommitUploadItem
	 * upload items}.
	 *
	 * @return a {@link PreCommit3Result}, or {@code null} if the used Teamscale
	 *         server is not configured
	 * @throws NetworkExceptionBase
	 *             on errors during the request service call
	 */
	@NonNull
	private PreCommitResult sendPreCommitRequest(PreCommitUploadTarget uploadTarget,
			PreCommitUploadData preCommitUploadData)
			throws NetworkExceptionBase, ApiNotSupportedException, ProjectNotFoundException {
		NamedServerDetails server = credentialsManager.getServer(uploadTarget.getServerUrl()).orElse(null);
		if (server == null) {
			notificationService.reportNotification(causeForFetchingFindings,
					Notifications.serverUrlNotFound(uploadTarget.getServerUrl()));
			throw new ProjectNotConfiguredCorrectlyException(
					"Server URL not found: " + uploadTarget.getServerUrl().toString());
		}

		PreCommitResult preCommitResult = connectorService.requestPreCommit3Result(server, uploadTarget.getProjectId(),
				uploadTarget.getBranch(), preCommitUploadData);

		if (preCommitResult.isSuccessful()) {
			LOGGER.debug("Got {} findings.", preCommitResult.getAllFindings().size());
		} else if (preCommitResult.isPending()) {
			LOGGER.debug("Got token: {}.", preCommitResult.getToken());
		} else if (preCommitResult.hasGlobalErrors()) {
			LOGGER.error("Error while fetching pre-commit findings: {}",
					preCommitResult.getSquashedGlobalErrors().replace("\n", "<br>"));
		}

		return preCommitResult;
	}

	/**
	 * Polls pending {@linkplain PreCommit3Result pre-commit results} until
	 * Teamscale finishes the respective analysis.
	 */
	private Collection<PreCommitResultItems> pollPreCommitResults(
			@Nullable Map<PreCommitUploadTarget, PreCommitResultItems> preCommitTargetsAndResults)
			throws InterruptedException {
		if (preCommitTargetsAndResults == null || preCommitTargetsAndResults.isEmpty()) {
			return Collections.emptyList();
		}

		Map<PreCommitUploadTarget, PreCommitResultItems> workingPreCommitTargetsAndResults = new HashMap<>(
				preCommitTargetsAndResults);

		try {
			return RestClientUtils.pollRepeatedly(() -> pollPreCommitResultsAttempt(workingPreCommitTargetsAndResults),
					pollingIntervalInSec, SECONDS);
		} catch (NetworkExceptionBase e) {
			// Something has gone really wrong. Should never happen.
			LOGGER.catching(e);
			preCommitTargetsAndResults.values().stream().flatMap(items -> items.uploadItems.stream())
					.forEach(item -> onFetchError(item.getLocalResource(), EFindingFetchError.NETWORK,
							e.getServerState().getDescription()));
			return Collections.emptyList();
		}
	}

	private Optional<Collection<PreCommitResultItems>> pollPreCommitResultsAttempt(
			Map<PreCommitUploadTarget, PreCommitResultItems> workingPreCommitTargetsAndResults)
			throws InterruptedException {
		checkIfTaskCanceled(workingPreCommitTargetsAndResults.values().stream()
				.map(PreCommitResultItems::getUploadItems).flatMap(Collection::stream)
				.map(PreCommitUploadItem::getLocalResource).collect(Collectors.toSet()));

		if (workingPreCommitTargetsAndResults.values().stream().noneMatch(PreCommitResultItems::isPending)) {
			// results finished > stop polling
			return Optional.of(workingPreCommitTargetsAndResults.values());
		}

		for (Map.Entry<PreCommitUploadTarget, PreCommitResultItems> targetAndResult : workingPreCommitTargetsAndResults
				.entrySet()) {
			PreCommitResultItems oldResults = targetAndResult.getValue();
			PreCommitResult preCommitResult = oldResults.getPreCommitResult();
			if (preCommitResult.isFinished()) {
				continue;
			}

			List<String> allUniformPaths = targetAndResult.getValue().getUploadItems().stream()
					.map(item -> item.getRemoteResource().getUniformPath().toString()).collect(Collectors.toList());

			PreCommitUploadTarget uploadTarget = targetAndResult.getKey();
			PreCommitResult newResult = pollPreCommitResult(uploadTarget, oldResults, preCommitResult.getToken(),
					allUniformPaths);

			LOGGER.debug("Pre-Commit polling results: {}", newResult);
			workingPreCommitTargetsAndResults.replace(uploadTarget,
					new PreCommitResultItems(newResult, oldResults.getUploadItems()));
		}

		// continue polling
		return Optional.empty();
	}

	/**
	 * Requests polling result from the Teamscale server for the given polling
	 * token.
	 */
	private PreCommitResult pollPreCommitResult(PreCommitUploadTarget uploadTarget, PreCommitResultItems oldResultItems,
			String token, Collection<String> allUniformPaths) {
		LOGGER.traceEntry("Attempting to poll pre-commit results for {} @ {} (Token: {})", uploadTarget.getProjectId(),
				uploadTarget.getServerUrl(), token);
		NamedServerDetails server = credentialsManager.getServer(uploadTarget.getServerUrl()).orElse(null);
		if (server == null) {
			notificationService.reportNotification(causeForFetchingFindings,
					Notifications.serverUrlNotFound(uploadTarget.getServerUrl()));
			String error = String.format("Server %s not reachable", uploadTarget.getServerUrl());
			return LOGGER.traceExit(PreCommitResult.finished()
					.addGlobalError(PreCommitResult.PreCommitErrorDetail.SERVER_NOT_REACHABLE, error, true).build());
		}

		return pollPreCommitResult(server, uploadTarget, oldResultItems, token, allUniformPaths);
	}

	private PreCommitResult pollPreCommitResult(NamedServerDetails server, PreCommitUploadTarget uploadTarget,
			PreCommitResultItems oldResultItems, String token, Collection<String> allUniformPaths) {
		LOGGER.traceEntry("Attempting to poll pre-commit results for {} @ {} (Token: {})", uploadTarget.getProjectId(),
				uploadTarget.getServerUrl(), token);
		PreCommitResult result;
		try {
			result = connectorService.pollPreCommit3Result(server, uploadTarget.getProjectId(), token, allUniformPaths);
		} catch (NetworkExceptionBase e) {
			LOGGER.catching(e);
			handleNetworkException(e, uploadTarget.getServerUrl(), oldResultItems.uploadItems);
			result = PreCommitResult.finished()
					.addGlobalError(PreCommitResult.PreCommitErrorDetail.SERVER_NOT_REACHABLE,
							e.getServerState().getDescription(), true)
					.build();
		} catch (ApiNotSupportedException e) {
			LOGGER.catching(e);
			handleApiNotSupportedException(e, uploadTarget.getServerUrl(), oldResultItems.uploadItems);
			result = PreCommitResult.finished()
					.addGlobalError(PreCommitResult.PreCommitErrorDetail.PRE_COMMIT_API_NOT_SUPPORTED,
							PRE_COMMIT_API_NOT_SUPPORTED_ERROR_MESSAGE, true)
					.build();
		}

		return LOGGER.traceExit(result);
	}

	/**
	 * Handles {@link NetworkExceptionBase network exceptions} while
	 * requesting/polling pre-commit results.
	 *
	 * @see #handleException(INotification, List, EFindingFetchError, String)
	 */
	private void handleNetworkException(NetworkExceptionBase exception, URI serverUrl,
			List<PreCommitUploadItem> uploadItems) {
		String message = exception.getMessage();
		if (exception instanceof HttpException) {
			message = ((HttpException) exception).getStatusMessage();
		}

		handleException(Notifications.connectionError(serverUrl, exception.getServerState()), uploadItems,
				EFindingFetchError.NETWORK, message);
	}

	/**
	 * Handles {@link ApiNotSupportedException}s while requesting/polling pre-commit
	 * results.
	 *
	 * @see #handleException(INotification, List, EFindingFetchError, String)
	 */
	private void handleApiNotSupportedException(ApiNotSupportedException exception, URI serverUrl,
			List<PreCommitUploadItem> uploadItems) {
		handleException(
				Notifications.apiNotSupported(serverUrl, exception.getLowestSupportedServerVersion(),
						PRE_COMMIT_API_NOT_SUPPORTED_ERROR_MESSAGE),
				uploadItems, EFindingFetchError.API, PRE_COMMIT_API_NOT_SUPPORTED_ERROR_MESSAGE);
	}

	/** Handles exceptions while requesting/polling pre-commit results. */
	private void handleException(INotification<?> notification, List<PreCommitUploadItem> uploadItems,
			EFindingFetchError errorType, String message) {
		notificationService.reportNotification(causeForFetchingFindings, notification);
		uploadItems.forEach(item -> onFetchError(item.getLocalResource(), errorType, message));
	}

	/**
	 * @return the content hash for the given local {@link ResourceReference} on the
	 *         given branch.
	 */
	private HashCode generateContentHashKey(@Nullable String branch, ResourceReference localResource) {
		StringBuilder contentKey = new StringBuilder();
		if (branch != null) {
			contentKey.append(branch);
		}
		contentKey.append(resourceResolutionService.readContents(localResource).orElse(""));
		return HASHER.hashBytes(contentKey.toString().getBytes(StandardCharsets.UTF_8));
	}

	/**
	 * A container object to group {@link PreCommit3Result} and the corresponding
	 * {@link PreCommitUploadItem}s.
	 */
	protected static class PreCommitResultItems {

		private final PreCommitResult preCommitResult;
		private final List<PreCommitUploadItem> uploadItems;

		private PreCommitResultItems(PreCommitResult preCommitResult, List<PreCommitUploadItem> uploadItems) {
			this.preCommitResult = preCommitResult;
			this.uploadItems = uploadItems;
		}

		PreCommitResult getPreCommitResult() {
			return preCommitResult;
		}

		List<PreCommitUploadItem> getUploadItems() {
			return uploadItems;
		}

		/**
		 * See {@link PreCommitResult#isPending()}.
		 */
		private boolean isPending() {
			return preCommitResult.isPending();
		}
	}
}
