/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.logging;

import java.util.Objects;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.Marker;

import com.google.common.base.Preconditions;
import com.teamscale.ide.commons.lang.ToStringHelpers;

/** The condition part of a {@link Rule}. */
public final class Condition {

	/** The logger name. May be {@code null}. */
	private final String logger;

	/** The marker. May be {@code null}. */
	private final Marker marker;

	/** The log level. Never {@code null}. */
	private final Level level;

	public Condition(String logger, Marker marker, Level level) {
		Preconditions.checkNotNull(level);

		this.logger = logger;
		this.marker = marker;
		this.level = level;
	}

	/** @see #logger */
	public String getLogger() {
		return logger;
	}

	/** @see #marker */
	public Marker getMarker() {
		return marker;
	}

	/** @see #level */
	public Level getLevel() {
		return level;
	}

	/**
	 * Finds the given rule {@linkplain Rule#matches(Condition) matching} this
	 * condition, if any.
	 */
	public Optional<Rule> findMatching(Rule... rules) {
		return Stream.of(rules).filter(rule -> rule.matches(this)).findFirst();
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		Condition that = (Condition) other;
		return Objects.equals(this.level, that.level) && Objects.equals(this.logger, that.logger)
				&& Objects.equals(this.marker, that.marker);
	}

	@Override
	public int hashCode() {
		return Objects.hash(level, logger, marker);
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
