/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.logging;

import java.util.Objects;

import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.Marker;
import org.apache.logging.log4j.core.LogEvent;
import org.apache.logging.log4j.core.config.Node;
import org.apache.logging.log4j.core.config.plugins.Plugin;
import org.apache.logging.log4j.core.config.plugins.PluginBuilderAttribute;
import org.apache.logging.log4j.core.config.plugins.PluginBuilderFactory;
import org.apache.logging.log4j.core.config.plugins.validation.constraints.Required;
import org.apache.logging.log4j.core.util.Builder;
import org.apache.logging.log4j.core.util.NameUtil;

import com.google.common.base.MoreObjects;
import com.google.common.base.Preconditions;

/**
 * A rule defines how {@link LogEvent}s are mapped to IDE-specific keys based on
 * their {@linkplain #logger logger}, {@linkplain #marker marker} (if any), and
 * {@linkplain #level log level}.
 */
@Plugin(name = "Rule", category = Node.CATEGORY, printObject = true)
public final class Rule {

	/**
	 * The IDE-specific key that corresponds to the {@link #logger},
	 * {@link #marker}, and {@link #level}.
	 */
	private final String key;

	/**
	 * The logger log events must be logged with to match this rule. May be
	 * {@literal null}, in which case this rules matches events with any logger.
	 */
	private final String logger;

	/**
	 * The {@link Marker} log events must be an instance of to match this rule. May
	 * be {@literal null}, in which case this rules matches events with any marker.
	 */
	private final String marker;

	/**
	 * The {@link Level} log event must at least be logged at to match this rule.
	 * May be {@literal null}, in which case this rules matches events with any
	 * level.
	 */
	private final Level level;

	private Rule(String key, String logger, String marker, Level level) {
		Preconditions.checkNotNull(key);

		this.key = key;
		this.logger = logger;
		this.marker = marker;
		this.level = level;
	}

	/** @see #key */
	public String getKey() {
		return key;
	}

	/** @see #logger */
	public String getLogger() {
		return logger;
	}

	/** @see #marker */
	public String getMarker() {
		return marker;
	}

	/** @see #level */
	public Level getLevel() {
		return level;
	}

	/** @return whether this rule matches the given {@link Condition}. */
	public boolean matches(Condition condition) {
		return matchesLevel(condition) && matchesMarker(condition) && matchesLogger(condition);
	}

	private boolean matchesLevel(Condition condition) {
		if (level == null) {
			return true;
		}
		return level.isLessSpecificThan(condition.getLevel());
	}

	private boolean matchesMarker(Condition condition) {
		if (marker == null) {
			return true;
		}
		return condition.getMarker() != null && condition.getMarker().isInstanceOf(marker);
	}

	private boolean matchesLogger(Condition condition) {
		if (logger == null) {
			return true;
		}
		String name = condition.getLogger();
		while (name != null && !name.equals(logger)) {
			name = NameUtil.getSubName(name);
		}

		return name != null;
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		Rule that = (Rule) other;
		return Objects.equals(this.key, that.key) && Objects.equals(this.logger, that.logger)
				&& Objects.equals(this.marker, that.marker) && Objects.equals(this.level, that.level);
	}

	@Override
	public int hashCode() {
		return Objects.hash(key, logger, marker, level);
	}

	@Override
	public String toString() {
		return MoreObjects.toStringHelper(this).add("key", key).add("level", level).add("logger", logger)
				.add("marker", marker).toString();
	}

	/**
	 * @return a builder to construct new instances of {@link Rule}
	 */
	@PluginBuilderFactory
	public static RuleBuilder newBuilder() {
		return new RuleBuilder();
	}

	/**
	 * A builder to construct new instances of {@link Rule}. Used by Log4J during
	 * plug-in configuration.
	 */
	public static class RuleBuilder implements Builder<Rule> {

		@PluginBuilderAttribute("key")
		@Required(message = "No key provided")
		private String key;

		@PluginBuilderAttribute("logger")
		private String logger;

		@PluginBuilderAttribute("marker")
		private String marker;

		@PluginBuilderAttribute("level")
		private Level level;

		/** @see Rule#key */
		public RuleBuilder setKey(String key) {
			this.key = key;
			return this;
		}

		/** @see Rule#logger */
		public RuleBuilder setLogger(String logger) {
			this.logger = logger;
			return this;
		}

		/** @see Rule#marker */
		public RuleBuilder setMarker(String marker) {
			this.marker = marker;
			return this;
		}

		/** @see Rule#level */
		public RuleBuilder setLevel(Level level) {
			this.level = level;
			return this;
		}

		@Override
		public Rule build() {
			return new Rule(key, logger, marker, level);
		}
	}
}
