/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.mappings;

import static java.util.Comparator.comparing;

import java.io.IOException;
import java.net.URI;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.checkerframework.checker.nullness.qual.Nullable;
import org.conqat.engine.resource.util.UniformPathUtils;
import org.conqat.lib.commons.uniformpath.RelativeUniformPath;
import org.conqat.lib.commons.uniformpath.UniformPath;
import org.conqat.lib.commons.uniformpath.UniformPathCompatibilityUtil;

import com.google.common.base.Preconditions;
import com.google.common.base.Strings;
import com.google.common.collect.ImmutableList;
import com.teamscale.ide.commons.lang.ToStringHelpers;
import com.teamscale.ide.commons.model.IIDEProject;
import com.teamscale.ide.commons.model.PrefixMapping;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.model.ResourceReferenceUtils;

/**
 * Container for all settings relevant to an {@link IIDEProject}.
 */
public class IDEProjectSettings implements ILegacyResourceMapping {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(IDEProjectSettings.class);

	/**
	 * Whether Teamscale is enabled for this project. Defaults to false.
	 */
	private boolean teamscaleEnabled = false;

	/** The Teamscale Server in use (may be null) */
	@Nullable
	@Deprecated
	private String teamscaleServer;

	/** The Teamscale Server url */
	@Nullable
	private URI url;

	/** The Teamscale project in use (may be null) */
	@Nullable
	private String teamscaleProject;

	/**
	 * {@link PrefixMapping}s for access by remote Teamscale prefix.
	 */
	private final Map<String, PrefixMapping> teamscaleToLocalMap = new HashMap<>();

	/**
	 * {@link PrefixMapping}s for access by local prefix.
	 */
	private final Map<String, PrefixMapping> localToTeamscaleMap = new HashMap<>();

	@Override
	public boolean isTeamscaleEnabled() {
		return teamscaleEnabled;
	}

	/**
	 * Sets whether Teamscale is enabled for the ide project.
	 */
	public void setTeamscaleEnabled(boolean enabled) {
		teamscaleEnabled = enabled;
	}

	/**
	 * Returns the Teamscale server used to load findings for the ide project.
	 * Returns a non-empty string or {@link Optional#empty()} (never returns
	 * {@code Optional.of("")}).
	 */
	@Override
	public String getTeamscaleServer() {
		return Strings.emptyToNull(teamscaleServer);
	}

	/**
	 * Sets the Teamscale server used to load findings for the ide project.
	 */
	public void setTeamscaleServer(String teamscaleServer) {
		this.teamscaleServer = teamscaleServer;
	}

	/**
	 * Returns the Teamscale project selected for the ide project. Returns a
	 * non-empty string or Optional.empty() (never returns Optional.of("")).
	 */
	@Override
	public String getTeamscaleProject() {
		return Strings.emptyToNull(teamscaleProject);
	}

	public URI getTeamscaleUrl() {
		return url;
	}

	/**
	 * Sets the Teamscale project for the ide project.
	 */
	public void setTeamscaleProject(String teamscaleProject) {
		this.teamscaleProject = teamscaleProject;
	}

	/**
	 * Return whether there are any mappings for this module.
	 */
	public boolean hasMappings() {
		return !localToTeamscaleMap.isEmpty();
	}

	/**
	 * Returns all registered prefix mappings for the module.
	 */
	public ImmutableList<PrefixMapping> getMappings() {
		return ImmutableList.sortedCopyOf(comparing(PrefixMapping::getLocalPrefix), localToTeamscaleMap.values());
	}

	/**
	 * Replace the currently active mappings with the given set.
	 */
	public void setMappings(Collection<PrefixMapping> mappings) {
		localToTeamscaleMap.clear();
		teamscaleToLocalMap.clear();

		for (PrefixMapping mapping : mappings) {
			localToTeamscaleMap.put(mapping.getLocalPrefix(), mapping);
			teamscaleToLocalMap.put(mapping.getTeamscalePrefix(), mapping);
		}
	}

	/**
	 * Register a new prefix mapping for the module.
	 */
	public void addMapping(PrefixMapping mapping) {
		PrefixMapping existingMapping = localToTeamscaleMap.get(mapping.getLocalPrefix());

		if (existingMapping != null && existingMapping.equals(mapping)) {
			// Mapping already exists
			LOGGER.warn("Mapping already exists: {}", mapping);
			return;
		}

		localToTeamscaleMap.put(mapping.getLocalPrefix(), mapping);
		teamscaleToLocalMap.put(mapping.getTeamscalePrefix(), mapping);
	}

	/**
	 * Return the most specific mapping for the given local relative path or
	 * <code>null</code> if no such mapping exists.
	 */
	public PrefixMapping findTeamscalePrefixMapping(String relativePath) {
		return findMapping(localToTeamscaleMap, relativePath);
	}

	/**
	 * Return the most specific prefix mapping for the given uniform path from the
	 * given mapping or null, if no matching mapping could be found.
	 */
	private static PrefixMapping findMapping(Map<String, PrefixMapping> mapping, String relativePath) {
		if (mapping.isEmpty()) {
			return null;
		}

		if (mapping.containsKey(relativePath)) {
			return mapping.get(relativePath);
		}

		String path = relativePath;
		do {
			path = UniformPathUtils.removeLastSegments(path, 1);
			if (mapping.containsKey(path)) {
				return mapping.get(path);
			}
		} while (!path.isEmpty());

		return null;
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}

	/**
	 * @return {@code true} if all fields are empty, {@code false} otherwise.
	 */
	public boolean isEmpty() {
		return getTeamscaleProject() == null && getTeamscaleServer() == null && !hasMappings();
	}

	/**
	 * @return True if settings have been configured otherwise false.
	 */
	public static boolean isConfigured(IDEProjectSettings settings) {
		return settings != null && settings.getTeamscaleServer() != null
				&& settings.getTeamscaleProject() != null;
	}

	/**
	 * Restores the mappings form the given XML string.
	 */
	public static List<PrefixMapping> loadMappingsFromXml(String xml) throws IOException {
		if (xml == null) {
			return Collections.emptyList();
		}
		return PrefixMapping.deserialize(xml);
	}

	/**
	 * Maps a project relative path below the project root, if possible.
	 */
	public ResourceMappingResult map(URI serverUrl, ResourceReference projectRoot,
			RelativeUniformPath projectRelativePath) {
		Preconditions.checkArgument(teamscaleProject != null);

		PrefixMapping applicableMapping = findTeamscalePrefixMapping(projectRelativePath.toString());
		if (applicableMapping == null) {
			return ResourceMappingResult.failure(MappingProblem.noMappingFound(projectRoot));
		}

		ResourceReference mappingRoot = ResourceReferenceUtils.resolveFolder(projectRoot,
				UniformPathCompatibilityUtil.convertRelative(applicableMapping.getLocalPrefix()));
		UniformPath mappingRootPath = UniformPathCompatibilityUtil.convert(applicableMapping.getTeamscalePrefix());

		return ResourceMappingResult.success(serverUrl, teamscaleProject, null, mappingRoot, mappingRootPath);
	}
}
