/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.mappings;

import static java.util.Objects.requireNonNull;

import java.nio.file.Path;
import java.util.Arrays;
import java.util.Objects;
import java.util.Optional;

import org.checkerframework.checker.nullness.qual.Nullable;

import com.teamscale.ide.commons.lang.ToStringHelpers;
import com.teamscale.ide.commons.model.ResourceReference;

/**
 * Indicates a problem that occurred when attempting to map a local resource to
 * a remote resource on a Teamscale server.
 */
public final class MappingProblem {

	/**
	 * The possible failure conditions that may occur when trying to map a local
	 * resource to a remote resource on the Teamscale server.
	 */
	public enum EType {

		/** Resource does not reside below a mapping root (e.g., IDE project). */
		RESOURCE_OUTSIDE_MAPPING_ROOT,

		/** No Teamscale servers have been configured in the IDE. */
		NO_SERVERS_CONFIGURED,

		/**
		 * The mapping root has not been configured for Teamscale use (for example if
		 * the Teamscale integration was explicitly disabled for this mapping root).
		 */
		MAPPING_ROOT_NOT_CONFIGURED,

		/** No Teamscale server has been set for this mapping root. */
		NO_TEAMSCALE_SERVER,

		/** No Teamscale project has been set for this mapping root. */
		NO_TEAMSCALE_PROJECT,

		/** The Teamscale server set for the mapping root has not been configured. */
		TEAMSCALE_SERVER_UNKNOWN,

		/** No mapping could be computed. */
		NO_MAPPING_FOUND,

		/**
		 * The config file is invalid, i.e. there may be required properties missing.
		 */
		TOML_CONFIG_INVALID
	}

	/** The type of problem. */
	private final EType type;

	/** The mapping root w.r.t. which this problem occurred. */
	private final @Nullable ResourceReference mappingRoot;

	/**
	 * The context that may be necessary to understand and, possibly, fix the
	 * problem. Depends on the problem's {@link #type}.
	 */
	private final String[] context;

	private MappingProblem(EType type, @Nullable ResourceReference mappingRoot, String... context) {
		this.type = requireNonNull(type);
		this.mappingRoot = mappingRoot;
		this.context = requireNonNull(context);
	}

	/** @see #type */
	public EType getType() {
		return type;
	}

	/** @see #context */
	public String getContext(int index) {
		return context[index];
	}

	/** @see #mappingRoot */
	public Optional<ResourceReference> getMappingRoot() {
		return Optional.ofNullable(mappingRoot);
	}

	/** Resource does not reside below a mapping root (e.g., IDE project). */
	public static MappingProblem resourceOutsideMappingRoot() {
		return new MappingProblem(EType.RESOURCE_OUTSIDE_MAPPING_ROOT, null);
	}

	/** No Teamscale servers have been configured in the IDE. */
	public static MappingProblem noServersConfigured() {
		return new MappingProblem(EType.NO_SERVERS_CONFIGURED, null);
	}

	/** The mapping root has not been configured for Teamscale use. */
	public static MappingProblem mappingRootNotConfigured(ResourceReference mappingRoot) {
		return new MappingProblem(EType.MAPPING_ROOT_NOT_CONFIGURED, mappingRoot);
	}

	/** No Teamscale project has been set for this mapping root. */
	public static MappingProblem noTeamscaleProject(ResourceReference mappingRoot) {
		return new MappingProblem(EType.NO_TEAMSCALE_PROJECT, mappingRoot);
	}

	/** No Teamscale server has been set for this mapping root. */
	public static MappingProblem teamscaleServerUnknown(ResourceReference mappingRoot, String unknownServerName) {
		return new MappingProblem(EType.TEAMSCALE_SERVER_UNKNOWN, mappingRoot, unknownServerName);
	}

	/** No Teamscale server has been set for this mapping root. */
	public static MappingProblem noTeamscaleServer(ResourceReference mappingRoot) {
		return new MappingProblem(EType.NO_TEAMSCALE_SERVER, mappingRoot);
	}

	/** No mapping could be computed. */
	public static MappingProblem noMappingFound() {
		return new MappingProblem(EType.NO_MAPPING_FOUND, null);
	}

	/** No mapping could be computed for the given mapping root. */
	public static MappingProblem noMappingFound(ResourceReference mappingRoot) {
		return new MappingProblem(EType.NO_MAPPING_FOUND, mappingRoot);
	}

	/**
	 * The config file is invalid, i.e. there may be required properties missing.
	 **/
	public static MappingProblem tomlConfigInvalid(Path invalidTomlFile, String message) {
		ResourceReference invalidTomlReference = ResourceReference.forFile(invalidTomlFile.toUri());
		return new MappingProblem(EType.TOML_CONFIG_INVALID, invalidTomlReference, message);
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		MappingProblem that = (MappingProblem) other;
		return this.type == that.type && Objects.equals(this.mappingRoot, that.mappingRoot)
				&& Arrays.equals(this.context, that.context);
	}

	@Override
	public int hashCode() {
		return Objects.hash(type, mappingRoot, Arrays.hashCode(context));
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
