/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.mappings;

import java.net.URI;

import com.google.common.base.Preconditions;
import com.teamscale.ide.commons.client.Uris;
import com.teamscale.ide.commons.findings.LineRange;
import com.teamscale.ide.commons.model.IdeCommit;
import com.teamscale.ide.commons.model.RemoteResourceReference;

/**
 * Utility methods for computing URIs for mapped resources on a Teamscale
 * server.
 */
public class MappingUris {

	/**
	 * Creates an {@link URI} reference to the Code perspective for the given
	 * resource.
	 */
	public static URI createUri(RemoteResourceReference remoteResource) {
		return createUri(remoteResource, LineRange.unspecified());
	}

	/**
	 * Creates an {@link URI} reference to the Code perspective for the given
	 * resource.
	 */
	public static URI createUri(RemoteResourceReference remoteResource, LineRange lineRange) {
		if (lineRange.isSpecified()) {
			Preconditions.checkArgument(!remoteResource.isFolder());
		}
		URI serverUrl = remoteResource.getServerUrl();
		StringBuilder uri = new StringBuilder(serverUrl.toString());
		if (uri.charAt(uri.length() - 1) != '/') {
			uri.append('/');
		}
		uri.append("metrics.html").append('#');
		uri.append('/').append(Uris.encodePathSegment(remoteResource.getTeamscaleProject()));
		uri.append('/').append(Uris.encodePathSegments(remoteResource.getUniformPath().getPathSegments()));
		if (remoteResource.getBranch().isPresent()) {
			String commitQueryParameter = "t=" + Uris.encodeQueryParameter(
					Uris.asTimestampQueryParameter(IdeCommit.head(remoteResource.getBranch().get())));
			uri.append('?').append(commitQueryParameter);
		}
		if (lineRange.isSpecified()) {
			if (!remoteResource.getBranch().isPresent()) {
				uri.append('?');
			} else {
				uri.append("&");
			}
			uri.append("selection=").append(lineRange.getFirstLine());
			if (!lineRange.isSingleLine()) {
				uri.append("-").append(lineRange.getLastLine());
			}
		}
		return URI.create(uri.toString());
	}
}
