package com.teamscale.ide.commons.mappings;

import static java.util.Objects.requireNonNull;

import java.net.URI;
import java.util.Objects;
import java.util.Optional;

import org.checkerframework.checker.nullness.qual.Nullable;
import org.conqat.lib.commons.uniformpath.RelativeUniformPath;
import org.conqat.lib.commons.uniformpath.UniformPath;

import com.google.common.annotations.VisibleForTesting;
import com.teamscale.ide.commons.lang.ToStringHelpers;
import com.teamscale.ide.commons.model.RemoteResourceReference;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.model.ResourceReferenceUtils;
import com.teamscale.ide.commons.model.VcsResourceReference;

/**
 * Stores result of mapping a local file/folder resource to a remote Teamscale
 * resource.
 */
public abstract class ResourceMappingResult {

	@VisibleForTesting
	static final class SuccessfulResourceMapping extends ResourceMappingResult {

		/** The URL of the Teamscale server. */
		private final URI serverUrl;

		/** The ID of the project on the Teamscale server. */
		private final String projectId;

		/** The branch override (if any) to apply. */
		private final @Nullable String branchOverride;
		/**
		 * The mapping root that has been used to perform the mapping. Always a
		 * {@linkplain ResourceReference#isFolder() folder}.
		 */
		private final ResourceReference mappingRoot;

		/** The uniform part associated with the used mapping root. */
		private final UniformPath mappingRootPath;

		private SuccessfulResourceMapping(URI serverUrl, String projectId, @Nullable String branchOverride,
				ResourceReference mappingRoot, UniformPath mappingRootPath) {
			this.serverUrl = requireNonNull(serverUrl);
			this.projectId = requireNonNull(projectId);
			this.branchOverride = branchOverride;
			this.mappingRoot = requireNonNull(mappingRoot);
			this.mappingRootPath = requireNonNull(mappingRootPath);
		}

		@Override
		public boolean isSuccessful() {
			return true;
		}

		@Override
		public URI getServerUrl() {
			return serverUrl;
		}

		@Override
		public String getProjectId() {
			return projectId;
		}

		@Override
		public Optional<String> getBranchOverride() {
			return Optional.ofNullable(branchOverride);
		}

		@Override
		public ResourceReference getMappingRoot() {
			return mappingRoot;
		}

		@Override
		public RemoteResourceReference getMappedResource(VcsResourceReference vcsResource) {
			ResourceReference localResource = vcsResource.getResource();
			String branch = vcsResource.getBranch().orElse(null);

			RelativeUniformPath relativePath = ResourceReferenceUtils.computeRelativeUniformPath(mappingRoot,
					localResource);
			UniformPath path = mappingRootPath.resolve(relativePath);
			if (localResource.isFolder()) {
				return RemoteResourceReference.forFolder(serverUrl, projectId, branch, path);
			} else {
				return RemoteResourceReference.forFile(serverUrl, projectId, branch, path);
			}
		}

		@Override
		public Optional<ResourceReference> getLocalFile(UniformPath remotePath) {
			if (!mappingRootPath.hasDescendant(remotePath)) {
				return Optional.empty();
			}

			RelativeUniformPath pathBelowMappingRoot = remotePath.getSubPath(mappingRootPath.getPathSegments().size());
			return Optional.of(ResourceReferenceUtils.resolveFile(mappingRoot, pathBelowMappingRoot));
		}

		@Override
		public MappingProblem getMappingFailure() {
			throw new IllegalStateException("Resource mapping was successful.");
		}

		@Override
		public boolean equals(Object other) {
			if (this == other) {
				return true;
			}
			if (other == null) {
				return false;
			}
			if (getClass() != other.getClass()) {
				return false;
			}
			SuccessfulResourceMapping that = (SuccessfulResourceMapping) other;
			return Objects.equals(serverUrl, that.serverUrl) && Objects.equals(projectId, that.projectId)
					&& Objects.equals(branchOverride, that.branchOverride)
					&& Objects.equals(mappingRoot, that.mappingRoot)
					&& Objects.equals(mappingRootPath, that.mappingRootPath);
		}

		@Override
		public int hashCode() {
			return Objects.hash(serverUrl, projectId, branchOverride, mappingRoot, mappingRootPath);
		}
	}

	@VisibleForTesting
	static final class FailedResourceMapping extends ResourceMappingResult {

		/** Indicates why the mapping failed. */
		private final MappingProblem reason;

		private FailedResourceMapping(MappingProblem reason) {
			this.reason = requireNonNull(reason);
		}

		@Override
		public MappingProblem getMappingFailure() {
			return reason;
		}

		@Override
		public boolean isSuccessful() {
			return false;
		}

		@Override
		public URI getServerUrl() {
			throw new IllegalStateException("Resource mapping was not successful.");
		}

		@Override
		public String getProjectId() {
			throw new IllegalStateException("Resource mapping was not successful.");
		}

		@Override
		public Optional<String> getBranchOverride() {
			throw new IllegalStateException("Resource mapping was not successful.");
		}

		@Override
		public ResourceReference getMappingRoot() {
			throw new IllegalStateException("Resource mapping was not successful.");
		}

		@Override
		public RemoteResourceReference getMappedResource(VcsResourceReference localResource) {
			throw new IllegalStateException("Resource mapping was not successful.");
		}

		@Override
		public Optional<ResourceReference> getLocalFile(UniformPath remotePath) {
			throw new IllegalStateException("Resource mapping was not successful.");
		}

		@Override
		public boolean equals(Object other) {
			if (this == other) {
				return true;
			}
			if (other == null) {
				return false;
			}
			if (getClass() != other.getClass()) {
				return false;
			}
			FailedResourceMapping that = (FailedResourceMapping) other;
			return Objects.equals(this.reason, that.reason);
		}

		@Override
		public int hashCode() {
			return Objects.hash(reason);
		}
	}

	/**
	 * Creates a result signifying success, i.e., a mapping root was found.
	 */
	public static ResourceMappingResult success(URI serverUrl, String projectId, @Nullable String branchOverride,
			ResourceReference mappingRoot, UniformPath mappingRootPath) {
		return new SuccessfulResourceMapping(serverUrl, projectId, branchOverride, mappingRoot, mappingRootPath);
	}

	/** Creates a result signifying failure. */
	public static ResourceMappingResult failure(MappingProblem reason) {
		return new FailedResourceMapping(reason);
	}

	/**
	 * Returns {@code true} if a mapping root was found and could be mapped to a
	 * remote resource on a Teamscale server, {@code false} otherwise.
	 */
	public abstract boolean isSuccessful();

	/**
	 * Returns the URL of a Teamscale server configured as part of the mapping.
	 * <p>
	 * Must only be called if the mapping was {@linkplain #isSuccessful()
	 * successful}.
	 */
	public abstract URI getServerUrl();

	/**
	 * Returns the ID of project on the Teamscale server configured as part of the
	 * mapping.
	 * <p>
	 * Must only be called if the mapping was {@linkplain #isSuccessful()
	 * successful}.
	 */
	public abstract String getProjectId();

	/**
	 * Returns the branch override (if any) configured as part of the mapping.
	 * <p>
	 * Must only be called if the mapping was {@linkplain #isSuccessful()
	 * successful}.
	 */
	public abstract Optional<String> getBranchOverride();

	/**
	 * Returns the mapping root that governs this mapping result.
	 * <p>
	 * Must only be called if the mapping was {@linkplain #isSuccessful()
	 * successful}.
	 */
	public abstract ResourceReference getMappingRoot();

	/**
	 * Returns the resource on a server that corresponds to a local file or folder,
	 * taking the local resource's branch into account. Does not apply any branch
	 * override, i.e., the resulting {@link RemoteResourceReference} resides on the
	 * branch as given in the {@link VcsResourceReference}.
	 * <p>
	 * Just only be called if the mapping was {@linkplain #isSuccessful()
	 * successful}.
	 */
	public abstract RemoteResourceReference getMappedResource(VcsResourceReference localResource);

	/**
	 * Returns the local file that a uniform path on a remote server corresponds to.
	 * <p>
	 * Must only be called if the mapping was {@linkplain #isSuccessful()
	 * successful}.
	 */
	public abstract Optional<ResourceReference> getLocalFile(UniformPath remotePath);

	/**
	 * Returns the reason why no mapping root could be found that can be mapped to a
	 * remote resource on a Teamscale server.
	 * <p>
	 * Must only be called if the mapping was not {@linkplain #isSuccessful()
	 * successful}.
	 */
	public abstract MappingProblem getMappingFailure();

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
