package com.teamscale.ide.commons.mappings;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.conqat.lib.commons.uniformpath.RelativeUniformPath;
import org.conqat.lib.commons.uniformpath.UniformPathCompatibilityUtil;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.ImmutableList;
import com.teamscale.ide.commons.model.PrefixMapping;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.model.ResourceReferenceUtils;
import com.teamscale.ide.commons.servers.NamedServerDetails;
import com.teamscale.ide.commons.servers.services.ITeamscaleServerManager;
import com.teamscale.ide.config.TeamscaleConfig;
import com.teamscale.ide.config.path.CodePath;

/**
 * Provides methods used to migrate legacy settings to new
 * {@code .teamscale.toml} configuration files.
 */
public class SettingsMigrationUtils {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(SettingsMigrationUtils.class);

	/** Link to the {@code .teamscale.toml} file format documentation. */
	public static final URI TEAMSCALE_TOML_DOCS = URI.create("https://docs.teamscale.com/reference/teamscale-toml/");

	/**
	 * Returns IDE configuration objects corresponding to the given legacy settings.
	 */
	public static List<TeamscaleConfig> convertLegacySettings(IDEProjectSettings legacySettings,
			ResourceReference projectRoot, ITeamscaleServerManager serverManager) {
		if (!legacySettings.isTeamscaleEnabled()) {
			return earlyExit(Collections.emptyList(), "Incomplete settings won't be migrated (Teamscale disabled).");
		}

		String serverId = legacySettings.getTeamscaleServer();
		if (serverId == null) {
			return earlyExit(Collections.emptyList(), "Incomplete settings won't be migrated (no server name set).");
		}
		NamedServerDetails server = serverManager.getNamedServerDetails(serverId);
		if (server == null) {
			return earlyExit(Collections.emptyList(), "Incomplete settings won't be migrated (unknown server name).");
		}
		String serverUrl = server.getUrl();
		if (serverUrl == null) {
			return earlyExit(Collections.emptyList(), "Incomplete settings won't be migrated (no server URL set).");
		}

		String projectId = legacySettings.getTeamscaleProject();
		if (projectId == null) {
			return earlyExit(Collections.emptyList(), "Incomplete settings won't be migrated (no project name set).");
		}

		ImmutableList<PrefixMapping> mappings = legacySettings.getMappings();
		if (mappings.isEmpty()) {
			return earlyExit(Collections.emptyList(),
					"Incomplete settings won't be migrated (no prefix mappings set).");
		}

		List<TeamscaleConfig> configs = new ArrayList<>();
		for (PrefixMapping mapping : mappings) {
			convertPrefixMapping(projectRoot, serverUrl, projectId, mapping).ifPresent(configs::add);
		}
		return configs;
	}

	@VisibleForTesting
	static Optional<TeamscaleConfig> convertPrefixMapping(ResourceReference projectRoot, String serverUrlString,
			String projectId, PrefixMapping mapping) {
		LOGGER.traceEntry(null, mapping, projectRoot, projectId, serverUrlString);

		URI serverUrl = parseUrl(serverUrlString).orElse(null);
		if (serverUrl == null) {
			return LOGGER.traceExit(earlyExit(Optional.empty(),
					"Malformed settings won't be migrated (server URL malformed): {}", serverUrlString));
		}

		String localPrefixString = mapping.getLocalPrefix();
		if (localPrefixString == null) {
			return LOGGER.traceExit(
					earlyExit(Optional.empty(), "Malformed settings won't be migrated (Local prefix is null)"));
		}

		String teamscalePrefixString = mapping.getTeamscalePrefix();
		if (teamscalePrefixString == null) {
			return LOGGER.traceExit(
					earlyExit(Optional.empty(), "Malformed settings won't be migrated (Teamscale prefix is null)"));
		}

		CodePath teamscalePrefix = parseCodePath(teamscalePrefixString).orElse(null);
		if (teamscalePrefix == null) {
			return LOGGER.traceExit(earlyExit(Optional.empty(),
					"Malformed settings won't be migrated (Teamscale prefix malformed): {}", teamscalePrefixString));
		}

		RelativeUniformPath localPrefix = UniformPathCompatibilityUtil.convertRelative(mapping.getLocalPrefix());
		Path mappingRoot = ResourceReferenceUtils.resolveFolder(projectRoot, localPrefix).toPath();

		return LOGGER
				.traceExit(Optional.of(new TeamscaleConfig(serverUrl, projectId, null, teamscalePrefix, mappingRoot)));
	}

	private static Optional<URI> parseUrl(String urlString) {
		try {
			return Optional.of(new URI(urlString));
		} catch (URISyntaxException e) {
			LOGGER.catching(e);
			return Optional.empty();
		}
	}

	private static Optional<CodePath> parseCodePath(String codePathString) {
		try {
			return Optional.of(CodePath.parse(UniformPathCompatibilityUtil.convert(codePathString).toString()));
		} catch (ParseException e) {
			LOGGER.catching(e);
			return Optional.empty();
		}
	}

	/**
	 * Writes {@code .teamscale.toml} files corresponding to the given
	 * configurations.
	 */
	public static List<Path> writeTeamscaleTomls(List<TeamscaleConfig> configs) {
		List<Path> writtenPaths = new ArrayList<>();
		for (TeamscaleConfig config : configs) {
			writeTeamscaleToml(config).ifPresent(writtenPaths::add);
		}
		return writtenPaths;
	}

	private static Optional<Path> writeTeamscaleToml(TeamscaleConfig config) {
		Path teamscaleTomlPath = config.getLocalBasePath().resolve(".teamscale.toml");

		if (Files.exists(teamscaleTomlPath)) {
			// the new teamscale config already exists.
			// No need to overwrite the already existing one
			return Optional.empty();
		}

		LOGGER.debug("Creating .teamscale.toml: {}", teamscaleTomlPath);
		try {
			Files.write(teamscaleTomlPath, Arrays.asList( //
					"# Teamscale Configuration", //
					String.format("# See <%s>", TEAMSCALE_TOML_DOCS), //
					"version = 1.0", //
					"", //
					"root = true", //
					"", //
					"[server]", //
					String.format("url = \"%s\"", config.getServerUrl()), //
					"", //
					"[project]", //
					String.format("id = \"%s\"", config.getProjectId()), //
					String.format("path = \"%s\"", config.getProjectPath()) //
			), StandardCharsets.UTF_8);
			return Optional.of(teamscaleTomlPath);
		} catch (IOException e) {
			LOGGER.catching(e);
			LOGGER.error("Could not write .teamscale.toml: {}", teamscaleTomlPath);
			return Optional.empty();
		}
	}

	private static <R> R earlyExit(R returnValue, String message, Object... parameters) {
		LOGGER.debug(message, parameters);
		return returnValue;
	}
}
