package com.teamscale.ide.commons.mappings.services.impl;

import static java.util.Objects.requireNonNull;

import java.io.IOException;
import java.net.URI;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.Collections;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.checkerframework.checker.nullness.qual.Nullable;
import org.conqat.lib.commons.uniformpath.UniformPathCompatibilityUtil;

import com.google.common.collect.Sets;
import com.teamscale.ide.commons.mappings.MappingProblem;
import com.teamscale.ide.commons.mappings.ResourceMappingResult;
import com.teamscale.ide.commons.mappings.services.IResourceMappingService;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.servers.services.IServerCredentialsManager;
import com.teamscale.ide.config.InvalidConfigException;
import com.teamscale.ide.config.TeamscaleConfig;
import com.teamscale.ide.config.TeamscaleConfigReader;

/**
 * A resource mapping implementation to support {@code .teamscale.toml}-based
 * configuration.
 */
public class TeamscaleTomlResourceMappingService implements IResourceMappingService {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(TeamscaleTomlResourceMappingService.class);

	private final IServerCredentialsManager credentialsManager;
	private final @Nullable String branchOverride;

	/** The teamscale (toml) configuration reader instance. */
	private final TeamscaleConfigReader configReader = new TeamscaleConfigReader();

	public TeamscaleTomlResourceMappingService(IServerCredentialsManager credentialsManager) {
		this(credentialsManager, null);
	}

	public TeamscaleTomlResourceMappingService(IServerCredentialsManager credentialsManager, String branchOverride) {
		this.credentialsManager = requireNonNull(credentialsManager);
		this.branchOverride = branchOverride;
	}

	@Override
	public ResourceMappingResult map(ResourceReference resource) {
		LOGGER.traceEntry(null, resource);

		if (!resource.hasPath()) {
			return LOGGER.traceExit(ResourceMappingResult.failure(MappingProblem.noMappingFound()));
		}

		TeamscaleConfig config = null;
		try {
			Path parent = folderOf(resource);
			if (parent != null) {
				config = configReader.read(parent).orElse(null);
			}
			if (config == null) {
				return LOGGER.traceExit(ResourceMappingResult.failure(MappingProblem.resourceOutsideMappingRoot()));
			}
		} catch (InvalidConfigException e) {
			LOGGER.catching(e);
			return LOGGER.traceExit(ResourceMappingResult
					.failure(MappingProblem.tomlConfigInvalid(e.getInvalidTomlFile(), e.getMessage())));
		} catch (IOException e) {
			LOGGER.catching(e);
			return LOGGER.traceExit(ResourceMappingResult.failure(MappingProblem.noMappingFound()));
		}

		LOGGER.debug("Mapping using Teamscale config {}.", config);

		URI mappingRootUri = config.getLocalBasePath().toUri();
		ResourceReference projectRoot = ResourceReference.forFolder(mappingRootUri);

		URI serverUrl = config.getServerUrl();
		if (!credentialsManager.getServer(serverUrl).isPresent()) {
			return LOGGER.traceExit(ResourceMappingResult
					.failure(MappingProblem.teamscaleServerUnknown(projectRoot, serverUrl.toString())));
		}

		String branch = determineBranch(config).orElse(null);
		ResourceMappingResult result = ResourceMappingResult.success(serverUrl, config.getProjectId(), branch,
				projectRoot, UniformPathCompatibilityUtil.convert(config.getProjectPath().toString()));
		return LOGGER.traceExit(result);
	}

	private Optional<String> determineBranch(TeamscaleConfig config) {
		if (branchOverride != null) {
			return Optional.of(branchOverride);
		}
		return config.getProjectBranchOverride();
	}

	@Override
	public Set<ResourceReference> findPotentialMappingRoots(ResourceReference resource) {
		if (!resource.hasPath()) {
			// some resources, like ADT-specific meta-files, have no real path in Teamscale
			return Collections.emptySet();
		}

		if (resource.isFolder()) {
			return findPotentialMappingRootsForFolder(resource);
		} else {
			return findPotentialMappingRootUpwards(resource);
		}
	}

	private static boolean isPotentialMappingRoot(TeamscaleConfig config, Path path) {
		return config.getLocalBasePath().equals(path);
	}

	/**
	 * Returns the folder's {@linkplain Path} of the resource. The returned folder
	 * is the first valid parent folder or null if there is none.
	 **/
	private static Path folderOf(ResourceReference resource) {
		LOGGER.traceEntry(null, resource);
		Path currentPath = resource.toPath();
		while (currentPath != null && !Files.isDirectory(currentPath)) {
			currentPath = currentPath.getParent();
		}
		return LOGGER.traceExit(currentPath);
	}

	private Set<ResourceReference> findPotentialMappingRootsForFolder(ResourceReference folder) {
		return Sets.union(findPotentialMappingRootsDownwards(folder), findPotentialMappingRootUpwards(folder));
	}

	private Set<ResourceReference> findPotentialMappingRootUpwards(ResourceReference resource) {
		for (Path currentPath = folderOf(resource); currentPath != null; currentPath = currentPath.getParent()) {
			try {
				TeamscaleConfig config = configReader.read(currentPath).orElse(null);
				if (config == null) {
					return Collections.emptySet();
				}
				if (isPotentialMappingRoot(config, currentPath)) {
					return Collections.singleton(ResourceReference.forFolder(currentPath.toUri()));
				}
			} catch (InvalidConfigException | IOException e) {
				return Collections.emptySet();
			}
		}
		return Collections.emptySet();
	}

	private Set<ResourceReference> findPotentialMappingRootsDownwards(ResourceReference folder) {
		try {
			TeamscaleConfigVisitor teamscaleConfigVisitor = new TeamscaleConfigVisitor();
			Files.walkFileTree(folder.toPath(), teamscaleConfigVisitor);
			return teamscaleConfigVisitor.getPotentialMappingRoots();
		} catch (IOException e) {
			return Collections.emptySet();
		}
	}

	/**
	 * File visitor that scans a directory tree and collects potential mapping
	 * roots, i.e. directories containing a
	 * {@linkplain TeamscaleConfigReader#TEAMSCALE_TOML_FILE_NAME} file with a
	 * filled ``
	 */
	private final class TeamscaleConfigVisitor extends SimpleFileVisitor<Path> {
		private final Set<ResourceReference> potentialMappingRoots = new HashSet<>();

		@Override
		public FileVisitResult preVisitDirectory(Path directory, BasicFileAttributes attrs) {
			if (isPotentialMappingRoot(directory)) {
				potentialMappingRoots.add(ResourceReference.forFolder(directory.toUri()));
			}
			return FileVisitResult.CONTINUE;
		}

		/**
		 * Returns the found mapping roots during traversal.
		 */
		public Set<ResourceReference> getPotentialMappingRoots() {
			return new HashSet<>(potentialMappingRoots);
		}

		private boolean isPotentialMappingRoot(Path directory) {
			try {
				TeamscaleConfig config = configReader.read(directory).orElse(null);
				if (config == null) {
					return false;
				}
				return TeamscaleTomlResourceMappingService.isPotentialMappingRoot(config, directory);
			} catch (InvalidConfigException | IOException e) {
				return false;
			}
		}
	}
}
