/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.mappings.tasks;

import static java.util.Objects.requireNonNull;

import java.net.URI;
import java.util.Collection;
import java.util.Collections;

import com.teamscale.ide.commons.findings.LineRange;
import com.teamscale.ide.commons.mappings.MappingUris;
import com.teamscale.ide.commons.mappings.MappingUtils;
import com.teamscale.ide.commons.mappings.ResourceMappingResult;
import com.teamscale.ide.commons.mappings.services.IResourceMappingService;
import com.teamscale.ide.commons.model.RemoteResourceReference;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.model.VcsResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.notification.Notifications;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.tasks.OpenInBrowserTaskBase;
import com.teamscale.ide.commons.precommit.PreCommitUtils;
import com.teamscale.ide.commons.servers.services.IServerCredentialsManager;
import com.teamscale.ide.commons.vcs.services.ICheckoutFactory;

/**
 * Task which opens the given resource in the system's default browser.
 */
public class OpenResourceInBrowserTask extends OpenInBrowserTaskBase {

	/** Checkout factory */
	protected final ICheckoutFactory checkoutFactory;

	/** Resource mapping service */
	protected final IResourceMappingService mappingService;

	/** The resource which should be shown in the browser. */
	protected final ResourceReference resource;

	private final IServerCredentialsManager credentialsManager;

	/**
	 * The {@link LineRange} of the selection or {@link LineRange#unspecified()}, if
	 * no line is selected
	 */
	protected final LineRange lineRange;

	/**
	 * Whether we should open the pre-commit branch computed from the current user
	 * and branch name. If false, we just use the "normal" branch name of the
	 * {@link RemoteResourceReference}.
	 */
	private final boolean openOnPreCommitBranch;

	public OpenResourceInBrowserTask(IServiceRegistry serviceRegistry, ResourceReference resource, LineRange lineRange,
			boolean openOnPreCommitBranch) {
		super(serviceRegistry, "Opening resource in browser");

		this.resource = requireNonNull(resource);
		this.lineRange = requireNonNull(lineRange);
		this.openOnPreCommitBranch = openOnPreCommitBranch;

		checkoutFactory = serviceRegistry.getService(ICheckoutFactory.class);
		mappingService = serviceRegistry.getService(IResourceMappingService.class);
		if (openOnPreCommitBranch) {
			credentialsManager = serviceRegistry.getService(IServerCredentialsManager.class);
		} else {
			credentialsManager = null;
		}
	}

	public OpenResourceInBrowserTask(IServiceRegistry serviceRegistry, ResourceReference resource) {
		this(serviceRegistry, resource, LineRange.unspecified(), false);
	}

	@Override
	protected Collection<URI> createUri() {
		ResourceMappingResult mappingResult = mappingService.map(resource);
		if (!mappingResult.isSuccessful()) {
			notificationService.reportNotification(ENotificationCause.DIRECT,
					Notifications.mappingProblem(resource, mappingResult.getMappingFailure()));
			return Collections.emptyList();
		}

		String branchName = MappingUtils.getBranch(checkoutFactory, resource, mappingResult).orElse(null);
		if (openOnPreCommitBranch) {
			URI serverUrl = mappingResult.getServerUrl();
			if (!credentialsManager.getServer(serverUrl).isPresent()) {
				notificationService.reportNotification(ENotificationCause.DIRECT,
						Notifications.serverUrlNotFound(serverUrl));
				return Collections.emptyList();
			}
			String username = credentialsManager.getServer(serverUrl).get().getUsername();
			branchName = PreCommitUtils.getPreCommit3BranchName(username, branchName);
		}
		RemoteResourceReference remoteResource = mappingResult
				.getMappedResource(new VcsResourceReference(resource, branchName));

		return Collections.singleton(MappingUris.createUri(remoteResource, lineRange));
	}
}
