package com.teamscale.ide.commons.mappings.tasks;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;

import com.google.common.collect.ImmutableCollection;
import com.teamscale.ide.commons.mappings.MappingUris;
import com.teamscale.ide.commons.mappings.MappingUtils;
import com.teamscale.ide.commons.mappings.ResourceMappingResult;
import com.teamscale.ide.commons.mappings.services.IResourceMappingService;
import com.teamscale.ide.commons.model.RemoteResourceReference;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.model.VcsResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.notification.Notifications;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.tasks.OpenInBrowserTaskBase;
import com.teamscale.ide.commons.precommit.PreCommitUtils;
import com.teamscale.ide.commons.servers.services.IServerCredentialsManager;
import com.teamscale.ide.commons.vcs.services.ICheckoutFactory;

/**
 * Task which opens the given resources in the system's default browser.
 */
public class OpenResourcesInBrowserTask extends OpenInBrowserTaskBase {

	/** The resources which should be opened in the browser. */
	private final ImmutableCollection<ResourceReference> resources;

	/**
	 * Whether we should open the pre-commit branch computed from the current user
	 * and branch name. If false, we just use the "normal" branch name of the
	 * {@link RemoteResourceReference}.
	 */
	private final boolean openOnPreCommitBranch;

	private final ICheckoutFactory checkoutFactory;
	private final IResourceMappingService mappingService;
	private final IServerCredentialsManager credentialsManager;

	public OpenResourcesInBrowserTask(IServiceRegistry serviceRegistry,
			ImmutableCollection<ResourceReference> resources, boolean openOnPreCommitBranch) {
		super(serviceRegistry, "Opening resources in browser");

		this.resources = resources;
		this.openOnPreCommitBranch = openOnPreCommitBranch;

		checkoutFactory = serviceRegistry.getService(ICheckoutFactory.class);
		mappingService = serviceRegistry.getService(IResourceMappingService.class);
		if (openOnPreCommitBranch) {
			credentialsManager = serviceRegistry.getService(IServerCredentialsManager.class);
		} else {
			credentialsManager = null;
		}
	}

	@Override
	protected Collection<URI> createUri() {
		Collection<URI> uris = new ArrayList<>();
		for (ResourceReference resource : resources) {
			ResourceMappingResult mappingResult = mappingService.map(resource);
			if (!mappingResult.isSuccessful()) {
				notificationService.reportNotification(ENotificationCause.DIRECT,
						Notifications.mappingProblem(resource, mappingResult.getMappingFailure()));
				continue;
			}

			String branchName = MappingUtils.getBranch(checkoutFactory, resource, mappingResult).orElse(null);
			if (openOnPreCommitBranch) {
				URI serverUrl = mappingResult.getServerUrl();
				if (!credentialsManager.getServer(serverUrl).isPresent()) {
					notificationService.reportNotification(ENotificationCause.DIRECT,
							Notifications.serverUrlNotFound(serverUrl));
					continue;
				}
				String username = credentialsManager.getServer(serverUrl).get().getUsername();
				branchName = PreCommitUtils.getPreCommit3BranchName(username, branchName);
			}
			RemoteResourceReference remoteResource = mappingResult
					.getMappedResource(new VcsResourceReference(resource, branchName));
			uris.add(MappingUris.createUri(remoteResource));
		}
		return uris;
	}
}