/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.model;

import java.util.Objects;
import java.util.Optional;

import org.checkerframework.checker.nullness.qual.Nullable;

import com.teamscale.ide.commons.lang.ToStringHelpers;

/**
 * A {@link ContentHashedResourceReference} points to a file resource on some
 * local file system with a specific file content digest hash.
 */
public class ContentHashedResourceReference {

	/** The referenced resource. */
	private final ResourceReference resource;

	/** The branch, if any, the referenced resource is on. */
	private final @Nullable String branch;

	/** The file content hash, if any, the referenced resource is on. */
	private final @Nullable String contentHash;

	public ContentHashedResourceReference(ResourceReference resource, @Nullable String branch,
			@Nullable String contentHash) {
		this.resource = Objects.requireNonNull(resource);
		this.branch = branch;
		this.contentHash = contentHash;
	}

	/** @see #resource */
	public ResourceReference getResource() {
		return resource;
	}

	/** @see #branch */
	public Optional<String> getBranch() {
		return Optional.ofNullable(branch);
	}

	/** @see #contentHash */
	public Optional<String> getContentHash() {
		return Optional.ofNullable(contentHash);
	}

	@Override
	public int hashCode() {
		return Objects.hash(resource, branch, contentHash);
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		ContentHashedResourceReference that = (ContentHashedResourceReference) other;
		return Objects.equals(resource, that.resource) && Objects.equals(branch, that.branch)
				&& Objects.equals(contentHash, that.contentHash);
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
