package com.teamscale.ide.commons.model;

import java.util.Comparator;
import java.util.Objects;
import java.util.Optional;

import org.checkerframework.checker.nullness.qual.Nullable;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.teamscale.ide.commons.lang.ToStringHelpers;

/**
 * Describes a single commit by its {@link #branchName} and {@link #timestamp}
 */
public final class IdeCommit implements Comparable<IdeCommit> {

	/** The {@link Comparator} used to compare {@link IdeCommit}s */
	private static final Comparator<IdeCommit> COMPARATOR = Comparator.nullsFirst(Comparator
			.comparingLong(IdeCommit::getTimestamp).thenComparing(commit -> commit.getBranchName().orElse(null)));

	/** The name of the JSON property name for {@link #branchName}. */
	private static final String BRANCH_NAME_PROPERTY = "branchName";

	/** The name of the JSON property name for {@link #timestamp}. */
	private static final String TIMESTAMP_PROPERTY = "timestamp";

	/** The name of the branch. */
	@JsonProperty(BRANCH_NAME_PROPERTY)
	private final @Nullable String branchName;

	/** The timestamp on the branch. */
	@JsonProperty(TIMESTAMP_PROPERTY)
	private final long timestamp;

	@JsonCreator
	private IdeCommit(@JsonProperty(BRANCH_NAME_PROPERTY) @Nullable String branchName,
			@JsonProperty(TIMESTAMP_PROPERTY) long timestamp) {
		this.branchName = branchName;
		this.timestamp = timestamp;
	}

	/**
	 * Creates a {@link IdeCommit} which represents the latest commit on the given
	 * branch
	 */
	public static IdeCommit head(@Nullable String branchName) {
		return new IdeCommit(branchName, Long.MAX_VALUE);
	}

	/**
	 * Creates a {@link IdeCommit} which represents the commit on the given branch
	 * at the given timestamp
	 */
	public static IdeCommit atTimestamp(@Nullable String branchName, long timestamp) {
		return new IdeCommit(branchName, timestamp);
	}

	/**
	 * @see #branchName
	 */
	public Optional<String> getBranchName() {
		return Optional.ofNullable(branchName);
	}

	/**
	 * @see #timestamp
	 */
	public long getTimestamp() {
		return timestamp;
	}

	@Override
	public int compareTo(IdeCommit that) {
		return COMPARATOR.compare(this, that);
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null || getClass() != other.getClass()) {
			return false;
		}

		IdeCommit that = (IdeCommit) other;
		return Objects.equals(branchName, that.branchName) //
				&& timestamp == that.timestamp;
	}

	@Override
	public int hashCode() {
		return Objects.hash(branchName, timestamp);
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}