/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.model;

import java.io.IOException;
import java.util.List;
import java.util.Objects;

import jakarta.xml.bind.annotation.XmlElement;

import org.conqat.engine.commons.util.JsonUtils;
import org.conqat.engine.resource.util.UniformPathUtils;
import org.conqat.lib.commons.uniformpath.UniformPath;
import org.conqat.lib.commons.uniformpath.UniformPathCompatibilityUtil;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.google.common.base.CharMatcher;
import com.google.common.base.Strings;
import com.teamscale.commons.service.XmlSerializationUtils;
import com.teamscale.ide.commons.lang.ToStringHelpers;

/**
 * Immutable mapping class from a local prefix (part of the resource path) to a
 * Teamscale prefix.
 * <p>
 * This class is serialized as part of
 * {@code teamscale_plugin_mappings.properties}. The fully-qualified class name
 * and any field names <strong>must</strong> not be changed.
 */
public final class PrefixMapping {

	/** A mapping <code>"" &lt;-&gt; ""</code>. */
	public static final PrefixMapping EMPTY_PREFIX_MAPPING = new PrefixMapping("", "", "");

	/** The project id of the project that the mapping is valid for. */
	private String projectId;

	/** The local prefix in the IDE. */
	private String localPrefix;

	/** The prefix as it is used by the Teamscale server. */
	private String teamscalePrefix;

	public PrefixMapping(String localPrefix, String teamscalePrefix) {
		this.localPrefix = localPrefix.trim();
		this.teamscalePrefix = teamscalePrefix.trim();
		projectId = "";
	}

	@JsonCreator
	public PrefixMapping(@JsonProperty("localPrefix") String localPrefix,
			@JsonProperty("teamscalePrefix") String teamscalePrefix, @JsonProperty("projectId") String projectId) {
		this.localPrefix = localPrefix;
		this.teamscalePrefix = teamscalePrefix;
		this.projectId = projectId;
	}

	/** Constructor. */
	public PrefixMapping(ProjectMapping mapping) {
		localPrefix = CharMatcher.is('/').trimTrailingFrom(mapping.from);
		teamscalePrefix = CharMatcher.is('/').trimTrailingFrom(mapping.to);
		projectId = mapping.projectId;
	}

	/**
	 * Receives the content of the "a" tag wrapper element that was added in an
	 * older xstream version.
	 */
	@XmlElement(name = "a")
	public void setLegacyMapping(PrefixMapping legacyMapping) {
		localPrefix = legacyMapping.localPrefix;
		teamscalePrefix = legacyMapping.teamscalePrefix;
		projectId = legacyMapping.projectId;
	}

	/**
	 * @see #projectId
	 * 
	 * @deprecated A prefix mapping as used in IDE Commons consists only of local
	 *             and Teamscale prefix. The project is implicit.
	 */
	@Deprecated
	public String getProjectId() {
		return projectId;
	}

	/** Returns the local prefix. */
	public String getLocalPrefix() {
		return localPrefix;
	}

	/** Returns the Teamscale prefix */
	public String getTeamscalePrefix() {
		return teamscalePrefix;
	}

	/**
	 * Transforms the path of a workspace resource into a uniform path for Teamscale
	 */
	public UniformPath makeUniformPath(String relativePath) {
		String commonPath = relativePath;
		if (commonPath.startsWith("/")) {
			commonPath = commonPath.substring("/".length());
		}
		if (commonPath.startsWith(localPrefix)) {
			commonPath = commonPath.substring(localPrefix.length());
		}

		if (Strings.isNullOrEmpty(teamscalePrefix)) {
			return UniformPathCompatibilityUtil.convert(UniformPathUtils.normalizeAllSeparators(commonPath));
		}

		return UniformPathCompatibilityUtil.convert(UniformPathUtils.concatenate(teamscalePrefix, commonPath));
	}

	/** Serialize a list of {@code PrefixMapping}s into an XML string */
	public static String serialize(List<PrefixMapping> mappings) {
		return XmlSerializationUtils.serializeToXML(mappings);
	}

	/** Deserialize a list of {@code PrefixMapping}s from an XML string */
	public static List<PrefixMapping> deserialize(String inputXml) throws IOException {
		LegacyPrefixMappingWrapper legacyMapping = XmlSerializationUtils.deserializeFromXML(inputXml,
				LegacyPrefixMappingWrapper.class);
		if (legacyMapping.mappings != null) {
			return legacyMapping.mappings;
		}
		return XmlSerializationUtils.deserializeFromXML(inputXml, JsonUtils.getJavaListType(PrefixMapping.class));
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		PrefixMapping that = (PrefixMapping) other;
		return Objects.equals(localPrefix, that.localPrefix) && Objects.equals(projectId, that.projectId)
				&& Objects.equals(teamscalePrefix, that.teamscalePrefix);
	}

	@Override
	public int hashCode() {
		return Objects.hash(localPrefix, projectId, teamscalePrefix);
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}

	/**
	 * Helper class for deserializing old versions of the prefix mappings that had
	 * an "a" tag wrapper element around the mappings.
	 */
	private static class LegacyPrefixMappingWrapper {

		@XmlElement(name = "a")
		private List<PrefixMapping> mappings;

	}
}
