/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.model;

import static java.util.Objects.requireNonNull;

import java.net.URI;
import java.util.Objects;
import java.util.Optional;

import org.checkerframework.checker.nullness.qual.Nullable;
import org.conqat.lib.commons.uniformpath.UniformPath;
import org.conqat.lib.commons.uniformpath.UniformPathCompatibilityUtil;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.teamscale.ide.commons.lang.ToStringHelpers;

/**
 * A reference to a remote resource, i.e., a file or folder on a Teamscale
 * server.
 * 
 * It is mirrored by a {@link VcsResourceReference}, which points to file or
 * folder in a local checkout from a version-control system.
 */
public final class RemoteResourceReference {

	/** The name of the JSON property name for {@link #serverUrl}. */
	protected static final String SERVER_URI_PROPERTY = "serverUrl";

	/** The name of the JSON property name for {@link #teamscaleProject}. */
	protected static final String TEAMSCALE_PROJECT_PROPERTY = "teamscaleProject";

	/** The name of the JSON property name for {@link #branch}. */
	protected static final String BRANCH_PROPERTY = "branch";

	/** The name of the JSON property name for {@link #uniformPath}. */
	protected static final String UNIFORM_PATH_PROPERTY = "uniformPath";

	/** The name of the JSON property name for {@link #isFolder}. */
	protected static final String IS_FOLDER_PROPERTY = "isFolder";

	/** The URL of the server this remote resource is located on. */
	@JsonProperty(SERVER_URI_PROPERTY)
	private final URI serverUrl;

	/**
	 * The name or alias of the Teamscale project this remote resource is located
	 * on.
	 */
	@JsonProperty(TEAMSCALE_PROJECT_PROPERTY)
	private final String teamscaleProject;

	/** The name of the branch (if any) this remote resource is located on. */
	@JsonProperty(BRANCH_PROPERTY)
	private final @Nullable String branch;

	/** The uniform path of this remote resource. */
	@JsonProperty(UNIFORM_PATH_PROPERTY)
	private final UniformPath uniformPath;

	/** Whether this remote resource is a folder (if false, then this is a file). */
	@JsonProperty(IS_FOLDER_PROPERTY)
	private final boolean isFolder;

	/**
	 * Creates a {@link RemoteResourceReference} for a regular file, i.e., not a
	 * folder, on a Teamscale server.
	 */
	public static RemoteResourceReference forFile(URI serverUrl, String teamscaleProject, @Nullable String branch,
			UniformPath uniformPath) {
		return new RemoteResourceReference(serverUrl, teamscaleProject, branch, uniformPath, false);
	}

	/**
	 * Creates a {@link RemoteResourceReference} for a regular file, i.e., not a
	 * folder, on a Teamscale server.
	 */
	public static RemoteResourceReference forFile(URI serverUrl, String teamscaleProject, @Nullable String branch,
			String uniformPath) {
		return new RemoteResourceReference(serverUrl, teamscaleProject, branch,
				UniformPathCompatibilityUtil.convert(uniformPath), false);
	}

	/**
	 * Create a {@link RemoteResourceReference} for a folder on a Teamscale server
	 */
	public static RemoteResourceReference forFolder(URI serverUrl, String teamscaleProject, @Nullable String branch,
			UniformPath uniformPath) {
		return new RemoteResourceReference(serverUrl, teamscaleProject, branch, uniformPath, true);
	}

	@JsonCreator
	private RemoteResourceReference(@JsonProperty(SERVER_URI_PROPERTY) URI serverUrl,
			@JsonProperty(TEAMSCALE_PROJECT_PROPERTY) String teamscaleProject,
			@JsonProperty(BRANCH_PROPERTY) @Nullable String branch,
			@JsonProperty(UNIFORM_PATH_PROPERTY) UniformPath uniformPath,
			@JsonProperty(IS_FOLDER_PROPERTY) boolean isFolder) {
		this.serverUrl = requireNonNull(serverUrl);
		this.teamscaleProject = requireNonNull(teamscaleProject);
		this.branch = branch;
		this.uniformPath = requireNonNull(uniformPath);
		this.isFolder = isFolder;

		if (!uniformPath.isCodePath()) {
			throw new IllegalArgumentException("Not a code resource");
		}
	}

	/**
	 * Returns a copy of this {@link RemoteResourceReference} but with the given
	 * branch name on the remote server.
	 */
	public RemoteResourceReference withNewBranch(String newBranchName) {
		if (isFolder()) {
			return forFolder(getServerUrl(), getTeamscaleProject(), newBranchName, getUniformPath());
		} else {
			return forFile(getServerUrl(), getTeamscaleProject(), newBranchName, getUniformPath());
		}
	}

	/** @see #serverUrl */
	public URI getServerUrl() {
		return serverUrl;
	}

	/** @see #teamscaleProject */
	public String getTeamscaleProject() {
		return teamscaleProject;
	}

	/** @see #branch */
	public Optional<String> getBranch() {
		return Optional.ofNullable(branch);
	}

	/** @see #uniformPath */
	public UniformPath getUniformPath() {
		return uniformPath;
	}

	/** @see #isFolder */
	public boolean isFolder() {
		return isFolder;
	}

	@Override
	public int hashCode() {
		return Objects.hash(serverUrl, teamscaleProject, branch, uniformPath, isFolder);
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		RemoteResourceReference that = (RemoteResourceReference) other;
		return Objects.equals(this.serverUrl, that.serverUrl)
				&& Objects.equals(this.teamscaleProject, that.teamscaleProject)
				&& Objects.equals(this.branch, that.branch) && Objects.equals(this.uniformPath, that.uniformPath)
				&& this.isFolder == that.isFolder;
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}

}
