/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.model;

import static java.util.Objects.requireNonNull;

import java.io.File;
import java.net.URI;
import java.nio.file.FileSystemNotFoundException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Objects;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.google.common.base.Preconditions;
import com.teamscale.ide.commons.lang.ToStringHelpers;

/**
 * A reference to a file or folder resource. It knows whether it refers to a
 * file or folder, respectively (albeit this may of course change on disk at any
 * time). The referenced file or folder may not exist and, if it does, it may
 * not exist on a physical disk but only in some kind of virtual filesystem.
 * <p>
 * As a resource reference is not backed by any IDE-specific implementation
 * (e.g., {@code FileObject}, {@code IFile}, {@code VirtualFile}), it can easily
 * be created in code common to all IDEs. Moreover, it does not have to satisfy
 * any of the restrictions that some IDEs place on their IDE-specific
 * implementations; for example, a NetBeans {@code FileObject} can only refer to
 * a file that exists on disk, which would preclude us from using it to refer to
 * file deletions as part of pre-commit analysis. A resource reference is also
 * guaranteed to have a very small memory footprint, unlike some IDE-specific
 * implementations; in contrast, an IntelliJ {@code VirtualFile} may keep an
 * entire virtual filesystem life.
 * <p>
 * Note that we use an {@link URI} rather than a {@link java.io.File} or
 * {@link java.nio.file.Path} to refer to the file or folder in question. This
 * allows for schemes other than {@code file}, which some IDEs use for special
 * purposes, e.g., to denote ABAP objects which do not reside on disk (but
 * behave file-like enough). This also implies that resource references may not
 * be portable across IDEs. While all IDEs share the same {@link URI}-backed
 * implementation of {@link ResourceReference}, not all URI schemes are
 * understood by all IDEs. This is not a problem, however, as
 * {@link ResourceReference} are not meant to be interchanged.
 */
public final class ResourceReference {

	/** The name of the JSON property name for {@link #uri}. */
	protected static final String URI_PROPERTY = "uri";

	/**
	 * A {@link URI} pointing to the referenced resource. If the resource is a
	 * folder, the URI's {@linkplain URI#getPath() paths} ends in a slash. If it is
	 * a file, it does not.
	 */
	@JsonProperty(URI_PROPERTY)
	private final URI uri;

	/**
	 * Creates a {@link ResourceReference} for a regular file, i.e., not a folder.
	 *
	 * @param file
	 *            a {@link URI} referring to the file in question. Must not end with
	 *            a slash.
	 */
	public static ResourceReference forFile(URI file) {
		Preconditions.checkArgument(!file.getPath().endsWith("/"), "File URIs must not end in a slash: %s", file);

		return new ResourceReference(file);
	}

	/**
	 * An overload of {@link #forFile(URI)} accepting a {@link Path} is deliberately
	 * not provided for the following reasons:
	 * <ul>
	 * <li>{@link Path#toUri()} performs I/O and may even throw an
	 * {@link java.io.IOError}. This may be unexpected to the caller.</li>
	 * <li>If the given {@link Path} refers to a file in a revision of a
	 * version-control system, that {@link Path} may now refer to a <em>folder</em>
	 * on disk. This discrepancy would result in the "no trailing slash"
	 * precondition of {@link #forFile(URI)} being violated.
	 * </ul>
	 */
	@SuppressWarnings("unused")
	private static ResourceReference forFile(Path file) {
		throw new UnsupportedOperationException("Deliberately not provided");
	}

	/**
	 * An overload of {@link #forFile(URI)} accepting a {@link File} is deliberately
	 * not provided for the following reasons:
	 * <ul>
	 * <li>{@link File#toURI()} performs I/O.</li>
	 * <li>If the given {@link File} refers to a file in a revision of a
	 * version-control system, that {@link File} may now refer to a <em>folder</em>
	 * on disk. This discrepancy would result in the "no trailing slash"
	 * precondition of {@link #forFile(URI)} being violated.
	 * </ul>
	 */
	@SuppressWarnings("unused")
	private static ResourceReference forFile(File file) {
		throw new UnsupportedOperationException("Deliberately not provided");
	}

	/**
	 * Create a {@link ResourceReference} for a folder.
	 *
	 * @param folder
	 *            a {@link URI} referring to the folder in question. Must end with a
	 *            slash to mark the reference as a folder.
	 */
	public static ResourceReference forFolder(URI folder) {
		Preconditions.checkArgument(folder.getPath().endsWith("/"), "Folder URIs must end in a slash: %s", folder);

		return new ResourceReference(folder);
	}

	/**
	 * An overload of {@link #forFolder(URI)} accepting a {@link Path} is
	 * deliberately not provided for the following reasons:
	 * <ul>
	 * <li>{@link Path#toUri()} performs I/O and may even throw an
	 * {@link java.io.IOError}. This may be unexpected to the caller.</li>
	 * <li>If given the {@link Path} refers to a folder in a revision of a
	 * version-control system, that {@link Path} may now refer to a <em>file</em> on
	 * disk. This discrepancy would result in the "trailing slash" precondition of
	 * {@link #forFolder(URI)} being violated.
	 * </ul>
	 */
	@SuppressWarnings("unused")
	private static ResourceReference forFolder(Path folder) {
		throw new UnsupportedOperationException("Deliberately not provided");
	}

	/**
	 * An overload of {@link #forFolder(URI)} accepting a {@link File} is
	 * deliberately not provided for the following reasons:
	 * <ul>
	 * <li>{@link File#toURI()} performs I/O.</li>
	 * <li>If the given {@link File} refers to a folder in a revision of a
	 * version-control system, that {@link File} may now refer to a <em>file</em> on
	 * disk. This discrepancy would result in the "trailing slash" precondition of
	 * {@link #forFolder(URI)} being violated.
	 * </ul>
	 */
	@SuppressWarnings("unused")
	private static ResourceReference forFolder(File folder) {
		throw new UnsupportedOperationException("Deliberately not provided");
	}

	@JsonCreator
	private ResourceReference(@JsonProperty(URI_PROPERTY) URI uri) {
		Preconditions.checkArgument(uri.isAbsolute(), "URI must be absolute: %s", uri);
		Preconditions.checkArgument(!uri.isOpaque(), "URI must be hierarchical: %s", uri);

		this.uri = requireNonNull(uri);
	}

	/**
	 * @return whether this {@code ResourceReference} referred to a folder when it
	 *         was created.
	 */
	public boolean isFolder() {
		return uri.getPath().endsWith("/");
	}

	/**
	 * @return whether this {@code ResourceReference} referred to a file when it was
	 *         created.
	 */
	public boolean isFile() {
		return !isFolder();
	}

	/**
	 * @return whether this {@code ResourceReference} is at or below the given other
	 *         {@code ResourceReference}.
	 */
	public boolean isAtOrBelow(ResourceReference other) {
		URI relativizedUri = other.uri.relativize(this.uri);
		// Relativizing this URI is a no-op if this URI is not at or below the
		// given other URI.
		return !relativizedUri.equals(this.uri);
	}

	/**
	 * @return an absolute URI with ({@code file} or some other scheme referring to
	 *         local resources) for this resource reference.
	 */
	public URI toUri() {
		return uri;
	}

	/**
	 * @return a {@link Path} if this {@link ResourceReference} uses a {@code file}
	 *         URI to refer to the file or folder.
	 * @throws IllegalStateException
	 *             if this {@link ResourceReference} does not have a path, as
	 *             determined by {@link #hasPath()}, or it cannot be obtained by
	 *             {@link Paths#get(URI)} for some reason.
	 */
	public Path toPath() {
		Preconditions.checkState(hasPath(), "URI must use 'file' scheme: %s", uri);

		try {
			return Paths.get(uri);
		} catch (IllegalArgumentException | FileSystemNotFoundException | SecurityException e) {
			throw new IllegalStateException(String.format("Cannot get path from URI: %s", uri), e);
		}
	}

	/**
	 * @return whether this {@code ResourceReference} uses a URI with {@code file}
	 *         scheme. Only if it does, one may call {@link #toPath()}.
	 */
	public boolean hasPath() {
		return "file".equalsIgnoreCase(uri.getScheme());
	}

	@Override
	public int hashCode() {
		return Objects.hashCode(uri);
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		ResourceReference that = (ResourceReference) other;
		return Objects.equals(this.uri, that.uri);
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}

	/**
	 * @return a string representation of this resource reference that can be used
	 *         for display purposes.
	 */
	public String toDisplayString() {
		// The 'file' scheme is always registered and we can produce a nicer
		// display string for such URIs.
		if (hasPath()) {
			return toPath().toString();
		}
		return uri.toString();
	}

	/**
	 * @return a short string representation of this resource reference that can be
	 *         used for display purposes. Note that it might have been shortened
	 *         such that it no longer unambiguously identifies a resource. If that
	 *         is a problem, use {@link #toDisplayString()} instead.
	 */
	public String toShortDisplayString() {
		// The 'file' scheme is always registered and we can produce a nicer
		// display string for such URIs.
		if (hasPath()) {
			return toPath().getFileName().toString();
		}
		return uri.toString();
	}
}
