/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.model;

import java.net.URI;
import java.util.List;

import org.conqat.lib.commons.uniformpath.RelativeUniformPath;
import org.conqat.lib.commons.uniformpath.UniformPath;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.google.common.net.UrlEscapers;
import com.teamscale.ide.commons.util.UriUtils;

/**
 * Utility methods for interfacing between {@link ResourceReference}s and
 * uniform paths.
 */
public class ResourceReferenceUtils {

	/**
	 * Creates a {@link RelativeUniformPath} from a given {@link ResourceReference}
	 * w.r.t. another base {@link ResourceReference}.
	 */
	public static RelativeUniformPath computeRelativeUniformPath(ResourceReference base, ResourceReference resource) {
		Preconditions.checkArgument(resource.isAtOrBelow(base), "Reference %s must be at or below base %s", resource,
				base);

		URI baseUri = base.toUri();
		URI resourceUri = resource.toUri();
		URI relativeUri = baseUri.relativize(resourceUri);

		List<String> unescapedSegments = UriUtils.getPathSegments(relativeUri);
		List<String> uniformPathSegments = Lists.transform(unescapedSegments, UniformPath::escapeSegment);

		if (resource.isFolder()) {
			// Drop the empty segment after a folder URI's trailing slash
			uniformPathSegments = uniformPathSegments.subList(0, uniformPathSegments.size() - 1);
		}
		return RelativeUniformPath.of(uniformPathSegments);
	}

	/**
	 * Resolves a {@link ResourceReference} from a given base folder
	 * {@link ResourceReference} and a {@link RelativeUniformPath}.
	 * <p>
	 * Assumes that the given {@link RelativeUniformPath} corresponds to a folder.
	 */
	public static ResourceReference resolveFolder(ResourceReference baseFolder,
			RelativeUniformPath relativeUniformPath) {
		return resolve(baseFolder, relativeUniformPath, true);
	}

	/**
	 * Resolves a {@link ResourceReference} from a given base folder
	 * {@link ResourceReference} and a {@link RelativeUniformPath}.
	 * <p>
	 * Assumes that the given {@link RelativeUniformPath} corresponds to a file.
	 */
	public static ResourceReference resolveFile(ResourceReference baseFolder, RelativeUniformPath relativeUniformPath) {
		return resolve(baseFolder, relativeUniformPath, false);
	}

	/**
	 * Resolves a {@link ResourceReference} from a given base folder
	 * {@link ResourceReference} and a {@link RelativeUniformPath}.
	 */
	private static ResourceReference resolve(ResourceReference baseFolder, RelativeUniformPath relativeUniformPath,
			boolean isFolder) {
		Preconditions.checkArgument(baseFolder.isFolder());

		List<String> unescapedSegments = Lists.transform(relativeUniformPath.getSegments(),
				UniformPath::unescapeSegment);
		List<String> urlEscapedPathComponents = Lists.transform(unescapedSegments,
				UrlEscapers.urlPathSegmentEscaper()::escape);

		URI baseFolderUri = baseFolder.toUri();
		URI relativeUri = URI.create("./" + String.join("/", urlEscapedPathComponents));
		URI resolvedUri = baseFolderUri.resolve(relativeUri);
		if (isFolder) {
			return ResourceReference.forFolder(UriUtils.ensurePathHasTrailingSlash(resolvedUri));
		} else {
			return ResourceReference.forFile(resolvedUri);
		}
	}
}
