/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.notification;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.NoSuchElementException;

import com.teamscale.ide.commons.findings.PreCommitResult;
import com.teamscale.ide.commons.mappings.MappingProblem;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.notification.impl.ApiNotSupportedNotification;
import com.teamscale.ide.commons.notification.impl.CheckDescriptionNotFoundNotification;
import com.teamscale.ide.commons.notification.impl.ConnectionErrorNotification;
import com.teamscale.ide.commons.notification.impl.FailedOpenUrisNotification;
import com.teamscale.ide.commons.notification.impl.HigherServerVersionWarningNotification;
import com.teamscale.ide.commons.notification.impl.NoMappingFoundNotification;
import com.teamscale.ide.commons.notification.impl.PreCommitErrorNotification;
import com.teamscale.ide.commons.notification.impl.ProjectNotFoundNotification;
import com.teamscale.ide.commons.notification.impl.ResourceOutsideMappingRootNotification;
import com.teamscale.ide.commons.notification.impl.SapTeamscaleSyncTimeoutNotification;
import com.teamscale.ide.commons.notification.impl.ServerUnknownNotification;
import com.teamscale.ide.commons.notification.impl.ServerUrlNotFoundNotification;
import com.teamscale.ide.commons.notification.impl.TeamscaleBranchNotFoundNotification;
import com.teamscale.ide.commons.notification.impl.TomlConfigInvalidNotification;
import com.teamscale.ide.commons.notification.impl.legacy.MappingRootNotConfiguredNotification;
import com.teamscale.ide.commons.notification.impl.legacy.NoServersConfiguredNotification;
import com.teamscale.ide.commons.notification.impl.legacy.NoTeamscaleProjectNotification;
import com.teamscale.ide.commons.notification.impl.legacy.NoTeamscaleServerNotification;
import com.teamscale.ide.commons.servers.ETeamscaleServerState;
import com.teamscale.ide.commons.util.TeamscaleVersion;

/**
 * Creates the implementations of {@link INotification}
 */
public final class Notifications {

	/** @see ApiNotSupportedNotification */
	public static INotification<?> apiNotSupported(URI serverUrl, String lowestSupportedServerVersion, String message) {
		return new ApiNotSupportedNotification(serverUrl, lowestSupportedServerVersion, message);
	}

	/** @see CheckDescriptionNotFoundNotification */
	public static INotification<?> higherServerVersionWarning(URI serverUrl, TeamscaleVersion serverVersion,
			TeamscaleVersion pluginVersion) {
		return new HigherServerVersionWarningNotification(serverUrl, serverVersion, pluginVersion);
	}

	/** @see ConnectionErrorNotification */
	public static INotification<?> connectionError(URI serverUrl, ETeamscaleServerState serverState) {
		return new ConnectionErrorNotification(serverUrl, serverState);
	}

	/**
	 * Creates the corresponding {@link INotification} for the given
	 * {@link MappingProblem}
	 */
	public static INotification<?> mappingProblem(ResourceReference resource, MappingProblem mappingProblem) {
		switch (mappingProblem.getType()) {
		case RESOURCE_OUTSIDE_MAPPING_ROOT:
			return new ResourceOutsideMappingRootNotification(resource);
		case TEAMSCALE_SERVER_UNKNOWN:
			return new ServerUnknownNotification(mappingProblem.getMappingRoot().orElse(null),
					mappingProblem.getContext(0));
		case NO_MAPPING_FOUND:
			return new NoMappingFoundNotification(resource);
		case TOML_CONFIG_INVALID:
			return new TomlConfigInvalidNotification(
					mappingProblem.getMappingRoot().orElseThrow(NoSuchElementException::new),
					mappingProblem.getContext(0));

		// Legacy notifications
		case NO_SERVERS_CONFIGURED:
			return new NoServersConfiguredNotification();
		case MAPPING_ROOT_NOT_CONFIGURED:
			return new MappingRootNotConfiguredNotification(
					mappingProblem.getMappingRoot().orElseThrow(NoSuchElementException::new));
		case NO_TEAMSCALE_SERVER:
			return new NoTeamscaleServerNotification(
					mappingProblem.getMappingRoot().orElseThrow(NoSuchElementException::new));
		case NO_TEAMSCALE_PROJECT:
			return new NoTeamscaleProjectNotification(
					mappingProblem.getMappingRoot().orElseThrow(NoSuchElementException::new));
		default:
			throw new IllegalArgumentException("Unknown mapping problem: " + mappingProblem);
		}
	}

	/**
	 * Creates a list of all {@link PreCommitErrorNotification notifications}
	 * necessary to report the error in the {@link PreCommitResult}.
	 * 
	 * @see PreCommitErrorNotification
	 */
	public static Collection<INotification<?>> preCommitError(PreCommitResult result) {
		Collection<INotification<?>> notifications = new ArrayList<>();

		for (PreCommitResult.PreCommitErrorDetail error : result.getGlobalErrorDetails()) {
			notifications.add(
					new PreCommitErrorNotification(error.getType(), error.getMessage(), error.invalidatesResults()));
		}
		for (String path : result.getAnalyzedPathsWithErrors()) {
			for (PreCommitResult.PreCommitErrorDetail error : result.getErrors(path)) {
				notifications.add(new PreCommitErrorNotification(error.getType(), error.getMessage(),
						error.invalidatesResults()));
			}
		}

		return notifications;
	}

	/** @see ProjectNotFoundNotification */
	public static INotification<?> projectNotFound(URI serverUrl, String projectId) {
		return new ProjectNotFoundNotification(serverUrl, projectId);
	}

	/** @see ServerUrlNotFoundNotification */
	public static INotification<?> serverUrlNotFound(URI serverUrl) {
		return new ServerUrlNotFoundNotification(serverUrl);
	}

	/** @see TeamscaleBranchNotFoundNotification */
	public static INotification<?> teamscaleBranchNotFound(URI serverUrl, String projectId, String branch) {
		return new TeamscaleBranchNotFoundNotification(serverUrl, projectId, branch);
	}

	/** @see FailedOpenUrisNotification */
	public static INotification<?> failedOpenUris(Collection<URI> failedUris) {
		return new FailedOpenUrisNotification(failedUris);
	}

	/** @see CheckDescriptionNotFoundNotification */
	public static INotification<?> checkDescriptionNotFound(String checkId) {
		return new CheckDescriptionNotFoundNotification(checkId);
	}

	/** @see SapTeamscaleSyncTimeoutNotification */
	public static INotification<?> sapTeamscaleSyncTimeout(String sapSystemId) {
		return new SapTeamscaleSyncTimeoutNotification(sapSystemId);
	}
}