/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.notification.impl;

import static java.util.Objects.requireNonNull;
import static java.util.stream.Collectors.joining;

import java.net.URI;
import java.text.MessageFormat;
import java.util.Collection;

import com.teamscale.ide.commons.lang.ToStringHelpers;
import com.teamscale.ide.commons.notification.INotification;

/**
 * Notifies the user that specific URIs could not be opened in the browser.
 */
public final class FailedOpenUrisNotification implements INotification<FailedOpenUrisNotification> {

	private final Collection<URI> failedUris;

	public FailedOpenUrisNotification(Collection<URI> failedUris) {
		this.failedUris = requireNonNull(failedUris);
	}

	@Override
	public boolean isSubsumedBy(FailedOpenUrisNotification that) {
		return that.getFailedUris().containsAll(failedUris);
	}

	@Override
	public void accept(INotificationVisitor visitor) {
		visitor.reportFailedOpenUris(this);
	}

	public Collection<URI> getFailedUris() {
		return failedUris;
	}

	/**
	 * Returns the {@link #getFailedUris() failed URIs} as a human-readable string
	 */
	public String getFailedUrisFormatted() {
		return getFailedUris().stream().map(uri -> MessageFormat.format("<{0}>", uri)).collect(joining(", "));
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
