/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.notification.impl;

import static com.teamscale.ide.commons.findings.PreCommitResult.PreCommitErrorDetail.PRE_COMMIT_API_NOT_SUPPORTED;
import static java.util.Objects.requireNonNull;
import static org.conqat.engine.index.shared.PreCommit3Result.EPrecommit3ErrorType.ANALYSIS_NOT_LIVE;
import static org.conqat.engine.index.shared.PreCommit3Result.EPrecommit3ErrorType.BRANCH_UNKNOWN;
import static org.conqat.engine.index.shared.PreCommit3Result.EPrecommit3ErrorType.FILE_COUNT_LIMIT;
import static org.conqat.engine.index.shared.PreCommit3Result.EPrecommit3ErrorType.FILE_EXCLUDED_BY_PROJECT_CONFIGURATION;
import static org.conqat.engine.index.shared.PreCommit3Result.EPrecommit3ErrorType.FILE_SIZE_LIMIT;
import static org.conqat.engine.index.shared.PreCommit3Result.EPrecommit3ErrorType.NO_VALID_CONNECTOR;
import static org.conqat.engine.index.shared.PreCommit3Result.EPrecommit3ErrorType.PRECOMMIT_BRANCH_NOT_AVAILABLE_DURING_POLL;
import static org.conqat.engine.index.shared.PreCommit3Result.EPrecommit3ErrorType.PRECOMMIT_COMMIT_ANALYSIS_DISABLED_ON_SERVER;
import static org.conqat.engine.index.shared.PreCommit3Result.EPrecommit3ErrorType.TEAMSCALE_LICENSE_ERROR;
import static org.conqat.engine.index.shared.PreCommit3Result.EPrecommit3ErrorType.USER_TIME_LIMIT;

import java.util.Objects;

import org.conqat.engine.index.shared.PreCommit3Result.EPrecommit3ErrorType;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.ImmutableSet;
import com.teamscale.ide.commons.lang.ToStringHelpers;
import com.teamscale.ide.commons.notification.INotification;

/**
 * Notifies the user about a
 * {@link com.teamscale.ide.commons.findings.PreCommitResult.PreCommitErrorDetail
 * PreCommitErrorDetail}
 */
public final class PreCommitErrorNotification implements INotification<PreCommitErrorNotification> {

	/**
	 * All types where the message should only be shown once without regard to the
	 * {@link #message}
	 */
	@VisibleForTesting
	static final ImmutableSet<String> SHOW_ONLY_ONE_PER_TYPE = ImmutableSet.of(
			PRECOMMIT_COMMIT_ANALYSIS_DISABLED_ON_SERVER.toString(), FILE_COUNT_LIMIT.toString(),
			FILE_SIZE_LIMIT.toString(), USER_TIME_LIMIT.toString(), TEAMSCALE_LICENSE_ERROR.toString(),
			NO_VALID_CONNECTOR.toString(), ANALYSIS_NOT_LIVE.toString(), PRE_COMMIT_API_NOT_SUPPORTED);

	/**
	 * All types where the message should only be shown once with regard to the
	 * {@link #message}
	 */
	@VisibleForTesting
	static final ImmutableSet<String> SHOW_ONLY_ONE_PER_MESSAGE = ImmutableSet.of(
			PRECOMMIT_BRANCH_NOT_AVAILABLE_DURING_POLL.toString(), FILE_EXCLUDED_BY_PROJECT_CONFIGURATION.toString(),
			BRANCH_UNKNOWN.toString());

	/**
	 * @see EPrecommit3ErrorType
	 */
	private final String type;
	private final String message;

	/**
	 * Indicates whether this notification should be displayed as error or warning.
	 */
	private final boolean error;

	public PreCommitErrorNotification(String type, String message, boolean isError) {
		this.type = requireNonNull(type);
		this.message = message != null ? message : "";
		this.error = isError;
	}

	@Override
	public boolean isSubsumedBy(PreCommitErrorNotification that) {
		if (!Objects.equals(type, that.getType())) {
			return false;
		}
		if (SHOW_ONLY_ONE_PER_TYPE.contains(type)) {
			return true;
		}
		if (SHOW_ONLY_ONE_PER_MESSAGE.contains(type)) {
			return Objects.equals(message, that.getMessage());
		}
		return false;
	}

	@Override
	public void accept(INotificationVisitor visitor) {
		visitor.reportPreCommitError(this);
	}

	public String getType() {
		return type;
	}

	public String getMessage() {
		return message;
	}

	public boolean isError() {
		return error;
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
