/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.notification.services;

import java.util.Set;

import com.google.common.base.Preconditions;
import com.google.common.collect.MultimapBuilder;
import com.google.common.collect.SetMultimap;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.notification.INotification;

/**
 * Remembers all {@link INotification notifications} and specifies if they
 * should be shown to the user.
 * <p>
 * The implementation of this interface has to be thread-safe.
 */
public interface INotificationMemory {

	/**
	 * Checks if the {@link INotification} should be shown to the user
	 */
	<N extends INotification<N>> boolean shouldShow(ENotificationCause cause, INotification<N> notification);

	/**
	 * An implementation where
	 * {@link #shouldShow(ENotificationCause, INotification)} will always return
	 * {@code true}
	 */
	static INotificationMemory createShowAlwaysMemory() {
		return new INotificationMemory() {
			@Override
			public <N extends INotification<N>> boolean shouldShow(ENotificationCause cause,
					INotification<N> notification) {
				return true;
			}
		};
	}

	/**
	 * An implementation where most {@link INotification notifications} are only
	 * shown once.
	 */
	static INotificationMemory createDefaultMemory() {
		return new DefaultNotificationMemory();
	}

	/**
	 * Implementation of {@link #createDefaultMemory()}
	 */
	class DefaultNotificationMemory implements INotificationMemory {

		private final SetMultimap<Class<?>, INotification<?>> memoryMap = MultimapBuilder.hashKeys().hashSetValues()
				.build();

		@Override
		public synchronized <N extends INotification<N>> boolean shouldShow(ENotificationCause cause,
				INotification<N> notification) {
			Set<INotification<?>> rememberedNotifications = memoryMap.get(notification.getClass());
			if (isSubsumedBy(notification, rememberedNotifications)) {
				return cause == ENotificationCause.DIRECT;
			}

			rememberedNotifications.add(notification);
			return true;
		}

		@SuppressWarnings("unchecked")
		private static <N extends INotification<N>> boolean isSubsumedBy(INotification<N> notification,
				Set<INotification<?>> rememberedNotifications) {
			for (INotification<?> memoryNotification : rememberedNotifications) {
				Preconditions.checkState(memoryNotification.getClass() == notification.getClass(),
						"Notification {} got stored in wrong compartment", notification);
				if (notification.isSubsumedBy((N) memoryNotification)) {
					return true;
				}
			}
			return false;
		}
	}
}
