/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.plugin;

/**
 * The registry for any {@linkplain IPluginService plug-in services} used by an
 * IDE plug-in.
 * <p>
 * In most cases, there will be a single {@link IServiceRegistry} instance
 * existing for the entire lifecycle of the plug-in, but this need not be the
 * case: The IntelliJ plug-in, for example, uses one {@link IServiceRegistry}
 * instance per open project open plus an additional {@link IServiceRegistry}
 * instance for those services that are available even if no project has been
 * opened.
 * <p>
 * How to obtain an {@link IServiceRegistry} instance is specific to each IDE
 * plug-in. In most case, it will itself be registered with the IDE in question
 * as some kind of service (not to be confused with an {@link IPluginService}).
 * <p>
 * Plug-in services are never removed from a particular {@link IServiceRegistry}
 * instance. This means that client code can keep references to services
 * {@link #getService(Class) retrieved from the registry} for as long as it
 * likes; there is no need to look up the same service interface again and
 * again.
 */
public interface IServiceRegistry {

	/**
	 * Checks whether a {@linkplain IPluginService plug-in services} implementing
	 * the given service interface has been registered.
	 *
	 * @throws IllegalArgumentException
	 *             if the given {@link Class} is not an interface
	 */
	<T extends IPluginService> boolean hasService(Class<T> serviceInterface);

	/**
	 * Returns the registered plug-in service implementing the given service
	 * interface.
	 *
	 * @throws IllegalArgumentException
	 *             if the given {@link Class} is not an interface
	 * @throws IllegalStateException
	 *             if no plug-in service implementing the given interface has been
	 *             registered
	 */
	<T extends IPluginService> T getService(Class<T> serviceInterface);
}
