/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.teamscale.ide.commons.plugin.impl;

import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

import com.google.common.base.Preconditions;
import com.teamscale.ide.commons.plugin.IPluginService;
import com.teamscale.ide.commons.plugin.IServiceRegistry;

/**
 * Default implementation of a registry for {@link IPluginService plug-in
 * services}.
 * <p>
 * All services have to be {@linkplain #registerService(IPluginService)
 * registered} during plug-in startup, before any services are retrieved, as the
 * implementation does not support concurrent modifications of the underlying
 * {@linkplain #registeredServices data-structure}.
 */
public class ServiceRegistry implements IServiceRegistry {

	/** Registry for services. */
	private final Map<Class<? extends IPluginService>, IPluginService> registeredServices = new HashMap<>();

	/** Adds a service to the registry. */
	public void registerService(IPluginService service) {
		Class<? extends IPluginService> serviceClass = service.getClass();
		Preconditions.checkState(!registeredServices.containsKey(serviceClass),
				"Duplicate service registration: " + serviceClass);

		registeredServices.put(serviceClass, service);
	}

	@Override
	public <T extends IPluginService> boolean hasService(Class<T> serviceInterface) {
		Preconditions.checkArgument(serviceInterface.isInterface());

		return findService(serviceInterface).isPresent();
	}

	@Override
	public <T extends IPluginService> T getService(Class<T> serviceInterface) {
		Preconditions.checkArgument(serviceInterface.isInterface());

		return findService(serviceInterface)
				.orElseThrow(() -> new IllegalStateException("Could not find service for " + serviceInterface));
	}

	private <T extends IPluginService> Optional<T> findService(Class<T> serviceClass) {
		for (Map.Entry<Class<? extends IPluginService>, IPluginService> entry : registeredServices.entrySet()) {
			if (serviceClass.isAssignableFrom(entry.getKey())) {
				T service = serviceClass.cast(entry.getValue());
				return Optional.of(service);
			}
		}
		return Optional.empty();
	}
}
