/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.plugin.services;

import com.teamscale.ide.commons.plugin.IPluginService;

/**
 * A plug-in service to execute tasks asynchronously using the IDE's underlying
 * task/job execution API.
 * <p>
 * Tasks with the same combination of {@linkplain Object#getClass()
 * implementation class} and {@linkplain TaskBase#getToken() token} are never
 * executed concurrently. Instead, depending on the task's desired
 * {@linkplain TaskBase#getInterruptionMode() interruption mode}, the currently
 * {@linkplain TaskBase#run() running} task with the same class/token
 * combination will either be
 * <ol>
 * <li>canceled, before the new task will be run (interruption mode
 * {@link ETaskInterruptionMode#CANCEL}).</li>
 * <li>finished, before the newest task with the same class/token combination
 * will be run afterwards (interruption mode
 * {@link ETaskInterruptionMode#WAIT}). All older waiting tasks with the same
 * class/token combination are discarded.</li>
 * </ol>
 * <p>
 * Tasks are typically executed asynchronously by an IDE-specific
 * {@link ITaskExecutor}, i.e., {@link #execute(TaskBase)} is non-blocking, the
 * notable exception being the implementation of {@code teamscale-dev}
 * command-line client.
 */
public interface ITaskExecutor extends IPluginService {

	/**
	 * Executes the given task using the IDE's underlying task/job execution API.
	 * Only one task with the same combination of {@linkplain Class implementation
	 * class} and {@linkplain TaskBase#getToken() token} will be
	 * {@linkplain TaskBase#run() run} at a time. Any other tasks previously
	 * {@linkplain ITaskExecutor#execute(TaskBase) executed} with the class/token
	 * combination will be either canceled (interruption mode
	 * {@link ETaskInterruptionMode#CANCEL}) or finished before the given task runs
	 * itself (interruption mode {@link ETaskInterruptionMode#WAIT}).
	 */
	void execute(TaskBase task);
}
