/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.plugin.services;

import static java.util.Objects.requireNonNull;

import com.teamscale.ide.commons.lang.ToStringHelpers;

/**
 * A task which can be {@link #run() run} and {@linkplain #cancel() canceled}.
 * Tasks should only been run indirectly by using
 * {@link ITaskExecutor#execute(TaskBase)}, which in turn which calls
 * {@link #start()}. This usually is triggered from an UI action like the click
 * of a button.
 */
public abstract class TaskBase {

	/** The desired visibility of this task. Never {@code null}. */
	private final ETaskVisibility visibility;

	/** The interruption mode of this task. Never {@code null}. */
	private final ETaskInterruptionMode interruptionMode;

	/**
	 * The token for this task. Will be used to identify tasks which should not be
	 * executed concurrently. Never {@code null}.
	 */
	private final Object token;

	/**
	 * The label of this task. It will be displayed in the UI and hence should help
	 * the end-user to identify the task and its purpose. Never {@code null}.
	 */
	private final String label;

	/** The thread on which this task is run. May be {@code null}. */
	private Thread thread;

	/** Whether cancellation of this task has been requested. */
	private volatile boolean cancelRequested = false;

	/**
	 * Creates a task with the given label, an
	 * {@linkplain ETaskVisibility#UNOBTRUSIVE unobtrusive} visibility, and an
	 * {@linkplain ETaskInterruptionMode#CANCEL} interruption mode.
	 * 
	 * @param token
	 *            tasks with the same token of the same task class are not executed
	 *            concurrently. If the given token is {@code null}, the task will be
	 *            executed independently of any already running tasks of the same
	 *            {@link Object#getClass() implementation class}.
	 * @param label
	 *            a label which is displayed in the UI and should help the end-user
	 *            identifying the task and its purpose
	 */
	protected TaskBase(Object token, String label) {
		this(token, label, ETaskVisibility.UNOBTRUSIVE, ETaskInterruptionMode.CANCEL);
	}

	/**
	 * Creates a task with the given label, the desired visibility, and an
	 * {@linkplain ETaskInterruptionMode#CANCEL} interruption mode.
	 *
	 * @param token
	 *            tasks with the same token of the same task class are not executed
	 *            concurrently. If the given token is {@code null}, the task will be
	 *            executed independently of any already running tasks of the same
	 *            {@link Object#getClass() implementation class}.
	 * @param label
	 *            a label which is displayed in the UI and should help the end-user
	 *            identifying the task and its purpose
	 * @param visibility
	 *            the desired visibility of this task's execution
	 */
	protected TaskBase(Object token, String label, ETaskVisibility visibility) {
		this(token, label, visibility, ETaskInterruptionMode.CANCEL);
	}

	/**
	 * Creates a task with the given label and desired visibility.
	 * 
	 * @param token
	 *            tasks with the same token of the same task class are not executed
	 *            concurrently. If the given token is {@code null}, the task will be
	 *            executed independently of any already running tasks of the same
	 *            {@link Object#getClass() implementation class}.
	 * @param label
	 *            a label which is displayed in the UI and should help the end-user
	 *            identifying the task and its purpose
	 * @param visibility
	 *            the desired visibility of this task's execution
	 * @param interruptionMode
	 *            the interruption mode of this task's execution
	 */
	protected TaskBase(Object token, String label, ETaskVisibility visibility, ETaskInterruptionMode interruptionMode) {
		if (token == null) {
			this.token = new Object();
		} else {
			this.token = token;
		}
		this.label = requireNonNull(label);
		this.visibility = requireNonNull(visibility);
		this.interruptionMode = requireNonNull(interruptionMode);
	}

	/**
	 * This method is only meant to be called by a {@link ITaskExecutor}
	 * implementation and only once when this task is started. The method has to be
	 * run in it's own thread using the IDE's underlying task/job execution API
	 */
	public final void start() {
		if (cancelRequested) {
			// Cancellation has been requested before this task was even
			// executed.
			return;
		}
		this.thread = Thread.currentThread();
		run();
	}

	/**
	 * Runs the task's business logic. Implementors should query
	 * {@link TaskBase#isCancelRequested()} regularly to check whether cancellation
	 * of the task has been is requested and it should cease running. If the
	 * execution requires waiting for results, {@link Thread#sleep(long)} can be
	 * used. If an {@link InterruptedException} is thrown while sleeping, the
	 * execution of this method should return immediately, since cancellation was
	 * requested. <b>Never call this method directly. Instead, use {@link #start()}.
	 * Otherwise cancellation will not work correctly. </b>
	 */
	protected abstract void run();

	/**
	 * Requests cancellation of this task.
	 */
	public final void cancel() {
		cancelRequested = true;
		if (thread != null) {
			thread.interrupt();
		}
	}

	/**
	 * Returns whether cancellation of this task has been requested. Subclasses
	 * should query this method regularly and react to a requested cancellation by
	 * aborting calculations immediately.
	 * 
	 * @return {@code true}, if cancellation of the task was requested and execution
	 *         should be stopped.
	 */
	public final boolean isCancelRequested() {
		return cancelRequested || (thread != null && thread.isInterrupted());
	}

	/** @see #token */
	public final Object getToken() {
		return token;
	}

	/** @see #label */
	public final String getLabel() {
		return label;
	}

	/** @see #visibility */
	public final ETaskVisibility getVisibility() {
		return visibility;
	}

	/** @see #interruptionMode */
	public final ETaskInterruptionMode getInterruptionMode() {
		return interruptionMode;
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
