/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.plugin.services.impl;

import java.awt.Desktop;
import java.io.IOException;
import java.net.URI;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import com.teamscale.ide.commons.plugin.services.IWebBrowserService;

/**
 * A generic {@link IWebBrowserService} that uses the {@link java.awt.Desktop}
 * API, if available. It does not respect any browser settings the user may have
 * configured within the IDE and should hence only be used if tighter IDE
 * integration is not possible.
 */
public class GenericWebBrowserService implements IWebBrowserService {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(GenericWebBrowserService.class);

	@Override
	public boolean open(URI uri) {
		LOGGER.traceEntry(null, uri);
		try {
			if (!isBrowsingSupported()) {
				return LOGGER.traceExit(false);
			}
			Desktop.getDesktop().browse(uri);
			return LOGGER.traceExit(true);
		} catch (IOException e) {
			LOGGER.catching(e);
			return LOGGER.traceExit(false);
		}
	}

	private static boolean isBrowsingSupported() {
		if (!Desktop.isDesktopSupported()) {
			LOGGER.debug("Desktop API is not supported.");
			return false;
		}
		if (!Desktop.getDesktop().isSupported(Desktop.Action.BROWSE)) {
			LOGGER.debug("Browse Action is not supported.");
			return false;
		}
		return true;
	}
}
