/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.plugin.tasks;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;

import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.notification.Notifications;
import com.teamscale.ide.commons.notification.services.INotificationService;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.ETaskVisibility;
import com.teamscale.ide.commons.plugin.services.IWebBrowserService;
import com.teamscale.ide.commons.plugin.services.TaskBase;

/**
 * Base class for tasks which open an {@link URI} in the default Web browser.
 */
public abstract class OpenInBrowserTaskBase extends TaskBase {

	/**
	 * The {@link INotificationService} subclasses should use to report errors.
	 */
	protected final INotificationService notificationService;

	private final IWebBrowserService webBrowser;

	protected OpenInBrowserTaskBase(IServiceRegistry serviceRegistry, String label) {
		super(null, label, ETaskVisibility.INVISIBLE);

		notificationService = serviceRegistry.getService(INotificationService.class);
		webBrowser = serviceRegistry.getService(IWebBrowserService.class);
	}

	@Override
	protected final void run() {
		Collection<URI> uris = createUri();
		Collection<URI> failedUris = new ArrayList<>();

		for (URI uri : uris) {
			boolean opened = webBrowser.open(uri);
			if (!opened) {
				failedUris.add(uri);
			}
		}

		if (!failedUris.isEmpty()) {
			notificationService.reportNotification(ENotificationCause.DIRECT,
					Notifications.failedOpenUris(failedUris));
		}
	}

	/**
	 * @return a set of {@link URI}s to be opened in the default Web browser or
	 *         {@link java.util.Collections#emptyList()} if no browser should be
	 *         opened, e.g., because some pre-condition is not met.
	 */
	protected abstract Collection<URI> createUri();
}
