/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.servers;

import java.net.URI;
import java.util.Objects;

import com.teamscale.ide.commons.lang.ToStringHelpers;

/**
 * Storage format of all needed data to connect to a Teamscale server.
 */
public final class CredentialsUrl {

	private final URI url;
	private final String username;
	private final String password;
	private final boolean trustAllCertificates;

	public CredentialsUrl(URI url, AuthenticationInfo authenticationInfo) {
		this(url, authenticationInfo.getUsername(), authenticationInfo.getPassword(),
				authenticationInfo.isTrustAllCertificates());
	}

	public CredentialsUrl(URI url, String username, String password, boolean trustAllCertificates) {
		this.url = Objects.requireNonNull(url);
		this.username = Objects.requireNonNull(username);
		this.password = Objects.requireNonNull(password);
		this.trustAllCertificates = trustAllCertificates;
	}

	public URI getUrl() {
		return url;
	}

	public String getUsername() {
		return username;
	}

	public String getPassword() {
		return password;
	}

	public boolean isTrustAllCertificates() {
		return trustAllCertificates;
	}

	/**
	 * Returns a new {@link CredentialsUrl} which is a copy of the current one with
	 * another trust policy.
	 */
	public CredentialsUrl withTrustAllCertificates(boolean trustAllCertificates) {
		return new CredentialsUrl(getUrl(), getUsername(), getPassword(), trustAllCertificates);
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null || getClass() != other.getClass()) {
			return false;
		}
		CredentialsUrl that = (CredentialsUrl) other;
		return Objects.equals(url, that.url) && Objects.equals(username, that.username)
				&& Objects.equals(password, that.password) && trustAllCertificates == that.trustAllCertificates;
	}

	@Override
	public int hashCode() {
		return Objects.hash(url, username, password, trustAllCertificates);
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
