/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.servers;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import jakarta.xml.bind.annotation.XmlElement;
import jakarta.xml.bind.annotation.XmlElementWrapper;
import jakarta.xml.bind.annotation.XmlRootElement;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.dataformat.xml.annotation.JacksonXmlElementWrapper;
import com.teamscale.commons.service.XmlSerializationUtils;
import com.teamscale.commons.service.client.ServerDetails;

/**
 * Adds a "symbolic name" to {@link ServerDetails} that is used to lookup the
 * ServerDetails object as configured by the user.
 */
public class NamedServerDetails extends ServerDetails {

	@XmlRootElement
	private static class NamedServerList {

		@XmlElement
		@JacksonXmlElementWrapper(useWrapping = false)
		private List<NamedServerDetails> details = new ArrayList<>();

		@XmlElementWrapper(name = "a")
		private List<NamedServerDetails> legacyDetails;

		@JsonCreator
		public NamedServerList() {
		}

		private NamedServerList(List<NamedServerDetails> serverDetails) {
			details = serverDetails;
		}

		/**
		 * @see #details
		 * @return the details or the legacy details when importing from an old version
		 *         which had an additional "a" tag around it.
		 */
		public List<NamedServerDetails> getDetails() {
			if (legacyDetails != null) {
				return legacyDetails;
			}
			return details;
		}
	}

	/**
	 * A symbolic name for the connection settings. Uniquely identifies a Teamscale
	 * server.
	 */
	private final String symbolicName;

	private final boolean trustAllCertificates;

	public NamedServerDetails(String symbolicName, String url, String username, String password) {
		this(symbolicName, url, username, password, false);
	}

	public NamedServerDetails(String symbolicName, String url, String username, String password,
			boolean trustAllCertificates) {
		super(url, username, password);
		this.symbolicName = symbolicName;
		this.trustAllCertificates = trustAllCertificates;
	}

	@JsonCreator
	public NamedServerDetails(@JsonProperty("symbolicName") String symbolicName, @JsonProperty("url") String url,
			@JsonProperty("username") String username, @JsonProperty("password") String password,
			@JsonProperty("timeoutSeconds") int timeoutSeconds,
			@JsonProperty("trustAllCertificates") boolean trustAllCertificates) {
		super(url, username, password, timeoutSeconds);
		this.symbolicName = symbolicName;
		this.trustAllCertificates = trustAllCertificates;
	}

	/** @see #symbolicName */
	public String getSymbolicName() {
		return symbolicName;
	}

	@Override
	public NamedServerDetails withTimeOut(int timeOutSeconds) {
		return new NamedServerDetails(getSymbolicName(), getUrl(), getUsername(), getPassword(), timeOutSeconds,
				isTrustAllCertificates());
	}

	@Override
	public String toString() {
		return symbolicName;
	}

	public boolean isTrustAllCertificates() {
		return trustAllCertificates;
	}

	/**
	 * Two NamedServerDetails are equal, if their symbolic name is the same.
	 */
	@Override
	public boolean equals(Object o) {
		if (o instanceof NamedServerDetails) {
			return ((NamedServerDetails) o).getSymbolicName().equals(symbolicName);
		}
		return false;
	}

	@Override
	public int hashCode() {
		return symbolicName.hashCode();
	}

	/** Serialize a list of {@code NamedServerDetails} into an XML string */
	public static String serialize(List<NamedServerDetails> serverDetails) {
		return XmlSerializationUtils.serializeToXML(new NamedServerList(serverDetails));
	}

	/** Deserialize a list of {@code NamedServerDetails} from an XML string */
	public static List<NamedServerDetails> deserialize(String serversXml) throws IOException {
		return XmlSerializationUtils.deserializeFromXML(serversXml, NamedServerList.class).getDetails();
	}

	/**
	 * Returns it as {@link CredentialsUrl} to persist it.
	 */
	public AuthenticationInfo asAuthenticationInfo() {
		return new AuthenticationInfo(getUsername(), getPassword(), isTrustAllCertificates());
	}
}
