/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.servers.services;

import java.net.URI;
import java.util.Collection;
import java.util.Map;
import java.util.Optional;

import com.google.common.collect.ImmutableMap;
import com.teamscale.ide.commons.servers.AuthenticationInfo;

/**
 * A {@link ITeamscaleServerManager} that maintains an in-memory snapshot of the
 * Teamscale servers managed by another {@link ITeamscaleServerManager}.
 */
public class CachingTeamscaleServerManager implements ITeamscaleServerManager {

	private ImmutableMap<URI, AuthenticationInfo> currentSnapshot = null;

	private final ITeamscaleServerManager delegate;

	public CachingTeamscaleServerManager(ITeamscaleServerManager delegate) {
		this.delegate = delegate;
	}

	@Override
	public Collection<URI> getAllServerUrls() {
		return getCache().keySet();
	}

	@Override
	public Optional<AuthenticationInfo> getAuthenticationInfo(URI url) {
		return Optional.ofNullable(getCache().get(url));
	}

	@Override
	public boolean hasServers() {
		return !getCache().isEmpty();
	}

	@Override
	public boolean storeCredentials(URI url, AuthenticationInfo authenticationInfo) {
		boolean persisted = delegate.storeCredentials(url, authenticationInfo);
		currentSnapshot = null;
		return persisted;
	}

	@Override
	public boolean removeServer(URI url) {
		boolean persisted = delegate.removeServer(url);
		currentSnapshot = null;
		return persisted;
	}

	private synchronized Map<URI, AuthenticationInfo> getCache() {
		if (currentSnapshot == null) {
			currentSnapshot = buildSnapshot();
		}
		return currentSnapshot;
	}

	private ImmutableMap<URI, AuthenticationInfo> buildSnapshot() {
		ImmutableMap.Builder<URI, AuthenticationInfo> builder = ImmutableMap.builder();

		for (URI url : delegate.getAllServerUrls()) {
			delegate.getAuthenticationInfo(url).ifPresent(info -> builder.put(url, info));
		}

		return builder.build();
	}
}
