/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.servers.services;

import java.net.URI;
import java.util.Collection;
import java.util.Objects;
import java.util.Optional;

import com.teamscale.ide.commons.servers.AuthenticationInfo;
import com.teamscale.ide.commons.servers.NamedServerDetails;
import com.teamscale.ide.commons.util.UriUtils;

/**
 * Manages the Teamscale servers configured in the IDE.
 */
public interface ITeamscaleServerManager extends IServerCredentialsManager {

	/**
	 * @return Returns all registered servers.
	 */
	Collection<URI> getAllServerUrls();

	/** Returns the {@link AuthenticationInfo} for the given url */
	Optional<AuthenticationInfo> getAuthenticationInfo(URI url);

	/**
	 * @return Whether there are any registered servers or not
	 */
	default boolean hasServers() {
		return !getAllServerUrls().isEmpty();
	}

	/**
	 * @param serverUrl
	 *            URL of server.
	 * @return The details of a configured server with the given URL, if it exists.
	 */
	@Override
	default Optional<NamedServerDetails> getServer(URI serverUrl) {
		// getAuthenticationInfo automatically corrects the key, but we need the
		// corrected key, too.
		Optional<URI> optKey = getAllServerUrls().stream().filter(key -> Objects
				.equals(UriUtils.ensurePathHasTrailingSlash(serverUrl), UriUtils.ensurePathHasTrailingSlash(key)))
				.findFirst();

		return optKey.flatMap(key -> getAuthenticationInfo(key).map(info -> info.asNamedServerDetails(key)));
	}

	/**
	 * Stores the {@link AuthenticationInfo}. Returns whether the changes were
	 * persisted to the disk.
	 */
	boolean storeCredentials(URI url, AuthenticationInfo authenticationInfo);

	/**
	 * Removes the {@link AuthenticationInfo} stored under the given {@link URI
	 * url}. Returns whether the changes were persisted to the disk.
	 */
	boolean removeServer(URI url);

	/**
	 * Returns {@link NamedServerDetails} which are stored in the old format.
	 */
	@Deprecated
	default NamedServerDetails getNamedServerDetails(String name) {
		throw new UnsupportedOperationException("This legacy method is not supported.");
	}
}
