/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.util;

import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.util.Objects;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import com.teamscale.ide.commons.servers.AuthenticationInfo;
import com.teamscale.ide.commons.servers.CredentialsUrl;

/**
 * Utils for {@link CredentialsUrl}
 */
public class CredentialsUrlUtils {

	private static final Logger LOGGER = LogManager.getLogger(CredentialsUrlUtils.class);

	private static final String TRUST_ALL_CERTIFICATES_FRAGMENT = "trust-all-certificates";

	/**
	 * Tries to deserialize the string to a {@link CredentialsUrl}
	 *
	 * @throws IllegalArgumentException
	 *             if the given string is not a valid {@link CredentialsUrl}
	 */
	public static CredentialsUrl deserialize(String credentialUrl) {
		try {
			return deserialize(new URI(credentialUrl));
		} catch (URISyntaxException e) {
			LOGGER.catching(e);
			throw new IllegalArgumentException(String.format("Not a valid URL: <%s>", credentialUrl), e);
		}
	}

	/**
	 * Tries to deserialize the {@link URI} to a {@link CredentialsUrl}
	 *
	 * @throws IllegalArgumentException
	 *             if the given {@link URI} is not a valid {@link CredentialsUrl}
	 */
	public static CredentialsUrl deserialize(URI credentialUrl) {
		if (credentialUrl.isOpaque()) {
			throw new IllegalArgumentException(String.format("Teamscale URL <%s> must not be opaque", credentialUrl));
		}

		if (!credentialUrl.isAbsolute()) {
			throw new IllegalArgumentException(String.format("Teamscale URL <%s> must be absolute", credentialUrl));
		}

		String urlScheme = credentialUrl.getScheme();
		if (!urlScheme.equalsIgnoreCase("http") && !urlScheme.equalsIgnoreCase("https")) {
			throw new IllegalArgumentException(String.format("Teamscale URL <%s> must be HTTP(S)", credentialUrl));
		}
		String rawUserInfo = credentialUrl.getRawUserInfo();
		if (rawUserInfo == null) {
			throw new IllegalArgumentException(
					String.format("No 'username:accesskey' given in Teamscale URL <%s>", credentialUrl));
		}

		int separator = rawUserInfo.indexOf(':');
		if (separator < 0) {
			throw new IllegalArgumentException(
					String.format("No 'accesskey' given in Teamscale URL <%s>", credentialUrl));
		}

		String username = urlDecode(rawUserInfo.substring(0, separator));
		String password = urlDecode(rawUserInfo.substring(separator + 1));
		boolean trustAllCertificates = Objects.equals(credentialUrl.getFragment(), TRUST_ALL_CERTIFICATES_FRAGMENT);
		URI simpleUrl = UriUtils.stripFragment(UriUtils.stripUserInfo(credentialUrl));

		return new CredentialsUrl(simpleUrl, username, password, trustAllCertificates);
	}

	/**
	 * Creates an {@link AuthenticationInfo} from the given {@link CredentialsUrl}
	 */
	public static AuthenticationInfo toAuthenticationInfo(CredentialsUrl credentialsUrl) {
		return new AuthenticationInfo(credentialsUrl.getUsername(), credentialsUrl.getPassword(),
				credentialsUrl.isTrustAllCertificates());
	}

	/**
	 * Serializes a {@link CredentialsUrl}
	 */
	public static String serialize(CredentialsUrl credentialsUrl) {
		URI url = credentialsUrl.getUrl();
		String rawUserInfo = urlEncode(credentialsUrl.getUsername()) + ":" + urlEncode(credentialsUrl.getPassword());
		String rawFragment = credentialsUrl.isTrustAllCertificates() ? TRUST_ALL_CERTIFICATES_FRAGMENT : null;
		return UriUtils.assemble(url.getScheme(), rawUserInfo, url.getHost(), url.getPort(), url.getRawPath(),
				url.getRawQuery(), rawFragment).toString();
	}

	private static String urlEncode(String unencoded) {
		try {
			return URLEncoder.encode(unencoded, StandardCharsets.UTF_8.name());
		} catch (UnsupportedEncodingException e) {
			throw new AssertionError("Never happens", e);
		}
	}

	private static String urlDecode(String encoded) {
		try {
			return URLDecoder.decode(encoded, StandardCharsets.UTF_8.name());
		} catch (UnsupportedEncodingException e) {
			throw new AssertionError("Never happens", e);
		}
	}
}
