package com.teamscale.ide.commons.util;

import java.util.Comparator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.checkerframework.checker.nullness.qual.NonNull;

import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.google.common.base.Strings;

/**
 * A version of a Teamscale server. For example, 9.5.6. If the last number is
 * not specified, it is filled with 0.
 * <p>
 * This IDE class corresponds to {@link org.conqat.lib.commons.version.Version}
 * in the Teamscale backend.
 */
public final class TeamscaleVersion implements Comparable<TeamscaleVersion> {

	private static final Comparator<TeamscaleVersion> COMPARATOR = Comparator
			.comparingInt((TeamscaleVersion v) -> v.major).thenComparingInt(v -> v.minor)
			.thenComparingInt(v -> v.patch);

	/**
	 * Pattern to parse Teamscale versions from a string. For example, "9.5.6" or
	 * "9.5".
	 */
	private static final Pattern VERSION_PATTERN = Pattern
			.compile("(?<major>\\d+)\\.(?<minor>\\d+)(\\.(?<patch>\\d+))?");

	/**
	 * The major version number.
	 */
	public final int major;

	/**
	 * The minor version number.
	 */
	public final int minor;

	/**
	 * The patch version number.
	 */
	public final int patch;

	public TeamscaleVersion(int major, int minor, int patch) {
		Preconditions.checkArgument(major >= 0 && minor >= 0 && patch >= 0, "Invalid version string: %d.%d.%d", major,
				minor, patch);
		this.major = major;
		this.minor = minor;
		this.patch = patch;
	}

	/**
	 * Parses a Teamscale version number from the given string. For example, from
	 * "9.5.6".
	 */
	public static TeamscaleVersion fromString(String version) {
		Matcher versionMatcher = VERSION_PATTERN.matcher(version);
		if (!versionMatcher.find()) {
			throw new IllegalArgumentException(String.format("Not a valid version string: <%s>", version));
		}

		int major = 0;
		if (!Strings.isNullOrEmpty(versionMatcher.group("major"))) {
			major = Integer.parseInt(versionMatcher.group("major"));
		}
		int minor = 0;
		if (!Strings.isNullOrEmpty(versionMatcher.group("minor"))) {
			minor = Integer.parseInt(versionMatcher.group("minor"));
		}
		int patch = 0;
		if (!Strings.isNullOrEmpty(versionMatcher.group("patch"))) {
			patch = Integer.parseInt(versionMatcher.group("patch"));
		}
		return new TeamscaleVersion(major, minor, patch);
	}

	@Override
	public String toString() {
		if (patch == 0) {
			return String.format("%d.%d", major, minor);
		}
		return String.format("%d.%d.%d", major, minor, patch);
	}

	@Override
	public int compareTo(@NonNull TeamscaleVersion o) {
		return COMPARATOR.compare(this, o);
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null || getClass() != other.getClass()) {
			return false;
		}

		TeamscaleVersion that = (TeamscaleVersion) other;
		return (major == that.major && minor == that.minor && patch == that.patch);
	}

	@Override
	public int hashCode() {
		return Objects.hashCode(major, minor, patch);
	}
}
