/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.util;

import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Objects;

import org.checkerframework.checker.nullness.qual.Nullable;

import com.google.common.base.Splitter;
import com.google.common.collect.Lists;

/**
 * Utilities for working with {@link URI}s.
 */
public class UriUtils {

	/**
	 * Assembles an opaque {@link URI} from its raw (i.e., percent-escaped) parts.
	 */
	public static URI assemble(@Nullable String scheme, @Nullable String rawSchemeSpecificPart,
			@Nullable String rawFragment) {
		return assemble(scheme, rawSchemeSpecificPart, null, null, null, rawFragment);
	}

	/**
	 * Assembles an non-opaque {@link URI} from its raw (i.e., percent-escaped)
	 * parts.
	 */
	public static URI assemble(@Nullable String scheme, @Nullable String rawUserInfo, @Nullable String rawHost,
			int port, @Nullable String rawPath, @Nullable String rawQuery, @Nullable String rawFragment) {
		return assemble(scheme, null, rawUserInfo, rawHost, port, rawPath, rawQuery, rawFragment);
	}

	/**
	 * Assembles an {@link URI} from its raw (i.e., percent-escaped) parts.
	 * <p>
	 * Modelled after
	 * {@link URI#toString(String,String,String,String,String,int,String,String,String)}.
	 */
	private static URI assemble(@Nullable String scheme, @Nullable String rawSchemeSpecificPart,
			@Nullable String rawUserInfo, @Nullable String rawHost, int port, @Nullable String rawPath,
			@Nullable String rawQuery, @Nullable String rawFragment) {

		String rawAuthority = null;

		if (rawHost != null) {
			StringBuilder rawAuthorityBuilder = new StringBuilder();

			if (rawUserInfo != null) {
				rawAuthorityBuilder.append(rawUserInfo).append('@');
			}
			rawAuthorityBuilder.append(rawHost);
			if (port >= 0) {
				rawAuthorityBuilder.append(':');
				rawAuthorityBuilder.append(port);
			}

			rawAuthority = rawAuthorityBuilder.toString();
		}

		return assemble(scheme, rawSchemeSpecificPart, rawAuthority, rawPath, rawQuery, rawFragment);
	}

	/**
	 * Assembles an {@link URI} from its raw (i.e., percent-escaped) parts.
	 * <p>
	 * Modelled after
	 * {@link URI#toString(String,String,String,String,String,int,String,String,String)}.
	 */
	private static URI assemble(@Nullable String scheme, @Nullable String rawSchemeSpecificPart,
			@Nullable String rawAuthority, @Nullable String rawPath, @Nullable String rawQuery,
			@Nullable String rawFragment) {
		StringBuilder uri = new StringBuilder();
		if (scheme != null) {
			uri.append(scheme).append(':');
		}
		if (rawSchemeSpecificPart != null) {
			uri.append(rawSchemeSpecificPart);
		} else {
			if (rawAuthority != null) {
				uri.append("//");
				uri.append(rawAuthority);
			}
			if (rawPath != null) {
				uri.append(rawPath);
			}
			if (rawQuery != null) {
				uri.append(rawQuery);
			}
			if (rawFragment != null) {
				uri.append('#').append(rawFragment);
			}
		}
		return URI.create(uri.toString());
	}

	/**
	 * Returns the given {@link URI}, with a path ending in a slash if it did not do
	 * so already.
	 */
	public static URI ensurePathHasTrailingSlash(URI uri) {
		if (uri.isOpaque()) {
			// Never has a dedicated path
			return uri;
		}
		return assemble(uri.getScheme(), uri.getRawUserInfo(), uri.getHost(), uri.getPort(),
				withTrailingSlash(uri.getRawPath()), uri.getRawQuery(), uri.getRawFragment());
	}

	private static String withTrailingSlash(@Nullable String rawPath) {
		if (rawPath == null) {
			return null;
		}
		if (!rawPath.endsWith("/")) {
			return rawPath + "/";
		}
		return rawPath;
	}

	/**
	 * Returns the given {@link URI}, with any {@linkplain URI#getUserInfo()
	 * user-information component} (if present) removed.
	 */
	public static URI stripUserInfo(URI uri) {
		if (uri.isOpaque()) {
			// Never has user-information
			return uri;
		}
		return assemble(uri.getScheme(), null, uri.getHost(), uri.getPort(), withTrailingSlash(uri.getRawPath()),
				uri.getRawQuery(), uri.getRawFragment());
	}

	/**
	 * Return the given {@link URI}, with the given authority.
	 */
	public static URI replaceAuthority(URI uri, String authority) {
		return assemble(uri.getScheme(), null, authority, uri.getRawPath(), uri.getRawQuery(), uri.getRawFragment());
	}

	/**
	 * Returns the given {@link URI}, with any {@linkplain URI#getFragment()}
	 * fragment} (if present) removed.
	 */
	public static URI stripFragment(URI uri) {
		return assemble(uri.getScheme(), uri.getRawUserInfo(), uri.getHost(), uri.getPort(),
				withTrailingSlash(uri.getRawPath()), uri.getRawQuery(), null);
	}

	/**
	 * @return the path segments of the given URI, in unescaped form.
	 */
	public static List<String> getPathSegments(URI uri) {
		List<String> uriEscapedSegments = Splitter.on('/').splitToList(uri.getRawPath());
		return Lists.transform(uriEscapedSegments, UriUtils::unescapeURIPathSegment);
	}

	private static String unescapeURIPathSegment(String pathSegment) throws AssertionError {
		try {
			return URLDecoder.decode(pathSegment, StandardCharsets.UTF_8.name());
		} catch (UnsupportedEncodingException e) {
			throw new AssertionError("Cannot happen; UTF-8 is always supported", e);
		}
	}

	/**
	 * Compares two {@link URI URIs} ignoring the trailing slashes of both.
	 */
	public static boolean compareUriIgnoringTrailingSlash(URI uri1, URI uri2) {
		return Objects.equals(UriUtils.ensurePathHasTrailingSlash(uri1), UriUtils.ensurePathHasTrailingSlash(uri2));
	}
}
