package com.teamscale.ide.commons.util.markdown;

import org.commonmark.node.HardLineBreak;
import org.commonmark.node.Node;
import org.commonmark.node.Paragraph;
import org.commonmark.node.Text;
import org.commonmark.node.Visitor;
import org.commonmark.parser.Parser;
import org.commonmark.parser.PostProcessor;
import org.commonmark.renderer.html.HtmlRenderer;

import com.google.common.html.HtmlEscapers;

/**
 * Utilities for working with (<a href="https://commonmark.org/">CommonMark</a>)
 * Markdown.
 */
public class MarkdownUtils {

	/**
	 * Renders the given (<a href="https://commonmark.org/">CommonMark</a>) Markdown
	 * text as HTML.
	 */
	public static String renderToHtml(String markdown) {
		return renderToHtml(markdown, 0);
	}

	/**
	 * Renders the given (<a href="https://commonmark.org/">CommonMark</a>) Markdown
	 * text as HTML, applying the given CSS rules.
	 */
	public static String renderToHtml(String markdown, String css) {
		return renderToHtml(markdown, css, 0);
	}

	/**
	 * Renders the given (<a href="https://commonmark.org/">CommonMark</a>) Markdown
	 * text as HTML and truncates all text after the specified character limit.
	 */
	public static String renderToHtml(String markdown, int charLimit) {
		return renderToHtml(markdown, "", charLimit);
	}

	/**
	 * Renders the given (<a href="https://commonmark.org/">CommonMark</a>) Markdown
	 * text as HTML, applying the given CSS rules, and truncates all text
	 * after the specified character limit.
	 */
	public static String renderToHtml(String markdown, String css, int charLimit) {
		Node document;
		if (charLimit == 0) {
			document = parseMarkdown(markdown);
		} else {
			document = parseMarkdownTruncated(markdown, charLimit);
		}
		HtmlRenderer renderer = HtmlRenderer.builder().build();
		String htmlHead;
		if (!css.isEmpty()) {
			htmlHead = String.format("<style>%s</style>", HtmlEscapers.htmlEscaper().escape(css));
		} else {
			htmlHead = "";
		}
		String htmlBody = renderer.render(document);
		return String.format("<html><head>%s</head><body>%s</body></html>", htmlHead, htmlBody);
	}

	/**
	 * Processes the given (<a href="https://commonmark.org/">CommonMark</a>)
	 * Markdown text with a {@link Visitor}.
	 */
	public static void visitMarkdown(String markdown, Visitor visitor) {
		Node document = parseMarkdown(markdown);
		document.accept(visitor);
	}

	/**
	 * Parse markdown and apply the provided post processors
	 */
	private static Node parseMarkdown(String markdown, PostProcessor... postProcessors) {
		Parser.Builder builder = Parser.builder();
		for (PostProcessor pp : postProcessors) {
			builder.postProcessor(pp);
		}
		Parser parser = builder.build();
		return parser.parse(markdown);
	}

	/**
	 * Parse markdown and insert newline Characters, when the character limit is
	 * exceeded.
	 */
	private static Node parseMarkdownTruncated(String markdown, int charLimit) {
		return parseMarkdown(markdown, new LineLengthLimiter(charLimit));
	}

	/**
	 * Post processor, that limits the maximum line length by iterating over all
	 * elements in the AST and inserting newlines when needed.
	 */
	private static class LineLengthLimiter implements PostProcessor {

		public LineLengthLimiter(int charLimit) {
			this.charLimit = charLimit;
		}

		private final int charLimit;

		private int currentLen = 0;

		@Override
		public Node process(Node node) {

			if (node instanceof Text) {
				limit((Text) node, charLimit);
			}

			Node child = node.getFirstChild();
			if (child != null) {
				process(child);
			}

			if (node instanceof HardLineBreak || node instanceof Paragraph) { // reset current character length
				// after line breaks and paragraphs
				currentLen = 0;
			}

			Node next = node.getNext();
			if (next != null) {
				process(next);
			}
			return node;
		}

		private void limit(Text node, int len) {
			String literal = node.getLiteral();

			while (currentLen + literal.length() > len) {
				Text newNode = new Text(literal.substring(0, len - currentLen));
				node.insertBefore(newNode);
				node.insertBefore(new HardLineBreak());
				literal = literal.substring(len - currentLen);
				currentLen = 0;
			}
			node.setLiteral(literal);
			currentLen += literal.length();
		}
	}
}
