/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.vcs;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.Set;

import com.google.common.base.MoreObjects;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.teamscale.ide.commons.model.ResourceReference;

/**
 * The set of changes between two revisions in a version-control system.
 * <p>
 * This class is immutable.
 */
public final class ChangeSet {

	/** {@code ChangeSet} that's completely empty */
	public static final ChangeSet EMPTY = new ChangeSet(ImmutableSet.of(), ImmutableSet.of(), ImmutableSet.of());

	/**
	 * A state a changed file can be in.
	 */
	private enum ChangeState {

		CREATED {

			@Override
			ChangeState andThen(ChangeState later) {
				switch (later) {
				case CREATED:
					throw new IllegalArgumentException(later.name());
				case REMOVED:
					return null; // this and the later state cancel each other
				default:
					return this;
				}
			}
		},
		MODIFIED {

			@Override
			ChangeState andThen(ChangeState later) {
				switch (later) {
				case CREATED:
					throw new IllegalArgumentException(later.name());
				default:
					return later;
				}
			}
		},
		REMOVED {

			@Override
			ChangeState andThen(ChangeState later) {
				switch (later) {
				case CREATED:
					return MODIFIED;
				default:
					throw new IllegalArgumentException(later.name());
				}
			}
		};

		/** Returns the resulting state if this state is followed by a later state. */
		abstract ChangeState andThen(ChangeState later);
	}

	/**
	 * As the state {@linkplain ChangeState#andThen(ChangeState) transition
	 * function} is rather complex, we keep track of the change state every
	 * individual resource is in and update that. The naive alternative, namely one
	 * set per state, is really complex to merge in
	 * {@link ChangeSet#andThen(ChangeSet)}.
	 */
	private final ImmutableMap<ResourceReference, ChangeState> changes;

	public ChangeSet(Set<ResourceReference> createdFiles, Set<ResourceReference> modifiedFiles,
			Set<ResourceReference> removedFiles) {
		ImmutableMap.Builder<ResourceReference, ChangeState> changes = ImmutableMap.builder();
		for (ResourceReference createdFile : createdFiles) {
			changes.put(createdFile, ChangeState.CREATED);
		}
		for (ResourceReference modifiedFile : modifiedFiles) {
			changes.put(modifiedFile, ChangeState.MODIFIED);
		}
		for (ResourceReference removedFile : removedFiles) {
			changes.put(removedFile, ChangeState.REMOVED);
		}
		this.changes = changes.build();
	}

	private ChangeSet(Map<ResourceReference, ChangeState> changes) {
		this.changes = ImmutableMap.copyOf(changes);
	}

	/**
	 * @return all newly created files in this change set
	 */
	public ImmutableSet<ResourceReference> getCreatedFiles() {
		return getFilesForState(ChangeState.CREATED);
	}

	/**
	 * @return all files that have been modified in this change set, but which were
	 *         not newly {@linkplain #getCreatedFiles() created}.
	 */
	public ImmutableSet<ResourceReference> getModifiedFiles() {
		return getFilesForState(ChangeState.MODIFIED);
	}

	/**
	 * @return all files that have been removed in this change set.
	 */
	public ImmutableSet<ResourceReference> getRemovedFiles() {
		return getFilesForState(ChangeState.REMOVED);
	}

	private ImmutableSet<ResourceReference> getFilesForState(ChangeState state) {
		return changes.asMultimap().inverse().get(state);
	}

	/**
	 * Whether this change set contains any created, modified, or removed files.
	 */
	public boolean isEmpty() {
		return changes.isEmpty();
	}

	/**
	 * Returns a new change set combining the changes from this change set with a
	 * later change set. For example, files that were modified in this change set
	 * but later removed will be considered removed in the returned change set.
	 */
	public ChangeSet andThen(ChangeSet later) {
		Map<ResourceReference, ChangeState> changes = new HashMap<>(this.changes);
		for (Entry<ResourceReference, ChangeState> entry : later.changes.entrySet()) {
			changes.merge(entry.getKey(), entry.getValue(), ChangeState::andThen);
		}
		return new ChangeSet(changes);
	}

	/**
	 * @return a new, filtered change set that only includes changes below the given
	 *         {@code roots}.
	 */
	public ChangeSet filter(Collection<ResourceReference> roots) {
		Set<ResourceReference> createdFiles = filterBySelection(getCreatedFiles(), roots);
		Set<ResourceReference> modifiedFiles = filterBySelection(getModifiedFiles(), roots);
		Set<ResourceReference> removedFiles = filterBySelection(getRemovedFiles(), roots);
		return new ChangeSet(createdFiles, modifiedFiles, removedFiles);
	}

	private static Set<ResourceReference> filterBySelection(Set<ResourceReference> references,
			Collection<ResourceReference> roots) {
		Set<ResourceReference> filteredReferences = new HashSet<>();
		for (ResourceReference reference : references) {
			for (ResourceReference root : roots) {
				if (reference.isAtOrBelow(root)) {
					filteredReferences.add(reference);
					break;
				}
			}
		}
		return filteredReferences;
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		ChangeSet that = (ChangeSet) other;
		return Objects.equals(this.changes, that.changes);
	}

	@Override
	public int hashCode() {
		return Objects.hash(changes);
	}

	@Override
	public String toString() {
		return MoreObjects.toStringHelper(this) //
				.add("created", getCreatedFiles()) //
				.add("modified", getModifiedFiles()) //
				.add("removed", getRemovedFiles()).toString();
	}
}
