/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.vcs;

import java.io.IOException;
import java.util.Optional;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.vcs.services.ICheckoutFactory;

/**
 * Utility class to make it easier to work with {@link ICheckout} if a checkout
 * is needed only temporarily.
 */
public class CheckoutUtils {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(CheckoutUtils.class);

	/**
	 * @return the branch that is currently checked out and contains the given
	 *         resource, if any
	 * 
	 * @see ICheckout#getCheckedOutBranch()
	 */
	public static Optional<String> getCheckedOutBranch(ICheckoutFactory checkoutFactory, ResourceReference resource) {
		return performOneOffCheckoutOperation(checkoutFactory, resource, ICheckout::getCheckedOutBranch,
				Optional.empty());
	}

	private static <T> T performOneOffCheckoutOperation(ICheckoutFactory factory, ResourceReference resource,
			CheckoutOperation<T> operation, T defaultOnFailure) {
		try (ICheckout checkout = factory.openCheckout(resource).orElse(null)) {
			if (checkout == null) {
				return defaultOnFailure;
			}
			return operation.perform(checkout);
		} catch (IOException e) {
			LOGGER.catching(e);
			// Catches CheckoutException and other possible IOExceptions.
			return defaultOnFailure;
		}
	}

}
