/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.vcs;

import java.net.URI;
import java.util.List;
import java.util.Optional;

import com.teamscale.ide.commons.model.ResourceReference;

/**
 * A local checkout from a version-control system (VCS).
 * <p>
 * It encompasses a file-system subtree that exists on a single branch, but not
 * subtrees from different branches (e.g., different subfolders in SVN's
 * {@code branches} folder will reside in different checkouts). This means that
 * it will be on a single {@link #getCheckedOutBranch() branch}. All files in
 * the subtree are assumed to have undergone the same series of
 * {@link #getMostRecentRevisions(int) revisions}.
 * <p>
 * It is advantageous, but not always possible, to make the file-system subtree
 * as large as possible while still staying on the same branch; this way,
 * multiple IDE projects can be represented by the same checkout and expensive
 * operations (e.g., {@link #getChanges(RevisionDescriptor)}) need only be
 * performed once, even if multiple IDE projects are of interest. The root of
 * this file-system subtree is class {@linkplain #getCheckoutRoot() checkout
 * root}.
 */
public interface ICheckout extends AutoCloseable {

	/**
	 * Gets the root {@linkplain ResourceReference#forFolder(URI) folder} of the
	 * file-system subtree managed by this checkout.
	 */
	ResourceReference getCheckoutRoot();

	/**
	 * If this checkout is on one or more branches, returns the name of the primary
	 * checked-out branch, i.e., the branch the user explicitly checked out.
	 */
	Optional<String> getCheckedOutBranch() throws CheckoutException;

	/**
	 * Gets the (VCS-specific) revision descriptors of the most recent revisions
	 * along the {@link #getCheckedOutBranch() currently checked-out branch}, up to
	 * and including the revision of this checkout. The revision of this checkout is
	 * returned as the first element, followed by its parent, grand-parent, etc. For
	 * an SVN repository where external repositories are attached, the method
	 * returns at most #limit revisions for each repository.
	 */
	List<RevisionDescriptor> getMostRecentRevisions(int limit) throws CheckoutException;

	/**
	 * Gets a list of all changes that have not yet been committed to this checkout.
	 * If changes included in one or more commits are also needed, use
	 * {@link #getChanges(RevisionDescriptor)} instead.
	 */
	ChangeSet getUncommittedChanges() throws CheckoutException;

	/**
	 * Gets a list of all changes that have been made to this checkout since the
	 * given revision. The change set includes all resources that have been added,
	 * changed, or removed in the meantime.
	 */
	ChangeSet getChanges(RevisionDescriptor sinceRevision) throws CheckoutException;

	/**
	 * Checks whether this checkout is equal to another one. Two checkouts may be
	 * equal if, e.g., because they are using the same {@code .git} or {@code .svn}
	 * directory or whether they connect to the same server.
	 */
	@Override
	boolean equals(Object other);

	@Override
	@SuppressWarnings("unused")
	default void close() {
		// Do nothing by default.
	}
}
