/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.vcs.git;

import static java.util.Objects.requireNonNull;

import java.io.IOException;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.eclipse.jgit.api.Git;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.diff.DiffEntry;
import org.eclipse.jgit.diff.DiffEntry.ChangeType;
import org.eclipse.jgit.lib.Constants;
import org.eclipse.jgit.lib.FileMode;
import org.eclipse.jgit.lib.ObjectId;
import org.eclipse.jgit.lib.ObjectReader;
import org.eclipse.jgit.lib.Repository;
import org.eclipse.jgit.revwalk.RevCommit;
import org.eclipse.jgit.revwalk.RevWalk;
import org.eclipse.jgit.storage.file.FileRepositoryBuilder;
import org.eclipse.jgit.storage.file.WindowCacheConfig;
import org.eclipse.jgit.treewalk.AbstractTreeIterator;
import org.eclipse.jgit.treewalk.CanonicalTreeParser;

import com.google.common.base.Preconditions;
import com.teamscale.ide.commons.lang.ToStringHelpers;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.util.UriPathUtils;
import com.teamscale.ide.commons.vcs.ChangeSet;
import com.teamscale.ide.commons.vcs.CheckoutException;
import com.teamscale.ide.commons.vcs.ICheckout;
import com.teamscale.ide.commons.vcs.RevisionDescriptor;

/**
 * A {@link ICheckout} for Git repositories implemented using plain JGit. It
 * does <strong>not</strong> use any configuration or state that a Git plug-in
 * for the IDE (like EGit for Eclipse) may have. Instead, it considers what is
 * in the {@code .git} directory on disk the sole truth about the repository's
 * state or configuration.
 */
public final class JGitCheckout implements ICheckout {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(JGitCheckout.class);

	private final Repository repository;

	private final ResourceReference checkoutRoot;

	@Override
	public ResourceReference getCheckoutRoot() {
		return checkoutRoot;
	}

	/**
	 * Finds and opens a {@link JGitCheckout} at or above the given search root
	 * path.
	 */
	public static Optional<JGitCheckout> open(Path searchRoot) throws CheckoutException {
		LOGGER.traceEntry(null, searchRoot);
		try {
			FileRepositoryBuilder builder = new FileRepositoryBuilder().readEnvironment()
					.findGitDir(searchRoot.toFile());
			if (builder.getGitDir() == null) {
				return LOGGER.traceExit(Optional.empty());
			}
			return LOGGER.traceExit(Optional.of(new JGitCheckout(builder.build())));
		} catch (IOException e) {
			throw LOGGER.throwing(new CheckoutException("Failed to open Git repository", e));
		}
	}

	private JGitCheckout(Repository repository) {
		this.repository = requireNonNull(repository);
		this.checkoutRoot = ResourceReference.forFolder(repository.getWorkTree().toURI());

		configureWindowCache(repository);
	}

	/**
	 * Configure the JGit {@link org.eclipse.jgit.internal.storage.file.WindowCache}
	 * so that it does <em>not</em> publish a JMX bean. This avoids a runtime
	 * dependency on package and module {@link javax.management}, which is not
	 * included in the jlink'ed variants {@code teamscale-dev} (to keep the
	 * installation size down).
	 */
	private static void configureWindowCache(Repository repository) {
		WindowCacheConfig windowCacheConfig = new WindowCacheConfig().fromConfig(repository.getConfig());
		windowCacheConfig.setExposeStatsViaJmx(false);
		windowCacheConfig.install();
	}

	@Override
	public Optional<String> getCheckedOutBranch() throws CheckoutException {
		LOGGER.traceEntry();
		try {
			return LOGGER.traceExit(Optional.ofNullable(repository.getBranch()));
		} catch (IOException e) {
			throw LOGGER.throwing(new CheckoutException("Could not determine primary branch", e));
		}
	}

	@Override
	public List<RevisionDescriptor> getMostRecentRevisions(int limit) throws CheckoutException {
		LOGGER.traceEntry(null, limit);
		try {
			ObjectId head = repository.resolve(Constants.HEAD);
			if (head == null) {
				// No commit yet in repository
				return LOGGER.traceExit(Collections.emptyList());
			}
			return LOGGER.traceExit(getMostRecentRevisions(head, limit));
		} catch (IOException e) {
			throw LOGGER.throwing(new CheckoutException("Could not get most recent revisions", e));
		}
	}

	private List<RevisionDescriptor> getMostRecentRevisions(ObjectId commitId, int limit) throws IOException {
		List<RevisionDescriptor> revisions = new ArrayList<>();

		try (RevWalk revWalk = new RevWalk(repository)) {
			RevCommit commit = revWalk.parseCommit(commitId);
			while (commit != null && revisions.size() < limit) {
				revisions.add(new RevisionDescriptor(ObjectId.toString(commit)));
				if (commit.getParentCount() == 0) {
					break;
				}
				commit = revWalk.parseCommit(commit.getParent(0).getId());
			}
			revWalk.dispose();
		}

		return revisions;
	}

	@Override
	public ChangeSet getUncommittedChanges() throws CheckoutException {
		LOGGER.traceEntry();
		ChangeSet changesFromIndex = getChangesFromIndex();
		ChangeSet changesFromWorkingDirectory = getChangesFromWorkingDirectory();
		ChangeSet changesFromStatus = changesFromIndex.andThen(changesFromWorkingDirectory);
		return LOGGER.traceExit(changesFromStatus);
	}

	@Override
	public ChangeSet getChanges(RevisionDescriptor sinceRevision) throws CheckoutException {
		LOGGER.traceEntry(null, sinceRevision);
		ChangeSet changesFromDiff = getChangesFromDiff(sinceRevision);
		ChangeSet changesFromIndex = getChangesFromIndex();
		ChangeSet changesFromWorkingDirectory = getChangesFromWorkingDirectory();
		ChangeSet changesFromStatus = changesFromIndex.andThen(changesFromWorkingDirectory);
		return LOGGER.traceExit(changesFromDiff.andThen(changesFromStatus));
	}

	/**
	 * Gets the changes between the {@code parentRevision} and the HEAD revision.
	 */
	private ChangeSet getChangesFromDiff(RevisionDescriptor parentRevision) throws CheckoutException {
		LOGGER.traceEntry(null, parentRevision);
		try (Git git = new Git(repository)) {
			List<DiffEntry> diff = git.diff().setShowNameAndStatusOnly(true) //
					.setOldTree(getTreeIterator(parentRevision.getRevision())) //
					.setNewTree(getTreeIterator(Constants.HEAD)).call();
			return LOGGER.traceExit(getChangesFromDiffEntries(diff));
		} catch (GitAPIException | IOException e) {
			throw LOGGER.throwing(new CheckoutException("Could not get changes from diff", e));
		}
	}

	/**
	 * Gets any staged changes, i.e., changes not yet part of the HEAD revision but
	 * already in the index.
	 */
	private ChangeSet getChangesFromIndex() throws CheckoutException {
		LOGGER.traceEntry();
		try (Git git = new Git(repository)) {
			List<DiffEntry> diff = git.diff().setCached(true).setShowNameAndStatusOnly(true).call();
			return LOGGER.traceExit(getChangesFromDiffEntries(diff));
		} catch (GitAPIException e) {
			throw LOGGER.throwing(new CheckoutException("Could not get change from index", e));
		}
	}

	/**
	 * Gets any changes from the working directory, i.e., changes that have not been
	 * staged yet.
	 */
	private ChangeSet getChangesFromWorkingDirectory() throws CheckoutException {
		LOGGER.traceEntry();
		try (Git git = new Git(repository)) {
			List<DiffEntry> diff = git.diff().setShowNameAndStatusOnly(true).call();
			return LOGGER.traceExit(getChangesFromDiffEntries(diff));
		} catch (GitAPIException e) {
			throw LOGGER.throwing(new CheckoutException("Could not get change from working directory", e));
		}
	}

	/**
	 * Taken from {@code org.eclipse.jgit.api.DiffCommandTest#getTreeIterator}.
	 */
	private AbstractTreeIterator getTreeIterator(String name) throws IOException {
		ObjectId id = repository.resolve(name);
		Preconditions.checkArgument(id != null, name);
		CanonicalTreeParser treeIterator = new CanonicalTreeParser();
		try (ObjectReader reader = repository.newObjectReader(); RevWalk revWalk = new RevWalk(repository)) {
			treeIterator.reset(reader, revWalk.parseTree(id));
			return treeIterator;
		}
	}

	private ChangeSet getChangesFromDiffEntries(List<DiffEntry> diff) {
		Set<ResourceReference> newFiles = new HashSet<>();
		Set<ResourceReference> modifiedFiles = new HashSet<>();
		Set<ResourceReference> removedFiles = new HashSet<>();

		for (DiffEntry entry : diff) {
			LOGGER.trace("Considering {}", entry);

			Optional<ChangeType> changeType = determineChangeType(entry);
			if (!changeType.isPresent()) {
				// Change does not affect an actual file, i.e., entry is a symlink or similar.
				continue;
			}
			switch (changeType.get()) {
			case ADD:
				// fall-through
			case COPY:
				newFiles.add(toResourceReference(entry.getNewPath()));
				break;
			case MODIFY:
				modifiedFiles.add(toResourceReference(entry.getNewPath()));
				break;
			case DELETE:
				removedFiles.add(toResourceReference(entry.getOldPath()));
				break;
			case RENAME:
				removedFiles.add(toResourceReference(entry.getOldPath()));
				newFiles.add(toResourceReference(entry.getNewPath()));
				break;
			default:
				throw LOGGER.throwing(new IllegalStateException("Unknown type: " + changeType));
			}
		}

		return new ChangeSet(newFiles, modifiedFiles, removedFiles);
	}

	/**
	 * Determines the change type for a given entry. This method respects the mode
	 * flags for the old and new file (if present) and uses them to override the
	 * stored mode. For example, changing a symlink to a plain file would be treated
	 * as an {@link DiffEntry.ChangeType#ADD}. Returns {@link Optional#empty()} if
	 * the entry was and still is not an actual file and should hence be ignored
	 * entirely.
	 */
	private static Optional<DiffEntry.ChangeType> determineChangeType(DiffEntry diff) {
		boolean wasActualFile = isActualFile(diff.getOldMode());
		boolean isActualFile = isActualFile(diff.getNewMode());

		if (wasActualFile && isActualFile) {
			return Optional.of(diff.getChangeType());
		} else if (!wasActualFile && isActualFile) {
			return Optional.of(DiffEntry.ChangeType.ADD);
		} else if (wasActualFile) {
			return Optional.of(DiffEntry.ChangeType.DELETE);
		}
		return Optional.empty();
	}

	/**
	 * Returns whether a "file" with given file mode is an actual file and not,
	 * e.g., a symlink which is not tracked by Teamscale.
	 */
	private static boolean isActualFile(FileMode mode) {
		int modeBits = mode.getBits();
		return FileMode.REGULAR_FILE.equals(modeBits) || FileMode.EXECUTABLE_FILE.equals(modeBits);
	}

	private ResourceReference toResourceReference(String pathInWorkTree) {
		// JGit always uses slashes as path separator, even on Windows, so using
		// UriPathUtils is safe.
		String uriPathInWorkTree = UriPathUtils.toRelativeUriPath(pathInWorkTree);
		return ResourceReference.forFile(repository.getWorkTree().toURI().resolve(uriPathInWorkTree));
	}

	@Override
	public void close() {
		LOGGER.traceEntry();
		try {
			repository.close();
		} finally {
			LOGGER.traceExit();
		}
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		JGitCheckout that = (JGitCheckout) other;
		return Objects.equals(this.checkoutRoot, that.checkoutRoot);
	}

	@Override
	public int hashCode() {
		return Objects.hash(checkoutRoot);
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
