/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.vcs.services;

import java.util.Optional;

import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.vcs.CheckoutException;
import com.teamscale.ide.commons.vcs.ICheckout;
import com.teamscale.ide.commons.vcs.git.JGitCheckout;
import com.teamscale.ide.commons.vcs.svn.SvnKitCheckout;

/**
 * Generic checkout implementations that are usable by any IDE; they do not make
 * use of any functionality dependent on a particular IDE or its plug-ins.
 */
public enum GenericSupportedCheckouts implements ISupportedCheckout {

	/**
	 * A generic (IDE-independent) implementation of a Git {@link ICheckout} using
	 * plain JGit.
	 */
	GENERIC_GIT() {

		@Override
		public String getDisplayName() {
			return "Git";
		}

		@Override
		public Optional<? extends ICheckout> openCheckout(ResourceReference resource) throws CheckoutException {
			if (!resource.hasPath()) {
				// Cannot obtain a Path; resource reference must use a non-'file' URI scheme.
				return Optional.empty();
			}

			return JGitCheckout.open(resource.toPath());
		}
	},

	/**
	 * A generic (IDE-independent) implementation of a Subversion {@link ICheckout}
	 * using plain SVNKit. This implementation is <strong>limited</strong>; it only
	 * offers those operations needed for Pre-Commit 3.
	 */
	GENERIC_SVN() {

		@Override
		public String getDisplayName() {
			return "Subversion";
		}

		@Override
		public Optional<? extends ICheckout> openCheckout(ResourceReference resource) {
			if (!resource.hasPath()) {
				// Cannot obtain a Path; resource reference must use a non-'file' URI scheme.
				return Optional.empty();
			}

			return SvnKitCheckout.open(resource.toPath());
		}
	};
}
