/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.vcs.services;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;

import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.plugin.IPluginService;
import com.teamscale.ide.commons.vcs.CheckoutException;
import com.teamscale.ide.commons.vcs.ICheckout;

/**
 * Determines the version-control system checkout a resource is contained in.
 */
public interface ICheckoutFactory extends IPluginService {

	/**
	 * @return the version-control systems supported by this service.
	 */
	List<ISupportedCheckout> getSupportedCheckouts();

	/**
	 * Opens the version-control system checkout the given resource is contained in,
	 * if any.
	 * <p>
	 * Clients should <strong>not</strong> make any assumptions whether resources
	 * are contained in the same checkout based on their paths. Instead, clients
	 * should ask this factory for the {@link ICheckout}s for both resources and
	 * then invoke {@link ICheckout#equals(Object)}.
	 *
	 * @return the {@link ICheckout} the given resource is contained in or
	 *         {@link Optional#empty()} if it is not under version control.
	 */
	Optional<ICheckout> openCheckout(ResourceReference resource) throws CheckoutException;

	/**
	 * Opens the version-control system checkouts the given resources are contained
	 * in, if any. This method either returns a set o open checkouts or throws an
	 * exception with no checkouts left unclosed.
	 *
	 * @see #openCheckout(ResourceReference)
	 */
	default Set<ICheckout> openCheckouts(Collection<ResourceReference> resources) throws CheckoutException {
		Set<ICheckout> openCheckouts = new HashSet<>();
		try {
			for (ResourceReference selectedPath : resources) {
				openCheckout(selectedPath).ifPresent(openCheckouts::add);
			}
		} catch (CheckoutException | RuntimeException e) {
			for (ICheckout checkout : openCheckouts) {
				checkout.close();
			}
			throw e;
		}
		return openCheckouts;
	}
}
