/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.vcs.services;

import java.util.Optional;

import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.vcs.CheckoutException;
import com.teamscale.ide.commons.vcs.ICheckout;

/**
 * An interface that allows to {@linkplain #openCheckout(ResourceReference)
 * open} {@link ICheckout}s for a particular, supported version-control system.
 * <p>
 * Clients should usually use
 * {@link ICheckoutFactory#openCheckout(ResourceReference)} instead, to
 * automatically open an {@link ICheckout} for a resource as long as it is
 * managed by <em>any</em> supported version-control system.
 */
public interface ISupportedCheckout {

	/** @return a human-readable name of the version-control system. */
	String getDisplayName();

	/**
	 * Opens an {@link ICheckout} if the given resource is managed by this
	 * version-control system.
	 * <p>
	 * Clients should <strong>not</strong> make any assumptions whether two
	 * resources are contained in the same checkout based on their paths. Instead,
	 * clients should ask this factory for the {@link ICheckout}s for both resources
	 * and then invoke {@link ICheckout#equals(Object)}.
	 * 
	 * @return the {@link ICheckout} the given resource is contained in or
	 *         {@link Optional#empty()} if it is not under version control by this
	 *         version-control system.
	 */
	Optional<? extends ICheckout> openCheckout(ResourceReference resource) throws CheckoutException;
}
