/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.vcs.services.impl;

import java.util.Optional;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import com.google.common.collect.ImmutableList;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.vcs.CheckoutException;
import com.teamscale.ide.commons.vcs.ICheckout;
import com.teamscale.ide.commons.vcs.services.ICheckoutFactory;
import com.teamscale.ide.commons.vcs.services.ISupportedCheckout;

/**
 * A {@link ICheckoutFactory} that tries any number of
 * {@link ISupportedCheckout}s, picking the first that is able to open a
 * {@link ICheckout} for a given resource.
 */
public class CheckoutFactoryImpl implements ICheckoutFactory {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(CheckoutFactoryImpl.class);

	private final ImmutableList<ISupportedCheckout> supportedCheckouts;

	public CheckoutFactoryImpl(ISupportedCheckout... supportedCheckouts) {
		this.supportedCheckouts = ImmutableList.copyOf(supportedCheckouts);
	}

	@Override
	public Optional<ICheckout> openCheckout(ResourceReference resource) throws CheckoutException {
		LOGGER.debug("Searching checkout for {}.", resource);

		for (ISupportedCheckout supportedCheckout : getSupportedCheckouts()) {
			ICheckout checkout = supportedCheckout.openCheckout(resource).orElse(null);
			if (checkout != null) {
				LOGGER.debug("Found {} checkout rooted at {}.", supportedCheckout.getDisplayName(),
						checkout.getCheckoutRoot());
				return Optional.of(checkout);
			}
			LOGGER.debug("Did not find {} checkout.", supportedCheckout.getDisplayName());
		}
		LOGGER.debug("Search found no checkouts.");
		return Optional.empty();
	}

	@Override
	public ImmutableList<ISupportedCheckout> getSupportedCheckouts() {
		return supportedCheckouts;
	}
}
