/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.vcs.svn;

import static java.lang.Math.max;
import static java.util.stream.Collectors.toList;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.stream.LongStream;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import com.teamscale.ide.commons.vcs.CheckoutException;
import com.teamscale.ide.commons.vcs.ICheckout;
import com.teamscale.ide.commons.vcs.RevisionDescriptor;

/**
 * Base class for {@link ICheckout}s for Subversion.
 */
public abstract class SubversionCheckoutBase implements ICheckout {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(SubversionCheckoutBase.class);

	@Override
	public final List<RevisionDescriptor> getMostRecentRevisions(int limit) throws CheckoutException {
		LOGGER.traceEntry(null, limit);
		List<RevisionDescriptor> revisions = new LinkedList<>();
		for (RevisionDescriptor revisionDescriptor : doReturnMostRecentRevisionDescriptor()) {
			revisions.addAll(buildDescendingRevisions(revisionDescriptor, limit));
		}
		return LOGGER.traceExit(revisions);
	}

	/**
	 * Returns a descending list of revision descriptors with revision numbers X and
	 * the given repository identifier, where X is in [<given revision number>,
	 * max(<revision number> - limit, 1)[.
	 */
	private List<RevisionDescriptor> buildDescendingRevisions(RevisionDescriptor revisionDescriptor, int limit) {
		Long revisionIdentifier = Long.parseLong(revisionDescriptor.getRevision());
		return LongStream //
				.rangeClosed(max(1, revisionIdentifier - limit + 1), revisionIdentifier) //
				.boxed().sorted(Collections.reverseOrder()) // reverse order to check most recent revision first
				.map(currentRevisionIdentifier -> new RevisionDescriptor(Long.toString(currentRevisionIdentifier),
						revisionDescriptor.getRepositoryUri())) //
				.collect(toList());
	}

	/**
	 * Returns the most recent revision data (a pair of revision number and
	 * repository identifier) for every repository URL (there are multiple
	 * repository URLs if externals are used) contained <em>anywhere</em> in this
	 * checkout. This may not necessarily be the revision of the checkout's root
	 * directory.
	 */
	protected abstract List<RevisionDescriptor> doReturnMostRecentRevisionDescriptor() throws CheckoutException;
}
