/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.commons.vcs.svn;

import java.net.URI;
import java.util.Optional;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.google.common.base.Preconditions;
import com.google.common.collect.Iterables;
import com.teamscale.ide.commons.util.UriUtils;

/**
 * Utility methods for working with Subversion within different IDEs.
 */
public class SubversionUtils {

	private static final Pattern CHECKOUT_ROOT_PATTERN = Pattern.compile("^([^/]+/)*?(trunk|branches/[^/]+)(?=(/|$))");

	/**
	 * Guesses the branch name from the resource URI and repository root URI.
	 * Handles URIs for repository layouts using a single project or multiple ones.
	 *
	 * @param repositoryRootUri
	 *            The URI to the root of the SVN repository. Must be an absolute
	 *            URI.
	 * @param resourceUri
	 *            The URI to the resource within the SVN repository. Must be an
	 *            absolute URI.
	 *
	 * @see <a href=
	 *      "http://svnbook.red-bean.com/en/1.7/svn.branchmerge.maint.html#svn.branchmerge.maint.layout">http://svnbook.red-bean.com/en/1.7/svn.branchmerge.maint.html#svn.branchmerge.maint.layout</a>
	 */
	public static Optional<String> guessBranchName(URI repositoryRootUri, URI resourceUri) {
		URI checkoutRoot = getBranchRoot(repositoryRootUri, resourceUri).orElse(null);
		if (checkoutRoot == null) {
			return Optional.empty();
		}
		return Optional.of(Iterables.getLast(UriUtils.getPathSegments(checkoutRoot)));
	}

	/**
	 * Determines the <em>branch root</em> of an SVN checkout, given a resource
	 * within the checkout. The branch root is a URI either of the {@code trunk}
	 * directory or of some {@code branches/some-branch} directory, depending on
	 * whether the checked out resource resides on the trunk or some non-trunk
	 * branch.
	 * 
	 * @param repositoryUri
	 *            The URI to the root of the SVN repository. Must be an absolute
	 *            URI.
	 * @param resourceUri
	 *            The URI to the resource within the SVN repository. Must be an
	 *            absolute URI.
	 */
	public static Optional<URI> getBranchRoot(URI repositoryUri, URI resourceUri) {
		Preconditions.checkArgument(repositoryUri.isAbsolute(), "SVN repository root URI is not absolute: <%s>",
				resourceUri);
		Preconditions.checkArgument(resourceUri.isAbsolute(), "SVN repository resource URI is not absolute: <%s>",
				resourceUri);
		URI relativeRepositoryPath = repositoryUri.relativize(resourceUri);
		Preconditions.checkArgument(!relativeRepositoryPath.isAbsolute(),
				"Expected relative URI but got <%s> from resource URI <%s> and expected prefixing root URI <%s>.",
				relativeRepositoryPath, resourceUri, repositoryUri);

		String path = relativeRepositoryPath.getRawPath();
		Matcher trunkMatcher = CHECKOUT_ROOT_PATTERN.matcher(path);
		if (trunkMatcher.find()) {
			return Optional.of(UriUtils.ensurePathHasTrailingSlash(repositoryUri).resolve(trunkMatcher.group()));
		}
		return Optional.empty();
	}
}
