package com.teamscale.ide.commons.vcs.svn;

import java.io.File;
import java.net.URI;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.tmatesoft.svn.core.SVNDepth;
import org.tmatesoft.svn.core.SVNException;
import org.tmatesoft.svn.core.SVNNodeKind;
import org.tmatesoft.svn.core.wc.SVNClientManager;
import org.tmatesoft.svn.core.wc.SVNInfo;
import org.tmatesoft.svn.core.wc.SVNRevision;
import org.tmatesoft.svn.core.wc.SVNStatus;
import org.tmatesoft.svn.core.wc.SVNStatusType;

import com.google.common.base.Preconditions;
import com.teamscale.ide.commons.lang.ToStringHelpers;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.util.UriUtils;
import com.teamscale.ide.commons.vcs.ChangeSet;
import com.teamscale.ide.commons.vcs.CheckoutException;
import com.teamscale.ide.commons.vcs.ICheckout;
import com.teamscale.ide.commons.vcs.RevisionDescriptor;

/**
 * A {@link ICheckout} for Subversion repositories implemented using plain
 * SVNKit. It does <strong>not</strong> use any configuration or state that a
 * Subversion plug-in for the IDE (like Subversive for Eclipse) may have.
 * <p>
 * At the moment, it only supports those operations needed for Pre-Commit 3,
 * i.e., determining the {@linkplain #getCheckedOutBranch() checked-out branch}
 * and any {@linkplain #getUncommittedChanges() uncommitted changes}. Both are
 * purely local operations that does not require contacting the remote
 * Subversion server.
 */
public final class SvnKitCheckout extends SubversionCheckoutBase {

	// Class argument necessary due to
	// <https://youtrack.jetbrains.com/issue/IDEA-220300>.
	private static final Logger LOGGER = LogManager.getLogger(SvnKitCheckout.class);

	private final SVNClientManager clientManager;

	/** The root directory of this checkout (contains a {@code .svn} directory). */
	private final File checkoutRootFile;

	private final ResourceReference checkoutRoot;

	/**
	 * Finds and opens a {@link SvnKitCheckout} when the given path is part of a SVN
	 * repository.
	 */
	public static Optional<SvnKitCheckout> open(Path path) {
		LOGGER.traceEntry(null, path);
		SVNClientManager clientManager = SVNClientManager.newInstance();
		try {
			Optional<SvnKitCheckout> checkout = doOpen(clientManager, path);
			if (!checkout.isPresent()) {
				clientManager.dispose();
			}
			return LOGGER.traceExit(checkout);
		} catch (SVNException e) {
			LOGGER.catching(e);
			clientManager.dispose();
			return LOGGER.traceExit(Optional.empty());
		}
	}

	private static Optional<SvnKitCheckout> doOpen(SVNClientManager clientManager, Path path) throws SVNException {
		// Search upwards until we find the first directory under version control; we
		// cannot use "svn info" otherwise.
		Path versionControlledPath = getVersionControlledPath(clientManager, path).orElse(null);
		if (versionControlledPath == null) {
			return Optional.empty();
		}

		SVNInfo infoForPath = clientManager.getWCClient().doInfo(versionControlledPath.toFile(), SVNRevision.WORKING);
		Path workingCopyRoot = infoForPath.getWorkingCopyRoot().toPath();

		URI pathRemoteUri = URI.create(infoForPath.getURL().toString());
		URI repositoryRootRemoteUri = URI.create(infoForPath.getRepositoryRootURL().toString());

		return determineCheckoutRoot(workingCopyRoot, versionControlledPath, pathRemoteUri, repositoryRootRemoteUri)
				.map(checkoutRoot -> new SvnKitCheckout(clientManager, checkoutRoot));
	}

	private static Optional<Path> getVersionControlledPath(SVNClientManager clientManager, Path path)
			throws SVNException {
		while (path != null) {
			SVNStatus status = clientManager.getStatusClient().doStatus(path.toFile(), false);
			if (status.getKind() != SVNNodeKind.UNKNOWN) {
				return Optional.of(path);
			}
			path = path.getParent();
		}
		return Optional.empty();
	}

	private static Optional<Path> determineCheckoutRoot(Path workingCopyRoot, Path path, URI pathRemoteUri,
			URI repositoryRemoteUri) {
		LOGGER.traceEntry(null, workingCopyRoot, path, pathRemoteUri, repositoryRemoteUri);

		URI branchRootUri = SubversionUtils.getBranchRoot(repositoryRemoteUri, pathRemoteUri).orElse(null);
		if (branchRootUri == null) {
			return LOGGER.traceExit(Optional.empty());
		}

		LOGGER.debug("Branch root URI: <{}>", branchRootUri);

		int directoriesToSkip = UriUtils.getPathSegments(pathRemoteUri).size()
				- UriUtils.getPathSegments(branchRootUri).size();
		Path checkoutRoot = path;
		while (directoriesToSkip > 0) {
			if (checkoutRoot.equals(workingCopyRoot)) {
				break;
			}
			checkoutRoot = checkoutRoot.getParent();
			directoriesToSkip--;
		}

		return LOGGER.traceExit(Optional.of(checkoutRoot));
	}

	private SvnKitCheckout(SVNClientManager clientManager, Path checkoutRootPath) {
		Preconditions.checkArgument(Files.isDirectory(checkoutRootPath));
		this.clientManager = clientManager;
		this.checkoutRootFile = checkoutRootPath.toFile();
		this.checkoutRoot = ResourceReference.forFolder(checkoutRootPath.toUri());
	}

	@Override
	public Optional<String> getCheckedOutBranch() throws CheckoutException {
		LOGGER.traceEntry();
		try {
			SVNInfo info = clientManager.getWCClient().doInfo(checkoutRootFile, SVNRevision.WORKING);
			URI checkoutRootRemoteUri = URI.create(info.getURL().toString());
			URI repositoryRemoteUri = URI.create(info.getRepositoryRootURL().toString());
			Optional<String> checkedOutBranch = SubversionUtils.guessBranchName(repositoryRemoteUri,
					checkoutRootRemoteUri);
			return LOGGER.traceExit(checkedOutBranch);
		} catch (SVNException e) {
			throw LOGGER.throwing(new CheckoutException("Could not determine primary branch", e));
		}
	}

	@Override
	public ChangeSet getUncommittedChanges() throws CheckoutException {
		LOGGER.traceEntry();
		try {
			ChangeSet uncommittedChanges = getChangesFromSvnStatus();
			return LOGGER.traceExit(uncommittedChanges);
		} catch (SVNException e) {
			throw LOGGER.throwing(new CheckoutException("Could not determine uncommitted changes", e));
		}
	}

	private ChangeSet getChangesFromSvnStatus() throws SVNException {
		LOGGER.traceEntry();
		List<SVNStatus> statuses = new ArrayList<>();
		clientManager.getStatusClient().doStatus(checkoutRootFile, SVNRevision.WORKING, SVNDepth.INFINITY, false, false,
				false, false, statuses::add, null);
		return LOGGER.traceExit(getChangesFromStatuses(statuses));
	}

	private ChangeSet getChangesFromStatuses(List<SVNStatus> statuses) {
		Set<ResourceReference> createdFiles = new HashSet<>();
		Set<ResourceReference> modifiedFiles = new HashSet<>();
		Set<ResourceReference> removedFiles = new HashSet<>();

		for (SVNStatus status : statuses) {
			LOGGER.trace("Considering status {}", () -> ToStringHelpers.toReflectiveStringHelper(status).toString());

			if (!affectsFile(status)) {
				continue;
			}

			SVNStatusType nodeStatus = status.getNodeStatus();
			if (nodeStatus.equals(SVNStatusType.STATUS_UNVERSIONED) || nodeStatus.equals(SVNStatusType.STATUS_ADDED)) {
				createdFiles.add(ResourceReference.forFile(status.getFile().toURI()));
			} else if (nodeStatus.equals(SVNStatusType.STATUS_MODIFIED)) {
				modifiedFiles.add(ResourceReference.forFile(status.getFile().toURI()));
			} else if (nodeStatus.equals(SVNStatusType.STATUS_MISSING)
					|| nodeStatus.equals(SVNStatusType.STATUS_DELETED)) {
				// If there is no committed revision, the file was never present in the
				// repository in the first place and thus cannot be removed.
				if (status.getCommittedRevision().isValid()) {
					removedFiles.add(ResourceReference.forFile(status.getFile().toURI()));
				}
			}
		}

		return new ChangeSet(createdFiles, modifiedFiles, removedFiles);
	}

	private static boolean affectsFile(SVNStatus status) {
		if (status.getKind().equals(SVNNodeKind.FILE)) {
			// SVNKit already knows that this is a file. No need to perform further I/O.
			return true;
		}
		return status.getFile().isFile();
	}

	@Override
	public ChangeSet getChanges(RevisionDescriptor sinceRevision) {
		throw new UnsupportedOperationException("Not needed by Pre-Commit 3");
	}

	@Override
	protected List<RevisionDescriptor> doReturnMostRecentRevisionDescriptor() {
		throw new UnsupportedOperationException("Not needed by Pre-Commit 3");
	}

	@Override
	public void close() {
		clientManager.dispose();
	}

	@Override
	public ResourceReference getCheckoutRoot() {
		return checkoutRoot;
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		SvnKitCheckout that = (SvnKitCheckout) other;
		return Objects.equals(this.checkoutRootFile, that.checkoutRootFile);
	}

	@Override
	public int hashCode() {
		return Objects.hash(checkoutRootFile);
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
