package com.teamscale.ide.config;

import static java.util.Objects.requireNonNull;

import java.net.URI;
import java.nio.file.Path;
import java.util.Objects;
import java.util.Optional;

import org.checkerframework.checker.nullness.qual.Nullable;

import com.google.common.base.MoreObjects;
import com.google.common.base.Preconditions;
import com.teamscale.ide.config.path.CodePath;

/**
 * The logical Teamscale configuration obtained by merging one or more
 * {@code .teamscale.toml} configuration files.
 */
public final class TeamscaleConfig {

	/** The URL of the Teamscale server to connect to. */
	private final URI serverUrl;

	/** The ID of the Teamscale Project to use. */
	private final String projectId;

	/**
	 * If set, the branch to use, e.g., overriding the branch as determined from the
	 * version-control system.
	 */
	@Nullable
	private final String projectBranchOverride;

	/**
	 * The code path the {@linkplain #localBasePath local path} maps to within the
	 * Teamscale project.
	 */
	private final CodePath projectPath;

	/**
	 * The local path with respect to which the {@linkplain #projectPath project
	 * path} will be interpreted.
	 */
	private final Path localBasePath;

	public TeamscaleConfig(URI serverUrl, String projectId, @Nullable String projectBranchOverride,
			CodePath projectPath, Path localBasePath) {
		Preconditions.checkArgument(serverUrl.isAbsolute(), "Server URL must be absolute: {}", serverUrl);
		Preconditions.checkArgument(localBasePath.isAbsolute(), "Local base path must be absolute: {}", localBasePath);

		this.serverUrl = requireNonNull(serverUrl).normalize();
		this.projectId = requireNonNull(projectId);
		this.projectBranchOverride = projectBranchOverride;
		this.projectPath = requireNonNull(projectPath);
		this.localBasePath = requireNonNull(localBasePath);
	}

	/** @see #serverUrl */
	public URI getServerUrl() {
		return serverUrl;
	}

	/** @see #projectId */
	public String getProjectId() {
		return projectId;
	}

	/** @see #projectBranchOverride */
	public Optional<String> getProjectBranchOverride() {
		return Optional.ofNullable(projectBranchOverride);
	}

	/** @see #projectPath */
	public CodePath getProjectPath() {
		return projectPath;
	}

	/** @see #localBasePath */
	public Path getLocalBasePath() {
		return localBasePath;
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		TeamscaleConfig that = (TeamscaleConfig) other;
		return Objects.equals(serverUrl, that.serverUrl) && Objects.equals(projectId, that.projectId)
				&& Objects.equals(projectBranchOverride, that.projectBranchOverride)
				&& Objects.equals(projectPath, that.projectPath) && Objects.equals(localBasePath, that.localBasePath);
	}

	@Override
	public int hashCode() {
		return Objects.hash(serverUrl, projectId, projectBranchOverride, projectPath, localBasePath);
	}

	@Override
	public String toString() {
		return MoreObjects.toStringHelper(this).omitNullValues() //
				.add("serverUrl", serverUrl) //
				.add("projectId", projectId) //
				.add("projectBranchOverride", projectBranchOverride) //
				.add("projectPath", projectPath) //
				.add("localBasePath", localBasePath).toString();
	}
}
