package com.teamscale.ide.config;

import java.net.URI;
import java.nio.file.Path;
import java.util.Optional;

import org.checkerframework.checker.nullness.qual.Nullable;

import com.google.common.base.Preconditions;
import com.teamscale.ide.config.path.CodePath;
import com.teamscale.ide.config.toml.TeamscaleTomlV1;

/**
 * Helper class to build an immutable {@link TeamscaleConfig}.
 */
/* package */ class TeamscaleConfigBuilder {

	/** The recently merged Teamscale .toml configuration file. **/
	private Path recentTomlConfigurationFile;

	/** The server URL, if set so far. */
	@Nullable
	private URI serverUrl;

	/** The project ID, if set so far. */
	@Nullable
	private String projectId;

	/** The project branch, if set so far. */
	@Nullable
	private String projectBranch;

	/** The project path, if set so far. */
	@Nullable
	private CodePath projectPath;

	/** The local base-path corresponding to the project subpath. */
	@Nullable
	private Path localBasePath;

	/**
	 * Whether there have been any {@link #merge(Path, TeamscaleTomlV1, Path)
	 * merges} performed so far.
	 */
	private boolean anyMergesPerformed;

	/* package */ TeamscaleConfigBuilder merge(Path directory, TeamscaleTomlV1 teamscaleToml, Path teamscaleTomlFile) {
		anyMergesPerformed = true;

		recentTomlConfigurationFile = teamscaleTomlFile;

		TeamscaleTomlV1.Server server = teamscaleToml.getServer();
		TeamscaleTomlV1.Project project = teamscaleToml.getProject();

		if (serverUrl == null && server.getUrl().isPresent()) {
			serverUrl = server.getUrl().get();
		}

		if (projectId == null && project.getId().isPresent()) {
			projectId = project.getId().get();
		}

		if (projectBranch == null && project.getBranch().isPresent()) {
			projectBranch = project.getBranch().get();
		}

		if (projectPath == null && project.getPath().isPresent()) {
			projectPath = project.getPath().get();
			localBasePath = directory;
		}

		// If project.path is not set explicitly, the presence of project.id implicitly
		// sets it to the empty path.
		if (projectPath == null && project.getId().isPresent()) {
			projectPath = CodePath.of();
			localBasePath = directory;
		}

		return this;
	}

	/**
	 * Attempts to build a {@link TeamscaleConfig} <strong>if</strong> there have
	 * been any merges performed so far.
	 * 
	 * @throws InvalidConfigException
	 *             in case a {@code .teamscale.toml} configuration is invalid.
	 */
	/* package */ Optional<TeamscaleConfig> build() throws InvalidConfigException {
		if (!anyMergesPerformed) {
			return Optional.empty();
		}

		try {
			Preconditions.checkState(serverUrl != null, "server.url is missing");
			Preconditions.checkState(projectId != null, "project.id is missing");
			Preconditions.checkState(projectPath != null, "project.path is missing");
			Preconditions.checkState(localBasePath != null, "Local base path is missing");
		} catch (IllegalStateException e) {
			throw new InvalidConfigException(recentTomlConfigurationFile,
					"Not all required properties have been set: " + e.getMessage(), e);
		}

		TeamscaleConfig teamscaleConfig = new TeamscaleConfig(serverUrl, projectId, projectBranch, projectPath,
				localBasePath);
		return Optional.of(teamscaleConfig);
	}
}
