package com.teamscale.ide.config;

import java.io.IOException;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Optional;
import java.util.stream.Collectors;

import org.checkerframework.checker.nullness.qual.Nullable;

import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.exc.InvalidFormatException;
import com.teamscale.ide.config.toml.TeamscaleTomlV1;

/**
 * Implements the lookup strategy described in
 * {@code architecture-decisions/0016-ide-config-format.md}
 */
public class TeamscaleConfigReader {

	/** The name of the configuration files in the hierarchy */
	public static final String TEAMSCALE_TOML_FILE_NAME = ".teamscale.toml";

	/**
	 * Attempts to create a logical {@link TeamscaleConfig} by lookup up one or more
	 * {@code .teamscale.toml} files, starting from the given {@code searchStart}
	 * directory upwards until a {@code root=true} configuration property is found
	 * or the root directory is reached.
	 *
	 * @throws InvalidConfigException
	 *             in case a {@code .teamscale.toml} is invalid.
	 * @throws IOException
	 *             in case a {@code .teamscale.toml} file cannot be read or parsed.
	 */
	public Optional<TeamscaleConfig> read(Path searchStart) throws InvalidConfigException, IOException {
		return read(searchStart, null);
	}

	/**
	 * Attempts to create a logical {@link TeamscaleConfig} by lookup up one or more
	 * {@code .teamscale.toml} files, starting from the given {@code searchStart}
	 * directory upwards until a {@code root=true} configuration property is found
	 * or the given {@code searchEnd} directory (if any) is reached.
	 *
	 * @throws InvalidConfigException
	 *             in case a {@code .teamscale.toml} file is invalid.
	 * @throws IOException
	 *             in case a {@code .teamscale.toml} file cannot be read or parsed.
	 */
	public Optional<TeamscaleConfig> read(Path searchStart, @Nullable Path searchEnd)
			throws InvalidConfigException, IOException {
		if (!Files.isDirectory(searchStart)) {
			throw new IllegalArgumentException("Search must start at a directory");
		}
		if (searchEnd != null && !Files.isDirectory(searchStart)) {
			throw new IllegalArgumentException("Search must end at a directory");
		}

		TeamscaleConfigBuilder builder = new TeamscaleConfigBuilder();

		for (Path searchDirectory = searchStart; searchDirectory != null
				&& !searchDirectory.equals(searchEnd); searchDirectory = searchDirectory.getParent()) {

			Path teamscaleTomlFile = searchDirectory.resolve(TEAMSCALE_TOML_FILE_NAME);
			if (Files.exists(teamscaleTomlFile)) {
				TeamscaleTomlV1 teamscaleToml = parse(teamscaleTomlFile);

				builder.merge(searchDirectory, teamscaleToml, teamscaleTomlFile);

				if (teamscaleToml.isRoot()) {
					break;
				}
			}
		}

		return builder.build();
	}

	private static TeamscaleTomlV1 parse(Path teamscaleTomlFile) throws InvalidConfigException, IOException {
		return doParse(teamscaleTomlFile);
	}

	private static TeamscaleTomlV1 doParse(Path teamscaleTomlFile) throws InvalidConfigException, IOException {
		try {
			TeamscaleTomlV1 parsedTeamscaleToml = TeamscaleConfigReader.getParsedToml(teamscaleTomlFile);

			Optional<URI> url = parsedTeamscaleToml.getServer().getUrl();
			if (url.isPresent() && !url.get().isAbsolute()) {
				throw new InvalidConfigException(teamscaleTomlFile,
						String.format("URL must be absolute: %s", url.get()));
			}

			String version = parsedTeamscaleToml.getVersion();
			if (!version.startsWith("1.")) {
				throw new InvalidConfigException(teamscaleTomlFile, String.format("Unexpected version: %s", version));
			}
			return parsedTeamscaleToml;
		} catch (IOException e) {
			throw new IOException(String.format("Failed to parse '%s': %s", teamscaleTomlFile, e.getMessage()), e);
		}
	}

	private static TeamscaleTomlV1 getParsedToml(Path teamscaleTomlFile) throws InvalidConfigException, IOException {
		String teamscaleToml = new String(Files.readAllBytes(teamscaleTomlFile), StandardCharsets.UTF_8);
		try {
			return TeamscaleTomlV1.parse(teamscaleToml);
		} catch (InvalidFormatException e) {
			String fieldName = e.getPath().stream().map(JsonMappingException.Reference::getFieldName)
					.collect(Collectors.joining("."));
			throw new InvalidConfigException(teamscaleTomlFile,
					String.format("Failure to parse TOML: %s => %s", fieldName, e.getValue()), e);
		}
	}
}
