package com.teamscale.ide.config.path;

import static java.util.stream.Collectors.joining;

import java.text.ParseException;
import java.util.Arrays;
import java.util.stream.Stream;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonValue;
import com.google.common.base.Preconditions;

/**
 * An (absolute) uniform path on a Teamscale server of "code" type.
 *
 * This is deliberately more restrictive than what is supported on the server
 * side, as other types (e.g., "architecture", "test-execution", or "non-code")
 * are not needed when mapping local to remote paths.
 */
public final class CodePath {

	/** The unescaped segments this code path consists of. */
	private final String[] unescapedSegments;

	/**
	 * The argument has been checked for validity and has to be an array owned by
	 * this object, i.e., it must not be shared.
	 *
	 * Package-visible as an optimization, to avoid the check and copy of
	 * {@link CodePath#of(String...)}.
	 */
	/* package */ CodePath(String[] unescapedSegments) {
		this.unescapedSegments = unescapedSegments;
	}

	/**
	 * Creates a new code path made up of the given non-empty, non-relative
	 * segments. Slashes in segments must not be escaped.
	 */
	public static CodePath of(String... unescapedSegments) {
		CodePathSegments.check(unescapedSegments);

		return new CodePath(Arrays.copyOf(unescapedSegments, unescapedSegments.length));
	}

	/**
	 * Parses a code path with optional leading slash. Slashes <em>within</em>
	 * segments must be escaped with backslashes.
	 */
	@JsonCreator
	public static CodePath parse(String codePath) throws ParseException {
		// Drop optional leading slash to simplify processing.
		int startOffset = codePath.startsWith("/") ? 1 : 0;
		return new CodePath(CodePathSegments.parseSegments(codePath, startOffset));
	}

	/** Gets this code path's segments in their unescaped form. */
	public Stream<String> segments() {
		return Arrays.stream(unescapedSegments);
	}

	/** Determines whether this code path has the given code path as prefix. */
	public boolean hasPrefix(CodePath potentialPrefix) {
		if (potentialPrefix.unescapedSegments.length > unescapedSegments.length) {
			return false;
		}
		for (int index = 0; index < potentialPrefix.unescapedSegments.length; index++) {
			if (!potentialPrefix.unescapedSegments[index].equals(unescapedSegments[index])) {
				return false;
			}
		}

		return true;
	}

	/**
	 * Removes the given code path prefix, producing a code path fragment.
	 *
	 * @throws IllegalArgumentException
	 *             if the given code path is not a prefix of this code path
	 */
	public CodePathFragment removePrefix(CodePath prefix) {
		Preconditions.checkArgument(hasPrefix(prefix), "Not a prefix of '%s': %s", this, prefix);

		return new CodePathFragment(
				Arrays.copyOfRange(unescapedSegments, prefix.unescapedSegments.length, unescapedSegments.length));
	}

	/**
	 * Adds the given code path fragment suffix, producing a new code path.
	 */
	public CodePath addSuffix(CodePathFragment suffix) {
		return new CodePath(Stream.concat(segments(), suffix.segments()).toArray(String[]::new));
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		CodePath that = (CodePath) other;
		return Arrays.equals(unescapedSegments, that.unescapedSegments);
	}

	@Override
	public int hashCode() {
		return Arrays.hashCode(unescapedSegments);
	}

	/**
	 * Renders this (absolute) code path in its canonical form, i.e., properly
	 * escaped and with a leading slash.
	 */
	@JsonValue
	@Override
	public String toString() {
		return segments().map(CodePathSegments::escape).collect(joining("/", "/", ""));
	}
}
