package com.teamscale.ide.config.path;

import static java.util.stream.Collectors.joining;

import java.text.ParseException;
import java.util.Arrays;
import java.util.stream.Stream;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonValue;

/**
 * A (relative) uniform path on a Teamscale server.
 */
public final class CodePathFragment {

	/** The unescaped segments this code path fragment consists of. */
	private final String[] unescapedSegments;

	/**
	 * The argument has been checked for validity and has to be an array owned by
	 * this object, i.e., it must not be shared.
	 *
	 * Package-visible as an optimization, to avoid the check and copy of
	 * {@link CodePathFragment#of(String...)}.
	 */
	/* package */ CodePathFragment(String[] unescapedSegments) {
		this.unescapedSegments = unescapedSegments;
	}

	/**
	 * Creates a new code path fragment made up of the given non-empty, non-relative
	 * segments. Slashes in segments must not be escaped.
	 */
	public static CodePathFragment of(String... unescapedSegments) {
		CodePathSegments.check(unescapedSegments);

		return new CodePathFragment(Arrays.copyOf(unescapedSegments, unescapedSegments.length));
	}

	/**
	 * Parses a code path fragment. Slashes <em>within</em> segments must be escaped
	 * with backslashes.
	 */
	@JsonCreator
	public static CodePathFragment parse(String codePathFragment) throws ParseException {
		if (codePathFragment.startsWith("/")) {
			throw new ParseException("Absolute code path not expected: " + codePathFragment, 0);
		}
		return new CodePathFragment(CodePathSegments.parseSegments(codePathFragment, 0));
	}

	/** Gets this code path fragment's segments in their unescaped form. */
	public Stream<String> segments() {
		return Arrays.stream(unescapedSegments);
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		CodePathFragment that = (CodePathFragment) other;
		return Arrays.equals(unescapedSegments, that.unescapedSegments);
	}

	@Override
	public int hashCode() {
		return Arrays.hashCode(unescapedSegments);
	}

	/**
	 * Renders this code path fragment in its canonical form, i.e., properly
	 * escaped.
	 */
	@JsonValue
	@Override
	public String toString() {
		return segments().map(CodePathSegments::escape).collect(joining("/"));
	}
}
