package com.teamscale.ide.config.path;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Utilities for working with code path segments, shared between
 * {@link CodePath} and {@link CodePathFragment}.
 */
/* package */ class CodePathSegments {

	private static final String EMPTY_SEGMENT_ERROR = "Empty segments are not allowed";
	private static final String RELATIVE_SEGMENT_ERROR = "Relative segments are not allowed";

	/** Pattern that matches a non-escaped slash in a uniform path. */
	private static final Pattern UNESCAPED_SLASH = Pattern.compile("(?<!\\\\)/");

	/** Escapes any slashes in an (unescaped) code path segment. */
	/* package */ static String escape(String unescapedSegment) {
		return unescapedSegment.replace("/", "\\/");
	}

	/** Unescapes any escaped slashes in an (escaped) code path segment. */
	/* package */ static String unescape(CharSequence escapedSegment) {
		return escapedSegment.toString().replace("\\/", "/");
	}

	/**
	 * Parses a code path or fragment starting at the given offset into its
	 * (unescaped) segments.
	 */
	/* package */ static String[] parseSegments(CharSequence codePath, int startOffset) throws ParseException {
		List<String> segments = new ArrayList<>();

		// Inner segments, delimited by a trailing slash
		Matcher matcher = CodePathSegments.UNESCAPED_SLASH.matcher(codePath);
		int segmentStart = startOffset;
		while (matcher.find(segmentStart)) {
			segments.add(parseSegment(codePath, segmentStart, matcher.start()));
			segmentStart = matcher.end();
		}

		// Final segment
		if (segmentStart < codePath.length()) {
			segments.add(parseSegment(codePath, segmentStart, codePath.length()));
		}

		return segments.toArray(new String[0]);
	}

	/**
	 * Parses a single escaped segment, delimited by the given start and end
	 * offsets, within a code path or fragment into its unescaped form.
	 */
	private static String parseSegment(CharSequence codePath, int startInclusive, int endExclusive)
			throws ParseException {
		if (startInclusive == endExclusive) {
			throw new ParseException(EMPTY_SEGMENT_ERROR + ": " + codePath, startInclusive);
		}
		CharSequence escapedSegment = codePath.subSequence(startInclusive, endExclusive);
		if (CodePathSegments.isRelative(escapedSegment)) {
			throw new ParseException(RELATIVE_SEGMENT_ERROR + ": " + codePath, startInclusive);
		}
		return CodePathSegments.unescape(escapedSegment);
	}

	/**
	 * Checks whether the given unescaped segments conform to the expected internal
	 * representation of {@link CodePath} or {@link CodePathFragment}. Throws an
	 * (unchecked) {@link IllegalArgumentException} otherwise.
	 */
	/* package */ static void check(CharSequence[] unescapedSegments) {
		for (CharSequence segment : unescapedSegments) {
			checkSegment(segment);
		}
	}

	private static void checkSegment(CharSequence unescapedSegment) {
		if (unescapedSegment.length() == 0) {
			throw new IllegalArgumentException(EMPTY_SEGMENT_ERROR);
		}
		if (isRelative(unescapedSegment)) {
			throw new IllegalArgumentException(RELATIVE_SEGMENT_ERROR);
		}
	}

	private static boolean isRelative(CharSequence segment) {
		return segment.equals(".") || segment.equals("..");
	}
}
