package com.teamscale.ide.config.toml;

import java.io.IOException;
import java.net.URI;
import java.text.ParseException;
import java.util.Objects;
import java.util.Optional;

import org.checkerframework.checker.nullness.qual.Nullable;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.dataformat.toml.TomlMapper;
import com.google.common.base.MoreObjects;
import com.teamscale.ide.config.path.CodePath;

/**
 * Representation of a single {@code .teamscale.toml} configuration file in
 * version 1.0. See {@code architecture-decisions/0016-ide-config-format.md} for
 * a description of the format.
 */
public final class TeamscaleTomlV1 {

	/**
	 * The config-file format version.
	 */
	private final String version;

	/**
	 * Whether this file is the "root" of the config hierarchy or the parser should
	 * continue searching upwards.
	 */
	private final boolean root;

	/** @see Server */
	private final Server server;

	/** @see Project */
	private final Project project;

	public TeamscaleTomlV1(@JsonProperty("version") String version, @JsonProperty("root") boolean root,
			@JsonProperty("server") Server server, @JsonProperty("project") Project project) {
		this.version = version != null ? version : "1.0"; // If no version is specified, "1.0" is the default.
		this.root = root;
		this.server = server != null ? server : Server.emptySection();
		this.project = project != null ? project : Project.emptySection();
	}

	public String getVersion() {
		return version;
	}

	public boolean isRoot() {
		return root;
	}

	public Server getServer() {
		return server;
	}

	public Project getProject() {
		return project;
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		TeamscaleTomlV1 that = (TeamscaleTomlV1) other;
		return Objects.equals(version, that.version) && root == that.root && Objects.equals(server, that.server)
				&& Objects.equals(project, that.project);
	}

	@Override
	public int hashCode() {
		return Objects.hash(version, root, server, project);
	}

	@Override
	public String toString() {
		return MoreObjects.toStringHelper(this).omitNullValues() //
				.add("version", version) //
				.add("root", root) //
				.add("server", server) //
				.add("project", project).toString();
	}

	/**
	 * Configuration properties for Teamscale server that we can connect to.
	 */
	public static final class Server {

		@Nullable
		private final URI url;

		@JsonCreator
		public Server(@JsonProperty("url") @Nullable URI url) {
			this.url = url;
		}

		private static Server emptySection() {
			return new Server(null);
		}

		public Optional<URI> getUrl() {
			return Optional.ofNullable(url);
		}

		@Override
		public boolean equals(Object other) {
			if (this == other) {
				return true;
			}
			if (other == null) {
				return false;
			}
			if (getClass() != other.getClass()) {
				return false;
			}
			Server that = (Server) other;
			return Objects.equals(url, that.url);
		}

		@Override
		public int hashCode() {
			return Objects.hash(url);
		}

		@Override
		public String toString() {
			return MoreObjects.toStringHelper(this).omitNullValues() //
					.add("url", url).toString();
		}
	}

	/**
	 * Configuration properties for the Teamscale project on the server.
	 */
	public static final class Project {

		@Nullable
		private final String id;

		@Nullable
		private final String branch;

		@Nullable
		private final CodePath path;

		public Project(@Nullable String id, @Nullable String branch, @Nullable CodePath path) {
			this.id = id;
			this.branch = branch;
			this.path = path;
		}

		@JsonCreator
		private Project(@JsonProperty("id") @Nullable String id, @JsonProperty("branch") @Nullable String branch,
				@JsonProperty("path") @Nullable String path) throws ParseException {
			this.id = id;
			this.branch = branch;
			this.path = path == null ? null : CodePath.parse(path);
		}

		private static Project emptySection() {
			return new Project(null, null, (CodePath) null);
		}

		public Optional<String> getId() {
			return Optional.ofNullable(id);
		}

		public Optional<String> getBranch() {
			return Optional.ofNullable(branch);
		}

		public Optional<CodePath> getPath() {
			return Optional.ofNullable(path);
		}

		@Override
		public boolean equals(Object other) {
			if (this == other) {
				return true;
			}
			if (other == null) {
				return false;
			}
			if (getClass() != other.getClass()) {
				return false;
			}
			Project that = (Project) other;
			return Objects.equals(id, that.id) && Objects.equals(branch, that.branch)
					&& Objects.equals(path, that.path);
		}

		@Override
		public int hashCode() {
			return Objects.hash(id, branch, path);
		}

		@Override
		public String toString() {
			return MoreObjects.toStringHelper(this).omitNullValues() //
					.add("id", id) //
					.add("branch", branch) //
					.add("path", path).toString();
		}
	}

	/** Parse a string in the {@code .teamscale.toml} format. */
	public static TeamscaleTomlV1 parse(String toml) throws IOException {
		TomlMapper mapper = TomlMapper.builder() //
				.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false).build();
		return mapper.readValue(toml, TeamscaleTomlV1.class);
	}
}
