/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.addon;

import static java.util.Arrays.asList;

import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.List;

import javax.annotation.PostConstruct;
import javax.inject.Inject;

import org.eclipse.e4.core.di.annotations.Optional;
import org.eclipse.e4.core.di.extensions.EventTopic;
import org.eclipse.e4.ui.model.application.MApplication;
import org.eclipse.e4.ui.model.application.ui.MUIElement;
import org.eclipse.e4.ui.model.application.ui.basic.MPart;
import org.eclipse.e4.ui.model.application.ui.menu.MItem;
import org.eclipse.e4.ui.workbench.UIEvents;
import org.eclipse.e4.ui.workbench.modeling.EModelService;
import org.osgi.service.event.Event;

/**
 * The basis for an add-on which synchronizes the {@linkplain MItem#isSelected()
 * selection state} of {@link MItem}s tagged with one of the tags passed to
 * {@link ItemSelectionSynchronizationAddonBase#ItemSelectionSynchronizationAddonBase(String...)}
 * with some external source.
 * <p>
 * Subclasses need to implement {@link #getSelection(String)} retrieve the
 * desired selection state for a given tag.
 */
@SuppressWarnings("restriction")
public abstract class ItemSelectionSynchronizationAddonBase implements PropertyChangeListener {

	private EModelService modelService;

	private MApplication application;

	private final List<String> tags;

	public ItemSelectionSynchronizationAddonBase(String... tags) {
		this.tags = new ArrayList<>(asList(tags));
	}

	/**
	 * Injects all objects necessary to work with the E4 Application Model. Not done
	 * in the constructor simply to keep subclasses unburdened by the knowledge
	 * about the exact objects needed.
	 */
	@PostConstruct
	public void injectUiModel(EModelService modelService, MApplication application) {
		this.modelService = modelService;
		this.application = application;
	}

	/**
	 * Synchronizes the selection state of all {@link MItem}s found in the
	 * Application Model when startup is complete. This does <strong>not</strong>
	 * include items that exist only in part descriptors, but no concrete parts yet.
	 * These items will be synchronized when the part gets created and
	 * {@link #partActivated(Event)}.
	 */
	@Inject
	@Optional
	public void applicationStartupComplete(
			@EventTopic(UIEvents.UILifeCycle.APP_STARTUP_COMPLETE) @SuppressWarnings("unused") Event event) {
		synchronizeSelections(application, tags,
				EModelService.ANYWHERE | EModelService.IN_MAIN_MENU | EModelService.IN_PART);
	}

	/**
	 * Synchronizes the selection state of all {@link MItem}s found in the
	 * Application Model for the part that was just activated.
	 */
	@Inject
	@Optional
	public void partActivated(@EventTopic(UIEvents.UILifeCycle.ACTIVATE) Event event) {
		MPart part = (MPart) event.getProperty(UIEvents.EventTags.ELEMENT);
		synchronizeSelections(part, tags, EModelService.IN_PART);
	}

	/**
	 * Synchronizes the selection state of all {@link MItem}s with the given tags.
	 * Should be called by the subclass when the value changes in the external
	 * source.
	 */
	protected void synchronizeSelections(List<String> tags) {
		synchronizeSelections(application, tags,
				EModelService.ANYWHERE | EModelService.IN_MAIN_MENU | EModelService.IN_PART);
	}

	private void synchronizeSelections(MUIElement searchScope, List<String> tags, int searchFlags) {
		List<MItem> items = modelService.findElements(searchScope, null, MItem.class, tags, searchFlags);
		for (MItem item : items) {
			for (String tag : tags) {
				if (item.getTags().contains(tag)) {
					item.setSelected(getSelection(tag));
				}
			}
		}
	}

	/**
	 * Gets the current selection state associated with the given tag.
	 */
	protected abstract boolean getSelection(String tag);
}
