/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.addon;

import static java.util.Collections.singletonList;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.inject.Inject;

import com.teamscale.ide.eclipse.IUiPreferences;

/**
 * An add-on that synchronizes certain the (boolean) state of certain UI
 * preferences with items in the UI (e.g., checked menu items, toggle buttons).
 */
public class UiPreferencesSynchronizationAddon extends ItemSelectionSynchronizationAddonBase
		implements PropertyChangeListener {

	/**
	 * The tag associated with the "Enable Automatic Pre-Commit Analysis" UI preference. All
	 * items in the Application Model whose selection state should be synchronized
	 * with this preference must be tagged with
	 * {@value UiPreferencesSynchronizationAddon#AUTOMATIC_PRE_COMMIT_ANALYSIS_TAG}.
	 */
	private static final String AUTOMATIC_PRE_COMMIT_ANALYSIS_TAG = "com.teamscale.ide.eclipse.tags.automaticPreCommitAnalysis";

	private final IUiPreferences uiPreferences;

	@Inject
	public UiPreferencesSynchronizationAddon(IUiPreferences uiPreferences) {
		super(AUTOMATIC_PRE_COMMIT_ANALYSIS_TAG);

		this.uiPreferences = uiPreferences;
	}

	/**
	 * Registers this add-on as a {@link PropertyChangeListener} with the
	 * {@link IUiPreferences} bean.
	 */
	@PostConstruct
	public void register() {
		uiPreferences.addPropertyChangeListener(this);
	}

	/**
	 * Unregisters this add-on as a {@link PropertyChangeListener} with the
	 * {@link IUiPreferences} bean.
	 */
	@PreDestroy
	public void unregister() {
		uiPreferences.removePropertyChangeListener(this);
	}

	@Override
	public void propertyChange(PropertyChangeEvent event) {
		switch (event.getPropertyName()) {
		case IUiPreferences.AUTOMATIC_PRE_COMMIT_ANALYSIS_ENABLED:
			synchronizeSelections(singletonList(AUTOMATIC_PRE_COMMIT_ANALYSIS_TAG));
			break;
		default:
			// Property without a registered tag. Do nothing.
			break;
		}
	}

	@Override
	protected boolean getSelection(String tag) {
		switch (tag) {
		case AUTOMATIC_PRE_COMMIT_ANALYSIS_TAG:
			return uiPreferences.isAutomaticPreCommitAnalysisEnabled();
		default:
			throw new IllegalArgumentException("Unexpected tag: " + tag);
		}
	}
}
