package com.teamscale.ide.eclipse.findings;

import java.util.HashSet;
import java.util.Objects;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceDeltaVisitor;

import com.google.common.collect.ImmutableSet;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.eclipse.model.ResourceReferences;

/**
 * Used to traverse a {@link IResourceDelta} and collect all additions, changes
 * and deletions.
 */
class ChangeDeltaVisitor implements IResourceDeltaVisitor {

	private static final Logger LOGGER = LogManager.getLogger();

	private Set<IResource> addedOrChanged;
	private Set<IResource> removed;

	public ChangeDeltaVisitor() {
		addedOrChanged = new HashSet<>();
		removed = new HashSet<>();
	}

	/** {@inheritDoc} */
	@Override
	public boolean visit(IResourceDelta delta) {
		IResource resource = delta.getResource();
		LOGGER.traceEntry(null, resource);
		if (!(resource instanceof IFile)) {
			LOGGER.debug("Resource {} is not a file", resource);
			return LOGGER.traceExit(true);
		}
		if ((delta.getFlags() & IResourceDelta.CONTENT) == 0 && (delta.getFlags() & IResourceDelta.REPLACED) == 0) {
			LOGGER.debug("Ignoring changed resource {}: flags {} are not relevant", resource, delta.getFlags());
			return LOGGER.traceExit(false);
		}
		if (delta.getKind() == IResourceDelta.ADDED || delta.getKind() == IResourceDelta.CHANGED) {
			LOGGER.debug("Found added/changed resource {}", resource);
			addedOrChanged.add(resource);
		} else if (delta.getKind() == IResourceDelta.REMOVED) {
			LOGGER.debug("Found removed resource {}", resource);
			removed.add(resource);
		} else {
			LOGGER.debug("Resource {} has some other change: {}", resource, delta.getKind());
		}
		return LOGGER.traceExit(true);
	}

	/**
	 * Returns all added or changed files which were visited by this visitor.
	 */
	public Set<ResourceReference> getAddedOrChangedFiles() {
		return addedOrChanged.stream().map(ResourceReferences::fromIResource).map(opt -> opt.orElse(null))
				.filter(Objects::nonNull).collect(ImmutableSet.toImmutableSet());
	}

	/**
	 * Returns all removed files which were visited by this visitor.
	 */
	public Set<ResourceReference> getRemovedFiles() {
		return removed.stream().map(ResourceReferences::fromIResource).map(opt -> opt.orElse(null))
				.filter(Objects::nonNull).collect(ImmutableSet.toImmutableSet());
	}
}
