package com.teamscale.ide.eclipse.findings;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.conqat.ide.commons.ui.editor.EditorUtils;
import org.conqat.ide.commons.ui.logging.LoggingUtils;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;

import com.google.common.collect.Sets;
import com.teamscale.ide.TeamscaleIDEActivator;
import com.teamscale.ide.commons.findings.tasks.FetchFindingsTask;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.ITaskExecutor;
import com.teamscale.ide.eclipse.IUiPreferences;
import com.teamscale.ide.eclipse.model.ResourceReferences;
import com.teamscale.ide.eclipse.util.ui.WorkbenchUtils;

/**
 * Listens to file changes and additions.
 */
public class FileSaveListener implements IResourceChangeListener {

	private static final Logger LOGGER = LogManager.getLogger();

	private final ITaskExecutor taskExecutor;
	private final IServiceRegistry serviceRegistry;

	public FileSaveListener(IServiceRegistry serviceRegistry) {
		this.serviceRegistry = serviceRegistry;
		taskExecutor = serviceRegistry.getService(ITaskExecutor.class);
	}

	@Override
	public void resourceChanged(IResourceChangeEvent event) {
		ChangeDeltaVisitor visitor = new ChangeDeltaVisitor();
		try {
			IResourceDelta delta = event.getDelta();
			if (delta != null) {
				delta.accept(visitor);
			}
		} catch (CoreException e) {
			LoggingUtils.error(TeamscaleIDEActivator.getDefault(), e);
		}

		Set<ResourceReference> changedFiles = visitor.getAddedOrChangedFiles();
		if (changedFiles.isEmpty()) {
			return;
		}
		LOGGER.debug("Files have been changed: {}", changedFiles);

		Set<ResourceReference> openFiles = getOpenInEditorReferences();
		LOGGER.debug("Files currently open in editors: {}", changedFiles);

		Set<ResourceReference> relevantFiles = Sets.intersection(changedFiles, openFiles);
		if (relevantFiles.isEmpty()) {
			return;
		}

		taskExecutor.execute(FetchFindingsTask.createFetchTask(serviceRegistry, ENotificationCause.INDIRECT,
				relevantFiles, IUiPreferences.getInstance().isAutomaticPreCommitAnalysisEnabled()));
	}

	private Set<ResourceReference> getOpenInEditorReferences() {
		IWorkbenchPage activePage = WorkbenchUtils.getActiveWorkbenchPage().orElse(null);
		if (activePage == null) {
			return Collections.emptySet();
		}

		Set<ResourceReference> result = new HashSet<>();
		for (IEditorReference editorRef : activePage.getEditorReferences()) {
			try {
				ResourceReferences.fromIResource(EditorUtils.getEditorFile(editorRef.getEditorInput()))
						.ifPresent(result::add);
			} catch (PartInitException e) {
				LOGGER.catching(e);
				// Silently ignore. The file referenced in the editor may not longer exist.
			}
		}

		return result;
	}
}
