package com.teamscale.ide.eclipse.findings.descriptions;

import java.net.URI;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.inject.Inject;

import org.checkerframework.checker.guieffect.qual.UIEffect;
import org.checkerframework.checker.nullness.qual.Nullable;
import org.eclipse.e4.ui.di.UISynchronize;
import org.eclipse.jface.layout.GridDataFactory;
import org.eclipse.jface.layout.GridLayoutFactory;
import org.eclipse.jface.resource.JFaceColors;
import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTError;
import org.eclipse.swt.browser.Browser;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.ImmutableSet;
import com.google.common.net.UrlEscapers;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.IdeCheckDescription;
import com.teamscale.ide.commons.findings.services.ICheckDescriptionListener;
import com.teamscale.ide.commons.findings.services.ICheckDescriptionStoreService;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.util.markdown.MarkdownUtils;

/**
 * The "Check Descriptions" view displaying the description of a check.
 */
public class CheckDescriptionsPart implements ICheckDescriptionListener {

	/**
	 * The (Markdown) message shown initially (i.e., when no check description was
	 * fetched yet)
	 */
	@VisibleForTesting
	static final String NO_CHECK_DESCRIPTION_MARKDOWN = "No check description to display";

	/**
	 * Of the browser engines supported by Eclipse, at least Internet Explorer
	 * (sadly <a href="https://www.eclipse.org/swt/faq.php#browserplatforms">the
	 * default</a> under Windows), does not support {@code about:} URLs, as produced
	 * by {@link #htmlAsDataUrl(String)}, as input to
	 * {@link Browser#setUrl(String)}. We hence resort to an explicit allow list.
	 * 
	 * @see #SUPPORTED_ENGINES_MASK
	 */
	private static final ImmutableSet<String> SUPPORTED_ENGINES = ImmutableSet.of("webkit", "edge");

	/**
	 * The {@link SWT} constant used for the WebKit renderer. Copied here for
	 * consistency with {@link #ENGINE_EDGE}.
	 * 
	 * @see SWT#WEBKIT
	 */
	private static final int ENGINE_WEBKIT = 1 << 16;

	/**
	 * The {@link SWT} constant used for Edge renderer. Only available under Windows
	 * and <a href="https://www.eclipse.org/swt/faq.php#howuseedge">since Eclipse
	 * 4.19</a>.
	 * 
	 * @see SWT#EDGE
	 */
	private static final int ENGINE_EDGE = 1 << 18;

	/** @see #SUPPORTED_ENGINES */
	private static final int SUPPORTED_ENGINES_MASK = ENGINE_WEBKIT | ENGINE_EDGE;

	private final UISynchronize uiThread;
	private final ICheckDescriptionStoreService checkDescriptionStoreService;

	/**
	 * The {@link Browser} widget to display the check description on. May be
	 * {@code null} in case there are issues with the underlying renderer (e.g.,
	 * version incompatibilities).
	 */
	private @Nullable Browser browser;

	@Inject
	public CheckDescriptionsPart(IServiceRegistry serviceRegistry, UISynchronize uiThread) {
		this.uiThread = uiThread;

		checkDescriptionStoreService = serviceRegistry.getService(ICheckDescriptionStoreService.class);
	}

	/** Called (on the UI thread) if the view is opened. */
	@PostConstruct
	@UIEffect
	public void onOpen(Composite parent) {
		try {
			browser = new Browser(parent, SUPPORTED_ENGINES_MASK);
		} catch (SWTError e) {
			showError(parent, "Could not load browser engine; check description cannot be displayed.");
			return;
		}

		String browserType = browser.getBrowserType();
		if (!SUPPORTED_ENGINES.contains(browserType)) {
			browser.dispose();

			showError(parent, String.format("Unsupported browser engine (%s); check description cannot be displayed.",
					browserType));
			return;
		}

		GridDataFactory.fillDefaults().grab(true, true).applyTo(browser);
		browser.setEnabled(true);

		checkDescriptionStoreService.addListener(this);

		showMarkdown(NO_CHECK_DESCRIPTION_MARKDOWN);
	}

	private void showError(Composite parent, String message) {
		Label errorText = new Label(parent, SWT.CENTER | SWT.WRAP);
		GridLayoutFactory.swtDefaults().applyTo(parent);
		GridDataFactory.fillDefaults().align(SWT.CENTER, SWT.CENTER).grab(true,  true).applyTo(errorText);
		errorText.setText(message);
	}

	/** Called (on the UI thread) if the view is closed. */
	@PreDestroy
	@UIEffect
	public void onClose() {
		checkDescriptionStoreService.removeListener(this);

		if (browser != null) {
			browser.dispose();
		}
	}

	@Override
	public void onCheckDescriptionLoaded(URI serverUri, Finding finding, IdeCheckDescription description) {
		uiThread.asyncExec(() -> showMarkdown(description.getDescription()));
	}

	/** Parses the given Markdown to HTML and shows it in the browser. */
	@UIEffect
	private void showMarkdown(String markdown) {
		Display display = browser.getDisplay();
		Color foregroundColor = JFaceColors.getInformationViewerForegroundColor(display);
		Color backgroundColor = JFaceColors.getInformationViewerBackgroundColor(display);
		Color activeLinkColor = JFaceColors.getActiveHyperlinkText(display);
		Color passiveLinkColor = JFaceColors.getHyperlinkText(display);

		String css = String.format("body { color: %s; background-color: %s; } a:active { color: %s; } a { color: %s}",
				toCssValue(foregroundColor), toCssValue(backgroundColor), toCssValue(activeLinkColor),
				toCssValue(passiveLinkColor));
		String html = MarkdownUtils.renderToHtml(markdown, css);
		browser.setUrl(htmlAsDataUrl(html));
	}

	private static String toCssValue(Color color) {
		return String.format("rgb(%d, %d, %d)", color.getRed(), color.getGreen(), color.getBlue());
	}

	/**
	 * Encodes the given HTML string as a {@code data:} URL. Modern browsers (see
	 * the <a href="https://www.eclipse.org/swt/faq.php#browserplatforms">FAQ</a>
	 * for the renderers used by SWT) support {@code data:} URLs of multiple
	 * megabytes in size, so encoding complete check descriptions like this should
	 * not pose a problem.
	 * 
	 * Using {@code data:} URLs together with {@link Browser#setUrl(String)} has one
	 * advantage over using {@link Browser#setText(String)} directly: The check
	 * description is properly added to the browser's history, so the browser's back
	 * button works correctly when the user clicks on a link in a check description
	 * and then wants to navigate back.
	 * 
	 * @see <a href="https://www.rfc-editor.org/rfc/rfc2397.html">RFC 2397: The
	 *      "data" URL scheme</a>
	 */
	private static String htmlAsDataUrl(String html) {
		return "data:text/html," + UrlEscapers.urlPathSegmentEscaper().escape(html);
	}
}
