/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.findings.handlers;

import javax.inject.Named;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.eclipse.e4.core.di.annotations.CanExecute;
import org.eclipse.e4.core.di.annotations.Execute;
import org.eclipse.e4.ui.services.IServiceConstants;
import org.eclipse.jface.dialogs.InputDialog;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.widgets.Shell;

import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.ITaskExecutor;
import com.teamscale.ide.commons.plugin.services.TaskBase;

/**
 * Base class for handlers that, given a single selected {@link Finding}, offer
 * to black-list it.
 */
public abstract class BlacklistFindingHandlerBase {

	// Logger is *not* static as package of *concrete* subclass must determine the
	// logging category.
	private final Logger logger = LogManager.getLogger(getClass());

	private final String dialogTitle;

	private final IServiceRegistry serviceRegistry;
	private final ITaskExecutor taskExecutor;

	public BlacklistFindingHandlerBase(IServiceRegistry serviceRegistry, String dialogTitle) {
		this.dialogTitle = dialogTitle;

		this.serviceRegistry = serviceRegistry;
		taskExecutor = serviceRegistry.getService(ITaskExecutor.class);
	}

	/**
	 * Creates a {@link TaskBase task} which blacklists the finding.
	 */
	protected abstract TaskBase createBlacklistTask(IServiceRegistry serviceRegistry, Finding finding,
			String rationale);

	@Execute
	public final void execute(@Named(IServiceConstants.ACTIVE_SELECTION) Finding finding, Shell shell) {
		logger.traceEntry(null, finding, shell);
		InputDialog dialog = new InputDialog(shell, dialogTitle, "Rationale:", null, null);
		int choice = dialog.open();
		if (choice == Window.OK) {
			String rationale = dialog.getValue();

			taskExecutor.execute(createBlacklistTask(serviceRegistry, finding, rationale));
		}
		logger.traceExit();
	}

	/**
	 * This handler can only execute when the selected finding has a
	 * {@linkplain Finding#getBaseBranch() base branch}.
	 */
	@CanExecute
	public boolean canExecute(@Named(IServiceConstants.ACTIVE_SELECTION) Finding finding) {
		logger.traceEntry(null, finding);
		boolean canExecute = finding.getBaseBranch().isPresent();
		return logger.traceExit(canExecute);
	}
}
