
/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.findings.handlers;

import java.net.URI;

import javax.inject.Named;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.e4.core.di.annotations.Execute;
import org.eclipse.e4.core.di.annotations.Optional;
import org.eclipse.e4.ui.services.IServiceConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IDocument;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.ide.IDE;
import org.eclipse.ui.texteditor.ITextEditor;

import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.eclipse.util.handlers.AdaptingSelectionHandlerBase;

/**
 * A handler that, given a single selected {@link Finding}, opens the
 * corresponding source file in an editor, with the finding centered.
 */
public class OpenFindingInEditorHandler {

	private static final Logger LOGGER = LogManager.getLogger();

	@Execute
	public void execute(@Named(IServiceConstants.ACTIVE_SELECTION) Finding finding, IWorkbenchPage page, @Optional Shell activeShell) {
		LOGGER.traceEntry(null, finding, page);
		URI fileUri = finding.getLocalResource().toUri();
		IFile[] files = ResourcesPlugin.getWorkspace().getRoot().findFilesForLocationURI(fileUri);
		if (files.length > 0) {
			try {
				IEditorPart editor = IDE.openEditor(page, files[0]);
				jumpToLine(editor, finding.getPrimaryLineRange().getFirstLine());
			} catch (PartInitException e) {
				LOGGER.catching(e);
				MessageDialog.openError(activeShell, "Error", "Failed to open editor.");
			}
		}
		LOGGER.traceExit();
	}

	/**
	 * Jump to the given line number (starting at 1) in the given editor.
	 */
	private static void jumpToLine(IEditorPart editor, int lineNumber) {
		if (!(editor instanceof ITextEditor)) {
			return;
		}

		ITextEditor textEditor = (ITextEditor) editor;
		IDocument document = textEditor.getDocumentProvider().getDocument(editor.getEditorInput());
		try {
			int start = document.getLineOffset(lineNumber - 1);
			textEditor.selectAndReveal(start, 0);
		} catch (BadLocationException | NumberFormatException e) {
			// can be ignored, as then we just don't reach the line
			LOGGER.catching(e);
		}
	}

	/**
	 * @see OpenFindingInEditorHandler
	 */
	public static class AdaptingHandler extends AdaptingSelectionHandlerBase<Finding> {

		public AdaptingHandler() {
			super(OpenFindingInEditorHandler.class, Finding.class);
		}
	}
}
