/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.logging;

import java.io.Serializable;
import java.util.Hashtable;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;

import org.apache.logging.log4j.core.Appender;
import org.apache.logging.log4j.core.Filter;
import org.apache.logging.log4j.core.Layout;
import org.apache.logging.log4j.core.LogEvent;
import org.apache.logging.log4j.core.appender.AbstractAppender;
import org.apache.logging.log4j.core.config.Node;
import org.apache.logging.log4j.core.config.Property;
import org.apache.logging.log4j.core.config.plugins.Plugin;
import org.apache.logging.log4j.core.config.plugins.PluginBuilderFactory;
import org.eclipse.osgi.service.debug.DebugOptions;
import org.eclipse.osgi.service.debug.DebugOptionsListener;
import org.eclipse.osgi.service.debug.DebugTrace;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.FrameworkUtil;
import org.osgi.framework.ServiceRegistration;

import com.google.common.base.Preconditions;
import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.collect.Sets;
import com.teamscale.ide.commons.logging.AppenderBuilderBase;
import com.teamscale.ide.commons.logging.Condition;
import com.teamscale.ide.commons.logging.Rule;
import com.teamscale.ide.eclipse.EclipseIds;

/**
 * A Log4J 2.x {@link Appender} backed by the Eclipse {@link DebugTrace}
 * mechanism, as configured on the General > Tracing preference page.
 * 
 * @see <a href=
 *      "https://wiki.eclipse.org/FAQ_How_do_I_use_the_platform_debug_tracing_facility">FAQ:
 *      How do I use the platform debug tracing facility</a>
 */
@Plugin(name = "EclipseDebugTrace", category = Node.CATEGORY, elementType = Appender.ELEMENT_TYPE, printObject = true)
public class EclipseDebugTraceAppender extends AbstractAppender implements DebugOptionsListener {

	/**
	 * The rules this filter uses to map {@link LogEvent}s to {@link DebugOptions
	 * options}.
	 */
	private final Rule[] rules;

	/**
	 * The registration of this appender as a {@link DebugOptionsListener} OSGi
	 * service.
	 */
	private ServiceRegistration<DebugOptionsListener> serviceRegistration;

	/** The current {@link DebugTrace} this appender logs to. */
	private final AtomicReference<DebugTrace> debugTrace = new AtomicReference<>();

	/** The {@link DebugOptions options} that are currently enabled. */
	private final Set<String> enabledOptions = Sets.newConcurrentHashSet();

	private final LoadingCache<Condition, String> resolvedOptions = CacheBuilder.newBuilder()
			.build(new CacheLoader<Condition, String>() {

				@Override
				public String load(Condition condition) {
					Optional<Rule> matchingRule = condition.findMatching(rules);
					return matchingRule.map(Rule::getKey).orElse(null);
				}
			});

	private EclipseDebugTraceAppender(String name, Filter filter, Layout<? extends Serializable> layout,
			boolean ignoreExceptions, Rule... rules) {
		super(name, filter, layout, ignoreExceptions, Property.EMPTY_ARRAY);

		this.rules = rules;
	}

	@Override
	public void start() {
		super.start();

		serviceRegistration = registerDebugOptionsListener(this);
	}

	private static ServiceRegistration<DebugOptionsListener> registerDebugOptionsListener(
			DebugOptionsListener listener) {
		Bundle bundle = FrameworkUtil.getBundle(listener.getClass());
		BundleContext context = bundle.getBundleContext();

		Hashtable<String, String> properties = new Hashtable<>();
		properties.put(DebugOptions.LISTENER_SYMBOLICNAME, bundle.getSymbolicName());
		return context.registerService(DebugOptionsListener.class, listener, properties);
	}

	@Override
	public void stop() {
		Preconditions.checkState(serviceRegistration != null);

		serviceRegistration.unregister();

		super.stop();
	}

	@Override
	public void optionsChanged(DebugOptions options) {
		debugTrace.set(options.newDebugTrace(EclipseIds.PLUGIN));

		for (Rule rule : rules) {
			String option = rule.getKey();
			boolean enabled = options.getBooleanOption(option, false);
			if (enabled) {
				enabledOptions.add(option);
			} else {
				enabledOptions.remove(option);
			}
		}
	}

	@Override
	public void append(LogEvent event) {
		Condition condition = new Condition(event.getLoggerName(), event.getMarker(), event.getLevel());
		String option = resolvedOptions.getUnchecked(condition);
		if (option == null || !enabledOptions.contains(option)) {
			return;
		}

		String message = new String(getLayout().toByteArray(event));
		// We leave logging the exception to Log4J's Layout rather than passing it as a
		// third argument to the DebugTrace, as the Layout provides more details (JAR
		// file names).
		debugTrace.get().trace(option.substring(option.indexOf('/')), message);
	}

	/**
	 * Builds new {@link EclipseDebugTraceAppender}s.
	 * <p>
	 * Entrypoint for Log4J's plug-in mechanism.
	 * 
	 * @see <a href=
	 *      "https://logging.apache.org/log4j/2.x/manual/plugins#Core">Plugins:
	 *      Core</a>
	 */
	@PluginBuilderFactory
	public static Builder newBuilder() {
		return new Builder();
	}

	/**
	 * A builder for {@link EclipseDebugTraceAppender}s.
	 */
	public static class Builder extends AppenderBuilderBase<EclipseDebugTraceAppender> {

		@Override
		public EclipseDebugTraceAppender build() {
			return new EclipseDebugTraceAppender(getName(), getFilter(), getLayout(), isIgnoreExceptions(), getRules());
		}
	}
}
