package com.teamscale.ide.eclipse.mappings.services;

import java.net.URI;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.conqat.ide.commons.ui.adt.AdtUtils;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ResourcesPlugin;

import com.teamscale.ide.client.PrefixMappingRegistry;
import com.teamscale.ide.client.adt.SuccessfullAdtResourceMappingResult;
import com.teamscale.ide.commons.mappings.MappingProblem;
import com.teamscale.ide.commons.mappings.ResourceMappingResult;
import com.teamscale.ide.commons.model.IIDEProject;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.plugin.services.IResourceResolutionService;
import com.teamscale.ide.commons.servers.NamedServerDetails;
import com.teamscale.ide.commons.servers.services.IServerStateBroadcastService;
import com.teamscale.ide.commons.servers.services.ITeamscaleServerManager;
import com.teamscale.ide.eclipse.model.EclipseProject;
import com.teamscale.ide.eclipse.model.ResourceReferences;
import com.teamscale.ide.eclipse.plugin.EclipseServiceRegistry;

/**
 * A wrapper for the {@link LegacyResourceMappingService} to support ADT (ABAP)
 * object references.
 */
public class EclipseResourceMappingService extends LegacyResourceMappingService {

	private static final Logger LOGGER = LogManager.getLogger();

	public EclipseResourceMappingService(IResourceResolutionService resolutionService,
			ITeamscaleServerManager serverManager, IServerStateBroadcastService broadcastService) {
		super(resolutionService, serverManager, broadcastService);
	}

	@Override
	public ResourceMappingResult map(ResourceReference resource) {
		LOGGER.traceEntry(null, resource);

		if (serverManager.getAllServerUrls().isEmpty()) {
			return LOGGER.traceExit(ResourceMappingResult.failure(MappingProblem.noServersConfigured()));
		}

		IIDEProject ideProject = resolutionService.getContainingProject(resource).orElse(null);
		if (ideProject == null) {
			return LOGGER.traceExit(ResourceMappingResult.failure(MappingProblem.resourceOutsideMappingRoot()));
		}

		IProject project = ((EclipseProject) ideProject).getProject();
		if (AdtUtils.isAdtProject(project)) {
			return LOGGER.traceExit(mapAdt(resource));
		}

		// no ADT project => let the the base class handle the mapping
		return LOGGER.traceExit(super.map(resource));
	}

	/**
	 * Maps the local resource to a Teamscale resource.
	 */
	private ResourceMappingResult mapAdt(ResourceReference resourceReference) {
		if (!resourceReference.isFile()) {
			// For ADT findings are only available for files
			return ResourceMappingResult.failure(MappingProblem.noMappingFound());
		}

		IFile[] files = ResourcesPlugin.getWorkspace().getRoot().findFilesForLocationURI(resourceReference.toUri());
		if (files.length != 1) {
			return ResourceMappingResult.failure(MappingProblem.noMappingFound());
		}
		IFile file = files[0];

		PrefixMappingRegistry registry = PrefixMappingRegistry.getInstance(file.getProject());
		if (registry == null) {
			return ResourceMappingResult.failure(MappingProblem.noMappingFound());
		}
		if (!registry.isTeamscaleEnabled()) {
			return ResourceMappingResult.failure(MappingProblem
					.mappingRootNotConfigured(ResourceReferences.fromIResource(file.getProject()).orElse(null)));
		}

		NamedServerDetails server = EclipseServiceRegistry.getInstance().getService(ITeamscaleServerManager.class)
				.getNamedServerDetails(registry.getTeamscaleServer());
		if (server == null) {
			return ResourceMappingResult
					.failure(MappingProblem.teamscaleServerUnknown(null, registry.getTeamscaleServer()));
		}

		return new SuccessfullAdtResourceMappingResult(URI.create(server.getUrl()), registry.getTeamscaleProject());
	}
}
