package com.teamscale.ide.eclipse.mappings.services;

import static java.util.Objects.requireNonNull;

import java.net.URI;
import java.util.Collections;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import com.teamscale.ide.client.PrefixMappingRegistry;
import com.teamscale.ide.commons.mappings.IDEProjectSettings;
import com.teamscale.ide.commons.mappings.ILegacyResourceMapping;
import com.teamscale.ide.commons.mappings.MappingProblem;
import com.teamscale.ide.commons.mappings.ResourceMappingResult;
import com.teamscale.ide.commons.mappings.services.IResourceMappingService;
import com.teamscale.ide.commons.model.IIDEProject;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.model.ResourceReferenceUtils;
import com.teamscale.ide.commons.plugin.services.IResourceResolutionService;
import com.teamscale.ide.commons.servers.ETeamscaleServerState;
import com.teamscale.ide.commons.servers.NamedServerDetails;
import com.teamscale.ide.commons.servers.services.IServerStateBroadcastService;
import com.teamscale.ide.commons.servers.services.ITeamscaleServerManager;
import com.teamscale.ide.eclipse.model.EclipseProject;

/**
 * A compatibility layer between the old {@link IDEProjectSettings}-based way to
 * configure mappings and the new [@link {@link IResourceMappingService} that is
 * aimed at supporting a {@code .teamscale.toml}-based configuration.
 */
public class LegacyResourceMappingService implements IResourceMappingService {

	private static final Logger LOGGER = LogManager.getLogger();

	private final IServerStateBroadcastService broadcastService;

	/** Resolution service */
	protected final IResourceResolutionService resolutionService;

	/** Server manager */
	protected final ITeamscaleServerManager serverManager;

	public LegacyResourceMappingService(IResourceResolutionService resolutionService,
			ITeamscaleServerManager serverManager, IServerStateBroadcastService broadcastService) {
		this.broadcastService = requireNonNull(broadcastService);
		this.resolutionService = requireNonNull(resolutionService);
		this.serverManager = requireNonNull(serverManager);
	}

	@Override
	public ResourceMappingResult map(ResourceReference resource) {
		LOGGER.traceEntry(null, resource);

		if (serverManager.getAllServerUrls().isEmpty()) {
			return LOGGER.traceExit(ResourceMappingResult.failure(MappingProblem.noServersConfigured()));
		}

		IIDEProject ideProject = resolutionService.getContainingProject(resource).orElse(null);
		if (ideProject == null) {
			return LOGGER.traceExit(ResourceMappingResult.failure(MappingProblem.resourceOutsideMappingRoot()));
		}

		ResourceReference projectRoot = ideProject.getRootDirectory();
		if (!resource.isAtOrBelow(projectRoot)) {
			return LOGGER.traceExit(ResourceMappingResult.failure(MappingProblem.resourceOutsideMappingRoot()));
		}
		
		ILegacyResourceMapping mapping = getMapping(ideProject);
		if (mapping == null) {
			return LOGGER
					.traceExit(ResourceMappingResult.failure(MappingProblem.mappingRootNotConfigured(projectRoot)));
		}
		if (!mapping.isTeamscaleEnabled()) {
			// Teamscale is not enabled for the project containing the current file.
			// So either the user made an error (expecting Teamscale on a project where he
			// disabled it) or our plugin reacts without the user intending this.
			// Unfortunately, we can't distinguish these cases. But the notification service
			// shows a warning message only the first time this happens for a given project,
			// so this is not too intrusive.
			return LOGGER
					.traceExit(ResourceMappingResult.failure(MappingProblem.mappingRootNotConfigured(projectRoot)));
		}
		

		String serverId = mapping.getTeamscaleServer();
		if (serverId == null) {
			return LOGGER.traceExit(ResourceMappingResult.failure(MappingProblem.noTeamscaleServer(projectRoot)));
		}

		NamedServerDetails server = serverManager.getNamedServerDetails(serverId);
		if (server == null) {
			return LOGGER.traceExit(
					ResourceMappingResult.failure(MappingProblem.teamscaleServerUnknown(projectRoot, serverId)));
		}

		String projectId = mapping.getTeamscaleProject();
		if (projectId == null) {
			broadcastService.updateServerState(server, ETeamscaleServerState.NOT_CONFIGURED);
			return LOGGER.traceExit(ResourceMappingResult.failure(MappingProblem.noTeamscaleProject(projectRoot)));
		}

		ResourceMappingResult result = mapping.map(URI.create(server.getUrl()), projectRoot,
				ResourceReferenceUtils.computeRelativeUniformPath(projectRoot, resource));

		return LOGGER.traceExit(result);
	}

	@Override
	public Set<ResourceReference> findPotentialMappingRoots(ResourceReference resource) {
		IIDEProject ideProject = resolutionService.getContainingProject(resource).orElse(null);
		if (ideProject == null) {
			return Collections.emptySet();
		}
		return Collections.singleton(ideProject.getRootDirectory());
	}
	
	/**
	 * Workaround to enable testing without a running eclipse project.
	 */
	protected ILegacyResourceMapping getMapping(IIDEProject ideProject) {
		return PrefixMappingRegistry.getInstance(((EclipseProject) ideProject).getProject());
	}
}
