/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.notification.services;

import java.net.MalformedURLException;
import java.net.URL;
import java.text.MessageFormat;
import java.util.Objects;

import org.conqat.ide.commons.ui.logging.LoggingUtils;
import org.eclipse.core.resources.IProject;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.preference.PreferenceDialog;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.events.TypedEvent;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.dialogs.PreferencesUtil;

import com.teamscale.ide.TeamscaleIDEActivator;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.notification.impl.ApiNotSupportedNotification;
import com.teamscale.ide.commons.notification.impl.CheckDescriptionNotFoundNotification;
import com.teamscale.ide.commons.notification.impl.ConnectionErrorNotification;
import com.teamscale.ide.commons.notification.impl.FailedOpenUrisNotification;
import com.teamscale.ide.commons.notification.impl.HigherServerVersionWarningNotification;
import com.teamscale.ide.commons.notification.impl.NoMappingFoundNotification;
import com.teamscale.ide.commons.notification.impl.PreCommitErrorNotification;
import com.teamscale.ide.commons.notification.impl.ProjectNotFoundNotification;
import com.teamscale.ide.commons.notification.impl.ResourceOutsideMappingRootNotification;
import com.teamscale.ide.commons.notification.impl.SapTeamscaleSyncTimeoutNotification;
import com.teamscale.ide.commons.notification.impl.ServerUnknownNotification;
import com.teamscale.ide.commons.notification.impl.ServerUrlNotFoundNotification;
import com.teamscale.ide.commons.notification.impl.TeamscaleBranchNotFoundNotification;
import com.teamscale.ide.commons.notification.impl.TomlConfigInvalidNotification;
import com.teamscale.ide.commons.notification.impl.legacy.MappingRootNotConfiguredNotification;
import com.teamscale.ide.commons.notification.impl.legacy.NoServersConfiguredNotification;
import com.teamscale.ide.commons.notification.impl.legacy.NoTeamscaleProjectNotification;
import com.teamscale.ide.commons.notification.impl.legacy.NoTeamscaleServerNotification;
import com.teamscale.ide.commons.notification.services.INotificationMemory;
import com.teamscale.ide.commons.notification.services.INotificationService;
import com.teamscale.ide.commons.notification.services.impl.NotificationServiceBase;
import com.teamscale.ide.commons.plugin.services.IResourceResolutionService;
import com.teamscale.ide.eclipse.EclipseIds;
import com.teamscale.ide.eclipse.util.dialogs.MessageDialogWithLink;
import com.teamscale.ide.eclipse.util.popup.MessagePopup;
import com.teamscale.ide.eclipse.util.popup.MessagePopupWithLink;
import com.teamscale.ide.eclipse.util.swt.DisplayUtils;

/**
 * Eclipse-specific implementation of {@link INotificationService}.
 */
public class EclipseNotificationService extends NotificationServiceBase {

	private final IResourceResolutionService resourceResolutionService;

	public EclipseNotificationService(IResourceResolutionService resourceResolutionService) {
		super(INotificationMemory.createDefaultMemory());
		this.resourceResolutionService = Objects.requireNonNull(resourceResolutionService);
	}

	@Override
	public void reportPreCommitError(PreCommitErrorNotification notification) {
		String notificationType = notification.isError() ? "Error" : "Warning";
		String title = MessageFormat.format("Pre-Commit {0}", notificationType);
		DisplayUtils.asyncExec(display -> MessagePopup.show(display, title, notification.getMessage()));

		if (notification.isError()) {
			LoggingUtils.error(TeamscaleIDEActivator.getDefault(), title + ": " + notification.getMessage(), null);
		} else {
			LoggingUtils.warn(TeamscaleIDEActivator.getDefault(), title + ": " + notification.getMessage(), null);
		}
	}

	@Override
	public void reportServerUrlNotFound(ServerUrlNotFoundNotification notification) {
		DisplayUtils.asyncExec(display -> MessageDialogWithLink.openError(display.getActiveShell(), "Server Not Found",
				MessageFormat.format("Teamscale server {0} could not be found anymore. Maybe it has been deleted?",
						notification.getServerUrl()),
				"Check Teamscale servers on the <a>'Teamscale'</a> preference page.",
				SelectionListener.widgetSelectedAdapter(EclipseNotificationService::openTeamscaleServerPreferences)));
	}

	@Override
	public void reportConnectionError(ConnectionErrorNotification notification) {
		DisplayUtils.asyncExec(display -> MessageDialogWithLink.openError(display.getActiveShell(), "Connnection Error",
				notification.getMessage(), "Check Teamscale servers on the <a>'Teamscale'</a> preference page.",
				SelectionListener.widgetSelectedAdapter(EclipseNotificationService::openTeamscaleServerPreferences)));
	}

	@Override
	public void reportApiNotSupported(ApiNotSupportedNotification notification) {
		DisplayUtils.asyncExec(display -> MessageDialog.openError(display.getActiveShell(), //
				MessageFormat.format("Service call requires Teamscale server version {0} or higher",
						notification.getLowestSupportedServerVersion()), //
				notification.getMessage()));
	}

	@Override
	public void reportHigherServerVersionWarning(HigherServerVersionWarningNotification notification) {		
		String url = "https://cqse.eu/get-eclipse";
		DisplayUtils.asyncExec(display -> MessagePopupWithLink.show(display,
				"Teamscale Plugin Update Recommended",
				MessageFormat.format(
						"The plugin version {0} which is currently installed is lower than version {1} of the server {2}. "
								+ "This may lead to compatibility issues. We strongly recommend updating the plugin.",
								notification.getPluginVersion(), notification.getServerVersion(),
						notification.getServerUrl()),
				"<a>Teamscale Eclipse plugin documentation</a>",
				SelectionListener.widgetSelectedAdapter(event -> doOpenURL(url))));
	}
	
	private void doOpenURL(String url) {
		try {
			PlatformUI.getWorkbench().getBrowserSupport().getExternalBrowser().openURL(new URL(url));
		} catch (PartInitException | MalformedURLException e) {
			LoggingUtils.error(TeamscaleIDEActivator.getDefault(), "Could not create URL for server warning: " + e.getMessage(), null);
		}
	}
	
	@Override
	public void reportProjectNotFound(ProjectNotFoundNotification notification) {
		DisplayUtils.asyncExec(display -> MessageDialog.openError(display.getActiveShell(), //
				"Unknown Teamscale Project",
				MessageFormat.format(
						"{0}: The project {1} is not known to the Teamscale server. Please specify an existing project.",
						notification.getServerUrl(), notification.getProjectId())));
	}

	@Override
	public void reportTeamscaleBranchNotFound(TeamscaleBranchNotFoundNotification notification) {
		DisplayUtils.asyncExec(display -> MessageDialog.openError(display.getActiveShell(), //
				"Unknown Teamscale Branch",
				MessageFormat.format("{0}: The branch {1} of the project {2} is not known to the Teamscale server."
						+ " Please publish your branch for Teamscale or check out a branch known to the Teamscale server.",
						notification.getServerUrl(), notification.getBranch(), notification.getProjectId())));
	}

	@Override
	public void reportResourceOutsideMappingRoot(ResourceOutsideMappingRootNotification notification) {
		DisplayUtils.asyncExec(display -> MessageDialog.openWarning(display.getActiveShell(), //
				"Warning",
				MessageFormat.format("The selected resource {0} does not reside within a project which has Teamscale enabled.",
						notification.getResource().toDisplayString())));
	}

	@Override
	public void reportServerUnknown(ServerUnknownNotification notification) {
		DisplayUtils.asyncExec(display -> MessageDialogWithLink.openWarning(display.getActiveShell(),
				"Teamscale Server Unknown", //
				"The Teamscale server " + notification.getServerUrl()
						+ " specified in the project mapping is unknown.\n" + "Please update your project mappings.",
				"Configure your Teamscale servers on the <a>'Teamscale'</a> property page.",
				SelectionListener.widgetSelectedAdapter(
						event -> openTeamscaleProjectProperties(event, notification.getMappingRoot()))));
	}

	@Override
	public void reportNoMappingFound(NoMappingFoundNotification notification) {
		ResourceReference resource = notification.getResource();
		DisplayUtils.asyncExec(display -> MessageDialogWithLink.openWarning(display.getActiveShell(), //
				"No Mapping Found", //
				MessageFormat.format("Could not map the local resource {0} to a resource on the Teamscale server.",
						resource.toDisplayString()), //
				"Configure your project on the <a>'Teamscale'</a> property page.", //
				SelectionListener.widgetSelectedAdapter(event -> openTeamscaleProjectProperties(event, resource))));
	}

	@Override
	public void reportTomlConfigInvalid(TomlConfigInvalidNotification notification) {
		throw new UnsupportedOperationException("No TOML support");
	}

	@Override
	public void reportFailedOpenUris(FailedOpenUrisNotification notification) {
		DisplayUtils.asyncExec(display -> MessageDialog.openError(display.getActiveShell(), "Error",
				"Failed to open the following URIs in your browser: " + notification.getFailedUrisFormatted()));
	}

	@Override
	public void reportCheckDescriptionNotFound(CheckDescriptionNotFoundNotification notification) {
		DisplayUtils.asyncExec(display -> MessageDialog.openError(display.getActiveShell(), "Error",
				"No check description found for type ID '" + notification.getCheckId() + "'"));
	}

	@Override
	public void reportNoServerConfigured(NoServersConfiguredNotification notification) {
		DisplayUtils.asyncExec(display -> MessageDialogWithLink.openWarning(display.getActiveShell(), //
				"No Teamscale Servers Configured", //
				"No Teamscale servers have been configured.", //
				"Configure your Teamscale servers on the <a>'Teamscale'</a> preference page.", //
				SelectionListener.widgetSelectedAdapter(EclipseNotificationService::openTeamscaleServerPreferences)));
	}

	@Override
	public void reportSapTeamscaleSyncTimeout(SapTeamscaleSyncTimeoutNotification notification) {
		DisplayUtils.asyncExec(display -> MessageDialog.openError(display.getActiveShell(), "Error",
				"Timeout while updating the SAP system: " + notification.getSapSystemId()));
	}

	@Override
	public void reportMappingRootNotConfigured(MappingRootNotConfiguredNotification notification) {
		DisplayUtils.asyncExec(display -> MessageDialogWithLink.openWarning(display.getActiveShell(), //
				"Project Not Configured For Teamscale", //
				MessageFormat.format("The project at {0} has not been configured for Teamscale.",
						notification.getMappingRoot().toDisplayString()), //
				"Configure your project on the <a>'Teamscale'</a> property page.", //
				SelectionListener.widgetSelectedAdapter(
						event -> openTeamscaleProjectProperties(event, notification.getMappingRoot()))));
	}

	@Override
	public void reportNoTeamscaleServer(NoTeamscaleServerNotification notification) {
		DisplayUtils.asyncExec(display -> MessageDialogWithLink.openWarning(display.getActiveShell(), //
				"No Teamscale Server", //
				MessageFormat.format("No Teamscale server has been set for the project at {0}.",
						notification.getMappingRoot().toDisplayString()), //
				"Configure your project on the <a>'Teamscale'</a> property page.", //
				SelectionListener.widgetSelectedAdapter(
						event -> openTeamscaleProjectProperties(event, notification.getMappingRoot()))));
	}

	@Override
	public void reportNoTeamscaleProject(NoTeamscaleProjectNotification notification) {
		DisplayUtils.asyncExec(display -> MessageDialogWithLink.openWarning(display.getActiveShell(), //
				"No Teamscale Project", //
				MessageFormat.format("No Teamscale project has been set for the project at {0}.",
						notification.getMappingRoot().toDisplayString()), //
				"Configure your project on the <a>'Teamscale'</a> property page.", //
				SelectionListener.widgetSelectedAdapter(
						event -> openTeamscaleProjectProperties(event, notification.getMappingRoot()))));

	}

	private static void openTeamscaleServerPreferences(TypedEvent event) {
		PreferenceDialog dialog = PreferencesUtil.createPreferenceDialogOn(event.display.getActiveShell(),
				EclipseIds.PreferencePages.TEAMSCALE, null, null);
		dialog.open();
	}

	private void openTeamscaleProjectProperties(TypedEvent event, ResourceReference resource) {
		resourceResolutionService.getContainingProject(resource).ifPresent(project -> {
			openTeamscaleProjectProperties(event, (IProject) project.getProject());
		});
	}

	private static void openTeamscaleProjectProperties(TypedEvent event, IProject project) {
		PreferenceDialog dialog = PreferencesUtil.createPropertyDialogOn(event.display.getActiveShell(), project,
				EclipseIds.PropertyPages.TEAMSCALE, null, null);
		dialog.open();
	}
}
