/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.plugin;

import java.util.List;
import java.util.Locale;
import java.util.function.Consumer;
import java.util.function.Function;

import org.checkerframework.checker.nullness.qual.Nullable;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IProduct;
import org.eclipse.core.runtime.Platform;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.Constants;
import org.osgi.framework.FrameworkUtil;
import org.osgi.framework.Version;

import com.teamscale.ide.commons.client.services.impl.TeamscaleConnectorService;
import com.teamscale.ide.commons.findings.EFindingFetchError;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.services.IFindingsListener;
import com.teamscale.ide.commons.findings.services.impl.CheckDescriptionStoreService;
import com.teamscale.ide.commons.findings.services.impl.FetchFindingsTaskCache;
import com.teamscale.ide.commons.findings.services.impl.FindingFetchStateService;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.notification.services.INotificationService;
import com.teamscale.ide.commons.plugin.IPluginService;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.impl.ServiceRegistry;
import com.teamscale.ide.commons.plugin.services.IResourceResolutionService;
import com.teamscale.ide.commons.servers.services.IServerStateBroadcastService;
import com.teamscale.ide.commons.servers.services.ITeamscaleServerManager;
import com.teamscale.ide.commons.servers.services.impl.ServerStateBroadcastService;
import com.teamscale.ide.commons.util.TeamscaleVersion;
import com.teamscale.ide.commons.vcs.services.GenericSupportedCheckouts;
import com.teamscale.ide.commons.vcs.services.impl.CheckoutFactoryImpl;
import com.teamscale.ide.eclipse.findings.FileDeleteListener;
import com.teamscale.ide.eclipse.findings.FileSaveListener;
import com.teamscale.ide.eclipse.mappings.services.EclipseResourceMappingService;
import com.teamscale.ide.eclipse.notification.services.EclipseNotificationService;
import com.teamscale.ide.eclipse.plugin.services.EclipseResourceResolutionService;
import com.teamscale.ide.eclipse.plugin.services.EclipseTaskExecutor;
import com.teamscale.ide.eclipse.plugin.services.EclipseWebBrowserService;
import com.teamscale.ide.eclipse.servers.services.EclipseTeamscaleServerManager;
import com.teamscale.ide.eclipse.vcs.services.EclipseSupportedCheckouts;
import com.teamscale.ide.findings.FindingsMarkerUtils;

/**
 * An {@link IServiceRegistry} that populates, upon creation, a delegate with
 * {@link IPluginService}s. It is made available as an OSGi declarative service,
 * registered in {@code OSGI-INF/serviceRegistration.xml}.
 */
public class EclipseServiceRegistry extends ServiceRegistry {

	public EclipseServiceRegistry() {
		// Core plugin services
		registerService(new EclipseResourceResolutionService());
		registerService(new EclipseNotificationService(getService(IResourceResolutionService.class)));
		registerService(new EclipseTaskExecutor());
		registerService(new EclipseWebBrowserService());

		// VCS services
		registerService(new CheckoutFactoryImpl(GenericSupportedCheckouts.GENERIC_GIT,
				EclipseSupportedCheckouts.SUBCLIPSE, EclipseSupportedCheckouts.SUBVERSIVE));

		// Server-management services
		registerService(new EclipseTeamscaleServerManager());
		registerService(new ServerStateBroadcastService());

		// Service-client services
		registerService(createTeamscaleConnectorService(getService(IServerStateBroadcastService.class),
				getService(INotificationService.class)));

		// Mapping-related services
		registerService(new EclipseResourceMappingService(getService(IResourceResolutionService.class),
				getService(ITeamscaleServerManager.class), getService(IServerStateBroadcastService.class)));

		// Finding-related and pre-commit services
		registerService(new FetchFindingsTaskCache());
		registerService(createFindingFetchStateService());
	}

	/**
	 * Callback invoked when the {@linkplain IWorkspace workspace} is opened.
	 * Registered in {@code OSGI-INF/serviceRegistration.xml}.
	 */
	public void onWorkspaceOpened(IWorkspace workspace) {
		workspace.addResourceChangeListener(new FileSaveListener(this), IResourceChangeEvent.POST_CHANGE);
		workspace.addResourceChangeListener(new FileDeleteListener(this), IResourceChangeEvent.PRE_DELETE);
		registerService(new CheckDescriptionStoreService());
	}

	private static TeamscaleConnectorService createTeamscaleConnectorService(
			IServerStateBroadcastService serverStateBroadcastService, INotificationService notificationService) {
		String pluginVersion = getFromNullable(FrameworkUtil.getBundle(EclipseServiceRegistry.class),
				bundle -> formatVersion(bundle.getVersion()));
		String pluginUserAgent = TeamscaleConnectorService.buildUserAgent(
				getFromNullable(FrameworkUtil.getBundle(EclipseServiceRegistry.class), Bundle::getSymbolicName),
				pluginVersion, getFromNullable(FrameworkUtil.getBundle(EclipseServiceRegistry.class),
						bundle -> bundle.getHeaders(Locale.ENGLISH.toString()).get(Constants.BUNDLE_NAME)));

		String ideUserAgent = TeamscaleConnectorService.buildUserAgent(
				getFromNullable(Platform.getProduct(), IProduct::getId),
				getFromNullable(Platform.getProduct(),
						product -> formatVersion(product.getDefiningBundle().getVersion())),
				getFromNullable(Platform.getProduct(), IProduct::getName));

		String javaUserAgent = TeamscaleConnectorService.buildJavaUserAgent();

		// Our plugin versioning scheme is aligned with the Teamscale versions.
		// Therefore, we can parse a Teamscale version from the plugin version here.
		// This is the newest Teamscale version we had when the current plugin was
		// compiled (i.e., the newest version to which we are compatible).
		TeamscaleVersion pluginTeamscaleVersion = TeamscaleVersion.fromString(pluginVersion);

		return new TeamscaleConnectorService(pluginTeamscaleVersion, serverStateBroadcastService, notificationService,
				pluginUserAgent, ideUserAgent, javaUserAgent);
	}

	/**
	 * Calls the getter on the object. If the object is null, this method returns
	 * null.
	 */
	private static <S, T> @Nullable T getFromNullable(@Nullable S object, Function<S, T> getter) {
		if (object == null) {
			return null;
		}
		return getter.apply(object);
	}

	private static @Nullable String formatVersion(@Nullable Version version) {
		if (version == null) {
			return null;
		}
		return String.format("%s.%s.%s-%s", version.getMajor(), version.getMinor(), version.getMicro(),
				version.getQualifier());
	}

	private FindingFetchStateService createFindingFetchStateService() {
		FindingFetchStateService service = new FindingFetchStateService();
		service.addListener(new IdeFindingsToMarkersListener());
		return service;
	}

	/**
	 * Exposes any received {@link IdeFinding}s as Eclipse {@link IMarker}s. This
	 * one half of the bridge between our IDE Commons API and the direct (and
	 * legacy) use of Eclipse APIs. (The other half is
	 * {@link FindingsMarkerUtils#getCommonsFindingFromMarker(IMarker)}.)
	 */
	private static final class IdeFindingsToMarkersListener implements IFindingsListener {

		@Override
		public void onFindingsFetchSucceeded(ResourceReference fetchResource, List<Finding> findings) {
			forEachAffectedFile(fetchResource, file -> {
				FindingsMarkerUtils.removeAllMarkers(file);
				FindingsMarkerUtils.createMarkers(findings, file);
			});
		}

		@Override
		public void onFindingsFetchFailed(ResourceReference fetchResource, EFindingFetchError error,
				@Nullable String message, List<Finding> findingsOfPreviousCommit) {
			forEachAffectedFile(fetchResource, file -> {
				FindingsMarkerUtils.removeAllMarkers(file);
				FindingsMarkerUtils.createMarkers(findingsOfPreviousCommit, file);
			});
		}

		@Override
		public void onFindingsFetchOutdated(ResourceReference resource) {
			forEachAffectedFile(resource, FindingsMarkerUtils::removeAllMarkers);
		}

		private static void forEachAffectedFile(ResourceReference fetchResource, Consumer<IFile> action) {
			IFile[] affectedFiles = ResourcesPlugin.getWorkspace().getRoot()
					.findFilesForLocationURI(fetchResource.toUri());
			for (IFile file : affectedFiles) {
				action.accept(file);
			}
		}
	}

	/**
	 * @return the {@link IServiceRegistry} registered as an OSGi declarative
	 *         service. Only call from within E3 code. When already using E4 APIs,
	 *         use injection instead.
	 */
	public static IServiceRegistry getInstance() {
		BundleContext bundleContext = FrameworkUtil.getBundle(EclipseServiceRegistry.class).getBundleContext();
		return bundleContext.getService(bundleContext.getServiceReference(IServiceRegistry.class));
	}
}
