/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.plugin.services;

import static java.util.Comparator.comparing;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.URI;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.eclipse.core.filebuffers.FileBuffers;
import org.eclipse.core.filebuffers.ITextFileBuffer;
import org.eclipse.core.filebuffers.ITextFileBufferManager;
import org.eclipse.core.filebuffers.LocationKind;
import org.eclipse.core.filesystem.EFS;
import org.eclipse.core.filesystem.IFileInfo;
import org.eclipse.core.filesystem.IFileStore;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;

import com.google.common.io.CharStreams;
import com.teamscale.ide.commons.model.IIDEProject;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.plugin.services.IResourceResolutionService;
import com.teamscale.ide.eclipse.model.EclipseProject;

/**
 * Eclipse-specific implementation of {@link IResourceResolutionService}.
 */
public class EclipseResourceResolutionService implements IResourceResolutionService {

	private static final Logger LOGGER = LogManager.getLogger();

	/**
	 * Determines the Eclipse project containing the given resource, if any.
	 */
	// TODO Move to EclipseCheckoutFactory once
	// IResourceResoluationService.getContainingProject is no longer used in
	// CollectUploadItemsTask TS-31234
	public static Optional<IProject> getContainingIdeProject(ResourceReference reference) {
		Map<URI, IProject> uriToProject = new HashMap<>();
		Stream.of(ResourcesPlugin.getWorkspace().getRoot().findContainersForLocationURI(reference.toUri()))
				.map(IContainer::getProject) //
				.filter(Objects::nonNull) //
				.forEach(project -> uriToProject.put(project.getLocationURI().relativize(reference.toUri()), project));
		Optional<URI> closestMatchingURI = uriToProject.keySet().stream()
				.min(comparing(uri -> uri.toString().length()));
		return closestMatchingURI.map(uriToProject::get);
	}

	@Override
	public Optional<IIDEProject> getContainingProject(ResourceReference resource) {
		LOGGER.traceEntry(null, resource);
		Optional<IIDEProject> containingProject = getContainingIdeProject(resource).map(EclipseProject::new);
		return LOGGER.traceExit(containingProject);
	}

	@Override
	public Optional<String> readContents(ResourceReference resource) {
		LOGGER.traceEntry(null, resource);
		try {
			IFileStore store = EFS.getStore(resource.toUri());
			if (!store.fetchInfo().exists()) {
				return LOGGER.traceExit(Optional.empty());
			}

			ITextFileBufferManager manager = FileBuffers.getTextFileBufferManager();
			ITextFileBuffer fileBuffer;
			IFile file = store.getAdapter(IFile.class);
			if (file != null) {
				manager.connect(file.getFullPath(), LocationKind.IFILE, null);
				fileBuffer = manager.getTextFileBuffer(file.getFullPath(), LocationKind.IFILE);
			} else {
				manager.connectFileStore(store, null);
				fileBuffer = manager.getFileStoreTextFileBuffer(store);
			}

			try (Reader reader = new InputStreamReader(new BufferedInputStream(store.openInputStream(EFS.NONE, null)),
					fileBuffer.getEncoding())) {
				String contents = CharStreams.toString(reader);
				LOGGER.traceExit(contents.length()); // Don't log the full contents themselves
				return Optional.of(contents);
			}
		} catch (CoreException | IOException | IllegalArgumentException e) {
			LOGGER.catching(e);
			return LOGGER.traceExit(Optional.empty());
		}
	}

	@Override
	public long determineSize(ResourceReference resource) {
		LOGGER.traceEntry(null, resource);
		try {
			IFileStore store = EFS.getStore(resource.toUri());
			IFileInfo file = store.fetchInfo();
			if (file.exists()) {
				return LOGGER.traceExit(file.getLength());
			}
		} catch (CoreException e) {
			LOGGER.catching(e);
			return UNKNOWN_SIZE;
		}
		return LOGGER.traceExit(UNKNOWN_SIZE);
	}
}
