/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.plugin.services;

import java.text.MessageFormat;
import java.util.Objects;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.IJobChangeEvent;
import org.eclipse.core.runtime.jobs.ISchedulingRule;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.core.runtime.jobs.JobChangeAdapter;

import com.teamscale.ide.commons.plugin.services.ETaskVisibility;
import com.teamscale.ide.commons.plugin.services.ITaskExecutor;
import com.teamscale.ide.commons.plugin.services.TaskBase;
import com.teamscale.ide.eclipse.EclipseIds;

/**
 * Executes tasks using Eclipse's {@link Job} framework.
 * 
 * @see <a href=
 *      "https://www.eclipse.org/articles/Article-Concurrency/jobs-api.html">On
 *      the Job: The Eclipse Jobs API</a>
 */
public class EclipseTaskExecutor implements ITaskExecutor {

	private static final Logger LOGGER = LogManager.getLogger();

	@Override
	public void execute(TaskBase task) {
		LOGGER.traceEntry(null, task);
		switch (task.getInterruptionMode()) {
		case CANCEL:
			// Cancel all jobs from the same job family, i.e., with the same token/class
			// combination...
			Job.getJobManager().cancel(JobFamily.create(task));
			break;
		case WAIT:
			// Cancel waiting jobs from the same job family, i.e., with the same token/class
			// combination...
			for (Job job : Job.getJobManager().find(JobFamily.create(task))) {
				job.addJobChangeListener(new CancelWaitingJobListener());
			}
			break;
		default:
			throw new UnsupportedOperationException(task.getInterruptionMode().name());
		}
		// ...and schedule a new job that waits until the previous task with this
		// token/class combination has finished, as its scheduling rule is conflicting.
		new TaskJob(task).schedule();

		LOGGER.traceExit();
	}

	private static class TaskJob extends Job {

		private final TaskBase task;
		private final Object family;

		private TaskJob(TaskBase task) {
			super(task.getLabel());

			super.setRule(JobFamilySchedulingRule.create(task));

			super.setSystem(ETaskVisibility.INVISIBLE.equals(task.getVisibility()));
			// Shows a progress dialog unless the "General" > "Always run in background"
			// preference is set (which it is by default since Eclipse Oxygen (4.7).
			// See <https://bugs.eclipse.org/bugs/show_bug.cgi?id=499992> for discussion on
			// a reliable replacement.
			super.setUser(ETaskVisibility.OBVIOUS.equals(task.getVisibility()));

			this.task = task;
			this.family = JobFamily.create(task);
		}

		@Override
		protected IStatus run(IProgressMonitor monitor) {
			monitor.beginTask(task.getLabel(), IProgressMonitor.UNKNOWN);
			try {
				task.start();

				if (task.isCancelRequested()) {
					return Status.CANCEL_STATUS;
				}
				return Status.OK_STATUS;
			} catch (Exception e) {
				return new Status(IStatus.ERROR, EclipseIds.PLUGIN,
						MessageFormat.format("An unexpected error occurred while executing ''{0}''.", task.getLabel()),
						e);
			} finally {
				monitor.done();
			}
		}

		@Override
		protected void canceling() {
			task.cancel();
		}

		@Override
		public boolean belongsTo(Object family) {
			if (!(family instanceof JobFamily)) {
				return false;
			}
			return Objects.equals(this.family, family);
		}
	}

	/**
	 * A {@linkplain Job#belongsTo(Object)} job family} that considers two
	 * {@linkplain TaskBase tasks} to be in the same family iff they have same
	 * token/class combination.
	 */
	private static final class JobFamily {

		private final Class<? extends TaskBase> taskClass;
		private final Object token;

		private JobFamily(Class<? extends TaskBase> taskClass, Object token) {
			this.taskClass = taskClass;
			this.token = token;
		}

		@Override
		public boolean equals(Object other) {
			if (this == other) {
				return true;
			}
			if (other == null) {
				return false;
			}
			if (getClass() != other.getClass()) {
				return false;
			}
			JobFamily that = (JobFamily) other;
			return Objects.equals(this.taskClass, that.taskClass) && Objects.equals(this.token, that.token);
		}

		@Override
		public int hashCode() {
			return Objects.hash(taskClass, token);
		}

		private static JobFamily create(TaskBase task) {
			return new JobFamily(task.getClass(), task.getToken());
		}
	}

	/**
	 * A {@link ISchedulingRule} that considers two jobs to be
	 * {@linkplain #isConflicting(ISchedulingRule) conflicting} iff they are of the
	 * same {@linkplain Job#belongsTo(Object)} job family}, as indicated by the
	 * family {@code Object} passed to the constructor.
	 */
	private static class JobFamilySchedulingRule implements ISchedulingRule {

		/**
		 * The {@linkplain Job#belongsTo(Object)} job family} with which this rule is in
		 * conflict.
		 */
		private final JobFamily family;

		public JobFamilySchedulingRule(JobFamily family) {
			this.family = family;
		}

		@Override
		public boolean contains(ISchedulingRule other) {
			if (!(other instanceof JobFamilySchedulingRule)) {
				return false;
			}
			JobFamilySchedulingRule that = (JobFamilySchedulingRule) other;
			return Objects.equals(this.family, that.family);
		}

		@Override
		public boolean isConflicting(ISchedulingRule other) {
			return contains(other);
		}

		private static ISchedulingRule create(TaskBase task) {
			return new JobFamilySchedulingRule(JobFamily.create(task));
		}
	}

	/**
	 * A listener that will cancel jobs that are waiting, once they are about to
	 * run.
	 */
	private static class CancelWaitingJobListener extends JobChangeAdapter {
		@Override
		public void aboutToRun(IJobChangeEvent event) {
			event.getJob().cancel();
		}
	}
}
