/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.servers.services;

import java.net.URI;
import java.util.Collection;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;

import com.teamscale.ide.client.TeamscaleConnectionRegistry;
import com.teamscale.ide.commons.servers.AuthenticationInfo;
import com.teamscale.ide.commons.servers.NamedServerDetails;
import com.teamscale.ide.commons.servers.services.ITeamscaleServerManager;
import com.teamscale.ide.eclipse.util.LegacyConversionUtils;

/**
 * Manages a list of available Teamscale servers. The server list is stored
 * securely in the IDE's secure store.
 */
public class EclipseTeamscaleServerManager implements ITeamscaleServerManager {

	@Override
	public boolean hasServers() {
		return TeamscaleConnectionRegistry.getInstance().hasServers();
	}

	@Override
	public Collection<URI> getAllServerUrls() {
		return TeamscaleConnectionRegistry.getInstance().getServers().stream()
				.map(server -> URI.create(server.getUrl())).collect(Collectors.toList());
	}

	@Override
	public Optional<AuthenticationInfo> getAuthenticationInfo(URI url) {
		return TeamscaleConnectionRegistry.getInstance().getServers().stream()
				.filter(server -> Objects.equals(URI.create(server.getUrl()), url))
				.findFirst()
				.map(LegacyConversionUtils::toCommonsNamedServerDetails)
				.map(NamedServerDetails::asAuthenticationInfo);
	}

	@Override
	public boolean storeCredentials(URI url, AuthenticationInfo authenticationInfo) {
		throw new UnsupportedOperationException("Not yet implemented");
	}

	@Override
	public boolean removeServer(URI url) {
		throw new UnsupportedOperationException("Not yet implemented");
	}
	
	@Override
	public NamedServerDetails getNamedServerDetails(String name) {
		com.teamscale.ide.client.NamedServerDetails legacyServer = TeamscaleConnectionRegistry.getInstance()
				.getServer(name);
		if (legacyServer == null) {
			return null;
		}
		return LegacyConversionUtils.toCommonsNamedServerDetails(legacyServer);
	}
}
