/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.util.dialogs;

import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Link;
import org.eclipse.swt.widgets.Shell;

/**
 * A dialog for showing messages to the user which also contains a link.
 */
public class MessageDialogWithLink extends MessageDialog {

	private final String linkText;
	private final SelectionListener linkSelectionListener;

	/**
	 * Convenience method for opening a new error dialog containing a link.
	 * To render parts of the link text clickable like an HTML
	 * link, wrap the desired text in an &lt;a&gt; tag. This renders the text clickable
	 * so that the action defined in the selection listener is executed when the text
	 * is clicked on. Note that this tag is only for rendering and does not incur any
	 * (HTML-) functionality.
	 */
	public static MessageDialogWithLink openError(Shell parentShell, String title, String message, String linkText,
			SelectionListener linkSelectionListener) {
		return openMessage(parentShell, title, message, linkText, MessageDialog.ERROR, linkSelectionListener);
	}

	/**
	 * Convenience method for opening a new warning dialog containing a link.
	 * To render parts of the link text clickable like an HTML
	 * link, wrap the desired text in an &lt;a&gt; tag. This renders the text clickable
	 * so that the action defined in the selection listener is executed when the text
	 * is clicked on. Note that this tag is only for rendering and does not incur any
	 * (HTML-) functionality.
	 */
	public static MessageDialogWithLink openWarning(Shell parentShell, String title, String message, String linkText,
			SelectionListener linkSelectionListener) {
		return openMessage(parentShell, title, message, linkText, MessageDialog.WARNING, linkSelectionListener);
	}

	private static MessageDialogWithLink openMessage(Shell parentShell, String title, String message, String linkText,
			int dialogImageType, SelectionListener linkSelectionListener) {
		MessageDialogWithLink dialog = new MessageDialogWithLink(parentShell, title, null, message, dialogImageType,
				new String[] { IDialogConstants.OK_LABEL }, 0, linkText, linkSelectionListener);
		dialog.open();
		return dialog;
	}

	/**
	 * Extends the
	 * {@link MessageDialog#MessageDialog(Shell, String, Image, String, int, String[], int)}
	 * constructor by the two parameters needed for the link.
	 */
	public MessageDialogWithLink(Shell parentShell, String dialogTitle, Image dialogTitleImage, String dialogMessage,
			int dialogImageType, String[] dialogButtonLabels, int defaultIndex, String linkText,
			SelectionListener linkSelectionListener) {
		super(parentShell, dialogTitle, dialogTitleImage, dialogMessage, dialogImageType, dialogButtonLabels,
				defaultIndex);
		this.linkText = linkText;
		this.linkSelectionListener = linkSelectionListener;
	}

	@Override
	protected Control createCustomArea(Composite parent) {
		Link link = new Link(parent, SWT.NONE);
		link.setText(linkText);
		link.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				close();
				linkSelectionListener.widgetSelected(e);
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e) {
				close();
				linkSelectionListener.widgetDefaultSelected(e);
			}
		});
		return link;
	}
}
