/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.util.handlers;

import java.util.Optional;

import javax.inject.Inject;
import javax.inject.Named;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.eclipse.e4.core.contexts.IEclipseContext;
import org.eclipse.e4.core.di.annotations.CanExecute;
import org.eclipse.e4.core.di.annotations.Execute;
import org.eclipse.e4.core.services.adapter.Adapter;
import org.eclipse.e4.ui.services.IServiceConstants;

import com.teamscale.ide.eclipse.util.contexts.InjectionUtils;

/**
 * A handler base class that takes an (Eclipse 4) domain object selection, tries
 * to adapt it, and (if successful) delegates to an instance of
 * {@code handlerClass}.
 */
@SuppressWarnings("restriction")
public abstract class AdaptingSelectionHandlerBase<T> {

	// Logger is *not* static as package of *concrete* subclass must determine the
	// logging category.
	private final Logger logger = LogManager.getLogger(getClass());

	/**
	 * The handler class to which execution will be delegated. Must be an Eclipse 4
	 * handler, i.e., have methods executed with {@link Execute} and (optionally)
	 * {@link CanExecute}.
	 */
	private final Class<?> handlerClass;

	private final Class<T> adapterClass;

	@Inject
	private Adapter adapter;

	protected AdaptingSelectionHandlerBase(Class<?> handlerClass, Class<T> adapterClass) {
		this.handlerClass = handlerClass;
		this.adapterClass = adapterClass;
	}

	@Execute
	public void execute(@Named(IServiceConstants.ACTIVE_SELECTION) Object selection, IEclipseContext context) {
		logger.traceEntry(null, selection, context);
		getAdaptedSelection(selection) //
				.map(element -> InjectionUtils.makeAndInvokeWithSelection(handlerClass, Execute.class, context, element,
						null));
		logger.traceExit();
	}

	@CanExecute
	public boolean canExecute(@Named(IServiceConstants.ACTIVE_SELECTION) Object selection, IEclipseContext context) {
		logger.traceEntry(null, selection, context);
		boolean canExecute = getAdaptedSelection(selection) //
				.map(element -> (Boolean) InjectionUtils.makeAndInvokeWithSelection(handlerClass, CanExecute.class,
						context, element, Boolean.TRUE))
				.orElse(Boolean.FALSE);
		return logger.traceExit(canExecute);
	}

	protected Optional<T> getAdaptedSelection(Object selection) {
		return Optional.ofNullable(adapter.adapt(selection, adapterClass));
	}
}
