/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.util.handlers;

import static java.util.Objects.requireNonNull;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.eclipse.e4.core.contexts.ContextInjectionFactory;
import org.eclipse.e4.core.contexts.IEclipseContext;
import org.eclipse.e4.core.di.annotations.CanExecute;
import org.eclipse.e4.core.di.annotations.Execute;
import org.eclipse.e4.core.internal.contexts.ContextObjectSupplier;
import org.eclipse.e4.core.services.contributions.IContributionFactory;
import org.eclipse.e4.ui.model.application.MApplication;
import org.eclipse.e4.ui.model.application.commands.MCommand;
import org.eclipse.e4.ui.model.application.commands.MHandler;
import org.eclipse.e4.ui.model.application.commands.MHandlerContainer;
import org.eclipse.e4.ui.model.application.ui.MContext;
import org.eclipse.e4.ui.workbench.modeling.EModelService;

import com.google.common.base.Preconditions;

/**
 * A generic handler that acts as a dispatcher for other handlers.
 * <p>
 * This class solves two problems:
 * <ol>
 * <li>The user is not exposed to a plethora of individual commands (e.g., "Open
 * Finding in Web Browser", "Open Resource in Web Browser"), but sees only a
 * single, logical command (e.g., "Open in Web Browser") and can assign a single
 * keyboard shortcut to it.</li>
 * <li>Using different handlers for different selections below the {@code MPart}
 * level (e.g., in a part's tree viewer) becomes more manageable, as only a
 * single {@code MHandledItem} needs to be entered into the
 * {@code fragment.e4xmi}.</li>
 * </ol>
 * <p>
 * A new target handler can be added by passing its associated
 * {@link MCommand#getElementId() command ID} into the constructor of this
 * class. However, every registered target handler <strong>must</strong> have a
 * {@link CanExecute} method that limits the target handler's applicability.
 * Otherwise, a single target handler could displace all other target handlers.
 */
@SuppressWarnings("restriction")
public class DispatchingHandlerBase {

	// Logger is *not* static as package of *concrete* subclass must determine the
	// logging category.
	private final Logger logger = LogManager.getLogger(getClass());

	private final List<Object> registeredTargetHandlers = new ArrayList<>();

	/**
	 * Constructor searching for and instantiating all {@link MHandler}s with the
	 * given IDs.
	 * 
	 * @param targetCommands The IDs of the {@link MCommand}s which should be
	 *                       considered targets of this dispatcher.
	 * 
	 * @see org.eclipse.e4.ui.internal.workbench.addons.HandlerProcessingAddon#postConstruct(MApplication,
	 *      EModelService)
	 */
	public DispatchingHandlerBase(MApplication application, EModelService modelService, String... targetCommands) {
		requireNonNull(application);
		requireNonNull(modelService);
		for (String targetCommand : targetCommands) {
			Preconditions.checkArgument(application.getCommand(targetCommand) != null, "Command '%s' does not exist",
					targetCommand);
		}

		Set<String> targetCommandsSet = new HashSet<>(Arrays.asList(targetCommands));
		List<MHandlerContainer> handlerContainers = modelService.findElements(application, null,
				MHandlerContainer.class, null);
		for (MHandlerContainer handlerContainer : handlerContainers) {
			if (!(handlerContainer instanceof MContext)) {
				continue;
			}

			MContext context = (MContext) handlerContainer;
			IEclipseContext eclipseContext = context.getContext();
			if (eclipseContext == null) {
				continue;
			}
			for (MHandler handler : handlerContainer.getHandlers()) {
				if (!targetCommandsSet.contains(handler.getCommand().getElementId())) {
					continue;
				}
				Object targetHandler = handler.getObject();
				if (targetHandler == null) {
					IContributionFactory contributionFactory = eclipseContext.get(IContributionFactory.class);
					targetHandler = contributionFactory.create(handler.getContributionURI(), eclipseContext);
					handler.setObject(targetHandler);
				}
				registeredTargetHandlers.add(targetHandler);
			}
		}
	}

	/**
	 * Executes the registered target handler which
	 * {@linkplain #canExecute(IEclipseContext) can execute} with the the current
	 * {@linkplain IEclipseContext execution context}.
	 * <p>
	 * Using {@link ContextObjectSupplier} is a workaround for which no solution
	 * using non-internal API is known yet. See the corresponding <a href=
	 * "https://stackoverflow.com/questions/75335858/how-to-get-the-local-context-injected-in-an-eclipse-4-command-handler">StackOverflow
	 * question</a>.
	 */
	@Execute
	public void execute(IEclipseContext context, ContextObjectSupplier localContextSupplier) {
		logger.traceEntry(() -> context, () -> localContextSupplier.getContext());
		IEclipseContext localContext = localContextSupplier.getContext();
		getExecutableTargetHandler(context, localContext).ifPresent(targetHandler -> ContextInjectionFactory
				.invoke(targetHandler, Execute.class, context, localContext, null));
		logger.traceExit();
	}

	/**
	 * Checks whether one registered target handler
	 * {@linkplain #canExecute(IEclipseContext) can be executed} in the current
	 * {@linkplain IEclipseContext execution context}.
	 * <p>
	 * <strong>Note:</strong> that this handler does not detect when multiple
	 * conflicting target handlers can execute in the same context; it simply picks
	 * the first executable target handler.
	 * <p>
	 * Using {@link ContextObjectSupplier} is a workaround for which no solution
	 * using non-internal API is known yet. See the corresponding <a href=
	 * "https://stackoverflow.com/questions/75335858/how-to-get-the-local-context-injected-in-an-eclipse-4-command-handler">StackOverflow
	 * question</a>.
	 */
	@CanExecute
	public boolean canExecute(IEclipseContext context, ContextObjectSupplier localContextSupplier) {
		logger.traceEntry(() -> context, () -> localContextSupplier.getContext());
		IEclipseContext localContext = localContextSupplier.getContext();
		boolean canExecute = getExecutableTargetHandler(context, localContext).isPresent();
		return logger.traceExit(canExecute);
	}

	private Optional<Object> getExecutableTargetHandler(IEclipseContext context, IEclipseContext localContext) {
		for (Object targetHandler : registeredTargetHandlers) {
			logger.debug("Considering registered target handlers: {}", targetHandler);
			Boolean result = (Boolean) ContextInjectionFactory.invoke(targetHandler, CanExecute.class, context,
					localContext, Boolean.FALSE);

			// Be paranoid of null results here
			if (Boolean.TRUE.equals(result)) {
				logger.debug("Found target handler to execute: {}", targetHandler);
				return Optional.of(targetHandler);
			}
		}
		return Optional.empty();
	}
}
