package com.teamscale.ide.eclipse.util.markdown;

import static java.util.Objects.requireNonNull;

import org.commonmark.node.AbstractVisitor;
import org.commonmark.node.Code;
import org.commonmark.node.Emphasis;
import org.commonmark.node.StrongEmphasis;
import org.commonmark.node.Text;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StyleRange;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.graphics.Color;

import com.teamscale.ide.commons.util.markdown.MarkdownUtils;

/**
 * Utility class for appending Markdown styled text to {@link StyledText}.
 */
public class MarkdownTextStyler {

	private final String markdown;
	private final Color defaultForegroundColor;
	private final Color codeColor;

	public MarkdownTextStyler(String markdown, Color defaultForegroundColor, Color codeColor) {
		this.markdown = requireNonNull(markdown);
		this.defaultForegroundColor = requireNonNull(defaultForegroundColor);
		this.codeColor = requireNonNull(codeColor);
	}

	/**
	 * Parses the {@link #markdown} and appends it to the given {@link StyledText}
	 * as rendered Markdown.
	 */
	public void style(StyledText styledText) {
		MarkdownUtils.visitMarkdown(markdown, new MarkdownStyledTextVisitor(styledText));
	}

	/**
	 * Commonmark visitor to style a {@link StyledText} based on a Markdown string.
	 */
	private class MarkdownStyledTextVisitor extends AbstractVisitor {

		private StyleRange currentStyleRange = new StyleRange(0, 0, defaultForegroundColor, null, SWT.NORMAL);

		/** The {@link StyledText text} to style. */
		private final StyledText styledText;

		private MarkdownStyledTextVisitor(StyledText styledText) {
			this.styledText = styledText;
		}

		@Override
		public void visit(Text text) {
			appendText(text.getLiteral());
		}

		@Override
		public void visit(Code code) {
			appendText(code.getLiteral(), codeColor);
		}

		@Override
		public void visit(StrongEmphasis strongEmphasis) {
			withStyle(SWT.BOLD, () -> super.visit(strongEmphasis));
		}

		@Override
		public void visit(Emphasis emphasis) {
			withStyle(SWT.ITALIC, () -> super.visit(emphasis));
		}

		/**
		 * Appends the given text.
		 */
		void appendText(String text) {
			appendText(text, currentStyleRange.foreground, currentStyleRange.background, currentStyleRange.fontStyle);
		}

		/**
		 * Appends the given text and applies the given foreground color.
		 */
		void appendText(String text, Color foregroundColor) {
			appendText(text, foregroundColor, currentStyleRange.background, currentStyleRange.fontStyle);
		}

		/**
		 * Appends the given text and applies the given foreground color, background
		 * color as well as the font style.
		 */
		void appendText(String text, Color foregroundColor, Color backgroundColor, int fontStyle) {
			int start = styledText.getCharCount();
			int textLength = text.length();
			styledText.append(text);
			styledText.setStyleRange(new StyleRange(start, textLength, foregroundColor, backgroundColor, fontStyle));
		}

		private void withStyle(int fontStyle, Runnable body) {
			StyleRange savedStyleRange = currentStyleRange;
			currentStyleRange = new StyleRange(currentStyleRange.start, currentStyleRange.length,
					currentStyleRange.foreground, currentStyleRange.background,
					currentStyleRange.fontStyle | fontStyle);

			body.run();

			currentStyleRange = savedStyleRange;
		}
	}
}
