package com.teamscale.ide.eclipse.util.popup;

import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.layout.GridDataFactory;
import org.eclipse.jface.notifications.AbstractNotificationPopup;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Monitor;
import org.eclipse.swt.widgets.Shell;

/**
 * A simple notification popup with a text.
 */
public class MessagePopup extends AbstractNotificationPopup {

	/**
	 * Displays a popup with the given title and text.
	 */
	public static MessagePopup show(Display display, String title, String text) {
		MessagePopup popup = new MessagePopup(display, title, text);
		popup.open();
		return popup;
	}
	
	private final String title;
	
	private final String text;

	public MessagePopup(Display display, String title, String text) {
		super(display);
		this.title = title;
		this.text = text;
		setParentShell(display.getActiveShell());
	}

	@Override
	protected String getPopupShellTitle() {
		return title;
	}
	
	@Override
	protected void initializeBounds() {
		/*
		 * By default, the popup is put in the bottom right corner of the
		 * current display. If the display is not the root Eclipse pane, but
		 * e.g. the server settings page, this might cause a blocking behavior,
		 * as buttons will be overlayed. This should not be the case. Hence, we
		 * need to override initializeBounds to place the popup in the very bottom
		 * right corner of the Eclipse root view independent of the given display.
		 */
		super.initializeBounds();
		
		int paddingX = IDialogConstants.HORIZONTAL_MARGIN;
		int paddingY = IDialogConstants.VERTICAL_MARGIN;
		
		// taken from AbstractNotificationPopup.MIN_HEIGHT
		int minHeight = 100;
		// taken from AbstractNotificationPopup.MAX_WIDTH
		int maxWidth = 400;
		Rectangle clArea = getPrimaryClientArea();
		Point initialSize = getShell().computeSize(SWT.DEFAULT, SWT.DEFAULT);
		int height = Math.max(initialSize.y, minHeight);
		int width = Math.min(initialSize.x, maxWidth);

		Point size = new Point(width, height);
		getShell().setLocation(clArea.width + clArea.x - size.x - paddingX,
				clArea.height + clArea.y - size.y - paddingY);
	}
	
	/** 
	 * See {@link AbstractNotificationPopup#getPrimaryClientArea}
	 */
	private Rectangle getPrimaryClientArea() {
		Shell parentShell = getParentShell();
		// Unlike in AbstractNotificationPopup, we do not want the Rectangle 
		// based on the given display shell but based on the root Eclipse view
		while (parentShell != null && parentShell.getParent() != null) {
			parentShell = (Shell) parentShell.getParent();
		}
		
		if (parentShell != null) {
			// calculate client area in display-relative coordinates
			// (i.e. without window border / decorations)
			Rectangle bounds = parentShell.getBounds();
			Rectangle trim = parentShell.computeTrim(0, 0, 0, 0);
			return new Rectangle(bounds.x - trim.x, bounds.y - trim.y, bounds.width - trim.width,
					bounds.height - trim.height);
		}
		// else display on primary monitor
		Monitor primaryMonitor = getShell().getDisplay().getPrimaryMonitor();
		if (primaryMonitor == null) {
			return getShell().getDisplay().getClientArea();
		}
		return primaryMonitor.getClientArea();
	}

	@Override
	protected void createContentArea(Composite parent) {
		Label label = new Label(parent, SWT.WRAP);
		label.setText(text);
		GridDataFactory.fillDefaults().grab(true, true).hint(IDialogConstants.MINIMUM_MESSAGE_AREA_WIDTH, SWT.DEFAULT).applyTo(label);
	}
}
