/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.util.popup;

import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.layout.GridDataFactory;
import org.eclipse.jface.layout.GridLayoutFactory;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Link;

/**
 * A simple notification popup with a text and a link.
 */
public class MessagePopupWithLink extends MessagePopup {
	private final String text;
	private final String linkText;
	private final SelectionListener linkSelectionListener;

	/**
	 * Displays a popup with the given title, text and link in a custom area at the
	 * bottom of the popup. To render parts of the link text clickable like an HTML
	 * link, wrap the desired text in an &lt;a&gt; tag. This renders the text clickable
	 * so that the action defined in the selection listener is executed when the text
	 * is clicked on. Note that this tag is only for rendering and does not incur any
	 * (HTML-) functionality.
	 */
	public static MessagePopupWithLink show(Display display, String title, String text, String linkText, SelectionListener linkSelectionListener) {
		MessagePopupWithLink popup = new MessagePopupWithLink(display, title, text, linkText, linkSelectionListener);
		popup.open();
		return popup;
	}

	public MessagePopupWithLink(Display display, String title, String text, String linkText, SelectionListener linkSelectionListener) {
		super(display, title, text);
		this.text = text;
		this.linkText = linkText;
		this.linkSelectionListener = linkSelectionListener;
		setParentShell(display.getActiveShell());
	}

	@Override
	protected void createContentArea(Composite parent) {
		Composite container = new Composite(parent, SWT.NONE);
		GridDataFactory.fillDefaults().grab(true, true).hint(IDialogConstants.MINIMUM_MESSAGE_AREA_WIDTH, SWT.DEFAULT).applyTo(container);
	    GridLayoutFactory.fillDefaults().applyTo(container);
	    
		Label label = new Label(container, SWT.WRAP);
		label.setText(text);
		GridDataFactory.fillDefaults().grab(true, true).applyTo(label);
		
		Link link = new Link(container, SWT.WRAP);
		link.setText(linkText);
		GridDataFactory.fillDefaults().grab(true, true).applyTo(link);
		link.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				close();
				linkSelectionListener.widgetSelected(e);
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e) {
				close();
				linkSelectionListener.widgetDefaultSelected(e);
			}
		});
	}
	
}

