package com.teamscale.ide.eclipse.util.viewers;

import java.util.Optional;

import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.ui.model.IWorkbenchAdapter;

/**
 * Utilities for working with {@link IAdaptable}s.
 */
public class AdaptableUtils {

	/**
	 * Returns the first matching object that can be adapted from a given
	 * {@link IAdaptable} element to the given type.
	 * 
	 * @param adaptableElement
	 *            the element to be adapted from. May be {@code null}.
	 * @param clazz
	 *            the class the element should be adapted to.
	 * @param tryAncestors
	 *            whether to try to adapt the {@code adaptableElement}'s ancestors
	 *            to the given class, if it cannot be adapted itself.
	 */
	public static <T> Optional<T> getMatchingAdapter(IAdaptable adaptableElement, Class<T> clazz,
			boolean tryAncestors) {
		if (adaptableElement == null) {
			return Optional.empty();
		}
		T adapted = adaptableElement.getAdapter(clazz);
		if (adapted != null) {
			return Optional.of(adapted);
		}
		if (!tryAncestors) {
			return Optional.empty();
		}
		Optional<IAdaptable> parent = getParent(adaptableElement);
		if (!parent.isPresent()) {
			return Optional.empty();
		}
		return getMatchingAdapter(parent.get(), clazz, tryAncestors);
	}

	/**
	 * Returns the parent of a given {@link IAdaptable} iff the parent can be casted
	 * to an {@link IAdaptable}. Otherwise returns {@link Optional#empty()}
	 */
	private static Optional<IAdaptable> getParent(IAdaptable adaptableElement) {
		IWorkbenchAdapter workbenchAdapter = adaptableElement.getAdapter(IWorkbenchAdapter.class);
		if (workbenchAdapter == null) {
			return Optional.empty();
		}
		return Optional.ofNullable(workbenchAdapter.getParent(adaptableElement)).filter(IAdaptable.class::isInstance)
				.map(IAdaptable.class::cast);
	}

}
