/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.vcs.services;

import java.util.Optional;

import org.eclipse.core.resources.IProject;
import org.eclipse.team.core.RepositoryProvider;

import com.teamscale.ide.commons.model.IIDEProject;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.vcs.CheckoutException;
import com.teamscale.ide.commons.vcs.ICheckout;
import com.teamscale.ide.commons.vcs.services.ISupportedCheckout;
import com.teamscale.ide.eclipse.plugin.services.EclipseResourceResolutionService;
import com.teamscale.ide.eclipse.vcs.subclipse.EclipseSubclipseCheckout;
import com.teamscale.ide.eclipse.vcs.subversive.EclipseSubversiveCheckout;

/** Eclipse-specific checkout implementations. */
public enum EclipseSupportedCheckouts implements ISupportedCheckout {

	/**
	 * An implementation of a Subversion {@link ICheckout} using the Subclipse
	 * Eclipse plug-in.
	 */
	SUBCLIPSE(EclipseSubclipseCheckout.REPOSITORY_PROVIDER_ID, "SVN (using Subclipse)") {

		@Override
		protected EclipseSubclipseCheckout newCheckout(IProject eclipseProject) throws CheckoutException {
			return new EclipseSubclipseCheckout(eclipseProject);
		}
	},

	/**
	 * An implementation of a Subversion {@link ICheckout} using the Eclipse
	 * Subversive plug-in.
	 */
	SUBVERSIVE(EclipseSubversiveCheckout.REPOSITORY_PROVIDER_ID, "SVN (using Subversive)") {

		@Override
		protected EclipseSubversiveCheckout newCheckout(IProject eclipseProject) throws CheckoutException {
			return new EclipseSubversiveCheckout(eclipseProject);
		}
	};

	private final String providerId;
	private final String displayName;

	EclipseSupportedCheckouts(String providerId, String displayName) {
		this.providerId = providerId;
		this.displayName = displayName;
	}

	@Override
	public String getDisplayName() {
		return displayName;
	}

	/**
	 * Attempts to open a {@link ICheckout} for the given resource.
	 * <p>
	 * This method <strong>must</strong> ensure that no classes from plug-ins
	 * declared <em>optional</em> in the {@code META-INF/MANIFEST.MF} are loaded
	 * until it has been ensured that the respective plug-in (e.g., Subclipse,
	 * Subversive) is actually installed. Otherwise, {@code ClassNotFoundException}s
	 * ensue. Hence, if checks for the provider ID (a {@code String}) first.
	 */
	@Override
	public Optional<ICheckout> openCheckout(ResourceReference resource) throws CheckoutException {
		IProject eclipseProject = EclipseResourceResolutionService.getContainingIdeProject(resource).orElse(null);
		if (eclipseProject == null) {
			return Optional.empty();
		}
		if (RepositoryProvider.getProvider(eclipseProject, providerId) == null) {
			return Optional.empty();
		}
		return Optional.of(newCheckout(eclipseProject));
	}

	/**
	 * Create a new {@link Checkout} given an {@linkplain IProject Eclipse project}.
	 */
	protected abstract ICheckout newCheckout(IProject eclipseProject) throws CheckoutException;
}
