/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.vcs.subclipse;

import static java.util.Arrays.stream;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.Path;
import org.tigris.subversion.subclipse.core.ISVNRepositoryLocation;
import org.tigris.subversion.subclipse.core.SVNException;
import org.tigris.subversion.subclipse.core.resources.SVNWorkspaceRoot;
import org.tigris.subversion.svnclientadapter.ISVNClientAdapter;
import org.tigris.subversion.svnclientadapter.ISVNInfo;
import org.tigris.subversion.svnclientadapter.ISVNStatus;
import org.tigris.subversion.svnclientadapter.SVNClientException;
import org.tigris.subversion.svnclientadapter.SVNRevision;
import org.tigris.subversion.svnclientadapter.utils.SVNStatusUtils;

import com.teamscale.ide.commons.lang.ToStringHelpers;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.util.UriPathUtils;
import com.teamscale.ide.commons.util.UriUtils;
import com.teamscale.ide.commons.vcs.ChangeSet;
import com.teamscale.ide.commons.vcs.CheckoutException;
import com.teamscale.ide.commons.vcs.ICheckout;
import com.teamscale.ide.commons.vcs.RevisionDescriptor;
import com.teamscale.ide.commons.vcs.svn.SubversionCheckoutBase;
import com.teamscale.ide.commons.vcs.svn.SubversionUtils;

/**
 * A {@link ICheckout} based on the Subclipse Eclipse plug-in.
 * <p>
 * Its {@link #checkoutRoot} is always an Eclipse workspace resource, as
 * Subclipse (unlike Subversive) cannot operate on directories outside the
 * Eclipse workspace.
 */
public class EclipseSubclipseCheckout extends SubversionCheckoutBase {

	/** Repository provider ID for Subclipse. */
	public static final String REPOSITORY_PROVIDER_ID = "org.tigris.subversion.subclipse.core.svnnature";

	private static final Logger LOGGER = LogManager.getLogger();

	private final File checkoutRootDir;

	private final ResourceReference checkoutRoot;

	public EclipseSubclipseCheckout(IProject project) throws CheckoutException {
		LOGGER.traceEntry(null, project);
		try {
			File projectDir = new File(project.getLocationURI());
			LOGGER.debug("Project directory: {}", projectDir);

			ISVNClientAdapter client = getClient(projectDir).orElseThrow(() -> new CheckoutException(
					String.format("Could not get SVN client for project directory %s", projectDir)));

			ISVNInfo info = client.getInfoFromWorkingCopy(projectDir);
			if (info == null) {
				throw LOGGER.throwing(new CheckoutException(
						String.format("SVN info unexpectedly failed for project directory %s", projectDir)));
			}

			URI projectUri = new URI(info.getUrl().toString());
			LOGGER.debug("Project URI: <{}>)", projectUri);

			URI repositoryRootUri = new URI(info.getRepository().toString());
			LOGGER.debug("Repository root URI: <{}>", repositoryRootUri);

			checkoutRootDir = determineCheckoutRoot(projectDir, projectUri, repositoryRootUri);
			LOGGER.debug("Checkout directory: {})", checkoutRootDir);

			checkoutRoot = ResourceReference.forFolder(checkoutRootDir.toURI());

			LOGGER.traceExit();
		} catch (SVNException | SVNClientException | URISyntaxException e) {
			throw LOGGER.throwing(new CheckoutException("Could not create Sublipse checkout", e));
		} catch (CheckoutException e) {
			throw LOGGER.throwing(e);
		}
	}

	private static File determineCheckoutRoot(File projectDir, URI projectUri, URI repositoryRootUri) {
		URI branchRootUri = SubversionUtils.getBranchRoot(repositoryRootUri, projectUri).orElse(repositoryRootUri);
		LOGGER.debug("Branch root URI: <{}>", branchRootUri);

		int directoriesToSkip = UriUtils.getPathSegments(projectUri).size()
				- UriUtils.getPathSegments(branchRootUri).size();

		File checkoutDir = projectDir;
		while (directoriesToSkip > 0) {
			File parentDir = checkoutDir.getParentFile();
			if (!getRepository(parentDir).isPresent()) {
				LOGGER.debug("Hit directory not managed by Subclipse: {}", parentDir);
				// Parent directory not managed by Subclipse; hence, it cannot become our
				// checkout root, as we wouldn't be able to obtain an client for it.
				break;
			}
			checkoutDir = parentDir;
			directoriesToSkip--;
		}
		return checkoutDir;
	}

	private static Optional<ISVNClientAdapter> getClient(File resource) throws SVNException {
		ISVNRepositoryLocation repository = getRepository(resource).orElse(null);
		if (repository == null) {
			return Optional.empty();
		}
		return Optional.of(repository.getSVNClient());
	}

	/**
	 * Get the Subclipse-managed SVN repository {@code file} is in, if any.
	 * 
	 * @param resource Must be an Eclipse workspace resource; Subclipse doesn't work
	 *                 for files outside Eclipse projects, even if they are in an
	 *                 SVN checkout.
	 */
	private static Optional<ISVNRepositoryLocation> getRepository(File resource) {
		return Optional.ofNullable(SVNWorkspaceRoot.getRepositoryFor(Path.fromOSString(resource.toString())));
	}

	@Override
	public ResourceReference getCheckoutRoot() {
		return checkoutRoot;
	}

	@Override
	public Optional<String> getCheckedOutBranch() throws CheckoutException {
		LOGGER.traceEntry();
		try {
			ISVNClientAdapter client = getClient(checkoutRootDir).orElseThrow(() -> new CheckoutException(
					String.format("Could not get SVN client for checkout root %s", checkoutRootDir)));

			ISVNInfo info = client.getInfoFromWorkingCopy(checkoutRootDir);
			if (info == null) {
				throw LOGGER.throwing(new CheckoutException(
						String.format("SVN info unexpectedly failed for checkout root %s", checkoutRootDir)));
			}

			URI checkoutRootUri = new URI(info.getUrl().toString());
			LOGGER.debug("Checkout root URI: <{}>)", checkoutRootUri);

			URI remoteRepositoryRoot = new URI(info.getUrl().toString());
			LOGGER.debug("Repository root URI: <{}>", remoteRepositoryRoot);

			return LOGGER.traceExit(SubversionUtils.guessBranchName(remoteRepositoryRoot, checkoutRootUri));
		} catch (SVNException | SVNClientException | URISyntaxException e) {
			throw LOGGER.throwing(new CheckoutException("Could not get primary branch", e));
		} catch (CheckoutException e) {
			throw LOGGER.throwing(e);
		}
	}

	@Override
	protected List<RevisionDescriptor> doReturnMostRecentRevisionDescriptor() throws CheckoutException {
		LOGGER.traceEntry();
		List<RevisionDescriptor> mostRecentRevisions = new LinkedList<>();
		try {
			ISVNClientAdapter client = getClient(checkoutRootDir).orElseThrow(() -> new CheckoutException(
					String.format("Could not get SVN client for checkout root %s", checkoutRootDir)));

			ISVNStatus[] statuses = client.getStatus(checkoutRootDir, /* recursive */ true, /* includeUnchanged */ true,
					/* contactRemote */ false, /* ignoreExternals */ true);
			long mostRecentRevision = stream(statuses) //
					.map(ISVNStatus::getRevision) //
					.filter(Objects::nonNull) //
					.mapToLong(SVNRevision.Number::getNumber) //
					// There is at least one directory under version control, namely checkoutRoot,
					// so there is at least one status.
					.max().getAsLong();
			String revisionIdentifier = Long.toString(mostRecentRevision);
			mostRecentRevisions.add(new RevisionDescriptor(revisionIdentifier));
			return LOGGER.traceExit(mostRecentRevisions);
		} catch (SVNException | SVNClientException e) {
			throw LOGGER.throwing(new CheckoutException("Could not get most recent revisions", e));
		} catch (CheckoutException e) {
			throw LOGGER.throwing(e);
		}
	}

	@Override
	public ChangeSet getUncommittedChanges() throws CheckoutException {
		throw new UnsupportedOperationException("Not needed by Legacy Pre-Commit");
	}

	@Override
	public ChangeSet getChanges(RevisionDescriptor sinceRevision) throws CheckoutException {
		LOGGER.traceEntry(null, sinceRevision);
		try {
			ISVNClientAdapter client = getClient(checkoutRootDir).orElseThrow(() -> new CheckoutException(
					String.format("Could not get SVN client for checkout root %s", checkoutRootDir)));

			ISVNStatus[] statuses = client.getStatus(checkoutRootDir, /* recursive */ true, /* includeUnchanged */ true,
					/* contactRemote */ true, /* ignoreExternals */ true);

			Set<ResourceReference> createdFiles = new HashSet<>();
			Set<ResourceReference> modifiedFiles = new HashSet<>();
			Set<ResourceReference> removedFiles = new HashSet<>();

			for (ISVNStatus status : statuses) {
				LOGGER.trace("Considering {} (text status: {}, property status: {})", status.getPath(),
						status.getTextStatus(), status.getPropStatus());

				if (SVNStatusUtils.isIgnored(status)) {
					continue;
				}

				String path = status.getPath();
				if (!Files.isRegularFile(Paths.get(path))) {
					continue;
				}
				// Subclipse always uses slashes as path separator, even on Windows, so using
				// UriPathUtils is safe.
				String uriPath = UriPathUtils.toAbsoluteUriPath(path);
				ResourceReference resourceReference = ResourceReference.forFile(URI.create("file://" + uriPath));
				if (!SVNStatusUtils.isManaged(status)) {
					createdFiles.add(resourceReference);
				} else if (SVNStatusUtils.isDeleted(status)) {
					removedFiles.add(resourceReference);
				} else if (SVNStatusUtils.isReadyForCommit(status)) {
					modifiedFiles.add(resourceReference);
				}
			}

			return LOGGER.traceExit(new ChangeSet(createdFiles, modifiedFiles, removedFiles));
		} catch (SVNException | SVNClientException e) {
			throw LOGGER.throwing(new CheckoutException("Could not get changes", e));
		} catch (CheckoutException e) {
			throw LOGGER.throwing(e);
		}
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		EclipseSubclipseCheckout that = (EclipseSubclipseCheckout) other;
		return Objects.equals(this.checkoutRoot, that.checkoutRoot);
	}

	@Override
	public int hashCode() {
		return Objects.hash(checkoutRoot);
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
