/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.vcs.subversive;

import static java.util.Arrays.stream;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.team.svn.core.SVNTeamPlugin;
import org.eclipse.team.svn.core.connector.SVNChangeStatus;
import org.eclipse.team.svn.core.connector.SVNEntryInfo;
import org.eclipse.team.svn.core.operation.AbstractActionOperation;
import org.eclipse.team.svn.core.operation.IActionOperation;
import org.eclipse.team.svn.core.operation.file.LocalStatusOperation;
import org.eclipse.team.svn.core.operation.file.RemoteStatusOperation;
import org.eclipse.team.svn.core.utility.ProgressMonitorUtility;
import org.eclipse.team.svn.core.utility.SVNUtility;

import com.teamscale.ide.commons.lang.ToStringHelpers;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.util.UriPathUtils;
import com.teamscale.ide.commons.util.UriUtils;
import com.teamscale.ide.commons.vcs.ChangeSet;
import com.teamscale.ide.commons.vcs.CheckoutException;
import com.teamscale.ide.commons.vcs.ICheckout;
import com.teamscale.ide.commons.vcs.RevisionDescriptor;
import com.teamscale.ide.commons.vcs.svn.SubversionCheckoutBase;
import com.teamscale.ide.commons.vcs.svn.SubversionUtils;

/**
 * A {@link ICheckout} based on the <a
 * href="https://www.eclipse.org/subversive/>Subversive</a> Eclipse plug-in.
 */
public class EclipseSubversiveCheckout extends SubversionCheckoutBase {

	/** Repository provider ID for Subversive. */
	public static final String REPOSITORY_PROVIDER_ID = SVNTeamPlugin.NATURE_ID;

	private static final Logger LOGGER = LogManager.getLogger();

	private final File checkoutRootDir;
	private ResourceReference checkoutRoot;

	public EclipseSubversiveCheckout(IProject project) throws CheckoutException {
		LOGGER.traceEntry(null, project);
		try {
			File projectDir = new File(project.getLocationURI());
			LOGGER.debug("Project directory: {}", projectDir);

			SVNEntryInfo info = getSvnInfo(projectDir);
			if (!info.hasWcInfo) {
				throw LOGGER.throwing(new CheckoutException(
						String.format("SVN info has no working-copy information for directory %s", projectDir)));
			}

			URI projectUri = new URI(info.url);
			LOGGER.debug("Project URI: <{}>)", projectUri);

			File workingCopyDir = new File(info.wcRoot);
			LOGGER.debug("Working copy directory: {}", workingCopyDir);

			URI repositoryRootUri = new URI(info.reposRootUrl);
			LOGGER.debug("Repository root URI: <{}>", repositoryRootUri);

			checkoutRootDir = determineCheckoutRoot(projectDir, projectUri, workingCopyDir, repositoryRootUri);
			LOGGER.debug("Checkout directory: {})", checkoutRootDir);

			checkoutRoot = ResourceReference.forFolder(checkoutRootDir.toURI());

			LOGGER.traceExit();
		} catch (URISyntaxException e) {
			throw LOGGER.throwing(new CheckoutException(e));
		} catch (CheckoutException e) {
			throw LOGGER.throwing(e);
		}
	}

	private static File determineCheckoutRoot(File projectDir, URI projectUri, File workingCopyDir,
			URI repositoryRootUri) {
		URI branchRootUri = SubversionUtils.getBranchRoot(repositoryRootUri, projectUri).orElse(repositoryRootUri);
		LOGGER.debug("Determined branch root URI: <{}>", branchRootUri);

		int directoriesToSkip = UriUtils.getPathSegments(projectUri).size()
				- UriUtils.getPathSegments(branchRootUri).size();

		File checkoutDir = projectDir;
		while (directoriesToSkip > 0) {
			if (checkoutDir.equals(workingCopyDir)) {
				LOGGER.debug("Hit working copy directory: {}", workingCopyDir);
				// The checked out working copy does not include the entire branch, so the
				// checkout root is below the branch root.
				break;
			}
			checkoutDir = checkoutDir.getParentFile();
			directoriesToSkip--;
		}
		return checkoutDir;
	}

	private static SVNEntryInfo getSvnInfo(File directory) throws CheckoutException {
		SVNEntryInfo info = SVNUtility.getSVNInfo(directory);
		if (info == null) {
			throw LOGGER.throwing(
					new CheckoutException(String.format("SVN info unexpectedly failed for directory %s", directory)));
		}
		return info;
	}

	@Override
	public ResourceReference getCheckoutRoot() {
		return checkoutRoot;
	}

	@Override
	public Optional<String> getCheckedOutBranch() throws CheckoutException {
		LOGGER.traceEntry();
		try {
			SVNEntryInfo info = getSvnInfo(checkoutRootDir);

			URI checkoutRootUri = new URI(info.url);
			LOGGER.debug("Checkout root URI: <{}>)", checkoutRootUri);

			URI repositoryRootUri = new URI(info.reposRootUrl);
			LOGGER.debug("Repository root URI: <{}>", repositoryRootUri);

			return LOGGER.traceExit(SubversionUtils.guessBranchName(repositoryRootUri, checkoutRootUri));
		} catch (URISyntaxException e) {
			throw LOGGER.throwing(new CheckoutException(e));
		} catch (CheckoutException e) {
			throw LOGGER.throwing(e);
		}
	}

	@Override
	protected List<RevisionDescriptor> doReturnMostRecentRevisionDescriptor() throws CheckoutException {
		LOGGER.traceEntry();
		List<RevisionDescriptor> mostRecentRevisions = new LinkedList<>();
		try {
			LocalStatusOperation operation = executeOperation(
					new EnhancedLocalStatusOperation(new File[] { checkoutRootDir }, true, true));
			long mostRecentRevision = stream(operation.getStatuses()) //
					.mapToLong(status -> status.revision) //
					// There is at least one directory under version control, namely checkoutRoot,
					// so there is at least one status.
					.max().getAsLong();
			String revisionIdentifier = Long.toString(mostRecentRevision);
			mostRecentRevisions.add(new RevisionDescriptor(revisionIdentifier));
			return LOGGER.traceExit(mostRecentRevisions);
		} catch (CheckoutException e) {
			throw LOGGER.throwing(e);
		}
	}

	@Override
	public ChangeSet getUncommittedChanges() throws CheckoutException {
		throw new UnsupportedOperationException("Not needed by Legacy Pre-Commit");
	}

	@Override
	public ChangeSet getChanges(RevisionDescriptor sinceRevision) throws CheckoutException {
		LOGGER.traceEntry(null, sinceRevision);
		try {
			RemoteStatusOperation operation = executeOperation(
					new RemoteStatusOperation(new File[] { checkoutRootDir }, true));

			if (operation.getStatuses() == null) {
				return LOGGER.traceExit(ChangeSet.EMPTY);
			}

			Set<ResourceReference> createdFiles = new HashSet<>();
			Set<ResourceReference> modifiedFiles = new HashSet<>();
			Set<ResourceReference> removedFiles = new HashSet<>();

			for (SVNChangeStatus status : operation.getStatuses()) {
				LOGGER.trace("Considering {} (text status: {}, property status: {})", status.path, status.textStatus,
						status.propStatus);

				if (SubversiveStatusUtils.isIgnored(status)) {
					continue;
				}

				String path = status.path;
				if (!Files.isRegularFile(Paths.get(path))) {
					continue;
				}
				// Subversive always uses slashes as path separator, even on Windows, so using
				// UriPathUtils is safe.
				String uriPath = UriPathUtils.toAbsoluteUriPath(path);
				ResourceReference resourceReference = ResourceReference.forFile(URI.create("file://" + uriPath));
				if (!SubversiveStatusUtils.isManaged(status)) {
					createdFiles.add(resourceReference);
				} else if (SubversiveStatusUtils.isDeleted(status)) {
					removedFiles.add(resourceReference);
				} else if (SubversiveStatusUtils.isReadyForCommit(status)) {
					modifiedFiles.add(resourceReference);
				}
			}

			return LOGGER.traceExit(new ChangeSet(createdFiles, modifiedFiles, removedFiles));
		} catch (CheckoutException e) {
			throw LOGGER.throwing(e);
		}
	}

	private static <T extends AbstractActionOperation> T executeOperation(T operation) throws CheckoutException {
		ProgressMonitorUtility.doTaskExternal(operation, null);
		if (operation.getExecutionState() != IActionOperation.OK) {
			IStatus status = operation.getStatus();
			throw new CheckoutException(status.getMessage(), status.getException());
		}
		return operation;
	}

	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		EclipseSubversiveCheckout that = (EclipseSubversiveCheckout) other;
		return Objects.equals(this.checkoutRoot, that.checkoutRoot);
	}

	@Override
	public int hashCode() {
		return Objects.hash(checkoutRoot);
	}

	@Override
	public String toString() {
		return ToStringHelpers.toReflectiveStringHelper(this).toString();
	}
}
