/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.vcs.subversive;

import java.io.File;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.team.svn.core.connector.ISVNConnector;
import org.eclipse.team.svn.core.connector.ISVNEntryStatusCallback;
import org.eclipse.team.svn.core.connector.ISVNProgressMonitor;
import org.eclipse.team.svn.core.connector.SVNDepth;
import org.eclipse.team.svn.core.operation.file.LocalStatusOperation;

/**
 * An enhanced version of {@link LocalStatusOperation}, which unfortunately does
 * not allow to retrieve the local status of unmodified files, i.e.,
 * {@code svn status --verbose}.
 */
/* package */ class EnhancedLocalStatusOperation extends LocalStatusOperation {

	private static final Logger LOGGER = LogManager.getLogger();

	private final boolean includeUnchanged;

	/* package */ EnhancedLocalStatusOperation(File[] files, boolean recursive, boolean includeUnchanged) {
		super(files, recursive);
		this.includeUnchanged = includeUnchanged;
	}

	@Override
	protected void reportStatuses(ISVNConnector delegate, ISVNEntryStatusCallback callback, File current,
			IProgressMonitor monitor, int tasks) {
		// ISVNConnector is an ugly interface to wrap as it has dozens of methods.
		// Hence, we use reflection and a dynamic proxy here.
		Method statusMethod = getMethod(ISVNConnector.class, "status", String.class, SVNDepth.class, Long.TYPE,
				String[].class, ISVNEntryStatusCallback.class, ISVNProgressMonitor.class);
		ISVNConnector proxy = (ISVNConnector) Proxy.newProxyInstance(getClass().getClassLoader(),
				new Class<?>[] { ISVNConnector.class }, new InvocationHandler() {

					private static final int OPTIONS_INDEX = 2;

					@Override
					public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
						if (method.equals(statusMethod) && includeUnchanged) {
							long options = (long) args[OPTIONS_INDEX];
							options |= ISVNConnector.Options.INCLUDE_UNCHANGED;
							args[OPTIONS_INDEX] = options;
						}
						return method.invoke(delegate, args);
					}
				});
		super.reportStatuses(proxy, callback, current, monitor, tasks);
	}

	private static Method getMethod(Class<?> declaringClass, String methodName, Class<?>... parameterTypes) {
		try {
			return declaringClass.getMethod(methodName, parameterTypes);
		} catch (NoSuchMethodException | SecurityException e) {
			throw LOGGER.throwing(new AssertionError("Unexpected Subversive API change", e));
		}
	}
}
