/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.findings;

import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.conqat.ide.commons.ui.editor.EditorUtils;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchPage;

import com.google.common.base.Objects;
import com.sap.adt.activation.ui.IActivationSuccessListener;
import com.sap.adt.tools.abapsource.sourcefileretrieval.AdtSourceFileRetrievalServiceFactory;
import com.sap.adt.tools.abapsource.sourcefileretrieval.IAdtSourceFileRetrievalService;
import com.sap.adt.tools.core.IAdtObjectReference;
import com.sap.adt.tools.core.urimapping.AdtUriMappingServiceFactory;
import com.sap.adt.tools.core.urimapping.IAdtUriMappingService;
import com.sap.adt.tools.core.urimapping.UriMappingContext;
import com.teamscale.ide.commons.findings.tasks.FetchFindingsTask;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.ITaskExecutor;
import com.teamscale.ide.eclipse.IUiPreferences;
import com.teamscale.ide.eclipse.model.ResourceReferences;
import com.teamscale.ide.eclipse.plugin.EclipseServiceRegistry;
import com.teamscale.ide.eclipse.util.ui.WorkbenchUtils;

/**
 * Registers {@link IActivationSuccessListener} to trigger loading current
 * finding markers for the activated files. This functionality heavily depends
 * upon internal APIs from SAP, so we should be prepared for breaking changes.
 * <p>
 * Enabled by extension point org.eclipse.ui.startup.
 */
public class ActivationSuccessListener implements IActivationSuccessListener {

	/**
	 * Properties files in Adt start with `ap`, e.g. `apclass`, `approg`, or
	 * `apfunc`.
	 */
	private static final String PROPERTIES_FILE_EXTENSION_PREFIX = "ap";

	private final IAdtUriMappingService uriMappingService;
	private final IAdtSourceFileRetrievalService fileRetrievalService;

	public ActivationSuccessListener() {
		uriMappingService = AdtUriMappingServiceFactory.createUriMappingService();
		fileRetrievalService = AdtSourceFileRetrievalServiceFactory.createSourceFileRetrievalService();
	}

	/**
	 * Overridden to be informed about ABAP code activations to trigger a code
	 * update and fetching current findings from Teamscale for the file in the open
	 * editor.
	 */
	@Override
	public void onActivationSuccess(List<IAdtObjectReference> references, IProject project) {
		UriMappingContext uriMappingContext = new UriMappingContext(project);
		Set<IFile> updatedFiles = new HashSet<>();

		for (IAdtObjectReference reference : references) {
			IFile file = getSourceFile(reference, uriMappingContext);
			if (updatedFiles.add(file) && isOpenInActiveEditor(file)) {
				fetchFindings(file);
				return;
			}
		}
	}

	private void fetchFindings(IFile file) {
		ResourceReference resource = ResourceReferences.fromIResource(file).orElse(null);
		if (resource == null) {
			return;
		}
		boolean usePreCommit = IUiPreferences.getInstance().isAutomaticPreCommitAnalysisEnabled();

		IServiceRegistry serviceRegistry = EclipseServiceRegistry.getInstance();
		serviceRegistry.getService(ITaskExecutor.class).execute(FetchFindingsTask.createFetchTask(serviceRegistry, ENotificationCause.INDIRECT,
				Collections.singleton(resource), usePreCommit));
	}

	/**
	 * Returns the source file for the given {@link IAdtObjectReference}.
	 * <p>
	 * We need this functionality, as on (single) activation of ABAP objects, the
	 * listener does not return the source references, which are opened in the
	 * editor, but the properties file. Thus, we must first retrieve the source file
	 * corresponding to the properties file.
	 */
	private IFile getSourceFile(IAdtObjectReference reference, UriMappingContext uriMappingContext) {
		IFile file = (IFile) uriMappingService.getPlatformResource(uriMappingContext, reference.getUri());
		if (isAdtPropertiesFile(file)) {
			file = fileRetrievalService.getSourceFile(file);
		}
		return file;
	}

	/**
	 * Returns {@code true} if the file has a properties file extension.
	 */
	private static boolean isAdtPropertiesFile(IFile file) {
		return file.getFileExtension().startsWith(PROPERTIES_FILE_EXTENSION_PREFIX);
	}

	/**
	 * Returns {@code true}, if the given (ADT source) {@link IFile} is currently
	 * open in the active workbench editor, otherwise {@code false}.
	 */
	private static boolean isOpenInActiveEditor(IFile activatedFile) {
		IWorkbenchPage activePage = WorkbenchUtils.getActiveWorkbenchPage().orElse(null);
		if (activePage == null) {
			return false;
		}

		IEditorPart openEditor = activePage.getActiveEditor();
		if (openEditor == null) {
			return false;
		}

		IFile openFile = EditorUtils.getEditorFile(openEditor.getEditorInput());
		return Objects.equal(activatedFile, openFile);
	}
}