/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.findings;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.conqat.ide.commons.ui.editor.EditorUtils;
import org.conqat.ide.commons.ui.logging.LoggingUtils;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.e4.core.contexts.IEclipseContext;
import org.eclipse.e4.core.di.annotations.CanExecute;
import org.eclipse.e4.core.di.annotations.Execute;
import org.eclipse.jface.action.ContributionItem;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.jface.resource.LocalResourceManager;
import org.eclipse.jface.resource.ResourceManager;
import org.eclipse.jface.text.source.IVerticalRulerInfo;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchPart;

import com.teamscale.ide.TeamscaleIDEActivator;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.eclipse.findings.FindingIcons;
import com.teamscale.ide.eclipse.findings.handlers.FlagFindingAsFalsePositiveHandler;
import com.teamscale.ide.eclipse.findings.handlers.FlagFindingAsToleratedHandler;
import com.teamscale.ide.eclipse.findings.handlers.OpenCheckDescriptionHandler;
import com.teamscale.ide.eclipse.findings.handlers.OpenFindingInBrowserHandler;
import com.teamscale.ide.eclipse.util.contexts.InjectionUtils;
import com.teamscale.ide.eclipse.util.ui.WorkbenchUtils;
import com.teamscale.ide.findings.view.FindingsMarkerView;

/**
 * A {@link ContributionItem} that dynamically creates the context menu item
 * when the user clicks on a finding annotation in the vertical ruler of the
 * editor or on a finding in the "findings view" windows.
 * 
 * See extension of org.eclipse.ui.menus in plugin.xml.
 */
public class FindingsContributionItem extends ContributionItem {

	private final ResourceManager resourceManager = new LocalResourceManager(JFaceResources.getResources());

	@Override
	public void dispose() {
		resourceManager.dispose();
	}

	@Override
	public void fill(Menu menu, int index) {
		IWorkbenchPart activePart = WorkbenchUtils.getActiveWorkbenchPart().orElseThrow(IllegalStateException::new);

		List<IMarker> findingMarkers;
		boolean editorSidebarSelected = activePart instanceof IEditorPart;
		if (activePart instanceof FindingsMarkerView) {
			FindingsMarkerView view = (FindingsMarkerView) activePart;
			findingMarkers = view.getSelectedMarkersFromView();
		} else if (editorSidebarSelected) {
			IEditorPart editor = (IEditorPart) activePart;
			findingMarkers = getSelectedFindingMarkers(editor);
		} else {
			findingMarkers = Collections.emptyList();
		}

		List<Finding> findings = FindingsMarkerUtils.convertMarkersToFindingListAndSort(findingMarkers);
		IEclipseContext context = activePart.getSite().getService(IEclipseContext.class);

		if (findings.size() > 1) {
			addSubMenusForFindings(findings, menu, context);
		} else if (findings.size() == 1) {
			if (editorSidebarSelected) {
				// Add a dedicated sub-menu for the single finding; in the editor sidebar it is
				// otherwise not clear what finding we are talking about.
				addSubMenusForFindings(findings, menu, context);
			} else {
				addItemsForFinding(findings.get(0), menu, context);
			}
		} else {
			addItemForNoFindings(menu);
		}
	}

	private static List<IMarker> getSelectedFindingMarkers(IEditorPart editor) {
		IVerticalRulerInfo rulerInfo = editor.getAdapter(IVerticalRulerInfo.class);
		if (rulerInfo == null || rulerInfo.getLineOfLastMouseButtonActivity() < 0) {
			return Collections.emptyList();
		}

		List<IMarker> selectedMarkers = new ArrayList<>();
		for (IMarker marker : getAllFindingMarkers(editor)) {
			if (isMarkerSelected(marker, rulerInfo)) {
				selectedMarkers.add(marker);
			}
		}
		return selectedMarkers;
	}

	/** Get all attached finding markers for the given editor. */
	private static IMarker[] getAllFindingMarkers(IEditorPart editor) {
		try {
			IFile file = EditorUtils.getEditorFile(editor.getEditorInput());
			if (file != null) {
				return FindingsMarkerUtils.getFindingMarkers(file);
			}
		} catch (CoreException e) {
			LoggingUtils.error(TeamscaleIDEActivator.getDefault(), e);
		}
		return new IMarker[0];
	}

	/** Determine whether the marker is what the user selected. */
	private static boolean isMarkerSelected(IMarker marker, IVerticalRulerInfo info) {
		int startLine = marker.getAttribute(IMarker.LINE_NUMBER, -1);
		int endLine = marker.getAttribute(FindingsMarkerUtils.END_LINE_NUMBER_KEY, -1);

		// the count for the line of last mouse activity begins with 0
		int clickedLine = info.getLineOfLastMouseButtonActivity() + 1;
		return clickedLine >= startLine && clickedLine <= endLine;
	}

	/** Creates a sub-menu for each finding as well as its entries */
	private void addSubMenusForFindings(List<Finding> findings, Menu parent, IEclipseContext context) {
		for (Finding finding : findings) {
			Menu child = createSubMenuForFinding(finding, parent);
			addItemsForFinding(finding, child, context);
		}
	}

	/** Creates a sub-menu item with the finding description. */
	private Menu createSubMenuForFinding(Finding finding, Menu parent) {
		Menu child = new Menu(parent);
		MenuItem item = new MenuItem(parent, SWT.CASCADE);
		item.setText(finding.getMessage());
		item.setMenu(child);
		item.setImage(resourceManager.createImage(FindingIcons.getImageDescriptor(finding)));
		return child;
	}

	/** Creates the open in browser and blacklisting options in the menu */
	private static void addItemsForFinding(Finding finding, Menu parent, IEclipseContext context) {
		createMenuItemWithHandler(parent, "Open in Web Browser", OpenFindingInBrowserHandler.AdaptingHandler.class, finding, context);
		createMenuItemWithHandler(parent, "Open Check Description", OpenCheckDescriptionHandler.AdaptingHandler.class, finding,
				context);
		new MenuItem(parent, SWT.SEPARATOR);
		createMenuItemWithHandler(parent, "Flag as Tolerated...", FlagFindingAsToleratedHandler.AdaptingHandler.class, finding,
				context);
		createMenuItemWithHandler(parent, "Flag as False Positive...", FlagFindingAsFalsePositiveHandler.AdaptingHandler.class, finding,
				context);
	}

	/**
	 * Creates a disabled menu item to make clear that there are no findings and to
	 * prevent that Eclipse automatically disables the 'Findings' menu <i>after</i>
	 * it is opened. It would probably be better to hide or disable the menu, but
	 * this doesn't seem to work with neither {@link #isVisible()} nor
	 * {@link #isEnabled()}.
	 */
	private static void addItemForNoFindings(Menu parent) {
		MenuItem item = new MenuItem(parent, SWT.NONE);
		item.setText("No Findings in This Line");
		item.setEnabled(false);
	}

	/**
	 * Creates a menu item in <code>parent</code> and attaches a selection listener
	 * to it which will invoke the handler's method annotated with {@link Execute}.
	 * If it cannot be executed (checked by comparing the result from the handler's
	 * method annotated with {@link CanExecute}) this method will disable the given
	 * {@link MenuItem}.
	 */
	private static void createMenuItemWithHandler(Menu parent, String label, Class<?> handler, Object selection,
			IEclipseContext context) {
		MenuItem menuItem = new MenuItem(parent, SWT.NONE);
		menuItem.setText(label);

		boolean canExecute = (boolean) InjectionUtils.makeAndInvokeWithSelection(handler, CanExecute.class, context,
				selection, true);
		menuItem.setEnabled(canExecute);
		menuItem.addSelectionListener(new SelectionAdapter() {

			@Override
			public void widgetSelected(SelectionEvent event) {
				InjectionUtils.makeAndInvokeWithSelection(handler, Execute.class, context, selection, null);
			}
		});
	}
}
