/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.findings;

import static java.util.Comparator.comparing;
import static java.util.Comparator.reverseOrder;
import static java.util.stream.Collectors.toList;

import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.conqat.ide.commons.ui.logging.LoggingUtils;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.ICoreRunnable;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;

import com.google.common.base.Preconditions;
import com.google.common.base.Strings;
import com.google.common.collect.Range;
import com.teamscale.ide.TeamscaleIDEActivator;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.LineRange;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.eclipse.model.ResourceReferences;

/**
 * Utilities for Finding Markers (of type {@value #MARKER_ID}).
 */
public class FindingsMarkerUtils {

	private static final Logger LOGGER = LogManager.getLogger();

	/** The marker ID as set in the plugin.xml. */
	public static final String MARKER_ID = "com.teamscale.ide.findingsMarker";

	/** Key for marker attribute storing the last line of a finding. */
	public static final String END_LINE_NUMBER_KEY = "endLineNumber";

	/** Key for marker attribute storing the commons finding as serialized JSON. */
	private static final String COMMONS_FINDING_KEY = "commonsFinding";

	/** Key for marker attribute storing finding group name. */
	public static final String GROUP_KEY = "group";

	/** Key for marker attribute storing finding category name. */
	public static final String CATEGORY_KEY = "category";

	/** Key for marker attribute storing the finding's assessment. */
	public static final String ASSESSMENT_KEY = "assessment";

	/** Key for marker attribute storing the finding's origin. */
	public static final String ORIGIN_KEY = "origin";

	/** Returns all finding markers contained in the given file. */
	public static IMarker[] getFindingMarkers(IFile file) throws CoreException {
		return file.findMarkers(FindingsMarkerUtils.MARKER_ID, false, IResource.DEPTH_ZERO);
	}

	/** Removes all finding markers from the given resource. */
	public static void removeAllMarkers(IResource resource) {
		LOGGER.traceEntry(null, resource);
		try {
			if (resource != null && resource.isAccessible()) {
				resource.deleteMarkers(MARKER_ID, true, IResource.DEPTH_INFINITE);
			}
		} catch (CoreException e) {
			LOGGER.catching(e);
			LoggingUtils.error(TeamscaleIDEActivator.getDefault(), "Markers could not be deleted!", e);
		}
		LOGGER.traceExit();
	}

	/**
	 * Creates markers on a file (if the resource is an {@link IFile}) or on the
	 * children of the resource, if it is a container (directory, package, ...).
	 *
	 * Will never adjust the marker positions.
	 */
	public static void createMarkers(Collection<Finding> findings, IResource resource) {
		LOGGER.traceEntry(null, findings, resource);
		try {
			ResourcesPlugin.getWorkspace().run(new ICoreRunnable() {

				@Override
				public void run(IProgressMonitor monitor) throws CoreException {
					createMarkersBatched(findings, resource);
				}
			}, null, IWorkspace.AVOID_UPDATE, null);
		} catch (CoreException e) {
			LOGGER.catching(e);
			LoggingUtils.error(TeamscaleIDEActivator.getDefault(), "Markers could not be created", e);
		}
		LOGGER.traceExit();
	}

	private static void createMarkersBatched(Collection<Finding> findings, IResource resource) {
		if (resource instanceof IContainer) {
			createMarkersForFolder(findings, (IContainer) resource);
		} else {
			for (Finding finding : findings) {
				createMarkerForFile(finding, resource);
			}
		}
	}

	private static void createMarkersForFolder(Collection<Finding> findings, IContainer folder) {
		ResourceReference localReference = ResourceReferences.fromIResource(folder).orElse(null);
		if (localReference == null) {
			return;
		}

		for (Finding finding : findings) {
			IFile file = toFile(folder.getProject(), finding.getLocalResource()).orElse(null);
			if (file == null) {
				continue;
			}
			createMarkerForFile(finding, file);
		}
	}

	private static void createMarkerForFile(Finding finding, IResource file) {
		if (!file.isAccessible()) {
			return;
		}

		try {
			Map<String, Object> markerAttributes = new HashMap<>();
			markerAttributes.put(COMMONS_FINDING_KEY, Finding.toJson(finding));

			markerAttributes.put(IMarker.MESSAGE, finding.getMessage());

			markerAttributes.putAll(generateIMarkerAttributesForLineRange(finding.getPrimaryLineRange()));

			markerAttributes.put(CATEGORY_KEY, finding.getCategory());
			markerAttributes.put(GROUP_KEY, finding.getGroup());
			markerAttributes.put(ASSESSMENT_KEY, finding.getAssessment().toString());
			markerAttributes.put(ORIGIN_KEY, finding.getBaseBranch());

			// Do not store markers because findings are fetched when a file is opened.
			markerAttributes.put(IMarker.TRANSIENT, true);

			IMarker marker = file.createMarker(FindingsMarkerUtils.MARKER_ID);
			marker.setAttributes(markerAttributes);
		} catch (CoreException e) {
			LOGGER.catching(e);
			LoggingUtils.error(TeamscaleIDEActivator.getDefault(), "Marker could not be created", e);
		}
	}

	private static Optional<IFile> toFile(IProject project, ResourceReference localResource) {
		IPath absolutePath = Path.fromOSString(localResource.toPath().toAbsolutePath().toString());
		IPath relativePath = absolutePath.makeRelativeTo(project.getLocation());

		IResource resource = project.findMember(relativePath);

		if (resource instanceof IFile) {
			return Optional.ofNullable((IFile) resource);
		}
		return Optional.empty();
	}

	/**
	 * Splits the location string produced by {@link #getLocationString(int, int)}.
	 * Note that the returned {@link Range} may contain the same number twice for
	 * one-line findings
	 */
	public static Range<Integer> splitLocationString(String locationString) {
		Preconditions.checkArgument(!Strings.isNullOrEmpty(locationString));
		String[] parts = locationString.split("-");
		if (parts.length == 1) {
			return Range.closed(Integer.valueOf(parts[0]), Integer.valueOf(parts[0]));
		}
		return Range.closed(Integer.valueOf(parts[0]), Integer.valueOf(parts[1]));

	}

	/** Returns a human readable string for the finding location. */
	public static String getLocationString(int startLine, int endLine) {
		if (startLine == endLine) {
			return String.valueOf(startLine);
		}
		return startLine + "-" + endLine;
	}

	private static Map<String, Object> generateIMarkerAttributesForLineRange(LineRange lineRange) {
		Map<String, Object> locationAttributes = new HashMap<>();
		int startLine = 1;
		int endLine = 1;
		if (lineRange.isSpecified()) {
			startLine = lineRange.getFirstLine();
			endLine = lineRange.getLastLine();
		}

		locationAttributes.put(IMarker.LOCATION, getLocationString(startLine, endLine));
		locationAttributes.put(IMarker.LINE_NUMBER, startLine);
		locationAttributes.put(END_LINE_NUMBER_KEY, endLine);

		return locationAttributes;
	}

	/**
	 * Generates a list of findings sorted by their assessment and then by their
	 * message from a given list of markers.
	 */
	public static List<Finding> convertMarkersToFindingListAndSort(List<IMarker> markers) {
		return markers.stream() //
				.map(FindingsMarkerUtils::getCommonsFindingFromMarker) //
				.filter(Optional::isPresent).map(Optional::get) //
				.sorted(comparing(Finding::getAssessment, reverseOrder()).thenComparing(Finding::getMessage)) //
				.collect(toList());
	}

	/**
	 * Retrieves previously serialized commons finding from marker. May return
	 * {@link Optional#empty} in case of IO errors or if no finding is attached to
	 * the marker.
	 */
	private static Optional<Finding> getCommonsFindingFromMarker(IMarker marker) {
		try {
			String value = marker.getAttribute(COMMONS_FINDING_KEY, null);
			if (value == null) {
				LoggingUtils.error(
						TeamscaleIDEActivator.getDefault(), "No persisted finding found in marker with message '"
								+ marker.getAttribute(IMarker.MESSAGE) + "' in resource '" + marker.getResource() + "'",
						null);
				return Optional.empty();
			}
			return Optional.of(Finding.fromJson(value));
		} catch (CoreException | IOException e) {
			LOGGER.catching(e);
			LoggingUtils.error(TeamscaleIDEActivator.getDefault(), e);
			return Optional.empty();
		}
	}
}
