/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.findings;

import static java.util.Objects.requireNonNull;

import java.util.Iterator;

import org.conqat.ide.commons.ui.editor.EditorUtils;
import org.eclipse.core.resources.IMarker;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.preference.PreferenceConverter;
import org.eclipse.jface.text.source.AbstractRulerColumn;
import org.eclipse.jface.text.source.Annotation;
import org.eclipse.jface.text.source.AnnotationBarHoverManager;
import org.eclipse.jface.text.source.IAnnotationHover;
import org.eclipse.jface.text.source.IAnnotationModel;
import org.eclipse.jface.text.source.IAnnotationModelListener;
import org.eclipse.jface.text.source.IVerticalRulerInfo;
import org.eclipse.jface.text.source.IVerticalRulerInfoExtension;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.editors.text.EditorsUI;
import org.eclipse.ui.texteditor.AnnotationPreference;
import org.eclipse.ui.texteditor.ITextEditor;
import org.eclipse.ui.texteditor.MarkerAnnotation;
import org.eclipse.ui.texteditor.rulers.IContributedRulerColumn;
import org.eclipse.ui.texteditor.rulers.RulerColumnDescriptor;

import com.google.common.base.Preconditions;
import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;
import com.teamscale.ide.commons.notification.ENotificationCause;

/**
 * A vertical ruler column contributed to text editors via the
 * {@code org.eclipse.ui.workbench.texteditor.rulerColumns} extension point that
 * paints a vertical bar for those lines that have findings.
 * <p>
 * This class also implements the {@link IVerticalRulerInfo} and
 * {@link IVerticalRulerInfoExtension} interfaces so that
 * {@link AnnotationBarHoverManager#getHover} requests the custom hover from us.
 */
public class FindingsRulerColumn extends AbstractRulerColumn
		implements IContributedRulerColumn, IAnnotationModelListener {

	/** The width of a vertical bar (in pixels). */
	private static final int SINGLE_ANNOTATION_WIDTH = 3;

	/** The width of the space between vertical bars (in pixels). */
	private static final int MULTI_ANNOTATION_SPACE_WIDTH = 1;

	/**
	 * The width of the ruler (in pixels). Should be at least 2 *
	 * {@link #SINGLE_ANNOTATION_WIDTH} + {@link #MULTI_ANNOTATION_SPACE_WIDTH}.
	 */
	private static final int RULER_COLUMN_WIDTH = 2 * SINGLE_ANNOTATION_WIDTH + MULTI_ANNOTATION_SPACE_WIDTH;

	/** The contribution descriptor. */
	private RulerColumnDescriptor descriptor;

	/** The target editor. */
	private ITextEditor editor;

	/** The color to paint the vertical bars in. */
	private Color barColor;

	/** Maps line numbers to the number of finding annotations. */
	private Multiset<Integer> lineAnnotationCount = HashMultiset.create();

	public FindingsRulerColumn() {
		setWidth(RULER_COLUMN_WIDTH);
	}

	@Override
	public void setDescriptor(RulerColumnDescriptor descriptor) {
		Preconditions.checkState(this.descriptor == null);
		this.descriptor = requireNonNull(descriptor);
	}

	@Override
	public RulerColumnDescriptor getDescriptor() {
		return descriptor;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * Called <em>once</em> when the editor opens.
	 */
	@Override
	public void setEditor(ITextEditor editor) {
		Preconditions.checkState(this.editor == null);
		this.editor = requireNonNull(editor);

		EditorUtils.fetchFindingsForEditor(editor, ENotificationCause.INDIRECT);
	}

	@Override
	public ITextEditor getEditor() {
		return editor;
	}

	/**
	 * Called when the column is enabled, either implicitly when the editor opens or
	 * explicitly via the "Findings" toggle in the ruler's context menu.
	 */
	@Override
	public void columnCreated() {
		// Do nothing
	}

	/**
	 * Called when the column is disabled via the "Findings" toggle in the ruler's
	 * context menu.
	 */
	@Override
	public void columnRemoved() {
		// Do nothing
	}

	@Override
	public void setModel(IAnnotationModel model) {
		if (getModel() != null) {
			getModel().removeAnnotationModelListener(this);
		}
		if (model != null) {
			model.addAnnotationModelListener(this);
		}
		super.setModel(model);
	}

	@Override
	public void modelChanged(IAnnotationModel model) {
		refreshLineToAnnotationCountMap(model.getAnnotationIterator());
		queueRedraw();
	}

	/**
	 * Refreshes the {@link #lineAnnotationCount} from the given annotation
	 * iterator.
	 */
	private void refreshLineToAnnotationCountMap(Iterator<Annotation> iterator) {
		lineAnnotationCount = HashMultiset.create();
		while (iterator.hasNext()) {
			Annotation annotation = iterator.next();
			if (!isRelevant(annotation)) {
				continue;
			}

			IMarker marker = ((MarkerAnnotation) annotation).getMarker();
			int startLine = marker.getAttribute(IMarker.LINE_NUMBER, -1);
			int endLine = marker.getAttribute(FindingsMarkerUtils.END_LINE_NUMBER_KEY, -1);
			for (int i = startLine - 1; i <= endLine - 1; i++) {
				lineAnnotationCount.add(i);
			}
		}
	}

	/**
	 * Returns <code>true</code> if the given annotation is relevant to this ruler,
	 * i.e. if it is a MarkerAnnotation of type
	 * {@value FindingsMarkerUtils#MARKER_ID}.
	 */
	private boolean isRelevant(Annotation annotation) {
		return annotation.getType().equals(FindingsMarkerUtils.MARKER_ID) && annotation instanceof MarkerAnnotation;
	}

	@Override
	protected int getCanvasStyle() {
		// enable drawing of background to prevent artifacts under Windows/Mac
		return SWT.NONE;
	}

	/** Posts a redraw request to the UI thread. */
	private void queueRedraw() {
		PlatformUI.getWorkbench().getDisplay().asyncExec(this::redraw);
	}

	@Override
	protected void paintLine(GC gc, int modelLine, int widgetLine, int linePixel, int lineHeight) {
		int numberOfAnnotations = lineAnnotationCount.count(modelLine);
		if (numberOfAnnotations == 0) {
			return;
		}

		gc.setLineWidth(SINGLE_ANNOTATION_WIDTH);
		gc.setForeground(computeForeground(modelLine));

		int x = SINGLE_ANNOTATION_WIDTH / 2;
		gc.drawLine(x, linePixel, x, linePixel + lineHeight);

		if (numberOfAnnotations > 1) {
			x += SINGLE_ANNOTATION_WIDTH + MULTI_ANNOTATION_SPACE_WIDTH;
			gc.drawLine(x, linePixel, x, linePixel + lineHeight);
		}
	}

	@Override
	protected Color computeForeground(int line) {
		if (barColor == null) {
			barColor = doComputeForeground();
		}
		return barColor;
	}

	private Color doComputeForeground() {
		AnnotationPreference annotationPreference = EditorsUI.getAnnotationPreferenceLookup()
				.getAnnotationPreference(FindingsMarkerUtils.MARKER_ID);
		RGB rgb = getColorPreference(annotationPreference, EditorsUI.getPreferenceStore());
		return new Color(getControl().getDisplay(), rgb);
	}

	/**
	 * Reads the {@link AnnotationPreference}'s color preference, using the value
	 * from the preference store (if set) and falling back to the default from the
	 * {@link AnnotationPreference} (set in the {@code plugin.xml}).
	 * <p>
	 * That one cannot simply use
	 * {@link AnnotationPreference#getColorPreferenceValue()} is a piece of very
	 * confusing API.
	 */
	private static RGB getColorPreference(AnnotationPreference annotationPreference, IPreferenceStore preferenceStore) {
		String colorPreferenceKey = annotationPreference.getColorPreferenceKey();
		if (preferenceStore.contains(colorPreferenceKey) && !preferenceStore.isDefault(colorPreferenceKey)) {
			return PreferenceConverter.getColor(preferenceStore, colorPreferenceKey);
		}
		return annotationPreference.getColorPreferenceValue();
	}

	/** @see MultilineAnnotationHover */
	@Override
	public IAnnotationHover getHover() {
		return new MultilineAnnotationHover();
	}
}
