/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.findings;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.resources.IMarker;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IInformationControl;
import org.eclipse.jface.text.IInformationControlCreator;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.source.Annotation;
import org.eclipse.jface.text.source.DefaultAnnotationHover;
import org.eclipse.jface.text.source.IAnnotationHover;
import org.eclipse.jface.text.source.IAnnotationHoverExtension;
import org.eclipse.jface.text.source.IAnnotationModel;
import org.eclipse.jface.text.source.ILineRange;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.jface.text.source.ISourceViewerExtension2;
import org.eclipse.jface.text.source.LineRange;
import org.eclipse.jface.text.source.projection.AnnotationBag;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.texteditor.MarkerAnnotation;
import org.eclipse.ui.texteditor.SimpleMarkerAnnotation;

import com.google.common.base.Strings;
import com.teamscale.ide.eclipse.findings.FindingsInformationControl;

/**
 * This is mostly a copy of {@link DefaultAnnotationHover} with a change in
 * {@link #isRulerLine(Position, IDocument, int)} so that tooltips get shown for
 * annotations that span multiple lines.
 * 
 * Actually, this should be fixed upstream in DefaultAnnotationHover, but the
 * Eclipse guys don't consider it a bug. See
 * <a href="https://bugs.eclipse.org/bugs/show_bug.cgi?id=427630">Bugreport at
 * eclipse.org</a>.
 */
public class MultilineAnnotationHover implements IAnnotationHover, IAnnotationHoverExtension {

	@Override
	public IInformationControlCreator getHoverControlCreator() {
		return new IInformationControlCreator() {
			@Override
			public IInformationControl createInformationControl(Shell parent) {
				return new FindingsInformationControl(parent);
			}
		};
	}

	@Override
	public boolean canHandleMouseCursor() {
		return false;
	}

	@Override
	public String getHoverInfo(ISourceViewer sourceViewer, int lineNumber) {
		throw new UnsupportedOperationException("Rendering is controlled by a custom IInformationControlCreator");
	}

	@Override
	public Object getHoverInfo(ISourceViewer sourceViewer, ILineRange lineRange, int visibleNumberOfLines) {
		List<Annotation> annotations = getAnnotationsForLine(sourceViewer, lineRange.getStartLine());
		List<IMarker> markers = new ArrayList<>();
		for (Annotation annotation : annotations) {
			SimpleMarkerAnnotation markerAnnotation = (SimpleMarkerAnnotation) annotation;
			markers.add(markerAnnotation.getMarker());
		}
		return FindingsMarkerUtils.convertMarkersToFindingListAndSort(markers);
	}

	/**
	 * @return A list of annotations for a certain line in the given source viewer.
	 */
	private static List<Annotation> getAnnotationsForLine(ISourceViewer viewer, int line) {
		IAnnotationModel model = getAnnotationModel(viewer);
		if (model == null) {
			return null;
		}

		IDocument document = viewer.getDocument();
		if (document == null) {
			return null;
		}

		List<Annotation> annotations = new ArrayList<>();

		Iterator<Annotation> annotationIterator = model.getAnnotationIterator();
		while (annotationIterator.hasNext()) {
			Annotation annotation = annotationIterator.next();

			Position position = model.getPosition(annotation);
			if (position == null || !isRulerLine(position, document, line)) {
				continue;
			}

			if (annotation instanceof AnnotationBag) {
				annotations.addAll(getAnnotationsFromAnnotationBag((AnnotationBag) annotation, model));
				continue;
			}

			if (includeAnnotation(annotation)) {
				annotations.add(annotation);
			}
		}

		return annotations;
	}

	/** @return The {@link IAnnotationModel} for the given source viewer. */
	private static IAnnotationModel getAnnotationModel(ISourceViewer viewer) {
		if (viewer instanceof ISourceViewerExtension2) {
			ISourceViewerExtension2 extension = (ISourceViewerExtension2) viewer;
			return extension.getVisualAnnotationModel();
		}
		return viewer.getAnnotationModel();
	}

	/**
	 * @return <code>true</code> if the given line is in the line range of the given
	 *         position.
	 */
	private static boolean isRulerLine(Position position, IDocument document, int line) {
		if (position.getOffset() > -1 && position.getLength() > -1) {
			try {
				return line >= document.getLineOfOffset(position.getOffset())
						&& line <= document.getLineOfOffset(position.getOffset() + position.getLength());
			} catch (BadLocationException e) {
				// Ignore
			}
		}
		return false;
	}

	/**
	 * @return The annotations from an {@link AnnotationBag} that return
	 *         <code>true</code> for {@link #includeAnnotation(Annotation)} as a
	 *         {@link List}.
	 */
	private static List<Annotation> getAnnotationsFromAnnotationBag(AnnotationBag bag, IAnnotationModel model) {
		List<Annotation> annotations = new ArrayList<>();

		Iterator<Annotation> bagIterator = bag.iterator();
		while (bagIterator.hasNext()) {
			Annotation annotation = bagIterator.next();
			Position position = model.getPosition(annotation);
			if (position != null && includeAnnotation(annotation)) {
				annotations.add(annotation);
			}
		}
		return annotations;
	}

	/** Decides whether to include the given annotation or not. */
	private static boolean includeAnnotation(Annotation annotation) {
		return isFindingAnnotation(annotation) && !hasEmptyMessage(annotation);
	}

	/** Returns whether the annotation originates in a findings marker. */
	private static boolean isFindingAnnotation(Annotation annotation) {
		return annotation instanceof MarkerAnnotation
				&& ((MarkerAnnotation) annotation).getType().equals(FindingsMarkerUtils.MARKER_ID);
	}

	/** Returns whether the annotation has an empty message. */
	private static boolean hasEmptyMessage(Annotation annotation) {
		return Strings.isNullOrEmpty(annotation.getText());
	}

	@Override
	public ILineRange getHoverLineRange(ISourceViewer viewer, int lineNumber) {
		return new LineRange(lineNumber, 1);
	}

}
