/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright (c) 2009-2023 CQSE GmbH                                        |
|                                                                          |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.findings.view;

import org.conqat.ide.commons.ui.adt.AdtUtils;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.jobs.IJobChangeEvent;
import org.eclipse.core.runtime.jobs.JobChangeAdapter;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;

import com.teamscale.ide.EclipseUtils;
import com.teamscale.ide.client.NamedServerDetails;
import com.teamscale.ide.client.PrefixMappingRegistry;
import com.teamscale.ide.client.TeamscaleConnectionRegistry;
import com.teamscale.ide.eclipse.util.ui.WorkbenchUtils;
import com.teamscale.ide.properties.AutoMapping;

/**
 * A {@link SelectionAdapter} used for performing auto mapping of the given
 * {@linkplain IResource resource}.
 */
final class AutoMappingButtonListener extends SelectionAdapter {
	/** The given {@link IResource} */
	private final IResource resource;

	/** Constructor expecting the related {@link IResource}. */
	AutoMappingButtonListener(IResource resource) {
		this.resource = resource;
	}

	/** {@inheritDoc} */
	@Override
	public void widgetSelected(SelectionEvent event) {
		// Enable Teamscale for this project
		PrefixMappingRegistry.getInstance(resource.getProject()).setTeamscaleEnabled(true);

		if (AdtUtils.isAdtResource(resource)) {
			EclipseUtils.openTeamscaleProjectPreferencePage(resource.getProject());
			FindingsMarkerView.refreshView(resource);
			return;
		}

		performAutoMapping(resource);
	}

	/**
	 * Determines the default branch and performs auto mapping of resources.
	 */
	private void performAutoMapping(final IResource resource) {
		IProject eclipseProject = resource.getProject();
		PrefixMappingRegistry registry = PrefixMappingRegistry.getInstance(eclipseProject);
		NamedServerDetails serverDetails = TeamscaleConnectionRegistry.getInstance()
				.getServer(registry.getTeamscaleServer());

		if (serverDetails == null) {
			displayErrorAndOpenProjectPreferences(eclipseProject, "server");
			return;
		}
		if (registry.getTeamscaleProject() == null) {
			displayErrorAndOpenProjectPreferences(eclipseProject, "project");
			return;
		}

		String path = null;
		if (resource instanceof IFile) {
			path = resource.getProjectRelativePath().toString();
		}

		JobChangeAdapter persistAndRefreshAdapter = new JobChangeAdapter() {
			/** {@inheritDoc} */
			@Override
			public void done(IJobChangeEvent event) {
				if (event.getResult().isOK()) {
					registry.persist();
					FindingsMarkerView.refreshView(resource);
				}
			}
		};
		new AutoMapping(resource.getProject(), path, true).tryAutoMapping(persistAndRefreshAdapter);
	}

	/**
	 * Shows an error dialog for the missing configuration option and opens the
	 * Teamscale project preference page afterwards.
	 */
	public static void displayErrorAndOpenProjectPreferences(IProject project, String missingConfigurationOption) {
		MessageDialog.openError(WorkbenchUtils.getActiveShell().orElse(null),
				"No Teamscale " + missingConfigurationOption + " configured", "There is no Teamscale "
						+ missingConfigurationOption + " configured. Please configure the server manually.");
		EclipseUtils.openTeamscaleProjectPreferencePage(project);
	}
}