/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.findings.view;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.conqat.ide.commons.ui.logging.LoggingUtils;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.ui.IMemento;
import org.eclipse.ui.views.markers.MarkerFieldFilter;
import org.eclipse.ui.views.markers.MarkerItem;

import com.google.common.base.Strings;
import com.teamscale.ide.TeamscaleIDEActivator;

/**
 * A {@link MarkerFieldFilter} that excludes markers based on a list of strings
 * matching a given marker attribute.
 */
public abstract class FieldFilterBase extends MarkerFieldFilter {

	/** Key for storing the filter values in the memento. */
	private static final String FILTER_VALUES_MEMENTO_KEY = "filter-values";

	/** The values to filter markers. */
	private List<String> filteredValues = new ArrayList<>();

	/** The name of the marker attribute to match against. */
	private final String filterAttributeName;

	/** Constructor. */
	public FieldFilterBase(String filterAttributeName) {
		this.filterAttributeName = filterAttributeName;
	}

	/** Sets a new list of values that should be used to exclude markers. */
	public void setFilteredValues(List<String> filteredValues) {
		this.filteredValues = filteredValues;
	}

	/** Gets the list of values that should be used to exclude markers. */
	public List<String> getFilteredValues() {
		return this.filteredValues;
	}

	/** {@inheritDoc} */
	@Override
	public void populateWorkingCopy(MarkerFieldFilter filter) {
		super.populateWorkingCopy(filter);
		if (filter instanceof FieldFilterBase) {
			((FieldFilterBase) filter).filteredValues = this.filteredValues;
		}
	}

	/** {@inheritDoc} */
	@Override
	public boolean select(MarkerItem item) {
		try {
			return !filteredValues.contains(item.getMarker().getAttribute(filterAttributeName));
		} catch (CoreException e) {
			LoggingUtils.error(TeamscaleIDEActivator.getDefault(), e);
			return true;
		}
	}

	/** {@inheritDoc} */
	@Override
	public void saveSettings(IMemento memento) {
		memento.putString(FILTER_VALUES_MEMENTO_KEY, String.join(";", filteredValues));
	}

	/** {@inheritDoc} */
	@Override
	public void loadSettings(IMemento memento) {
		String value = memento.getString(FILTER_VALUES_MEMENTO_KEY);
		if (!Strings.isNullOrEmpty(value)) {
			filteredValues = Arrays.asList(value.split(";"));
		}
	}

}