/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.findings.view;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.function.Consumer;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.checkerframework.checker.nullness.qual.Nullable;
import org.conqat.ide.commons.ui.adt.AdtUtils;
import org.conqat.ide.commons.ui.editor.EditorUtils;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.jface.layout.GridDataFactory;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.TreePath;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StackLayout;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IMemento;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IViewSite;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.views.markers.MarkerItem;
import org.eclipse.ui.views.markers.MarkerSupportView;

import com.teamscale.ide.EclipseUtils;
import com.teamscale.ide.client.CheckTemporarilyDisabledConnectionJob;
import com.teamscale.ide.client.ETeamscaleGlobalServerStatus;
import com.teamscale.ide.client.ITeamscaleServerStatusChangeListener;
import com.teamscale.ide.client.NamedServerDetails;
import com.teamscale.ide.client.PrefixMappingRegistry;
import com.teamscale.ide.client.TeamscaleConnectionRegistry;
import com.teamscale.ide.client.TeamscaleServerStatus;
import com.teamscale.ide.commons.findings.EFindingFetchError;
import com.teamscale.ide.commons.findings.Finding;
import com.teamscale.ide.commons.findings.services.IFindingFetchStateService;
import com.teamscale.ide.commons.findings.services.IFindingsListener;
import com.teamscale.ide.commons.findings.tasks.FetchFindingsTask;
import com.teamscale.ide.commons.findings.tasks.TeamscaleSapSyncTask;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.ITaskExecutor;
import com.teamscale.ide.eclipse.EclipseIds;
import com.teamscale.ide.eclipse.IUiPreferences;
import com.teamscale.ide.eclipse.model.ResourceReferences;
import com.teamscale.ide.eclipse.plugin.EclipseServiceRegistry;
import com.teamscale.ide.eclipse.util.ui.WorkbenchUtils;
import com.teamscale.ide.eclipse.util.viewers.AdaptableUtils;
import com.teamscale.ide.eclipse.util.viewers.SelectionUtils;
import com.teamscale.ide.findings.FindingsMarkerUtils;

/**
 * Findings view based on MarkerSupportView (available since Eclipse 3.4). Shows
 * markers of the <code>findingmarkers</code> type as defined in plugin.xml. The
 * view overrides {@link #createPartControl(Composite)} and uses a
 * {@link StackLayout} in order to display an additional message besides the
 * marker view's content. We differ between error messages and "normal"
 * messages. An error message is displayed, if e.g. a specific server is not
 * reachable. If no error message is set, either a message depending on a
 * globally aggregated server status or findings are shown.
 */
public class FindingsMarkerView extends MarkerSupportView
		implements ISelectionListener, ITeamscaleServerStatusChangeListener, IFindingsListener {

	private static final Logger LOGGER = LogManager.getLogger();

	/**
	 * The used layout to either display an error message or the marker view itself.
	 */
	private StackLayout stackLayout;

	/** The parent composite of the view. */
	private Composite parentComposite;

	/** The composite which contains the marker view. */
	private Composite markerComposite;

	/** The composite which contains the error label. */
	private Composite errorComposite;

	/** Stores the markers which are selected in the view. */
	private List<IMarker> selectedMarkers = new ArrayList<>();

	/**
	 * If currently an error message is displayed. If this is false a message
	 * depending on the globally aggregated server status will be displayed.
	 */
	private boolean isErrorMessageDisplayed;

	/** The label to display an message. */
	private Label messageLabel;

	/** The button to display below an error message. */
	private Button errorButton;

	/** The listener to react on error button clicks. */
	private SelectionListener errorButtonListener;

	/**
	 * The label to display info message like the current branch. We show the branch
	 * in the UI because it can happen that a branch is used locally that has not
	 * been pushed to the server. In this case the default branch is used
	 */
	private Label infoLabel;

	/**
	 * The resource for which findings are currently shown. May be <code>null</code>
	 * when the user selects a non-resource e.g. dependency containers or similar
	 * (in which case the {@link #currentProject} is <code>non-null</code>)
	 */
	private IResource currentResource = null;

	private ResourceReference currentResourceReference = null;

	/**
	 * The project which is currently selected or is associated with the currently
	 * selected resource. This is either the project associated with the current
	 * selection or the project associated with the currently open editor window.
	 * May be <code>null</code>
	 */
	private IProject currentProject = null;

	private ITaskExecutor taskExecutor = null;
	private IServiceRegistry serviceRegistry = null;

	/** Constructor. */
	public FindingsMarkerView() {
		super("com.teamscale.ide.findingsMarkerContentGenerator");
	}

	/** {@inheritDoc} */
	@SuppressWarnings("restriction")
	@Override
	public void init(IViewSite site, IMemento memento) throws PartInitException {
		super.init(site, memento);
		site.getWorkbenchWindow().getSelectionService().addSelectionListener(this);
		TeamscaleServerStatus.getInstance().addStatusChangeListener(this);

		serviceRegistry = EclipseServiceRegistry.getInstance();
		taskExecutor = serviceRegistry.getService(ITaskExecutor.class);
		serviceRegistry.getService(IFindingFetchStateService.class).addListener(this);
	}

	/**
	 * Updates the view when the current selection is changed. If the event was
	 * triggered by <code>this</code>, indicating the user pressed the refresh
	 * button, the selection is empty and therefore this returns early because we
	 * can not update the resource and project, and refresh is called anyway
	 */
	@Override
	public void selectionChanged(IWorkbenchPart sourcePart, ISelection selection) {
		handleSelectedMarkers(selection);

		if (sourcePart == this) {
			return;
		}

		IResource resource;
		IProject project;

		if (isEditorSelected()) {
			// obtain current resource and project from the active editor window
			resource = EditorUtils.getEditorFile(
					(IEditorPart) WorkbenchUtils.getActiveWorkbenchPart().orElseThrow(IllegalStateException::new));
			project = AdaptableUtils.getMatchingAdapter(resource, IProject.class, true).orElse(null);
		} else {
			// obtain current resource and project from the selection
			IAdaptable adaptableFromSelection = SelectionUtils.getSingleElement(selection)
					.filter(IAdaptable.class::isInstance).map(IAdaptable.class::cast).orElse(null);
			resource = AdaptableUtils.getMatchingAdapter(adaptableFromSelection, IResource.class, false).orElse(null);
			project = AdaptableUtils.getMatchingAdapter(adaptableFromSelection, IProject.class, true).orElse(null);
		}

		boolean shouldRefresh = false;
		if (resource != null && !resource.equals(currentResource)) {
			currentResource = resource;
			currentResourceReference = ResourceReferences.fromIResource(resource).orElse(null);
			shouldRefresh = true;
		}
		if (project != null && !project.equals(currentProject)) {
			currentProject = project;
			shouldRefresh = true;
		}
		if (shouldRefresh) {
			refresh();
		}
	}

	/** Returns whether the editor currently selected by the user */
	private static boolean isEditorSelected() {
		IWorkbenchPart activePart = WorkbenchUtils.getActiveWorkbenchPart().orElse(null);
		// null instanceof IEditorPart is always false
		return (activePart instanceof IEditorPart);
	}

	/**
	 * Refreshes the view based on {@link #currentResource} and
	 * {@link #currentProject}
	 */
	public void refresh() {
		refresh(false);
	}

	/**
	 * Refreshes the view based on {@link #currentResource} and
	 * {@link #currentProject}. In comparison to {@link #refresh()}, the findings
	 * will always be pulled from Teamscale and never from the cache.
	 */
	public void forceRefresh() {
		refresh(true);
	}

	/**
	 * Refreshes the view based on {@link #currentResource} and
	 * {@link #currentProject}.
	 * 
	 * @param forceFetch When {@code true}, a fetch without cache will be done.
	 */
	private void refresh(boolean forceFetch) {
		if (isProjectClosed(currentProject)) {
			handleClosedProject();
		} else if (isProjectDisabled(currentProject)) {
			handleDisabledProject(currentProject);
		} else if (currentResource == null) {
			handleNonResource();
		} else if (AdtUtils.isAdtProject(currentProject)) {
			refreshAdt(forceFetch);
		} else {
			handleResource(currentResource, forceFetch);
		}
	}

	/**
	 * Refreshes the view based on {@link #currentResource} and
	 * {@link #currentProject}. It is expected that {@link #currentResource} is not
	 * {@code null} and {@link #currentProject} is an ADT project.
	 * 
	 * @param forceFetch When {@code true}, a fetch without cache will be done.
	 */
	private void refreshAdt(boolean forceFetch) {
		if (currentResource == currentProject) {
			// We are not able to resolve the findings if someone clicked on the project.
			// We shouldn't fetch any because it will only throw an error.
			handleNonResource();
		} else if (!AdtUtils.isAdtProjectConnected(currentProject)) {
			// The URIs change if the project is not connected. There is no reason to even
			// try to do a pre-commit or findings fetch.
			handleNonConnectedAbapProject();
		} else {
			handleResource(currentResource, forceFetch);
		}
	}

	/**
	 * Returns whether the given project is closed and therefore not accessible
	 */
	private static boolean isProjectClosed(IProject project) {
		return project != null && !project.isAccessible();
	}

	/** Returns whether the given project is disabled */
	private static boolean isProjectDisabled(IProject project) {
		return project != null && !PrefixMappingRegistry.getInstance(project).isTeamscaleEnabled();
	}

	/** Handles the closed project by displaying the corresponding error message */
	private static void handleClosedProject() {
		setErrorMessage("Project is closed. Open it to see findings.");
	}

	/**
	 * Handles the given disabled project by displaying the corresponding error
	 * message
	 */
	private void handleDisabledProject(IProject project) {
		FindingsMarkerUtils.removeAllMarkers(project);
		showErrorMessageWithButton("Teamscale is disabled for this project.", "Enable Teamscale for this project",
				getAutoMappingButtonListener(project));
	}

	/** Handles the non-resource by showing the corresponding error message */
	private static void handleNonResource() {
		setErrorMessage("There are no findings for this resource.");
	}

	/**
	 * Handles the non connected ABAP project by showing the corresponding error
	 * message
	 */
	private static void handleNonConnectedAbapProject() {
		setErrorMessage("Project is not logged on.");
	}

	/** Fetches the findings by scheduling a {@link LoadFindingMarkersJob} */
	private void handleResource(IResource resource, boolean forceFetch) {
		ResourceReference resourceReference = ResourceReferences.fromIResource(resource).orElse(null);

		if (resourceReference == null) {
			return;
		}

		boolean usePreCommit = IUiPreferences.getInstance().isAutomaticPreCommitAnalysisEnabled();
		if (forceFetch) {
			if (AdtUtils.isAdtResource(resource)) {
				// A force fetch in ABAP/SAP means a full sync
				taskExecutor.execute(new TeamscaleSapSyncTask(serviceRegistry, resourceReference,
						PrefixMappingRegistry.getInstance(currentProject).getTeamscaleSapSyncTimeout()));
			} else {
				taskExecutor.execute(FetchFindingsTask.createForceFetchTask(serviceRegistry, ENotificationCause.DIRECT,
						Collections.singleton(resourceReference), usePreCommit));
			}
		} else {
			taskExecutor.execute(FetchFindingsTask.createFetchTask(serviceRegistry, ENotificationCause.INDIRECT,
					Collections.singleton(resourceReference), usePreCommit));
		}
	}

	/** Stores the selected marker. */
	private void handleSelectedMarkers(ISelection selection) {
		if (selection instanceof TreeSelection) {
			TreeSelection treeSelection = (TreeSelection) selection;
			List<IMarker> markers = new ArrayList<>();
			for (TreePath path : treeSelection.getPaths()) {
				Object first = path.getFirstSegment();

				if (first instanceof MarkerItem) {
					MarkerItem item = (MarkerItem) first;
					markers.add(item.getMarker());
				}
			}
			selectedMarkers = markers;
		}
	}

	/**
	 * @see #selectedMarkers
	 */
	public List<IMarker> getSelectedMarkersFromView() {
		return selectedMarkers;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("restriction")
	@Override
	public void dispose() {
		TeamscaleServerStatus.getInstance().removeStatusChangeListener(this);
		getSite().getWorkbenchWindow().getSelectionService().removeSelectionListener(this);
		serviceRegistry.getService(IFindingFetchStateService.class).removeListener(this);
		super.dispose();
	}

	/** {@inheritDoc} */
	@Override
	public void statusChanged() {
		final IWorkbench workbench = PlatformUI.getWorkbench();
		workbench.getDisplay().asyncExec(this::updateDisplayedMessage);
	}

	/** Updates the displayed message, depending on the global server status. */
	public void updateDisplayedMessage() {
		if (!isErrorMessageDisplayed) {
			if (!TeamscaleConnectionRegistry.getInstance().hasServers()) {
				showMessage("Currently no servers are configured!");
			}

			final ETeamscaleGlobalServerStatus status = TeamscaleServerStatus.getInstance().getGlobalStatus();
			if (status == ETeamscaleGlobalServerStatus.DISABLED) {
				showMessage("Currently no servers are enabled!");
			} else if (status == ETeamscaleGlobalServerStatus.UNREACHABLE) {
				showMessage("No server is reachable!");
			} else {
				hideMessage();
			}
		}
	}

	/**
	 * Removes the currently displayed error message and displays the marker view.
	 */
	private void hideErrorMessage() {
		if (isErrorMessageDisplayed) {
			isErrorMessageDisplayed = false;
			updateDisplayedMessage();
		}
	}

	/**
	 * Displays the given custom error messages and a button with the given text.
	 * The given listener handles button click events.
	 */
	private void showErrorMessageWithButton(String errorMessage, String buttonText, SelectionListener listener) {
		isErrorMessageDisplayed = true;
		showMessage(errorMessage, buttonText, listener);
	}

	/**
	 * Displays the given error message on the currently opened FindingsMarkerView.
	 */
	public static void setErrorMessage(String errorMessage) {
		changeView(view -> view.showErrorMessageWithButton(errorMessage, null, null));
	}

	/** Display the given error message. */
	private void showMessage(String errorMessage) {
		showMessage(errorMessage, null, null);
	}

	/**
	 * Display the given error message. If buttonText is not null, a button with
	 * this text will be displayed below the error message. The given listener will
	 * react on button click events.
	 */
	private void showMessage(String errorMessage, String buttonText, SelectionListener listener) {
		messageLabel.setText(errorMessage);

		if (buttonText != null) {
			errorButton.setText(buttonText);
			if (errorButtonListener != null) {
				errorButton.removeSelectionListener(errorButtonListener);
			}
			errorButtonListener = listener;
			errorButton.addSelectionListener(errorButtonListener);
			errorButton.pack();
		}

		stackLayout.topControl = errorComposite;
		parentComposite.pack();
		parentComposite.layout();
		errorButton.setVisible(buttonText != null);
		parentComposite.layout(true, true);
	}

	/** Removes the displayed message and displays the marker view. */
	private void hideMessage() {
		if (stackLayout == null) {
			return;
		}
		stackLayout.topControl = markerComposite;
		parentComposite.pack();
		parentComposite.layout();

	}

	/** {@inheritDoc} */
	@SuppressWarnings("restriction")
	@Override
	public void createPartControl(Composite parent) {
		this.parentComposite = parent;

		stackLayout = new StackLayout();
		parent.setLayout(stackLayout);

		GridLayout markerLayout = new GridLayout(1, true);
		markerComposite = new Composite(parent, SWT.NONE);
		markerComposite.setLayout(markerLayout);

		infoLabel = new Label(markerComposite, SWT.NONE);
		infoLabel.setToolTipText(
				"The current branch shown may differ from the actual branch in Eclipse if it does not exist in Teamscale.");
		Composite markerDetailsView = new Composite(markerComposite, SWT.NONE);
		super.createPartControl(markerDetailsView);
		GridDataFactory.fillDefaults().grab(true, true).applyTo(markerDetailsView);

		markerComposite.layout();
		markerComposite.pack();

		createMessageView(parent);
		// check server status and set an appropriate error message
		updateDisplayedMessage();
	}

	/** Creates the view to show messages and errors. */
	private void createMessageView(Composite parent) {
		stackLayout.topControl = markerComposite;

		GridLayout messageLayout = new GridLayout();

		errorComposite = new Composite(parent, SWT.NONE);
		errorComposite.setLayout(messageLayout);
		errorComposite.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false));

		messageLabel = new Label(errorComposite, SWT.CENTER);
		messageLabel.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false));

		errorButton = new Button(errorComposite, SWT.PUSH);
		errorButton.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false));
	}

	@Override
	public void onFindingsFetchSucceeded(ResourceReference fetchResource, List<Finding> findings) {
		if (isNotRelevantForCurrentlySelectedResource(fetchResource)) {
			return;
		}
		removeErrorMessage();
	}

	@Override
	public void onFindingsFetchFailed(ResourceReference fetchResource, EFindingFetchError error,
			@Nullable String message, List<Finding> findingsOfPreviousCommit) {
		if (isNotRelevantForCurrentlySelectedResource(fetchResource)) {
			return;
		}
		removeErrorMessage();
	}

	/**
	 * Returns whether the given {@link ResourceReference reference} is not relevant
	 * to the {@link #currentlySelectedResource} . The
	 * {@link #currentlySelectedResource} might change if several service calls are
	 * active in parallel (a trigger-happy user or a slow Teamscale server).
	 * <p>
	 * Use this method when handling the service-call response to check whether the
	 * response is actually for the requested resource. If it is not, ignore it.
	 * <p>
	 * Since {@link #currentlySelectedResource} will be the most recently requested
	 * resource, we should wait until that response comes (and then stop the spinner
	 * etc.).
	 */
	private boolean isNotRelevantForCurrentlySelectedResource(ResourceReference resource) {
		if (this.currentResourceReference == null) {
			return true;
		}

		if (this.currentResourceReference.isFolder()) {
			return !isBelowOrEqual(resource, currentResourceReference);
		} else {
			return !currentResourceReference.equals(resource);
		}
	}

	/**
	 * Checks if possibleChild is equals to possibleParent or if possibleChild is a
	 * child of possibleParent.
	 */
	private static boolean isBelowOrEqual(@Nullable ResourceReference possibleChild, ResourceReference possibleParent) {
		return possibleChild != null && possibleChild.isAtOrBelow(possibleParent);
	}

	/** Changes the view in the UI thread. */
	public static void changeView(Consumer<FindingsMarkerView> function) {
		PlatformUI.getWorkbench().getDisplay().asyncExec(() -> WorkbenchUtils.getViewPart(EclipseIds.Parts.FINDINGS)
				.map(FindingsMarkerView.class::cast).ifPresent(function));
	}

	/**
	 * Removes the given error message from the currently opened FindingsMarkerView.
	 */
	public static void removeErrorMessage() {
		LOGGER.traceEntry();
		changeView(FindingsMarkerView::hideErrorMessage);
		LOGGER.traceExit();
	}

	/**
	 * Displays the temporarily disabled error message. Further a button for
	 * re-enabling the given server is added.
	 */
	public static void displayTemporarilyDisabledMessage(final String symbolicName) {
		SelectionAdapter buttonListener = new SelectionAdapter() {

			@Override
			public void widgetSelected(SelectionEvent event) {
				NamedServerDetails details = TeamscaleConnectionRegistry.getInstance().getServer(symbolicName);
				new CheckTemporarilyDisabledConnectionJob(details).schedule();
			}
		};

		changeView(view -> view.showErrorMessageWithButton(
				"Server '" + symbolicName + "' is temporarily not available!", "Re-enable server", buttonListener));
	}

	/**
	 * Displays that no mapping is found for the resource. Further a button for
	 * auto-mapping.
	 */
	public static void displayNoMatchingFound(final String errorMessage, final IResource resource) {

		changeView(view -> view.showErrorMessageWithButton(errorMessage,
				"Automatically configure Teamscale for this project", getAutoMappingButtonListener(resource)));

		FindingsMarkerUtils.removeAllMarkers(resource.getProject());
	}

	/** Returns a button listener, which handles the auto-mapping. */
	private static SelectionAdapter getAutoMappingButtonListener(final IResource resource) {
		return new AutoMappingButtonListener(resource);
	}

	/** Refreshes the view and reloads the findings. */
	public static void refreshView(IResource resource) {
		changeView(view -> {
			view.currentResource = resource;
			view.currentResourceReference = ResourceReferences.fromIResource(resource).orElse(null);
			view.refresh();
		});
	}

	/** Sets the info message. */
	public static void setInfoMessage(String infoMessage) {
		LOGGER.traceEntry(null, infoMessage);
		changeView(view -> {
			view.infoLabel.setText(infoMessage);
			view.markerComposite.layout();
		});
		LOGGER.traceExit();
	}

	/**
	 * Shows an error massage with a button to open the project property page.
	 */
	public static void displayErrorWithPropertyPageButton(String message, IResource resource) {
		SelectionAdapter buttonListener = new SelectionAdapter() {

			@Override
			public void widgetSelected(SelectionEvent event) {
				EclipseUtils.openTeamscaleProjectPreferencePage(resource.getProject());
				refreshView(resource);
			}
		};

		changeView(view -> view.showErrorMessageWithButton(message, "Open properties", buttonListener));
	}
}
