/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.findings.view;

import java.util.Map;

import org.conqat.ide.commons.ui.logging.LoggingUtils;
import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.Command;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.IParameter;
import org.eclipse.core.commands.NotEnabledException;
import org.eclipse.core.commands.NotHandledException;
import org.eclipse.core.commands.Parameterization;
import org.eclipse.core.commands.ParameterizedCommand;
import org.eclipse.core.commands.common.NotDefinedException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.commands.ICommandService;
import org.eclipse.ui.commands.IElementUpdater;
import org.eclipse.ui.handlers.IHandlerService;
import org.eclipse.ui.menus.UIElement;

import com.teamscale.ide.TeamscaleIDEActivator;
import com.teamscale.ide.client.ETeamscaleGlobalServerStatus;
import com.teamscale.ide.client.ITeamscaleServerStatusChangeListener;
import com.teamscale.ide.client.TeamscaleServerStatus;
import com.teamscale.ide.eclipse.EclipseIds;
import com.teamscale.ide.ui.images.ETeamscaleImage;

/**
 * Handler for the server status button which opens the Teamscale Server
 * preferences page.
 */
public class OpenServerPreferencesHandler extends AbstractHandler
		implements IElementUpdater, ITeamscaleServerStatusChangeListener {

	/** Constructor. */
	public OpenServerPreferencesHandler() {
		TeamscaleServerStatus.getInstance().addStatusChangeListener(this);
	}

	/** {@inheritDoc} */
	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {
		@SuppressWarnings("cast") // Cast required for Eclipse 4.4.2 or older
		ICommandService commandService = (ICommandService) PlatformUI.getWorkbench().getService(ICommandService.class);
		@SuppressWarnings("cast") // Cast required for Eclipse 4.4.2 or older
		IHandlerService handlerService = (IHandlerService) PlatformUI.getWorkbench().getService(IHandlerService.class);
		if (commandService != null && handlerService != null) {
			Command openPreferences = commandService.getCommand("org.eclipse.ui.window.preferences");
			try {
				IParameter param = openPreferences.getParameter("preferencePageId");
				Parameterization[] parameters = { new Parameterization(param, EclipseIds.PreferencePages.TEAMSCALE) };
				ParameterizedCommand command = new ParameterizedCommand(openPreferences, parameters);
				handlerService.executeCommand(command, null);
			} catch (NotDefinedException | NotEnabledException | NotHandledException e) {
				LoggingUtils.error(TeamscaleIDEActivator.getDefault(), e);
			}
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public void updateElement(UIElement element, @SuppressWarnings("rawtypes") Map parameters) {
		ETeamscaleGlobalServerStatus status = TeamscaleServerStatus.getInstance().getGlobalStatus();

		ETeamscaleImage image;
		switch (status) {
		case REACHABLE:
			image = ETeamscaleImage.REACHABLE;
			break;
		case MIXED:
			image = ETeamscaleImage.MIXED;
			break;
		case UNREACHABLE:
			image = ETeamscaleImage.UNREACHABLE;
			break;
		default: /* DISABLED */
			image = ETeamscaleImage.DISABLED;
			break;
		}

		element.setIcon(image.getImageDescriptor());
	}

	/** {@inheritDoc} */
	@Override
	public void statusChanged() {
		updateTeamscaleConnectionStatus();
	}

	/** Update the server status icon. */
	public static void updateTeamscaleConnectionStatus() {
		@SuppressWarnings("cast") // Cast required for Eclipse 4.4.2 or older
		ICommandService commandService = (ICommandService) PlatformUI.getWorkbench().getService(ICommandService.class);
		if (commandService != null) {
			commandService.refreshElements(EclipseIds.Commands.OPEN_SERVER_PREFERENCES, null);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void dispose() {
		TeamscaleServerStatus.getInstance().removeStatusChangeListener(this);
		super.dispose();
	}
}
