/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.prefs;

import org.conqat.ide.commons.ui.dialog.FormInputDialog;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.IMessageProvider;
import org.eclipse.jface.fieldassist.ControlDecoration;
import org.eclipse.jface.fieldassist.FieldDecorationRegistry;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

import com.teamscale.commons.service.ServiceUtils;
import com.teamscale.ide.client.ConnectionValidationJob;
import com.teamscale.ide.client.NamedServerDetails;
import com.teamscale.ide.client.TeamscaleConnectionRegistry;

/**
 * A dialog for editing Teamscale server connections.
 */
public class ServerDetailsDialog extends FormInputDialog {

	/** Button id of the "Validate" button. */
	private static final int VALIDATE_BUTTON_ID = IDialogConstants.CLIENT_ID;

	/** Symbolic name of the server. */
	private String name;

	/** URL of the server. */
	private String url;

	/** Username for connection. */
	private String username;

	/** Password for connection. */
	private String password;

	/** Whether the server is enabled. */
	private boolean enabled;

	/** Whether the server trusts all certificates. */
	private boolean trustAllCertificates;

	/** Text field for the symbolic name. */
	private Text nameText;

	/** Text field for the URL. */
	private Text urlText;

	/** Text field for the user name. */
	private Text usernameText;

	/** Text field for the password. */
	private Text passwordText;

	/** The enabled check-box. */
	private Button enabledCheckBox;

	/** The trustAllCertificates check-box. */
	private Button trustAllCertificatesCheckBox;

	/** The check connection button. */
	private Button validationButton;

	/** Job used to validate the server connection. */
	private ConnectionValidationJob connectionValidationJob;

	/** Whether the enable checkbox is enabled to click or not. */
	private Boolean allowToAddDisabledServer = true;

	/** Constructor. */
	public ServerDetailsDialog(Shell parentShell, String title) {
		this(parentShell, "", "", "", "", true, false, title);
	}

	/** Constructor. */
	public ServerDetailsDialog(Shell parentShell, String title, Boolean allowToAddDisabledServer) {
		this(parentShell, "", "", "", "", true, false, title);
		this.allowToAddDisabledServer = allowToAddDisabledServer;
	}

	/** Constructor. */
	public ServerDetailsDialog(Shell parentShell, String initialServerName, String initialUrl, String initialUsername,
			String initialPassword, boolean initialEnabled, boolean initialTrustAllCertificates, String title) {
		super(parentShell, title, title, "Please enter the server details.");
		setShellStyle(getShellStyle() | SWT.RESIZE);
		setHelpAvailable(false);

		this.name = initialServerName;
		this.url = initialUrl;
		this.username = initialUsername;
		this.password = initialPassword;
		this.enabled = initialEnabled;
		this.trustAllCertificates = initialTrustAllCertificates;
	}

	/** Returns the symbolic name of the server. */
	public String getServerName() {
		return name;
	}

	/** Returns the URL of the server. */
	public String getServerUrl() {
		return url;
	}

	/**
	 * @see #username
	 */
	public String getUsername() {
		return username;
	}

	/** Returns the password. */
	public String getPassword() {
		return password;
	}

	/** Returns enabled. */
	public boolean isEnabled() {
		return enabled;
	}

	public boolean isTrustAllCertificates() {
		return trustAllCertificates;
	}

	/** {@inheritDoc} */
	@Override
	protected void createFields(Composite formGrid) {
		createNameField();
		createUrlField();
		createUsernameField();
		createPasswordField();
		createEnabledField();
		createTrustAllCertificatesField();
	}

	/** Creates text field for symbolic server name. */
	private void createNameField() {
		nameText = addText("Name:", name, 0, new TextValidatorBase() {
			@Override
			protected String validate(String text) {
				if (text.isEmpty()) {
					return "Missing name";
				}
				if (!text.equals(name) && TeamscaleConnectionRegistry.getInstance().containsServer(text)) {
					return "Name already used.";
				}
				return null;
			}
		});
	}

	/** Creates text field for URL. */
	private void createUrlField() {
		urlText = addText("URL:", url, 0, new TextValidatorBase() {
			@Override
			protected String validate(String text) {
				if (text.isEmpty()) {
					return "Missing URL";
				}
				if (!ServiceUtils.isValidServerAddress(text)) {
					return "Malformed URL";
				}
				return null;
			}
		});
	}

	/** Creates text field for user name. */
	private void createUsernameField() {
		usernameText = addText("Username:", username, 0, new TextValidatorBase() {
			@Override
			protected String validate(String text) {
				if (text.isEmpty()) {
					return "Missing username";
				}
				return null;
			}
		});
	}

	/** Creates text field for password. */
	private void createPasswordField() {
		passwordText = addText("IDE Access Key:", password, SWT.PASSWORD, new TextValidatorBase() {
			@Override
			protected String validate(String text) {
				if (text.isEmpty()) {
					return "Missing password";
				}
				return null;
			}
		});
	}

	/** Creates enabled check-box. */
	private void createEnabledField() {
		enabledCheckBox = addCheckBox("Enabled", enabled);
		enabledCheckBox.setEnabled(allowToAddDisabledServer);
	}

	/** Creates trustAllCertificates check-box. */
	private void createTrustAllCertificatesField() {
		trustAllCertificatesCheckBox = addCheckBox("Trust all certificates", trustAllCertificates);

		ControlDecoration decoration = new ControlDecoration(trustAllCertificatesCheckBox, SWT.LEFT | SWT.TOP);
		decoration.setImage(FieldDecorationRegistry.getDefault().getFieldDecoration(FieldDecorationRegistry.DEC_WARNING)
				.getImage());
		String warning = "Without TLS certificate validation, the security of server communication cannot be guaranteed.";
		decoration.setDescriptionText(warning);

		showDecoration(decoration, trustAllCertificates);
		trustAllCertificatesCheckBox.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				boolean trustAllCertificates = trustAllCertificatesCheckBox.getSelection();
				showDecoration(decoration, trustAllCertificates);
				if (trustAllCertificates) {
					// Explicitly open warning
					decoration.showHoverText(warning);
				}
			}
		});
	}

	private static void showDecoration(ControlDecoration decoration, boolean show) {
		if (show) {
			decoration.show();
		} else {
			decoration.hide();
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void formValidationStatusChanged(boolean allValid) {
		getButton(IDialogConstants.OK_ID).setEnabled(allValid);
		validationButton.setEnabled(allValid);
	}

	/** {@inheritDoc} */
	@Override
	protected void createButtonsForButtonBar(Composite parent) {
		super.createButtonsForButtonBar(parent);
		validationButton = createButton(parent, VALIDATE_BUTTON_ID, "Validate", false);
	}

	/** {@inheritDoc} */
	@Override
	protected void buttonPressed(int buttonId) {
		if (buttonId == VALIDATE_BUTTON_ID) {
			validateConnection();
		} else {
			super.buttonPressed(buttonId);
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void cancelPressed() {
		cancelConnectionValidationJob();
		super.cancelPressed();
	}

	/** {@inheritDoc} */
	@Override
	protected void okPressed() {
		cancelConnectionValidationJob();
		name = nameText.getText();
		url = urlText.getText();
		username = usernameText.getText();
		password = passwordText.getText();
		enabled = enabledCheckBox.getSelection();
		trustAllCertificates = trustAllCertificatesCheckBox.getSelection();
		super.okPressed();
	}

	/** Start a {@link Job} to validate the connection configuration. */
	private void validateConnection() {
		setMessage("Validating connection settings...");
		validationButton.setEnabled(false);
		setFormEnabled(false);
		connectionValidationJob = new ConnectionValidationJob("Validate Teamscale server connection",
				new NamedServerDetails("dummy", urlText.getText(), usernameText.getText(), passwordText.getText(),
						false, trustAllCertificatesCheckBox.getSelection()),
				new ConnectionValidationJobCallback());
		connectionValidationJob.schedule();
	}

	/** Callback function object class for the ConnectionValidationJob. */
	private class ConnectionValidationJobCallback implements ConnectionValidationJob.ValidationFinishedCallback {
		/** {@inheritDoc} */
		@Override
		public void execute(NamedServerDetails serverDetails, final boolean isReachable, final String errorMessage) {
			ServerDetailsDialog.this.getShell().getDisplay().asyncExec(() -> {
				if (validationButton == null || validationButton.isDisposed()) {
					// nothing to do, dialog has been closed already.
					return;
				}
				validationButton.setEnabled(true);
				setFormEnabled(true);

				if (isReachable) {
					setMessage("Successfully connected to server.", IMessageProvider.INFORMATION);
				} else {
					setMessage("Could not connect to server: " + errorMessage, IMessageProvider.ERROR);
				}
			});
		}

	}

	/**
	 * Cancel validation job if it exists. This does not actually end a running job,
	 * as there is no way of canceling the only blocking operation in the job.
	 * Rather this cancels possible scheduled jobs that are not yet running.
	 */
	private void cancelConnectionValidationJob() {
		if (connectionValidationJob != null) {
			connectionValidationJob.cancel();
			connectionValidationJob = null;
		}
	}
}
