/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.prefs;

import static java.util.Comparator.comparing;
import static java.util.stream.Collectors.toList;

import java.text.MessageFormat;
import java.util.List;

import org.conqat.ide.commons.ui.jface.CRUDListWidget;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.jface.resource.LocalResourceManager;
import org.eclipse.jface.resource.ResourceManager;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;

import com.teamscale.commons.service.client.ServerDetails;
import com.teamscale.ide.EclipseUtils;
import com.teamscale.ide.client.ETeamscaleServerStatus;
import com.teamscale.ide.client.ITeamscaleServerStatusChangeListener;
import com.teamscale.ide.client.NamedServerDetails;
import com.teamscale.ide.client.TeamscaleConnectionRegistry;
import com.teamscale.ide.client.TeamscaleServerStatus;
import com.teamscale.ide.findings.view.FindingsMarkerView;
import com.teamscale.ide.ui.images.ETeamscaleImage;

/**
 * A list of registered Teamscale servers that can be edited using create,
 * update, delete operations.
 */
class ServerDetailsList extends CRUDListWidget<NamedServerDetails> implements ITeamscaleServerStatusChangeListener {

	/** Constructor. */
	public ServerDetailsList(Composite parent) {
		super(parent);

		addColumn("Name");
		addColumn("URL");
		addColumn("Username");
		addColumn("Status");

		setContentProvider(new TeamscaleConnectionRegistryContentProvider());
		setLabelProvider(new TeamscaleConnectionLabelProvider());

		setInput(getRegistry());

		TeamscaleServerStatus.getInstance().addStatusChangeListener(this);
	}

	/** {@inheritDoc} */
	@Override
	public void dispose() {
		TeamscaleServerStatus.getInstance().removeStatusChangeListener(this);
		super.dispose();
	}

	/** Revert the changes. */
	protected void revertChanges() {
		getRegistry().restoreServers();
	}

	/** Perisits the changes. */
	protected void persistTemporaryChanges() {
		getRegistry().persistServers();
	}

	/** {@inheritDoc} */
	@Override
	protected void add() {
		NamedServerDetails server = EclipseUtils.openServerDetailsDialog("Add new Teamscale Server");
		if (server != null) {
			getRegistry().putServerTemporarily(server.getSymbolicName(), server);

		}
	}

	/** {@inheritDoc} */
	@Override
	protected void edit(NamedServerDetails selectedElement) {
		ServerDetailsDialog dialog = new ServerDetailsDialog(getShell(), selectedElement.getSymbolicName(),
				selectedElement.getUrl(), selectedElement.getUsername(), selectedElement.getPassword(),
				selectedElement.isEnabled(), selectedElement.isTrustAllCertificates(), "Edit Teamscale Server");

		if (dialog.open() == Window.OK) {
			getRegistry().removeServerTemporarily(selectedElement.getSymbolicName());
			NamedServerDetails serverDetails = createServerDetailsFromDialog(dialog);
			getRegistry().putServerTemporarily(dialog.getServerName(), serverDetails);
		}
	}

	/** Create {@link ServerDetails} from user input in dialog. */
	private static NamedServerDetails createServerDetailsFromDialog(ServerDetailsDialog dialog) {
		return new NamedServerDetails(dialog.getServerName(), dialog.getServerUrl(), dialog.getUsername(),
				dialog.getPassword(), dialog.isEnabled(), dialog.isTrustAllCertificates());
	}

	/** {@inheritDoc} */
	@Override
	protected void delete(NamedServerDetails selectedElement) {
		boolean deleteConfirmed = MessageDialog.openConfirm(getShell(), "Delete Teamscale Server",
				MessageFormat.format("Really delete ‘{0}’?", selectedElement.getSymbolicName()));

		if (deleteConfirmed) {
			getRegistry().removeServerTemporarily(selectedElement.getSymbolicName());
		}
	}

	/** Returns the Teamscale registry instance. */
	private static TeamscaleConnectionRegistry getRegistry() {
		return TeamscaleConnectionRegistry.getInstance();
	}

	/** Clears the registry temporarily. */
	protected void clear() {
		getRegistry().removeAllServersTemporarily();
		super.refresh();
	}

	/** {@inheritDoc} */
	@Override
	public void statusChanged() {
		FindingsMarkerView.removeErrorMessage();
	}

	/**
	 * Content provider for connection configurations from
	 * {@link TeamscaleConnectionRegistry}.
	 */
	private static class TeamscaleConnectionRegistryContentProvider implements IStructuredContentProvider {
		/** {@inheritDoc} */
		@Override
		public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
			// do nothing
		}

		/** {@inheritDoc} */
		@Override
		public void dispose() {
			// do nothing
		}

		/** {@inheritDoc} */
		@Override
		public Object[] getElements(Object inputElement) {
			return getNamedServers((TeamscaleConnectionRegistry) inputElement).toArray();
		}

		/**
		 * Returns a sorted list of the servers configured in
		 * {@link TeamscaleConnectionRegistry}.
		 */
		private static List<NamedServerDetails> getNamedServers(TeamscaleConnectionRegistry registry) {
			return registry.getAllServers().values().stream().sorted(comparing(NamedServerDetails::getSymbolicName))
					.collect(toList());
		}
	}

	/** Label provider for the entries of the registry. */
	private static class TeamscaleConnectionLabelProvider extends LabelProvider implements ITableLabelProvider {

		private final ResourceManager resourceManager = new LocalResourceManager(JFaceResources.getResources());

		/** {@inheritDoc} */
		@Override
		public Image getColumnImage(Object element, int columnIndex) {
			if (columnIndex == 3) {
				String symbolicName = ((NamedServerDetails) element).getSymbolicName();
				ETeamscaleServerStatus status = TeamscaleServerStatus.getInstance().getStatus(symbolicName)
						.orElse(ETeamscaleServerStatus.DISABLED);
				switch (status) {
				case REACHABLE:
					return resourceManager.createImage(ETeamscaleImage.REACHABLE.getImageDescriptor());
				case UNREACHABLE:
					return resourceManager.createImage(ETeamscaleImage.UNREACHABLE.getImageDescriptor());
				case TEMPORARILY_DISABLED:
					return resourceManager.createImage(ETeamscaleImage.MIXED.getImageDescriptor());
				default: /* DISABLED */
					return resourceManager.createImage(ETeamscaleImage.DISABLED.getImageDescriptor());
				}
			}
			return null;
		}

		/** {@inheritDoc} */
		@Override
		public String getColumnText(Object element, int columnIndex) {
			NamedServerDetails serverDetails = (NamedServerDetails) element;
			switch (columnIndex) {
			case 0:
				return serverDetails.getSymbolicName();
			case 1:
				return serverDetails.getUrl();
			case 2:
				return serverDetails.getUsername();
			case 3:
				return null;
			default:
				throw new IndexOutOfBoundsException("Invalid column index: " + columnIndex);
			}
		}

		@Override
		public void dispose() {
			resourceManager.dispose();
		}
	}

}