/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.prefs;

import org.conqat.ide.commons.ui.editor.EditorUtils;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.jface.preference.PreferencePage;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchPreferencePage;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.texteditor.ITextEditor;

import com.teamscale.ide.TeamscaleIDEActivator;
import com.teamscale.ide.commons.notification.ENotificationCause;
import com.teamscale.ide.eclipse.EclipseIds;
import com.teamscale.ide.findings.FindingsMarkerUtils;
import com.teamscale.ide.findings.view.FindingsMarkerView;

/**
 * Preference page for managing Teamscale servers.
 */
public class TeamscaleServerPreferencePage extends PreferencePage implements IWorkbenchPreferencePage {

	/** The ServerDetailsList-Widget. */
	private ServerDetailsList serverList;

	/** Constructor */
	public TeamscaleServerPreferencePage() {
		setPreferenceStore(TeamscaleIDEActivator.getDefault().getPreferenceStore());
		setDescription("Teamscale Servers");
		setTitle("Server preferences");
	}

	/** {@inheritDoc} */
	@Override
	protected Control createContents(Composite parent) {
		serverList = new ServerDetailsList(parent);
		return serverList;
	}

	/** {@inheritDoc} */
	@Override
	public void init(IWorkbench workbench) {
		// Do nothing
	}

	/** {@inheritDoc} */
	@Override
	public boolean performOk() {
		serverList.persistTemporaryChanges();
		removeAllFindingMarkersFromWorkspace();
		refreshFindingsView();
		refreshFindingsForActiveEditor();
		return super.performOk();
	}

	/** {@inheritDoc} */
	@Override
	public boolean performCancel() {
		serverList.revertChanges();
		return super.performCancel();
	}

	/** {@inheritDoc} */
	@Override
	protected void performDefaults() {
		serverList.clear();
		super.performDefaults();
	}


	/** Removes all finding markers from all workspace projects. */
	private static void removeAllFindingMarkersFromWorkspace() {
		for (IProject project : getWorkspaceProjects()) {
			FindingsMarkerUtils.removeAllMarkers(project);
		}
	}

	/** Refresh findings view. */
	private static void refreshFindingsView() {
		IWorkbenchPage page = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();
		FindingsMarkerView view = (FindingsMarkerView) page.findView(EclipseIds.Parts.FINDINGS);
		if (view != null) {
			view.refresh();
		}
	}

	/** Refresh findings for active editor. */
	private static void refreshFindingsForActiveEditor() {
		IEditorPart editor = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getActiveEditor();
		if (editor instanceof ITextEditor) {
			EditorUtils.fetchFindingsForEditor((ITextEditor) editor, ENotificationCause.INDIRECT);
		}
	}

	/** @return An array of all projects in the workspace. */
	private static IProject[] getWorkspaceProjects() {
		return ResourcesPlugin.getWorkspace().getRoot().getProjects();
	}
}
