/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.properties;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Queue;
import java.util.Set;

import org.conqat.ide.commons.ui.adt.AdtUtils;
import org.conqat.ide.commons.ui.jface.DialogUtils;
import org.conqat.ide.commons.ui.logging.LoggingUtils;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.ui.progress.UIJob;

import com.google.common.collect.ImmutableSet;
import com.teamscale.commons.service.client.ServerDetails;
import com.teamscale.ide.EclipseUtils;
import com.teamscale.ide.TeamscaleIDEActivator;
import com.teamscale.ide.client.NamedServerDetails;
import com.teamscale.ide.client.TeamscaleConnectionRegistry;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.services.ITeamscaleConnectorService;
import com.teamscale.ide.commons.model.PrefixMapping;
import com.teamscale.ide.commons.model.ProjectMapping;
import com.teamscale.ide.eclipse.plugin.EclipseServiceRegistry;
import com.teamscale.ide.eclipse.util.LegacyConversionUtils;

/**
 * Calls the service with a file as argument and returns the possible prefixes
 * to the callback.
 */
public class AutoMappingJob extends UIJob {

	/** How many files should be used to search for mappings. */
	private static final int NUMBER_OF_FILES_FOR_AUTODETECT = 5;

	/**
	 * Possible (default) extensions for the languages for which we support
	 * auto-mapping.
	 */
	private static final Set<String> AUTOMAPPABLE_EXTENSIONS = ImmutableSet.of("java", // Java
			"c", "h", "inc", // C
			"cpp", "hpp", "cxx", "hxx", "cc", "hh", "h", "inc", "inl", // C++
			"py" // Python
	);

	/** The handler to call after the thread has finished. */
	private final AutoMappingResultHandler handler;

	/** The paths to files which are used to find the proper prefixes. */
	private final List<String> paths = new ArrayList<>();

	/** The Teamscale servername. */
	private String servername;

	/** The corresponding project. */
	private final IProject project;

	/** A Teamscale project. */
	private final String teamscaleProject;

	private final ITeamscaleConnectorService connectorService;

	/**
	 * Constructor.
	 *
	 * @param path
	 *            A path to a local resource, it can be <code>null</code>. If it is
	 *            <code>null</code>, then this job tries to find a source file
	 *            within the given project.
	 *
	 *            The path is used to find a matching file on the Teamscale server.
	 *
	 * @param teamscaleProject
	 *            A project on the Teamscale server, can be <code>null</code>. Itf
	 *            it is <code>null</code>, then the job searches every project for
	 *            matching mappings.
	 */
	public AutoMappingJob(String servername, IProject project, String path, AutoMappingResultHandler handler,
			String teamscaleProject) {
		super("Calculating project mappings...");
		this.handler = handler;
		if (path != null) {
			paths.add(path);
		}
		this.servername = servername;
		this.project = project;
		this.teamscaleProject = teamscaleProject;

		connectorService = EclipseServiceRegistry.getInstance().getService(ITeamscaleConnectorService.class);
	}

	/** Search a source file from a folder for the service call. */
	private static String getFileFromFolder(IFolder parent) throws CoreException {
		Queue<IResource> projectMembers = new LinkedList<>();
		projectMembers.addAll(Arrays.asList(parent.members()));

		while (projectMembers.size() > 0) {
			IResource resource = projectMembers.poll();
			if (resource instanceof IFile && isSourceFile((IFile) resource)) {
				return resource.getProjectRelativePath().toString();
			} else if (resource instanceof IFolder) {
				IFolder folder = (IFolder) resource;
				projectMembers.addAll(Arrays.asList(folder.members()));
			}
		}
		return null;
	}

	/**
	 * Search source files from a project for the service call.
	 * 
	 * @param project
	 *            the project to be searched for files
	 * @param count
	 *            maximum number of files searched (must be > 0)
	 */
	public static List<String> getFilesFromProject(IProject project, int count) {
		List<String> filePaths = new ArrayList<>();
		try {
			Queue<IResource> folders = new LinkedList<>();
			folders.addAll(Arrays.asList(project.members(IResource.FOLDER)));

			while (folders.size() > 0 && filePaths.size() < count) {
				IResource resource = folders.poll();
				if (resource instanceof IFolder) {
					IFolder folder = (IFolder) resource;
					String path = getFileFromFolder(folder);
					if (path != null) {
						filePaths.add(path);
					}
					folders.addAll(Arrays.asList(folder.members(IResource.FOLDER)));
				}
			}

		} catch (CoreException e) {
			LoggingUtils.error(TeamscaleIDEActivator.getDefault(), e);
		}

		return filePaths;
	}

	/** Checks by the extension if a file is a source file. */
	public static boolean isSourceFile(IFile resource) {
		String extension = resource.getFileExtension();
		if (extension == null) {
			return false;
		}
		return AUTOMAPPABLE_EXTENSIONS.contains(extension.toLowerCase(Locale.ROOT));
	}

	/** {@inheritDoc} */
	@Override
	public IStatus runInUIThread(IProgressMonitor monitor) {
		monitor.beginTask("Calculating project mappings...", IProgressMonitor.UNKNOWN);
		Set<PrefixMapping> mappings = new HashSet<>();
		try {

			servername = getServerName();
			if (servername == null) {
				monitor.done();
				return Status.CANCEL_STATUS;
			}
			if (paths.isEmpty()) {
				paths.addAll(getFilesFromProject(project, NUMBER_OF_FILES_FOR_AUTODETECT));
			}
			if (paths.isEmpty()) {
				if (AdtUtils.isAdtProject(project)) {
					mappings.add(AdtUtils.ABAP_DEFAULT_PREFIX_MAPPING);
					monitor.done();
					return Status.OK_STATUS;
				}
				// if no source file is found.
				handler.handleMapping(Collections.emptySet(), servername);
				monitor.done();
				return Status.CANCEL_STATUS;
			}

			ServerDetails serverDetails = TeamscaleConnectionRegistry.getInstance().getServer(servername);
			if (serverDetails == null) {
				monitor.done();
				return Status.CANCEL_STATUS;
			}
			// FIXME Use correct branch information instead of default branch
			for (ProjectMapping mapping : connectorService.getProjectMappings(
					LegacyConversionUtils.toCommonsNamedServerDetails((NamedServerDetails) serverDetails),
					teamscaleProject, "", paths)) {
				mappings.add(new PrefixMapping(mapping));
			}
		} catch (NetworkExceptionBase e) {
			LoggingUtils.error(TeamscaleIDEActivator.getDefault(), e);
		}
		handler.handleMapping(mappings, servername);
		monitor.done();
		return Status.OK_STATUS;
	}

	/**
	 * If there is no server configured then open the server preference page.
	 */
	private String getServerName() {
		if (!getEnabledServerList().contains(servername)) {
			servername = null;
		}

		if (servername == null || servername.isEmpty()) {
			Collection<String> serverlist = getEnabledServerList();
			if (serverlist.isEmpty()) {
				NamedServerDetails server = EclipseUtils.openServerDetailsDialog("Add new Teamscale Server", false);
				if (server != null) {
					TeamscaleConnectionRegistry.getInstance().putServer(server.getSymbolicName(), server);
					return server.getSymbolicName();
				}
			} else if (serverlist.size() == 1) {
				return (String) serverlist.toArray()[0];
			} else {
				return DialogUtils.showListDialog(new ArrayList<>(serverlist),
						"Select a server to search for the project.", "Select the Teamscale server for this project",
						new LabelProvider());
			}
		}
		return servername;
	}

	/** Gets the enabled servers. */
	private static Collection<String> getEnabledServerList() {
		return TeamscaleConnectionRegistry.getInstance().getEnabledServerNames();
	}
}
