/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.properties;

import org.conqat.ide.commons.ui.dialog.FormInputDialog;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

import com.teamscale.ide.client.PrefixMappingRegistry;
import com.teamscale.ide.commons.model.PrefixMapping;

/**
 * Dialog for displaying and editing a {@link PrefixMapping}.
 */
public class PrefixMappingDetailsDialog extends FormInputDialog {

	/**
	 * Error message that is shown when user enters an already existing prefix.
	 */
	private static final String PREFIX_ALREADY_EXISTS = "Prefix already exists";

	/** Text widget for the local prefix. */
	private Text localPrefixText;

	/** Text widget for the remote Teamscale prefix. */
	private Text teamscalePrefixText;

	/** The local prefix. */
	private String localPrefix;

	/** The remote Teamscale prefix. */
	private String teamscalePrefix;

	/**
	 * In edit mode the dialog will not show an error if the entered values match
	 * the initial values.
	 */
	private boolean editMode;

	/**
	 * Instance of {@link PrefixMappingRegistry} used to check that an entered
	 * prefix is not already used
	 */
	private final PrefixMappingRegistry mappingRegistry;

	/** Constructor for dialog without pre-filled values. */
	public PrefixMappingDetailsDialog(Shell parentShell, PrefixMappingRegistry mappingRegistry, String title) {
		this(parentShell, mappingRegistry, title, "", "");
	}

	public PrefixMappingDetailsDialog(Shell parentShell, PrefixMappingRegistry mappingRegistry, String title,
			String initialLocalPrefix, String initialTeamscalePrefix) {
		super(parentShell, title, title, "Please enter the mapping's details.");
		setShellStyle(getShellStyle() | SWT.RESIZE);
		setHelpAvailable(false);

		this.localPrefix = initialLocalPrefix;
		this.teamscalePrefix = initialTeamscalePrefix;

		this.mappingRegistry = mappingRegistry;
	}

	@Override
	public void create() {
		super.create();
		Point size = new Point(350, 280);
		getShell().setSize(size);
		getShell().setMinimumSize(size);
	}

	@Override
	protected void createFields(Composite formGrid) {
		localPrefixText = addText("Local Prefix:", localPrefix, SWT.NONE, new TextValidatorBase() {
			@Override
			protected String validate(String text) {
				if (!(editMode && text.equals(localPrefix)) && mappingRegistry.localPrefixExists(text)) {
					return PREFIX_ALREADY_EXISTS;
				}
				return null;
			}
		});

		teamscalePrefixText = addText("Teamscale Prefix:", teamscalePrefix, SWT.NONE, null);
	}

	/** Sets edit mode. See {@link #editMode}. */
	public void setEditMode(boolean editMode) {
		this.editMode = editMode;
	}

	@Override
	protected void formValidationStatusChanged(boolean valid) {
		getButton(Window.OK).setEnabled(valid);
	}

	@Override
	protected void okPressed() {
		localPrefix = localPrefixText.getText();
		teamscalePrefix = teamscalePrefixText.getText();
		super.okPressed();
	}

	/** Returns local prefix as entered by the user. */
	public String getLocalPrefix() {
		return localPrefix;
	}

	/** Returns Teamscale prefix as entered by the user. */
	public String getTeamscalePrefix() {
		return teamscalePrefix;
	}
}
