/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.properties;

import java.text.MessageFormat;

import org.conqat.ide.commons.ui.jface.CRUDListWidget;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;

import com.teamscale.ide.client.PrefixMappingRegistry;
import com.teamscale.ide.commons.model.PrefixMapping;

/**
 * A list that allows create, update, delete operations on prefix mappings for a
 * project.
 */
public class PrefixMappingListWidget extends CRUDListWidget<PrefixMapping> {

	/** The mapping registry for the project. */
	private final PrefixMappingRegistry mappingRegistry;

	/** Constructor */
	public PrefixMappingListWidget(Composite parent, PrefixMappingRegistry mappingRegistry) {

		super(parent);
		this.mappingRegistry = mappingRegistry;

		addColumn("Local Prefix");
		addColumn("Teamscale Prefix");

		setContentProvider(new PrefixMappingRegistryContentProvider(mappingRegistry));
		setLabelProvider(new PrefixMappingLabelProvider());

		setInput(mappingRegistry);
	}

	/** {@inheritDoc} */
	@Override
	public void refresh() {
		// make refresh public available
		super.refresh();
	}

	/** {@inheritDoc} */
	@Override
	protected void add() {
		PrefixMappingDetailsDialog dialog = new PrefixMappingDetailsDialog(getShell(), mappingRegistry,
				"Add new Teamscale Mapping");

		if (dialog.open() == Window.OK) {
			PrefixMapping mapping = createPrefixMapping(dialog);
			mappingRegistry.addMapping(mapping);
			refresh();
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void edit(PrefixMapping selectedElement) {
		PrefixMappingDetailsDialog dialog = new PrefixMappingDetailsDialog(getShell(), mappingRegistry,
				"Edit Teamscale Mapping", selectedElement.getLocalPrefix(), selectedElement.getTeamscalePrefix());
		dialog.setEditMode(true);

		if (dialog.open() == Window.OK) {
			PrefixMapping mapping = createPrefixMapping(dialog);
			mappingRegistry.removeMapping(selectedElement);
			mappingRegistry.addMapping(mapping);
			refresh();
		}

	}

	/** Creates {@link PrefixMapping} from current widget contents in dialog. */
	private static PrefixMapping createPrefixMapping(PrefixMappingDetailsDialog dialog) {
		return new PrefixMapping(dialog.getLocalPrefix(), dialog.getTeamscalePrefix());
	}

	/** {@inheritDoc} */
	@Override
	protected void delete(PrefixMapping selectedElement) {
		boolean deleteConfirmed = MessageDialog.openConfirm(getShell(), "Delete Teamscale Mapping",
				MessageFormat.format("Really delete mapping ‘{0}’ → ‘{1}’?", selectedElement.getLocalPrefix(),
						selectedElement.getTeamscalePrefix()));

		if (deleteConfirmed) {
			mappingRegistry.removeMapping(selectedElement);
			refresh();
		}
	}

	/**
	 * Content provider for list of {@link PrefixMapping}s from
	 * {@link PrefixMappingRegistry}.
	 */
	private static class PrefixMappingRegistryContentProvider implements IStructuredContentProvider {

		/** Registry to get mappings from. */
		private final PrefixMappingRegistry registry;

		/** Constructor. */
		public PrefixMappingRegistryContentProvider(PrefixMappingRegistry registry) {
			this.registry = registry;
		}

		/** {@inheritDoc} */
		@Override
		public void dispose() {
			// Do nothing
		}

		/** {@inheritDoc} */
		@Override
		public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
			// Do nothing
		}

		/** {@inheritDoc} */
		@Override
		public Object[] getElements(Object inputElement) {
			return registry.getMappings().toArray();
		}
	}

	/**
	 * Label provider that maps {@link PrefixMapping}s to text in the respective
	 * columns.
	 */
	private static class PrefixMappingLabelProvider extends LabelProvider implements ITableLabelProvider {

		/** {@inheritDoc} */
		@Override
		public Image getColumnImage(Object element, int columnIndex) {
			return null;
		}

		/** {@inheritDoc} */
		@Override
		public String getColumnText(Object element, int columnIndex) {
			PrefixMapping prefixMapping = (PrefixMapping) element;
			switch (columnIndex) {
			case 0:
				return prefixMapping.getLocalPrefix();
			case 1:
				return prefixMapping.getTeamscalePrefix();
			default:
				throw new IndexOutOfBoundsException("Invalid column index: " + columnIndex);
			}
		}

	}
}
