/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package com.teamscale.ide.properties;

import static java.util.Objects.requireNonNull;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import org.conqat.ide.commons.ui.adt.AdtUtils;
import org.conqat.ide.commons.ui.logging.LoggingUtils;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.IJobChangeEvent;
import org.eclipse.core.runtime.jobs.JobChangeAdapter;
import org.eclipse.jface.layout.GridDataFactory;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.dialogs.PropertyPage;

import com.teamscale.ide.TeamscaleIDEActivator;
import com.teamscale.ide.client.NamedServerDetails;
import com.teamscale.ide.client.PrefixMappingRegistry;
import com.teamscale.ide.client.ServerDetailsJobBase;
import com.teamscale.ide.client.TeamscaleConnectionRegistry;
import com.teamscale.ide.commons.client.exceptions.NetworkExceptionBase;
import com.teamscale.ide.commons.client.services.ITeamscaleConnectorService;
import com.teamscale.ide.commons.model.PrefixMapping;
import com.teamscale.ide.eclipse.EclipseIds;
import com.teamscale.ide.eclipse.plugin.EclipseServiceRegistry;
import com.teamscale.ide.eclipse.util.LegacyConversionUtils;

/**
 * A property page that allows to configure the project specific settings of
 * Teamscale.
 */
public class TeamscalePropertyPage extends PropertyPage {

	/** Text to show if the server does not support branches. */
	public static final String NO_BRANCH_SUPPORT_TEXT = "This server does not support branches";

	/** A checkbox for turning Teamscale for this project on or off. */
	private Button teamscaleEnabledCheckBox;

	/** The widget for editing the {@link PrefixMapping}s for this project. */
	private PrefixMappingListWidget mappingList;

	/** Combo box for setting up server name. */
	private Combo serverNameCombo;

	/** Combo box for setting up Teamscale project name. */
	private Combo projectNameCombo;

	/** Combo box for selecting the branch */
	private Text overrideBranchNameTextField;

	/** The timeout for an SAP system update */
	private Combo timeoutCombo;

	/** List of configured servers. */
	private final List<String> serverList;

	/** Button to auto-map the project. **/
	private Button autoProjectMapping;

	/** Constructor. */
	public TeamscalePropertyPage() {
		serverList = TeamscaleConnectionRegistry.getInstance().getEnabledServerNames().stream().sorted()
				.collect(Collectors.toCollection(ArrayList::new));
	}

	/** {@inheritDoc} */
	@Override
	protected Control createContents(Composite parent) {
		Composite composite = new Composite(parent, SWT.NONE);
		composite.setLayout(new GridLayout(2, false));

		Label teamscaleEnabLabel = new Label(composite, SWT.NONE);
		teamscaleEnabLabel.setText("Enable Teamscale:");

		teamscaleEnabledCheckBox = createTeamscaleEnabledCheckBox(composite);
		GridDataFactory.fillDefaults().align(SWT.FILL, SWT.BEGINNING).applyTo(teamscaleEnabledCheckBox);

		Label serverNameLabel = new Label(composite, SWT.NONE);
		serverNameLabel.setText("Server Name:");

		createServerNameCombo(composite);

		Label projectNameLabel = new Label(composite, SWT.NONE);
		projectNameLabel.setText("Project Name:");

		projectNameCombo = new Combo(composite, SWT.NONE);
		GridDataFactory.fillDefaults().hint(50, SWT.DEFAULT).applyTo(projectNameCombo);

		createOverrideBranchNameField(composite);
		createSapSystemUpdateTimeoutCombo(composite);
		createTryAutoMappingButton(composite);
		createMappingListWidget(composite);

		restoreUIState();

		return composite;
	}

	private void createMappingListWidget(Composite composite) {
		if (isAbap()) {
			return;
		}
		mappingList = new PrefixMappingListWidget(composite, getRegistry());
		GridDataFactory.fillDefaults().span(2, 1).grab(true, true).applyTo(mappingList);
	}

	private void createOverrideBranchNameField(Composite composite) {
		if (isAbap()) {
			return;
		}
		Label overrideBranchNameLabel = new Label(composite, SWT.NONE);
		overrideBranchNameLabel.setText("Override Branch Name (?):");
		overrideBranchNameLabel.setToolTipText(
				"This is the branch name for retrieving data from the Teamscale server. If empty, the plugin uses "
						+ "auto-discovery for the currently checked out branch from VCS with a fallback to the "
						+ "Teamscale default branch. If non-empty, the plugin turns off auto-discovery of the branch "
						+ "from BCS and always uses the specified branch.");

		overrideBranchNameTextField = new Text(composite, SWT.SINGLE | SWT.BORDER);
		GridDataFactory.fillDefaults().hint(50, SWT.DEFAULT).applyTo(overrideBranchNameTextField);
	}

	/** Create the combo box for selecting an SAP system update timeout. */
	private void createSapSystemUpdateTimeoutCombo(Composite parent) {
		if (!isAbap()) {
			return;
		}
		Label timeoutComboLabel = new Label(parent, NONE);
		timeoutComboLabel.setText("SAP system update timeout (s):");
		timeoutComboLabel.setToolTipText(
				"The Eclipse plugin supports fetching findings of recently activated code, that is not yet known to "
						+ "the Teamscale server. Therefore, the plugin initiates an update process of all "
						+ "relevant Teamscale projects and waits until all projects have been updated before "
						+ "fetching findings. "
						+ "This parameter specifies the time in seconds the plugin waits for a successful "
						+ "update. When the update process could not finish before this timeout, a warning "
						+ "appears indicating that the findings might be outdated.");

		timeoutCombo = new Combo(parent, NONE);

		timeoutCombo.setItems(IntStream.of(1, 3, 5, 10, 20, 30, 60, 90, 120, 180, 300, 600).mapToObj(Integer::toString)
				.toArray(String[]::new));
		timeoutCombo.setText(Integer.toString(getRegistry().getTeamscaleSapSyncTimeout()));

		GridDataFactory.fillDefaults().hint(50, SWT.DEFAULT).applyTo(timeoutCombo);

		timeoutCombo.addSelectionListener(new SelectionListener() {
			@Override
			public void widgetSelected(SelectionEvent event) {
				getRegistry().setTeamscaleSapSyncTimeout(Integer.parseInt(timeoutCombo.getText()));
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent event) {
				widgetSelected(event);
			}
		});
	}

	/** Create the auto-mapping-button. */
	private void createTryAutoMappingButton(Composite parent) {
		if (isAbap()) {
			return;
		}
		autoProjectMapping = new Button(parent, SWT.NONE);
		autoProjectMapping.setText("Automatically configure Teamscale");
		autoProjectMapping.addListener(SWT.Selection, new Listener() {

			@Override
			public void handleEvent(Event event) {
				doAutoMapping();
			}
		});
	}

	/** Create the combo box for the server name. */
	private void createServerNameCombo(Composite parent) {
		serverNameCombo = new Combo(parent, SWT.NONE);
		serverNameCombo.setItems(serverList.toArray(new String[0]));
		GridDataFactory.fillDefaults().hint(50, SWT.DEFAULT).applyTo(serverNameCombo);

		serverNameCombo.addSelectionListener(new SelectionListener() {
			@Override
			public void widgetSelected(SelectionEvent event) {
				loadProjects(null);
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent event) {
				widgetSelected(event);
			}
		});
	}

	/** Gets project properties and sets state of UI widgets accordingly. */
	private void restoreUIState() {
		boolean teamscaleEnabled = getRegistry().isTeamscaleEnabled();
		teamscaleEnabledCheckBox.setSelection(teamscaleEnabled);
		setEnabled(teamscaleEnabled);

		String serverName = getRegistry().getTeamscaleServer();
		if (serverName != null) {
			serverNameCombo.setText(serverName);
		} else {
			serverNameCombo.select(0);
		}
		loadProjects(getRegistry().getTeamscaleProject());
	}

	/**
	 * Loads the projects for the selected server into the project combo box.
	 *
	 * @param projectName
	 *            Project name to select in the combo box. If <code>null</code> the
	 *            first project returned by the server will be selected.
	 */
	private void loadProjects(String projectName) {
		NamedServerDetails serverDetails = TeamscaleConnectionRegistry.getInstance()
				.getServer(serverNameCombo.getText());
		if (serverDetails != null) {
			getRegistry().setTeamscaleServer(serverNameCombo.getText());
			new LoadProjectsJob(serverDetails, projectName).schedule();
		}
	}

	/** Job that loads the projects list from a server. */
	private class LoadProjectsJob extends ServerDetailsJobBase {

		/**
		 * Project name to set in the combo box after loading the list. May be
		 * <code>null</code>.
		 */
		private final String projectName;

		private final ITeamscaleConnectorService connectorService;

		/**
		 * Constructor.
		 *
		 * @param projectName
		 *            The projectName that should be chosen if it exists in the new list
		 *            of projects. May be <code>null</code> in which case the first
		 *            element is chosen or the combo stays empty if there are no
		 *            projects on the server.
		 */
		public LoadProjectsJob(NamedServerDetails serverDetails, String projectName) {
			super("Loading project list", serverDetails);
			this.projectName = projectName;

			connectorService = EclipseServiceRegistry.getInstance().getService(ITeamscaleConnectorService.class);
		}

		@Override
		protected IStatus run(IProgressMonitor monitor) {
			try {
				Set<String> publicProjectIds = connectorService
						.getProjects(LegacyConversionUtils.toCommonsNamedServerDetails(serverDetails));
				String[] projectList = publicProjectIds.stream().toArray(String[]::new);
				setProjectsComboItems(projectList);
				setOverrideBranch();
				return Status.OK_STATUS;
			} catch (NetworkExceptionBase e) {
				LoggingUtils.error(TeamscaleIDEActivator.getDefault(), e);
				return new Status(IStatus.ERROR, EclipseIds.PLUGIN, e.toString(), e);
			}
		}

		/**
		 * Sets the items of the projects combo to the given values and sets the
		 * selected project. Runs in the UI thread.
		 */
		private void setProjectsComboItems(final String[] projects) {
			Display.getDefault().asyncExec(() -> {
				if (projectNameCombo.isDisposed()) {
					return;
				}
				projectNameCombo.setItems(projects);

				if (projectName != null && Arrays.asList(projects).contains(projectName)) {
					projectNameCombo.setText(projectName);
				} else if (projectNameCombo.getItemCount() > 0) {
					projectNameCombo.select(0);
				}
			});
		}

		/**
		 * Sets the override branch text field. Runs in the UI thread.
		 */
		private void setOverrideBranch() {
			Display.getDefault().asyncExec(() -> {
				PrefixMappingRegistry registry = getRegistry();
				if (registry.getOverrideBranch() != null && overrideBranchNameTextField != null) {
					overrideBranchNameTextField.setText(registry.getOverrideBranch());
				}
			});
		}

	}

	/** Create the checkbox that allows turning Teamscale on or off. */
	private Button createTeamscaleEnabledCheckBox(Composite parent) {
		final Button checkBox = new Button(parent, SWT.CHECK);
		checkBox.setText("Enable Teamscale for this project");

		checkBox.addSelectionListener(new SelectionListener() {
			@Override
			public void widgetSelected(SelectionEvent event) {
				widgetDefaultSelected(event);

				if (checkBox.getSelection()) {
					doAutoMapping();
				}
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent event) {
				setEnabled(((Button) event.getSource()).getSelection());
			}
		});

		return checkBox;
	}

	/** Calls the automapper and refreshes the GUI. */
	private void doAutoMapping() {
		IProject project = getElement().getAdapter(IProject.class);
		AutoMapping autoMapper = new AutoMapping(project, serverNameCombo.getText(), projectNameCombo.getText());

		autoMapper.tryAutoMapping(new JobChangeAdapter() {
			@Override
			public void done(IJobChangeEvent event) {
				restoreUIState();
				if (mappingList != null) {
					mappingList.refresh();
				}
			}
		});
		getRegistry().setTeamscaleEnabled(true);
	}

	private boolean isAbap() {
		return AdtUtils.isAdtProject(getProject());
	}

	/**
	 * Enables or disables the widgets of this page, except for the checkbox.
	 */
	private void setEnabled(boolean enabled) {
		serverNameCombo.setEnabled(enabled);
		projectNameCombo.setEnabled(enabled);

		setEnabledIfPresent(mappingList, enabled);
		setEnabledIfPresent(autoProjectMapping, enabled);
		setEnabledIfPresent(overrideBranchNameTextField, enabled);
		setEnabledIfPresent(timeoutCombo, enabled);
	}

	/** If the control is present, sets the enabled flag of the control */
	private static <T extends Control> void setEnabledIfPresent(T control, boolean enabled) {
		if (control != null) {
			control.setEnabled(enabled);
		}
	}

	/** {@inheritDoc} */
	@Override
	public boolean performOk() {
		PrefixMappingRegistry registry = getRegistry();
		registry.setTeamscaleEnabled(teamscaleEnabledCheckBox.getSelection());
		registry.setTeamscaleServer(serverNameCombo.getText());
		registry.setTeamscaleProject(projectNameCombo.getText());
		String branch = AdtUtils.ABAP_DEFAULT_BRANCH;
		if (overrideBranchNameTextField != null) {
			branch = overrideBranchNameTextField.getText();
		}
		registry.setOverrideBranch(branch);
		registry.persist();
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean performCancel() {
		getRegistry().restore();
		return true;
	}

	/**
	 * Returns the {@link PrefixMappingRegistry} associated with the project of this
	 * property page.
	 */
	private PrefixMappingRegistry getRegistry() {
		return PrefixMappingRegistry.getInstance(getProject());
	}

	/** Gets the corresponding project */
	private IProject getProject() {
		return requireNonNull(getElement().getAdapter(IProject.class));
	}
}
