package com.teamscale.index.metadata.abap;

import org.conqat.lib.commons.collections.CounterSet;
import org.conqat.lib.commons.js_export.ExportToTypeScript;
import org.conqat.lib.commons.test.IndexValueClass;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;

/**
 * Lists the possible types of ABAP programs. Read <a href=
 * "https://help.sap.com/doc/saphelp_nw73ehp1/7.31.19/en-us/fc/eb2d5a358411d1829f0000e829fbfe/content.htm?no_cache=true">docs</a>
 * for more infos.
 */
@ExportToTypeScript
@IndexValueClass
public enum EAbapProgramType {

	/**
	 * Used to communicate that some ABAP file has no program type assigned.
	 */
	NOT_ASSIGNED(' ', "Not assigned"),

	/**
	 * Represents an unknown (and potentially new) program type whose details can be
	 * retrieved from Teamscale's worker log and used to update this enumeration
	 */
	UNKNOWN_TYPE('~', "Unknown"),

	BEHAVIOR_DEFINITION('B', "Behavior definition"),

	FUNCTION_GROUP('F', "Function pool"),

	INCLUDE_PROGRAM('I', "Include program"),

	INTERFACE_POOL('J', "Interface pool"),

	CLASS_POOL('K', "Class pool"),

	MODULE_POOL('M', "Module pool"),

	DATABASE_PROCEDURE_PROXY('Q', "Database procedure proxy"),

	SUBROUTINE_POOL('S', "Subroutine pool"),

	TYPE_POOL('T', "Type pool"),

	XSLT_OR_ST_PROGRAM_TRANSFORMATION('X', "Program transformation"),

	EXECUTABLE_PROGRAM('1', "Executable program");

	private static final String TYPE_VALUE_PROPERTY = "typeValue";
	private static final String HUMAN_READABLE_NAME_PROPERTY = "humanReadableName";

	@JsonProperty(TYPE_VALUE_PROPERTY)
	private final char typeValue;

	@JsonProperty(HUMAN_READABLE_NAME_PROPERTY)
	private final String humanReadableName;

	@JsonCreator
	EAbapProgramType(@JsonProperty(TYPE_VALUE_PROPERTY) char typeValue,
			@JsonProperty(HUMAN_READABLE_NAME_PROPERTY) String humanReadableName) {
		this.typeValue = typeValue;
		this.humanReadableName = humanReadableName;
	}

	/**
	 * Derives a {@link EAbapProgramType} from some given type value. Returns
	 * <code>null</code> if the type value is not defined in this enumeration. For
	 * new program types, check the details in SAP GUI and update this enumeration.
	 * Check the details at transaction SE11 -> Open the domain SUBC -> click "Value
	 * Range". Read <a href=
	 * "https://answers.sap.com/questions/4359585/which-table-contains-the-text-to-program-type.html">docs</a>
	 * for more information.
	 */
	public static EAbapProgramType getProgramTypeFromValue(char typeValue) {
		for (EAbapProgramType programType : EAbapProgramType.values()) {
			if (programType != UNKNOWN_TYPE && programType.typeValue == typeValue) {
				return programType;
			}
		}

		return UNKNOWN_TYPE;
	}

	public char getTypeValue() {
		return typeValue;
	}

	/**
	 * Converts the enum to a human-readable metric.
	 * <p>
	 * We are using {@link CounterSet counter sets} because we don't lose the amount
	 * of files with the specific enum on aggregation in a folder.
	 */
	public static CounterSet<String> toMetric(EAbapProgramType type) {
		if (type == null || type == NOT_ASSIGNED) {
			return new CounterSet<>(NOT_ASSIGNED.humanReadableName, 1);
		}
		return new CounterSet<>(type.humanReadableName + " (" + type.typeValue + ")", 1);
	}
}
