package eu.cqse.check.framework.scanner.ambiguous_language;

import java.util.EnumSet;
import java.util.regex.Pattern;

import org.checkerframework.checker.nullness.qual.Nullable;
import org.conqat.lib.commons.string.LineSplitter;

import eu.cqse.check.framework.scanner.ELanguage;

/**
 * Searches for typical patterns of C++ and Objective-C in the given file text.
 * The two pattern-recognition methods are summarized in this class since we
 * need to skip C-Like comments for both heuristics.
 */
class CLikeLanguageDetector {

	/** Keywords in code that are very likely only used in C++. */
	private static final String[] CPP_COMMON_KEYWORDS = { "::", "using namespace", "namespace ", "try {", "constexpr ",
			"template<", "virtual " };

	/**
	 * Pattern to detect keywords that, if found, are very likely only used in C++.
	 * <p>
	 * The pattern is applied to each code line separately.
	 * <p>
	 * The "(?<!@)class" is a negative lookbehind. It ensures that there is no "@"
	 * before the "class" (that would be objective-c).
	 * <p>
	 * "(^|\s)" matches on a space or the beginning of the input (i.e., the
	 * beginning of a code line).
	 */
	private static final Pattern CPP_PATTERN = Pattern.compile("(?<!@)class\\s[A-Z]|=\\snew\\s|(^|\\s)throw\\s[A-Z]");

	/**
	 * Keywords in code that indicate Objective-C code uniquely. Note, that #import
	 * is a valid directive for MSVC. However, we assume separated Objective-C(++)
	 * and Windows C++ code, such that the appropriate keywords are picked.
	 */
	private static final String[] OBJECTIVE_C_COMMON_KEYWORDS = { "@end", "@interface", "@property", "@protocol",
			"@import", "@class", "#import" };

	/**
	 * Pattern to detect types that, if found, are very likely used in Objective-C.
	 */
	private static final Pattern OBJECTIVE_C_COMMON_TYPES_HINT_PATTERN = Pattern
			.compile("\\W?(NSString|NSMutableString|NSMutableArray|NSSet|NSDictionary|NSMutableDictionary|NSArray)\\W");

	/**
	 * Whether current line ends inside a multiline C/C++ comment (starting with
	 * <code>/*</code>). And therefore, whether the next line starts in a comment.
	 */
	private boolean currentlyInMultilineComment = false;

	/**
	 * The position in the currently analyzed line that until which we scanned. This
	 * is advanced in {@link #advancePositionInLine(String)} and has to be reset to
	 * 0 when starting on a new line.
	 */
	private int currentPositionInLine = 0;

	private final EnumSet<ELanguage> languagesToSearchFor;

	/**
	 * The languages for which we identified patterns in the code. This is always a
	 * subset of {@link #languagesToSearchFor}. This list can contain multiple
	 * languages if our patterns are ambiguous or if the code actually contains a
	 * mix of languages (e.g., C++ and Objective-C in an Objective-C++ file).
	 */
	private final EnumSet<ELanguage> detectedLanguagePatterns = EnumSet.noneOf(ELanguage.class);

	private CLikeLanguageDetector(EnumSet<ELanguage> languagesToSearchFor) {
		this.languagesToSearchFor = languagesToSearchFor;
	}

	/**
	 * Searches for typical patterns of C++ and Objective-C in the given file text.
	 * The two pattern-recognition methods are summarized in this class since we
	 * need to skip C-Like comments for both heuristics.
	 * <p>
	 * We stop iterating, if we could identify both {@link ELanguage#CPP} and
	 * {@link ELanguage#OBJECTIVE_C} or if we only need to check for one language
	 * and could identify it.
	 * 
	 * @param fileContent
	 *            The text of the file to analyze
	 * @param languagesToSearchFor
	 *            the languages to search for (either {@link ELanguage#OBJECTIVE_C},
	 *            {@link ELanguage#CPP}, or both).
	 * @return Returns the found languages (subset of
	 *         {@link #languagesToSearchFor}).
	 */
	public static EnumSet<ELanguage> scanForClikeLanguagePatterns(String fileContent,
			EnumSet<ELanguage> languagesToSearchFor) {
		CLikeLanguageDetector detector = new CLikeLanguageDetector(languagesToSearchFor);
		LineSplitter lineSplitter = new LineSplitter(fileContent);
		lineSplitter.setIncludeTrailingEmptyLine(true);
		for (String line : lineSplitter) {
			detector.advanceLine(line);
			if (detector.detectedLanguagePatterns.equals(languagesToSearchFor)) {
				// already found all patterns, no need to continue scanning
				return detector.detectedLanguagePatterns;
			}
		}
		return detector.detectedLanguagePatterns;
	}

	/**
	 * Scans the given line and tries to detect C++ or Objective-C code in it.
	 * <p>
	 * We first check whether the current line contains an end-of-line comment. If
	 * so, we select all code before the comment. If not, we check whether the
	 * current line contains starting and ending multi line comment literals. If
	 * both exist we filter out the comment and select the code before and after it.
	 * If only starting comment literals exist, we select all code before, if
	 * present, and store that we are now in a multi line comment.
	 */
	private void advanceLine(String line) {
		currentPositionInLine = 0;
		while (currentPositionInLine < line.length()) {
			String code = advancePositionInLine(line);
			if (code == null || code.isEmpty()) {
				continue;
			}
			if (languagesToSearchFor.contains(ELanguage.CPP) && !detectedLanguagePatterns.contains(ELanguage.CPP)
					&& isLikelyCppCode(code)) {
				detectedLanguagePatterns.add(ELanguage.CPP);
			}
			if (languagesToSearchFor.contains(ELanguage.OBJECTIVE_C)
					&& !detectedLanguagePatterns.contains(ELanguage.OBJECTIVE_C) && isLikelyObjectiveCCode(code)) {
				detectedLanguagePatterns.add(ELanguage.OBJECTIVE_C);
			}
		}
	}

	/**
	 * Continues scanning the code in the current line. Skips comments and returns
	 * non-comment code if any is found. Returns null if the method skips a comment.
	 */
	@Nullable
	private String advancePositionInLine(String line) {
		String code = null;
		if (currentlyInMultilineComment) {
			int commentEndIndex = line.indexOf("*/", currentPositionInLine);
			if (commentEndIndex != -1) {
				// the current multiline comment ends
				currentlyInMultilineComment = false;
				currentPositionInLine = commentEndIndex + 2;
			} else {
				// the current multiline comment extends until the end of the line
				currentPositionInLine = line.length();
				currentlyInMultilineComment = true;
			}
		} else {
			int multilineCommentStartIndex = line.indexOf("/*", currentPositionInLine);
			int endOfLineCommentStartIndex = line.indexOf("//", currentPositionInLine);
			if (endOfLineCommentStartIndex != -1
					&& (endOfLineCommentStartIndex < multilineCommentStartIndex || multilineCommentStartIndex == -1)) {
				// There is an end-of-line comment (terminating this line). The next line is
				// code again.
				code = line.substring(currentPositionInLine, endOfLineCommentStartIndex);
				currentlyInMultilineComment = false;
				currentPositionInLine = line.length();
			} else if (multilineCommentStartIndex != -1) {
				// a multiline comment starting here
				code = line.substring(currentPositionInLine, multilineCommentStartIndex);
				currentlyInMultilineComment = true;
				currentPositionInLine = multilineCommentStartIndex + 2;
			} else {
				// no multiline comment and no end-of-line comment. Everything is code.
				code = line.substring(currentPositionInLine);
				currentPositionInLine = line.length();
				currentlyInMultilineComment = false;
			}
		}
		return code;
	}

	/**
	 * Returns <code>true</code>, if the given line content represents C++ code. We
	 * search for indicator keywords that are unique to C++.
	 *
	 * @see #CPP_COMMON_KEYWORDS
	 * @see #CPP_PATTERN
	 */
	private static boolean isLikelyCppCode(String lineContent) {
		return containsOneOfKeywords(lineContent, CPP_COMMON_KEYWORDS) || CPP_PATTERN.matcher(lineContent).find();
	}

	/**
	 * Returns <code>true</code>, if the given line content represents Objective-C
	 * code. We search for indicator keywords that are unique to Objective-C.
	 *
	 * @see #OBJECTIVE_C_COMMON_KEYWORDS
	 * @see #OBJECTIVE_C_COMMON_TYPES_HINT_PATTERN
	 */
	private static boolean isLikelyObjectiveCCode(String lineContent) {
		return containsOneOfKeywords(lineContent, OBJECTIVE_C_COMMON_KEYWORDS)
				|| OBJECTIVE_C_COMMON_TYPES_HINT_PATTERN.matcher(lineContent).find();
	}

	/**
	 * Iterates through the given language-specific keywords and returns whether the
	 * given line contains any of them.
	 * <p>
	 * If a keyword starts with a letter, we also check that the match is not
	 * preceded by another char that is allowed as part of identifiers. This avoids
	 * matching, for example, "namespace" on "identifier_ending_with_namespace".
	 */
	private static boolean containsOneOfKeywords(String lineContent, String[] keywords) {
		for (String keyword : keywords) {
			int matchOffset = lineContent.indexOf(keyword);
			while (matchOffset != -1) {
				// We assume that the set of chars allowed in java identifiers is the same set
				// as in C/C++/Objective-C. That is reasonably close to the truth.
				if (!Character.isJavaIdentifierPart(keyword.charAt(0))) {
					return true;
				}
				boolean matchPreceededByIdentifierChar = matchOffset != 0
						&& Character.isJavaIdentifierPart(lineContent.charAt(matchOffset - 1));
				if (!matchPreceededByIdentifierChar) {
					return true;
				}
				matchOffset = lineContent.indexOf(keyword, matchOffset + 1);
			}
		}
		return false;
	}
}
