/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package eu.cqse.check.framework.shallowparser;

import java.util.Collections;
import java.util.Set;

import org.conqat.lib.commons.collections.CollectionUtils;

import eu.cqse.check.framework.shallowparser.framework.ShallowEntity;

/**
 * This class collects constants used for subtypes in a {@link ShallowEntity}.
 * The list of constants is not complete, as often the subtype is also taken
 * directly from the parsing context (e.g. a matched keyword). The purpose of
 * this class is only to eliminate redundancy between the parser and code
 * traversing the AST.
 * <p>
 * Convention is to have the name of the constant and the content the same
 * (while respecting naming conventions). Hence, changes to a value should be
 * reflected in the name.
 */
public class SubTypeNames {

	/** {@value} */
	public static final String ANGULAR_JS = "AngularJS";

	/** {@value} */
	public static final String ANGULAR_JS_MODULE_DECLARATION = "AngularJS module declaration";

	/** {@value} */
	public static final String YUI_MODULE = "YUI module";

	/** {@value} */
	public static final String FIELD = "field";

	/** {@value} */
	public static final String FILTER_FUNCTION = "filter function";

	/** {@value} */
	public static final String ASSIGNMENT = "assignment";

	/** {@value} */
	public static final String EVENT = "event";

	/** {@value} */
	public static final String SCRIPT = "script";

	/** {@value} */
	public static final String SCHEMA = "schema";

	/** {@value} */
	public static final String TEMPLATE = "template";

	/** {@value} */
	public static final String ANNOTATION = "annotation";

	/** {@value} */
	public static final String NAMESPACE = "namespace";

	/** {@value} */
	public static final String CLASS = "class";

	/** {@value} */
	public static final String OBJECT = "object";

	/** {@value} */
	public static final String ENUM = "enum";

	/** {@value} */
	public static final String ENUM_LITERAL = "enum literal";

	/** {@value} */
	public static final String STRUCT = "struct";

	/** {@value} */
	public static final String DELEGATE = "delegate";

	/** {@value} */
	public static final String ACTOR = "actor";

	/** {@value} */
	public static final String EXTERN_C_PREFIX = "extern C prefix";

	/** {@value} */
	public static final String CLASS_PUBLICATION = "class publication";

	/** {@value} */
	public static final String CLASS_DEFINITION = "class definition";

	/** {@value} */
	public static final String CLASS_IMPLEMENTATION = "class implementation";

	/** {@value} */
	public static final String INTERFACE_CLASS_NAME = "interface class";

	/** {@value} */
	public static final String INTERFACE_PUBLICATION = "interface publication";

	/** {@value} */
	public static final String INTERFACE_DEFINITION = "interface definition";

	/** {@value} */
	public static final String METHOD_IMPLEMENTATION = "method implementation";

	/** {@value} */
	public static final String DECLARATION = "declaration";

	/**
	 * Suffix for subtypes that declare something (e.g.,
	 * {@link #METHOD_DECLARATION})
	 */
	public static final String DECLARATION_SUFFIX = "declaration";

	/** {@value} */
	public static final String DECLARE = "declare";

	/** {@value} */
	public static final String METHOD_DECLARATION = "method " + DECLARATION_SUFFIX;

	/** {@value} */
	public static final String METHOD_CALL_SIGNATURE = "method call signature";

	/** {@value} */
	public static final String IMPLICIT_MEMBER = "implicit member";

	/** {@value} */
	public static final String FUNCTION = "function";

	/** {@value} */
	public static final String PROCEDURE = "procedure";

	/** {@value} */
	public static final String PROCESS = "process";

	/** {@value} */
	public static final String ON_CHANGE = "on change";

	/** {@value} */
	public static final String NATIVE_SQL = "native SQL";

	/** {@value} */
	public static final String SQL_CMD = "SQL cmd";

	/** ABAP Entity Manipulation Language statement */
	public static final Object EML_STATEMENT = "EML statement";

	/** {@value} */
	public static final String SELECT_BLOCK = "select block";

	/** {@value} */
	public static final String MODULE_INPUT = "module input";

	/** {@value} */
	public static final String MODULE_OUTPUT = "module output";

	/** {@value} */
	public static final String GET_LATE = "get late";

	/** {@value} */
	public static final String TOP_OF_PAGE_DURING_LINE_SELECTION = "top-of-page during line-selection";

	/** {@value} */
	public static final String PF_EVENT = "pf event";

	/** {@value} */
	public static final String FORM = "form";

	/** {@value} */
	public static final String SINGLE_SELECT = "single select";

	/** {@value} */
	public static final String MACRO = "macro";

	/** {@value} */
	public static final String VISIBILITY = "Visibility";

	/** {@value} */
	public static final String EMPTY_STATEMENT = "empty statement";

	/** {@value} */
	public static final String LOCAL_VARIABLE = "local variable";

	/** {@value} */
	public static final String INTERFACE = "interface";

	/** {@value} */
	public static final String IF = "if";

	/** {@value} */
	public static final String IF_LET = "if let";

	/** {@value} */
	public static final String IF_VAR = "if var";

	/** {@value} */
	public static final String ELSE_IF = "else if";

	/** {@value} */
	public static final String ELSE_IF_NOSPACE = "elseif";

	/** {@value} */
	public static final String DO = "do";

	/** {@value} */
	public static final String FOR = "for";

	/** {@value} */
	public static final String WHILE = "while";

	/** {@value} */
	public static final String WHILE_LET = "while let";

	/** {@value} */
	public static final String WORKFLOW_FUNCTION = "workflow function";

	/** {@value} */
	public static final String FOREACH = "foreach";

	/** {@value} */
	public static final String FOR_EACH = "for each";

	/** {@value} */
	public static final String ECHO = "echo";

	/** {@value} */
	public static final String REMOVE = "remove";

	/** {@value} */
	public static final String ELSIF = "elsif";

	/** {@value} */
	public static final String ELIF = "elif";

	/** {@value} */
	public static final String LOOP = "loop";

	/** {@value} */
	public static final String CONTINUE = "continue";

	/** {@value} */
	public static final String BREAK = "break";

	/** {@value} */
	public static final String BROKER_SCHEMA = "broker schema";

	/** {@value} */
	public static final String CASE = "case";

	/** {@value} */
	public static final String SWITCH = "switch";

	/** {@value} */
	public static final String SWITCH_EXPRESSION = "switch expression";

	/** {@value} */
	public static final String DEFAULT = "default";

	/** {@value} */
	public static final String OTHERWISE = "otherwise";

	/** {@value} */
	public static final String WHEN = "when";

	/** {@value} */
	public static final String ELSE = "else";

	/** {@value} */
	public static final String FINALLY = "finally";

	/** {@value} */
	public static final String TRY = "try";

	/** {@value} */
	public static final String CATCH = "catch";

	/** {@value} */
	public static final String BLOCK = "block";

	/** {@value} */
	public static final String AT_CATCH = "@catch";

	/** {@value} */
	public static final String AT_INTERFACE = "@interface";

	/** {@value} */
	public static final String AT_IMPLEMENTATION = "@implementation";

	/** {@value} */
	public static final String EXIT = "exit";

	/** {@value} */
	public static final String SIMPLE_STATEMENT = "simple statement";

	/** {@value} */
	public static final String CLEAN = "clean";

	/** {@value} */
	public static final String CLEANUP = "cleanup";

	/** {@value} */
	public static final String PROVIDE = "provide";

	/** {@value} */
	public static final String PROVIDES = "provides";

	/** {@value} */
	public static final String GOOG_PROVIDE = "goog.provide";

	/** {@value} */
	public static final String GOOG_MODULE = "goog.module";

	/** {@value} */
	public static final String GOOG_REQUIRE = "goog.require";

	/** {@value} */
	public static final String COMMONJS_REQUIRE = "commonjs require";

	/** {@value} */
	public static final String COMMONJS_EXPORTS = "commonjs exports";

	/** {@value} */
	public static final String NODEJS_EXPORTS = "nodejs module.exports";

	/** {@value} */
	public static final String LABEL = "label";

	/** {@value} */
	public static final String ANONYMOUS_BLOCK = "anonymous block";

	/** {@value} */
	public static final String ANONYMOUS_CLASS = "anonymous class";

	/** {@value} */
	public static final String CHECKED = "checked";

	/** {@value} */
	public static final String UNCHECKED = "unchecked";

	/** {@value} */
	public static final String UNSAFE = "unsafe";

	/** {@value} */
	public static final String LOCK = "lock";

	/** {@value} */
	public static final String USING = "using";

	/** {@value} */
	public static final String USING_ASSEMBLY = "using assembly";

	/** {@value} */
	public static final String USING_NAMESPACE = "using namespace";

	/** {@value} */
	public static final String USING_MODULE = "using module";

	/** {@value} */
	public static final String USE = "use";

	/** {@value} */
	public static final String FIXED = "fixed";

	/** {@value} */
	public static final String YIELD = "yield";

	/** {@value} */
	public static final String GUARD = "guard";

	/** {@value} */
	public static final String GOTO = "goto";

	/** {@value} */
	public static final String LAMBDA = "lambda";

	/** {@value} */
	public static final String LAMBDA_EXPRESSION = "lambda expression";

	/** {@value} */
	public static final String SYNCHRONIZED = "synchronized";

	/** {@value} */
	public static final String DATA = "data";

	/** {@value} */
	public static final String STATICS = "statics";

	/** {@value} */
	public static final String CONSTANTS = "constants";

	/** {@value} */
	public static final String FIELD_SYMBOLS = "field-symbols";

	/** {@value} */
	public static final String PARAMETER = "parameter";

	/** {@value} */
	public static final String PARAMETERS = "parameters";

	/** {@value} */
	public static final String TABLE = "table";

	/** {@value} */
	public static final String TABLES = "tables";

	/** {@value} */
	public static final String NODES = "nodes";

	/** {@value} */
	public static final String SUBMIT = "submit";

	/** {@value} */
	public static final String LEAVE = "leave";

	/** {@value} */
	public static final String RETURN = "return";

	/** {@value} */
	public static final String ROUTINE = "routine";

	/** {@value} */
	public static final String STOP = "stop";

	/** {@value} */
	public static final String REJECT = "reject";

	/** {@value} */
	public static final String RAISE = "raise";

	/** {@value} */
	public static final String RAISERROR = "raiserror";

	/** {@value} */
	public static final String PRINT = "print";

	/** {@value} */
	public static final String TYPES = "types";

	/** {@value} */
	public static final String INCLUDE = "include";

	/** {@value} */
	public static final String TYPE_POOLS = "type-pools";

	/** {@value} */
	public static final String STATIC_INITIALIZER = "static initializer";

	/** {@value} */
	public static final String NON_STATIC_INITIALIZER = "non-static initializer";

	/** {@value} */
	public static final String ATTRIBUTE = "attribute";

	/** {@value} */
	public static final String SELECT = "select";

	/** {@value} */
	public static final String CREATE = "create";

	/** {@value} */
	public static final String ALTER = "alter";

	/** {@value} */
	public static final String DROP = "drop";

	/** {@value} */
	public static final String DYNAMIC_PARAMETERS = "dynamicparam";

	/** {@value} */
	public static final String PROPERTY = "property";

	/** {@value} */
	public static final String PROTOCOL = "protocol";

	/** {@value} */
	public static final String SET = "set";

	/** {@value} */
	public static final String META_CLASS = "meta class";

	/** {@value} */
	public static final String RANGE = "range";

	/** {@value} */
	public static final String ANONYMOUS_METHOD = "anonymous method";

	/** {@value} */
	public static final String ATTRIBUTE_FUNCTION = "attribute function";

	/** {@value} */
	public static final String NAMED_FUNCTION = "named function";

	/** {@value} */
	public static final String ANONYMOUS_FUNCTION = "anonymous function";

	/** {@value} */
	public static final String ASSIGNED_FUNCTION = "assigned function";

	/** {@value} */
	public static final String ASSIGNED_LAMBDA = "assigned lambda";

	/** {@value} */
	public static final String GLOBAL_BEGIN = "global begin";

	/** {@value} */
	public static final String FUNCTION_POINTER = "function pointer";

	/** {@value} */
	public static final String CONSTANT = "constant";

	/** {@value} */
	public static final String CONSTRUCTOR = "constructor";

	/** {@value} */
	public static final String DESTRUCTOR = "destructor";

	/** {@value} */
	public static final String OPENS = "opens";

	/** {@value} */
	public static final String OPERATOR = "operator";

	/** {@value} */
	public static final String OPERATOR_DECLARATION = "operator " + DECLARATION_SUFFIX;

	/** {@value} */
	public static final String CAST_OPERATOR = "cast operator";

	/** {@value} */
	public static final String CASE_LABEL = "case label";

	/** {@value} */
	public static final String ALIAS = "alias";

	/** {@value} */
	public static final String TYPE_POINTER = "type pointer";

	/** {@value} */
	public static final String VARIABLE = "variable";

	/** {@value} */
	public static final String CASE_ELSE = "case else";

	/** {@value} */
	public static final String IEC_SV_VARIABLE = "IEC SV Variable";

	/** {@value} */
	public static final String IEC_SV_ELEMENT = "IEC SV Element";

	/** {@value} */
	public static final String FUNCTION_DECLARATION = "function " + DECLARATION_SUFFIX;

	/** {@value} */
	public static final String FUNCTION_POINTER_DECLARATION = "function pointer " + DECLARATION_SUFFIX;

	/** {@value} */
	public static final String IMPORT = "import";

	/** {@value} */
	public static final String ALIAS_IMPORT = "alias import";

	/** {@value} */
	public static final String AT_IMPORT = "@import";

	/** {@value} */
	public static final String STATIC_CONSTRUCTOR = "static constructor";

	/** {@value} */
	public static final String PACKAGE = "package";

	/** {@value} */
	public static final String METHOD = "method";

	/** {@value} */
	public static final String STATIC_METHOD = "static method";

	/** {@value} */
	public static final String DESTRUCTOR_DECLARATION = "destructor " + DECLARATION_SUFFIX;

	/** {@value} */
	public static final String CONSTRUCTOR_DECLARATION = "constructor " + DECLARATION_SUFFIX;

	/** {@value} */
	public static final String EXECUTE = "execute";

	/** {@value} */
	public static final String EXEC = "exec";

	/** {@value} */
	public static final String EXTENSION = "extension";

	/** {@value} */
	public static final String GO = "go";

	/** {@value} */
	public static final String INSERT = "insert";

	/** {@value} */
	public static final String BULK_INSERT = "bulk insert";

	/** {@value} */
	public static final String GRANT = "grant";

	/** {@value} */
	public static final String MERGE = "merge";

	/** {@value} */
	public static final String MODIFY = "modify";

	/** {@value} */
	public static final String UPDATE = "update";

	/** {@value} */
	public static final String DELETE = "delete";

	/** {@value} */
	public static final String FETCH = "fetch";

	/** {@value} */
	public static final String VIEW = "view";

	/** {@value} */
	public static final String TRIGGER = "trigger";

	/** {@value} */
	public static final String AT_LINE_SELECTION = "at line-selection";

	/** {@value} */
	public static final String AT_SELECTION_SCREEN = "at selection-screen";

	/** {@value} */
	public static final String AT_SELECTION_OUTPUT = "at selection-output";

	/** {@value} */
	public static final String AT_USER_COMMAND = "at user-command";

	/** {@value} */
	public static final String END_OF_PAGE = "end-of-page";

	/** {@value} */
	public static final String END_OF_SELECTION = "end-of-selection";

	/** {@value} */
	public static final String INITIALIZATION = "initialization";

	/** {@value} */
	public static final String LOAD_OF_PROGRAM = "load-of-program";

	/** {@value} */
	public static final String START_OF_SELECTION = "start-of-selection";

	/** {@value} */
	public static final String TOP_OF_PAGE = "top-of-page";

	/** {@value} */
	public static final String GET = "get";

	/** {@value} */
	public static final String REPEAT_WHILE = "repeat while";

	/** {@value} */
	public static final String REPEAT_UNTIL = "repeat until";

	/** {@value} */
	public static final String DO_LOOP = "do loop";

	/** {@value} */
	public static final String DO_UNTIL = "do until";

	/** {@value} */
	public static final String DO_WHILE = "do while";

	/** {@value} */
	public static final String REPORT = "report";

	/** {@value} */
	public static final String PROGRAM = "program";

	/** {@value} */
	public static final String GENERATE = "generate";

	/** {@value} */
	public static final String CALL = "call";

	/** {@value} */
	public static final String ASSERT = "assert";

	/** {@value} */
	public static final String SELECT_OPTIONS = "select-options";

	/** {@value} */
	public static final String DEFINE = "define";

	/** {@value} */
	public static final String FIELD_GROUPS = "field-groups";

	/** {@value} */
	public static final String RANGES = "ranges";

	/** {@value} */
	public static final String CLASS_DATA = "class-data";

	/** {@value} */
	public static final String DOCUMENT_ROOT = "document-root";

	/** {@value} */
	public static final String IMPORTING = "importing";

	/** {@value} */
	public static final String EXPORTING = "exporting";

	/** {@value} */
	public static final String CHANGING = "changing";

	/** {@value} */
	public static final String RETURNING = "returning";

	/** {@value} */
	public static final String PUBLIC = "public";

	/** {@value} */
	public static final String PROTECTED = "protected";

	/** {@value} */
	public static final String PRIVATE = "private";

	/** {@value} */
	public static final String ENUM_CLASS = "enum class";

	/** {@value} */
	public static final String ENUM_STRUCT = "enum struct";

	/** {@value} */
	public static final String GLOBAL_VARIABLE = "global variable";

	/** {@value} */
	public static final String GLOBAL_FUNCTION = "global function";

	/** {@value} */
	public static final String LOCAL_FUNCTION = "local function";

	/** {@value} */
	public static final String STATIC_IMPORT = "static import";

	/** {@value} */
	public static final String EMPTY_GET = "empty get";

	/** {@value} */
	public static final String EMPTY_SET = "empty set";

	/** {@value} */
	public static final String EMPTY_INIT = "empty init";

	/** {@value} */
	public static final String TRAIT = "trait";

	/** {@value} */
	public static final String TRAP = "trap";

	/** {@value} */
	public static final String REQUIRE = "require";

	/** {@value} */
	public static final String REQUIRES = "requires";

	/** {@value} */
	public static final String REQUIRE_ONCE = "require_once";

	/** {@value} */
	public static final String INCLUDE_ONCE = "include_once";

	/** {@value} */
	public static final String BLOCK_LITERAL = "block literal";

	/** {@value} */
	public static final String UNIT_STRUCT = "unit struct";

	/** {@value} */
	public static final String UNION = "union";

	/** {@value} */
	public static final String MATCH_CLAUSE = "match clause";

	/** {@value} */
	public static final String ATTRIBUTE_DIRECTIVE = "attribute directive";

	/** {@value} */
	public static final String MODULE = "module";

	/** {@value} */
	public static final String IMPLEMENTATION = "implementation";

	/** {@value} */
	public static final String UNSAFE_BLOCK = "unsafe block";

	/** {@value} */
	public static final String TYPE_ALIAS = "type alias";

	/** {@value} */
	public static final String ASSOCIATED_TYPE = "associated type";

	/** {@value} */
	public static final String TUPLE_STRUCT = "tuple struct";

	/** {@value} */
	public static final String IFS_SEARCH_BLOCK = "ifs search block";

	/** {@value} */
	public static final String PREPROCESSOR_CONDITION = "preprocessor condition";

	/** {@value} */
	public static final String STRUCTURE = "structure";

	/** {@value} */
	public static final String DECORATOR = "decorator";

	/** {@value} */
	public static final String CURSOR = "cursor";

	/** {@value} */
	public static final String SQL = "SQL";

	/** {@value} */
	public static final String WITH = "with";

	/** {@value} */
	public static final String EXCEPT = "except";

	/** {@value} */
	public static final String PREPROCESSOR_DIRECTIVE = "preprocessor directive";

	/** {@value} */
	public static final String USES = "uses";

	/** {@value} */
	public static final String EXPORTED_VARIABLE = "exported variable";

	/** {@value} */
	public static final String TYPEDEF = "typedef";

	/** {@value} */
	public static final String ABSTRACT = "abstract";

	/** {@value} */
	public static final String FROM = "from";

	/** {@value} */
	public static final String CTE = "common table expression";

	/** {@value} */
	public static final String EVALUATE = "evaluate";

	/** {@value} */
	public static final String OTHER = "other";

	/** {@value} */
	public static final String STAR_IMPORT = "star import";

	/** {@value} */
	public static final String DEFAULT_EXPORT_IMPORT = "default export import";

	/** {@value} */
	public static final String TRIPLE_SLASH_DIRECTIVE_IMPORT = "triple slash directive import";

	/** {@value} */
	public static final String EXPORT = "export";

	/** {@value} */
	public static final String EXPORTS = "exports";

	/** {@value} */
	public static final String RE_EXPORT = "re-export";

	/** {@value} */
	public static final String DEFAULT_EXPORT = "default export";

	/** {@value} */
	public static final String OBJECT_EXPORT = "object export";

	/** {@value} */
	public static final String OBJECT_EXPORT_IMPORT = "object export import";

	/** {@value} */
	public static final String INLINE_ASSEMBLER_BLOCK = "inline assembler block";

	/** {@value} */
	public static final String INLINE_ASSEMBLER_CALL = "inline assembler call";

	/** {@value} */
	public static final String INLINE_ASSEMBLER_FUNCTION = "inline assembler function";

	/** {@value} */
	public static final String BEGIN = "begin";

	/** {@value} */
	public static final String END = "end";

	/** {@value} */
	public static final String WRITE = "write";

	/** {@value} */
	public static final String CLEAR = "clear";

	/** {@value} */
	public static final String FREE = "free";

	/** {@value} */
	public static final String REFRESH = "refresh";

	/** {@value} */
	public static final String EXPANDED_MACRO = "expanded macro";

	/** {@value} */
	public static final String COND = "cond";

	/** {@value} */
	public static final String CHECK = "check";

	/**
	 * {@value}
	 *
	 * Subtype name for code that we assume is intended to be a macro name. But the
	 * macro name was not supplied to the preprocessor, so the Macro name remained
	 * in the code parsed by our Shallow Parser. Usually this hints at a
	 * configuration problem (Macro was not supplied). The detection of these cases
	 * is based on a crude heuristic (Identifier written in all-uppercase). If
	 * possible, this subtype name should not be used in analysis.
	 */
	public static final String PREPROCESSOR_CODE = "preprocessor code";

	/** {@value} */
	public static final String METHODS = "methods";

	/** Sub type names of loops. */
	public static final Set<String> LOOP_SUB_TYPE_NAMES = Collections
			.unmodifiableSet(CollectionUtils.asHashSet(FOR, WHILE, DO, FOREACH));

	/**
	 * Sub type names of if-conditions, including <code>ELSE</code>.
	 *
	 * @see #IF_CONDITION_CONDITIONAL_SUB_TYPE_NAMES
	 */
	public static final Set<String> IF_CONDITION_SUB_TYPE_NAMES = Collections
			.unmodifiableSet(CollectionUtils.asHashSet(IF, ELSE_IF, ELSE_IF_NOSPACE, ELSIF, ELSE));

	/**
	 * Sub type names of if-conditions, excluding <code>ELSE</code>.
	 *
	 * @see #IF_CONDITION_SUB_TYPE_NAMES
	 */
	public static final Set<String> IF_CONDITION_CONDITIONAL_SUB_TYPE_NAMES = Collections
			.unmodifiableSet(CollectionUtils.asHashSet(IF, ELSE_IF, ELSE_IF_NOSPACE, ELSIF));

	/**
	 * Attributes introduced in C++17 and C#. We call these attribute annotations,
	 * to differentiate from class attributes (aka fields).
	 *
	 * @see "http://en.cppreference.com/w/cpp/language/attributes"
	 * @see "https://docs.microsoft.com/de-de/dotnet/csharp/programming-guide/concepts/attributes/"
	 */
	public static final String ATTRIBUTE_ANNOTATION = "attribute annotation";

	/** Subtype for java annotation interfaces. */
	public static final String ANNOTATION_INTERFACE = "@interface";

	/** {@value} */
	public static final String TYPE = "TYPE";

	/** {@value} */
	public static final String JAVA_CODE = "java code";

	/** {@value} */
	public static final String INFOTYPES = "INFOTYPES";

	/** {@value} */
	public static final String RECORD = "record";

	/** SQL actions */

	/** {@value} */
	public static final String TRAN = "tran";

	/** {@value} */
	public static final String COMMIT = "commit";

	/** {@value} */
	public static final String ROLLBACK = "rollback";

	/** {@value} */
	public static final String INIT = "init";

	/** {@value} */
	public static final String PASSTHRU = "passthru";

	/**
	 * A static assertion in C/C++ (evaluated at compile time). We parse it as
	 * {@link eu.cqse.check.framework.shallowparser.framework.EShallowEntityType#META}
	 * entity since it is not really part of the executable code but might be
	 * interesting for analysis.
	 *
	 * Can occur in or outside of methods.
	 */
	public static final String STATIC_ASSERT = "static_assert";

	/** {@value} */
	public static final String ARRAY = "array";

	/**
	 * Pragma preprocessor directives in C/C++/Objectiv-C/Objective-C++. The parser
	 * will parse them as
	 * {@link eu.cqse.check.framework.shallowparser.framework.EShallowEntityType#META}
	 * shallow entity with this subtype.
	 * <p>
	 * Note that Shallow entities for pragma directives don't necessarily extend to
	 * the end of the line (there may be other entities following in the same line)
	 * since we have pragmas in macro expansions. For an example, see
	 * <code>CppShallowParserTest#testPragmaInFullCodeExample()</code>.
	 */
	public static final String PRAGMA_DIRECTIVE = "pragma directive";

	/**
	 * C++ declaration of a "friend" class. ("section 3" and "section 4" on
	 * https://en.cppreference.com/w/cpp/language/friend). This is always a
	 * {@link eu.cqse.check.framework.shallowparser.framework.EShallowEntityType#META}
	 * entity.
	 */
	public static final String FRIEND_CLASS_DECLARATION = "Friend Class Declaration";
}
