/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package eu.cqse.check.framework.shallowparser.framework;

import static eu.cqse.check.framework.shallowparser.framework.EShallowEntityType.META;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import org.apache.commons.lang3.ArrayUtils;
import org.checkerframework.checker.nullness.qual.NonNull;

import eu.cqse.check.framework.scanner.IToken;
import eu.cqse.check.framework.shallowparser.SubTypeNames;

/** Utility methods for working with shallow entities. */
public class ShallowEntityTraversalUtils {

	/** Lists all entities. */
	public static List<ShallowEntity> listAllEntities(Collection<ShallowEntity> entities) {
		return listEntitiesOfTypes(entities, EnumSet.allOf(EShallowEntityType.class));
	}

	/**
	 * Lists all entities of the given type, includes matching child entities
	 * recursively.
	 */
	public static List<ShallowEntity> listEntitiesOfType(Collection<ShallowEntity> entities, EShallowEntityType type) {
		return listEntitiesOfTypes(entities, EnumSet.of(type));
	}

	/**
	 * Lists all entities of the given types, includes matching child entities
	 * recursively.
	 */
	public static List<ShallowEntity> listEntitiesOfTypes(Collection<ShallowEntity> entities,
			Set<EShallowEntityType> types) {
		return new CollectingVisitorBase() {
			@Override
			protected boolean collect(ShallowEntity entity) {
				return types.contains(entity.getType());
			}
		}.apply(entities);
	}

	/**
	 * Lists all entities of the given types. Recursively looks into child entities
	 * (regardless of whether they match the types/subtypes).
	 */
	public static List<ShallowEntity> listEntitiesOfTypesWithSubtypes(Collection<ShallowEntity> entities,
			Set<EShallowEntityType> types, Set<String> shallowEntitySubtypes) {
		return new CollectingVisitorBase() {
			@Override
			protected boolean collect(ShallowEntity entity) {
				return types.contains(entity.getType()) && shallowEntitySubtypes.contains(entity.getSubtype());
			}
		}.apply(entities);
	}

	/**
	 * Lists all entities that are selected by the given predicate. The search runs
	 * recursively (checks child entities if parents don't match).
	 */
	public static List<ShallowEntity> selectEntities(Collection<ShallowEntity> entities,
			Predicate<ShallowEntity> predicate) {
		return new CollectingVisitorBase() {
			@Override
			protected boolean collect(ShallowEntity entity) {
				return predicate.test(entity);
			}
		}.apply(entities);
	}

	/**
	 * Returns the first incomplete entity found (or null). Unclosed entities
	 * correspond to parsing errors.
	 */
	private static ShallowEntity findIncompleteEntity(ShallowEntity entity) {
		if (!entity.isCompleted()) {
			return entity;
		}
		return findIncompleteEntity(entity.getChildren());
	}

	/**
	 * Returns the first incomplete entity found (or null). Unclosed entities
	 * correspond to parsing errors.
	 */
	public static ShallowEntity findIncompleteEntity(List<ShallowEntity> entities) {
		for (ShallowEntity entity : entities) {
			ShallowEntity incomplete = findIncompleteEntity(entity);
			if (incomplete != null) {
				return incomplete;
			}
		}
		return null;
	}

	/**
	 * Traverses the given collection of entities and returns a flat list of
	 * entities.
	 */
	public static List<ShallowEntity> getAllEntities(Collection<ShallowEntity> entities) {
		return new CollectingVisitorBase() {
			@Override
			protected boolean collect(ShallowEntity entity) {
				return true;
			}
		}.apply(entities);
	}

	/**
	 * Lists all entities of the given type, but does not traverse into the
	 * entities, i.e. entities contained in the returned ones are not returned.
	 */
	public static List<ShallowEntity> listEntitiesOfTypeNonRecursive(Collection<ShallowEntity> entities,
			EShallowEntityType type) {
		List<ShallowEntity> methods = new ArrayList<>();
		ShallowEntity.traverse(entities, new ShallowEntityVisitorBase() {
			/** {@inheritDoc} */
			@Override
			public boolean visit(ShallowEntity entity) {
				if (entity.getType() == type) {
					methods.add(entity);
					// do not recurse into methods
					return false;
				}
				return true;
			}
		});
		return methods;
	}

	/**
	 * Returns all methods found in the given entities, but does not traverse into
	 * the entities, i.e. methods contained in other methods are not returned.
	 */
	public static List<ShallowEntity> listMethodsNonRecursive(List<ShallowEntity> entities) {
		return listEntitiesOfTypeNonRecursive(entities, EShallowEntityType.METHOD);
	}

	/**
	 * Recursively traverses the given shallow entities and returns all entities
	 * that match the given predicate.
	 */
	public static List<ShallowEntity> listMatchingEntitiesRecursive(List<ShallowEntity> entities,
			Predicate<ShallowEntity> matchPredicate) {
		return new CollectingVisitorBase() {
			@Override
			protected boolean collect(ShallowEntity entity) {
				return matchPredicate.test(entity);
			}
		}.apply(entities);
	}

	/**
	 * Finds child {@link ShallowEntity} objects in the given rootEntities that are
	 * also children (not necessarily direct children) of the selected ancestors.
	 * <p>
	 * The returned list contains each ShallowEntity only once (for example, if you
	 * search for "while->break" in a <code>while(){while(){break;}}</code>, we
	 * return the break only once, even though it is a child of both while nodes).
	 * <p>
	 * The returned list is ordered by start offset of the entities.
	 * <p>
	 * Note: this is a rather expensive method since it needs to search the entire
	 * tree (and parts of it several times). If your scenario can be implemented in
	 * a cheaper, more focused way, then better avoid this method.
	 *
	 * @param rootEntities
	 *            the root entities of the {@link ShallowEntity} tree that should be
	 *            searched
	 * @param ancestorEntityType
	 *            types of ancestor that should be considered (i.e., that should be
	 *            searched)
	 * @param ancestorEntitySelector
	 *            {@link Predicate} that determines whether an ancestor should be
	 *            considered (the #ancestorEntityTypes are already considered, no
	 *            need to check them again here)
	 * @param targetEntityType
	 *            filter for the types of the children that should be returned
	 */
	public static List<ShallowEntity> findNestedEntities(List<ShallowEntity> rootEntities,
			EShallowEntityType ancestorEntityType, Predicate<ShallowEntity> ancestorEntitySelector,
			EShallowEntityType targetEntityType) {
		return findNestedEntities(rootEntities, EnumSet.of(ancestorEntityType), ancestorEntitySelector,
				EnumSet.of(targetEntityType));
	}

	/**
	 * Finds child {@link ShallowEntity} objects in the given rootEntities that are
	 * also children (not necessarily direct children) of the selected ancestors.
	 * <p>
	 * The returned list contains each ShallowEntity only once (for example, if you
	 * search for "while->break" in a <code>while(){while(){break;}}</code>, we
	 * return the break only once, even though it is a child of both while nodes).
	 * <p>
	 * The returned list is ordered by start offset of the entities.
	 * <p>
	 * Note: this is a rather expensive method since it needs to search the entire
	 * tree (and parts of it several times). If your scenario can be implemented in
	 * a cheaper, more focused way, then better avoid this method.
	 *
	 * @param rootEntities
	 *            the root entities of the {@link ShallowEntity} tree that should be
	 *            searched
	 * @param ancestorEntityTypes
	 *            types of ancestors that should be considered (i.e., that should be
	 *            searched)
	 * @param ancestorEntitySelector
	 *            {@link Predicate} that determines whether an ancestor should be
	 *            considered (the #ancestorEntityTypes are already considered, no
	 *            need to check them again here)
	 * @param targetEntityTypes
	 *            filter for the types of the children that should be returned
	 */
	public static List<ShallowEntity> findNestedEntities(List<ShallowEntity> rootEntities,
			Set<EShallowEntityType> ancestorEntityTypes, Predicate<ShallowEntity> ancestorEntitySelector,
			Set<EShallowEntityType> targetEntityTypes) {
		Set<ShallowEntity> selectedTargetEntities = new HashSet<>();
		List<ShallowEntity> selectedAncestorEntities = listEntitiesOfTypes(rootEntities, ancestorEntityTypes);
		for (ShallowEntity ancestorEntity : selectedAncestorEntities) {
			if (!ancestorEntitySelector.test(ancestorEntity)) {
				continue;
			}
			selectedTargetEntities.addAll(listEntitiesOfTypes(ancestorEntity.getChildren(), targetEntityTypes));
		}
		return selectedTargetEntities.stream().sorted(Comparator.comparing(ShallowEntity::getStartOffset))
				.collect(Collectors.toList());
	}

	/** Empty default implementation of {@link IShallowEntityVisitor}. */
	public static abstract class ShallowEntityVisitorBase implements IShallowEntityVisitor {
		/** {@inheritDoc} */
		@Override
		public boolean visit(ShallowEntity entity) {
			return true;
		}

		/** {@inheritDoc} */
		@Override
		public void endVisit(ShallowEntity entity) {
			// nothing
		}
	}

	/**
	 * Base class for visitors that collect shallow entities, includes matching
	 * child entities recursively.
	 */
	public static abstract class CollectingVisitorBase extends ShallowEntityVisitorBase {

		/** The collected entities. */
		private final List<ShallowEntity> entities = new ArrayList<>();

		/** {@inheritDoc} */
		@Override
		public boolean visit(ShallowEntity entity) {
			if (collect(entity)) {
				entities.add(entity);
			}
			return true;
		}

		/** Template method that returns true if the entity should be collected. */
		protected abstract boolean collect(ShallowEntity entity);

		/**
		 * Applies this collecting visitor depth-first to the given entities and returns
		 * the collected result.
		 */
		public List<ShallowEntity> apply(Collection<ShallowEntity> entities) {
			ShallowEntity.traverse(entities, this);
			return this.entities;
		}
	}

	/** Utility method for traversing multiple entities. */
	public static void traverseWithKey(List<ShallowEntity> entities, IShallowEntityKeyVisitor visitor) {
		traverseWithKey(entities, visitor, new int[0]);
	}

	/**
	 * Utility method for traversing multiple entities.
	 *
	 * @param parentKey
	 *            Specifies the unique key of the entity lists parent.
	 */
	private static void traverseWithKey(List<ShallowEntity> entities, IShallowEntityKeyVisitor visitor,
			int[] parentKey) {
		int length = parentKey.length;
		for (int i = 0; i < entities.size(); i++) {
			ShallowEntity entity = entities.get(i);
			int[] key = Arrays.copyOf(parentKey, length + 1);
			key[length] = i;
			if (visitor.visit(key, entity)) {
				traverseWithKey(entity.getChildren(), visitor, key);
			}
		}
	}

	/** Empty default implementation of {@link IShallowEntityVisitor}. */
	public interface IShallowEntityKeyVisitor {

		/**
		 * Denotes that visiting the entity begins.
		 *
		 * @param key
		 *            Key, which uniquely identifies the entity for the traversed list
		 *            of {@link ShallowEntity}s.
		 * @return true if the children of this entity are to be visited as well.
		 */
		boolean visit(int[] key, ShallowEntity entity);
	}

	/**
	 * Returns a {@link ShallowEntity} from the given list of entities, which is
	 * identified by the given key.
	 *
	 * @see ShallowEntityTraversalUtils#traverseWithKey(List,
	 *      IShallowEntityKeyVisitor)
	 */
	public static ShallowEntity getEntityFromListWithKey(List<ShallowEntity> entities, int[] key) {
		if (key.length == 0 || key[0] >= entities.size()) {
			return null;
		}
		ShallowEntity entity = entities.get(key[0]);
		if (key.length == 1) {
			return entity;
		}
		return getEntityFromListWithKey(entity.getChildren(), Arrays.copyOfRange(key, 1, key.length));
	}

	/**
	 * Returns the next {@link ShallowEntity} after the given entity. Returns
	 * <code>null</code> if there is no subsequent entity.
	 */
	public static ShallowEntity getSubsequentEntity(ShallowEntity entity) {
		boolean found = false;
		while (entity != null && entity.getParent() != null) {
			List<ShallowEntity> neighboringEntities = entity.getParent().getChildren();
			for (ShallowEntity neighboringEntity : neighboringEntities) {
				if (found) {
					return neighboringEntity;
				} else if (neighboringEntity.equals(entity)) {
					found = true;
				}
			}
			entity = entity.getParent();
			found = false;
		}
		return null;
	}

	/**
	 * Returns the previous {@link ShallowEntity} before the given entity. This is
	 * either the previous sibling entity or the parent entity if the given entity
	 * is its first child. Returns <code>null</code> if there is no previous entity.
	 */
	public static ShallowEntity getPreviousEntity(@NonNull ShallowEntity entity) {
		ShallowEntity previousEntity = entity.getParent();
		while (entity != null && entity.getParent() != null) {
			List<ShallowEntity> neighboringEntities = entity.getParent().getChildren();
			for (ShallowEntity neighboringEntity : neighboringEntities) {
				if (neighboringEntity.equals(entity)) {
					return previousEntity;
				}
				previousEntity = neighboringEntity;
			}
			entity = entity.getParent();
		}
		return null;
	}

	/**
	 * Returns the outermost entity containing the given (one-based) line number.
	 */
	public static Optional<ShallowEntity> findEntityForLineNonRec(int line, List<ShallowEntity> entities) {
		return findEntityForLine(line, false, entities, false);
	}

	/**
	 * Returns the innermost entity containing the given (one-based) line number.
	 */
	public static Optional<ShallowEntity> findEntityForLine(int line, List<ShallowEntity> entities) {
		return findEntityForLine(line, true, entities, false);
	}

	/**
	 * Returns the innermost entity containing the given (one-based) line number.
	 *
	 * @param returnSuccessor
	 *            Returns the next entity after the given line if the exact line has
	 *            no matching entity
	 */
	public static Optional<ShallowEntity> findEntityForLine(int line, List<ShallowEntity> entities,
			boolean returnSuccessor) {
		return findEntityForLine(line, true, entities, returnSuccessor);
	}

	/**
	 * Returns the entity containing the given (one-based) line number.
	 *
	 * @param returnSuccessor
	 *            Returns the next entity after the given line if the exact line has
	 *            no matching entity
	 */
	private static Optional<ShallowEntity> findEntityForLine(int line, boolean recursive, List<ShallowEntity> entities,
			boolean returnSuccessor) {
		// we know that entities are sorted according to line number, so we can
		// binary search them
		@SuppressWarnings("serial")
		ShallowEntity comparisonEntity = new ShallowEntity(null, null, null, null, 0) {
			@Override
			public int getStartLine() {
				return line;
			}
		};
		int index = Collections.binarySearch(entities, comparisonEntity,
				Comparator.comparingInt(ShallowEntity::getStartLine));
		if (index < 0) {
			// we want insertion point +1 in this case
			index = -index - 2;
		}
		if (index >= 0 && index < entities.size() && entities.get(index).getStartLine() <= line
				&& line <= entities.get(index).getEndLine()) {
			ShallowEntity match = entities.get(index);
			if (recursive && match.hasChildren()) {
				Optional<ShallowEntity> betterMatch = findEntityForLine(line, match.getChildren(), returnSuccessor);
				if (betterMatch.isPresent()) {
					return betterMatch;
				}
			}
			return Optional.of(match);
		}
		if (returnSuccessor && index + 1 < entities.size()) {
			ShallowEntity match = entities.get(index + 1);
			return Optional.of(match);
		}

		return Optional.empty();
	}

	/**
	 * Traverses the parent hierarchy of the given startEntity upwards. Returns the
	 * first parent that matches the given predicate or Optional.empty if no entity
	 * matches the predicate.
	 */
	public static Optional<ShallowEntity> findParentEntity(ShallowEntity startEntity,
			Predicate<ShallowEntity> shallowEntityPredicate) {
		ShallowEntity current = startEntity;
		while (current.getParent() != null) {
			current = current.getParent();
			if (shallowEntityPredicate.test(current)) {
				return Optional.of(current);
			}
		}
		return Optional.empty();
	}

	/**
	 * Traverses the parent hierarchy of the given startEntity upwards. Returns the
	 * entity itself or the first parent that matches the given predicate or
	 * Optional.empty if no entity matches the predicate.
	 */
	public static Optional<ShallowEntity> findEntityOrParentEntity(ShallowEntity startEntity,
			Predicate<ShallowEntity> shallowEntityPredicate) {
		ShallowEntity current = startEntity;
		while (current != null && !shallowEntityPredicate.test(current)) {
			current = current.getParent();
		}
		return Optional.ofNullable(current);
	}

	/**
	 * Traverses the parent hierarchy of the given startEntity upwards. Returns all
	 * the shallow entities that match the given predicate, or an empty list if no
	 * entity matches the predicate.
	 */
	public static List<ShallowEntity> findMatchingParentEntities(ShallowEntity startEntity,
			Predicate<ShallowEntity> shallowEntityPredicate) {
		List<ShallowEntity> parentEntities = new ArrayList<>();
		ShallowEntity current = startEntity;
		while (current.getParent() != null) {
			current = current.getParent();
			if (shallowEntityPredicate.test(current)) {
				parentEntities.add(current);
			}
		}
		return parentEntities;
	}

	/**
	 * Traverses the parent hierarchy of the given startEntity upwards. Returns the
	 * first parent that matches the given subtype (see {@link SubTypeNames}) or
	 * Optional.empty if no entity matches the subtype.
	 */
	public static Optional<ShallowEntity> findParentEntityWithSubType(ShallowEntity startEntity, String subtype) {
		return findParentEntity(startEntity, entity -> entity.getSubtype().equals(subtype));
	}

	/**
	 * Returns the previous sibling {@link ShallowEntity} that matches the given
	 * subtype (see {@link SubTypeNames}). Optional.empty() is returned if no entity
	 * matches the subtype.
	 */
	public static Optional<ShallowEntity> getPreviousEntityWithSubtype(ShallowEntity entity, String... subtypes) {
		ShallowEntity parent = entity.getParent();
		Optional<ShallowEntity> previousEntity = Optional.empty();
		if (parent != null) {
			List<ShallowEntity> neighboringEntities = parent.getChildren();
			for (ShallowEntity neighboringEntity : neighboringEntities) {
				if (neighboringEntity == entity) {
					return previousEntity;
				}
				if (ArrayUtils.contains(subtypes, neighboringEntity.getSubtype())) {
					previousEntity = Optional.of(neighboringEntity);
				}
			}
		}
		return Optional.empty();
	}

	/**
	 * Searches recursively for and returns the shallow entity in which
	 * {@code nextToken} is contained in. Meta Entities get ignored.
	 */
	public static Optional<ShallowEntity> findContainingEntityDepthFirst(IToken nextToken,
			List<ShallowEntity> shallowEntities) {
		for (ShallowEntity entity : shallowEntities) {
			int index = entity.includedTokens().indexOf(nextToken);
			switch (index) {
			case -1:
				continue;
			case 0:
				if (entity.getType() == META) {
					// This is a package entity
					continue;
				}
				return Optional.of(entity);
			default:
				return findContainingEntityDepthFirst(nextToken, entity.getChildren());
			}
		}
		return Optional.empty();
	}

}
