/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.conqat.engine.commons.findings.location;

import java.util.Set;

import org.conqat.lib.commons.region.LineBasedRegion;

/**
 * Simple line adjuster is just a specialized form of {@link LocationAdjuster}
 * that performs way faster when only line adjustments are needed and the
 * content to be adjusted between is exactly the same. It basically only does
 * discard lines that are outside of the file content.
 */
public class SimpleValidLinesFilter implements ILineAdjuster {

	private final int totalLineCount;

	public SimpleValidLinesFilter(int totalLineCount) {
		this.totalLineCount = totalLineCount;
	}

	/** Returns whether the given line is valid for the underlying content. */
	public boolean isValidLine(int line) {
		return line > 0 && line <= totalLineCount;
	}

	@Override
	public LineBasedRegion adjustLine(int line, Set<Integer> invalidLines) {
		if (!isValidLine(line)) {
			invalidLines.add(line);
			return null;
		}
		return new LineBasedRegion(line, line);
	}

	/** Returns the line count of the original text */
	@Override
	public int getOriginalLineCount() {
		return totalLineCount;
	}

	@Override
	public ElementLocation adjustLocation(ElementLocation location) {
		if (location instanceof TextRegionLocation) {
			TextRegionLocation textRegionLocation = (TextRegionLocation) location;
			if (textRegionLocation.getRawEndLine() > totalLineCount) {
				return null;
			}
		}
		return location;
	}
}
