/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.conqat.engine.core.configuration;

import org.conqat.lib.commons.js_export.ExportToTypeScript;

import io.swagger.v3.oas.annotations.Hidden;

/**
 * Enum containing all available feature toggles in Teamscale. Feature toggles
 * are enabled via System properties.
 */
@ExportToTypeScript
public enum EFeatureToggle {

	/**
	 * Internal value to indicate no feature toggle in annotations (not an actual
	 * feature toggle intended to be used).
	 */
	@Hidden
	NONE(""),

	/**
	 * Enables the code usage widget that shows execution counts per method. This is
	 * currently only usable for ABAP projects with direct coverage report upload.
	 * The toggle will be removed once code usage analysis supports more languages.
	 */
	CODE_USAGE_WIDGET("com.teamscale.feature-toggle.code-usage-widget"),

	/**
	 * Feature configuration for coverage computations (filter coverable volumes).
	 */
	BMW_COVERABLE_VOLUME_FILTER("bmw.teamscale.feature-toggle.coverable-volume-filter"),

	/** Enables the landing page that is used for our own demo instance. */
	ENABLE_DEMO_LANDING_PAGE("com.teamscale.landing.enable"),

	/**
	 * Disable the check for sufficient memory. Doesn't make sense in some scenarios
	 * (e.g. cloud)
	 */
	DISABLE_MEMORY_CHECK("com.teamscale.disable-memory-check"),

	/**
	 * Disables the commit chart widget that shows the commits per user.
	 */
	DISABLE_COMMIT_CHART_WIDGET("com.teamscale.feature-toggle.disable-commit-chart-widget"),

	/**
	 * Overwrite coverable lines calculated by Teamscale with information from
	 * coverage reports. As this only makes sense for preprocessed languages this
	 * feature toggle only applies in case a file contains code of a preprocessed
	 * language like C, C++ or Objective-C.
	 */
	USE_COVERABLE_LINES_FROM_COVERAGE_REPORTS("com.teamscale.feature-toggle.use-coverable-lines-from-coverage-reports"),

	/**
	 * If this is true, class attributes are skipped for calculation of coverable
	 * lines.
	 */
	SKIP_CLASS_ATTRIBUTES_FOR_COVERAGE("com.teamscale.feature-toggle.skip-class-attributes-for-coverage"),

	/**
	 * Enables support for reading legacy ConQAT in-file review ratings.
	 */
	ENABLE_CONQAT_INFILE_RATING("com.teamscale.feature-toggle.conqat-infile-rating"),

	/** Enables support for monitoring uploads to PDash. */
	ENABLE_MONITORING_UPLOAD("com.teamscale.enable-monitoring-uploads"),

	/**
	 * Enables the execution of Roslyn Analysers from within Teamscale using RAËX.
	 */
	ENABLE_RAEX("com.teamscale.feature-toggle.raex"),

	/** Enables the developer mode. */
	ENABLE_DEV_MODE("com.teamscale.dev-mode"),

	/**
	 * Disables automatic updating of data classes when dev mode is on. This is used
	 * during UI tests on the build server.
	 */
	DISABLE_UI_DEV_SERVER_MODE("com.teamscale.ui-dev-server-mode"),

	/**
	 * Enables the teamscale.io mode, which enables making repositories public, adds
	 * a special role for provisioned users etc.
	 */
	TEAMSCALE_IO_MODE("teamscale.io-mode"),

	/**
	 * Allows disabling simulink dependency extraction as a work-around for
	 * TS-21836.
	 */
	DISABLE_SIMULINK_DEPENDENCY_EXTRACTION("com.teamscale.simulink.dependency-extracion.disable"),

	/**
	 * Sets the normally hidden option for the instance label server option to
	 * visible. See the InstanceLabelOption class for more info.
	 */
	INSTANCE_LABEL_OPTION_VISIBLE("com.teamscale.instance-label-option.visible"),

	/** Disable creation of default analysis profiles */
	DISABLE_DEFAULT_ANALYSIS_PROFILES_CREATION(
			"com.teamscale.feature-toggle.disable-default-analysis-profiles-creation"),

	/** Show no warning, when the admin user admin/admin is present. */
	NO_WARN_ON_UNSAFE_ADMIN_USER("com.teamscale.feature-toggle.no-warn-on-unsafe-admin-user"),

	/**
	 * Enables a check-category in Model Advisor reports for simulink that is
	 * specific to Nordex and should be visible in only their analysis profiles.
	 */
	ENABLE_NORDEX_MODEL_ADVISOR_REPORT_ITEM("com.teamscale.feature-toggle.enable-nordex-model-advisor-report-item"),

	/** Whether to log performance details or not. */
	ENABLE_LOG_PERFORMANCE_DETAILS("com.teamscale.log_performance_details"),

	/**
	 * Whether to log performance details for the scheduler or not (this is a lot of
	 * data). Only is considered when ENABLE_LOG_PERFORMANCE_DETAILS is also
	 * enabled.
	 */
	ENABLE_LOG_SCHEDULER_PERFORMANCE_DETAILS("com.teamscale.log_scheduler_performance_details"),

	/**
	 * Enable the automatic regular garbage collection in the git repository
	 * connection.
	 */
	ENABLE_GIT_AUTO_GC("com.teamscale.experimental.git-auto-gc"),

	/**
	 * Enables features that only make sense for Apple customer, such as Apple
	 * Connect integration.
	 */
	@Hidden // Apple must not be present in OpenApi
	ENABLE_APPLE_SPECIFIC_FEATURES("com.teamscale.customers.enable-apple-features"),

	/** Enables the test mode. */
	ENABLE_TEST_MODE("com.teamscale.test-mode"),

	/**
	 * Enabling this will add a daily maintenance trigger that removes orphaned data
	 * deduplication entries, which allows removing historic upload data without a
	 * full reanalysis while still being able to reclaim storage space.
	 */
	CLEANUP_ORPHANED_DATABASE_ENTIRES("com.teamscale.cleanup-database"),

	/**
	 * Disable PersistenceMaintenanceTrigger. In some scenarios (very large DBs) the
	 * maintenance can take a very long time and has almost no impact on the DBs
	 * performance. So disabling it improves the overall system's performance.
	 */
	DISABLE_PERSISTENCE_MAINTENANCE_TRIGGER("com.teamscale.disable-persistence-maintenance"),

	/**
	 * Can be used to configure the amount of files that are allowed in a gerrit
	 * patch set. If the patch set contains more changed files than configured here,
	 * we will skip the patch set to avoid long-running analysis and blocking
	 * progress on the instance. Defaults to 1000
	 */
	GERRIT_PATCHSET_MAX_FILES_CHANGED_FILTER_COUNT("com.teamscale.gerrit.patchset-max-files"),

	/**
	 * Disables the dead connector check in the scheduler in case this causes more
	 * problems than it solves.
	 */
	DISABLE_DEAD_CONNECTOR_CHECK("com.teamscale.disable-dead-connector-check"),

	/**
	 * Enable the use of pre-emptive basic auth in the git connections. This saves
	 * one roundtrip to the git server, if using a username/password for fetching
	 * changes.
	 */
	ENABLE_PREEMPTIVE_GIT_BASIC_AUTH("com.teamscale.enable-preemptive-git-basic-auth"),

	/** Disables extracting test case implementations from code. */
	DISABLE_TEST_CASE_EXTRACTION("com.teamscale.disable-test-case-extraction"),

	/**
	 * Enable the use of auto importing LDAP groups. This applies to all LDAP
	 * connectors. The auto import is triggered when the groups are synchronized.
	 * The newly created groups can have a base group for copying the initial
	 * permissions by defining the
	 * {@link LDAPSynchronizeGroupsAsyncStrategy#PERMISSIONS_BASE_GROUP_PROPERTY_NAME}
	 * system property.
	 */
	ENABLE_LDAP_GROUP_AUTO_IMPORT("com.teamscale.enable-ldap-group-auto-import"),

	/**
	 * Enabled more detailed debugging output for rollbacks in case of store
	 * overlaps.
	 */
	ENABLE_SCHEDULER_STORE_OVERLAP_DEBUG("com.teamscale.scheduler-store-overlap-debug"),

	/**
	 * Allows to disable the truncation of the worker log (the detailed worker log
	 * is truncated if too many messages are logged due to memory reasons. This
	 * toggle allows to skip this truncation if memory consumption is not an issue
	 * to be able to get the full log messages).
	 */
	DISABLE_WORKER_LOG_TRUNCATION("com.teamscale.disable-worker-log-truncation"),

	/**
	 * Flag that indicates that full deltas should be recomputed by every single
	 * commit. This is a work-around to compensate the effects of external uploads
	 * with missing parent commit.
	 */
	FORCE_FULL_EXTERNAL_UPLOAD_DELTA("com.teamscale.external-uploads.enable-full-delta"),

	/**
	 * Restricts providing the detailed version number via
	 * TeamscaleVersionApiInfoService and OpenApiService. When set, the version is
	 * shortened to Major.Minor.0.
	 */
	DISABLE_VERSION_DISPLAY("com.teamscale.feature-toggle.disable-version-display"),
	/**
	 * Allows to disable registering FileSystemRepositoryConnectorDescriptor and
	 * MultiVersionFileSystemRepositoryConnectorDescriptor
	 */
	DISABLE_FILE_SYSTEM_CONNECTOR("com.teamscale.feature-toggle.disable-file-system-connector"),

	/**
	 * Allows to show or hide languages in the UI that we support under the hood,
	 * but do not want to appear in the UI for now.
	 */
	SHOW_HIDDEN_LANGUAGES("com.teamscale.feature-toggle.show-hidden-languages"),

	/**
	 * Disables reparsing of unchanged C/C++/Objective-C files if any of their
	 * included headers changed.
	 * <p>
	 * In normal operation (without this Feature toggle), Teamscale will analyze
	 * #include relations between C-preprocessor files. If a commit changes file A
	 * that is (transitively) included into file B, but does not change B, then we
	 * will also preprocess/scan/parse B again. If the resulting TokenElementInfo
	 * object for B changes, we will run the analysis pipeline for B, too.
	 * <p>
	 * By enabling this feature toggle, you can remove this behavior (then a file
	 * will be analyzed only if its own content is changed in a commit).
	 * <p>
	 * Enabling this feature toggle on a running instance requires no reanalysis (as
	 * reparsing is not needed any more).
	 * <p>
	 * Disabling the feature toggle on a running instance without reanalysis might
	 * have unintended effects: Some files that would need reparsing will not be
	 * reparsed because their dependencies will be missing from the
	 * {@link org.conqat.engine.index.resource.reparsing_dependency.PreprocessorIncludeReparsingDependencyIndex}.
	 */
	CPP_DISABLE_REPARSE_ON_HEADER_CHANGE("com.teamscale.experimental.cpp-disable-reparse-on-header-change"),

	/**
	 * Disables automatic precommit analysis (precommit3) on the server. Any
	 * precommit3-calls by IDE plugins will be rejected with a corresponding error
	 * message.
	 */
	PRECOMMIT3_DISABLED("com.teamscale.diable_automatic_precommit_analysis"),

	/**
	 * Whether to disable the auto-repair logic that tries to merge inconsistent
	 * commits in the {@link BackupReader}. See TS-33213
	 */
	DISABLE_AUTO_REPAIR_INCONSISTENT_COMMITS_ON_BACKUP_IMPORT(
			"com.teamscale.index.backup.disable-auto-repair-backup-on-import"),

	/**
	 * A feature toggle that allows to disable merging of coverage. If the toggle is
	 * set, coverage is no longer merged from source to target branches.
	 */
	DISABLE_COVERAGE_MERGING("com.teamscale.disable-coverage-merging"),

	/**
	 * Changes the common behaviour of custom finding description provided in the
	 * config folder overriding the original description. Instead, if this option is
	 * set, the provided custom finding description is added as prefix followed by
	 * original finding description.
	 */
	PREFIX_CUSTOM_FINDING_DESCRIPTION("com.teamscale.feature-toggle.prefix-custom-finding-description"),

	/**
	 * Enables running the SonarLint analysis using an independent external service.
	 * This means that SonarLint executions will no longer block other analysis from
	 * continuing, but Teamscale still needs to wait for the SonarLint analysis
	 * results before the analysis on a commit can be completed. This may loop back
	 * to the Teamscale server itself, as it also implements the API of an
	 * independent analysis service. This is useful in scenarios where we otherwise
	 * would not optimally utilize system resources. (TS-34643)
	 */
	ENABLE_INDEPENDENT_SONAR_LINT_ANALYSIS("com.teamscale.enable-independent-sonar-lint-analysis"),

	/**
	 * Enables outsourcing the Clang-tidy analysis tasks to an independent external
	 * server. This means that Clang-tidy executions will no longer block other
	 * analysis from continuing, but Teamscale still needs to wait for the
	 * Clang-tidy analysis results before the analysis on a commit can be completed.
	 * This may loop back to the Teamscale server itself, as it also implements the
	 * API of an independent analysis service. This is useful in scenarios where we
	 * otherwise would not optimally utilize system resources. (TS-25912)
	 */
	ENABLE_OUTSOURCED_CLANG_TIDY_ANALYSIS("com.teamscale.enable-outsourced-clang-tidy-analysis"),

	/**
	 * Disables the bullseye coverage report parsing using control flow analysis and
	 * falls back to a simpler heuristic.
	 */
	DISABLE_BULLSEYE_CONTROL_FLOW_PARSING("com.teamscale.disable-bullseye-control-flow-parsing"),

	/** Whether audit features should be enabled */
	@Hidden
	AUDIT_FEATURES("com.teamscale.audit-features"),

	/**
	 * Enables extra thorough checks whether we can access the repository root, as
	 * necessary for TS-15203.
	 */
	SVN_THOROUGH_AUTH_CHECK_SUPPORT("com.teamscale.enable-svn-thorough-auth-check-support"),

	/** Enable the project wizard for importing example projects. */
	ENABLE_PROJECT_WIZARD("com.teamscale.project_wizard.enabled"),

	/** Enables the use of multithreading for database reads and scans. */
	ENABLE_MULTITHREADED_DB_READS("com.teamscale.multithreaded_db_reads.enabled"),

	/** Shows retrospectives on left sidebar under Quality Control tab */
	ENABLE_RETROSPECTIVES_UI("com.teamscale.show_retrospectives_ui"),

	/** Enables support for AI features. */
	AI_SUPPORT("com.teamscale.ai-support.enabled"),

	/**
	 * Enables the {@link PatternMatchingCache} for architecture analysis. It is
	 * disabled by default it scales quite badly with number files * number of
	 * patterns and might decrease performance or lead to OOM's in case of many
	 * architectures.
	 */
	ARCHITECTURE_CACHING("com.teamscale.architecture-cache");

	/**
	 * The feature toggle's id. Also used as key for the corresponding System
	 * property.
	 */
	private final String id;

	/**
	 * Determines whether Webpack should be used to compile sources in watch mode
	 * and hot reload modules.
	 */
	public static boolean isInUiDevServerMode() {
		return ENABLE_DEV_MODE.isEnabled() && !DISABLE_UI_DEV_SERVER_MODE.isEnabled();
	}

	/**
	 * Determines whether the server is executed in the UI test mode (UI tests).
	 */
	public static boolean isInUiTestServerMode() {
		return ENABLE_DEV_MODE.isEnabled() && DISABLE_UI_DEV_SERVER_MODE.isEnabled();
	}

	/** Returns whether the given feature toggle is enabled or not. */
	public boolean isEnabled() {
		return Boolean.getBoolean(id);
	}

	/** Constructor */
	EFeatureToggle(String id) {
		this.id = id;
	}

	/** @see #id */
	public String getId() {
		return id;
	}

	/** Allows enabling/disabling the feature toggle. */
	public void setFeatureToggle(boolean value) {
		System.setProperty(id, Boolean.toString(value));
	}
}
