/*-------------------------------------------------------------------------+
 |                                                                          |
 | Copyright 2005-2011 the ConQAT Project                                   |
 |                                                                          |
 | Licensed under the Apache License, Version 2.0 (the "License");          |
 | you may not use this file except in compliance with the License.         |
 | You may obtain a copy of the License at                                  |
 |                                                                          |
 |    http://www.apache.org/licenses/LICENSE-2.0                            |
 |                                                                          |
 | Unless required by applicable law or agreed to in writing, software      |
 | distributed under the License is distributed on an "AS IS" BASIS,        |
 | WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
 | See the License for the specific language governing permissions and      |
 | limitations under the License.                                           |
 +-------------------------------------------------------------------------*/
package org.conqat.ide.commons.ui.adt;

import java.util.Arrays;

import org.checkerframework.checker.nullness.qual.Nullable;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;

import com.google.common.base.Preconditions;
import com.sap.adt.communication.destinations.AdtDestinationRegistryFactory;
import com.sap.adt.communication.destinations.IDestinationRegistry;
import com.sap.adt.project.AdtCoreProjectServiceFactory;
import com.sap.adt.project.IAdtCoreProjectService;
import com.teamscale.ide.commons.model.PrefixMapping;

/**
 * Utility functions for ADT projects.
 */
public class AdtUtils {

	/** The ABAP nature to identify an ABAP-project. */
	private static final String ABAP_NATURE = "com.sap.adt.abapnature";

	/** The default branch of an ABAP project. */
	public static final String ABAP_DEFAULT_BRANCH = "master";

	/** The default prefix mapping of an ABAP project. */
	public static final PrefixMapping ABAP_DEFAULT_PREFIX_MAPPING = new PrefixMapping("", "");

	/**
	 * Checks if the the resource is an ADT-Resource. If resource is
	 * <code>null</code>, or the containing project is closed or non-existing, false
	 * is returned.
	 */
	public static boolean isAdtResource(@Nullable IResource resource) {
		if (resource == null) {
			return false;
		}
		return isAdtProject(resource.getProject());
	}

	/**
	 * Checks if the the project is an ADT-project. If the project is closed or
	 * non-existing, false is returned.
	 */
	public static boolean isAdtProject(@Nullable IProject project) {
		if (project == null) {
			return false;
		}
		try {
			IProjectDescription description = project.getDescription();
			return Arrays.asList(description.getNatureIds()).contains(ABAP_NATURE);
		} catch (CoreException e) {
			// exception indicate closed or non-existing projects
			return false;
		}
	}
	
	/**
	 * Checks if the ADT-project is connected to a SAP system.
	 * <p>
	 * The project is connected to a SAP system if the user has logged into the project.
	 */
	public static boolean isAdtProjectConnected(IProject project) {
		Preconditions.checkArgument(isAdtProject(project), "The project is expected to be an ADT project.");
		
		IAdtCoreProjectService adtProjectService = AdtCoreProjectServiceFactory.createCoreProjectService();
		String destinationId = adtProjectService.getDestinationId(project);
		if (destinationId == null) {
			return false;
		}
		IDestinationRegistry destinationRegistry = AdtDestinationRegistryFactory.getDestinationRegistry();
		return destinationRegistry.isRegistered(destinationId);
	}
}
