/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |

|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package org.conqat.ide.commons.ui.dialog;

import static java.util.Objects.requireNonNull;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.jface.dialogs.TitleAreaDialog;
import org.eclipse.jface.fieldassist.ControlDecoration;
import org.eclipse.jface.fieldassist.FieldDecorationRegistry;
import org.eclipse.jface.layout.GridDataFactory;
import org.eclipse.jface.layout.GridLayoutFactory;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.FocusAdapter;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

/**
 * A convenience class that allows the quick creation of form-style dialogs.
 * Each form field in the dialog consists of a label and an input area.
 */
public abstract class FormInputDialog extends TitleAreaDialog {

	/**
	 * Implementation of the {@link IFieldValidator} interface to be used with a
	 * text field.
	 */
	public static abstract class TextValidatorBase implements IFieldValidator<Text> {

		/** {@inheritDoc} */
		@Override
		public String validate(Text target) {
			return validate(target.getText());
		}

		/**
		 * Validates the text of a text field.
		 * 
		 * @param text
		 *            The text of the text field that should be validated.
		 * @return Null if text is valid, error message if it isn't.
		 */
		protected abstract String validate(String text);

	}

	/**
	 * Implementation of the {@link IFieldValidator} interface to be used with a
	 * combo box.
	 */
	public static abstract class ComboValidatorBase implements IFieldValidator<Combo> {

		/** {@inheritDoc} */
		@Override
		public String validate(Combo target) {
			int index = target.getSelectionIndex();
			String itemValue = target.getText();
			return validate(index, itemValue);
		}

		/**
		 * Validates the selection of a combo box.
		 * 
		 * @param index
		 *            The index of the selected item, -1 if none is selected.
		 * @param itemValue
		 *            The string value of the selected item, null if none is selected.
		 * @return Null if the selection is valid, error message if it isn't.
		 */
		protected abstract String validate(int index, String itemValue);

	}

	/** A field and its associated decoration and validator */
	private static class FormField<T extends Control> {

		/** The field (i.e. a control that can hold a value) */
		private final T field;

		/** The validator */
		private final IFieldValidator<T> validator;

		/** The decoration that should be shown when the validation fails */
		private final ControlDecoration decoration;

		/** Constructor */
		public FormField(T field, IFieldValidator<T> validator, ControlDecoration decoration) {
			this.field = field;
			this.validator = validator;
			this.decoration = decoration;
		}

		/**
		 * Validate field
		 * 
		 * @param displayErrors
		 *            Flag, true if found errors should be displayed.
		 * @return True if the field is valid, false otherwise.
		 */
		private boolean validate(boolean displayErrors) {
			String errorMessage = null;

			if (validator != null) {
				errorMessage = validator.validate(field);
			}

			if (errorMessage == null) {
				decoration.hide();
				return true;
			}

			if (displayErrors) {
				decoration.setDescriptionText(errorMessage);
				decoration.show();
			}
			return false;
		}

		/** Set if field is enabled/disabled. */
		public void setEnabled(boolean enabled) {
			field.setEnabled(enabled);
		}
	}

	/** List of form fields shown in this form */
	private final List<FormField<?>> formFields = new ArrayList<>();

	/** The composite with grid layout used to display the fields */
	private Composite formGrid;

	/** The title of the dialog window (shell text) */
	private final String windowTitle;

	/** The text for the dialog's message. */
	private final String messageText;

	/** The text for the dialog's title area. */
	private final String titleText;

	/** Constructor */
	public FormInputDialog(Shell parentShell, String windowTitle, String titleText, String message) {
		super(parentShell);
		this.windowTitle = requireNonNull(windowTitle);
		this.titleText = requireNonNull(titleText);
		this.messageText = requireNonNull(message);
	}

	/**
	 * In this methods fields should be created by calling addText, addCombo etc. It
	 * is also valid to add other controls to the composite (it has a grid layout).
	 * 
	 * @param formGrid
	 *            The grid into which the fields are placed.
	 */
	protected abstract void createFields(Composite formGrid);

	/**
	 * Is called, when the validation status of the form is calculated. Can be used
	 * for example for enabling/disabling buttons.
	 * 
	 * @param valid
	 *            True if all form fields are valid, false if not.
	 */
	protected abstract void formValidationStatusChanged(boolean valid);

	/** {@inheritDoc} */
	@Override
	protected Control createContents(Composite parent) {
		Control contents = super.createContents(parent);
		validateFields(false);
		return contents;
	}

	/** {@inheritDoc} */
	@Override
	protected Control createDialogArea(Composite parent) {
		formGrid = new Composite(parent, SWT.NULL);
		formGrid.setLayout(GridLayoutFactory.fillDefaults().numColumns(2).spacing(10, 5).margins(10, 10).create());
		GridDataFactory.fillDefaults().grab(true, true).applyTo(formGrid);
		createFields(formGrid);

		getShell().setText(windowTitle);
		setMessage(messageText);
		setTitle(titleText);

		return formGrid;
	}

	/**
	 * Create a combo box and add it to the form composite.
	 * 
	 * @param labelText
	 *            Text of the label for the combo
	 * @param values
	 *            The values in the combo
	 * @param defaultSelectionIndex
	 *            The index of the default item (-1 if none should be selected)
	 * @param validator
	 *            Validator for the combo
	 * @return The create combo
	 */
	protected Combo addCombo(String labelText, String[] values, int defaultSelectionIndex, int style,
			ComboValidatorBase validator) {

		createFormLabel(formGrid, labelText);
		Combo combo = new Combo(formGrid, SWT.DROP_DOWN | style);
		combo.setItems(values);
		if (defaultSelectionIndex >= 0) {
			combo.select(defaultSelectionIndex);
		}

		initField(combo, validator);

		combo.addModifyListener(new ModifyListener() {
			@Override
			public void modifyText(ModifyEvent e) {
				validateFields(true);
			}
		});

		combo.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				// This is called when the user selects an item
				// Do nothing
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e) {
				// This is called when the user presses enter
				if (validateFields(true)) {
					okPressed();
				}
			}
		});
		return combo;
	}

	/**
	 * Create a text field and add it to the form composite.
	 * 
	 * @param labelText
	 *            Text of the label for the combo
	 * @param defaultValue
	 *            The default text
	 * @param textStyle
	 *            Style for the text (for password use SWT.PASSWORD)
	 * @param validator
	 *            The validator for the text
	 * @return The created text
	 */
	protected Text addText(String labelText, String defaultValue, int textStyle, TextValidatorBase validator) {
		createFormLabel(formGrid, labelText);

		Text textField = new Text(formGrid, textStyle | SWT.BORDER);
		GridDataFactory.fillDefaults().grab(true, false).applyTo(textField);

		if (defaultValue != null) {
			textField.setText(defaultValue);
		}

		initField(textField, validator);

		textField.addModifyListener(new ModifyListener() {
			@Override
			public void modifyText(ModifyEvent e) {
				validateFields(true);
			}
		});

		textField.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				// Do nothing
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e) {
				okPressed();
			}
		});
		return textField;
	}

	/**
	 * Create a check-box and add it to the form composite.
	 * 
	 * @param labelText
	 *            text of the label for the combo
	 * @param defaultValue
	 *            the default selection value
	 * @return the created check-box
	 */
	protected Button addCheckBox(String labelText, boolean defaultValue) {
		createFormLabel(formGrid, labelText);

		Button checkBox = new Button(formGrid, SWT.CHECK);
		checkBox.setSelection(defaultValue);

		return checkBox;
	}

	/** Initialize and register a field */
	private <T extends Control> void initField(T field, IFieldValidator<T> validator) {

		GridDataFactory.fillDefaults().grab(true, false).applyTo(field);
		final ControlDecoration nameDecoration = new ControlDecoration(field, SWT.LEFT | SWT.TOP);

		Image errorDecorationImage = FieldDecorationRegistry.getDefault()
				.getFieldDecoration(FieldDecorationRegistry.DEC_ERROR).getImage();
		nameDecoration.setImage(errorDecorationImage);
		nameDecoration.hide();
		final FormField<T> formField = new FormField<>(field, validator, nameDecoration);
		formFields.add(formField);

		field.addFocusListener(new FocusAdapter() {

			@Override
			public void focusLost(FocusEvent e) {
				formField.validate(true);
			}
		});
	}

	/**
	 * Checks if all fields displayed in this form are valid.
	 * 
	 * @param displayErrors
	 *            Flag, true if found errors should be displayed.
	 * @return True if all fields are valid, false if not.
	 */
	protected boolean validateFields(boolean displayErrors) {
		boolean allValid = true;
		for (FormField<?> formField : formFields) {
			if (!formField.validate(displayErrors)) {
				allValid = false;
			}
		}
		formValidationStatusChanged(allValid);
		return allValid;
	}

	/** Create label with default properties and {@link GridData} */
	private static Label createFormLabel(Composite composite, String text) {
		Label label = new Label(composite, SWT.NONE);
		GridDataFactory.fillDefaults().align(SWT.RIGHT, SWT.CENTER).applyTo(label);
		label.setText(text);
		return label;
	}

	/** Set whether the form fields should be enabled */
	protected void setFormEnabled(boolean enabled) {
		for (FormField<?> formField : formFields) {
			formField.setEnabled(enabled);
		}
	}
}
