/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package org.conqat.ide.commons.ui.jface;

import org.eclipse.jface.layout.GridDataFactory;
import org.eclipse.jface.layout.GridLayoutFactory;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IContentProvider;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.ViewerSorter;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;

/**
 * A composite widget that supports Create Update Delete (CRUD) operations on a
 * list of generic elements. The widget consists of a TableViewer and buttons
 * for adding, editing, and deleting items.
 * 
 * Clients should subclass and override the abstract methods.
 * 
 * @author bader svejda
 */
public abstract class CRUDListWidget<T> extends Composite {

	/** Table for displaying the elements */
	private TableViewer tableViewer;

	/** The "add element" button. */
	private Button addButton;

	/** The "delete element" button */
	private Button deleteButton;

	/** The "edit element" button */
	private Button editButton;

	/** The type of the underlying model. */
	Class<T> modelType;

	/** Constructor. */
	public CRUDListWidget(Composite parent) {
		super(parent, SWT.NONE);
		setLayout(GridLayoutFactory.fillDefaults().numColumns(2).create());
		createTableViewer();
		createAddButton();
		createEditButton();
		createDeleteButton();
	}

	/** Creates the table viewer for the element list. */
	private void createTableViewer() {
		tableViewer = new TableViewer(this, SWT.FULL_SELECTION | SWT.BORDER);
		GridDataFactory.fillDefaults().span(1, 7).grab(true, true).applyTo(tableViewer.getControl());

		tableViewer.getTable().setHeaderVisible(true);

		addTableViewerListeners();
	}

	/** Hooks up the listeners of the table viewer. */
	private void addTableViewerListeners() {
		tableViewer.addSelectionChangedListener(new ISelectionChangedListener() {
			@Override
			public void selectionChanged(SelectionChangedEvent event) {
				boolean contextButtonsEnabled = !event.getSelection().isEmpty();
				editButton.setEnabled(contextButtonsEnabled);
				deleteButton.setEnabled(contextButtonsEnabled);
			}
		});

		tableViewer.addDoubleClickListener(new IDoubleClickListener() {
			@Override
			public void doubleClick(DoubleClickEvent event) {
				edit(getSelectedElement());
				refresh();
			}
		});
	}

	/** Add column with the given caption. */
	public void addColumn(String caption) {
		TableViewerColumn column = new TableViewerColumn(tableViewer, SWT.LEFT);
		column.getColumn().setText(caption);
		column.getColumn().setWidth(120);
	}

	/** Create the "Add" button */
	private void createAddButton() {
		addButton = createButton(this, "Add", SWT.PUSH, new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				add();
				refresh();
			}
		});
	}

	/** Create the "Edit" button */
	private void createEditButton() {
		editButton = createButton(this, "Edit", SWT.PUSH, new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				edit(getSelectedElement());
				refresh();
			}
		});
		editButton.setEnabled(false);
	}

	/** Create the "Delete" button */
	private void createDeleteButton() {
		deleteButton = createButton(this, "Delete", SWT.PUSH, new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				delete(getSelectedElement());
				refresh();
			}
		});
		deleteButton.setEnabled(false);
	}

	/**
	 * Creates and returns a {@link Button} in the given parent with the given text,
	 * the given style and with the given selection listener. A default (grid)
	 * layout is applied to the button.
	 */
	private static Button createButton(Composite parent, String text, int style, SelectionListener selectionListener) {
		Button button = new Button(parent, style);
		GridDataFactory.defaultsFor(button).applyTo(button);
		button.setText(text);
		button.addSelectionListener(selectionListener);
		return button;
	}

	/** Returns the currently selected list element. */
	@SuppressWarnings("unchecked")
	private T getSelectedElement() {
		IStructuredSelection structuredSelection = getTableSelection();
		return (T) structuredSelection.getFirstElement();
	}

	/** Returns selection of the table viewer. */
	private IStructuredSelection getTableSelection() {
		return (IStructuredSelection) tableViewer.getSelection();
	}

	/** {@inheritDoc} */
	@Override
	public void setEnabled(boolean enabled) {
		super.setEnabled(enabled);
		tableViewer.getTable().setEnabled(enabled);
		addButton.setEnabled(enabled);
		boolean nonEmptySelection = !getTableSelection().isEmpty();
		editButton.setEnabled(enabled && nonEmptySelection);
		deleteButton.setEnabled(enabled && nonEmptySelection);
	}

	/**
	 * Set the content provider of the element list.
	 * 
	 * @see org.eclipse.jface.viewers.ContentViewer#setContentProvider(org.eclipse.jface.viewers.IContentProvider)
	 */
	protected void setContentProvider(IContentProvider provider) {
		tableViewer.setContentProvider(provider);
	}

	/** Set {@link ITableLabelProvider} on the {@link TableViewer}. */
	protected void setLabelProvider(ITableLabelProvider labelProvider) {
		tableViewer.setLabelProvider(labelProvider);
	}

	/**
	 * Set the sorter of the element list.
	 * 
	 * @see org.eclipse.jface.viewers.StructuredViewer
	 */
	protected void setSorter(ViewerSorter sorter) {
		tableViewer.setSorter(sorter);
	}

	/**
	 * Set the input provider of the element list.
	 * 
	 * @see org.eclipse.jface.viewers.Viewer#setInput(java.lang.Object)
	 */
	protected void setInput(Object input) {
		tableViewer.setInput(input);
	}

	/** Refresh the contents of the element list. */
	protected void refresh() {
		tableViewer.refresh();
	}

	/**
	 * Called when a new element should be added to the list. The list will refresh
	 * automatically after this method returns.
	 */
	protected abstract void add();

	/**
	 * Called if an element is selected and the "Edit" button is clicked OR if an
	 * element in the list is double clicked. The list will refresh automatically
	 * after this method returns.
	 * 
	 * @param selectedElement
	 *            the element that should be edited
	 */
	protected abstract void edit(T selectedElement);

	/**
	 * Called if an element is selected and the "Delete" button is clicked. The list
	 * will refresh automatically after this method returns.
	 * 
	 * @param selectedElement
	 *            the element that should be deleted
	 */
	protected abstract void delete(T selectedElement);
}
