/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.conqat.lib.commons.assessment;

import java.util.ArrayList;
import java.util.List;

import org.conqat.lib.commons.js_export.ExportToTypeScript;
import org.conqat.lib.commons.test.IndexValueClass;

/**
 * Enum for traffic light colors.
 *
 * Note that the order is relevant: The first Color (RED) is considered the most
 * dominant color (see {@link Assessment#getDominantColor()}).
 *
 * The mapping to actual colors is defined in
 * {@link AssessmentUtils#getColor(ETrafficLightColor)}, so typically this
 * method should be adjusted when a new constant is introduced here.
 */
@ExportToTypeScript
@IndexValueClass(containedInBackup = true)
public enum ETrafficLightColor {

	/** Red signals errors or incompleteness. */
	RED("#dd513a", "Red"),

	/**
	 * Orange is an intermediate state between {@link #RED} and {@link #YELLOW}.
	 */
	ORANGE("#ff9966", "Orange"),

	/** Yellow signals warning or lack of control. */
	YELLOW("#fde725", "Yellow"),

	/** Green signals the absence of errors or correctness. */
	GREEN("#44bf70", "Green"),

	/**
	 * Baseline indicates a baseline entry, i.e. there has been no change compared
	 * to a given baseline and thus no color applies.
	 */
	BASELINE("#dfeaf4", "Baseline"),

	/** This is used if no information is available. */
	UNKNOWN;

	/** Display text of the color (name with normal casing) */
	private final String displayText;

	/** Holds a display name used in the context of assessments. */
	private final String assessmentDisplayName;

	/** Hex representation of the color */
	private final String hexValue;

	/**
	 * This is hex value is used if no color is associated with the
	 * ETrafficLightColor.
	 */
	private static final String HEX_VALUE_BLACK = "#000000";

	/**
	 * Constructor for ETrafficLightColors that do not have an associated color.
	 */
	ETrafficLightColor() {
		this(HEX_VALUE_BLACK, "");
	}

	/** Constructor for ETrafficLightColors that have an associated color. */
	ETrafficLightColor(String hexValue, String assessmentDisplayName) {
		displayText = name().substring(0, 1) + name().substring(1).toLowerCase();
		this.assessmentDisplayName = assessmentDisplayName;
		this.hexValue = hexValue;
	}

	/** @see #displayText */
	public String getDisplayText() {
		return displayText;
	}

	public String getAssessmentDisplayName() {
		return assessmentDisplayName;
	}

	/** @see #hexValue */
	public String getHexValue() {
		return hexValue;
	}

	/**
	 * Returns the more dominant color, which is the enum literal with smaller index
	 * (as they are sorted by dominance).
	 */
	public static ETrafficLightColor getDominantColor(ETrafficLightColor color1, ETrafficLightColor color2) {
		if (color2.ordinal() < color1.ordinal()) {
			return color2;
		}
		return color1;
	}

	/**
	 * Returns the traffic light colors red, yellow and green in this particular
	 * order.
	 */
	public static List<ETrafficLightColor> getTrafficLightColors() {
		List<ETrafficLightColor> trafficLights = new ArrayList<>();
		trafficLights.add(RED);
		trafficLights.add(YELLOW);
		trafficLights.add(GREEN);
		return trafficLights;
	}

}
