package org.conqat.lib.commons.collections;

import java.util.EnumSet;
import java.util.Set;
import java.util.function.BiConsumer;
import java.util.function.BinaryOperator;
import java.util.function.Function;
import java.util.stream.Collector;

import org.checkerframework.checker.nullness.qual.NonNull;
import org.conqat.lib.commons.assertion.CCSMAssert;

/**
 * Base class for stream {@link Collector}s of {@link CollectionMap}s.
 * 
 * @param <T>
 *            the type of input elements to the reduction operation
 * @param <K>
 *            the key type
 * @param <V>
 *            the value type (i.e. the values stored in the collections)
 * @param <R>
 *            the resulting {@link CollectionMap} type.
 */
/* package */ abstract class CollectionMapCollectorBase<T, K, V, R extends CollectionMap<K, V, ?>>
		implements Collector<T, R, R> {

	/** Mapper from stream objects to {@link CollectionMap} keys. */
	private final Function<T, K> keyMapper;

	/** Mapper from stream objects to {@link CollectionMap} values. */
	private final Function<T, V> valueMapper;

	protected CollectionMapCollectorBase(@NonNull Function<T, K> keyMapper, @NonNull Function<T, V> valueMapper) {
		CCSMAssert.isNotNull(keyMapper, () -> String.format("Expected \"%s\" to be not null", "keyMapper"));
		CCSMAssert.isNotNull(valueMapper, () -> String.format("Expected \"%s\" to be not null", "valueMapper"));
		this.keyMapper = keyMapper;
		this.valueMapper = valueMapper;
	}

	@Override
	public BiConsumer<R, T> accumulator() {
		return (collection, element) -> collection.add(keyMapper.apply(element), valueMapper.apply(element));
	}

	@Override
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public BinaryOperator<R> combiner() {
		return (collection1, collection2) -> {
			// raw type required, because addAll (needlessly) expects exactly the same
			// collection type
			collection1.addAll((CollectionMap) collection2);
			return collection1;
		};
	}

	@Override
	public Function<R, R> finisher() {
		return Function.identity();
	}

	@Override
	public Set<Characteristics> characteristics() {
		return EnumSet.of(Characteristics.IDENTITY_FINISH);
	}
}
