/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.conqat.lib.commons.collections;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

/**
 * Methods for sorting instances of {@link CounterSet}.
 * <p>
 * This code stems originally from the frontend and was moved to the backend to
 * allow for server-side sorting of huge tables.
 */
public class CounterSetSorting {

	/**
	 * Build a comparator to compare two instances of {@code CounterSet}. The
	 * ordering of keys is determined by {@code sortOrder}.
	 */
	public static Comparator<CounterSet<Object>> comparator(Object[] allKeyCandidatesOrdered, Object[][] sortOrder) {
		int[][] sortOrderIndices = new int[sortOrder.length][];
		for (int i = 0; i < sortOrder.length; i++) {
			sortOrderIndices[i] = keysToKeyIndicies(allKeyCandidatesOrdered, sortOrder[i]);
		}
		return comparator(allKeyCandidatesOrdered, sortOrderIndices);
	}

	/**
	 * Build a comparator to compare two instances of {@code CounterSet}. The
	 * ordering of keys is determined by {@code sortOrder}, which are indices in
	 * {@code allKeyCandidatesOrdered}.
	 */
	public static <E> Comparator<CounterSet<E>> comparator(E[] allKeyCandidatesOrdered, int[][] sortOrder) {
		return (set1, set2) -> {
			boolean firstIsEmpty = set1 == null || set1.getKeys().isEmpty();
			boolean secondIsEmpty = set2 == null || set2.getKeys().isEmpty();
			if (firstIsEmpty && !secondIsEmpty) {
				return -1;
			} else if (!firstIsEmpty && secondIsEmpty) {
				return 1;
			} else if (firstIsEmpty && secondIsEmpty) {
				return 0;
			}

			Pair<double[], Long> percentages1 = toPercentages(set1.transformToList(allKeyCandidatesOrdered));
			Pair<double[], Long> percentages2 = toPercentages(set2.transformToList(allKeyCandidatesOrdered));

			if (percentages1.getSecond() == 0 || percentages2.getSecond() == 0) {
				return Math.round(percentages1.getSecond() - percentages2.getSecond());
			}

			return compareSeverities(percentages1.getFirst(), percentages2.getFirst(), sortOrder);
		};
	}

	private static int[] keysToKeyIndicies(Object[] orderedKeys, Object[] returnIndicesOf) {
		int[] result = new int[returnIndicesOf.length];
		for (int i = 0; i < returnIndicesOf.length; i++) {
			Object indexOf = returnIndicesOf[i];
			boolean found = false;
			for (int keyIndex = 0; keyIndex < orderedKeys.length; keyIndex++) {
				if (orderedKeys[keyIndex].equals(indexOf)) {
					result[i] = keyIndex;
					found = true;
				}
			}

			if (!found) {
				throw new IllegalArgumentException(String.format("Was not able to lookup key index for %s", indexOf));
			}
		}

		return result;
	}

	private static Pair<double[], Long> toPercentages(int[] list) {
		long sum = Arrays.stream(list).sum();
		double[] result = new double[list.length];
		for (int i = 0; i < list.length; i++) {
			int element = list[i];
			if (sum == 0) {
				result[i] = Double.NaN;
			} else {
				result[i] = (double) element / sum;
			}
		}
		return Pair.createPair(result, sum);
	}

	private static int compareSeverities(double[] lhs, double[] rhs, int[][] sortOrder) {
		List<double[]> groupedLhs = accumulateByGroup(lhs, sortOrder);
		List<double[]> groupedRhs = accumulateByGroup(rhs, sortOrder);

		// Compare the two arrays of arrays lexicographically
		return compare3(groupedLhs, groupedRhs);
	}

	private static List<double[]> accumulateByGroup(double[] inputArray, int[][] sortOrders) {
		List<double[]> result = new ArrayList<>();
		for (int[] subSortOrder : sortOrders) {
			double[] permutedArray = copyByIndex(inputArray, subSortOrder);
			result.add(accumulateLeft(permutedArray));
		}
		return result;
	}

	private static double[] accumulateLeft(double[] array) {
		double[] result = new double[array.length];
		double sum = 0.0;
		for (int i = 0; i < array.length; i++) {
			sum += array[i];
			result[array.length - i - 1] = sum;
		}
		return result;
	}

	private static double[] copyByIndex(double[] inputArray, int[] sortOrder) {
		double[] result = new double[sortOrder.length];
		for (int i = 0; i < sortOrder.length; i++) {
			result[i] = inputArray[sortOrder[i]];
		}
		return result;
	}

	private static int compare3(List<double[]> list1, List<double[]> list2) {
		for (int i = 0; i < Math.min(list1.size(), list2.size()); i++) {
			int comparisonResult = compareArrays(list1.get(i), list2.get(i));
			if (comparisonResult != 0) {
				return comparisonResult;
			}
		}
		return Integer.compare(list1.size(), list2.size());
	}

	// This method compares two arrays lexicographically.
	private static int compareArrays(double[] arr1, double[] arr2) {
		for (int i = 0; i < Math.min(arr1.length, arr2.length); i++) {
			int comparisonResult = Double.compare(arr1[i], arr2[i]);
			if (comparisonResult != 0) {
				return comparisonResult;
			}
		}
		return Integer.compare(arr1.length, arr2.length);
	}

}