/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.conqat.lib.commons.collections;

import org.conqat.lib.commons.test.IndexValueClass;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import org.conqat.lib.commons.string.LineRangeStringParser;

import com.fasterxml.jackson.annotation.JsonCreator;

/**
 * A really simple list for storing ints. This exists as it is both more
 * efficient and uses way less memory than a List of Integers.
 */
@IndexValueClass(containedInBackup = true)
public class IntList extends ManagedIntArray {

	private static final long serialVersionUID = 1L;

	public IntList() {
	}

	public IntList(Collection<Integer> collection) {
		addArrayElements(collection.size());
		Iterator<Integer> iterator = collection.iterator();
		for (int i = 0; iterator.hasNext(); i++) {
			array[i] = iterator.next();
		}
	}

	/**
	 * Creates an IntList from a string containing ranges (denoted with '-'). E.g.,
	 * used for the Testwise Coverage report format.
	 */
	@JsonCreator
	public IntList(String lineNumbersString) {
		addAll(new LineRangeStringParser().parse(lineNumbersString));
	}

	/** Returns the size of the list. */
	public int getSize() {
		return size;
	}

	/**
	 * Returns <code>true</code> if the list does not contain any element,
	 * <code>false</code> otherwise.
	 */
	public boolean isEmpty() {
		return size == 0;
	}

	/**
	 * Returns the element at the given index. No range checking is performed, thus
	 * you might sometimes get an {@link ArrayIndexOutOfBoundsException}, and
	 * sometimes just 0 returned, depending on if you reached existing memory by
	 * chance.
	 */
	public int get(int index) {
		return array[index];
	}

	/**
	 * Set the element at the given index. No range checking is performed, thus you
	 * might sometimes get an {@link ArrayIndexOutOfBoundsException}, and sometimes
	 * not for illegal indexes, depending on whether you hit memory allocated by the
	 * exponential growth strategy by chance.
	 */
	public void set(int index, int value) {
		array[index] = value;
	}

	/** Adds an element to the end of the list. */
	public void add(int value) {
		int index = size;
		addArrayElement();
		array[index] = value;
	}

	/** Adds all elements of the given list to the end of the list. */
	public void addAll(IntList values) {
		for (int i = 0; i < values.getSize(); i++) {
			add(values.get(i));
		}
	}

	/** Returns the int list as an actual list. This is only used for testing. */
	public List<Integer> getAsList() {
		List<Integer> result = new ArrayList<>();
		for (int i = 0; i < getSize(); i++) {
			result.add(get(i));
		}
		return result;
	}
}