/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.conqat.lib.commons.io;

import java.io.IOException;
import java.io.InputStream;
import java.io.InvalidClassException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectStreamClass;
import java.io.OutputStream;
import java.util.Arrays;
import java.util.HashSet;

import org.conqat.lib.commons.collections.CollectionUtils;
import org.conqat.lib.commons.collections.UnmodifiableSet;

import com.google.common.annotations.VisibleForTesting;

/**
 * Used to filter allowed classes in object streams, i.e., during
 * (de)serialization. We only allow certain classes to be (de)serialized, in
 * order to guard against attacks where malicious code is injected by importing
 * a backup that contains an unsafe class.
 */
public class FilteredObjectStreams {

	private static final UnmodifiableSet<String> ALLOWED_PACKAGES = getAllowedPackages();
	private static final UnmodifiableSet<String> BLACKLISTED_NAMES = getBlacklistedNames();

	private static UnmodifiableSet<String> getAllowedPackages() {
		return CollectionUtils.asUnmodifiable(new HashSet<>(Arrays.asList("java", "org.conqat", "com.teamscale",
				"eu.cqse", "com.fasterxml.jackson.databind.node", "com.google.common.collect")));
	}

	@VisibleForTesting
	static void verifyClassIsAllowed(ObjectStreamClass desc) throws InvalidClassException {
		String name = getName(desc);
		if (name.isEmpty()) {
			// Can happen for primitive array types
			return;
		}
		if (BLACKLISTED_NAMES.stream().anyMatch(name::startsWith)) {
			throw new InvalidClassException("Class not supported for (de)serialization, blacklisted", desc.getName());
		}
		if (packageNameIsNotAllowed(name)) {
			throw new InvalidClassException("Class not supported for (de)serialization", desc.getName());
		}
	}

	private static boolean packageNameIsNotAllowed(String className) {
		return ALLOWED_PACKAGES.stream().noneMatch(allowedPackage -> className.startsWith(allowedPackage + "."));
	}

	private static String getName(ObjectStreamClass desc) {
		String name = desc.getName();
		if (name.startsWith("[")) {
			// The JVM names for array classes start with a number of '[' symbols,
			// representing the dimension of the array, followed by a letter or string
			// representing the array type. For instance, [I for int array and [[I for int
			// array array. Object arrays are represented as [L<className>;, such as
			// "[Ljava.lang.Object;" for Object array. We want to strip away all opening
			// brackets, as well as the type info, but keep possible object types. Empty
			// string return values will be accepted as unproblematic.
			return name.replaceFirst("\\[+.(.*);?", "$1");
		}
		return name;
	}

	/**
	 * Returns an input stream that performs filtering on the classes that can be
	 * (de)serialized. Only packages used in Teamscale are allowed.
	 */
	public static ObjectInputStream filteredInputStream(InputStream in) throws IOException {
		return new FilteredObjectInputStream(in);
	}

	/**
	 * Returns an output stream that performs filtering on the classes that can be
	 * (de)serialized. Only packages used in Teamscale are allowed.
	 */
	public static ObjectOutputStream filteredOutputStream(OutputStream out) throws IOException {
		return new FilteredObjectOutputStream(out);
	}

	private static final class FilteredObjectInputStream extends ObjectInputStream {

		public FilteredObjectInputStream(InputStream in) throws IOException {
			super(in);
		}

		@Override
		protected Class<?> resolveClass(ObjectStreamClass desc) throws IOException, ClassNotFoundException {
			verifyClassIsAllowed(desc);
			return super.resolveClass(desc);
		}

	}

	private static final class FilteredObjectOutputStream extends ObjectOutputStream {

		public FilteredObjectOutputStream(OutputStream out) throws IOException {
			super(out);
		}

		@Override
		protected void writeClassDescriptor(ObjectStreamClass desc) throws IOException {
			verifyClassIsAllowed(desc);
			super.writeClassDescriptor(desc);
		}
	}

	/**
	 * Blacklist of known attacks taken from the <a href=
	 * "https://github.com/ikkisoft/SerialKiller/blob/master/config/serialkiller.conf">SerialKiller
	 * config</a>.
	 */
	private static UnmodifiableSet<String> getBlacklistedNames() {
		HashSet<String> blacklistedNames = new HashSet<>(Arrays.asList(
				// ysoserial's BeanShell1 payload
				"bsh.XThis", //
				"bsh.Interpreter", //
				// ysoserial's C3P0 payload
				"com.mchange.v2.c3p0.impl.PoolBackedDataSourceBase", //
				// ysoserial's CommonsBeanutils1 payload
				"org.apache.commons.beanutils.BeanComparator", //
				// ysoserial's CommonsCollections1,3,5,6 payload
				"org.apache.commons.collections.Transformer", //
				"org.apache.commons.collections.functors.InvokerTransformer", //
				"org.apache.commons.collections.functors.ChainedTransformer", //
				"org.apache.commons.collections.functors.ConstantTransformer", //
				"org.apache.commons.collections.functors.InstantiateTransformer", //
				// ysoserial's CommonsCollections2,4 payload
				"org.apache.commons.collections4.functors.InvokerTransformer", //
				"org.apache.commons.collections4.functors.ChainedTransformer", //
				"org.apache.commons.collections4.functors.ConstantTransformer", //
				"org.apache.commons.collections4.functors.InstantiateTransformer", //
				"org.apache.commons.collections4.comparators.TransformingComparator", //
				// ysoserial's FileUpload1,Wicket1 payload
				"org.apache.commons.fileupload.disk.DiskFileItem", //
				"org.apache.wicket.util.upload.DiskFileItem", //
				// ysoserial's Groovy payload
				"org.codehaus.groovy.runtime.ConvertedClosure", //
				"org.codehaus.groovy.runtime.MethodClosure", //
				// ysoserial's Hibernate1,2 payload
				"org.hibernate.engine.spi.TypedValue", //
				"org.hibernate.tuple.component.AbstractComponentTuplizer", //
				"org.hibernate.tuple.component.PojoComponentTuplizer", //
				"org.hibernate.type.AbstractType", //
				"org.hibernate.type.ComponentType", //
				"org.hibernate.type.Type", //
				"com.sun.rowset.JdbcRowSetImpl", //
				// ysoserial's JBossInterceptors1, JavassistWeld1 payload
				"org.jboss.interceptor.builder.InterceptionModelBuilder", //
				"org.jboss.interceptor.builder.MethodReference", //
				"org.jboss.interceptor.proxy.DefaultInvocationContextFactory", //
				"org.jboss.interceptor.proxy.InterceptorMethodHandler", //
				"org.jboss.interceptor.reader.ClassMetadataInterceptorReference", //
				"org.jboss.interceptor.reader.DefaultMethodMetadata", //
				"org.jboss.interceptor.reader.ReflectiveClassMetadata", //
				"org.jboss.interceptor.reader.SimpleInterceptorMetadata", //
				"org.jboss.interceptor.spi.instance.InterceptorInstantiator", //
				"org.jboss.interceptor.spi.metadata.InterceptorReference", //
				"org.jboss.interceptor.spi.metadata.MethodMetadata", //
				"org.jboss.interceptor.spi.model.InterceptionModel", //
				"org.jboss.interceptor.spi.model.InterceptionType", //
				"org.jboss.weld.interceptor.builder.InterceptionModelBuilder", //
				"org.jboss.weld.interceptor.builder.MethodReference", //
				"org.jboss.weld.interceptor.proxy.DefaultInvocationContextFactory", //
				"org.jboss.weld.interceptor.proxy.InterceptorMethodHandler", //
				"org.jboss.weld.interceptor.reader.ClassMetadataInterceptorReference", //
				"org.jboss.weld.interceptor.reader.DefaultMethodMetadata", //
				"org.jboss.weld.interceptor.reader.ReflectiveClassMetadata", //
				"org.jboss.weld.interceptor.reader.SimpleInterceptorMetadata", //
				"org.jboss.weld.interceptor.spi.instance.InterceptorInstantiator", //
				"org.jboss.weld.interceptor.spi.metadata.InterceptorReference", //
				"org.jboss.weld.interceptor.spi.metadata.MethodMetadata", //
				"org.jboss.weld.interceptor.spi.model.InterceptionModel", //
				"org.jboss.weld.interceptor.spi.model.InterceptionType", //
				// ysoserial's JRMPClient payload
				"java.rmi.registry.Registry", //
				"java.rmi.server.ObjID", //
				"java.rmi.server.RemoteObjectInvocationHandler", //
				// ysoserial's JSON1 payload
				"net.sf.json.JSONObject", //
				// ysoserial's Jdk7u21 payload
				"javax.xml.transform.Templates", //
				// ysoserial's Jython1 payload
				"org.python.core.PyObject", //
				"org.python.core.PyBytecode", //
				"org.python.core.PyFunction", //
				// ysoserial's MozillaRhino1 payload
				"org.mozilla.javascript..*", //
				// ysoserial's Myfaces1,2 payload
				"org.apache.myfaces.context.servlet.FacesContextImpl", //
				"org.apache.myfaces.context.servlet.FacesContextImplBase", //
				"org.apache.myfaces.el.CompositeELResolver", //
				"org.apache.myfaces.el.unified.FacesELContext", //
				"org.apache.myfaces.view.facelets.el.ValueExpressionMethodExpression", //
				// ysoserial's ROME payload
				"com.sun.syndication.feed.impl.ObjectBean", //
				// ysoserial's Spring1,2 payload
				"org.springframework.beans.factory.ObjectFactory", //
				"org.springframework.core.SerializableTypeWrapper$MethodInvokeTypeProvider", //
				"org.springframework.aop.framework.AdvisedSupport", //
				"org.springframework.aop.target.SingletonTargetSource", //
				"org.springframework.aop.framework.JdkDynamicAopProxy", //
				"org.springframework.core.SerializableTypeWrapper$TypeProvider", //
				// other trigger gadgets or payloads
				"java.util.PriorityQueue", //
				"java.lang.reflect.Proxy", //
				"javax.management.MBeanServerInvocationHandler", //
				"javax.management.openmbean.CompositeDataInvocationHandler", //
				"org.springframework.aop.framework.JdkDynamicAopProxy", //
				"java.beans.EventHandler", //
				"java.util.Comparator", //
				"org.reflections.Reflections"));
		// Performance optimization: remove blacklist items that are already disallowed
		// by the whitelist anyway. Since the whitelist is already quite restrictive,
		// we can avoid most explicit blacklist checks this way.
		blacklistedNames.removeIf(FilteredObjectStreams::packageNameIsNotAllowed);
		return CollectionUtils.asUnmodifiable(blacklistedNames);
	}

}
