/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.conqat.lib.commons.io;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;

import org.conqat.lib.commons.filesystem.FileSystemUtils;

/**
 * {@link AutoCloseable} wrapper around
 * {@link Files#createTempDirectory(String, java.nio.file.attribute.FileAttribute...)}
 * that automatically creates and deletes the temp directory when closed or on
 * shutdown.
 *
 * The directory will always be deleted on auto-close if the object is created
 * via a constructor. This can easily be used with try-resource blocks and it is
 * ensured that the directory is deleted after leaving the try block:
 *
 * <code>
 *     try (SelfDeletingTempDirectory tmp = new SelfDeletingTempDirectory("foo") { ...}
 * </code>
 *
 * If a directory is needed for the whole JVM life-cylce, use one of the
 * <code>SelfDeletingTempDirectory.createAndDeleteOnShutdown</code> factory
 * methods instead.
 */
public class SelfDeletingTempDirectory extends File implements AutoCloseable {
	private static final long serialVersionUID = 1L;

	/**
	 * Whether to keep the directory till JVM shutdown instead of deletion on
	 * auto-close.
	 */
	private final boolean keepUntilShutdown;

	/**
	 * Creates a new temp directory using the given prefix to determine its name.
	 * The directory will be deleted on auto-close.
	 */
	public SelfDeletingTempDirectory(String prefix) throws IOException {
		this(prefix, false);
	}

	/**
	 * Creates a new temp directory using the given prefix to determine its name.
	 */
	private SelfDeletingTempDirectory(String prefix, boolean keepUntilShutdown) throws IOException {
		super(Files.createTempDirectory(prefix).toUri());
		this.keepUntilShutdown = keepUntilShutdown;
		if (keepUntilShutdown) {
			deleteOnShutdown();
		}
	}

	/**
	 * Creates a new temp directory using the given prefix and deletes the directory
	 * on JVM shutdown.
	 */
	public static SelfDeletingTempDirectory createAndDeleteOnShutdown(String prefix) throws IOException {
		return new SelfDeletingTempDirectory(prefix, true);
	}

	/**
	 * Recursively deletes the temp directory. In case that fails (e.g. if some
	 * files are still being used), registers the directory for deletion on JVM
	 * shutdown.
	 */
	@Override
	public void close() {
		if (keepUntilShutdown) {
			return;
		}

		FileSystemUtils.deleteRecursively(this);
		if (exists()) {
			deleteOnShutdown();
		}
	}

	private void deleteOnShutdown() {
		Runtime.getRuntime().addShutdownHook(new Thread(() -> FileSystemUtils.deleteRecursively(this)));
	}
}
