/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.conqat.lib.commons.io;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

import org.conqat.lib.commons.assertion.CCSMAssert;
import org.conqat.lib.commons.filesystem.FileSystemUtils;

/**
 * Utility methods for serialization.
 */
public class SerializationUtils {

	/** Serializes an object to byte array */
	public static byte[] serializeToByteArray(Serializable object) throws IOException {
		ByteArrayOutputStream outBuffer = new ByteArrayOutputStream();
		ObjectOutputStream out = FilteredObjectStreams.filteredOutputStream(outBuffer);
		out.writeObject(object);
		// no need to put this in finally, as we do not block file handles.
		// Let the GC do the work
		out.close();
		return outBuffer.toByteArray();
	}

	/**
	 * <b>Make use of StorageUtils#deserialize(byte[]) whenever possible. Only use
	 * this method as last resort.</b> <br>
	 * <br>
	 * Deserializes an object from byte array using a custom/given class loader.
	 * <b>Important:</b>The caller has to ensure that the supplied class loader
	 * "knows" all classes to be deserialized. Returns <code>null</code> if a
	 * <code>null</code> value is passed.
	 *
	 */
	public static Serializable deserializeFromByteArray(byte[] bytes) throws IOException, ClassNotFoundException {
		if (bytes == null) {
			return null;
		}

		ObjectInputStream in = FilteredObjectStreams.filteredInputStream(new ByteArrayInputStream(bytes));

		try {
			return (Serializable) in.readObject();
		} finally {
			FileSystemUtils.close(in);
		}
	}

	/**
	 * Returns a copy of the given object obtained by serialization and
	 * deserialization in memory.
	 *
	 */
	@SuppressWarnings("unchecked")
	public static <T extends Serializable> T cloneBySerialization(T t) {
		try {
			return (T) deserializeFromByteArray(serializeToByteArray(t));
		} catch (IOException e) {
			CCSMAssert.fail("This should be impossible as we are working in memory!");
			return null;
		} catch (ClassNotFoundException e) {
			CCSMAssert.fail("This should be impossible as we just had the object available!");
			return null;
		}
	}

	/**
	 * Inserts an int value to the given position in the byte array. The storage
	 * will require 4 bytes in big endian byte order.
	 */
	public static void insertInt(int i, byte[] bytes, int position) {
		bytes[position++] = (byte) (i >> 24 & 0xff);
		bytes[position++] = (byte) (i >> 16 & 0xff);
		bytes[position++] = (byte) (i >> 8 & 0xff);
		bytes[position] = (byte) (i & 0xff);
	}

	/**
	 * Extracts an int value from the given array position (4 bytes in big endian).
	 * This is the counter part to {@link #insertInt(int, byte[], int)} .
	 */
	public static int extractInt(byte[] bytes, int position) {
		int result = bytes[position++] & 0xff;
		result <<= 8;
		result |= bytes[position++] & 0xff;
		result <<= 8;
		result |= bytes[position++] & 0xff;
		result <<= 8;
		result |= bytes[position++] & 0xff;
		return result;
	}

}