/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.conqat.lib.commons.serialization.utils;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;

import org.conqat.lib.commons.assertion.CCSMAssert;
import org.conqat.lib.commons.io.ByteArrayUtils;
import org.conqat.lib.commons.io.SerializationUtils;
import org.conqat.lib.commons.serialization.SerializedEntityBase;
import org.conqat.lib.commons.serialization.SerializedEntityPool;
import org.conqat.lib.commons.serialization.SerializedEntitySerializer;
import org.conqat.lib.commons.serialization.classes.SerializedClass;
import org.conqat.lib.commons.serialization.classes.SerializedIntField;
import org.conqat.lib.commons.serialization.objects.SerializedClassValues;
import org.conqat.lib.commons.serialization.objects.SerializedObject;
import org.conqat.lib.commons.serialization.objects.SerializedStringObject;

/**
 * Utility methods for migrating serialized objects.
 */
public class SerializationMigrationUtils {

	/** Transforms all strings in a (de)serialized string set. */
	public static void transformStringsInHashSet(SerializedObject setObject, IStringTransformation transformation,
			SerializedEntityPool pool) throws IOException {

		SerializedClass setClass = findClassInHierarchy(setObject, HashSet.class);

		SerializedClassValues fieldSets = setObject.getFieldSet(setObject.getPlainClassHierarchy().indexOf(setClass));

		for (Object data : fieldSets.getPreFieldData()) {
			if (!(data instanceof Integer)) {
				continue;
			}

			SerializedEntityBase containedElement = pool.getEntity((Integer) data, SerializedEntityBase.class);
			if (containedElement instanceof SerializedStringObject) {
				SerializedStringObject stringContent = (SerializedStringObject) containedElement;
				stringContent.setValue(transformation.transform(stringContent.getValue()));
			}
		}
	}

	/** Returns all objects in a (de)serialized {@link ArrayList}. */
	@SuppressWarnings("unchecked")
	public static <T extends SerializedEntityBase> List<T> getObjectsFromArrayList(SerializedObject listObject,
			SerializedEntityPool pool) throws IOException {

		SerializedClass listClass = findClassInHierarchy(listObject, ArrayList.class);

		SerializedClassValues fieldSets = listObject
				.getFieldSet(listObject.getPlainClassHierarchy().indexOf(listClass));
		List<T> objects = new ArrayList<>();
		for (Object data : fieldSets.getPostFieldData()) {
			if (!(data instanceof Integer)) {
				continue;
			}
			SerializedEntityBase containedElement = pool.getEntity((Integer) data, SerializedEntityBase.class);
			objects.add((T) containedElement);
		}
		return objects;
	}

	/**
	 * Tries to find a given class in the {@link SerializedObject}'s classHierarchy.
	 * Fails with an assertion error if the given class is not in the class
	 * hierarchy of the given serialized object.
	 *
	 * @param setObject
	 *            The object that defines the class hierarchy.
	 * @param classToFind
	 *            A class in the class hierarchy of the runtime type of #setObject.
	 *
	 */
	public static SerializedClass findClassInHierarchy(SerializedObject setObject, Class<?> classToFind)
			throws IOException {
		for (SerializedClass classObject : setObject.getPlainClassHierarchy()) {
			if (classToFind.getName().equals(classObject.getName())) {
				return classObject;
			}
		}
		CCSMAssert.fail("Only works on " + classToFind.getName() + " objects!");
		return null;
	}

	/**
	 * Renames the given entity class, reads it from the pool and returns it. The
	 * given class must be a root entity in the pool.
	 */
	@SuppressWarnings("unchecked")
	public static <T> T serializeRootEntities(SerializedEntityPool entityPool)
			throws AssertionError, IOException, ClassNotFoundException {
		List<SerializedEntityBase> rootEntities = entityPool.getRootEntities();
		byte[] newValue = SerializedEntitySerializer.serializeToBytes(rootEntities);
		return (T) SerializationUtils.deserializeFromByteArray(newValue);
	}

	/** Interface for transforming strings. */
	public interface IStringTransformation {

		/** Transforms a single string. */
		String transform(String input);
	}

	/**
	 * Creates a SerializedObject for an ArrayList with the given objects as
	 * entries.
	 */
	public static SerializedObject createSerializedArrayListObject(List<SerializedEntityBase> entries,
			SerializedEntityPool entityPool) throws IOException {
		int arrayListClassHandle = getArrayListClassHandle(entityPool);
		SerializedClassValues classValues = new SerializedClassValues(1);

		classValues.appendData(ByteArrayUtils.intToByteArray(entries.size()), false);
		for (SerializedEntityBase entry : entries) {
			classValues.appendData(entry.getHandle(), false);
		}

		SerializedObject serializedObject = new SerializedObject(arrayListClassHandle,
				Collections.singletonList(classValues), entityPool);

		serializedObject.setFieldValue("size", entries.size());

		return serializedObject;
	}

	private static int getArrayListClassHandle(SerializedEntityPool entityPool) {
		SerializedClass arrayListClass = entityPool.findClass(ArrayList.class.getName());

		if (arrayListClass == null) {
			arrayListClass = new SerializedClass(ArrayList.class.getName(), 8683452581122892189L, (byte) 3, entityPool);
			arrayListClass.addField(new SerializedIntField("size"));
		}
		return arrayListClass.getHandle();
	}
}
