/*
 * Copyright (c) CQSE GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teamscale.ide.eclipse.mappings.handlers;

import java.util.Optional;

import javax.inject.Inject;
import javax.inject.Named;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.eclipse.core.resources.IResource;
import org.eclipse.e4.core.di.annotations.CanExecute;
import org.eclipse.e4.core.di.annotations.Execute;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.ISources;
import org.eclipse.ui.ide.ResourceUtil;
import org.eclipse.ui.texteditor.ITextEditor;

import com.teamscale.ide.commons.findings.LineRange;
import com.teamscale.ide.commons.mappings.tasks.OpenResourceInBrowserTask;
import com.teamscale.ide.commons.model.ResourceReference;
import com.teamscale.ide.commons.plugin.IServiceRegistry;
import com.teamscale.ide.commons.plugin.services.ITaskExecutor;
import com.teamscale.ide.eclipse.model.ResourceReferences;

/**
 * A handler that, given the active editor (which may be
 * {@linkplain org.eclipse.ui.part.MultiPageEditorPart multi-page}), opens the
 * configured default browser to view the resource currently being displayed on
 * the Teamscale server. If possible, the current selection is highlighted.
 * <p>
 * When "editing text", the associated command is bound to Ctrl+Shift+F2 (or
 * Cmd+Shift+F2 on macOS), alluding to the key binding of the "Open Attached
 * Javadoc (in Browser)" command.
 */
public class OpenEditorContentsInBrowserHandler {

	private static final Logger LOGGER = LogManager.getLogger();

	private final IServiceRegistry serviceRegistry;
	private final ITaskExecutor taskExecutor;

	@Inject
	public OpenEditorContentsInBrowserHandler(IServiceRegistry serviceRegistry) {
		this.serviceRegistry = serviceRegistry;
		taskExecutor = serviceRegistry.getService(ITaskExecutor.class);
	}

	@Execute
	public void execute(@Named(ISources.ACTIVE_EDITOR_NAME) IEditorPart editor) {
		LOGGER.traceEntry(null, editor);
		editor = getUnderlyingEditor(editor);

		IResource resource = getEditorInputResource(editor).orElse(null);
		if (resource == null) {
			return;
		}
		ResourceReference reference = ResourceReferences.fromIResource(resource).orElse(null);
		
		if (reference == null) {
			return;
		}

		LineRange lineRange = getTextSelection(editor).map(OpenEditorContentsInBrowserHandler::toLineRange)
				.orElse(LineRange.unspecified());

		taskExecutor.execute(new OpenResourceInBrowserTask(serviceRegistry, reference, lineRange, false));

		LOGGER.traceExit();
	}

	@CanExecute
	public static boolean canExecute(@Named(ISources.ACTIVE_EDITOR_NAME) IEditorPart editor) {
		LOGGER.traceEntry(null, editor);
		editor = getUnderlyingEditor(editor);

		boolean canExecute = getEditorInputResource(editor).isPresent();

		return LOGGER.traceExit(canExecute);
	}

	/**
	 * Returns the "underlying" editor of the given input editor. In case the input
	 * is a {@link org.eclipse.ui.part.MultiPageEditorPart}, this returns the editor
	 * from the currently active editor page. The implementation exploits the
	 * fallback logic of
	 * {@link org.eclipse.ui.part.MultiPageEditorPart#getAdapter(Class)} and the
	 * fact that a multi-page editor itself is never a text editor.
	 */
	private static IEditorPart getUnderlyingEditor(IEditorPart editor) {
		ITextEditor textEditor = editor.getAdapter(ITextEditor.class);
		return textEditor == null ? editor : textEditor;
	}

	private static Optional<IResource> getEditorInputResource(IEditorPart editor) {
		return Optional.ofNullable(ResourceUtil.getResource(editor.getEditorInput()));
	}

	private static Optional<ITextSelection> getTextSelection(IEditorPart editor) {
		return getSelectionProvider(editor).map(ISelectionProvider::getSelection) //
				.filter(ITextSelection.class::isInstance).map(ITextSelection.class::cast);
	}

	private static Optional<ISelectionProvider> getSelectionProvider(IEditorPart editor) {
		if (editor instanceof ITextEditor) {
			return Optional.of((ITextEditor) editor).map(ITextEditor::getSelectionProvider);
		}
		return Optional.ofNullable(editor.getEditorSite()).map(IEditorSite::getSelectionProvider);
	}

	private static LineRange toLineRange(ITextSelection textSelection) {
		// Convert zero-based line numbers (Eclipse) to one-based (Teamscale)
		int firstLine = textSelection.getStartLine() + 1;
		int lastLine = textSelection.getEndLine() + 1;
		return LineRange.specified(firstLine, lastLine);
	}
}
